
/**MOD+***********************************************************************/
/* Module:    crtmain.c                                                      */
/*                                                                           */
/* Purpose:   Certificate Revocation Tool                                    */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.9                $ $Modtime::   Jan 11 2001 15:56:16   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#define CRT_DEFINE_VARS
#include <crtincl.h>
#undef CRT_DEFINE_VARS

/**PROC+**********************************************************************/
/* Name:      main                                                           */
/*                                                                           */
/* Purpose:   Main entry point for CRT                                       */
/*                                                                           */
/* Returns:   0           - success                                          */
/*            other       - failure                                          */
/*                                                                           */
/* Params:    Standard main() parameters                                     */
/*                                                                           */
/**PROC-**********************************************************************/

int main(int argc, char *argv[])
{
  int opt_char;
  char *config_file;
  char *subject_name;
  long crldays;
  long crlhours;
  long crlmins;
  LHASH *config;
  long error_line;
  char *host;
  char *username;
  char *password;
  char *dn_suffix;
  char *ca_name;
  char *identity;
  char *pkey_file_name;
  FILE *pkey_file;
  EVP_PKEY *private_key;
  LDAP *directory;
  X509_OBJECT x509_obj;
  X509_CRL *crl;
  char *digest_name;
  int digest_nid;
  EVP_MD *crl_digest;
  char temp_buf[200];
  int error;
  int ii;
  int op_type;
  int forceupdate;

  /***************************************************************************/
  /* Initialize local variables.                                             */
  /***************************************************************************/
  config_file = NULL;
  subject_name = NULL;
  crldays = 0;
  crlhours = 0;
  crlmins = 0;
  config = NULL;
  host = NULL;
  username = NULL;
  password = NULL;
  dn_suffix = NULL;
  ca_name = NULL;
  identity = NULL;
  pkey_file_name = NULL;
  pkey_file = NULL;
  private_key = NULL;
  directory = NULL;
  crl = NULL;
  digest_name = NULL;
  crl_digest = NULL;
  op_type = CRT_OP_ERROR;
  forceupdate = 0;

  /***************************************************************************/
  /* Initialize OpenSSL stuff.                                               */
  /***************************************************************************/
  OpenSSL_add_all_algorithms();
  ERR_load_crypto_strings();

  /***************************************************************************/
  /* Read options.                                                           */
  /***************************************************************************/
  opterr = 0;
  while ((opt_char = getopt(argc, argv, "c:n:d:h:m:u")) != EOF)
  {
    switch (opt_char)
    {
      case 'c':
        config_file = optarg;
        break;

      case 'n':
        subject_name = optarg;
        break;

      case 'd':
        crldays = atol(optarg);
        break;

      case 'h':
        crlhours = atol(optarg);
        break;

      case 'm':
        crlmins = atol(optarg);
        break;

      case 'u':
        forceupdate = 1;
        break;

      default:
        crt_print_usage(argv[0]);
        exit(-1);
    }
  }

  /***************************************************************************/
  /* Deal with parameter errors and set defaults.                            */
  /***************************************************************************/
  if ((config_file == NULL) ||
      ((subject_name != NULL) && (forceupdate)))
  {
    crt_print_usage(argv[0]);
    exit(-1);
  }

  if ((crldays == 0) && (crlhours == 0) && (crlmins == 0))
  {
    crldays = 30;
  }

  printf("\n");

  /***************************************************************************/
  /* Open config file and read values.                                       */
  /***************************************************************************/
  config = CONF_load(NULL, config_file, &error_line);
  if (config == NULL)
  {
    printf("Error in config file '%s' on line %ld\n", config_file, error_line);
    exit(-1);
  }

  host = CONF_get_string(config, "openosp", "ca_directory");
  if (host == NULL)
  {
    printf("Couldn't find ca_directory in the config file\n");
    exit(-1);
  }

  username = CONF_get_string(config, "openosp", "ca_directory_user");
  if (username == NULL)
  {
    username = "";
  }

  password = CONF_get_string(config, "openosp", "ca_directory_password");
  if (password == NULL)
  {
    password = "";
  }

  dn_suffix = CONF_get_string(config, "openosp", "ca_directory_dn_suffix");
  if (dn_suffix == NULL)
  {
    dn_suffix = "";
  }

  ca_name = CONF_get_string(config, "openosp", "ca_identity");
  if (ca_name == NULL)
  {
    printf("Couldn't find ca_identity in the config file\n");
    exit(-1);
  }

  /***************************************************************************/
  /* Open the LDAP directory.                                                */
  /***************************************************************************/
  directory = ldap_init(host, LDAP_PORT);
  if (directory == NULL)
  {
    printf("Couldn't initialize LDAP access to host '%s'\n", host);
    exit(-1);
  }

  error = ldap_simple_bind_s(directory, username, password);
  if (error != LDAP_SUCCESS)
  {
    printf("Failed to connect to LDAP host '%s'\n", host);
    exit(-1);
  }

  /***************************************************************************/
  /* Look up the current CRL.                                                */
  /***************************************************************************/
  printf("Looking up current CRL for CA '%s'...\n", ca_name);
  error = crt_ldap_lookup(directory, ca_name, dn_suffix, &x509_obj,
                                                               CRT_LOOKUP_CRL);
  if (error == 0)
  {
    printf("Couldn't get current CRL for CA\n");
    exit(-1);
  }

  if (error < 0)
  {
    /*************************************************************************/
    /* The CA doesn't have a CRL. Offer to generate one.                     */
    /*************************************************************************/
    printf("\nCA does not have a CRL in the directory.\n");
    printf("Do you want to create a new, empty CRL?");
    if (!crt_yes_no('y'))
    {
      printf("No changes made\n");
      goto EXIT_LABEL;
    }
    op_type = CRT_OP_CREATE_NEW;
  }
  else
  {
    /*************************************************************************/
    /* There's a CRL.                                                        */
    /*************************************************************************/
    crl = x509_obj.data.crl;

    /*************************************************************************/
    /* See what operation we're supposed to be doing.                        */
    /*************************************************************************/
    if (subject_name != NULL)
    {
      op_type = CRT_OP_REVOKE_CERT;
    }
    else if (forceupdate)
    {
      op_type = CRT_OP_UPDATE;
    }
    else
    {
      op_type = CRT_OP_DISPLAY;
    }

    /*************************************************************************/
    /* Verify the signature on the CRL.                                      */
    /*************************************************************************/
    printf("Looking up CA certificate to verify CRL...\n", ca_name);
    error = crt_ldap_lookup(directory, ca_name, dn_suffix, &x509_obj,
                                                            CRT_LOOKUP_CACERT);
    if (error <= 0)
    {
      printf("Couldn't get CA's certificate to verify CRL\n");
      exit(-1);
    }

    printf("Verifying CRL...\n", ca_name);
    error = X509_CRL_verify(crl, X509_get_pubkey(x509_obj.data.x509));
    X509_free(x509_obj.data.x509);
    if (error <= 0)
    {
      printf("The CRL's signature is wrong: it may be corrupt\n");
      printf("Proceed anyway?");
      if (!crt_yes_no('n'))
      {
        if (op_type != CRT_OP_DISPLAY)
        {
          printf("No changes made\n");
        }
        goto EXIT_LABEL;
      }
    }
    else
    {
      printf("CRL signature is OK\n");
    }
  }

  if (op_type == CRT_OP_DISPLAY)
  {
    /*************************************************************************/
    /* We're not changing the CRL in any way; just print out the details and */
    /* exit gracefully.                                                      */
    /*************************************************************************/
    X509_CRL_print_fp(stdout, crl);
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* At this point we can either be creating a new CRL, revoking a           */
  /* certificate or just updating the existing CRL.  First, find the CA      */
  /* identity in the ID list so that we can get the private key filename.    */
  /***************************************************************************/
  printf("\n");
  for (ii = 0; ; ii++)
  {
    sprintf(temp_buf, "identity_%d_dn", ii);
    identity = CONF_get_string(config, "identities", temp_buf);
    if (identity == NULL)
    {
      printf("Couldn't find ca_identity value in identity list\n");
      exit(-1);
    }

    if (strcmp(identity, ca_name) == 0)
    {
      break;
    }
  }

  sprintf(temp_buf, "identity_%d_privkey", ii);
  pkey_file_name = CONF_get_string(config, "identities", temp_buf);
  if (pkey_file_name == NULL)
  {
    printf("No private key configured for CA identity\n");
    exit(-1);
  }

  pkey_file = fopen(pkey_file_name, "r");
  if (pkey_file == NULL)
  {
    printf("Couldn't open private key file '%s'\n", pkey_file_name);
    exit(-1);
  }
  private_key = PEM_read_PrivateKey(pkey_file, NULL, NULL, NULL);
  fclose(pkey_file);

  if (private_key == NULL)
  {
    printf("Couldn't read private key from file '%s'\n", pkey_file_name);
    exit(-1);
  }

  printf("Using CA private key from '%s'\n", pkey_file_name);

  /***************************************************************************/
  /* Decide what signing digest to use.                                      */
  /***************************************************************************/
  if (private_key->type == EVP_PKEY_DSA)
  {
    crl_digest = EVP_dss1();
    digest_nid = NID_sha1;
  }
  else
  {
    digest_name = CONF_get_string(config, "openosp", "ca_rsa_signing_digest");
    if (digest_name == NULL)
    {
      digest_nid = NID_md5;
    }
    else
    {
      digest_nid = OBJ_txt2nid(digest_name);
    }

    if (digest_nid == NID_md5)
    {
      crl_digest = EVP_md5();
    }
    else if (digest_nid == NID_sha1)
    {
      crl_digest = EVP_sha1();
    }
    else
    {
      printf("Unrecognized RSA signing digest '%s' in config file\n",
                                                                  digest_name);
      exit(-1);
    }
  }

  /***************************************************************************/
  /* Check the operation we're meant to be doing.                            */
  /***************************************************************************/
  switch (op_type)
  {
    case CRT_OP_CREATE_NEW:
      crl = crt_create_new_crl(directory, ca_name, dn_suffix);
      if (crl == NULL)
      {
        goto EXIT_LABEL;
      }
      break;

    case CRT_OP_REVOKE_CERT:
      if (!crt_revoke_cert(directory, subject_name, dn_suffix, crl))
      {
        goto EXIT_LABEL;
      }
      break;

    case CRT_OP_UPDATE:
      /***********************************************************************/
      /* We already have the existing CRL to hand, and the code below will   */
      /* update its validity dates, so there's nothing to do here.           */
      /***********************************************************************/
      break;

    default:
      printf("Internal error: can't determine operation type\n");
      exit(-1);
      break;
  }

  /***************************************************************************/
  /* We now have a CRL that just needs to have its validity dates set and to */
  /* be signed.                                                              */
  /***************************************************************************/
  printf("The new CRL will expire in %ld days, %ld hours and %ld minutes\n",
                                                  crldays, crlhours, crlmins);
  printf("The signature digest will be %s\n", OBJ_nid2sn(digest_nid));
  printf("The changes are about to be written to the directory. OK?");
  if (!crt_yes_no('n'))
  {
    printf("No changes made\n");
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the lastUpdate and nextUpdate fields and sign it.                   */
  /***************************************************************************/
  X509_gmtime_adj(crl->crl->lastUpdate, 0);

  if (crl->crl->nextUpdate == NULL)
  {
    crl->crl->nextUpdate = ASN1_UTCTIME_new();
    if (crl->crl->nextUpdate == NULL)
    {
      printf("Failed to set nextUpdate field in CRL\n");
      exit(-1);
    }
  }
  X509_gmtime_adj(crl->crl->nextUpdate,
                         ((((crldays * 24) + crlhours) * 60) + crlmins) * 60);

  error = X509_CRL_sign(crl, private_key, crl_digest);
  if (error == 0)
  {
    printf("Failed to sign new CRL\n");
    ERR_print_errors_fp(stdout);
    exit(-1);
  }

  /***************************************************************************/
  /* Put the new CRL in the directory.                                       */
  /***************************************************************************/
  printf("Storing CRL in directory...\n");
  if (!crt_replace_crl(directory, ca_name, dn_suffix, crl))
  {
    printf("Failed to store CRL in directory\n");
    if (op_type == CRT_OP_REVOKE_CERT)
    {
      printf("\nCertificate has NOT been revoked\n\n");
    }
    exit(-1);
  }

  if (op_type == CRT_OP_REVOKE_CERT)
  {
    /*************************************************************************/
    /* Delete the revoked certificate from the directory.                    */
    /*************************************************************************/
    printf("Deleting revoked certificate...\n");
    if (!crt_delete_user_cert(directory, subject_name, dn_suffix))
    {
      printf("Failed to delete revoked certificate from the directory\n");
      printf("However, the CRL has been successfully updated\n");
    }
    printf("\nSuccessfully revoked certificate for '%s'\n\n", subject_name);
  }
  else
  {
    printf("\nSuccessfully created new CRL for '%s'\n\n", ca_name);
    if (subject_name != NULL)
    {
      printf("NOTE: You must run this tool again with the same parameters\n");
      printf("in order to revoke the certificate you specified\n\n");
    }
  }

EXIT_LABEL:

  if (crl != NULL)
  {
    X509_CRL_free(crl);
  }

  if (private_key != NULL)
  {
    EVP_PKEY_free(private_key);
  }

  ldap_unbind(directory);
  CONF_free(config);
  ERR_free_strings();

  return(0);

} /* main */

/**PROC+**********************************************************************/
/* Name:      crt_ldap_lookup                                                */
/*                                                                           */
/* Purpose:   Look up a certificate or CRL in the LDAP directory             */
/*                                                                           */
/* Returns:   1           - success                                          */
/*            -1          - found object, but it had no cert/CRL             */
/*            0           - failure                                          */
/*                                                                           */
/* Params:    IN     directory      - the LDAP directory handle              */
/*            IN     name           - the distinguished name to look up      */
/*            IN     dn_suffix      - the suffix to add for the lookup       */
/*            IN     x509_obj       - the X509_OBJECT to fill in with the    */
/*                                      result                               */
/*            IN     type           - the type of object being requested     */
/*                                                                           */
/**PROC-**********************************************************************/

int crt_ldap_lookup(LDAP *directory,
                    char *name,
                    char *dn_suffix,
                    X509_OBJECT *x509_obj,
                    int type)
{
  int rc = 0;
  int error;
  int ii;
  char *object_name = NULL;
  char *lookup_name = NULL;
  char *attrs[2] = {NULL, NULL};
  char *filter;
  LDAPMessage *search_result = NULL;
  LDAPMessage *result_entry = NULL;
  struct berval **attrib_values = NULL;
  unsigned char *data = NULL;
  long data_len;

  /***************************************************************************/
  /* Initialize.                                                             */
  /***************************************************************************/
  x509_obj->data.ptr = NULL;
  lookup_name = malloc(strlen(name) + 1 + strlen(dn_suffix) + 1);
  strcpy(lookup_name, name);
  if ((dn_suffix != NULL) && (strlen(dn_suffix) > 0))
  {
    strcat(lookup_name, ",");
    strcat(lookup_name, dn_suffix);
  }

  /***************************************************************************/
  /* Do the search.                                                          */
  /***************************************************************************/
  switch (type)
  {
    case CRT_LOOKUP_USERCERT:
      attrs[0] = CRT_LDAP_ATTR_USERCERT;
      filter = CRT_LDAP_FILTER_SAUSER;
      object_name = "user certificate";
      x509_obj->type = X509_LU_X509;
      break;

    case CRT_LOOKUP_CACERT:
      attrs[0] = CRT_LDAP_ATTR_CACERT;
      filter = CRT_LDAP_FILTER_CA;
      object_name = "CA certificate";
      x509_obj->type = X509_LU_X509;
      break;

    case CRT_LOOKUP_CRL:
      attrs[0] = CRT_LDAP_ATTR_CRL;
      filter = CRT_LDAP_FILTER_CA;
      object_name = "CRL";
      x509_obj->type = X509_LU_CRL;
      break;

    default:
      printf("Internal error: unknown lookup type\n");
      exit(-1);
      break;
  }

  error = ldap_search_s(directory,
                        lookup_name,
                        LDAP_SCOPE_BASE,
                        filter,
                        attrs,
                        0,
                        &search_result);
  if (error != LDAP_SUCCESS)
  {
    printf("LDAP search error %d: %s\n", error, ldap_err2string(error));
    if (error == LDAP_NO_SUCH_ATTRIBUTE)
    {
      rc = -1;
    }
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* From now on any failure is interpreted as 'cert/CRL not found'.         */
  /***************************************************************************/
  rc = -1;

  if (search_result == NULL)
  {
    printf("LDAP search returned no results\n");
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Extract the result.                                                     */
  /***************************************************************************/
  ii = ldap_count_entries(directory, search_result);
  if (ii != 1)
  {
    printf("Found %d directory entries; expecting 1\n", ii);
    goto EXIT_LABEL;
  }

  result_entry = ldap_first_entry(directory, search_result);
  attrib_values = ldap_get_values_len(directory,
                                      result_entry,
                                      attrs[0]);
  if (attrib_values == NULL)
  {
    printf("LDAP search returned no attribute values\n");
    goto EXIT_LABEL;
  }

  ii = ldap_count_values_len(attrib_values);
  if (ii != 1)
  {
    printf("Got %d %ss; expecting 1\n", ii, object_name);
  }

  data = (unsigned char *)attrib_values[0]->bv_val;
  data_len = (long)attrib_values[0]->bv_len;

  /***************************************************************************/
  /* Convert result.                                                         */
  /***************************************************************************/
  if (type == CRT_LOOKUP_CRL)
  {
    x509_obj->data.crl = d2i_X509_CRL(NULL, &data, data_len);
    if (x509_obj->data.crl == NULL)
    {
      printf("CRL is corrupt\n");
      goto EXIT_LABEL;
    }
  }
  else
  {
    x509_obj->data.x509 = d2i_X509(NULL, &data, data_len);
    if (x509_obj->data.x509 == NULL)
    {
      printf("Certificate is corrupt\n");
      goto EXIT_LABEL;
    }
  }

  rc = 1;

EXIT_LABEL:

  if (lookup_name != NULL)
  {
    free(lookup_name);
  }

  return(rc);

} /* crt_ldap_lookup */

/**PROC+**********************************************************************/
/* Name:      crt_create_new_crl                                             */
/*                                                                           */
/* Purpose:   Create a new, empty CRL                                        */
/*                                                                           */
/* Returns:   X509_CRL    - success                                          */
/*            NULL        - failure                                          */
/*                                                                           */
/* Params:    IN     directory      - the LDAP directory handle              */
/*            IN     ca_name        - the CA's distinguished name            */
/*            IN     dn_suffix      - the suffix to add for the LDAP op      */
/*                                                                           */
/**PROC-**********************************************************************/

X509_CRL *crt_create_new_crl(LDAP *directory,
                             char *ca_name,
                             char *dn_suffix)
{
  X509_CRL *new_crl = NULL;
  X509_OBJECT x509_obj;
  X509 *ca_cert = NULL;
  int error;

  new_crl = X509_CRL_new();
  if (new_crl == NULL)
  {
    printf("Failed to create new CRL\n");
    exit(-1);
  }

  printf("Looking up CA certificate...\n", ca_name);
  error = crt_ldap_lookup(directory, ca_name, dn_suffix, &x509_obj,
                                                            CRT_LOOKUP_CACERT);
  if (error <= 0)
  {
    printf("Couldn't find CA certificate\n");
    exit(-1);
  }

  ca_cert = x509_obj.data.x509;

  new_crl->crl->issuer = X509_NAME_dup(X509_get_subject_name(ca_cert));
  if (new_crl->crl->issuer == NULL)
  {
    printf("Failed to set new CRL's issuer name\n");
    exit(-1);
  }

EXIT_LABEL:

  if (ca_cert != NULL)
  {
    X509_free(ca_cert);
  }

  return(new_crl);

} /* crt_create_new_crl */

/**PROC+**********************************************************************/
/* Name:      crt_revoke_cert                                                */
/*                                                                           */
/* Purpose:   Revoke a certificate by adding its details to the supplied CRL */
/*                                                                           */
/* Returns:   1           - success                                          */
/*            0           - failure                                          */
/*                                                                           */
/* Params:    IN     directory      - the LDAP directory handle              */
/*            IN     subject_name   - the subject name of the certificate    */
/*                                      to be revoked                        */
/*            IN     dn_suffix      - the DN suffix to add for the LDAP op   */
/*            IN     crl            - the CRL to update                      */
/*                                                                           */
/**PROC-**********************************************************************/

int crt_revoke_cert(LDAP *directory,
                    char *subject_name,
                    char *dn_suffix,
                    X509_CRL *crl)
{
  int rc = 0;
  X509_OBJECT x509_obj;
  X509 *revoke_cert = NULL;
  X509_REVOKED *revoked = NULL;
  int error;

  /***************************************************************************/
  /* Look up the certificate to find out its serial number.                  */
  /***************************************************************************/
  printf("Looking up certificate to be revoked for '%s'...\n", subject_name);
  error = crt_ldap_lookup(directory, subject_name, dn_suffix, &x509_obj,
                                                          CRT_LOOKUP_USERCERT);
  if (error <= 0)
  {
    printf("Couldn't find a valid certificate to be revoked\n");
    exit(-1);
  }

  /***************************************************************************/
  /* Print the certificate details and get confirmation.                     */
  /***************************************************************************/
  revoke_cert = x509_obj.data.x509;
  X509_print_fp(stdout, revoke_cert);
  printf("\nThe above certificate is to be revoked\n");
  printf("Are you sure?");
  if (!crt_yes_no('n'))
  {
    printf("No changes made\n");
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We've got confirmation to go ahead.  Add the certificate to be revoked  */
  /* to the CRL.                                                             */
  /***************************************************************************/
  printf("Updating CRL...\n");
  revoked = X509_REVOKED_new();
  if (revoked == NULL)
  {
    printf("Out of memory\n");
    exit(-1);
  }

  revoked->serialNumber = ASN1_INTEGER_dup(X509_get_serialNumber(revoke_cert));
  X509_gmtime_adj(revoked->revocationDate, 0);

  sk_X509_REVOKED_push(X509_CRL_get_REVOKED(crl), revoked);

  rc = 1;

EXIT_LABEL:

  if (revoke_cert != NULL)
  {
    X509_free(revoke_cert);
  }

  return(rc);

} /* crt_revoke_cert */

/**PROC+**********************************************************************/
/* Name:      crt_replace_crl                                                */
/*                                                                           */
/* Purpose:   Replace a CRL in the directory with the one supplied           */
/*                                                                           */
/* Returns:   1           - success                                          */
/*            0           - failure                                          */
/*                                                                           */
/* Params:    IN     directory      - the LDAP directory handle              */
/*            IN     name           - the CA's distinguished name            */
/*            IN     dn_suffix      - the DN suffix to add for the LDAP op   */
/*            IN     crl            - the new CRL to put in the directory    */
/*                                                                           */
/**PROC-**********************************************************************/

int crt_replace_crl(LDAP *directory,
                    char *name,
                    char *dn_suffix,
                    X509_CRL *crl)
{
  int rc = 0;
  char *lookup_name = NULL;
  unsigned char *mem = NULL;
  char *crl_asn1 = NULL;
  unsigned int crl_asn1_len;
  struct berval crl_value;
  struct berval *crl_values[2] = {NULL, NULL};
  LDAPMod crl_mod;
  LDAPMod *mods[2] = {NULL, NULL};
  int error;

  /***************************************************************************/
  /* Initialize.                                                             */
  /***************************************************************************/
  lookup_name = malloc(strlen(name) + 1 + strlen(dn_suffix) + 1);
  strcpy(lookup_name, name);
  if ((dn_suffix != NULL) && (strlen(dn_suffix) > 0))
  {
    strcat(lookup_name, ",");
    strcat(lookup_name, dn_suffix);
  }

  /***************************************************************************/
  /* Convert CRL to ASN.1 DER-encoded form.                                  */
  /***************************************************************************/
  crl_asn1_len = i2d_X509_CRL(crl, NULL);
  mem = (unsigned char *)malloc(crl_asn1_len);
  if (mem == NULL)
  {
    printf("Out of memory\n");
    goto EXIT_LABEL;
  }
  crl_asn1 = (char *)mem;
  i2d_X509_CRL(crl, &mem);

  /***************************************************************************/
  /* Do the modify operation.                                                */
  /***************************************************************************/
  crl_value.bv_val = crl_asn1;
  crl_value.bv_len = crl_asn1_len;
  crl_values[0] = &crl_value;

  crl_mod.mod_op = LDAP_MOD_REPLACE | LDAP_MOD_BVALUES;
  crl_mod.mod_type = CRT_LDAP_ATTR_CRL;
  crl_mod.mod_vals.modv_bvals = crl_values;
  crl_mod.mod_next = NULL;
  mods[0] = &crl_mod;

  error = ldap_modify_s(directory, lookup_name, mods);
  if (error != LDAP_SUCCESS)
  {
    printf("LDAP modify error %d: %s\n", error, ldap_err2string(error));
    goto EXIT_LABEL;
  }

  rc = 1;

EXIT_LABEL:

  if (crl_asn1 != NULL)
  {
    free(crl_asn1);
  }

  if (lookup_name != NULL)
  {
    free(lookup_name);
  }

  return(rc);

} /* crt_replace_crl */

/**PROC+**********************************************************************/
/* Name:      crt_delete_user_cert                                           */
/*                                                                           */
/* Purpose:   Delete the specified entity's certificate from the directory   */
/*                                                                           */
/* Returns:   1           - success                                          */
/*            0           - failure                                          */
/*                                                                           */
/* Params:    IN     directory      - the LDAP directory handle              */
/*            IN     name           - the entity's distinguished name        */
/*            IN     dn_suffix      - the suffix to add for the lookup       */
/*                                                                           */
/**PROC-**********************************************************************/

int crt_delete_user_cert(LDAP *directory,
                         char *name,
                         char *dn_suffix)
{
  int rc = 0;
  char *lookup_name = NULL;
  LDAPMod cert_mod;
  LDAPMod *mods[2] = {NULL, NULL};
  int error;

  /***************************************************************************/
  /* Initialize.                                                             */
  /***************************************************************************/
  lookup_name = malloc(strlen(name) + 1 + strlen(dn_suffix) + 1);
  strcpy(lookup_name, name);
  if ((dn_suffix != NULL) && (strlen(dn_suffix) > 0))
  {
    strcat(lookup_name, ",");
    strcat(lookup_name, dn_suffix);
  }

  /***************************************************************************/
  /* Do the modify operation.                                                */
  /***************************************************************************/
  cert_mod.mod_op = LDAP_MOD_DELETE | LDAP_MOD_BVALUES;
  cert_mod.mod_type = CRT_LDAP_ATTR_USERCERT;
  cert_mod.mod_values = NULL;
  cert_mod.mod_next = NULL;
  mods[0] = &cert_mod;

  error = ldap_modify_s(directory, lookup_name, mods);
  if (error != LDAP_SUCCESS)
  {
    printf("LDAP modify error %d: %s\n", error, ldap_err2string(error));
    goto EXIT_LABEL;
  }

  rc = 1;

EXIT_LABEL:

  if (lookup_name != NULL)
  {
    free(lookup_name);
  }

  return(rc);

} /* crt_delete_user_cert */

/**PROC+**********************************************************************/
/* Name:      crt_print_usage                                                */
/*                                                                           */
/* Purpose:   Show usage information                                         */
/*                                                                           */
/* Returns:   None                                                           */
/*                                                                           */
/* Params:    IN     cmd_name       - the name of the executable             */
/*                                                                           */
/**PROC-**********************************************************************/

void crt_print_usage(char *cmd_name)
{
  printf("Usage: %s -c <config_file> [-n <subject_name> | -u]\n", cmd_name);
  printf("          [-d <days_to_next_crl>] [-h <hours_to_next_crl>]\n");
  printf("          [-m <minutes_to_next_crl>]\n");
  printf("\n");
  printf("       Specify -n to revoke the certificate for subject_name.\n");
  printf("       If you don't specify -n, information about the CRL\n");
  printf("       is displayed. If there is no CRL, you have the option\n");
  printf("       to generate a new, empty one.\n");
  printf("\n");
  printf("       Use -u to force an existing CRL to be updated with\n");
  printf("       a new expiry date\n");

  return;

} /* crt_print_usage */

/**PROC+**********************************************************************/
/* Name:      crt_yes_no                                                     */
/*                                                                           */
/* Purpose:   Read stdin to determine the user's answer to a yes/no question */
/*                                                                           */
/* Returns:   None                                                           */
/*                                                                           */
/* Params:    IN     default_answer - the default option to use              */
/*                                                                           */
/**PROC-**********************************************************************/

int crt_yes_no(char default_answer)
{
  int rc = 0;
  char temp_buf[10];

  printf(" [%c] ", default_answer);
  fgets(temp_buf, 10, stdin);

  if ((temp_buf[0] == '\r') || (temp_buf[0] == '\n'))
  {
    /*************************************************************************/
    /* Nothing was supplied -- use the default.                              */
    /*************************************************************************/
    temp_buf[0] = default_answer;
  }

  if ((temp_buf[0] == 'y') || (temp_buf[0] == 'Y'))
  {
    rc = 1;
  }

  return(rc);

} /* crt_yes_no */

