
/**MOD+***********************************************************************/
/* Module:    ospd.c                                                         */
/*                                                                           */
/* Purpose:   Implementation of main() for sample OSP daemon                 */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.21               $ $Modtime::   Jan 11 2001 15:56:16   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <xmlparse.h>
#include <openssl/conf.h>
#include <openosp.h>
#include <trcextrn.h>
#include <comextrn.h>
#include <scmextrn.h>
#include <scaextrn.h>
#include <suaextrn.h>
#include <smextrn.h>
#include <aibextrn.h>
#ifdef OSP_DEBUG
#include <rmglbl.h>
#endif

/*****************************************************************************/
/* Prototypes.                                                               */
/*****************************************************************************/

OSP_RC cv_cb(unsigned long   remote_ip,
             void          **ver_cookie,
             char           *cert,
             unsigned int    cert_len,
             unsigned int    chain_len,
             unsigned int    chain_pos);

void cd_cb(OSP_CLIENT_ID *client_id);

/*****************************************************************************/
/* Main entry point.                                                         */
/*****************************************************************************/

int main(int argc, char *argv[])
{
  OSP_RC rc;
  OSP_STATS stats;
  int init_osp;
  int init_scm;
  int init_sca;
  int init_sua;
  sigset_t set;
  char *config_file_name;
  long config_error_line;
  LHASH *config = NULL;
  int terminate_server;

  int processing_count;
  struct timeval processing_time;
  int io_count;
  struct timeval io_time;

  TRC_ENTRY("main");

  printf("ospd starting...\n");

  /***************************************************************************/
  /* Mask all signals - we don't want to handle any of them at this point    */
  /***************************************************************************/
  sigfillset(&set);
  pthread_sigmask(SIG_SETMASK, &set, NULL);

  /***************************************************************************/
  /* Set up state                                                            */
  /***************************************************************************/
  init_osp = FALSE;
  init_scm = FALSE;
  init_sca = FALSE;
  init_sua = FALSE;

  /***************************************************************************/
  /* Init OSP                                                                */
  /***************************************************************************/
  rc = osp_init();
  TRC_DETAIL((TB, "osp_init() rc %d", rc));
  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }
  init_osp = TRUE;

  /***************************************************************************/
  /* Open the config file for the sample application components.  Currently  */
  /* we use the same file as for the stack.                                  */
  /***************************************************************************/
  config_file_name = osp_common_get_config_filename();
  if (config_file_name == NULL)
  {
    TRC_ERROR((TB, "No configuration file name"));
    rc = OSP_ERR_CONFIG;
    goto EXIT_POINT;
  }

  TRC_DETAIL((TB, "Open configuration file: \"%s\"", config_file_name));
  config = CONF_load(NULL, config_file_name, &config_error_line);
  if (config == NULL)
  {
    TRC_ERROR((TB, "Error loading configuration: file=\"%s\" line=%ld",
                                         config_file_name, config_error_line));
    rc = OSP_ERR_CONFIG;
    goto EXIT_POINT;
  }

  /***************************************************************************/
  /* Initialize the sample apps                                              */
  /***************************************************************************/
  rc = scm_init(config);
  TRC_DETAIL((TB, "scm_init() rc %d", rc));
  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }
  init_scm = TRUE;

  rc = sca_init(config);
  TRC_DETAIL((TB, "sca_init() rc %d", rc));
  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }
  init_sca = TRUE;

  rc = sua_init(config);
  TRC_DETAIL((TB, "sua_init() rc %d", rc));
  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }
  init_sua = TRUE;

  /***************************************************************************/
  /* Register callbacks                                                      */
  /***************************************************************************/
  rc = osp_client_verify_register(cv_cb, cd_cb);

  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }

  /***************************************************************************/
  /* Listen                                                                  */
  /***************************************************************************/
  rc = osp_listen();

  TRC_DETAIL((TB, "osp_listen() rc %d", rc));

  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }

  /***************************************************************************/
  /* The stack is up and running -- wait around for connections until        */
  /* someone presses a key.                                                  */
  /***************************************************************************/
  printf("ospd ready\n");
  printf("Command options:\n");
  printf("q - terminate server\n");
  printf("s - connection stats\n");
#ifdef OSP_DEBUG
  printf("m - memory stats\n");
#endif
#ifdef OSP_MEASURE_PERFORMANCE
  printf("p - perf stats\n");
#endif
#ifdef OSP_DATA_COVERAGE
  printf("d - data coverage stats\n");
#endif
  printf("\n");

  terminate_server = FALSE;
  while (!terminate_server)
  {
    switch (getchar())
    {
      case EOF:
      case 'q':
        terminate_server = TRUE;
        break;

#ifdef OSP_MEASURE_PERFORMANCE
      case 'p':
        sm_print_performance_measurements(stdout);
        break;
#endif

#ifdef OSP_DATA_COVERAGE
      case 'd':
        aib_write_data_coverage(stdout);
        break;
#endif

      case 's':
        rc = osp_get_stack_statistics(&stats);
        if (rc == OSP_SUCCESS)
        {
          printf("Stack statistics:\n");
          printf("\tNon-SSL connections:         %u\n", stats.nonssl_conns);
          printf("\tSSL connections:             %u\n", stats.ssl_conns);
          printf("\tOSP requests:                %u\n", stats.osp_requests);
          printf("\tSCEP requests:               %u\n", stats.scep_requests);
        }
        else
        {
          printf("Stack statistics call failed: rc=%d\n", rc);
        }
      break;

#ifdef OSP_DEBUG
      case 'm':
        OSP_GET_LOCK(&rm_count_mutex);
        printf("Memory usage:\n");
        printf("\tGeneral purpose blocks:  %lu\n", count_mem);
        printf("\tThread structures:       %lu\n", count_struct[1]);
        printf("\tConnection structures:   %lu\n", count_struct[2]);
        printf("\tTransaction structures:  %lu\n", count_struct[3]);
        OSP_RELEASE_LOCK(&rm_count_mutex);
        break;
#endif
    }
  }

  rc = osp_get_stack_statistics(&stats);

  TRC_DETAIL((TB, "get stats rc %d", rc));
  TRC_DETAIL((TB, "stats.nonssl_conns  - %u", stats.nonssl_conns ));
  TRC_DETAIL((TB, "stats.osp_requests  - %u", stats.osp_requests ));
  TRC_DETAIL((TB, "stats.scep_requests - %u", stats.scep_requests));
  TRC_DETAIL((TB, "stats.ssl_conns     - %u", stats.ssl_conns    ));

  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }

  rc = osp_client_verify_register(NULL, NULL);

  TRC_DETAIL((TB, "cv reg rc %d", rc));

  if (rc != OSP_SUCCESS)
  {
    goto EXIT_POINT;
  }

EXIT_POINT:

  if (config != NULL)
  {
    /*************************************************************************/
    /* Free the configuration file handle.                                   */
    /*************************************************************************/
    CONF_free(config);
    config = NULL;
  }

  /***************************************************************************/
  /* Terminate OSP                                                           */
  /***************************************************************************/
  printf("ospd stopping...\n");
  if (init_sua)
  {
    sua_term();
  }
  if (init_sca)
  {
    sca_term();
  }
  if (init_scm)
  {
    scm_term();
  }
  if (init_osp)
  {
    osp_term();
  }

  printf("ospd stopped\n");
#ifdef OSP_DEBUG
  printf("Final memory usage:\n");
  printf("\tGeneral purpose blocks:  %lu\n", count_mem);
  printf("\tThread structures:       %lu\n", count_struct[1]);
  printf("\tConnection structures:   %lu\n", count_struct[2]);
  printf("\tTransaction structures:  %lu\n", count_struct[3]);
#endif

  TRC_EXIT();
  return(0);

} /* main */

/*****************************************************************************/
/* Client verification callback.                                             */
/*****************************************************************************/

OSP_RC cv_cb(unsigned long   remote_ip,
             void          **ver_cookie,
             char           *cert,
             unsigned int    cert_len,
             unsigned int    chain_len,
             unsigned int    chain_pos)
{
  TRC_ENTRY("cv_cb");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((ver_cookie != NULL), (TB, "ver_cookie is null"));

  /***************************************************************************/
  /* Trace parameters and set ver_cookie to a recognizable value.            */
  /***************************************************************************/
  TRC_NORMAL((TB, "cv_cb called!"));
  TRC_NORMAL((TB, "  remote IP: %08x", remote_ip));
  TRC_NORMAL((TB, "  ver_cookie at %p is %p; setting to 0x5a5a5a5a",
                                                     ver_cookie, *ver_cookie));
  *ver_cookie = (void *)0x5a5a5a5a;
  if (cert != NULL)
  {
    TRC_NORMAL((TB, "  cert at %p", cert));
  }
  TRC_NORMAL((TB, "  cert_len: %u", cert_len));
  TRC_NORMAL((TB, "  chain_len: %u", chain_len));
  TRC_NORMAL((TB, "  chain_pos: %u", chain_pos));

  TRC_EXIT();
  return(OSP_SUCCESS);

} /* cv_cb */

/*****************************************************************************/
/* Client disconnection callback.                                            */
/*****************************************************************************/

void cd_cb(OSP_CLIENT_ID *client_id)
{
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("cd_cb");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((client_id != NULL), (TB, "client_id is null"));

  /***************************************************************************/
  /* Trace parameters.                                                       */
  /***************************************************************************/
  TRC_NORMAL((TB, "cd_cb called! client_id %p", client_id));
  TRC_NORMAL((TB, "  ip_address %08x", client_id->ip_address));
  TRC_NORMAL((TB, "  ver_cookie %p", client_id->ver_cookie));

  TRC_EXIT();
  return;

} /* cd_cb */

