
/**MOD+***********************************************************************/
/* Module:    scaarcb.c                                                      */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Authorization Request processing                               */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.19               $ $Modtime::   Jan 11 2001 15:56:16   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_auth_rq_callback                                           */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/*                                                                           */
/* Operation: Decide on the requested combination of authentication,         */
/*            authorization,and routing                                      */
/*                                                                           */
/*            Store parameters required for response                         */
/*            Build and submit routing query to LAM                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI sca_auth_rq_callback(OSP_CORRELATOR corr,
                                   char *timestamp,
                                   OSP_CALL_ID *call_id_list,
                                   OSP_ADDRESS *source_info,
                                   OSP_ADDRESS *source_alt_list,
                                   OSP_ADDRESS *dest_info,
                                   OSP_ADDRESS *dest_alt_list,
                                   OSP_SERVICE_INFO *service,
                                   char *max_dests,
                                   OSP_TOKEN *token_list,
                                   OSP_AUTH_INFO *authentication_info,
                                   OSP_CLIENT_ID *client_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_STATUS status;
  OSP_RC rc = OSP_SUCCESS;
  int authen_req;
  int author_req;
  int routing_req;
  char transaction_id[SCA_TRANS_ID_LEN];
  OSP_ADDRESS *subscriber_id;

  TRC_ENTRY("sca_auth_rq_callback");

  /***************************************************************************/
  /* Check parameters                                                        */
  /***************************************************************************/
  TRC_ASSERT((corr != NULL), (TB, "Passed NULL correlator"));
  TRC_ASSERT((source_info != NULL), (TB, "Passed NULL source address"));
  TRC_ASSERT((dest_info != NULL), (TB, "Passed NULL dest address"));
  TRC_ASSERT((max_dests != NULL), (TB, "Passed NULL max dests"));
  TRC_DETAIL((TB, "Auth Rq for source '%s' type '%s'",
                                     source_info->address, source_info->type));
  TRC_DETAIL((TB, "Destination address '%s' type '%s'",
                                         dest_info->address, dest_info->type));

  /***************************************************************************/
  /* Set up the default status                                               */
  /***************************************************************************/
  status.code = OSP_STATUS_SUCCESS;
  status.desc = sca_operation_success;

  /***************************************************************************/
  /* Whatever happens, we must have a valid transaction ID.                  */
  /***************************************************************************/
  sca_get_trans_id(transaction_id);
  TRC_DETAIL((TB, "Got transaction id: %s", transaction_id));

  /***************************************************************************/
  /* We only handle e164 numbers in this sample application                  */
  /***************************************************************************/
  if ((strcmp(source_info->type, sca_e164_type) != 0) ||
                                 (strcmp(dest_info->type, sca_e164_type) != 0))
  {
    /*************************************************************************/
    /* Set up the failure code                                               */
    /*************************************************************************/
    TRC_ERROR((TB, "Got non-e164 address.  Src type '%s', Dest '%s'",
                                          source_info->type, dest_info->type));
    status.code = OSP_STATUS_NOT_IMPLEMENTED;
    status.desc = (char *)sca_e164_only;

    /*************************************************************************/
    /* As far as the lower layers are concerned, we've still processed the   */
    /* message successfully, so keep the good return code                    */
    /*************************************************************************/
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Next, we need to work out exactly what we're being asked to do, but to  */
  /* understand what follows, you need some background...                    */
  /*                                                                         */
  /* In the first public release of the OSP spec (V1.4.2), there is no       */
  /* mechanism for subscriber authentication (which is used particularly in  */
  /* supporting roaming subscribers).  To get around this, the Authorization */
  /* Request may be overloaded to provide                                    */
  /*                                                                         */
  /* - authorization                                                         */
  /*                                                                         */
  /* - authentication and authorization                                      */
  /*                                                                         */
  /* - routing                                                               */
  /*                                                                         */
  /* - authorization and routing                                             */
  /*                                                                         */
  /* - all three!                                                            */
  /*                                                                         */
  /*                                                                         */
  /* The information to be used to authenticate the subscriber is supplied   */
  /* as a SourceAlternate of type "subscriber", and typically would consist  */
  /* of an identification card number and pin                                */
  /*                                                                         */
  /* The V2.0.9 release of the spec added a new request/response pair to     */
  /* explicitly support subscriber authentication, and also added optional   */
  /* lists of tokens and of subscriber authentication information to the     */
  /* elements of the AuthorizationRequest.  The elements of the new          */
  /* SubscriberAuthenticationRequest map directly on to those in a           */
  /* subscriber authentication plus authorization AuthorizationRequest.      */
  /*                                                                         */
  /* So how do we distinguish what we're being asked to do?                  */
  /*                                                                         */
  /*   if max_dests is non-zero                                              */
  /*     we need to do routing                                               */
  /*   fi                                                                    */
  /*                                                                         */
  /*   if src alt contains subscriber and subs info is NULL                  */
  /*     we need to do authentication                                        */
  /*     Note that if we need to to authentication, then                     */
  /*      we need to do authorization too                                    */
  /*     etoN                                                                */
  /*   fi                                                                    */
  /*                                                                         */
  /*   if subs info is non-NULL                                              */
  /*     we need to do authorization                                         */
  /*   fi                                                                    */
  /*                                                                         */
  /***************************************************************************/

  /***************************************************************************/
  /* We start with the assumption that we need to do nothing                 */
  /***************************************************************************/
  authen_req = FALSE;
  author_req = FALSE;
  routing_req = FALSE;
  subscriber_id = NULL;

  /***************************************************************************/
  /* Check max dests                                                         */
  /***************************************************************************/
  if (strcmp(max_dests, "0") != 0)
  {
    TRC_DETAIL((TB, "Routing will be required"));
    routing_req = TRUE;
  }

  /***************************************************************************/
  /* Check for a source alt of type subscriber                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Looking for subscriber info..."));
  subscriber_id = sca_find_subs_info(source_alt_list);
  if ((subscriber_id != NULL) && (authentication_info == NULL))
  {
    TRC_DETAIL((TB, "Found subscriber info and no authen info supplied"));
    TRC_DETAIL((TB, "Authorization and authentication required"));
    authen_req = TRUE;
    author_req = TRUE;
  }

  /***************************************************************************/
  /* If previously acquired subscriber authentication is supplied, then we   */
  /* now need to do authorization                                            */
  /***************************************************************************/
  if (authentication_info != NULL)
  {
    TRC_DETAIL((TB, "Authorization required"));
    author_req = TRUE;
  }

  /***************************************************************************/
  /* Authentication requires us to look up the subscriber in our directory   */
  /* before doing anything else - we'll do the authorization and any routing */
  /* in the callback from that lookup                                        */
  /***************************************************************************/
  if (authen_req)
  {
    TRC_DETAIL((TB, "Doing authentication..."));
    rc = sca_do_authentication(corr,
                               SCA_ROUTE_AUTH_RQ,
                               call_id_list,
                               source_info,
                               subscriber_id,
                               dest_info,
                               dest_alt_list,
                               max_dests);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Authentication failed, rc = %d", rc));
      status.code = OSP_STATUS_AUTHENTICATION_UNSUCCESSFUL;
      status.desc = (char *)sca_authen_failed;
    }
  }
  else
  {
    /*************************************************************************/
    /* Do we need to authorise this person?  This is a synchronous operation */
    /*************************************************************************/
    if (author_req)
    {
      rc = sca_do_authorization(source_info,
                                source_alt_list,
                                authentication_info);
      if (rc == OSP_SUCCESS)
      {
        /*********************************************************************/
        /* Set up the status description here in case there's no routing to  */
        /* be done                                                           */
        /*********************************************************************/
        TRC_DETAIL((TB, "Authorization successfull"));
        status.desc = (char *)sca_author_success;
      }
      else
      {
        TRC_ERROR((TB, "Authorization failed, rc = %d", rc));
        status.code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
        status.desc = (char *)sca_author_failed;
        goto EXIT_LABEL;
      }
    }

    /*************************************************************************/
    /* Do any routing required - this is asynchronous, as it requires a      */
    /* lookup of the route in the directory.                                 */
    /*************************************************************************/
    if (routing_req)
    {
      TRC_DETAIL((TB, "Doing routing..."));
      rc = sca_do_routing(corr, SCA_ROUTE_AUTH_RQ, call_id_list, source_info,
                                 dest_info, dest_alt_list, service, max_dests);
      if (rc != OSP_SUCCESS)
      {
        TRC_ERROR((TB, "routing failed with rc %d", rc));
        status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
        status.desc = (char *)scm_subs_server_failure;
        goto EXIT_LABEL;
      }
    }

    /*************************************************************************/
    /* Should never be asked to do nothing!                                  */
    /*************************************************************************/
    if (!author_req && !routing_req)
    {
      TRC_ERROR((TB, "Auth Request with no operations requested"));
      status.code = OSP_STATUS_BAD_REQUEST;
      status.desc = (char *)sca_no_operation;
      goto EXIT_LABEL;
    }
  }

EXIT_LABEL:
  /***************************************************************************/
  /* If something went wrong, then we need to send a (failure) response now  */
  /***************************************************************************/
  if (status.code != OSP_STATUS_SUCCESS)
  {
    TRC_ALERT((TB, "Immediate AR failure, code %d", status.code));
    rc = osp_ar_auth_response(corr,
                              &status,
                              transaction_id,
                              NULL,
                              NULL,
                              scm_sign_responses);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We absolutely should NOT get a bad parameter error; otherwise just  */
      /* trace the result.                                                   */
      /***********************************************************************/
      TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
      TRC_ERROR((TB, "Auth response call failed, rc = %d", rc));
    }
  }

  TRC_EXIT();

  return(rc);

} /* sca_auth_rq_callback */

