
/**MOD+***********************************************************************/
/* Module:    scaauthr.c                                                     */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Authorization processing                                       */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.5                $ $Modtime::   Jan 11 2001 15:56:16   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_do_authorization                                           */
/*                                                                           */
/* Purpose:   Authorize the request                                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS - authorization successful                         */
/*            other       - a failure occured                                */
/*                                                                           */
/* Params:    IN     source_info  - pointer to source info structure         */
/*            IN     source_alt_list                                         */
/*                                - list of source alternates                */
/*            IN     authentication_info                                     */
/*                                - supplied subscriber authentication       */
/*                                  tokens                                   */
/*                                                                           */
/* Operation: Verfiy supplied authentication info by                         */
/*            - 'cracking' the token                                         */
/*            - checking that it matches the supplied information about the  */
/*              source                                                       */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_do_authorization(OSP_ADDRESS *source_info,
                            OSP_ADDRESS *source_alt_list,
                            OSP_AUTH_INFO *authentication_info)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERR_FAILED_VERIFY;
  int valid_rc;
  OSP_TOKEN authen_token;
  SCA_TOKEN_INFO *cracked_token;
  OSP_STATUS status;

  TRC_ENTRY("sca_do_authorization");

  /***************************************************************************/
  /* Assuming that we issued it(!), the received authentication info is      */
  /* actually a normal authorization token minus the destination             */
  /* information.  We need to crack the token (ie verify the signature and   */
  /* parse the XML) then check the token is for the source we've just        */
  /* received it from.                                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Cracking token"));
  authen_token.next = NULL;
  authen_token.encoding = authentication_info->encoding;
  authen_token.token_len = 0;
  authen_token.token_data = authentication_info->auth_data;

  cracked_token = sca_crack_token(&status, &authen_token);
  if (cracked_token == NULL)
  {
    TRC_ERROR((TB, "Failed to crack authen info, status %d", status.code));
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Token cracked OK"));

  /***************************************************************************/
  /* Check the token is still valid                                          */
  /***************************************************************************/
  valid_rc = sca_check_validity(cracked_token->valid_after,
                                                   cracked_token->valid_until);
  if (valid_rc != SCA_VALIDITY_OK)
  {
    TRC_ERROR((TB, "Authentication info is not valid, rc %d", valid_rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now check what we got back                                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Examining authen info, source '%s'", source_info->address));
  TRC_DETAIL((TB, "Authen info for '%s'",cracked_token->source_info->address));

  if (!strcmp(cracked_token->source_info->address, source_info->address))
  {
    TRC_DETAIL((TB, "Token sources match - authorized"));
    rc = OSP_SUCCESS;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Tidy up                                                                 */
  /***************************************************************************/
  if (cracked_token != NULL)
  {
    TRC_DETAIL((TB, "Freeing token"));
    sca_free_token_info(cracked_token);
  }

  TRC_EXIT();

  return(rc);

} /* sca_do_authorization */

