
/**MOD+***********************************************************************/
/* Module:    scacicb.c                                                      */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Caps Indication processing                                     */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.5                $ $Modtime::   Jan 11 2001 15:56:16   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_caps_ind_callback                                          */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/*                                                                           */
/* Operation: if directory update required, initiate update (OOR or protocol */
/*              support settings provided)                                   */
/*            else send immediate caps confirm                               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI sca_caps_ind_callback(OSP_CORRELATOR corr,
                                    OSP_DEVICE_INFO *dev_info_list,
                                    char *osp_version,
                                    OSP_CAPS *caps_list,
                                    OSP_RESOURCE *resource_list,
                                    OSP_CLIENT_ID *client_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_STATUS status;
  OSP_RC rc = OSP_SUCCESS;
  int oor_flag;
  int prot_count;
  OSP_PROTOCOL_TYPE *prot_list;

  TRC_ENTRY("sca_caps_ind_callback");

  /***************************************************************************/
  /* Check parameters                                                        */
  /***************************************************************************/
  TRC_ASSERT((corr != NULL), (TB, "Passed NULL correlator"));
  TRC_ASSERT((osp_version != NULL), (TB, "Passed NULL version string"));
  TRC_DETAIL((TB, "Caps Ind received"));

  /***************************************************************************/
  /* Check for OOR updates                                                   */
  /***************************************************************************/
  if (sca_maybe_update_entry(resource_list, &oor_flag, &prot_list,&prot_count))
  {
    /*************************************************************************/
    /* Need to update settings before sending confirm                        */
    /*************************************************************************/
    rc = sca_update_entry(oor_flag, prot_list, prot_count, corr, dev_info_list,
                                                       osp_version, caps_list);
    if (rc == OSP_SUCCESS)
    {
      /***********************************************************************/
      /* Don't want to delete the protocol list yet!                         */
      /***********************************************************************/
      TRC_DETAIL((TB, "Entry update started"));
      prot_list = NULL;
    }
  }
  else
  {
    /*************************************************************************/
    /* No OOR updates - send confirm now                                     */
    /*************************************************************************/
    rc = sca_send_caps_confirm(corr, osp_version, caps_list);
  }

  /***************************************************************************/
  /* Do we need to send a response straightaway?                             */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Set the return code                                                   */
    /*************************************************************************/
    if (rc == OSP_ERR_BAD_REQUEST)
    {
      /***********************************************************************/
      /* This means that no transport address was provided                   */
      /***********************************************************************/
      status.code = OSP_STATUS_BAD_REQUEST;
      status.desc = (char *)sca_caps_no_transport;
    }
    else
    {
      status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status.desc = (char *)scm_subs_resource_failure;
    }

    TRC_NORMAL((TB, "Immediate Caps Confirm, code %d", status.code));
    rc = osp_ar_caps_confirm(corr,
                             &status,
                             NULL,
                             NULL,
                             NULL,
                             NULL,
                             scm_sign_responses);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We absolutely should NOT get a bad parameter error; otherwise just  */
      /* trace the result.                                                   */
      /***********************************************************************/
      TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
      TRC_ERROR((TB, "caps confirm call failed, rc = %d", rc));
    }
  }

  /***************************************************************************/
  /* Maybe free up the protocol list                                         */
  /***************************************************************************/
  if (prot_list != NULL)
  {
    TRC_DETAIL((TB, "Freeing protocol list"));
    sca_free_prot_list(prot_list);
  }

  TRC_EXIT();

  return(rc);

} /* sca_caps_ind_callback */

/**PROC+**********************************************************************/
/* Name:      sca_maybe_update_entry                                         */
/*                                                                           */
/* Purpose:   scan supplied resource list for any with the OOR flag present  */
/*            or with supported protocols specified                          */
/*                                                                           */
/* Returns:   TRUE if update required                                        */
/*            FALSE otherwise                                                */
/*                                                                           */
/* Params:    IN     resource_list                                           */
/*                                - list of resources to check               */
/*            OUT    oor_flag     - returned OOR setting                     */
/*            OUT    prot_list    - returned list of supported protocols     */
/*            OUT    prot_count   - count of entries in prot_list            */
/*                                                                           */
/* Operation: scan the supplied resource list for entries with an            */
/*              almost out of resources element.                             */
/*            If any entries have the element, an update is required; if     */
/*            any have the element set to true, then the returned oor flag   */
/*            is set TRUE.                                                   */
/*                                                                           */
/*            Similarly, if any resources have supported protcols set,       */
/*              we need to add these to the entry.  Note that an empty       */
/*              supported protocols element means 'delete all'               */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_maybe_update_entry(OSP_RESOURCE *resource_list,
                           int *oor_flag,
                           OSP_PROTOCOL_TYPE **prot_list,
                           int *prot_count)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int update_required;
  OSP_RESOURCE *curr_item;
  OSP_SUPP_PROTOCOL *curr_supp_prot;
  OSP_PROTOCOL_TYPE *new_prot;
  SLIST_ROOT prot_list_root;
  int prot_list_count;
  int success;
  int prot_delete_all;

  TRC_ENTRY("sca_maybe_update_entry");

  /***************************************************************************/
  /* Initialization                                                          */
  /***************************************************************************/
  success = TRUE;
  update_required = FALSE;
  prot_delete_all = FALSE;
  *oor_flag = SCA_OOR_NO_CHANGE;
  *prot_list = NULL;
  SL_INIT_ROOT(prot_list_root);
  prot_list_count = 0;

  /***************************************************************************/
  /* Walk the list                                                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Checking resource list"));
  for (curr_item = resource_list;
       curr_item != NULL;
       curr_item = curr_item->next)
  {
    /*************************************************************************/
    /* Is the almost out flag present?                                       */
    /*************************************************************************/
    if (curr_item->almost_out != NULL)
    {
      /***********************************************************************/
      /* An update is required - now test the value                          */
      /***********************************************************************/
      update_required = TRUE;
      if (*oor_flag == SCA_OOR_NO_CHANGE)
      {
        *oor_flag = SCA_OOR_MAKE_FALSE;
      }
      TRC_DETAIL((TB, "Resource has almost out flag '%s'",
                                                       curr_item->almost_out));
      if (!strcasecmp(curr_item->almost_out, sca_aoor_true))
      {
        *oor_flag = SCA_OOR_MAKE_TRUE;
        TRC_DETAIL((TB, "Found OOR True"));
      }
    }

    /*************************************************************************/
    /* Check for supported protocols                                         */
    /*************************************************************************/
    TRC_DETAIL((TB, "Checking for supported protocols"));
    for (curr_supp_prot = curr_item->supp_protocol_list;
         curr_supp_prot != NULL;
         curr_supp_prot = curr_supp_prot->next)
    {
      /***********************************************************************/
      /* Check the protocol information                                      */
      /***********************************************************************/
      update_required = TRUE;
      if (curr_supp_prot->prot_type == NULL)
      {
        TRC_DETAIL((TB, "Empty protocol type - delete all"));
        prot_delete_all = TRUE;
        break;
      }
      else
      {
        TRC_ASSERT((curr_supp_prot->prot_type->next == NULL),
                                       (TB, "Only one protocol type allowed"));
        TRC_ASSERT((curr_supp_prot->prot_type->prot_name != NULL),
                                           (TB, "Protocol type without name"));
        TRC_DETAIL((TB, "Got protocol '%s'",
                                        curr_supp_prot->prot_type->prot_name));

        /*********************************************************************/
        /* Save off the protocol name                                        */
        /*********************************************************************/
        new_prot = rm_get_mem(sizeof(OSP_PROTOCOL_TYPE));
        if (new_prot == NULL)
        {
          TRC_ERROR((TB, "Failed to get memory for protocol type"));
          success = FALSE;
          goto EXIT_LABEL;
        }
        new_prot->next = NULL;

        OSP_STRDUP(new_prot->prot_name, curr_supp_prot->prot_type->prot_name);
        if (new_prot->prot_name == NULL)
        {
          TRC_ERROR((TB, "Failed to copy protocol name"));
          rm_release_mem(new_prot);
          success = FALSE;
          goto EXIT_LABEL;
        }

        SL_ADD_TO_TAIL(prot_list_root, new_prot);
        prot_list_count++;
      }
    }
  }

  /***************************************************************************/
  /* Everything went OK                                                      */
  /***************************************************************************/
  *prot_list = SL_HEAD(prot_list_root);
  if (prot_delete_all)
  {
    *prot_count = -1;
  }
  else
  {
    *prot_count = prot_list_count;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Did something go wrong?                                                 */
  /***************************************************************************/
  if (!success)
  {
    TRC_DETAIL((TB, "Tidy up after failure"));
    *oor_flag = FALSE;

    if (SL_HEAD(prot_list_root)!= NULL)
    {
      TRC_DETAIL((TB, "Freeing protocol list"));
      sca_free_prot_list(SL_HEAD(prot_list_root));
    }
  }

  TRC_EXIT();

  return(update_required);

} /* sca_maybe_update_entry */

/**PROC+**********************************************************************/
/* Name:      sca_update_entry                                               */
/*                                                                           */
/* Purpose:   Start updates of a gateways records in the directory           */
/*                                                                           */
/* Returns:   OSP_SUCCESS - update initiated                                 */
/*            other - OSP failure code                                       */
/*                                                                           */
/* Params:    IN     oor_flag     - new OOR setting                          */
/*            IN     prot_list    - list of supported protocols.  This is    */
/*                                  allocated by SCA, so we don't need to    */
/*                                  copy it to store it                      */
/*            IN     corr         - correlator to be used on subsequent      */
/*                                  confirm                                  */
/*            IN     dev_info_list                                           */
/*                                - device info which must contain the       */
/*                                  the transport address of the gateway     */
/*            IN     osp_version  - client osp version                       */
/*            IN     caps_list    - list of client caps for which service    */
/*                                  information should be returned           */
/*                                                                           */
/* Operation: Search for transport address in device info                    */
/*            Start directory lookup on address (modification will be done   */
/*            in result callback)                                            */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_update_entry(int oor_flag,
                        OSP_PROTOCOL_TYPE *prot_list,
                        int prot_count,
                        OSP_CORRELATOR corr,
                        OSP_DEVICE_INFO *dev_info_list,
                        char *osp_version,
                        OSP_CAPS *caps_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  SCA_CAPS_DATA *caps_data;
  OSP_CAPS *curr_item;
  OSP_CAPS *new_item;
  SLIST_ROOT caps_list_copy;
  OSP_DEVICE_INFO *curr_dev;
  char *transport_address;
  char filter[256];

  TRC_ENTRY("sca_update_entry");

  /***************************************************************************/
  /* Initialize locals                                                       */
  /***************************************************************************/
  caps_data = NULL;
  transport_address = NULL;

  /***************************************************************************/
  /* First, check there's a transport address in the supplied device info -  */
  /* without that, we don't know what gateway to update!                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Checking for transport address in device info"));
  for (curr_dev = dev_info_list;
       curr_dev != NULL;
       curr_dev = curr_dev->next)
  {
    TRC_DETAIL((TB, "Looking at type '%s'", curr_dev->type));
    if (!strcmp(curr_dev->type, sca_transport))
    {
      transport_address = curr_dev->data;
      TRC_DETAIL((TB, "Got transport address '%s'", transport_address));
      break;
    }
  }

  if (transport_address == NULL)
  {
    TRC_DETAIL((TB, "Didn't find transport address - can't do update"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We should now be able to go ahead with the update, so store the caps    */
  /* list etc for use in the callback                                        */
  /***************************************************************************/
  caps_data = rm_get_mem(sizeof(SCA_CAPS_DATA));
  if (caps_data == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for caps data"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }
  caps_data->corr = corr;
  caps_data->oor_flag = oor_flag;
  caps_data->osp_version = NULL;
  caps_data->caps_list = NULL;
  caps_data->prot_list = NULL;
  caps_data->prot_count = 0;

  OSP_STRDUP(caps_data->osp_version, osp_version);
  if (caps_data->osp_version == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for osp_version"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Copy the caps list                                                      */
  /***************************************************************************/
  SL_INIT_ROOT(caps_list_copy);
  for (curr_item = caps_list;
       curr_item != NULL;
       curr_item = curr_item->next)
  {
    new_item = rm_get_mem(sizeof(OSP_CAPS));
    if (new_item == NULL)
    {
      TRC_ERROR((TB, "Failed to get mem for new caps item"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }

    new_item->next = NULL;
    OSP_STRDUP(new_item->cap_name, curr_item->cap_name);

    if (new_item->cap_name == NULL)
    {
      TRC_ERROR((TB, "Failed to get mem for new cap name"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }

    SL_ADD_TO_TAIL(caps_list_copy, new_item);
  }
  caps_data->caps_list = SL_HEAD(caps_list_copy);

  /***************************************************************************/
  /* And the protocol list                                                   */
  /***************************************************************************/
  caps_data->prot_count = prot_count;
  caps_data->prot_list = prot_list;

  /***************************************************************************/
  /* We need to find the gateway in the directory so we can then apply the   */
  /* modification to the OOR flag.  The filter is simply                     */
  /* "ip_address=<gateway's address>"                                        */
  /***************************************************************************/
  OSP_SNPRINTF((filter, sizeof(filter), "(%s=%s)", sca_ip_address,
                                                           transport_address));

  rc = lam_search(sca_lam_handle,
                  (char *)sca_ar_base,
                  LDAP_SCOPE_SUBTREE,
                  filter,
                  (char **)sca_auth_rq_attrs,
                  FALSE,
                  (void *)caps_data,
                  sca_caps_search_callback);

  if (rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* now we wait for the LAM callback, where we'll do the modification     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Search initiated successfully"));
  }
  else
  {
    TRC_ERROR((TB, "Search operation failed with rc %d", rc));
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Any tidy up required?                                                   */
  /***************************************************************************/
  if ((rc != OSP_SUCCESS) && (caps_data != NULL))
  {
    /*************************************************************************/
    /* Caller will free the prot list in this instance                       */
    /*************************************************************************/
    caps_data->prot_list = NULL;
    TRC_DETAIL((TB, "Freeing caps data after failure"));
    sca_free_caps(caps_data);
  }

  TRC_EXIT();

  return(rc);

} /* sca_update_entry */

/**PROC+**********************************************************************/
/* Name:      sca_caps_search_callback                                       */
/*                                                                           */
/* Purpose:   Called by LAM on search completion                             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - userdata supplied to lam_search/modify   */
/*            IN     ld           - LDAP handle required to parse result     */
/*            IN     result       - results of the search operation.         */
/*                                                                           */
/**PROC-**********************************************************************/

void OSPAPI sca_caps_search_callback(void *user_data,
                                     LDAP *ld,
                                     LDAPMessage **result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  OSP_STATUS status;
  SCA_CAPS_DATA *caps_data;
  int result_count;
  char *dn;
  LDAPMod oor_mod;
  char * oor_values[2];
  LDAPMod prot_mod;
  char **prot_values;
  int mod_count;
  int val_count;
  LDAPMod *mods[3];
  int ldap_rc;
  OSP_PROTOCOL_TYPE *curr_prot;

  TRC_ENTRY("sca_caps_search_callback");

  /***************************************************************************/
  /* Extract the caps data from the user_data.  Without it there's nothing   */
  /* we can do - we don't even have the correlator required for the callback */
  /* into the AIB - hence the assert                                         */
  /***************************************************************************/
  caps_data = (SCA_CAPS_DATA *)user_data;
  TRC_ASSERT((caps_data != NULL), (TB, "Received NULL caps data in LAM cb"));

  /***************************************************************************/
  /* Initialze variables                                                     */
  /***************************************************************************/
  prot_values = NULL;

  /***************************************************************************/
  /* Did the search operation complete OK?                                   */
  /***************************************************************************/
  if ((ld == NULL) || (result == NULL) || (*result == NULL))
  {
    status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status.desc = (char *)scm_subs_resource_failure;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up default values for the status field                              */
  /***************************************************************************/
  status.code = OSP_STATUS_SUCCESS;
  status.desc = sca_pricing_success;

  /***************************************************************************/
  /* See if we found the gateway                                             */
  /***************************************************************************/
  result_count = ldap_count_entries(ld, *result);
  TRC_DETAIL((TB, "Got %d results", result_count));
  TRC_ASSERT((result_count <= 1), (TB, "Got multiple matches for gateway"));
  if (result_count == 1)
  {
    /*************************************************************************/
    /* Got a match: we can go ahead and modify the entry                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got match on gateway/destination combination"));

    /*************************************************************************/
    /* Get the dn                                                            */
    /*************************************************************************/
    dn = ldap_get_dn(ld, *result);
    if (dn == NULL)
    {
      /***********************************************************************/
      /* With no dn, there's nothing we can do                               */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to get dn - can't update price"));
      status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status.desc = (char *)scm_subs_resource_failure;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Build the required modify operations, first for the OOR flag          */
    /*************************************************************************/
    mod_count = 0;
    if (caps_data->oor_flag == SCA_OOR_NO_CHANGE)
    {
      TRC_DETAIL((TB, "No change to OOR flag"));
    }
    else
    {
      TRC_DETAIL((TB, "Updating OOR setting to..."));
      oor_mod.mod_op = LDAP_MOD_REPLACE;
      oor_mod.mod_type = (char *)sca_almost_out;
      if (caps_data->oor_flag == SCA_OOR_MAKE_TRUE)
      {
        TRC_DETAIL((TB, "TRUE"));
        oor_values[0] = (char *)sca_aoor_true;
      }
      else
      {
        TRC_DETAIL((TB, "FALSE"));
        oor_values[0] = (char *)sca_aoor_false;
      }
      oor_values[1] = NULL;
      oor_mod.mod_values = oor_values;
      mods[mod_count] = &oor_mod;
      mod_count++;
    }

    /*************************************************************************/
    /* Now add in any protocols                                              */
    /*************************************************************************/
    if (caps_data->prot_count == -1)
    {
      TRC_DETAIL((TB, "Deleting all protocols"));
      prot_mod.mod_op = LDAP_MOD_DELETE;
      prot_mod.mod_type = (char *)sca_prot_type;
      prot_mod.mod_values = NULL;
      mods[mod_count] = &prot_mod;
      mod_count++;
    }
    else if (caps_data->prot_list != NULL)
    {
      TRC_DETAIL((TB, "Updating protocols"));

      /***********************************************************************/
      /* Need to get memory for values                                       */
      /***********************************************************************/
      prot_values = rm_get_mem((caps_data->prot_count + 1) * sizeof(char *));
      if (prot_values == NULL)
      {
        TRC_ERROR((TB, "Failed to get memory for protocol values"));
        status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
        status.desc = (char *)scm_subs_resource_failure;
        goto EXIT_LABEL;
      }

      /***********************************************************************/
      /* Now add each value                                                  */
      /***********************************************************************/
      val_count = 0;
      for (curr_prot = caps_data->prot_list;
           curr_prot != NULL;
           curr_prot = curr_prot->next)
      {
        TRC_DETAIL((TB, "Adding protocol '%s'", curr_prot->prot_name));
        prot_values[val_count] = curr_prot->prot_name;
        val_count++;
      }
      prot_values[val_count] = NULL;

      /***********************************************************************/
      /* Build the mod                                                       */
      /***********************************************************************/
      TRC_DETAIL((TB, "Adding protocol mods"));
      prot_mod.mod_op = LDAP_MOD_REPLACE;
      prot_mod.mod_type = (char *)sca_prot_type;
      prot_mod.mod_values = prot_values;
      mods[mod_count] = &prot_mod;
      mod_count++;
    }
    mods[mod_count] = NULL;

    /*************************************************************************/
    /* And apply the modification                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Writing updates..."));
    ldap_rc = ldap_modify_s(ld, dn, mods);
    if (ldap_rc == LDAP_SUCCESS)
    {
      TRC_DETAIL((TB, "Updates successful"));
    }
    else
    {
      TRC_ERROR((TB, "LDAP failure writing updates"));
      status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status.desc = (char *)scm_subs_resource_failure;
      goto EXIT_LABEL;
    }
  }
  else
  {
    /*************************************************************************/
    /* Set the status code and send a failure response                       */
    /*************************************************************************/
    TRC_ALERT((TB, "Search found no matching destinations"));
    status.code = OSP_STATUS_ROUTE_AUTHORIZATION_UNSUCCESSFUL;
    status.desc = (char *)sca_route_not_found;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If we get here, send a good confirmation                                */
  /***************************************************************************/
  rc = sca_send_caps_confirm(caps_data->corr, caps_data->osp_version,
                                                         caps_data->caps_list);

EXIT_LABEL:
  /***************************************************************************/
  /* If something went wrong, we need to send a confirm                      */
  /***************************************************************************/
  if (status.code != OSP_STATUS_SUCCESS)
  {
    TRC_NORMAL((TB, "Caps Confirm failure, code %d", status.code));
    rc = osp_ar_caps_confirm(caps_data->corr,
                             &status,
                             NULL,
                             NULL,
                             NULL,
                             NULL,
                             scm_sign_responses);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We absolutely should NOT get a bad parameter error; otherwise just  */
      /* trace the result.                                                   */
      /***********************************************************************/
      TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
      TRC_ERROR((TB, "caps confirm call failed, rc = %d", rc));
    }
  }

  /***************************************************************************/
  /* Tidy up                                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing caps memory"));
  sca_free_caps(caps_data);
  rm_release_mem(prot_values);

  TRC_EXIT();

  return;

} /* sca_caps_search_callback */

/**PROC+**********************************************************************/
/* Name:      sca_send_caps_confirm                                          */
/*                                                                           */
/* Purpose:   Send a caps confirm message based on the supplied parameters   */
/*                                                                           */
/* Returns:   OSP_SUCCESS - confirm sent successfully                        */
/*            other - OSP failure code                                       */
/*                                                                           */
/* Params:    IN     corr         - correlator to be used on subsequent      */
/*                                  confirm                                  */
/*            IN     osp_version  - client osp version                       */
/*            IN     caps_list    - list of client caps for which service    */
/*                                  information should be returned           */
/*                                                                           */
/* Operation: Decide the version as min(client version, server version)      */
/*            Build an OSP_SERVICE entry for each client capability          */
/*            Build the server certificate chain                             */
/*            Send the confirmation                                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_send_caps_confirm(OSP_CORRELATOR corr,
                             char *osp_version,
                             OSP_CAPS *caps_list)

{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_SERVICE *service;
  OSP_SERVICE_URL service_url;
  SLIST_ROOT service_list;
  OSP_CAPS *curr_caps;
  char neg_version[16];
  OSP_STATUS status;

  TRC_ENTRY("sca_send_caps_confirm");

  /***************************************************************************/
  /* Set up the service URL                                                  */
  /***************************************************************************/
  service_url.next = NULL;
  service_url.url = sca_service_url;

  /***************************************************************************/
  /* Set up the OSP version support, based on the lower of our support and   */
  /* the client's                                                            */
  /***************************************************************************/
  if (osp_version[0] == '1')
  {
    strcpy(neg_version, "1.4.2");
  }
  else
  {
    strcpy(neg_version, "2.1.0");
  }
  TRC_DETAIL((TB, "Negotiated OSP version %s support", neg_version));

  /***************************************************************************/
  /* Build the service list based on the caps list                           */
  /***************************************************************************/
  SL_INIT_ROOT(service_list);
  for (curr_caps = caps_list;
       curr_caps != NULL;
       curr_caps = curr_caps->next)
  {
    TRC_DETAIL((TB, "Looking at service %s", curr_caps->cap_name));
    /*************************************************************************/
    /* Check its a service we provide                                        */
    /*************************************************************************/
    if (!sca_service_available(curr_caps->cap_name))
    {
      TRC_ALERT((TB, "Service %s not available", curr_caps->cap_name));
      continue;
    }

    /*************************************************************************/
    /* Get memory for the new service                                        */
    /*************************************************************************/
    service = rm_get_mem(sizeof(OSP_SERVICE));
    if (service == NULL)
    {
      TRC_ERROR((TB, "Failed to get mem for new service"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }
    service->next = NULL;
    service->capability = NULL;
    service->url_list = &service_url;
    service->sig_required = FALSE;

    /*************************************************************************/
    /* Set up the capability name                                            */
    /*************************************************************************/
    OSP_STRDUP(service->capability, curr_caps->cap_name);
    if ((service)->capability == NULL)
    {
      TRC_ERROR((TB, "Failed to get mem for new capability"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Add the new service to the list                                       */
    /*************************************************************************/
    SL_ADD_TO_TAIL(service_list, service);
  }

  /***************************************************************************/
  /* Send the confirm                                                        */
  /***************************************************************************/
  status.code = OSP_STATUS_SUCCESS;
  status.desc = sca_operation_success;
  rc = osp_ar_caps_confirm(corr,
                           &status,
                           neg_version,
                           SL_HEAD(service_list),
                           sca_token_signing_cert_list,
                           NULL,
                           scm_sign_responses);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We absolutely should NOT get a bad parameter error; otherwise just    */
    /* trace the result.                                                     */
    /*************************************************************************/
    TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                       (TB, "Bad parameter errors unexpected from OSP api"));
    TRC_ERROR((TB, "caps confirm call failed, rc = %d", rc));

    /*************************************************************************/
    /* Set success again or we'll send again on return from this fn          */
    /*************************************************************************/
    rc = OSP_SUCCESS;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Tidy up                                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing service list"));
  while ((service = SL_HEAD(service_list)) != NULL)
  {
    SL_REMOVE_HEAD(service_list)
    rm_release_mem(service->capability);
    rm_release_mem(service);
    /*************************************************************************/
    /* Note that we don't need to free the url 'list' or its string, since   */
    /* the URL is a stack's variable containing a a pointer to a global!     */
    /*************************************************************************/
  }

  TRC_EXIT();

  return(rc);

} /* sca_send_caps_confirm */

/**PROC+**********************************************************************/
/* Name:      sca_free_caps                                                  */
/*                                                                           */
/* Purpose:   Free members of supplied capabilites structure                 */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     caps_data    - caps structure to free                   */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_caps(SCA_CAPS_DATA *caps_data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_CAPS *curr_caps, *next_caps;

  TRC_ENTRY("sca_free_caps");

  if (caps_data != NULL)
  {
    /*************************************************************************/
    /* Free each element                                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing caps data %p", caps_data));
    rm_release_mem(caps_data->osp_version);

    curr_caps = caps_data->caps_list;
    while (curr_caps != NULL)
    {
      next_caps = curr_caps->next;
      rm_release_mem(curr_caps->cap_name);
      rm_release_mem(curr_caps);
      curr_caps = next_caps;
    }

    if (caps_data->prot_list)
    {
      TRC_DETAIL((TB, "Freeing prot list"));
      sca_free_prot_list(caps_data->prot_list);
    }

    rm_release_mem(caps_data);
  }

  TRC_EXIT();

  return;

} /* sca_free_caps */

/**PROC+**********************************************************************/
/* Name:      sca_service_available                                          */
/*                                                                           */
/* Purpose:   Check named service is one we provide                          */
/*                                                                           */
/* Returns:   TRUE if service provided                                       */
/*                                                                           */
/* Params:    IN     service_name - name of service to check                 */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_service_available(char *service_name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int available = FALSE;
  int ii;

  TRC_ENTRY("sca_service_available");

  /***************************************************************************/
  /* Run through our list of services, looking for the service name          */
  /***************************************************************************/
  TRC_DETAIL((TB, "checking for '%s'", service_name));
  for (ii = 0; ii < SCA_NUM_SERVICES; ii++)
  {
    if (!strcmp(sca_services_list[ii], service_name))
    {
      TRC_DETAIL((TB, "Found it"));
      available = TRUE;
      break;
    }
  }

  TRC_EXIT();

  return(available);

} /* sca_service_available */

/**PROC+**********************************************************************/
/* Name:      sca_free_prot_list                                             */
/*                                                                           */
/* Purpose:   Free a list of protocol types                                  */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN     prot_list    - protocol list to be freed                */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_prot_list(OSP_PROTOCOL_TYPE *prot_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_PROTOCOL_TYPE *curr_prot, *next;

  TRC_ENTRY("sca_free_prot_list");

  TRC_DETAIL((TB, "Freeing prot list at %p", prot_list));
  curr_prot = prot_list;
  while (curr_prot != NULL)
  {
    next = curr_prot->next;
    rm_release_mem(curr_prot->prot_name);
    rm_release_mem(curr_prot);
    curr_prot = next;
  }

  TRC_EXIT();

  return;

} /* sca_free_prot_list */
