
/**MOD+***********************************************************************/
/* Module:    scamain.c                                                      */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            API functions                                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.22               $ $Modtime::   Jan 11 2001 15:56:18   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_init                                                       */
/*                                                                           */
/* Purpose:   Initializes the CAR sample application                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS          - completed successfully                  */
/*            OSP_ERR_ALREADY_INIT - the application was already initialized */
/*            OSP_ERR_CONFIG       - missing configuration information       */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: Register our callbacks with the AIB                            */
/*            Get a handle to LAM for use on searches                        */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_init(LHASH *config)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  char *host;
  char *host_copy;
  char *colon;
  int port;
  char *signing_dn;
  char *service_url;
  char *token_signing;

  TRC_ENTRY("sca_init");

  /***************************************************************************/
  /* Check configuration parameter.                                          */
  /***************************************************************************/
  TRC_ASSERT((config != NULL), (TB, "NULL configuration parameter"));

  /***************************************************************************/
  /* Check we're not already initialized                                     */
  /***************************************************************************/
  if (sca_initialized)
  {
    TRC_ERROR((TB, "Authorization App already initialized!"));
    rc = OSP_ERR_ALREADY_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get our configuration information                                       */
  /***************************************************************************/
  host_copy = NULL;
  port = 0;
  signing_dn = NULL;
  sca_signing_dn = NULL;
  service_url = NULL;

  /***************************************************************************/
  /* Get the host name of the LDAP routing directory                         */
  /***************************************************************************/
  host = CONF_get_string(config, SCM_CONFIG_SECTION, SCA_CONFIG_DIRECTORY);
  if (host != NULL)
  {
    TRC_NORMAL((TB, "ldap host name: '%s'", host));
    OSP_STRDUP(host_copy, host);
  }

  /***************************************************************************/
  /* Check to see if we should sign tokens.  Default is yes                  */
  /***************************************************************************/
  sca_sign_tokens = TRUE;
  token_signing = CONF_get_string(config, SCM_CONFIG_SECTION,
                                                       SCA_CONFIG_SIGN_TOKENS);
  if (token_signing != NULL)
  {
    TRC_NORMAL((TB, "Token signing enabled: '%s'", token_signing));
    if (strcasecmp(token_signing, SCM_CONFIG_DO_SIGN_RESPONSES) != 0)
    {
      TRC_NORMAL((TB, "Token signing off"));
      sca_sign_tokens = FALSE;
    }
  }

  /***************************************************************************/
  /* Get the name we use if signing tokens                                   */
  /***************************************************************************/
  if (sca_sign_tokens)
  {
    TRC_DETAIL((TB, "Getting token signing dn..."));
    signing_dn = CONF_get_string(config, SCM_CONFIG_SECTION,
                                 SCA_CONFIG_SIGNING_ID);
    if (signing_dn != NULL)
    {
      TRC_NORMAL((TB, "Signing dn: '%s'", signing_dn));
      OSP_STRDUP(sca_signing_dn, signing_dn);
    }
  }

  /***************************************************************************/
  /* Get the server's service URL                                            */
  /***************************************************************************/
  service_url = CONF_get_string(config, SCM_CONFIG_SECTION,
                                                       SCA_CONFIG_SERVICE_URL);
  if (service_url != NULL)
  {
    TRC_NORMAL((TB, "Service URL: '%s'", service_url));
    OSP_STRDUP(sca_service_url, service_url);
  }

  /***************************************************************************/
  /* Get the spread of validity for authorizations                           */
  /***************************************************************************/
  sca_auth_validity = CONF_get_number(config, SCM_CONFIG_SECTION,
                                                     SCA_CONFIG_AUTH_VALIDITY);
  if (sca_auth_validity == 0)
  {
    sca_auth_validity = SCA_DEFAULT_AUTH_VALIDITY;
    TRC_NORMAL((TB, "Using default token duration %d", sca_auth_validity));
  }
  else
  {
    TRC_NORMAL((TB, "Using token duration %d", sca_auth_validity));
  }

  /***************************************************************************/
  /* and the duration of call authorizations                                 */
  /***************************************************************************/
  sca_auth_duration = CONF_get_string(config, SCM_CONFIG_SECTION,
                                                     SCA_CONFIG_AUTH_DURATION);
  if (sca_auth_duration != NULL)
  {
    TRC_NORMAL((TB, "Using auth duration %s", sca_auth_duration));
  }
  else
  {
    TRC_NORMAL((TB, "No auth duration supplied"));
  }

  /***************************************************************************/
  /* Check we got all we require                                             */
  /***************************************************************************/
  if (host_copy == NULL)
  {
    TRC_ERROR((TB, "failed to get LDAP host name"));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  colon = strchr(host_copy, ':');
  if (colon == NULL)
  {
    port = LDAP_PORT;
  }
  else
  {
    /*************************************************************************/
    /* Scan the port number from the part of the string after the colon.     */
    /*************************************************************************/
    if (1 != sscanf(colon + 1, "%d", &port))
    {
      TRC_ERROR((TB, "invalid host:port configuration value: %s", host_copy));
      rc = OSP_ERR_CONFIG;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Null-terminate the host name part of the string.                      */
    /*************************************************************************/
    *colon = 0;
  }
  TRC_DETAIL((TB, "using LDAP port number: %d", port));

  if (sca_sign_tokens && (sca_signing_dn == NULL))
  {
    TRC_ERROR((TB, "failed to get dn for token signing"));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  if (sca_service_url == NULL)
  {
    TRC_ERROR((TB, "failed to get service URL"));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get a handle to the ldap access module, LAM                             */
  /***************************************************************************/
  sca_lam_handle = NULL;
  rc = lam_open(host_copy, port, "", "", &sca_lam_handle);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get LAM handle, rc %d", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If needed, get the list of certificates used for token signing - this   */
  /* is returned on caps confirm                                             */
  /***************************************************************************/
  if (sca_sign_tokens)
  {
    TRC_DETAIL((TB, "Get token signing cert chain"));
    rc = ccm_get_cert_chain(sca_signing_dn, &sca_token_signing_cert_list);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to get cert chain, rc %d", rc));
      goto EXIT_LABEL;
    }
    TRC_ASSERT((sca_token_signing_cert_list != NULL),
                                           (TB, "cert chain null on good rc"));
  }

  /***************************************************************************/
  /* Register our callbacks with the AIB - these will be called when the     */
  /* corresponding request or indication is received                         */
  /***************************************************************************/
  rc = osp_ar_register(sca_auth_rq_callback,
                       sca_auth_ind_callback,
                       sca_reauth_rq_callback,
                       sca_pricing_ind_callback,
                       sca_caps_ind_callback);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We absolutely should NOT get a bad parameter error; otherwise just    */
    /* trace the result.                                                     */
    /*************************************************************************/
    TRC_ASSERT((rc != OSP_ERR_BAD_PARAM),
                        (TB, "Bad parameter errors unexpected from OSP api"));
    TRC_ERROR((TB, "Failure registering CAR callbacks, rc %d", rc));
    goto EXIT_LABEL;
  }

  rc = osp_sa_register(sca_subs_auth_callback);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We absolutely should NOT get a bad parameter error; otherwise just    */
    /* trace the result.                                                     */
    /*************************************************************************/
    TRC_ASSERT((rc != OSP_ERR_BAD_PARAM),
                        (TB, "Bad parameter errors unexpected from OSP api"));
    TRC_ERROR((TB, "Failure registering SA callbacks, rc %d", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize the count we use for transaction ids, and the mutex we use   */
  /* to protect it.  Note that this has to start at 1 to be interoperable    */
  /* with some routers                                                       */
  /***************************************************************************/
  if (osp_create_mutex(&sca_trans_count_mutex) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create trans count mutex"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }
  OSP_GET_LOCK(&sca_trans_count_mutex);
  sca_trans_count = 1;
  OSP_RELEASE_LOCK(&sca_trans_count_mutex);

  /***************************************************************************/
  /* Set the array of attributes we want to query                            */
  /***************************************************************************/
  sca_auth_rq_attrs[0] = sca_ip_address;
  sca_auth_rq_attrs[1] = sca_almost_out;
  sca_auth_rq_attrs[2] = sca_e164_attr;
  sca_auth_rq_attrs[3] = sca_pricing;
  sca_auth_rq_attrs[4] = NULL;

  /***************************************************************************/
  /* If we get here, we're up and running.                                   */
  /***************************************************************************/
  TRC_NORMAL((TB, "AR App initialized successfully"));
  sca_initialized = TRUE;

EXIT_LABEL:

  /***************************************************************************/
  /* If the callback registration failed, we need to release the lam handle  */
  /* (if we got one!)                                                        */
  /***************************************************************************/
  if ((rc != OSP_SUCCESS) && (sca_lam_handle != NULL))
  {
    TRC_DETAIL((TB, "AR init failed: closing LAM handle"));
    lam_close(&sca_lam_handle);
  }

  rm_release_mem(host_copy);

  TRC_EXIT();

  return(rc);

} /* sca_init */

/**PROC+**********************************************************************/
/* Name:      sca_term                                                       */
/*                                                                           */
/* Purpose:   Terminates the CAR sample application                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - completed successfully                      */
/*            OSP_ERR_NOT_INIT - the AIB was not initialized                 */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: Deregister the Auth Rq callback                                */
/*            Close the LAM handle                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("sca_term");

  /***************************************************************************/
  /* Check we're initialized!                                                */
  /***************************************************************************/
  if (!sca_initialized)
  {
    TRC_ERROR((TB, "AR Term called when not init"));
    rc = OSP_ERR_NOT_INIT;
  }

  /***************************************************************************/
  /* Deregister the callbacks                                                */
  /***************************************************************************/
  rc = osp_ar_register(NULL, NULL, NULL, NULL, NULL);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We absolutely should NOT get a bad parameter error; otherwise just    */
    /* trace the result.                                                     */
    /*************************************************************************/
    TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
    TRC_ERROR((TB, "Failure registering AR callbacks, rc %d", rc));
    goto EXIT_LABEL;
  }

  rc = osp_sa_register(NULL);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We absolutely should NOT get a bad parameter error; otherwise just    */
    /* trace the result.                                                     */
    /*************************************************************************/
    TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
    TRC_ERROR((TB, "Failure registering SA callback, rc %d", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Destroy the trans count mutex                                           */
  /***************************************************************************/
  osp_destroy_mutex(&sca_trans_count_mutex);

  /***************************************************************************/
  /* Stop using lam                                                          */
  /***************************************************************************/
  lam_close(&sca_lam_handle);

  /***************************************************************************/
  /* Free the certificate chain                                              */
  /***************************************************************************/
  ccm_free_cert_chain(sca_token_signing_cert_list);

  /***************************************************************************/
  /* Free memory pointed to by globals.                                      */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing sca globals"));
  rm_release_mem(sca_signing_dn);
  rm_release_mem(sca_auth_duration);
  rm_release_mem(sca_service_url);
  sca_signing_dn = NULL;
  sca_auth_duration = NULL;
  sca_service_url = NULL;

  /***************************************************************************/
  /* Now we're done                                                          */
  /***************************************************************************/
  TRC_NORMAL((TB, "AR App terminated"));
  sca_initialized = FALSE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sca_term */

