
/**MOD+***********************************************************************/
/* Module:    scaprcb.c                                                      */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Pricing Indication processing                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.5                $ $Modtime::   Jan 11 2001 15:56:18   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_pricing_ind_callback                                       */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/*                                                                           */
/* Operation: Ensure price is in US Dollars for a time interval              */
/*            Check source of type transport                                 */
/*            Check dest of type e164                                        */
/*            Check source gateway is in directory and has a route to dest   */
/*            Update pricing                                                 */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI sca_pricing_ind_callback(OSP_CORRELATOR corr,
                                       char *timestamp,
                                       OSP_ADDRESS *source_info,
                                       OSP_ADDRESS *dest_info,
                                       char *currency,
                                       char *amount,
                                       char *increment,
                                       char *unit,
                                       OSP_SERVICE_INFO *service,
                                       char *valid_after,
                                       char *valid_until,
                                       OSP_CLIENT_ID *client_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_STATUS status;
  OSP_RC rc = OSP_SUCCESS;
  int send_immediate;
  int numeric_increment;
  int numeric_amount;
  int numeric_price_tenth_cents;
  char *new_price;

  TRC_ENTRY("sca_pricing_ind_callback");

  /***************************************************************************/
  /* Check parameters                                                        */
  /***************************************************************************/
  TRC_ASSERT((corr != NULL), (TB, "Passed NULL correlator"));
  TRC_ASSERT((timestamp != NULL), (TB, "Passed NULL timestamp"));
  TRC_ASSERT((source_info != NULL), (TB, "Passed NULL source address"));
  TRC_ASSERT((dest_info != NULL), (TB, "Passed NULL dest address"));
  TRC_ASSERT((currency != NULL), (TB, "Passed NULL currency"));
  TRC_ASSERT((amount != NULL), (TB, "Passed NULL amount"));
  TRC_ASSERT((increment != NULL), (TB, "Passed NULL increment"));
  TRC_ASSERT((service != NULL), (TB, "Passed NULL service"));
  TRC_ASSERT((valid_after != NULL), (TB, "Passed NULL valid_after"));
  TRC_ASSERT((valid_until != NULL), (TB, "Passed NULL valid_until"));

  TRC_DETAIL((TB, "Pricing Ind for source '%s' type '%s'",
                                     source_info->address, source_info->type));
  TRC_DETAIL((TB, "Destination address '%s' type '%s'",
                                         dest_info->address, dest_info->type));

  /***************************************************************************/
  /* Initialize variables                                                    */
  /***************************************************************************/
  send_immediate = FALSE;
  new_price = NULL;

  /***************************************************************************/
  /* Set up the default status                                               */
  /***************************************************************************/
  status.code = OSP_STATUS_SUCCESS;
  status.desc = sca_pricing_success;

  /***************************************************************************/
  /* For the sample application, pricing must be in US Dollars per number of */
  /* seconds - we send a failure if not                                      */
  /***************************************************************************/
  if ((strcmp(currency, sca_dollars) != 0) || (strcmp(unit, sca_seconds) != 0))
  {
    TRC_ERROR((TB, "Pricing in '%s' per n '%s', not dollars/sec", currency,
                                                                        unit));
    status.code = OSP_STATUS_NOT_IMPLEMENTED;
    status.desc = (char *)sca_usd_only;
    send_immediate = TRUE;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Similarly, the destination address must be of type e164                 */
  /***************************************************************************/
  if (strcmp(dest_info->type, sca_e164_attr) != 0)
  {
    TRC_ALERT((TB, "Got non-e164 address for dest, type '%s'",
                                                             dest_info->type));
    status.code = OSP_STATUS_NOT_IMPLEMENTED;
    status.desc = (char *)sca_e164_only;
    send_immediate = TRUE;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We only handle updates for source addresses of type transport - send    */
  /* success but don't actually do anything if not                           */
  /***************************************************************************/
  if (strcmp(source_info->type, sca_transport) != 0)
  {
    TRC_ALERT((TB, "Got non-transport address for source, type '%s'",
                                                           source_info->type));
    send_immediate = TRUE;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Convert the pricing to dollars/minute, via 1/10 cents/minute to ensure  */
  /* we maintain the precision we'll use in the sort-on-pricing algorithm    */
  /***************************************************************************/
  TRC_DETAIL((TB, "Supplied price $%s per %s %s", amount, increment, unit));
  numeric_increment = atoi(increment);
  if (numeric_increment <= 0)
  {
    TRC_ALERT((TB, "Invalid increment: %s, %d", increment, numeric_increment));
    status.code = OSP_STATUS_BAD_REQUEST;
    status.desc = (char *)sca_bad_request;
    send_immediate = TRUE;
    goto EXIT_LABEL;
  }

  numeric_amount = sca_price_to_int(amount);
  numeric_price_tenth_cents = (numeric_amount * 60) / numeric_increment;

  new_price = sca_int_to_price(numeric_price_tenth_cents);
  if (new_price == NULL)
  {
    TRC_ERROR((TB, "Failed to convert price"));
    status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status.desc = (char *)scm_subs_resource_failure;
    send_immediate = TRUE;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Price to store : %s dollars/minute", new_price));

  /***************************************************************************/
  /* Now kick off the lookup on the source to confirm that it's a gateway we */
  /* know about                                                              */
  /***************************************************************************/
  rc = sca_start_pricing_update(corr, source_info, dest_info, new_price,
                                                     valid_after, valid_until);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to start pricing update operation"));
    status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status.desc = (char *)scm_subs_resource_failure;
    send_immediate = TRUE;
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Pricing update started."));

EXIT_LABEL:
  /***************************************************************************/
  /* Do we need to send a response straightaway?                             */
  /***************************************************************************/
  if (send_immediate)
  {
    TRC_NORMAL((TB, "Immediate Pricing Confirm, code %d", status.code));
    rc = osp_ar_pricing_confirm(corr,
                                &status,
                                scm_sign_responses);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We absolutely should NOT get a bad parameter error; otherwise just  */
      /* trace the result.                                                   */
      /***********************************************************************/
      TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
      TRC_ERROR((TB, "Price confirm call failed, rc = %d", rc));
    }
  }

  /***************************************************************************/
  /* Free resources                                                          */
  /***************************************************************************/
  rm_release_mem(new_price);

  TRC_EXIT();

  return(rc);

} /* sca_pricing_ind_callback */

/**PROC+**********************************************************************/
/* Name:      sca_start_pricing_update                                       */
/*                                                                           */
/* Purpose:   Initiate the directory search for supplied gateway and         */
/*            destination address                                            */
/*                                                                           */
/* Returns:   OSP_SUCCESS - the search was initiated successfully            */
/*            other - OSP failure code                                       */
/*                                                                           */
/* Params:    IN     source       - gateway pricing is for                   */
/*            IN     dest         - e164 destination address                 */
/*            IN     price        - price for destination                    */
/*            IN     valid_from   - time pricing valid from                  */
/*            IN     valid_until  - time pricing valid until                 */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_start_pricing_update(OSP_CORRELATOR corr,
                                OSP_ADDRESS *source,
                                OSP_ADDRESS *dest,
                                char *price,
                                char *valid_from,
                                char *valid_until)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  char filter[256];
  int filter_len;
  SCA_PRICE_DATA *price_data;

  TRC_ENTRY("sca_start_pricing_update");

  /***************************************************************************/
  /* Save off all the details                                                */
  /***************************************************************************/
  price_data = rm_get_mem(sizeof(SCA_PRICE_DATA));
  if (price_data == NULL)
  {
    TRC_ERROR((TB, "Failed to get price data structure"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  price_data->corr = corr;
  price_data->source_info = sca_copy_address(source);
  price_data->dest_info = sca_copy_address(dest);
  OSP_STRDUP(price_data->price, price);
  OSP_STRDUP(price_data->valid_from, valid_from);
  OSP_STRDUP(price_data->valid_until, valid_until);

  if ((price_data->source_info == NULL) ||
      (price_data->dest_info == NULL) ||
      (price_data->price == NULL) ||
      (price_data->valid_from == NULL) ||
      (price_data->valid_until == NULL))
  {
    TRC_ERROR((TB, "Failed to copy price data structure"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Build the search string                                                 */
  /***************************************************************************/
  filter_len = OSP_SNPRINTF((filter,
                             sizeof(filter),
                             "(&(%s=%s)(%s=%s))",
                             sca_ip_address,
                             source->address,
                             sca_e164_type,
                             dest->address));
  TRC_ASSERT((filter_len < sizeof(filter)),
             (TB, "Resulting filter would be too large at %d bytes",
              filter_len));

  TRC_DETAIL((TB, "Search filter is '%s'", filter));

  /***************************************************************************/
  /* Issue the search request                                                */
  /***************************************************************************/
  rc = lam_search(sca_lam_handle,
                  (char *)sca_ar_base,
                  LDAP_SCOPE_SUBTREE,
                  filter,
                  (char **)sca_auth_rq_attrs,
                  FALSE,
                  (void *)price_data,
                  sca_price_search_callback);

  if (rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* now we wait for the LAM callback, where we'll send the response       */
    /*************************************************************************/
    TRC_DETAIL((TB, "Search initiated successfully"));
  }
  else
  {
    TRC_ERROR((TB, "Search operation failed with rc %d", rc));
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Do any tidy up                                                          */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    TRC_DETAIL((TB, "Freeing price update memory"));
    sca_free_price(price_data);
  }

  TRC_EXIT();

  return(rc);

} /* sca_start_pricing_update */

/**PROC+**********************************************************************/
/* Name:      sca_price_search_callback                                      */
/*                                                                           */
/* Purpose:   Called by LAM on search completion                             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - userdata supplied to lam_search/modify   */
/*            IN     ld           - LDAP handle required to parse result     */
/*            IN     result       - results of the search operation.         */
/*                                                                           */
/**PROC-**********************************************************************/

void OSPAPI sca_price_search_callback(void *user_data,
                                      LDAP *ld,
                                      LDAPMessage **result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  SCA_PRICE_DATA *price_data;
  int result_count;
  OSP_STATUS status;
  char *dn;
  LDAPMod add_mod;
  LDAPMod del_mod;
  LDAPMod *mods[3];
  int mod_count;
  char *add_values[2];
  char *del_values[2];
  int ldap_rc;
  char pricing_string[256];
  char *old_price;

  TRC_ENTRY("sca_price_search_callback");

  /***************************************************************************/
  /* Initialize variables                                                    */
  /***************************************************************************/
  price_data = NULL;
  old_price = NULL;

  /***************************************************************************/
  /* Extract the price data from the user_data.  Without it there's nothing  */
  /* we can do - we don't even have the correlator required for the callback */
  /* into the AIB - hence the assert                                         */
  /***************************************************************************/
  price_data = (SCA_PRICE_DATA *)user_data;
  TRC_ASSERT((price_data != NULL), (TB, "Received NULL price data in LAM cb"));

  /***************************************************************************/
  /* Check we got some results to look at                                    */
  /***************************************************************************/
  if ((ld == NULL) || (result == NULL))
  {
    TRC_ALERT((TB, "Got bad result from search"));
    status.code = OSP_STATUS_ROUTE_AUTHORIZATION_UNSUCCESSFUL;
    status.desc = (char *)sca_route_not_found;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up default values for the status field                              */
  /***************************************************************************/
  status.code = OSP_STATUS_SUCCESS;
  status.desc = sca_pricing_success;

  /***************************************************************************/
  /* See if we got any destinations back                                     */
  /***************************************************************************/
  result_count = ldap_count_entries(ld, *result);
  TRC_DETAIL((TB, "Got %d results", result_count));
  TRC_ASSERT((result_count <= 1), (TB, "Got multiple matches for gateway"));
  if (result_count == 1)
  {
    /*************************************************************************/
    /* Got a match: we can go ahead and modify the pricing                   */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got match on gateway/destination combination"));

    /*************************************************************************/
    /* Get the dn                                                            */
    /*************************************************************************/
    dn = ldap_get_dn(ld, *result);
    if (dn == NULL)
    {
      /***********************************************************************/
      /* With no dn, there's nothing we can do                               */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to get dn - can't update price"));
      status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status.desc = (char *)scm_subs_resource_failure;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* See if we got an existing price for the route.  If we did, we need to */
    /* delete it                                                             */
    /*************************************************************************/
    TRC_DETAIL((TB, "Getting old pricing"));
    old_price = sca_find_price(price_data->dest_info->address, ld, *result);

    /*************************************************************************/
    /* Build the pricing string                                              */
    /*************************************************************************/
    OSP_SNPRINTF((pricing_string,
                  sizeof(pricing_string),
                  "%s@%s",
                  price_data->dest_info->address,
                  price_data->price));

    if (*(price_data->valid_from) != 0)
    {
      strcat(pricing_string, "F");
      TRC_ASSERT((strlen(price_data->valid_from)
                  + strlen(pricing_string) < sizeof(pricing_string)),
                  (TB, "Valid from doesn't fit in pricing string"));
      strcat(pricing_string, price_data->valid_from);

      if (*(price_data->valid_until) != 0)
      {
        strcat(pricing_string, "U");
        strcat(pricing_string, price_data->valid_until);
        TRC_ASSERT((strlen(price_data->valid_until)
                    + strlen(pricing_string) < sizeof(pricing_string)),
                    (TB, "Valid until doesn't fit in pricing string"));
      }
    }
    TRC_DETAIL((TB, "Revised price '%s'", pricing_string));

    /*************************************************************************/
    /* Build the modify operation, deleting the old pricing if necessary     */
    /*************************************************************************/
    mod_count = 0;
    if (old_price != NULL)
    {
      TRC_DETAIL((TB, "Building delete mods"));
      del_mod.mod_op = LDAP_MOD_DELETE;
      del_mod.mod_type = (char *)sca_pricing;
      del_values[0] = old_price;
      del_values[1] = NULL;
      del_mod.mod_values = del_values;
      mods[mod_count] = &del_mod;
      mod_count++;
    }

    TRC_DETAIL((TB, "Building add mods"));
    add_mod.mod_op = LDAP_MOD_ADD;
    add_mod.mod_type = (char *)sca_pricing;
    add_values[0] = pricing_string;
    add_values[1] = NULL;
    add_mod.mod_values = add_values;
    mods[mod_count] = &add_mod;
    mod_count++;
    mods[mod_count] = NULL;

    /*************************************************************************/
    /* And apply the modification                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Updating price..."));
    ldap_rc = ldap_modify_s(ld, dn, mods);
    if (ldap_rc == LDAP_SUCCESS)
    {
      TRC_DETAIL((TB, "price updated successfully"));
    }
    else
    {
      TRC_ERROR((TB, "LDAP failure %d updating price", ldap_rc));
      status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status.desc = (char *)scm_subs_resource_failure;
    }
  }
  else
  {
    /*************************************************************************/
    /* Set the status code and send an failure response                      */
    /*************************************************************************/
    TRC_ALERT((TB, "Search found no matching destinations"));
    status.code = OSP_STATUS_ROUTE_AUTHORIZATION_UNSUCCESSFUL;
    status.desc = (char *)sca_route_not_found;
  }

EXIT_LABEL:
  TRC_NORMAL((TB, "Pricing Confirm, code %d", status.code));
  rc = osp_ar_pricing_confirm(price_data->corr,
                              &status,
                              scm_sign_responses);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We absolutely should NOT get a bad parameter error; otherwise just    */
    /* trace the result.                                                     */
    /*************************************************************************/
    TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                       (TB, "Bad parameter errors unexpected from OSP api"));
    TRC_ERROR((TB, "Price confirm call failed, rc = %d", rc));
  }

  /***************************************************************************/
  /* Tidy up                                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing pricing memory"));
  sca_free_price(price_data);

  TRC_EXIT();

  return;

} /* sca_price_search_callback */

/**PROC+**********************************************************************/
/* Name:      sca_free_price                                                 */
/*                                                                           */
/* Purpose:   Free the supplied price structure                              */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     price        - the price information to be freed        */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_price(SCA_PRICE_DATA *price_data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sca_free_price");

  if (price_data != NULL)
  {
    /*************************************************************************/
    /* Free each element                                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing price data %p", price_data));
    sca_free_address(price_data->source_info);
    sca_free_address(price_data->dest_info);
    rm_release_mem(price_data->price);
    rm_release_mem(price_data->valid_from);
    rm_release_mem(price_data->valid_until);
    rm_release_mem(price_data);
  }

  TRC_EXIT();

  return;

} /* sca_free_price */

/**PROC+**********************************************************************/
/* Name:      sca_find_price                                                 */
/*                                                                           */
/* Purpose:   Find the price for a given destination given an LDAP search    */
/*            result.                                                        */
/*                                                                           */
/* Returns:   pointer to price, or NULL if not found.                        */
/*                                                                           */
/* Params:    IN     destination  - destination to find price for            */
/*            IN     ld           - ldap handle                              */
/*            IN     result       - search result to examine                 */
/*                                                                           */
/* Operation: Search the results for the pricing attribute.                  */
/*            For each pricing attribute, check the destination              */
/*                                                                           */
/**PROC-**********************************************************************/

char *sca_find_price(char *destination, LDAP *ld, LDAPMessage *result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char *price = NULL;
  char *curr_attr;
  char **attr_vals;
  char **matches = NULL;
  BerElement *ber = NULL;
  int ii;
  int len;

  TRC_ENTRY("sca_find_price");

  TRC_DETAIL((TB, "Finding old price for destination %s", destination));

  /***************************************************************************/
  /* Walk the attribute list                                                 */
  /***************************************************************************/
  for (curr_attr = ldap_first_attribute(ld, result, &ber);
       (curr_attr != NULL) && (price == NULL);
       curr_attr = ldap_next_attribute(ld, result, ber))
  {
    TRC_DETAIL((TB, "Processing attribute '%s'", curr_attr));

    /*************************************************************************/
    /* See if we've got the pricing attribute                                */
    /*************************************************************************/
    if (strcmp(curr_attr, sca_pricing) == 0)
    {
      /***********************************************************************/
      /* Now get the values for it                                           */
      /***********************************************************************/
      TRC_DETAIL((TB, "Found pricing"));
      attr_vals = ldap_get_values(ld, result, curr_attr);
      TRC_ASSERT((attr_vals != NULL),
                                (TB, "Attribute %s has no values", curr_attr));

      /***********************************************************************/
      /* Examine each attribute in turn                                      */
      /***********************************************************************/
      for (ii = 0; attr_vals[ii] != NULL; ii++)
      {
        /*********************************************************************/
        /* The pricing info is in the form                                   */
        /*                                                                   */
        /*     pricing=44181@3.0F2000-05-31T18:00:00ZU2000-06-01T08:00:00Z   */
        /*                                                                   */
        /* which means 'calls to numbers beginning 44181 cost 3 dollars per  */
        /* minute from 6pm on May 31 until 8am on June 1'.                   */
        /*                                                                   */
        /* We want to find a price for the supplied destination, so we can   */
        /* just compare the two directly                                     */
        /*********************************************************************/
        TRC_DETAIL((TB, "Got number@price %s", attr_vals[ii]));
        len = OSP_MIN(strlen(attr_vals[ii]), strlen(destination));
        if (strncmp(attr_vals[ii], destination, len) == 0)
        {
          TRC_DETAIL((TB, "Found old price"));
          price = attr_vals[ii];
          break;
        }
      }
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return(price);

} /* sca_find_price */
