
/**MOD+***********************************************************************/
/* Module:    scaroute.c                                                     */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Route generation code                                          */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.18               $ $Modtime::   Jan 11 2001 15:56:18   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_do_routing                                                 */
/*                                                                           */
/* Purpose:   Initiate the search phase of routing                           */
/*                                                                           */
/* Returns:   OSP_SUCCESS - search started successfully                      */
/*            other       - a failure occured                                */
/*                                                                           */
/* Params:    IN     corr         - correlator to be used on response        */
/*            IN     request_type - auth request or reauth                   */
/*            IN     call_id_list - list of call identifiers to use for      */
/*                                  returned destinations                    */
/*            IN     source_info  - pointer to source info structure         */
/*            IN     dest_info    - pointer to destination info structure    */
/*            IN     dest_alt_list- destination alternate list               */
/*            IN     service      - requested service information            */
/*            IN     max_dests    - maximum number of destinations to return */
/*                                                                           */
/* Operation: allocate memory to store the parameters required in building   */
/*            the response, and issue an LDAP search for gateways able to    */
/*            terminate the call                                             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_do_routing(OSP_CORRELATOR corr,
                      int request_type,
                      OSP_CALL_ID *call_id_list,
                      OSP_ADDRESS *source_info,
                      OSP_ADDRESS *dest_info,
                      OSP_ADDRESS *dest_alt_list,
                      OSP_SERVICE_INFO *service,
                      char *max_dests)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERR_MEMORY;
  SCA_ROUTE_DATA *route_data;
  char *filter;
  SLIST_ENTRY *curr_entry;

  TRC_ENTRY("sca_do_routing");

  /***************************************************************************/
  /* Initialize variables                                                    */
  /***************************************************************************/
  route_data = NULL;
  filter = NULL;

  /***************************************************************************/
  /* Store off the parameters                                                */
  /***************************************************************************/
  route_data = sca_store_routing_params(corr, request_type, call_id_list,
                             source_info, dest_info, dest_alt_list, max_dests);
  if (route_data == NULL)
  {
    TRC_ERROR((TB, "failed to store routing params"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* build the LAM query filter using the destination address and any        */
  /* service requirements                                                    */
  /***************************************************************************/
  filter = sca_route_filter(dest_info->address, service);
  if (filter == NULL)
  {
    TRC_ERROR((TB, "failed to build search filter"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Pass the filter onto LAM                                                */
  /***************************************************************************/
  rc = lam_search(sca_lam_handle,
                  (char *)sca_ar_base,
                  LDAP_SCOPE_SUBTREE,
                  filter,
                  (char **)sca_auth_rq_attrs,
                  FALSE,
                  (void *)route_data,
                  sca_search_callback);
  if (rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* now we wait for the LAM callback, where we'll send the response       */
    /*************************************************************************/
    TRC_DETAIL((TB, "Routing search initiated successfully"));
  }
  else
  {
    /*************************************************************************/
    /* Free any resources then return the failure                            */
    /*************************************************************************/
    TRC_ERROR((TB, "Search operation failed with rc %d", rc));
    sca_free_route_data(route_data);
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Tidy up                                                                 */
  /*                                                                         */
  /* If we didn't allocate the filter, then we need to free up the route     */
  /* data                                                                    */
  /***************************************************************************/
  if (filter == NULL)
  {
    TRC_DETAIL((TB, "Freeing route data"));
    sca_free_route_data(route_data);
  }
  else
  {
    TRC_DETAIL((TB, "Freeing filter"));
    rm_release_mem(filter);
  }

  TRC_EXIT();

  return(rc);

} /* sca_do_routing */

/**PROC+**********************************************************************/
/* Name:      sca_search_callback                                            */
/*                                                                           */
/* Purpose:   Called by LAM on search completion                             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - userdata supplied to lam_search/modify   */
/*            IN     ld           - LDAP handle required to parse result     */
/*            IN     result       - results of the search operation.         */
/*                                                                           */
/* Operation: Sort the list of destinations returned by the LDAP search      */
/*            and convert it to a full OSP destination list                  */
/*                                                                           */
/**PROC-**********************************************************************/

void OSPAPI sca_search_callback(void *user_data,
                                LDAP *ld,
                                LDAPMessage **result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  PSCA_ROUTE_DATA route_data;
  int result_count;
  LIST_ENTRY internal_destination_list;
  OSP_DESTINATION *final_destination_list;
  OSP_STATUS status;
  char transaction_id[SCA_TRANS_ID_LEN];

  TRC_ENTRY("sca_search_callback");

  /***************************************************************************/
  /* Extract the ar data from the user_data.  Without it there's nothing we  */
  /* can do - we don't even have the correlator required for the callback    */
  /* into the AIB - hence the assert                                         */
  /***************************************************************************/
  route_data = (PSCA_ROUTE_DATA)user_data;
  TRC_ASSERT((route_data != NULL), (TB, "Received NULL ar data in LAM c/b"));

  /***************************************************************************/
  /* Set up default values for the status field                              */
  /***************************************************************************/
  status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
  status.desc = (char *)scm_subs_resource_failure;

  /***************************************************************************/
  /* We've got no destinations to start with                                 */
  /***************************************************************************/
  final_destination_list = NULL;

  /***************************************************************************/
  /* Get a transaction id for this authorization                             */
  /***************************************************************************/
  sca_get_trans_id(transaction_id);

  /***************************************************************************/
  /* Check we got something                                                  */
  /***************************************************************************/
  if ((ld == NULL) || (result == NULL))
  {
    TRC_ALERT((TB, "No route found due to directory server problem"));
    status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status.desc = (char *)scm_subs_server_failure;
  }
  else
  {
    /*************************************************************************/
    /* See if we got any destinations back                                   */
    /*************************************************************************/
    result_count = ldap_count_entries(ld, *result);
    TRC_DETAIL((TB, "Got %d results", result_count));
    if (result_count != 0)
    {
      /***********************************************************************/
      /* Build a list of destinations in our internal format from the LDAP   */
      /* search results                                                      */
      /***********************************************************************/
      TRC_DETAIL((TB, "Building unsorted list"));
      status.code = sca_build_int_dest_list(route_data,
                                            &internal_destination_list,
                                            ld,
                                            result);
      if (status.code == OSP_STATUS_SUCCESS)
      {
        /*********************************************************************/
        /* We may get here with an empty list                                */
        /*********************************************************************/
        if (LIST_IS_EMPTY(&internal_destination_list))
        {
          TRC_DETAIL((TB, "List empty - no routes in dest alternates"));
          status.code = OSP_STATUS_ROUTE_AUTHORIZATION_UNSUCCESSFUL;
          status.desc = (char *)sca_route_not_found;
        }
        else
        {
          /*******************************************************************/
          /* Now convert that list to the final format                       */
          /*******************************************************************/
          TRC_DETAIL((TB, "Building sorted list"));
          status.code = sca_build_final_list(route_data,
                                             transaction_id,
                                             &internal_destination_list,
                                             &final_destination_list);
          if (status.code == OSP_STATUS_SUCCESS)
          {
            TRC_DETAIL((TB, "Sorted list successfully"));
            status.desc = (char *)scm_success;
          }
        }
      }
    }
    else
    {
      /***********************************************************************/
      /* Set the status code and send an empty response                      */
      /***********************************************************************/
      TRC_ALERT((TB, "Search found no matching destinations"));
      status.code = OSP_STATUS_ROUTE_AUTHORIZATION_UNSUCCESSFUL;
      status.desc = (char *)sca_route_not_found;
    }
  }

  /***************************************************************************/
  /* Pass the result back down the stack                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Calling into OSP - good return code"));
  if (route_data->request_type == SCA_ROUTE_AUTH_RQ)
  {
    TRC_DETAIL((TB, "Auth Rsp"));
    rc = osp_ar_auth_response(route_data->corr,
                              &status,
                              transaction_id,
                              NULL,
                              final_destination_list,
                              scm_sign_responses);
  }
  else if (route_data->request_type == SCA_ROUTE_REAUTH_RQ)
  {
    TRC_DETAIL((TB, "Reauth Rsp"));
    rc = osp_ar_reauth_response(route_data->corr,
                                &status,
                                transaction_id,
                                NULL,
                                final_destination_list,
                                scm_sign_responses);
  }
  else
  {
    TRC_ASSERT((FALSE), (TB, "Unknown request type %d",
                                                    route_data->request_type));
  }

  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We absolutely should NOT get a bad parameter error; otherwise just    */
    /* trace the result.                                                     */
    /*************************************************************************/
    TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
    TRC_ERROR((TB, "OSP (re)auth rsp call failed, rc = %d", rc));
  }

  /***************************************************************************/
  /* All done - just free our resources                                      */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing route request memory"));
  sca_free_route_data(route_data);

  /***************************************************************************/
  /* Free the up all the memory associated with the destination list         */
  /***************************************************************************/
  if (final_destination_list != NULL)
  {
    TRC_DETAIL((TB, "Free destination list"));
    scm_free_list((SLIST_ENTRY *)final_destination_list,
                                               (PFREE_ENTRY_FN)&sca_free_dest);
  }

  TRC_EXIT();

  return;

} /* sca_search_callback */

