
/**MOD+***********************************************************************/
/* Module:    scarteut.c                                                     */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Route generation utilities                                     */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.10               $ $Modtime::   Jan 11 2001 15:56:18   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_build_int_dest_list                                        */
/*                                                                           */
/* Purpose:   Build a list of destinations in our internal format from the   */
/*            raw LDAP result                                                */
/*                                                                           */
/* Returns:   OSP status code for the operation                              */
/*                                                                           */
/* Params:    IN     route_data   - stored auth request parameters           */
/*            OUT    int_dest_list                                           */
/*                                - head of returned list                    */
/*            IN     ld           - LDAP handle required to parse result     */
/*            IN     result       - results of the search operation.         */
/*                                                                           */
/* Operation: For each returned gateway,                                     */
/*              extract the best match destination                           */
/*              add it to the right sub list                                 */
/*            end for                                                        */
/*            merge the sublists to form the full list                       */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_build_int_dest_list(PSCA_ROUTE_DATA route_data,
                            LIST_ENTRY *int_dest_list,
                            LDAP *ld,
                            LDAPMessage **result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  unsigned int status = OSP_STATUS_SUCCESS;
  LDAPMessage *curr_result;
  SCA_ROUTE_OPTION *curr_dest;
  SCA_ROUTE_OPTION *new_dest;
  LIST_ENTRY not_oor_list;
  LIST_ENTRY oor_list;

  TRC_ENTRY("sca_build_int_dest_list");

  /***************************************************************************/
  /* Initialize the lists                                                    */
  /***************************************************************************/
  LIST_INIT(int_dest_list);
  LIST_INIT(&not_oor_list);
  LIST_INIT(&oor_list);

  /***************************************************************************/
  /* use the LDAP message processing to extract the results                  */
  /***************************************************************************/
  for (curr_result = ldap_first_entry(ld, *result);
       curr_result != NULL;
       curr_result = ldap_next_entry(ld, curr_result))
  {
    /*************************************************************************/
    /* Convert the result into our minimal destination format                */
    /*************************************************************************/
    TRC_DETAIL((TB, "Processing result %p", curr_result));

    new_dest = sca_make_sca_destination(route_data, ld, curr_result);
    if (new_dest != NULL)
    {
      /***********************************************************************/
      /* Add the new destination to the appropriate sublist                  */
      /***********************************************************************/
      TRC_DETAIL((TB, "Adding new dest %p", new_dest));
      sca_add_to_sub_list(route_data, new_dest, &not_oor_list, &oor_list);
    }
    else
    {
      TRC_ERROR((TB, "Failed to allocate new destination"));
      status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      break;
    }
  }

  /***************************************************************************/
  /* Now merge the sublists to create the internal list                      */
  /***************************************************************************/
  TRC_DETAIL((TB, "Merging sorted lists"));
  while (!LIST_IS_EMPTY(&not_oor_list))
  {
    curr_dest = (SCA_ROUTE_OPTION *)LIST_REMOVE_HEAD(&not_oor_list);
    LIST_INSERT_TAIL(int_dest_list, &curr_dest->links);
  }

  while (!LIST_IS_EMPTY(&oor_list))
  {
    curr_dest = (SCA_ROUTE_OPTION *)LIST_REMOVE_HEAD(&oor_list);
    LIST_INSERT_TAIL(int_dest_list, &curr_dest->links);
  }

EXIT_LABEL:

  /***************************************************************************/
  /* If something went wrong, we must free anything we did allocate          */
  /***************************************************************************/
  if (status != OSP_STATUS_SUCCESS)
  {
    TRC_NORMAL((TB, "Allocation failure - freeing what we got"));
    while (!LIST_IS_EMPTY(int_dest_list))
    {
      curr_dest = (SCA_ROUTE_OPTION *)LIST_REMOVE_HEAD(int_dest_list);
      rm_release_mem(curr_dest->ip_address);
      rm_release_mem(curr_dest);
    }
  }

  TRC_EXIT();

  return(status);

} /* sca_build_int_dest_list */

/**PROC+**********************************************************************/
/* Name:      sca_add_to_sub_list                                            */
/*                                                                           */
/* Purpose:   Add the supplied item to the appropriate sublist               */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     route_data   - stored auth request parameters           */
/*            IN     new_dest     - the destination to add to the lists      */
/*            IN     not_oor_list - list to store destinations which are     */
/*                                  not almost out of resources              */
/*            IN     oor_list     - list to store destinations which are     */
/*                                  almost out of resources                  */
/*                                                                           */
/* Operation: If any destination alternates were supplied, then we just drop */
/*            any items not in that list.  Otherwise, assign the new item to */
/*            one of the sublists, OOR or not OOR.                           */
/*                                                                           */
/*            Note that the add operation sorts the sublist by price then    */
/*            match length                                                   */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_add_to_sub_list(PSCA_ROUTE_DATA route_data,
                         SCA_ROUTE_OPTION *new_dest,
                         LIST_ENTRY *not_oor_list,
                         LIST_ENTRY *oor_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sca_add_to_sub_list");

  /***************************************************************************/
  /* If any destination alternates are provided, we drop any destinations    */
  /* that aren't in that list                                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Looking at dest gw %s", new_dest->ip_address));
  if ((route_data->dest_alt_list == NULL) ||
      (sca_dest_in_alt_list(new_dest->ip_address, route_data->dest_alt_list)))
  {
    /*************************************************************************/
    /* Split the destinations between not oor and oor                        */
    /*************************************************************************/
    if (!new_dest->almost_out)
    {
      TRC_DETAIL((TB, "not oor"));
      sca_add_in_sequence(not_oor_list, new_dest);
    }
    else
    {
      TRC_DETAIL((TB, "oor"));
      sca_add_in_sequence(oor_list, new_dest);
    }
  }
  else
  {
    TRC_DETAIL((TB, "Discarding"));
    rm_release_mem(new_dest->ip_address);
    rm_release_mem(new_dest);
  }

  TRC_EXIT();

  return;

} /* sca_add_to_sub_list */

/**PROC+**********************************************************************/
/* Name:      sca_build_final_list                                           */
/*                                                                           */
/* Purpose:   Convert the supplied list to the required number of            */
/*            OSP_DESTINATIONs, and free the entire list                     */
/*                                                                           */
/* Returns:   OSP status code for the operation                              */
/*                                                                           */
/* Params:    IN     route_data   - stored auth request parameters           */
/*            IN     int_dest_list                                           */
/*                                - the list of internal format destinations */
/*            IN/OUT final_destination_list                                  */
/*                                - pointer to start of final list           */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_build_final_list(PSCA_ROUTE_DATA route_data,
                         char *transaction_id,
                         LIST_ENTRY *int_dest_list,
                         OSP_DESTINATION **final_destination_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  unsigned int status = OSP_STATUS_SUCCESS;
  int dest_count;
  SCA_ROUTE_OPTION *curr_dest;
  SLIST_ROOT final_list;
  OSP_DESTINATION *new_osp_dest;
  OSP_CALL_ID *curr_call_id;

  TRC_ENTRY("sca_build_final_list");

  /***************************************************************************/
  /* Convert the minimal destinations in the sorted list into full osp       */
  /* destinations in the final list.                                         */
  /***************************************************************************/
  SL_INIT_ROOT(final_list);
  dest_count = 0;
  curr_call_id = route_data->call_id_list;

  TRC_DETAIL((TB, "Building list of OSP_DESTINATIONs"));
  curr_dest = (SCA_ROUTE_OPTION *)LIST_GET_FIRST(int_dest_list);
  while ((curr_dest != NULL) && (dest_count < route_data->max_dests))
  {
    /*************************************************************************/
    /* Remove the destination from the list                                  */
    /*************************************************************************/
    LIST_REMOVE_ITEM(&curr_dest->links);

    /*************************************************************************/
    /* Convert the item to an OSP_DESTINATION                                */
    /*************************************************************************/
    TRC_DETAIL((TB, "Converting destination %p", curr_dest));
    new_osp_dest = sca_make_osp_destination(route_data,
                                            transaction_id,
                                            curr_dest,
                                            curr_call_id);

    /*************************************************************************/
    /* Free the memory for the now redundant destination                     */
    /*************************************************************************/
    rm_release_mem(curr_dest->ip_address);
    rm_release_mem(curr_dest);

    /*************************************************************************/
    /* Check the conversion worked                                           */
    /*************************************************************************/
    if (new_osp_dest != NULL)
    {
      /***********************************************************************/
      /* Add it to the sorted list                                           */
      /***********************************************************************/
      dest_count++;
      SL_ADD_TO_TAIL(final_list, new_osp_dest);

      /***********************************************************************/
      /* We may have been supplied with be either just one call id or one    */
      /* per destination up to max_dests.  That this is so will already have */
      /* been checked by the AIB: we just need to move onto the next one if  */
      /* its present                                                         */
      /***********************************************************************/
      if (curr_call_id->next != NULL)
      {
        curr_call_id = (OSP_CALL_ID *)curr_call_id->next;
        TRC_DETAIL((TB, "New call id"));
      }
    }
    else
    {
      /***********************************************************************/
      /* Failed to do the conversion.                                        */
      /***********************************************************************/
      TRC_ERROR((TB, "Conversion to OSP dest failed"));
      status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      break;
    }

    /*************************************************************************/
    /* Get the next destination                                              */
    /*************************************************************************/
    curr_dest = (SCA_ROUTE_OPTION *)LIST_GET_FIRST(int_dest_list);
  }

  /***************************************************************************/
  /* Now free up the memory for any remaining destinations                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing remaining internal destinations"));
  while (curr_dest != NULL)
  {
    LIST_REMOVE_ITEM(&curr_dest->links);
    rm_release_mem(curr_dest->ip_address);
    rm_release_mem(curr_dest);

    curr_dest = (SCA_ROUTE_OPTION *)LIST_GET_FIRST(int_dest_list);
  }

  /***************************************************************************/
  /* Lastly set up the returned pointer to the sorted list                   */
  /***************************************************************************/
  *final_destination_list = (OSP_DESTINATION *)final_list.first;

  TRC_EXIT();

  return(status);

} /* sca_build_final_list  */

/**PROC+**********************************************************************/
/* Name:      sca_make_osp_destination                                       */
/*                                                                           */
/* Purpose:   Convert a destination stored in the sample app's minimal       */
/*            format to a full OSP_DESTINATION                               */
/*                                                                           */
/* Returns:   pointer to the new destination                                 */
/*                                                                           */
/* Params:    IN     route_data   - stored auth request parameters           */
/*            IN     dest         - destination to convert                   */
/*            IN     call_id      - H.323 call id to use for destination     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_DESTINATION *sca_make_osp_destination(PSCA_ROUTE_DATA route_data,
                                          char *transaction_id,
                                          SCA_ROUTE_OPTION *dest,
                                          OSP_CALL_ID *call_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_DESTINATION *osp_dest = NULL;
  int success;

  TRC_ENTRY("sca_make_osp_destination");

  /***************************************************************************/
  /* Initialize the flag we use to determine if something goes wrong along   */
  /* the way                                                                 */
  /***************************************************************************/
  success = FALSE;

  /***************************************************************************/
  /* Get memory for the destination                                          */
  /***************************************************************************/
  TRC_DETAIL((TB, "Get space for new destination"));
  osp_dest = rm_get_mem(sizeof(OSP_DESTINATION));
  if (osp_dest != NULL)
  {
    /*************************************************************************/
    /* Initialize the structure                                              */
    /*************************************************************************/
    TRC_DETAIL((TB, "Building destination"));
    osp_dest->next = NULL;
    osp_dest->info = NULL;
    osp_dest->alt_list = NULL;
    osp_dest->signal_address = NULL;
    osp_dest->token_list = NULL;
    osp_dest->valid_after = NULL;
    osp_dest->valid_until = NULL;
    osp_dest->usage_list = NULL;
    osp_dest->auth_url_list = NULL;
    osp_dest->call_id = NULL;

    /*************************************************************************/
    /* Set up the destination address                                        */
    /*************************************************************************/
    osp_dest->info = sca_copy_address(route_data->dest_info);
    if (osp_dest->info == NULL)
    {
      TRC_ERROR((TB, "Copy of dest address failed"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* We don't supply any alternate addresses                               */
    /*************************************************************************/
    osp_dest->alt_list = NULL;

    /*************************************************************************/
    /* Copy the call signalling address of the destination                   */
    /*************************************************************************/
    OSP_STRDUP(osp_dest->signal_address, dest->ip_address);
    if (osp_dest->signal_address == NULL)
    {
      TRC_ERROR((TB, "Sig address alloc failed"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Set up the authorization tokens (in fact we only ever supply one).    */
    /*                                                                       */
    /* We want the validity and duration in the destination to match that in */
    /* the token, but as the returned token is base64 encoded, its a lot     */
    /* easier to get the token creation code to fill in those fields for us! */
    /*************************************************************************/
    osp_dest->token_list = sca_create_token(route_data,
                                            transaction_id,
                                            call_id,
                                            osp_dest);
    if (osp_dest->token_list == NULL)
    {
      TRC_ERROR((TB, "Token create failed"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* No authorization URLs                                                 */
    /*************************************************************************/
    osp_dest->auth_url_list = NULL;

    /*************************************************************************/
    /* Copy over the supplied call identifier                                */
    /*************************************************************************/
    osp_dest->call_id = sca_copy_call_id(call_id);
    if (osp_dest->call_id == NULL)
    {
      TRC_ERROR((TB, "Call id allocation failed"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* If we get here, all is well                                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Completed osp destination %p", osp_dest));
    success = TRUE;
  }
  else
  {
    TRC_ERROR((TB, "Failed to get memory for OSP destination"));
  }

EXIT_LABEL:
  /***************************************************************************/
  /* If something went wrong along the way, free any memory we did get       */
  /***************************************************************************/
  if (!success && (osp_dest != NULL))
  {
    TRC_ALERT((TB, "Freeing allocated fields"));
    sca_free_dest(osp_dest);
    osp_dest = NULL;
  }

  TRC_EXIT();

  return(osp_dest);

} /* sca_make_osp_destination */

/**PROC+**********************************************************************/
/* Name:      sca_make_sca_destination                                       */
/*                                                                           */
/* Purpose:   Convert an OSP result to our local minimal destination format  */
/*                                                                           */
/* Returns:   pointer to new destination                                     */
/*                                                                           */
/* Params:    IN     route_data   - stored auth request parameters           */
/*            IN     ld           - LDAP handle required to parse result     */
/*            IN     result       - result to convert                        */
/*                                                                           */
/* Operation: Allocate memory for the new destination                        */
/*            Parse the attributes supplied in the result to get             */
/*            - the ip address                                               */
/*            - the out of resources setting                                 */
/*            - the best price to the destination, or in the absence of any  */
/*              pricing, the best (longest) match on the destination address */
/*                                                                           */
/**PROC-**********************************************************************/

SCA_ROUTE_OPTION *sca_make_sca_destination(PSCA_ROUTE_DATA route_data,
                                           LDAP *ld,
                                           LDAPMessage *result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  SCA_ROUTE_OPTION *new_dest;
  char *curr_attr;
  char **attr_vals;
  char **matches = NULL;
  BerElement *ber = NULL;
  int ii;
  int valid_rc;
  char *sep;
  char *price;
  char *dn;
  unsigned int this_price;
  unsigned int this_match;
  LDAPMod mod;
  LDAPMod *mods[2] = {NULL};
  char *values[2] = {NULL};
  int ldap_rc;

  TRC_ENTRY("sca_make_sca_destination");

  /***************************************************************************/
  /* Get some memory for the destination                                     */
  /***************************************************************************/
  new_dest = (SCA_ROUTE_OPTION *)rm_get_mem(sizeof(SCA_ROUTE_OPTION));
  if (new_dest == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate memory for new route option"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize the new destination                                          */
  /***************************************************************************/
  new_dest->links.prev = NULL;
  new_dest->links.next = NULL;
  new_dest->ip_address = NULL;
  new_dest->almost_out = FALSE;
  strcpy(new_dest->matching_number, "<no match>");
  new_dest->best_match = 0;
  strcpy(new_dest->price, "<none>");
  new_dest->best_price = UINT_MAX;

  /***************************************************************************/
  /* Process all the attributes returned with this destination               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Processing attributes"));

  for (curr_attr = ldap_first_attribute(ld, result, &ber);
       curr_attr != NULL;
       curr_attr = ldap_next_attribute(ld, result, ber))
  {
    TRC_DETAIL((TB, "Processing attribute '%s'", curr_attr));

    attr_vals = ldap_get_values(ld, result, curr_attr);
    TRC_ASSERT((attr_vals != NULL),
                                (TB, "Attribute %s has no values", curr_attr));

    /*************************************************************************/
    /* What we do next depends on the attribute we've got                    */
    /*************************************************************************/
    if (strcmp(curr_attr, sca_ip_address) == 0)
    {
      /***********************************************************************/
      /* Save off the IP address - we'll catch any allocation failure        */
      /* further down                                                        */
      /***********************************************************************/
      TRC_DETAIL((TB, "Found IP address %s", attr_vals[0]));
      OSP_STRDUP(new_dest->ip_address, attr_vals[0]);
    }
    else if (strcmp(curr_attr, sca_almost_out) == 0)
    {
      /***********************************************************************/
      /* Check for out of resources flag                                     */
      /***********************************************************************/
      if (strcmp(attr_vals[0], sca_aoor_true) == 0)
      {
        TRC_DETAIL((TB, "Gateway almost out of resources"));
        new_dest->almost_out = TRUE;
      }
    }
    else if (strcmp(curr_attr, sca_e164_attr) == 0)
    {
      /***********************************************************************/
      /* We're actually not interested in the list of matches unless there   */
      /* turns out to be no pricing information for the gateway.  Thus if we */
      /* haven't yet received the pricing, just store off the pointer to the */
      /* array                                                               */
      /***********************************************************************/
      if (new_dest->best_price == UINT_MAX)
      {
        TRC_DETAIL((TB, "Saving match array"));
        matches = attr_vals;

        /*********************************************************************/
        /* We don't want to free the attr vals yet...                        */
        /*********************************************************************/
        continue;
      }
    }
    else if (strcmp(curr_attr, sca_pricing) == 0)
    {
      /***********************************************************************/
      /* Pricing info: search for prices which begin with a substring match  */
      /* on the destination number and record the lowest price and its       */
      /* associated number                                                   */
      /*                                                                     */
      /* The pricing info is in the form                                     */
      /*                                                                     */
      /*     pricing=44181@3.0F2000-05-31T18:00:00ZU2000-06-01T08:00:00Z     */
      /*                                                                     */
      /* which means 'calls to numbers beginning 44181 cost 3 dollars per    */
      /* minute from 6pm on May 31 until 8am on June 1'.                     */
      /*                                                                     */
      /* Note that this sample code assumes that if a given gateway gives    */
      /* multiple matches, the gateway's pricing algorithm will also provide */
      /* the lowest price, and further, that the better matches will be      */
      /* cheaper than less good ones.                                        */
      /*                                                                     */
      /* E.g. a gateway may offer to terminate calls to 44 (the whole UK)    */
      /* for 1 dollar/minute, but to 4420 (London) for 50cents/minute.       */
      /* We'll pick the 50c/min option here, and assume the gateway will do  */
      /* the same.                                                           */
      /***********************************************************************/
      /* Of the prices which are substring matches on our destination, find  */
      /* the lowest                                                          */
      /***********************************************************************/
      TRC_DETAIL((TB, "Gateway pricing"));
      for (ii = 0; attr_vals[ii] != NULL; ii++)
      {
        /*********************************************************************/
        /* find the separator                                                */
        /*********************************************************************/
        TRC_DETAIL((TB, "Got number@price %s", attr_vals[ii]));
        sep = strchr(attr_vals[ii], '@');
        if (sep == NULL)
        {
          TRC_ALERT((TB, "Price without separator %s", attr_vals[ii]));
          continue;
        }

        /*********************************************************************/
        /* Replace the separator with a NULL in the attribute to enable us   */
        /* to do the substring match on the number part                      */
        /*********************************************************************/
        *sep = 0;
        TRC_DETAIL((TB, "Checking price for number %s", attr_vals[ii]));
        if (!strncmp(attr_vals[ii], route_data->dest_info->address,
                                                        strlen(attr_vals[ii])))
        {
          /*******************************************************************/
          /* Price IS for the right number - check it is current             */
          /*******************************************************************/
          *sep = '@';
          TRC_DETAIL((TB, "Checking validity for '%s'", attr_vals[ii]));
          valid_rc = sca_check_price_validity(attr_vals[ii], &price);
          if (valid_rc == SCA_PRICE_CURRENT)
          {
            /*****************************************************************/
            /* The price is current - convert it to a numeric value          */
            /*****************************************************************/
            TRC_DETAIL((TB, "Got match with valid times, price %s", price));
            this_price = sca_price_to_int(price);

            /*****************************************************************/
            /* Is it lower?                                                  */
            /*****************************************************************/
            if (this_price < new_dest->best_price)
            {
              /***************************************************************/
              /* Yes - record the price and the matching number              */
              /***************************************************************/
              strncpy(new_dest->price, price, SCA_MAX_PRICE_LEN);
              strncpy(new_dest->matching_number, attr_vals[ii],
                                                             SCA_MAX_E164_LEN);
              new_dest->best_price = this_price;
              new_dest->best_match = strlen(new_dest->matching_number);
              TRC_NORMAL((TB, "New best price %s on match %s",
                          new_dest->price,
                          new_dest->matching_number));
            }
          }
          else if (valid_rc == SCA_PRICE_EXPIRED)
          {
            /*****************************************************************/
            /* Pricing validity has expired - we should remove it from the   */
            /* directory                                                     */
            /*****************************************************************/
            TRC_NORMAL((TB, "Removing expired pricing"));

            dn = ldap_get_dn(ld, result);
            if (dn == NULL)
            {
              /***************************************************************/
              /* With no dn, there's nothing we can do.  This is not fatal   */
              /* however, so we'll carry on without deleting the pricing.    */
              /***************************************************************/
              TRC_ALERT((TB, "Failed to get dn - can't delete price"));
            }
            else
            {
              /***************************************************************/
              /* Now build the mods structure for the deletion.              */
              /***************************************************************/
              TRC_DETAIL((TB, "Got dn %s", dn));
              mod.mod_op = LDAP_MOD_DELETE;
              mod.mod_type = (char *)sca_pricing;
              values[0] = attr_vals[ii];
              mod.mod_values = values;
              mods[0] = &mod;

              TRC_DETAIL((TB, "Deleting price..."));
              ldap_rc = ldap_modify_s(ld, dn, mods);
              if (ldap_rc == LDAP_SUCCESS)
              {
                TRC_DETAIL((TB, "Expired price removed OK"));
              }
              else
              {
                /*************************************************************/
                /* We trace the error, but our interest is purely academic - */
                /* we won't fail the routing because we couldn't remove an   */
                /* expired price                                             */
                /*************************************************************/
                TRC_ERROR((TB, "LDAP failure removing Expired price"));
              }
            }
          }
          else
          {
            /*****************************************************************/
            /* Price is not yet current                                      */
            /*****************************************************************/
            TRC_NORMAL((TB, "Price not current yet"));
          }
        }
      }
    }
    else
    {
      TRC_ASSERT((FALSE), (TB, "Unexpected attribute %s", curr_attr));
    }

    ldap_value_free(attr_vals);
  }

  /***************************************************************************/
  /* Ensure we got a gateway address - the destination is no good without    */
  /* one!                                                                    */
  /***************************************************************************/
  if (new_dest->ip_address == NULL)
  {
    TRC_ERROR((TB, "No ip address for gateway - discarding"));
    rm_release_mem(new_dest);
    new_dest = NULL;
  }
  else
  {
    /*************************************************************************/
    /* If we didn't get any pricing info, then we need to run though the     */
    /* matches looking for the best                                          */
    /*************************************************************************/
    if (new_dest->best_price == UINT_MAX)
    {
      /***********************************************************************/
      /* If there were no matches, then we shouldn't be called               */
      /***********************************************************************/
      TRC_ASSERT((matches != NULL), (TB, "No matches and no pricing!"));

      /***********************************************************************/
      /* This sample code assumes that if a given gateway gives multiple     */
      /* matches, the better matches will be cheaper than less good ones.    */
      /*                                                                     */
      /* The best match is simply the longest - but note that not all the    */
      /* numbers returned for the gateway will be matches on the number we   */
      /* want, so we can't simply look for the longest number                */
      /***********************************************************************/
      TRC_NORMAL((TB, "No pricing - find best match for %s",
                                              route_data->dest_info->address));
      new_dest->best_match = 0;
      for (ii = 0; matches[ii] != NULL; ii++)
      {
        this_match = sca_match_len(matches[ii],
                                               route_data->dest_info->address);
        if (this_match > new_dest->best_match)
        {
          TRC_DETAIL((TB, "New best match"));
          new_dest->best_match = this_match;
          strncpy(new_dest->matching_number, matches[ii], SCA_MAX_E164_LEN);
        }
      }
    }
  }

  /***************************************************************************/
  /* If we saved off the matches, free them now                              */
  /***************************************************************************/
  if (matches != NULL)
  {
    TRC_DETAIL((TB, "Freeing match array"));
    ldap_value_free(matches);
  }

EXIT_LABEL:

  TRC_EXIT();

  return(new_dest);

} /* sca_make_sca_destination */

/**PROC+**********************************************************************/
/* Name:      sca_free_route_data                                            */
/*                                                                           */
/* Purpose:   Frees an SCA_ROUTE_DATA structure                              */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     route_data   - pointer to structure to free             */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_route_data(SCA_ROUTE_DATA *route_data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("sca_free_route_data");

  if (route_data != NULL)
  {
    /*************************************************************************/
    /* Free each element                                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing ar data %p", route_data));
    scm_free_list((SLIST_ENTRY *)route_data->call_id_list,
                                            (PFREE_ENTRY_FN)&sca_free_call_id);
    sca_free_address(route_data->source_info);
    sca_free_address(route_data->dest_info);
    scm_free_list((SLIST_ENTRY *)route_data->dest_alt_list,
                                            (PFREE_ENTRY_FN)&sca_free_address);
    rm_release_mem(route_data);
  }

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* sca_free_route_data */

/**PROC+**********************************************************************/
/* Name:      sca_store_routing_params                                       */
/*                                                                           */
/* Purpose:   Store the supplied params in a SCA_ROUTE_DATA structure        */
/*                                                                           */
/* Returns:   pointer to completed route data                                */
/*                                                                           */
/* Params:    IN     corr         - correlator to be used on response        */
/*            IN     request_type - auth request or reauth                   */
/*            IN     call_id_list - list of call identifiers to use for      */
/*                                  returned destinations                    */
/*            IN     source_info  - pointer to source info structure         */
/*            IN     dest_info    - pointer to destination info structure    */
/*            IN     dest_alt_list- destination alternate list               */
/*            IN     max_dests    - maximum number of destinations to return */
/*                                                                           */
/**PROC-**********************************************************************/

SCA_ROUTE_DATA *sca_store_routing_params(OSP_CORRELATOR corr,
                                         int request_type,
                                         OSP_CALL_ID *call_id_list,
                                         OSP_ADDRESS *source_info,
                                         OSP_ADDRESS *dest_info,
                                         OSP_ADDRESS *dest_alt_list,
                                         char *max_dests)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  SCA_ROUTE_DATA *route_data = NULL;
  int success = FALSE;

  TRC_ENTRY("sca_store_routing_params");

  /***************************************************************************/
  /* Get the memory to hold the parameters                                   */
  /***************************************************************************/
  route_data = (SCA_ROUTE_DATA *)rm_get_mem(sizeof(SCA_ROUTE_DATA));
  if (route_data == NULL)
  {
    /*************************************************************************/
    /* Fail the request                                                      */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to alloc mem for ar data"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize pointers                                                     */
  /***************************************************************************/
  route_data->call_id_list = NULL;
  route_data->source_info = NULL;
  route_data->dest_info = NULL;
  route_data->dest_alt_list = NULL;

  /***************************************************************************/
  /* Save the parameters                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Saving parameters"));
  route_data->request_type = request_type;
  route_data->corr = corr;
  route_data->max_dests = atoi(max_dests);

  if (source_info != NULL)
  {
    TRC_DETAIL((TB, "Copying source address"));
    route_data->source_info = sca_copy_address(source_info);
    if (route_data->source_info == NULL)
    {
      goto EXIT_LABEL;
    }
  }

  if (dest_info != NULL)
  {
    TRC_DETAIL((TB, "Copying dest address"));
    route_data->dest_info = sca_copy_address(dest_info);
    if (route_data->dest_info == NULL)
    {
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Copy the call id and destination alternate lists                        */
  /***************************************************************************/
  if (!scm_copy_list((SLIST_ENTRY *)call_id_list,
                     (SLIST_ENTRY **)&route_data->call_id_list,
                     (PCOPY_ENTRY_FN)sca_copy_call_id,
                     (PFREE_ENTRY_FN)sca_free_call_id))
  {
    TRC_ERROR((TB, "Failed copying call id list"));
    route_data->call_id_list = NULL;
    goto EXIT_LABEL;
  }

  if (!sca_copy_dest_alt_list((SLIST_ENTRY *)dest_alt_list,
                              (SLIST_ENTRY **)&route_data->dest_alt_list,
                              (PCOPY_ENTRY_FN)sca_copy_address,
                              (PFREE_ENTRY_FN)sca_free_address))
  {
    TRC_ERROR((TB, "Failed copying destination alternates"));
    route_data->dest_alt_list = NULL;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If we got here, we're OK                                                */
  /***************************************************************************/
  success = TRUE;

EXIT_LABEL:
  /***************************************************************************/
  /* If something went wrong, free what we got                               */
  /***************************************************************************/
  if (!success)
  {
    TRC_DETAIL((TB, "Freeing what we got after failure"));
    sca_free_route_data(route_data);
    route_data = NULL;
  }

  TRC_EXIT();

  return(route_data);

} /* sca_store_routing_params */

/**PROC+**********************************************************************/
/* Name:      sca_route_filter                                               */
/*                                                                           */
/* Purpose:   Build a routing search filter from the supplied address        */
/*                                                                           */
/* Returns:   pointer to filter string                                       */
/*                                                                           */
/* Params:    IN     address      - address to use                           */
/*            IN     service      - requested service information            */
/*                                                                           */
/**PROC-**********************************************************************/

char *sca_route_filter(char *address, OSP_SERVICE_INFO *service)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char *filter = NULL;
  int filter_len;
  int e164_len;
  int protocols_specified;
  int service_specified;
  int extra_paren;
  OSP_PROTOCOL_TYPE *prot_type;

  TRC_ENTRY("sca_route_filter");

  /***************************************************************************/
  /* Initialization                                                          */
  /***************************************************************************/
  extra_paren = FALSE;

  /***************************************************************************/
  /* Bulid the filters.                                                      */
  /*                                                                         */
  /* We want to end up with a filter of the form                             */
  /*                                                                         */
  /*     (&(e164 part)(protocol part)(service part))                         */
  /*                                                                         */
  /* but depending on what's specified, we might just end up with            */
  /*                                                                         */
  /*     (e164 part)                                                         */
  /*                                                                         */
  /* Hence we have to walk the lists of supplied information to determine    */
  /* the detail of the filter as well as just the length                     */
  /*                                                                         */
  /* e164 part                                                               */
  /* =========                                                               */
  /*                                                                         */
  /* For 441813661177, we want a search on (e164=441813661177)|              */
  /* (e164=44181366117)|(e164=4418136611)|...  (e164=4), for which the LDAP  */
  /* syntax is (|(e164=441813661177) (e164=44181366117)...(e164=4)           */
  /*                                                                         */
  /* If any protocol types are specified in the service information, then    */
  /* these have to be added on as (number filter) AND (service filters)      */
  /*                                                                         */
  /* Calculate the size of the number part of the filter                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Calculating e164 filter len"));
  e164_len = strlen(address);
  filter_len = 4; /* leading "(|", trailing ")\0" */
  while (e164_len)
  {
    filter_len += 7; /* "(e164=") + trailing ")" */
    filter_len += e164_len;
    e164_len--;
  }
  TRC_DETAIL((TB, "e164 filter len is %d", filter_len));

  /***************************************************************************/
  /* protocol part                                                           */
  /* =============                                                           */
  /*                                                                         */
  /* This one is much easier: we want a filter that returns just gateways    */
  /* with an attribute matching the protocol, or those with no protocol      */
  /* specified, which we treat as meaning 'supports all'                     */
  /*                                                                         */
  /*     (|(protocol_type=<protocol 1>)(protocol_type=<protocol 2>)          */
  /*                                                   (!(protocol_type=*))) */
  /*                                                                         */
  /***************************************************************************/
  protocols_specified = FALSE;
  if ((service != NULL) && (service->prot_type != NULL))
  {
    TRC_DETAIL((TB, "Adding space for protocol types"));
    filter_len += 3; /* initial "(&" and trailing ")" */
    filter_len += 20; /* "(!(protocol_type=*))"       */
    extra_paren = TRUE;
    protocols_specified = TRUE;
    prot_type = service->prot_type;
    while (prot_type != NULL)
    {
      filter_len += 16; /* "(protocol_type=" and ")" */
      filter_len += strlen(prot_type->prot_name);
      prot_type = prot_type->next;
    }
  }

  /***************************************************************************/
  /* service part                                                            */
  /* ============                                                            */
  /*                                                                         */
  /* This is similar to the protocol part, except                            */
  /* - there can be a maximum of one service                                 */
  /* - we treat not service type attributes as 'supports all services'       */
  /*                                                                         */
  /* so we need a search of the form                                         */
  /*                                                                         */
  /*    (|(service_type=<service type>)(!(service_type=*)))                  */
  /*                                                                         */
  /* where the second term means 'service attribute not present'             */
  /*                                                                         */
  /***************************************************************************/
  service_specified = FALSE;
  if ((service != NULL) && (service->service_type != NULL))
  {
    TRC_DETAIL((TB, "Adding space for service type"));
    extra_paren = TRUE;
    service_specified = TRUE;
    filter_len += 38;
    filter_len += strlen(service->service_type);
  }

  /***************************************************************************/
  /* If we've got multiple parts, we need to allow for the (&...) that wraps */
  /* them all                                                                */
  /***************************************************************************/
  if (extra_paren)
  {
    filter_len += 3; /* initial "(&" and trailing ")" */
  }

  /***************************************************************************/
  /* The whole filter is null terminated                                     */
  /***************************************************************************/
  filter_len++;

  /***************************************************************************/
  /* Get the memory for the filter                                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting %d bytes for filter", filter_len));
  filter = rm_get_mem(filter_len);
  if (filter == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory to build filter"));
  }
  else
  {
    /*************************************************************************/
    /* Build the filter                                                      */
    /*************************************************************************/
    TRC_DETAIL((TB, "Building filter..."));
    if (extra_paren)
    {
      strcpy(filter, "(&(|");
    }
    else
    {
      strcpy(filter, "(|");
    }

    /*************************************************************************/
    /* Number part                                                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Number part"));
    e164_len = strlen(address);
    while (e164_len)
    {
      strcat(filter, "(e164=");
      strncat(filter, address, e164_len);
      strcat(filter, ")");
      e164_len--;
    }
    strcat(filter, ")");

    /*************************************************************************/
    /* Protocol part                                                         */
    /*************************************************************************/
    if (protocols_specified)
    {
      TRC_DETAIL((TB, "Protocol part"));
      strcat(filter, "(|");
      prot_type = service->prot_type;
      while (prot_type != NULL)
      {
        strcat(filter, "(");
        strcat(filter, sca_prot_type);
        strcat(filter, "=");
        strcat(filter, prot_type->prot_name);
        strcat(filter, ")");
        prot_type = prot_type->next;
      }

      strcat(filter, "(!(");
      strcat(filter, sca_prot_type);
      strcat(filter, "=*)))");
    }

    /*************************************************************************/
    /* Service types                                                         */
    /*************************************************************************/
    if (service_specified)
    {
      TRC_DETAIL((TB, "Adding space for service type"));
      strcat(filter, "(|(");
      strcat(filter, sca_service_type);
      strcat(filter, "=");
      strcat(filter, service->service_type);
      strcat(filter, ")(!(");
      strcat(filter, sca_service_type);
      strcat(filter, "=*)))");
    }

    /*************************************************************************/
    /* Any trailing close parentheses                                        */
    /*************************************************************************/
    if (extra_paren)
    {
      strcat(filter, ")");
    }

    TRC_DETAIL((TB, "Dest <%s> gave search filter <%s>", address, filter));
  }

  TRC_EXIT();

  return(filter);

} /* sca_route_filter */
