
/**MOD+***********************************************************************/
/* Module:    scasacb.c                                                      */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            SubscriberAuthentication Request processing                    */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.8                $ $Modtime::   Jan 11 2001 15:56:18   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_subs_auth_callback                                         */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI sca_subs_auth_callback(OSP_CORRELATOR corr,
                                     char *timestamp,
                                     OSP_ADDRESS *source_info,
                                     OSP_ADDRESS *source_alt_list,
                                     OSP_ADDRESS *dest_info,
                                     OSP_SERVICE_INFO *service,
                                     OSP_AUTH_INFO *subs_authen_info,
                                     OSP_CLIENT_ID *client_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_STATUS status;
  OSP_RC rc = OSP_SUCCESS;
  OSP_ADDRESS *subscriber_id;

  TRC_ENTRY("sca_subs_auth_callback");

  /***************************************************************************/
  /* Check parameters                                                        */
  /***************************************************************************/
  TRC_ASSERT((corr != NULL), (TB, "Passed NULL correlator"));
  TRC_ASSERT((source_info != NULL), (TB, "Passed NULL source address"));
  TRC_DETAIL((TB, "Subs Auth Rq for source '%s' type '%s'",
                                     source_info->address, source_info->type));

  /***************************************************************************/
  /* Check for a source alt of type subscriber                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Looking for subscriber info..."));
  subscriber_id = sca_find_subs_info(source_alt_list);
  if (subscriber_id == NULL)
  {
    TRC_ALERT((TB, "No subscriber info supplied"));
    status.code = OSP_STATUS_NOT_IMPLEMENTED;
    status.desc = sca_authen_failed;
    rc = OSP_ERR_BAD_REQUEST;
  }
  else
  {
    /*************************************************************************/
    /* Start the authentication process.  This involves a directory lookup   */
    /* so completes asynchronously                                           */
    /*************************************************************************/
    rc = sca_do_authentication(corr, SCA_SUBS_AUTHEN_RQ, NULL, source_info,
                                        source_alt_list, dest_info, NULL, "0");
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "authentication failure, rc = %d", rc));
      status.code = OSP_STATUS_AUTHENTICATION_UNSUCCESSFUL;
      status.desc = sca_authen_failed;
    }
  }

  /***************************************************************************/
  /* If something went wrong, then we need to send a (failure) response now  */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    TRC_ALERT((TB, "Immediate SA failure, code %d", status.code));
    rc = osp_sa_response(corr,
                         &status,
                         NULL,
                         NULL,
                         NULL,
                         scm_sign_responses);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We absolutely should NOT get a bad parameter error; otherwise just  */
      /* trace the result.                                                   */
      /***********************************************************************/
      TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
      TRC_ERROR((TB, "SubsAuth response call failed, rc = %d", rc));
    }
  }

  TRC_EXIT();

  return(rc);

} /* sca_auth_rq_callback */

/**PROC+**********************************************************************/
/* Name:      sca_do_authentication                                          */
/*                                                                           */
/* Purpose:   Initiate authentication lookup                                 */
/*                                                                           */
/* Returns:   OSP_SUCCESS - lookup started                                   */
/*            other - OSP error code                                         */
/*                                                                           */
/* Params:    IN     corr         - correlator to be used on response        */
/*            IN     request_type - auth request or reauth                   */
/*            IN     call_id_list - list of call identifiers to use for      */
/*                                  returned destinations                    */
/*            IN     source_info  - pointer to source info structure         */
/*            IN     subscriber_id- subscriber identifier from src alt list  */
/*            IN     dest_info    - pointer to destination info structure    */
/*            IN     dest_alt_list- destination alternate list               */
/*            IN     max_dests    - maximum number of destinations to return */
/*                                                                           */
/* Operation: Find subscriber id in source alternate list                    */
/*            Start look up in subscriber directory                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_do_authentication(OSP_CORRELATOR corr,
                             int request_type,
                             OSP_CALL_ID *call_id_list,
                             OSP_ADDRESS *source_info,
                             OSP_ADDRESS *subscriber_id,
                             OSP_ADDRESS *dest_info,
                             OSP_ADDRESS *dest_alt_list,
                             char *max_dests)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  SCA_ROUTE_DATA *route_data;
  char filter[256];

  TRC_ENTRY("sca_do_authentication");

  /***************************************************************************/
  /* We MUST be passed the source and subscriber id                          */
  /***************************************************************************/
  TRC_ASSERT((source_info != NULL), (TB, "Passed NULL source address"));
  TRC_ASSERT((subscriber_id != NULL), (TB, "Called with no subs id"));

  /***************************************************************************/
  /* Store off the parameters that we'll need to perform routing later on    */
  /***************************************************************************/
  route_data = sca_store_routing_params(corr, request_type, call_id_list,
                             source_info, dest_info, dest_alt_list, max_dests);
  if (route_data == NULL)
  {
    TRC_ERROR((TB, "failed to store routing params"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now start the lookup based on the subscriber id                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Lookup subscriber id '%s'", subscriber_id->address));
  OSP_SNPRINTF((filter, sizeof(filter), "subscriber_id=%s",
                                                      subscriber_id->address));
  rc = lam_search(sca_lam_handle,
                  (char *)sca_ar_base,
                  LDAP_SCOPE_SUBTREE,
                  filter,
                  (char **)sca_auth_rq_attrs,
                  FALSE,
                  (void *)route_data,
                  sca_authen_callback);
  if (rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* now we wait for the LAM callback, where we'll do any required routing */
    /*************************************************************************/
    TRC_DETAIL((TB, "Subscriber lookup initiated successfully"));
  }
  else
  {
    /*************************************************************************/
    /* Free any resources then return the failure                            */
    /*************************************************************************/
    TRC_ERROR((TB, "Search operation failed with rc %d", rc));
    sca_free_route_data(route_data);
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sca_do_authentication */

/**PROC+**********************************************************************/
/* Name:      sca_authen_callback                                            */
/*                                                                           */
/* Purpose:   Called by LAM on subscriber lookup (authentication) completion */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - userdata supplied to lam_search/modify   */
/*            IN     ld           - LDAP handle required to parse result     */
/*            IN     result       - results of the search operation.         */
/*                                                                           */
/**PROC-**********************************************************************/

void OSPAPI sca_authen_callback(void *user_data,
                                LDAP *ld,
                                LDAPMessage **result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  SCA_ROUTE_DATA *route_data;
  int result_count;
  char *filter;
  OSP_TOKEN *authen_token;
  OSP_STATUS status;
  char transaction_id[SCA_TRANS_ID_LEN];
  LDAPMessage *routing_result;
  int ldap_rc;
  int send_now;

  TRC_ENTRY("sca_authen_callback");

  /***************************************************************************/
  /* Initialize variables                                                    */
  /***************************************************************************/
  route_data = NULL;
  authen_token = NULL;
  filter = NULL;
  routing_result = NULL;

  /***************************************************************************/
  /* Unless we're doing routing, we want to send the results from here.      */
  /***************************************************************************/
  send_now = TRUE;

  /***************************************************************************/
  /* Extract the ar data from the user_data.  Without it there's nothing we  */
  /* can do - we don't even have the correlator required for the callback    */
  /* into the AIB - hence the assert                                         */
  /***************************************************************************/
  route_data = (PSCA_ROUTE_DATA)user_data;
  TRC_ASSERT((route_data != NULL), (TB, "Received NULL ar data in LAM c/b"));

  /***************************************************************************/
  /* Set up default values for the status field                              */
  /***************************************************************************/
  status.code = OSP_STATUS_SUCCESS;
  status.desc = (char *)sca_authen_success;

  /***************************************************************************/
  /* Get a transaction id                                                    */
  /***************************************************************************/
  sca_get_trans_id(transaction_id);

  /***************************************************************************/
  /* Check we got something                                                  */
  /***************************************************************************/
  if ((ld == NULL) || (result == NULL))
  {
    TRC_ALERT((TB, "No subscriber found due to directory server problem"));
    status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status.desc = (char *)scm_subs_server_failure;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* See if we got any subscribers back                                      */
  /***************************************************************************/
  result_count = ldap_count_entries(ld, *result);
  TRC_DETAIL((TB, "Got %d results", result_count));
  if (result_count != 1)
  {
    /*************************************************************************/
    /* If we got more than one subscriber back, we don't know which one      */
    /* they're trying to authenticate!                                       */
    /*************************************************************************/
    TRC_ALERT((TB, "Search found %d matching subscribers", result_count));
    status.code = OSP_STATUS_AUTHENTICATION_UNSUCCESSFUL;
    status.desc = (char *)sca_authen_failed;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Build an authentication token                                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Subscriber authenticated; building token"));
  authen_token = sca_create_token(route_data, transaction_id, NULL, NULL);
  if (authen_token == NULL)
  {
    TRC_ERROR((TB, "Failed to create authentication token"));
    status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status.desc = (char *)scm_subs_server_failure;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /*                                                                         */
  /* NOTE: you should do any required authorization at this point.           */
  /*                                                                         */
  /* For the OpenOSP sample applications, authorization consists simply of   */
  /* validating the authentication token we've just created, so we don't do  */
  /* anything here.                                                          */
  /*                                                                         */
  /***************************************************************************/

  /***************************************************************************/
  /* If any routing is required, do it now                                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Authentication complete"));
  if (route_data->max_dests > 0)
  {
    /*************************************************************************/
    /* build the LAM query filter using the destination address              */
    /*************************************************************************/
    TRC_DETAIL((TB, "Doing routing"));
    filter = sca_route_filter(route_data->dest_info->address, NULL);
    if (filter == NULL)
    {
      TRC_ERROR((TB, "failed to build search filter"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Do the search                                                         */
    /*************************************************************************/
    ldap_rc = ldap_search_s(ld,
                            (char *)sca_ar_base,
                            LDAP_SCOPE_SUBTREE,
                            filter,
                            (char **)sca_auth_rq_attrs,
                            FALSE,
                            &routing_result);
    if (ldap_rc == LDAP_SUCCESS)
    {
      TRC_DETAIL((TB, "Search successful - do routing callback"));
      sca_search_callback((void *)route_data, ld, &routing_result);
      send_now = FALSE;

      /***********************************************************************/
      /* The search callback will have freed the route data for us           */
      /***********************************************************************/
      route_data = NULL;
    }
    else
    {
      TRC_ERROR((TB, "LDAP failure %d doing routing search", ldap_rc));
      status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status.desc = (char *)scm_subs_resource_failure;
    }
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Pass the result back down the stack if required                         */
  /***************************************************************************/
  if (send_now)
  {
    TRC_DETAIL((TB, "Calling into stack"));
    if (route_data->request_type == SCA_ROUTE_AUTH_RQ)
    {
      TRC_DETAIL((TB, "Auth Rsp"));
      rc = osp_ar_auth_response(route_data->corr,
                                &status,
                                transaction_id,
                                NULL,
                                NULL,
                                scm_sign_responses);
    }
    else if (route_data->request_type == SCA_ROUTE_REAUTH_RQ)
    {
      TRC_DETAIL((TB, "Reauth Rsp"));
      rc = osp_ar_reauth_response(route_data->corr,
                                  &status,
                                  transaction_id,
                                  NULL,
                                  NULL,
                                  scm_sign_responses);
    }
    else if (route_data->request_type == SCA_SUBS_AUTHEN_RQ)
    {
      TRC_DETAIL((TB, "SubsAuthen Rsp"));
      rc = osp_sa_response(route_data->corr,
                           &status,
                           authen_token,
                           NULL,
                           NULL,
                           scm_sign_responses);
    }
    else
    {
      TRC_ASSERT((FALSE), (TB, "Unknown request type %d",
                                                    route_data->request_type));
    }
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We absolutely should NOT get a bad parameter error; otherwise just  */
      /* trace the result.                                                   */
      /***********************************************************************/
      TRC_ASSERT(((rc != OSP_ERR_BAD_PARAM) && (rc != OSP_ERR_BAD_CORR)),
                         (TB, "Bad parameter errors unexpected from OSP api"));
      TRC_ERROR((TB, "OSP (re)auth rsp call failed, rc = %d", rc));
    }
  }

  /***************************************************************************/
  /* All done - just free our resources                                      */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing memory"));
  sca_free_route_data(route_data);
  rm_release_mem(filter);

  /***************************************************************************/
  /* Free any tokens                                                         */
  /***************************************************************************/
  if (authen_token != NULL)
  {
    TRC_DETAIL((TB, "Freeing authen token"));
    rm_release_mem(authen_token->encoding);
    rm_release_mem(authen_token->token_data);
    rm_release_mem(authen_token);
  }

  TRC_EXIT();

  return;

} /* sca_authen_callback */

