
/**MOD+***********************************************************************/
/* Module:    scatkut.c                                                      */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Token Functions                                                */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.19               $ $Modtime::   Jan 11 2001 15:56:20   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>
#include <openssl/err.h>

/**PROC+**********************************************************************/
/* Name:      sca_create_token                                               */
/*                                                                           */
/* Purpose:   Allocate memory for and create an authorization token          */
/*                                                                           */
/* Returns:   pointer to the token or NULL                                   */
/*                                                                           */
/* Params:    IN     route_data   - information used in the token creation   */
/*            IN     transaction_id                                          */
/*                                - trans id for the token                   */
/*            IN     call_id      - call id to use for the token             */
/*            IN     osp_dest     - destination to receive token validity    */
/*                                  and duration                             */
/*                                                                           */
/* Operation: Allocate memory for the token                                  */
/*            Write the token in XML format                                  */
/*            MD5 sign it, if required                                       */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_TOKEN *sca_create_token(SCA_ROUTE_DATA *route_data,
                            char *transaction_id,
                            OSP_CALL_ID *call_id,
                            OSP_DESTINATION *osp_dest)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_TOKEN *new_token = NULL;
  OSP_USAGE_DETAIL* osp_usage;
  char xml[128];
  int xml_len;
  char token_data[SCA_TOKEN_LEN];
  char *signed_token;
  unsigned int signed_token_len;
  unsigned long random_data;
  OSP_RC rc;

  TRC_ENTRY("sca_create_token");

  /***************************************************************************/
  /* Check what we've been passed                                            */
  /***************************************************************************/
  TRC_ASSERT((route_data != NULL),
                           (TB, "Asked to create token usig NULL route data"));

  /***************************************************************************/
  /* Initialize pointers                                                     */
  /***************************************************************************/
  signed_token = NULL;
  osp_usage = NULL;

  /***************************************************************************/
  /* alloc the token memory                                                  */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting memory for token"));
  new_token = rm_get_mem(sizeof(OSP_TOKEN));
  if (new_token == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate memory for token"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  new_token->next = NULL;
  new_token->encoding = NULL;
  new_token->token_data = NULL;

  /***************************************************************************/
  /* Get the random data                                                     */
  /***************************************************************************/
  ccm_random((void *)&random_data, sizeof(random_data));

  /***************************************************************************/
  /* Now set up the token                                                    */
  /***************************************************************************/
  OSP_STRDUP(new_token->encoding, "base64");
  if (new_token->encoding == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate memory for token encoding"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Write the XML header                                                    */
  /***************************************************************************/
  xml_len = OSP_SNPRINTF((xml,
                          sizeof(xml),
                          "<%s %s=\"%010lu\">\n",
                          scm_token_info,
                          scm_random,
                          random_data));
  TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
  TRC_ASSERT((xml_len < SCA_TOKEN_LEN), (TB, "New xml overflows token buff"));
  strcpy(token_data, xml);

  /***************************************************************************/
  /* SourceInfo comes from car data                                          */
  /***************************************************************************/
  xml_len = OSP_SNPRINTF((xml,
                          sizeof(xml),
                          "<%s %s=\"%s\">\n%s\n</%s>\n",
                          scm_source_info,
                          scm_type,
                          sca_e164_type,
                          route_data->source_info->address,
                          scm_source_info));
  TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
  TRC_ASSERT(((strlen(token_data) + xml_len) < SCA_TOKEN_LEN),
                                       (TB, "New xml overflows token buffer"));
  strcat(token_data, xml);

  /***************************************************************************/
  /* DestinationInfo is also from car data                                   */
  /***************************************************************************/
  if ((route_data->dest_info != NULL) &&
      (route_data->dest_info->address != NULL))
  {
    xml_len = OSP_SNPRINTF((xml,
                            sizeof(xml),
                            "<%s %s=\"%s\">\n%s\n</%s>\n",
                            scm_dest_info,
                            scm_type,
                            sca_e164_type,
                            route_data->dest_info->address,
                            scm_dest_info));
    TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
    TRC_ASSERT(((strlen(token_data) + xml_len) < SCA_TOKEN_LEN),
                                       (TB, "New xml overflows token buffer"));
    strcat(token_data, xml);
  }

  /***************************************************************************/
  /* CallId is supplied as a parameter                                       */
  /***************************************************************************/
  if (call_id != NULL)
  {
    xml_len = OSP_SNPRINTF((xml,
                            sizeof(xml),
                            "<%s %s=\"%s\">\n%s\n</%s>\n",
                            scm_call_id,
                            scm_encoding,
                            call_id->encoding,
                            call_id->data,
                            scm_call_id));
    TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
    TRC_ASSERT(((strlen(token_data) + xml_len) < SCA_TOKEN_LEN),
                                       (TB, "New xml overflows token buffer"));
    strcat(token_data, xml);
  }

  /***************************************************************************/
  /* Set up the spread of authorization validity                             */
  /***************************************************************************/
  rc = sca_add_validity(token_data, SCA_TOKEN_LEN, sca_auth_validity,osp_dest);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add validity"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* TransactionId                                                           */
  /***************************************************************************/
  xml_len = OSP_SNPRINTF((xml,
                          sizeof(xml),
                          "<%s>\n%s\n</%s>\n",
                          scm_trans_id,
                          transaction_id,
                          scm_trans_id));
  TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
  TRC_ASSERT(((strlen(token_data) + xml_len) < SCA_TOKEN_LEN),
                                       (TB, "New xml overflows token buffer"));
  strcat(token_data, xml);

  /***************************************************************************/
  /* Add in the usage                                                        */
  /***************************************************************************/
  if (sca_auth_duration == NULL)
  {
    TRC_DETAIL((TB, "No duration set so no usage details"));
  }
  else
  {
    /*************************************************************************/
    /* Build the duration into the token                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Build usage for auth duration %d", sca_auth_duration));
    xml_len = OSP_SNPRINTF((xml,
                            sizeof(xml),
                            "<%s>\n<%s/><%s>\n%s\n</%s>\n<%s>\n1\n</%s>\n" \
                            "<%s>\ns\n</%s>\n</%s>\n",
                            scm_usage_det,
                            com_service,
                            com_amount,
                            sca_auth_duration,
                            com_amount,
                            com_increment,
                            com_increment,
                            com_unit,
                            com_unit,
                            scm_usage_det));
    TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
    TRC_ASSERT(((strlen(token_data) + xml_len) < SCA_TOKEN_LEN),
                                       (TB, "New xml overflows token buffer"));
    strcat(token_data, xml);

    /*************************************************************************/
    /* And set up the same usage in the destination, if supplied             */
    /*************************************************************************/
    if (osp_dest != NULL)
    {
      TRC_DETAIL((TB, "Build usage in destination"));
      /***********************************************************************/
      /* Get the usage structure                                             */
      /***********************************************************************/
      osp_usage = rm_get_mem(sizeof(OSP_USAGE_DETAIL));
      if (osp_usage == NULL)
      {
        TRC_ERROR((TB, "Failed to get memory for auth usage"));
        goto EXIT_LABEL;
      }
      osp_usage->next = NULL;
      osp_usage->service = NULL;
      osp_usage->amount = NULL;
      osp_usage->increment = NULL;
      osp_usage->unit = NULL;
      osp_usage->start_time = NULL;
      osp_usage->end_time = NULL;
      osp_usage->term_cause = NULL;
      osp_usage->statistics = NULL;

      /***********************************************************************/
      /* Fill in the amount, increment and unit (1 second)                   */
      /***********************************************************************/
      OSP_STRDUP(osp_usage->amount, sca_auth_duration);
      OSP_STRDUP(osp_usage->increment, "1");
      OSP_STRDUP(osp_usage->unit, sca_seconds);
      if ((osp_usage->amount == NULL) ||
          (osp_usage->increment == NULL) ||
          (osp_usage->unit == NULL))
      {
        TRC_ERROR((TB, "Failure setting up auth usage"));
        goto EXIT_LABEL;
      }
      osp_dest->usage_list = osp_usage;
    }
  }

  /***************************************************************************/
  /* End the token                                                           */
  /***************************************************************************/
  xml_len = OSP_SNPRINTF((xml,
                          sizeof(xml),
                          "</%s>", scm_token_info));
  TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
  TRC_ASSERT(((strlen(token_data) + xml_len) < SCA_TOKEN_LEN),
                                       (TB, "New xml overflows token buffer"));
  strcat(token_data, xml);
  TRC_DETAIL((TB, "Created token\n%s", token_data));

  /***************************************************************************/
  /* Should we sign the token?                                               */
  /***************************************************************************/
  if (sca_sign_tokens)
  {
    TRC_DETAIL((TB, "Signing token"));
    rc = ccm_pkcs7_sign(sca_signing_dn, token_data, strlen(token_data),
                    CCM_DIGEST_MD5, NULL, 0, &signed_token, &signed_token_len);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Signing failed, rc = %d", rc));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Now we need to base64 encode the result                               */
    /*************************************************************************/
    TRC_DETAIL((TB, "Signed token len %u", signed_token_len));
    TRC_DETAIL((TB, "Base64-encoding token"));
    rc = osp_base64_encode(signed_token,
                           signed_token_len,
                           &new_token->token_data,
                           (unsigned int *)(&(new_token->token_len)));
  }
  else
  {
    /*************************************************************************/
    /* Just encode the token                                                 */
    /*************************************************************************/
    TRC_DETAIL((TB, "Not signing token"));
    rc = osp_base64_encode(token_data,
                           strlen(token_data),
                           &new_token->token_data,
                           (unsigned int *)(&(new_token->token_len)));
  }

  /***************************************************************************/
  /* Check the encoding result                                               */
  /***************************************************************************/
  if (rc == OSP_SUCCESS)
  {
    TRC_DETAIL((TB, "Encoded token successfully"));
  }
  else
  {
    TRC_ERROR((TB, "Base64 encoding failed, rc = %d", rc));
  }
  rm_release_mem(signed_token);

EXIT_LABEL:
  if (rc != OSP_SUCCESS)
  {
    if (new_token != NULL)
    {
      TRC_DETAIL((TB, "Freeing token after failure"));
      rm_release_mem(new_token->encoding);
      rm_release_mem(new_token);
      new_token = NULL;
    }

    if (osp_usage != NULL)
    {
      TRC_DETAIL((TB, "Freeing usage"));
      rm_release_mem(osp_usage->amount);
      rm_release_mem(osp_usage->increment);
      rm_release_mem(osp_usage->unit);
      rm_release_mem(osp_usage);
      osp_dest->usage_list = NULL;
    }
  }

  TRC_EXIT();

  return(new_token);

} /* sca_create_token */

/**PROC+**********************************************************************/
/* Name:      sca_crack_token                                                */
/*                                                                           */
/* Purpose:   Decode and parse the supplied token                            */
/*                                                                           */
/* Returns:   pointer to decoded token, or NULL                              */
/*                                                                           */
/* Params:    IN/OUT status       - status structure used to return any      */
/*                                  error                                    */
/*            IN     token_list   - token to crack                           */
/*                                                                           */
/* Operation: decode the token (it should be base64 encoded)                 */
/*            parse the resulting xml                                        */
/*                                                                           */
/**PROC-**********************************************************************/

SCA_TOKEN_INFO *sca_crack_token(OSP_STATUS *status, OSP_TOKEN *token_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char *verified_token_data;
  unsigned int verified_token_len;
  char *signed_token;
  unsigned int signed_token_len;
  OSP_RC check_rc;
  SCA_TOKEN_INFO *token_info;
  XML_Parser parser;
  TRANS_INFO *trans;
  enum XML_Error xml_error;
  char *xml_error_string;
  int parse_rc;

  TRC_ENTRY("sca_crack_token");

  /***************************************************************************/
  /* Initialize variables                                                    */
  /***************************************************************************/
  verified_token_data = NULL;
  signed_token = NULL;
  token_info = NULL;
  parser = NULL;
  trans = NULL;

  /***************************************************************************/
  /* Check what we've been supplied - we only expect one token, and that     */
  /* should be base64 encoded (since that's what we sent out!)               */
  /***************************************************************************/
  TRC_ASSERT((token_list != NULL), (TB, "Supplied with NULL token"));
  if (token_list->next != NULL)
  {
    TRC_ERROR((TB, "Supplied with multiple tokens"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_unrecognised_token;
    goto EXIT_LABEL;
  }

  if (strcmp(token_list->encoding, "base64"))
  {
    TRC_ERROR((TB, "Token is not base64 encoded"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_unrecognised_token;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Tokens as received don't have the data length supplied                  */
  /***************************************************************************/
  token_list->token_len = strlen(token_list->token_data);

  /***************************************************************************/
  /* Now we can decode the token and check the signature                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Decoding token len %d", token_list->token_len));
  check_rc = osp_base64_decode(token_list->token_data,
                               token_list->token_len,
                               &signed_token,
                               &signed_token_len);
  if (check_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Token decoding failed"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_unrecognised_token;
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Decoded token len %d", signed_token_len));
  check_rc = ccm_pkcs7_verify(signed_token,
                              signed_token_len,
                              NULL,
                              0,
                              0,
                              NULL,
                              &verified_token_data,
                              &verified_token_len,
                              NULL,
                              NULL);
  if (check_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Token signature not verified"));
    status->code = OSP_STATUS_SIGNATURE_INVALID;
    status->desc = sca_token_verify_failed;
    goto EXIT_LABEL;
  }

  TRC_DATA("Verified token:", verified_token_data, verified_token_len);

  /***************************************************************************/
  /* We should now have an XML format token, which we need to parse.  Create */
  /* the expat parser instance                                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Creating parser"));
  parser = XML_ParserCreate(NULL);
  if (parser == NULL)
  {
    TRC_ERROR((TB, "Failed to create parser instance"));
    status->code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status->desc = com_status_transient_problem_in_server;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize the parsing state information                                */
  /***************************************************************************/
  trans = rm_get_mem(sizeof(TRANS_INFO));
  if (trans == NULL)
  {
    TRC_ALERT((TB, "Couldn't allocate transaction for token parsing"));
    status->code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status->desc = com_status_transient_problem_in_server;
    goto EXIT_LABEL;
  }

  trans->http_status = OSP_STATUS_SUCCESS;
  trans->aib_state = NULL;
  trans->xml_state = &parser;
  trans->pool_position = 0;
  trans->pool_size = verified_token_len;
  trans->pool = rm_get_mem(trans->pool_size);
  if (trans->pool == NULL)
  {
    TRC_ALERT((TB, "Couldn't allocate string pool for token parsing"));
    status->code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status->desc = com_status_transient_problem_in_server;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up our expat handlers                                               */
  /***************************************************************************/
  XML_SetElementHandler(parser, sca_start_elem_handler, sca_end_elem_handler);
  XML_SetCharacterDataHandler(parser, sca_character_data_handler);
  XML_SetUserData(parser, trans);

  /***************************************************************************/
  /* And start parsing                                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Start parsing token"));
  parse_rc = XML_Parse(parser, verified_token_data, verified_token_len, TRUE);

  TRC_DETAIL((TB, "...finished parsing: rc=%d", parse_rc));

  /***************************************************************************/
  /* Now see what happened during parsing                                    */
  /***************************************************************************/
  if (parse_rc == 0)
  {
    xml_error = XML_GetErrorCode(parser);
    xml_error_string = (char *)XML_ErrorString(xml_error);
    TRC_ERROR((TB, "Got parsing error %d (%s)", xml_error,
                    (xml_error_string != NULL) ? xml_error_string : "NULL"));
    if (xml_error == XML_ERROR_NO_MEMORY)
    {
      TRC_DETAIL((TB, "Mapping expat no memory to http transient problem"));
      status->code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status->desc = com_status_transient_problem_in_server;
    }
    else
    {
      TRC_DETAIL((TB, "Return parsing unsucessful"));
      status->code = OSP_STATUS_PARSING_UNSUCCESSFUL;
      status->desc = com_status_parsing_unsuccessful;
    }
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* The XML may have been parsed OK, but something else might have gone     */
  /* wrong along the way                                                     */
  /***************************************************************************/
  token_info = trans->aib_state;
  if ((token_info != NULL) &&
                        (token_info->common.status.code == OSP_STATUS_SUCCESS))
  {
    TRC_DETAIL((TB, "Parsed token successfully"));
  }
  else
  {
    TRC_ERROR((TB, "Failure parsing token"));
    status->code = OSP_STATUS_PARSING_UNSUCCESSFUL;
    status->desc = com_status_parsing_unsuccessful;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* The parsed token should now be in trans->aib_state                      */
  /***************************************************************************/
  TRC_ASSERT((token_info != NULL), (TB, "Token info unexpectedly NULL"));
  TRC_DETAIL((TB, "Got token info %p", token_info));

EXIT_LABEL:
  /***************************************************************************/
  /* Free resources                                                          */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing signed/verified token"));
  rm_release_mem(signed_token);
  rm_release_mem(verified_token_data);

  if (status->code != OSP_STATUS_SUCCESS)
  {
    TRC_DETAIL((TB, "Freeing token info after failure"));
    if (trans != NULL)
    {
      sca_free_token_info(trans->aib_state);
    }
    token_info = NULL;
  }

  if (parser != NULL)
  {
    TRC_DETAIL((TB, "Free parser instance"));
    XML_ParserFree(parser);
  }

  TRC_EXIT();

  return(token_info);

} /* sca_crack_token */

/**PROC+**********************************************************************/
/* Name:      sca_start_elem_handler                                         */
/*                                                                           */
/* Purpose:   Start element handler called during token parsing              */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name,                                                   */
/*                   attrs        - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_start_elem_handler(void *user_data,
                            const char *name,
                            const char **attrs)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;
  SCA_TOKEN_INFO *token_info;
  OSP_STATUS status;
  XML_Parser *parser;

  TRC_ENTRY("sca_start_elem_handler");

  /***************************************************************************/
  /* Check the parameters                                                    */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_DETAIL((TB, "Start element for trans %p", trans));
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction info for start element"));
  parser = (XML_Parser *)trans->xml_state;
  TRC_ASSERT((parser != NULL), (TB, "NULL parser for start element"));

  /***************************************************************************/
  /* Check what we got - it should be TokenInfo                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Got start of element %s", name));
  if (!strcmp(name, sca_token_info))
  {
    TRC_DETAIL((TB, "handing off to aib"));
    aib_start_processing(trans,
                         parser,
                         "0",
                         sizeof(SCA_TOKEN_INFO),
                         sca_token_info_elements,
                         sca_end_elem_handler,
                         &status);

    if (trans->aib_state != NULL)
    {
      /***********************************************************************/
      /* Initialize the parsed data memory                                   */
      /***********************************************************************/
      token_info = (SCA_TOKEN_INFO *)trans->aib_state;
      TRC_DETAIL((TB, "Initializing parsed token data"));
      token_info->source_info = NULL;
      SL_INIT_ROOT(token_info->source_alt_list);
      token_info->dest_info = NULL;
      SL_INIT_ROOT(token_info->dest_alt_list);
      SL_INIT_ROOT(token_info->call_id_list);
      token_info->valid_after = NULL;
      token_info->valid_until = NULL;
      token_info->transaction_id = NULL;
      SL_INIT_ROOT(token_info->use_detail_list);
      SL_INIT_ROOT(token_info->auth_url_list);
      token_info->common.char_data.trans = trans;
    }
  }

  TRC_EXIT();

  return;

} /* sca_start_elem_handler */

/**PROC+**********************************************************************/
/* Name:      sca_end_elem_handler                                           */
/*                                                                           */
/* Purpose:   End element handler.                                           */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name         - see expat documentation                  */
/*                                                                           */
/* Operation: Nothing to do.                                                 */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_end_elem_handler(void *user_data, const char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;

  TRC_ENTRY("sca_end_elem_handler");

  /***************************************************************************/
  /* get the transaction info, trans, by casting the user data               */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_ASSERT((trans != NULL), (TB, "NULL user data in AuthRq end handler"));

  /***************************************************************************/
  /* Do common AIB end tag handling.                                         */
  /***************************************************************************/
  TRC_DETAIL((TB, "Got end of %s; calling aib", name));
  aib_end_element_handler(trans, name);

  TRC_EXIT();

  return;

} /* sca_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      sca_character_data_handler                                     */
/*                                                                           */
/* Purpose:   Called with XML character data                                 */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to the transaction's control     */
/*                                  structure                                */
/*            IN     data         - the character data (not NULL-terminated) */
/*            IN     len          - length of data                           */
/*                                                                           */
/* Operation: Nothing to do.                                                 */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_character_data_handler(void *user_data, const char *data, int len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sca_character_data_handler");

  /***************************************************************************/
  /* Just trace what we got                                                  */
  /***************************************************************************/
  TRC_DATA("Unhandled char data:", data, len);

  TRC_EXIT();

  return;

} /* sca_character_data_handler */

/**PROC+**********************************************************************/
/* Name:      sca_free_token_info                                            */
/*                                                                           */
/* Purpose:   Free the memory in the supplied token info                     */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     token        - token to be freed                        */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_token_info(SCA_TOKEN_INFO *token)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sca_free_token_info");

  if (token != NULL)
  {
    TRC_DETAIL((TB, "Freeing token"));

    aib_free_address(token->source_info);
    scm_free_list(SL_HEAD(token->source_alt_list),
                                            (PFREE_ENTRY_FN)&aib_free_address);
    aib_free_address(token->dest_info);
    scm_free_list(SL_HEAD(token->dest_alt_list),
                                            (PFREE_ENTRY_FN)&aib_free_address);

    scm_free_list(SL_HEAD(token->call_id_list),
                                            (PFREE_ENTRY_FN)&aib_free_call_id);

    scm_free_list(SL_HEAD(token->use_detail_list),
                                       (PFREE_ENTRY_FN)&aib_free_usage_detail);

    scm_free_list(SL_HEAD(token->auth_url_list),
                                              (PFREE_ENTRY_FN)&rm_release_mem);

    if (token->common.char_data.trans != NULL)
    {
      rm_release_mem(token->common.char_data.trans->pool);
      rm_release_mem(token->common.char_data.trans);
    }

    if (token->common.corr != NULL)
    {
      rm_release_mem(token->common.corr->comp_id);
      rm_release_mem(token->common.corr);
    }

    rm_release_mem(token);
  }

  TRC_EXIT();

  return;

} /* sca_free_token_info */

/**PROC+**********************************************************************/
/* Name:      sca_confirm_authorization                                      */
/*                                                                           */
/* Purpose:   Verify that a token matches the supplied call information      */
/*                                                                           */
/* Returns:   pointer to cracked token                                       */
/*                                                                           */
/* Params:    IN/OUT status       - used to return status information        */
/*            IN     token_list   - token to verify against supplied details */
/*            IN     source_info  - supplied source address                  */
/*            IN     dest_info    - supplied destination address             */
/*            IN     call_id      - supplied call id                         */
/*                                                                           */
/* Operation: Verify the token signature, decoding if required               */
/*            Compare the contents of the token with the supplied parameters */
/*                                                                           */
/**PROC-**********************************************************************/

SCA_TOKEN_INFO *sca_confirm_authorization(OSP_STATUS *status,
                                          OSP_TOKEN *token_list,
                                          OSP_ADDRESS *source_info,
                                          OSP_ADDRESS *dest_info,
                                          OSP_CALL_ID *call_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  SCA_TOKEN_INFO *token_info = NULL;
  OSP_CALL_ID *token_call_id;
  OSP_RC rc;
  int valid_rc;
  char *token_call_id_data;
  char *supplied_call_id_data;
  unsigned int call_id_len;

  TRC_ENTRY("sca_confirm_authorization");

  /***************************************************************************/
  /* Check that there's a token - since we confirm the authorization based   */
  /* on the token content, there's nothing we can do without one!            */
  /***************************************************************************/
  if (token_list == NULL)
  {
    TRC_ALERT((TB, "No tokens supplied!"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_no_token;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Similarly, we need source and destination information                   */
  /***************************************************************************/
  if (source_info == NULL)
  {
    TRC_ALERT((TB, "No source supplied!"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_author_failed;
    goto EXIT_LABEL;
  }

  if (dest_info == NULL)
  {
    TRC_ALERT((TB, "No destination supplied!"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_author_failed;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now we try to obtain the authorization token, which may be base64       */
  /* encoded                                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Cracking token"));
  token_info = sca_crack_token(status, token_list);
  if (token_info == NULL)
  {
    /*************************************************************************/
    /* If there was a failure, the status will have been recorded for us -   */
    /* just jump to the end and send the response                            */
    /*************************************************************************/
    TRC_ALERT((TB, "Failed to crack supplied token"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Check the token is still valid                                          */
  /***************************************************************************/
  valid_rc = sca_check_validity(token_info->valid_after,
                                                      token_info->valid_until);
  if (valid_rc != SCA_VALIDITY_OK)
  {
    TRC_ERROR((TB, "Authentication info is not valid"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_author_failed;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Having cracked the token successfully and confirmed that it is valid,   */
  /* we can check that the route the token is for matches the route details  */
  /* we've been given                                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Checking token vs supplied route"));
  TRC_DETAIL((TB, "Field     Token - Params"));
  TRC_DETAIL((TB, "Src info  %s - %s", token_info->source_info->address,
                                                        source_info->address));
  TRC_DETAIL((TB, "Dest info %s - %s", token_info->dest_info->address,
                                                          dest_info->address));

  if (strcmp(token_info->source_info->address, source_info->address))
  {
    TRC_ALERT((TB, "Sources don't match"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_invalid_token;
    goto EXIT_LABEL;
  }

  if (strcmp(token_info->dest_info->address, dest_info->address))
  {
    TRC_ALERT((TB, "Destinations don't match"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_invalid_token;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Call id is slightly more complex - either might be in CDATA or base 64  */
  /* encoded                                                                 */
  /***************************************************************************/
  token_call_id = SL_HEAD(token_info->call_id_list);
  TRC_DETAIL((TB, "Call id   %s - %s", token_call_id->data, call_id->data));
  if (strcmp(token_call_id->encoding, call_id->encoding) == 0)
  {
    /*************************************************************************/
    /* Same - just set up the pointers                                       */
    /*************************************************************************/
    TRC_DETAIL((TB, "Call id encodings are the same"));
    token_call_id_data = token_call_id->data;
    supplied_call_id_data = call_id->data;
    call_id_len = OSP_MIN(strlen(token_call_id_data),
                                                strlen(supplied_call_id_data));
  }
  else
  {
    /*************************************************************************/
    /* Different - decode the one that's base64 encoded                      */
    /*************************************************************************/
    TRC_DETAIL((TB, "Call id encodings different"));
    TRC_DETAIL((TB, "Token: '%s', supplied: '%s'", token_call_id->encoding,
                                                           call_id->encoding));
    if (strcmp(token_call_id->encoding, com_base64) == 0)
    {
      TRC_DETAIL((TB, "Decoding token call id"));
      rc = osp_base64_decode(token_call_id->data,
                             strlen(token_call_id->data),
                             &token_call_id_data,
                             &call_id_len);
      if (rc != OSP_SUCCESS)
      {
        TRC_ERROR((TB, "Failed to decode token's call id"));
        status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
        status->desc = sca_invalid_token;
        goto EXIT_LABEL;
      }
      supplied_call_id_data = call_id->data;
      call_id_len = OSP_MIN(call_id_len, strlen(supplied_call_id_data));
    }
    else if (strcmp(call_id->encoding, com_base64) == 0)
    {
      TRC_DETAIL((TB, "Decoding supplied call id"));
      rc = osp_base64_decode(call_id->data,
                             strlen(call_id->data),
                             &supplied_call_id_data,
                             &call_id_len);
      if (rc != OSP_SUCCESS)
      {
        TRC_ERROR((TB, "Failed to decode supplied call id"));
        status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
        status->desc = sca_invalid_token;
        goto EXIT_LABEL;
      }
      token_call_id_data = token_call_id->data;
      call_id_len = OSP_MIN(call_id_len, strlen(token_call_id_data));
    }
    else
    {
      TRC_ERROR((TB, "CallID Encodings different, but neither is base64"));
      status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
      status->desc = sca_invalid_token;
      goto EXIT_LABEL;
    }
  }

  if (memcmp(token_call_id_data, supplied_call_id_data, call_id_len))
  {
    TRC_ALERT((TB, "Call ids don't match"));
    status->code = OSP_STATUS_CALL_AUTHORIZATION_UNSUCCESSFUL;
    status->desc = sca_invalid_token;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If we get here, all is well.                                            */
  /***************************************************************************/
  TRC_NORMAL((TB, "Authorization confirmed"));

EXIT_LABEL:

  /***************************************************************************/
  /* If anything went wrong, we need to free the token info                  */
  /***************************************************************************/
  if (status->code != OSP_STATUS_SUCCESS)
  {
    TRC_DETAIL((TB, "Freeing token info"));
    sca_free_token_info(token_info);
    token_info = NULL;
  }

  TRC_EXIT();

  return(token_info);

} /* sca_confirm_authorization */
