
/**MOD+***********************************************************************/
/* Module:    scautil.c                                                      */
/*                                                                           */
/* Purpose:   Sample Call Authorization and Routing Component                */
/*            Helper Functions                                               */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.34               $ $Modtime::   Jan 11 2001 15:56:20   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scaincl.h>

/**PROC+**********************************************************************/
/* Name:      sca_dest_in_alt_list                                           */
/*                                                                           */
/* Purpose:   See if supplied destination address is in the supplied list of */
/*            destination alternates                                         */
/*                                                                           */
/* Returns:   TRUE if it is; FALSE otherwise                                 */
/*                                                                           */
/* Params:    IN     address      - the address to search for                */
/*            IN     alt_list     - list of destinations                     */
/*                                                                           */
/* Operation: walk the list of alternates, looking at each address in turn   */
/*                                                                           */
/**PROC-**********************************************************************/

unsigned int sca_dest_in_alt_list(char *address, OSP_ADDRESS *alt_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  unsigned int found = FALSE;
  OSP_ADDRESS *curr_alt;

  TRC_ENTRY("sca_dest_in_alt_list");

  /***************************************************************************/
  /* is it in the client's destination alternate list?                       */
  /***************************************************************************/
  for (curr_alt = alt_list;
       curr_alt != NULL;
       curr_alt = (OSP_ADDRESS *)curr_alt->next)
  {
    TRC_DETAIL((TB, "Checking alternate %s", curr_alt->address));
    if (strcmp(curr_alt->address, address) == 0)
    {
      TRC_DETAIL((TB, "Destination is in alt list"));
      found = TRUE;
      break;
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return(found);

} /* sca_dest_in_alt_list */

/**PROC+**********************************************************************/
/* Name:      sca_add_in_sequence                                            */
/*                                                                           */
/* Purpose:   Add the supplied item into the list, sorting by price then by  */
/*            match length                                                   */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     list         - list into which item should be inserted  */
/*            IN     new_item     - item to insert into list                 */
/*                                                                           */
/* Operation: Walk the list until                                            */
/*            - reach the end, when just add to end                          */
/*            - reach a price that's not less, when                          */
/*              continuing walking the list until                            */
/*              - reach a price that's more, when insert before              */
/*              - item's match is better than current match, when insert     */
/*                before                                                     */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_add_in_sequence(LIST_ENTRY *list,
                         SCA_ROUTE_OPTION *new_item)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  SCA_ROUTE_OPTION *curr_item;
  int added = FALSE;

  TRC_ENTRY("sca_add_in_sequence");

  TRC_DETAIL((TB, "adding item address %s, price %s, match len %d",
                 new_item->ip_address, new_item->price, new_item->best_match));

  /***************************************************************************/
  /* Walk the list till we reach a price that's not lower (or the end!)      */
  /***************************************************************************/
  curr_item = (SCA_ROUTE_OPTION *)LIST_GET_FIRST(list);
  while ((curr_item != NULL) && (curr_item->best_price < new_item->best_price))
  {
    TRC_DETAIL((TB, "Going past item price %s", curr_item->price));
    curr_item = (SCA_ROUTE_OPTION *)LIST_GET_NEXT(list, &curr_item->links);
  }

  /***************************************************************************/
  /* Did we reach the end of the list                                        */
  /***************************************************************************/
  if (curr_item != NULL)
  {
    /*************************************************************************/
    /* No, so we have reached a price that's the same or more.  Keep going   */
    /* until we get a price that is not the same, or a match that's shorter  */
    /*************************************************************************/
    TRC_DETAIL((TB, "Stopped at item price %s", curr_item->price));
    while ((curr_item != NULL) &&
           (curr_item->best_price == new_item->best_price) &&
           (curr_item->best_match > new_item->best_match))
    {
      TRC_DETAIL((TB, "Going past item match len %d", curr_item->best_match));
      curr_item = (SCA_ROUTE_OPTION *)LIST_GET_NEXT(list, &curr_item->links);
    }

    /*************************************************************************/
    /* Did we reach the end of the list                                      */
    /*************************************************************************/
    if (curr_item != NULL)
    {
      /***********************************************************************/
      /* No, so the current item is either more expensive or a worse match.  */
      /* Either way, we insert the new item before it.                       */
      /***********************************************************************/
      TRC_DETAIL((TB, "Got more expensive or worse match - insert before"));
      LIST_INSERT_BEFORE(&curr_item->links, &new_item->links);
      added = TRUE;
    }
  }

  /***************************************************************************/
  /* If we haven't already put the item in the list, then do so now          */
  /***************************************************************************/
  if (!added)
  {
    TRC_DETAIL((TB, "Adding item to end"));
    LIST_INSERT_TAIL(list, &new_item->links);
  }

  TRC_EXIT();

  return;

} /* sca_add_in_sequence */

/**PROC+**********************************************************************/
/* Name:      sca_get_trans_id                                               */
/*                                                                           */
/* Purpose:   Create a transaction id                                        */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     buffer       - Pointer to a buffer with allocated size  */
/*                                  at least SCA_TRANS_ID_LEN, to which the  */
/*                                  next unique transaction ID will be       */
/*                                  written.                                 */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_get_trans_id(char *buffer)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRC_ENTRY("sca_get_trans_id");

  /***************************************************************************/
  /* Now create a transaction id                                             */
  /***************************************************************************/
  OSP_GET_LOCK(&sca_trans_count_mutex);
  OSP_SNPRINTF((buffer, SCA_TRANS_ID_LEN - 1, "%d", sca_trans_count));
  if (sca_trans_count == INT_MAX)
  {
    TRC_DETAIL((TB, "Wrapping trans count at MAX INT"));
    sca_trans_count = 1;
  }
  else
  {
    sca_trans_count++;
  }

  OSP_RELEASE_LOCK(&sca_trans_count_mutex);

  TRC_DETAIL((TB, "Returning trans id %s", buffer));

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* sca_get_trans_id */

/**PROC+**********************************************************************/
/* Name:      sca_price_to_int                                               */
/*                                                                           */
/* Purpose:   Convert a price string in USD/minute to an integer value       */
/*            in 1/10 cents/minute                                           */
/*                                                                           */
/* Returns:   price in 1/10 cents                                            */
/*                                                                           */
/* Params:    IN     price        - price to convert                         */
/*                                                                           */
/* Operation: The price may contain decimal point, so to avoid using floats  */
/*            we convert to 1/10 cent increments.  The value is only used    */
/*            for price comparisons in routing, not in billing, so this      */
/*            restriction on precision is reasonable                         */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_price_to_int(char *price)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int int_price = 0;
  int ii, jj;

  TRC_ENTRY("sca_price_to_int");

  /***************************************************************************/
  /* Implement a simple atoi functionality to convert as far as any decimal  */
  /* point                                                                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Converting price %s", price));
  for (ii = 0;
       (price[ii] >= '0') && (price[ii] <= '9');
       ii++)
  {
    int_price = (10 * int_price) + (price[ii] - '0');
  }
  TRC_DETAIL((TB, "Got dollar part %d", int_price));

  /***************************************************************************/
  /* If we hit a decimal point, skip past it                                 */
  /***************************************************************************/
  if (price[ii] == '.')
  {
    TRC_DETAIL((TB, "Skipping decimal point"));
    ii++;
  }

  /***************************************************************************/
  /* Now convert the dollar value to 1/10 cents (ie multiply by 1000),       */
  /* accruing any decimal places along the way                               */
  /***************************************************************************/
  for (jj = 0; jj < 3; jj++)
  {
    int_price = (10 * int_price);
    if ((price[ii] >= '0') && (price[ii] <= '9'))
    {
      int_price += (price[ii] - '0');
      ii++;
    }
  }
  TRC_DETAIL((TB, "Returning price %d", int_price));

  TRC_EXIT();

  return(int_price);

} /* sca_price_to_int */

/**PROC+**********************************************************************/
/* Name:      sca_int_to_price                                               */
/*                                                                           */
/* Purpose:   Convert an integer value in 1/10 cents/minute to a price       */
/*            in USD/minute                                                  */
/*                                                                           */
/* Returns:   price in USD/Minute                                            */
/*                                                                           */
/* Params:    IN     price        - price in 1/10 cents/minute to convert    */
/*                                                                           */
/**PROC-**********************************************************************/

char *sca_int_to_price(int price)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int len;
  char price_dummy[32];
  char *ret_price;

  TRC_ENTRY("sca_int_to_price");

  /***************************************************************************/
  /* Find out how much space we need                                         */
  /***************************************************************************/
  len = OSP_SNPRINTF((price_dummy, sizeof(price_dummy), "%d.%03d",
                                              (price / 1000), (price % 1000)));
  if (len == 0)
  {
    TRC_ERROR((TB, "Got zero length for price %d", price));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Allow for the null terminator!                                          */
  /***************************************************************************/
  len++;
  TRC_DETAIL((TB, "Need %d bytes for price %d", len, price));

  /***************************************************************************/
  /* Get the memory                                                          */
  /***************************************************************************/
  ret_price = rm_get_mem(len);
  if (ret_price == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for price"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now do the actual printing                                              */
  /***************************************************************************/
  OSP_SNPRINTF((ret_price, len, "%d.%03d", price / 1000, price % 1000));
  TRC_DETAIL((TB, "Converted price %d to %s", price, ret_price));

EXIT_LABEL:
  TRC_EXIT();

  return(ret_price);

} /* sca_int_to_price */

/**PROC+**********************************************************************/
/* Name:      sca_check_price_validity                                       */
/*                                                                           */
/* Purpose:   Check the supplied price string contains a current validity,   */
/*                                                                           */
/* Returns:   SCA_PRICE_CURRENT     - the price is valid and current         */
/*            SCA_PRICE_NOT_FOUND   - the pricing information could not be   */
/*                                    found                                  */
/*            SCA_PRICE_NOT_CURRENT - price is not yet current               */
/*            SCA_PRICE_EXPIRED     - the price is no longer valid           */
/*                                                                           */
/* Params:    IN     price_info   - NULL terminated string of the form       */
/*                                  "<number>@<price>F<from>U<until>"        */
/*            OUT    price_ptr    - string pointer to receive the price      */
/*                                  when extracted from the price_info       */
/*                                                                           */
/* Operation: Extract valid from and valid until times                       */
/*            see if validity has expired,                                   */
/*            see if validity has started,                                   */
/*            Note that no valid from/till information is treated as meaning */
/*            immediate/indefinite                                           */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_check_price_validity(char *price_info, char **price_ptr)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int  rc;
  char *sep_at;
  char *sep_from;
  char *sep_until;
  char *number;
  char *price;
  char *valid_from;
  char *valid_until;

  TRC_ENTRY("sca_check_price_validity");

  /***************************************************************************/
  /* Check parameters                                                        */
  /***************************************************************************/
  TRC_ASSERT((price_info != NULL), (TB, "Null price info supplied"));
  TRC_ASSERT((price_ptr != NULL), (TB, "Null price ptr supplied"));

  /***************************************************************************/
  /* Get the consituent parts of the pricing information, number first       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Parsing price info '%s'", price_info));
  number = price_info;
  sep_at = strchr(price_info, '@');
  if (sep_at == NULL)
  {
    TRC_ALERT((TB, "Supplied price info had no @ separator"));
    rc = SCA_PRICE_NOT_FOUND;
    goto EXIT_LABEL;
  }
  *sep_at = 0;
  TRC_DETAIL((TB, "Got number part %s", number));

  /***************************************************************************/
  /* The price starts after the @ separator                                  */
  /***************************************************************************/
  price = sep_at + 1;
  *price_ptr = price;
  sep_from = strchr(price, 'F');
  if (sep_from == NULL)
  {
    TRC_DETAIL((TB, "No from separator; assume immediate/indefinite"));
    rc = SCA_PRICE_CURRENT;
    goto EXIT_LABEL;
  }
  *sep_from = 0;
  TRC_DETAIL((TB, "Got price part %s", price));

  /***************************************************************************/
  /* Valid from next                                                         */
  /***************************************************************************/
  valid_from = sep_from + 1;
  sep_until = strchr(valid_from, 'U');
  if (sep_until == NULL)
  {
    TRC_DETAIL((TB, "No until separator; assume immediate/indefinite"));
    rc = SCA_PRICE_CURRENT;
    goto EXIT_LABEL;
  }
  *sep_until = 0;
  TRC_DETAIL((TB, "Got from part %s", valid_from));

  /***************************************************************************/
  /* The last section is the valid until                                     */
  /***************************************************************************/
  valid_until = sep_until + 1;
  TRC_DETAIL((TB, "Got until part %s", valid_until));

  /***************************************************************************/
  /* Check the prices are still current                                      */
  /***************************************************************************/
  rc = sca_check_validity(valid_from, valid_until);
  if (rc == SCA_VALIDITY_NOT_VALID)
  {
    TRC_DETAIL((TB, "Price not valid"));
    rc = SCA_PRICE_NOT_FOUND;
  }
  else if (rc == SCA_VALIDITY_EXPIRED)
  {
    TRC_DETAIL((TB, "Price has expired"));
    *sep_at = '@';
    *sep_from = 'F';
    *sep_until = 'U';
    rc = SCA_PRICE_EXPIRED;
  }
  else if (rc == SCA_VALIDITY_NOT_YET)
  {
    TRC_DETAIL((TB, "Price not yet current"));
    rc = SCA_PRICE_NOT_CURRENT;
  }
  else
  {
    TRC_DETAIL((TB, "All OK - price is current"));
    rc = SCA_PRICE_CURRENT;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sca_check_price_validity */

/**PROC+**********************************************************************/
/* Name:      sca_check_validity                                             */
/*                                                                           */
/* Purpose:   Check the supplied validity range is current                   */
/*                                                                           */
/* Returns:   SCA_VALIDITY_OK        - validity is current                   */
/*            SCA_VALIDITY_EXPIRED   - validity has expired                  */
/*            SCA_VALIDITY_NOT_YET   - validity has not yet started          */
/*            SCA_VALIDITY_NOT_VALID - validity info is itself invalid       */
/*                                                                           */
/* Params:    IN     valid_from   - valid from time                          */
/*            IN     valid_until  - valid until time                         */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_check_validity(char *valid_from, char *valid_until)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = SCA_VALIDITY_NOT_VALID;
  time_t curr_time;
  struct tm gm_time;
  char current_time[21];
  int compare_rc;
  int time_len;
  int valid_len;

  TRC_ENTRY("sca_check_validity");

  /***************************************************************************/
  /* Valid from being NULL or empty means immediate validity                 */
  /***************************************************************************/
  if ((valid_from == NULL) || (strcmp(valid_from, "") == 0))
  {
    TRC_DETAIL((TB, "Valid from immediate"));
    valid_from = NULL;
  }

  /***************************************************************************/
  /* Valid until being NULL or empty means indefinite validity               */
  /***************************************************************************/
  if ((valid_until == NULL) || (strcmp(valid_until, "") == 0))
  {
    TRC_DETAIL((TB, "Valid until is indefinite"));
    valid_until = NULL;
  }

  /***************************************************************************/
  /* Get the current time                                                    */
  /***************************************************************************/
  time(&curr_time);
  gmtime_r(&curr_time, &gm_time);
  strftime(current_time, 21, "%Y-%m-%dT%H:%M:%SZ", &gm_time);
  TRC_DETAIL((TB, "Current time is %s", current_time));

  /***************************************************************************/
  /* Unless we already know the validity is immediate and/or indefinite,     */
  /* check the validities are the same length as the current time - we can't */
  /* meaningfully compare them if not, and will assume they are not valid.   */
  /***************************************************************************/
  time_len = strlen(current_time);

  if (valid_until != NULL)
  {
    TRC_DETAIL((TB, "Checking until len"));
    valid_len = strlen(valid_until);
    if (time_len != valid_len)
    {
      TRC_ALERT((TB, "valid to time '%s' not valid - assume not valid",
                                                                 valid_until));
      goto EXIT_LABEL;
    }
  }

  if (valid_from != NULL)
  {
    TRC_DETAIL((TB, "Checking from len"));
    valid_len = strlen(valid_from);
    if (time_len != valid_len)
    {
      TRC_ALERT((TB, "valid from time '%s' not valid - assume not valid",
                                                                  valid_from));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Check if the validity has expired                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Checking still valid..."));
  compare_rc = strcmp(current_time, valid_until);
  if (compare_rc > 0)
  {
    TRC_ALERT((TB, "Until time has passed"));
    rc = SCA_VALIDITY_EXPIRED;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Until time is still valid"));

  /***************************************************************************/
  /* Check if the validity has started yet                                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Checking validity started..."));
  compare_rc = strcmp(current_time, valid_from);
  if (compare_rc < 0)
  {
    TRC_ALERT((TB, "From time not yet reached"));
    rc = SCA_VALIDITY_NOT_YET;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If we get here, all is well and the validity is current                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "All OK - validity is current"));
  rc = SCA_VALIDITY_OK;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sca_check_validity */

/**PROC+**********************************************************************/
/* Name:      sca_copy_address                                               */
/*                                                                           */
/* Purpose:   Copy an OSP_ADDRESS structure                                  */
/*                                                                           */
/* Returns:   Pointer to copy or NULL                                        */
/*                                                                           */
/* Params:    IN     address      - address to copy                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_ADDRESS *sca_copy_address(OSP_ADDRESS *address)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_ADDRESS *copy = NULL;

  TRC_ENTRY("sca_copy_address");

  /***************************************************************************/
  /* Check what we were asked to copy                                        */
  /***************************************************************************/
  if (address == NULL)
  {
    TRC_ERROR((TB, "Asked to copy NULL address"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get some memory for a copy                                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Copying address %s", address->address));
  copy = (OSP_ADDRESS *)rm_get_mem(sizeof(OSP_ADDRESS));
  if (copy == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for address copy"));
    goto EXIT_LABEL;
  }
  copy->next = NULL;

  /***************************************************************************/
  /* Copy the data                                                           */
  /***************************************************************************/
  OSP_STRDUP(copy->type, address->type);
  if (copy->type == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for address type"));
    rm_release_mem(copy);
    copy = NULL;
    goto EXIT_LABEL;
  }

  OSP_STRDUP(copy->address, address->address);
  if (copy->address == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for address data"));
    rm_release_mem(copy->type);
    rm_release_mem(copy);
    copy = NULL;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(copy);

} /* sca_copy_address */

/**PROC+**********************************************************************/
/* Name:      sca_copy_call_id                                               */
/*                                                                           */
/* Purpose:   Copy call id                                                   */
/*                                                                           */
/* Returns:   Pointer to the copy or NULL                                    */
/*                                                                           */
/* Params:    IN     call_id      - call id to copy                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_CALL_ID *sca_copy_call_id(OSP_CALL_ID *call_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_CALL_ID *copy = NULL;

  TRC_ENTRY("sca_copy_call_id");

  /***************************************************************************/
  /* Check what we were asked to copy                                        */
  /***************************************************************************/
  if (call_id == NULL)
  {
    TRC_ERROR((TB, "Asked to copy NULL call id"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get some memory for a copy                                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Copying call_id %p", call_id));
  copy = (OSP_CALL_ID *)rm_get_mem(sizeof(OSP_CALL_ID));
  if (copy == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for call_id copy"));
    goto EXIT_LABEL;
  }
  copy->next = NULL;

  /***************************************************************************/
  /* Copy the data                                                           */
  /***************************************************************************/
  OSP_STRDUP(copy->encoding, call_id->encoding);
  if (copy->encoding == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for call_id encoding"));
    rm_release_mem(copy);
    copy = NULL;
    goto EXIT_LABEL;
  }

  OSP_STRDUP(copy->data, call_id->data);
  if (copy->data == NULL)
  {
    TRC_ERROR((TB, "Failed to get mem for call_id data"));
    rm_release_mem(copy->encoding);
    rm_release_mem(copy);
    copy = NULL;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(copy);

} /* sca_copy_call_id */

/**PROC+**********************************************************************/
/* Name:      sca_free_dest                                                  */
/*                                                                           */
/* Purpose:   Free all the memory in an OSP destination                      */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     dest         - destination to free                      */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_dest(OSP_DESTINATION *dest)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sca_free_dest");

  /***************************************************************************/
  /* Check what we've been passed                                            */
  /***************************************************************************/
  if (dest == NULL)
  {
    TRC_NORMAL((TB, "Passed NULL destination to free"));
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Freeing osp destination %p", dest));

  /***************************************************************************/
  /* There ought to be a destination info, but we may be freeing a partially */
  /* allocated structure in an error path                                    */
  /***************************************************************************/
  if (dest->info != NULL)
  {
    rm_release_mem(dest->info->type);
    rm_release_mem(dest->info->address);
    rm_release_mem(dest->info);
  }

  /***************************************************************************/
  /* There should never be any alternates                                    */
  /***************************************************************************/
  TRC_ASSERT((dest->alt_list == NULL), (TB, "Non-NULL alternate list"));

  /***************************************************************************/
  /* Signal address is simply a string                                       */
  /***************************************************************************/
  rm_release_mem(dest->signal_address);

  /***************************************************************************/
  /* Again, there should be a token, but we may be in an error path          */
  /***************************************************************************/
  if (dest->token_list != NULL)
  {
    /*************************************************************************/
    /* We only ever assign one token, so there's no list to free.  That      */
    /* said, we'll still make sure.                                          */
    /*************************************************************************/
    TRC_ASSERT((dest->token_list->next == NULL), (TB, "More than one token"));
    rm_release_mem(dest->token_list->encoding);
    rm_release_mem(dest->token_list->token_data);
    rm_release_mem(dest->token_list);
  }

  /***************************************************************************/
  /* Free validities                                                         */
  /***************************************************************************/
  rm_release_mem(dest->valid_after);
  rm_release_mem(dest->valid_until);

  /***************************************************************************/
  /* The authority URL list should be NULL                                   */
  /***************************************************************************/
  TRC_ASSERT((dest->auth_url_list == NULL), (TB, "Non-NULL auth url list"));

  /***************************************************************************/
  /* There might be a usage detail                                           */
  /***************************************************************************/
  if (dest->usage_list != NULL)
  {
    /*************************************************************************/
    /* There should be only one                                              */
    /*************************************************************************/
    TRC_ASSERT((dest->usage_list->next == NULL), (TB, "More than one usage!"));
    TRC_DETAIL((TB, "Freeing usage detail elements"));
    rm_release_mem(dest->usage_list->service);
    rm_release_mem(dest->usage_list->amount);
    rm_release_mem(dest->usage_list->increment);
    rm_release_mem(dest->usage_list->unit);
    rm_release_mem(dest->usage_list->start_time);
    rm_release_mem(dest->usage_list->end_time);

    /*************************************************************************/
    /* These two should always be NULL                                       */
    /*************************************************************************/
    TRC_ASSERT((dest->usage_list->term_cause == NULL),
                            (TB, "Usage with non-NULL term cause in routing"));
    TRC_ASSERT((dest->usage_list->statistics == NULL),
                                 (TB, "Usage with non-NULL stats in routing"));

    /*************************************************************************/
    /* Free the usage structure itself                                       */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing usage detail %p", dest->usage_list));
    rm_release_mem(dest->usage_list);
  }

  /***************************************************************************/
  /* As before, there ought to be a call id, but we might be in an error     */
  /* path so check                                                           */
  /***************************************************************************/
  if (dest->call_id != NULL)
  {
    rm_release_mem(dest->call_id->encoding);
    rm_release_mem(dest->call_id->data);
    rm_release_mem(dest->call_id);
  }

  /***************************************************************************/
  /* Finally release the structure                                           */
  /***************************************************************************/
  rm_release_mem(dest);

EXIT_LABEL:
  TRC_EXIT();

  return;

} /* sca_free_dest */

/**PROC+**********************************************************************/
/* Name:      sca_free_address                                               */
/*                                                                           */
/* Purpose:   Free the memory in an OSP_ADDRESS                              */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     address      - address to free                          */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_address(OSP_ADDRESS *address)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sca_free_address");

  /***************************************************************************/
  /* Check what we've been given                                             */
  /***************************************************************************/
  if (address != NULL)
  {
    rm_release_mem(address->type);
    rm_release_mem(address->address);
    rm_release_mem(address);
  }

  TRC_EXIT();

  return;

} /* sca_free_address */

/**PROC+**********************************************************************/
/* Name:      sca_free_call_id                                               */
/*                                                                           */
/* Purpose:   Free the memory in an OSP_CALL_ID                              */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     call_id      - call_id to free                          */
/*                                                                           */
/**PROC-**********************************************************************/

void sca_free_call_id(OSP_CALL_ID *call_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sca_free_call_id");

  /***************************************************************************/
  /* Check what we've been given                                             */
  /***************************************************************************/
  if (call_id != NULL)
  {
    rm_release_mem(call_id->encoding);
    rm_release_mem(call_id->data);
    rm_release_mem(call_id);
  }

  TRC_EXIT();

  return;

} /* sca_free_call_id */

/**PROC+**********************************************************************/
/* Name:      sca_find_subs_info                                             */
/*                                                                           */
/* Purpose:   Search the supplied source alternate list for subscriber       */
/*            information                                                    */
/*                                                                           */
/* Returns:   pointer to first subscriber information found, or NULL         */
/*                                                                           */
/* Params:    IN     src_alt_list - list of source alternates to search      */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_ADDRESS *sca_find_subs_info(OSP_ADDRESS *src_alt_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_ADDRESS *curr_alt;

  TRC_ENTRY("sca_find_subs_info");

  /***************************************************************************/
  /* Walk the list, checking the address type                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Checking alternate list for '%s'", sca_subscriber));
  curr_alt = src_alt_list;
  while (curr_alt != NULL)
  {
    TRC_DETAIL((TB, "Checking %s", curr_alt->type));
    if (!strcmp(curr_alt->type, sca_subscriber))
    {
      TRC_DETAIL((TB, "Found subscriber!"));
      break;
    }
    curr_alt = (OSP_ADDRESS *)curr_alt->next;
  }

  TRC_EXIT();

  return(curr_alt);

} /* sca_find_subs_info */

/**PROC+**********************************************************************/
/* Name:      sca_match_len                                                  */
/*                                                                           */
/* Purpose:   Find the length of the prefix of test_string which matches     */
/*            reference_string                                               */
/*                                                                           */
/* Returns:   length of match                                                */
/*                                                                           */
/* Params:    IN     test_string  - string to test                           */
/*            IN     reference_string                                        */
/*                                - string to test against                   */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_match_len(char *test_string, char *reference_string)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int match_len = 0;

  TRC_ENTRY("sca_match_len");

  TRC_ASSERT((test_string != NULL), (TB, "Passed NULL test string"));
  TRC_ASSERT((reference_string != NULL), (TB, "Passed NULL reference string"));
  TRC_DETAIL((TB, "Comparing %s with %s", test_string, reference_string));

  /***************************************************************************/
  /* Walk the strings until                                                  */
  /* - we reach the end of either                                            */
  /* - they differ                                                           */
  /***************************************************************************/
  while ((test_string[match_len] != 0) &&
         (reference_string[match_len] != 0) &&
         (test_string[match_len] == reference_string[match_len]))
  {
    match_len++;
  }

  TRC_DETAIL((TB, "Got match length %d", match_len));

  TRC_EXIT();

  return(match_len);

} /* sca_match_len */

/**PROC+**********************************************************************/
/* Name:      sca_add_validity                                               */
/*                                                                           */
/* Purpose:   Adds valid after/valid until elements to the supplied string   */
/*            (for the token) and osp_destination                            */
/*                                                                           */
/* Returns:   OSP_SUCCESS - validities added successfully                    */
/*                                                                           */
/* Params:    IN     dest         - destination string                       */
/*            IN     dest_len     - length of dest                           */
/*            IN     spread       - spread of validity in seconds            */
/*            IN     osp_dest     - destination to receive validity          */
/*                                                                           */
/* Operation: Get the current time                                           */
/*            Subtract the spread to get the from time                       */
/*            Convert it to the correct format to get the from time          */
/*            Add the spread to the current time                             */
/*            Convert it to the correct format to get the until time         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sca_add_validity(char *dest,
                        int dest_len,
                        time_t spread,
                        OSP_DESTINATION *osp_dest)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERR_MEMORY;
  time_t curr_time;
  time_t from_time;
  struct tm gm_time;
  char time_string[21];
  int ret_count;
  char xml[128];
  int xml_len;

  TRC_ENTRY("sca_add_validity");

  /***************************************************************************/
  /* Initial setup                                                           */
  /***************************************************************************/
  if (osp_dest != NULL)
  {
    osp_dest->valid_after = NULL;
    osp_dest->valid_until = NULL;
  }

  /***************************************************************************/
  /* Get the current time                                                    */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting time..."));
  time(&curr_time);

  /***************************************************************************/
  /* Subtract the spread to get the start time                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Spread is %d", spread));
  from_time = curr_time - spread;
  gmtime_r(&from_time, &gm_time);

  /***************************************************************************/
  /* Build the valid after time                                              */
  /***************************************************************************/
  ret_count = strftime(time_string, 21, "%Y-%m-%dT%H:%M:%SZ", &gm_time);
  TRC_ASSERT((ret_count != 0), (TB, "Formatted time didn't fit in array"));

  TRC_DETAIL((TB, "Valid after %s", time_string));
  xml_len = OSP_SNPRINTF((xml,
                          sizeof(xml),
                          "<%s>\n%s\n</%s>\n",
                          scm_valid_after,
                          time_string,
                          scm_valid_after));
  TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
  TRC_ASSERT(((strlen(dest) + xml_len) < dest_len),
                                        (TB, "New xml overflows dest buffer"));
  strcat(dest, xml);

  /***************************************************************************/
  /* Copy it into the destination, if supplied                               */
  /***************************************************************************/
  if (osp_dest != NULL)
  {
    TRC_DETAIL((TB, "Copying validity to osp dest"));
    OSP_STRDUP(osp_dest->valid_after, time_string)
    if (osp_dest->valid_after == NULL)
    {
      TRC_ERROR((TB, "Failed to alloc memory for validity"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Add on the spread to the current time                                   */
  /***************************************************************************/
  curr_time += spread;
  gmtime_r(&curr_time, &gm_time);

  /***************************************************************************/
  /* And build the valid until time                                          */
  /***************************************************************************/
  ret_count = strftime(time_string, 21, "%Y-%m-%dT%H:%M:%SZ", &gm_time);
  TRC_ASSERT((ret_count != 0), (TB, "Formatted time didn't fit in array"));

  TRC_DETAIL((TB, "Valid until %s", time_string));
  xml_len = OSP_SNPRINTF((xml,
                          sizeof(xml),
                          "<%s>\n%s\n</%s>\n",
                          scm_valid_until,
                          time_string,
                          scm_valid_until));
  TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));
  TRC_ASSERT(((strlen(dest) + xml_len) < dest_len),
                                        (TB, "New xml overflows dest buffer"));
  strcat(dest, xml);

  /***************************************************************************/
  /* Copy it into the destination, if supplied                               */
  /***************************************************************************/
  if (osp_dest != NULL)
  {
    TRC_DETAIL((TB, "Copying validity to osp dest"));
    OSP_STRDUP(osp_dest->valid_until, time_string)
    if (osp_dest->valid_until == NULL)
    {
      TRC_ERROR((TB, "Failed to alloc memory for validity"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* All went OK                                                             */
  /***************************************************************************/
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sca_add_validity */

/**PROC+**********************************************************************/
/* Name:      sca_copy_dest_alt_list                                         */
/*                                                                           */
/* Purpose:   Copy a list of dest alternates, dropping those which are not   */
/*            of type transport                                              */
/*                                                                           */
/* Returns:   TRUE if successful, FALSE otherwise                            */
/*                                                                           */
/* Params:    IN     source_list  - list to copy                             */
/*            IN     dest_ptr     - pointer to destination list              */
/*            IN     copy_fn      - function to use to copy each member      */
/*            IN     free_fn      - function to use to free any allocated    */
/*                                  copies in case of failure                */
/*                                                                           */
/**PROC-**********************************************************************/

int sca_copy_dest_alt_list(SLIST_ENTRY *source_list,
                           SLIST_ENTRY **dest_ptr,
                           PCOPY_ENTRY_FN copy_fn,
                           PFREE_ENTRY_FN free_fn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  SLIST_ENTRY *copy;
  SLIST_ENTRY *curr_entry;
  OSP_ADDRESS *curr_alt;
  SLIST_ROOT copied_list;

  TRC_ENTRY("sca_copy_dest_alt_list");

  /***************************************************************************/
  /* Initialize the copied items list                                        */
  /***************************************************************************/
  SL_INIT_ROOT(copied_list);

  /***************************************************************************/
  /* Walk the source list                                                    */
  /***************************************************************************/
  for (curr_entry = source_list;
       curr_entry != NULL;
       curr_entry = (SLIST_ENTRY *)curr_entry->next)
  {
    /*************************************************************************/
    /* If its not of type transport, then skip it                            */
    /*************************************************************************/
    curr_alt = (OSP_ADDRESS *)curr_entry;
    if (strcmp(curr_alt->type, sca_transport) != 0)
    {
      TRC_DETAIL((TB, "Dropping dest alt type %s", curr_alt->type));
      continue;
    }

    /*************************************************************************/
    /* Make a copy of the current call id                                    */
    /*************************************************************************/
    TRC_DETAIL((TB, "Copying entry %p", curr_entry));
    copy = (*copy_fn)(curr_entry);
    if (copy == NULL)
    {
      /***********************************************************************/
      /* Free any we did allocate                                            */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to copy entry - freeing what we've got"));
      scm_free_list(SL_HEAD(copied_list), free_fn);
      rc = FALSE;
      break;
    }

    /*************************************************************************/
    /* Add the copy to the list                                              */
    /*************************************************************************/
    SL_ADD_TO_TAIL(copied_list, copy);
  }

  /***************************************************************************/
  /* If all is ok, set up the return pointer                                 */
  /***************************************************************************/
  *dest_ptr = copied_list.first;

  TRC_EXIT();

  return(rc);

} /* sca_copy_dest_alt_list */

