
/**MOD+***********************************************************************/
/* Module:    scmlist.c                                                      */
/*                                                                           */
/* Purpose:   Sample Application Common list management routines             */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.3                $ $Modtime::   Jan 11 2001 15:56:20   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scmincl.h>

/**PROC+**********************************************************************/
/* Name:      scm_copy_list                                                  */
/*                                                                           */
/* Purpose:   Copy a list of structures, using the supplied function to copy */
/*            each structure.  The first element of the structure must be a  */
/*            pointer to the next structure in the list                      */
/*                                                                           */
/* Returns:   TRUE if successful, FALSE otherwise                            */
/*                                                                           */
/* Params:    IN     source_list  - list to copy                             */
/*            IN     dest_ptr     - pointer to destination list              */
/*            IN     copy_fn      - function to use to copy each member      */
/*            IN     free_fn      - function to use to free any allocated    */
/*                                  copies in case of failure                */
/*                                                                           */
/**PROC-**********************************************************************/

int scm_copy_list(SLIST_ENTRY *source_list,
                  SLIST_ENTRY **dest_ptr,
                  PCOPY_ENTRY_FN copy_fn,
                  PFREE_ENTRY_FN free_fn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  SLIST_ENTRY *copy;
  SLIST_ENTRY *curr_entry;
  SLIST_ROOT copied_list;

  TRC_ENTRY("scm_copy_list");

  /***************************************************************************/
  /* Initialize the copied items list                                        */
  /***************************************************************************/
  SL_INIT_ROOT(copied_list);

  /***************************************************************************/
  /* Walk the source list                                                    */
  /***************************************************************************/
  for (curr_entry = source_list;
       curr_entry != NULL;
       curr_entry = (SLIST_ENTRY *)curr_entry->next)
  {
    /*************************************************************************/
    /* Make a copy of the current call id                                    */
    /*************************************************************************/
    TRC_DETAIL((TB, "Copying entry %p", curr_entry));
    copy = (*copy_fn)(curr_entry);
    if (copy == NULL)
    {
      /***********************************************************************/
      /* Free any we did allocate                                            */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to copy entry - freeing what we've got"));
      scm_free_list(SL_HEAD(copied_list), free_fn);
      rc = FALSE;
      break;
    }

    /*************************************************************************/
    /* Add the copy to the list                                              */
    /*************************************************************************/
    SL_ADD_TO_TAIL(copied_list, copy);
  }

  /***************************************************************************/
  /* If all is ok, set up the return pointer                                 */
  /***************************************************************************/
  *dest_ptr = copied_list.first;

  TRC_EXIT();

  return(rc);

} /* scm_copy_list */

/**PROC+**********************************************************************/
/* Name:      scm_free_list                                                  */
/*                                                                           */
/* Purpose:   Free all the members in a list of structures, using the        */
/*            supplied function to free the individual members.  The first   */
/*            element of the structure must be a pointer to the next in the  */
/*            list.                                                          */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     list         - list to free                             */
/*            IN     free_fn      - function to use to free each member      */
/*                                                                           */
/**PROC-**********************************************************************/

void scm_free_list(SLIST_ENTRY *list, PFREE_ENTRY_FN free_fn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  SLIST_ENTRY *curr_entry, *next_entry;

  TRC_ENTRY("scm_free_list");

  /***************************************************************************/
  /* Check what we got                                                       */
  /***************************************************************************/
  TRC_ASSERT((free_fn != NULL), (TB, "Passed NULL ptr to free fn"));
  TRC_ASSERT((*free_fn != NULL), (TB, "Passed NULL free fn"));

  /***************************************************************************/
  /* Walk the list, freeing each item                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing list..."));
  curr_entry = list;
  while (curr_entry != NULL)
  {
    next_entry = (SLIST_ENTRY *)curr_entry->next;
    (*free_fn)(curr_entry);
    curr_entry = next_entry;
  }

  TRC_EXIT();

  return;

} /* scm_free_list */
