
/**MOD+***********************************************************************/
/* Module:    suamain.c                                                      */
/*                                                                           */
/* Purpose:   Sample USE application - code                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.17               $ $Modtime::   Jan 11 2001 15:56:22   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <suaincl.h>

/**PROC+**********************************************************************/
/* Name:      sua_init                                                       */
/*                                                                           */
/* Purpose:   Initialize USE component of sample application                 */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - initialized successfully                    */
/*            OSP_ERR_ALREADY_INIT                                           */
/*                             - USE component was already initialized       */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sua_init(LHASH *config)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  char *value;

  TRC_ENTRY("sua_init");

  /***************************************************************************/
  /* Check configuration parameter.                                          */
  /***************************************************************************/
  TRC_ASSERT((config != NULL), (TB, "NULL configuration parameter"));

  /***************************************************************************/
  /* Check for previous component initialization, to avoid initializing      */
  /* twice.                                                                  */
  /***************************************************************************/
  TRC_DETAIL((TB, "USE component initialized flag: %d", sua_initialized));
  if (!sua_initialized)
  {
    TRC_NORMAL((TB, "Initialize USE component of sample application"));

    /*************************************************************************/
    /* Create the record file access mutex.                                  */
    /*************************************************************************/
    rc = osp_create_mutex(&sua_file_mutex);
    TRC_DETAIL((TB, "Mutex creation: rc=%d", rc));
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to create file access mutex: rc=%d", rc));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Initialize global variables.                                          */
    /*************************************************************************/
    sua_initialized = TRUE;
    sua_file_name = NULL;
    sua_max_records = SUA_DEFAULT_MAX_RECORDS;
    sua_file = NULL;
    sua_records_written = 0;

    /*************************************************************************/
    /* Read the OSP config file to discover the desired location of the      */
    /* usage record file.                                                    */
    /*************************************************************************/
    value = CONF_get_string(config,
                            SCM_CONFIG_SECTION,
                            SUA_CONFIG_RECORD_FILE);
    if (value != NULL)
    {
      TRC_NORMAL((TB, "Record file name: \"%s\"", value));
      OSP_STRDUP(sua_file_name, value);
      if (sua_file_name == NULL)
      {
        TRC_ALERT((TB, "Couldn't allocate memory for file name"));
      }
    }
    else
    {
      TRC_ALERT((TB, "Couldn't read record file name from config"));
    }

    if (sua_file_name != NULL)
    {
      TRC_NORMAL((TB, "Record file name: \"%s\"", sua_file_name));
    }
    else
    {
      TRC_ALERT((TB, "Couldn't read record file name from config"));
    }

    value = CONF_get_string(config,
                            SCM_CONFIG_SECTION,
                            SUA_CONFIG_MAX_RECORDS);
    if ((value != NULL) && (sscanf(value, "%d", &sua_max_records) == 1))
    {
      TRC_NORMAL((TB, "Maximum number of records: %d", sua_max_records));
    }
    else
    {
      TRC_NORMAL((TB, "Default maximum number of records"));
      sua_max_records = SUA_DEFAULT_MAX_RECORDS;
    }

    /*************************************************************************/
    /* If we successfully established a record file name above, now see if   */
    /* we can access it.                                                     */
    /*************************************************************************/
    if (sua_file_name != NULL)
    {
      TRC_NORMAL((TB, "Usage record file name: \"%s\"", sua_file_name));

      /***********************************************************************/
      /* Try opening this file name for write access.                        */
      /***********************************************************************/
      OSP_GET_LOCK(&sua_file_mutex);
      sua_file = fopen(sua_file_name, "w");
      if (sua_file != NULL)
      {
        TRC_NORMAL((TB, "File opened successfully: %p", sua_file));

        /*********************************************************************/
        /* Since we managed to open the usage record file, it makes sense to */
        /* register our usage callback.  The only way that osp_um_register   */
        /* call fail is if the OSP Server stack has not yet been             */
        /* initialized.  In this case we might as well assert - the          */
        /* application code must be corrected so that osp_init is called     */
        /* earlier on.                                                       */
        /*********************************************************************/
        rc = osp_um_register(sua_cb);
        TRC_ASSERT((rc == OSP_SUCCESS),
                   (TB, "osp_um_register call failed: rc=%d", rc));
      }
      else
      {
        TRC_OS_ERROR((TB, "Couldn't open usage record file"));
      }
      OSP_RELEASE_LOCK(&sua_file_mutex);
    }
  }
  else
  {
    TRC_ALERT((TB, "USE component is already initialized"));
    rc = OSP_ERR_ALREADY_INIT;
  }

EXIT_LABEL:
  TRC_EXIT();
  return(rc);
}

/**PROC+**********************************************************************/
/* Name:      sua_term                                                       */
/*                                                                           */
/* Purpose:   Terminate USE component of sample application                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - terminated successfully                     */
/*            OSP_ERR_NOT_INIT - USE component was not initialized           */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sua_term()
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("sua_term");

  /***************************************************************************/
  /* Check that we have previously been initialized.                         */
  /***************************************************************************/
  TRC_DETAIL((TB, "USE component initialized flag: %d", sua_initialized));
  if (sua_initialized)
  {
    TRC_NORMAL((TB, "Terminate USE component of sample application"));

    /*************************************************************************/
    /* The sua_file variable indicates whether or not the usage record file  */
    /* is currently open, and hence needs closing, and also whether our      */
    /* usage callback is currently registered.                               */
    /*************************************************************************/
    OSP_GET_LOCK(&sua_file_mutex);
    if (sua_file != NULL)
    {
      TRC_NORMAL((TB, "Current usage record file handle: %p", sua_file));

      /***********************************************************************/
      /* Unregister our usage callback.  There are no documented ways for    */
      /* this call to fail.                                                  */
      /***********************************************************************/
      rc = osp_um_register(NULL);
      TRC_ASSERT((rc == OSP_SUCCESS),
                 (TB, "osp_um_register call failed: rc=%d", rc));

      /***********************************************************************/
      /* Close the usage record file.                                        */
      /***********************************************************************/
      if (fclose(sua_file) != 0)
      {
        TRC_OS_ERROR((TB, "Couldn't close usage record file"));
      }
      sua_file = NULL;
    }
    OSP_RELEASE_LOCK(&sua_file_mutex);

    /*************************************************************************/
    /* Terminate remaining global variables.                                 */
    /*************************************************************************/
    sua_initialized = FALSE;
    sua_records_written = 0;

    if (sua_file_name != NULL)
    {
      rm_release_mem(sua_file_name);
      sua_file_name = NULL;
    }

    /*************************************************************************/
    /* Destroy the file access mutex.                                        */
    /*************************************************************************/
    osp_destroy_mutex(&sua_file_mutex);
  }
  else
  {
    TRC_ALERT((TB, "USE component has not been initialized"));
    rc = OSP_ERR_NOT_INIT;
  }

  TRC_EXIT();
  return(rc);
}

/**PROC+**********************************************************************/
/* Name:      sua_cb                                                         */
/*                                                                           */
/* Purpose:   Handles UsageIndication information from OpenOSP Server        */
/*                                                                           */
/*            For parameters and return values, see the description of       */
/*            POSP_UM_CALLBACK in <openosp.h>.                               */
/*                                                                           */
/* Operation: sua_cb writes a usage record to the record file for            */
/*            each item in the _use_detail_list_.                            */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI sua_cb(OSP_CORRELATOR corr,
                     char *timestamp,
                     char *role,
                     char *transaction_id,
                     OSP_CALL_ID *call_id,
                     OSP_ADDRESS *source_info,
                     OSP_ADDRESS *source_alt_list,
                     OSP_ADDRESS *dest_info,
                     OSP_ADDRESS *dest_alt_list,
                     OSP_USAGE_DETAIL *use_detail_list,
                     OSP_CLIENT_ID *client_id)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_ADDRESS *alt_entry;
  OSP_USAGE_DETAIL *use_entry;
  OSP_STATUS status;

  TRC_ENTRY("sua_cb");

  /***************************************************************************/
  /* Verify and trace input parameters.                                      */
  /***************************************************************************/
  TRC_DETAIL((TB, "Correlator: %p", corr));

  TRC_ASSERT((timestamp != NULL),
             (TB, "Timestamp parameter is NULL"));
  TRC_DETAIL((TB, "Timestamp: %s", timestamp));

  TRC_ASSERT((role != NULL),
             (TB, "Role parameter is NULL"));
  TRC_DETAIL((TB, "Role: %s", role));

  TRC_ASSERT((transaction_id != NULL),
             (TB, "Transaction ID parameter is NULL"));
  TRC_DETAIL((TB, "Transaction ID: %s", transaction_id));

  TRC_ASSERT((call_id != NULL) &&
             (call_id->encoding != NULL) &&
             (call_id->data != NULL),
             (TB, "Call ID parameter (or one of its fields) is NULL"));
  TRC_DETAIL((TB, "Call ID encoding: %s", call_id->encoding));
  TRC_DETAIL((TB, "Call ID data: %s", call_id->data));

  TRC_ASSERT((source_info != NULL) &&
             (source_info->type != NULL) &&
             (source_info->address != NULL),
             (TB, "Source info parameter (or one of its fields) is NULL"));
  TRC_DETAIL((TB, "Source info type: %s", source_info->type));
  TRC_DETAIL((TB, "Source info address: %s", source_info->address));

  for (alt_entry = source_alt_list;
       alt_entry != NULL;
       alt_entry = alt_entry->next)
  {
    TRC_ASSERT((alt_entry->type != NULL) &&
               (alt_entry->address != NULL),
               (TB, "Source alternate info field is NULL"));
    TRC_DETAIL((TB, "Source alternate info type: %s", alt_entry->type));
    TRC_DETAIL((TB, "Source alternate info address: %s", alt_entry->address));
  }

  TRC_ASSERT((dest_info != NULL) &&
             (dest_info->type != NULL) &&
             (dest_info->address != NULL),
             (TB,"Destination info parameter (or one of its fields) is NULL"));
  TRC_DETAIL((TB, "Destination info type: %s", dest_info->type));
  TRC_DETAIL((TB, "Destination info address: %s", dest_info->address));

  for (alt_entry = dest_alt_list;
       alt_entry != NULL;
       alt_entry = alt_entry->next)
  {
    TRC_ASSERT((alt_entry->type != NULL) &&
               (alt_entry->address != NULL),
               (TB, "Destination alternate info field is NULL"));
    TRC_DETAIL((TB, "Destination alternate info type: %s", alt_entry->type));
    TRC_DETAIL((TB, "Destination alternate info address: %s",
                                                          alt_entry->address));
  }

  /***************************************************************************/
  /* A UsageIndication is permitted by the OSP spec to contain no            */
  /* UsageDetail elements.  But this would be rather strange, so it's worth  */
  /* an alert.                                                               */
  /***************************************************************************/
  if (use_detail_list == NULL)
  {
    TRC_ALERT((TB, "Usage detail parameter is NULL"));
  }

  for (use_entry = use_detail_list;
       use_entry != NULL;
       use_entry = use_entry->next)
  {
    TRC_ASSERT((use_entry->service != NULL) &&
               (use_entry->amount != NULL) &&
               (use_entry->increment != NULL) &&
               (use_entry->unit != NULL),
               (TB, "Usage detail field is NULL"));
    TRC_DETAIL((TB, "Usage detail service: %p", use_entry->service));
    if (use_entry->service->bandwidth != NULL)
    {
      TRC_DETAIL((TB, "Usage detail bandwidth: %s",
                                               use_entry->service->bandwidth));
    }
    TRC_DETAIL((TB, "Usage detail amount: %s", use_entry->amount));
    TRC_DETAIL((TB, "Usage detail increment: %s", use_entry->increment));
    TRC_DETAIL((TB, "Usage detail unit: %s", use_entry->unit));

    /*************************************************************************/
    /* StartTime and EndTime are optional.                                   */
    /*************************************************************************/
    if (use_entry->start_time != NULL)
    {
      TRC_DETAIL((TB, "Usage detail start time: %s", use_entry->start_time));
    }
    if (use_entry->end_time != NULL)
    {
      TRC_DETAIL((TB, "Usage detail end time: %s", use_entry->end_time));
    }

    /*************************************************************************/
    /* The TerminationCause element is optional.  But if present, it must    */
    /* have a TCCode subelement.                                             */
    /*************************************************************************/
    if (use_entry->term_cause != NULL)
    {
      TRC_ASSERT((use_entry->term_cause->term_code != NULL),
                 (TB, "Termination code field is NULL"));
      TRC_DETAIL((TB, "Usage detail cause: %p", use_entry->term_cause));
      TRC_DETAIL((TB, "Usage detail code: %d",
                                            use_entry->term_cause->term_code));
      if (use_entry->term_cause->term_desc != NULL)
      {
        TRC_DETAIL((TB, "Usage detail description: %s",
                                            use_entry->term_cause->term_desc));
      }
    }

    /*************************************************************************/
    /* The Statistics element is optional, as are its subelements LossSent,  */
    /* LossReceived, OneWayDelay and RoundTripDelay.  But whenever one of    */
    /* these subelements is present, all of the subelement's subsubelements  */
    /* are compulsory.                                                       */
    /*************************************************************************/
    if (use_entry->statistics != NULL)
    {
      if (use_entry->statistics->loss_sent != NULL)
      {
        TRC_ASSERT((use_entry->statistics->loss_sent->packets != NULL) &&
                   (use_entry->statistics->loss_sent->fraction != NULL),
                   (TB, "Loss sent field is NULL"));
        TRC_DETAIL((TB, "Usage: loss sent packets: %s",
                                   use_entry->statistics->loss_sent->packets));
        TRC_DETAIL((TB, "Usage: loss sent fraction: %s",
                                  use_entry->statistics->loss_sent->fraction));
      }

      if (use_entry->statistics->loss_received != NULL)
      {
        TRC_ASSERT((use_entry->statistics->loss_received->packets != NULL) &&
                   (use_entry->statistics->loss_received->fraction != NULL),
                   (TB, "Loss received field is NULL"));
        TRC_DETAIL((TB, "Usage: loss received packets: %s",
                               use_entry->statistics->loss_received->packets));
        TRC_DETAIL((TB, "Usage: loss received fraction: %s",
                              use_entry->statistics->loss_received->fraction));
      }

      if (use_entry->statistics->one_way_delay != NULL)
      {
        TRC_ASSERT((use_entry->statistics->one_way_delay->minimum != NULL) &&
                   (use_entry->statistics->one_way_delay->mean != NULL) &&
                   (use_entry->statistics->one_way_delay->variance != NULL) &&
                   (use_entry->statistics->one_way_delay->samples != NULL),
                   (TB, "One way delay field is NULL"));
        TRC_DETAIL((TB, "Usage: one way delay minimum: %s",
                               use_entry->statistics->one_way_delay->minimum));
        TRC_DETAIL((TB, "Usage: one way delay mean: %s",
                                  use_entry->statistics->one_way_delay->mean));
        TRC_DETAIL((TB, "Usage: one way delay variance: %s",
                              use_entry->statistics->one_way_delay->variance));
        TRC_DETAIL((TB, "Usage: one way delay samples: %s",
                               use_entry->statistics->one_way_delay->samples));
      }

      if (use_entry->statistics->round_trip_delay != NULL)
      {
        TRC_ASSERT((use_entry->statistics->round_trip_delay->minimum !=NULL) &&
                   (use_entry->statistics->round_trip_delay->mean != NULL) &&
                   (use_entry->statistics->round_trip_delay->variance!=NULL) &&
                   (use_entry->statistics->round_trip_delay->samples != NULL),
                   (TB, "Round trip delay field is NULL"));
        TRC_DETAIL((TB, "Usage: round trip delay minimum: %s",
                            use_entry->statistics->round_trip_delay->minimum));
        TRC_DETAIL((TB, "Usage: round trip delay mean: %s",
                               use_entry->statistics->round_trip_delay->mean));
        TRC_DETAIL((TB, "Usage: round trip delay variance: %s",
                           use_entry->statistics->round_trip_delay->variance));
        TRC_DETAIL((TB, "Usage: round trip delay samples: %s",
                            use_entry->statistics->round_trip_delay->samples));
      }
    }
  }

  TRC_ASSERT((client_id != NULL),
             (TB, "Client ID parameter is NULL"));
  TRC_DETAIL((TB, "Client ID IP address: %#x", client_id->ip_address));
  TRC_DETAIL((TB, "Client ID cookie: %p", client_id->ver_cookie));

  /***************************************************************************/
  /* Serialize access to the usage record file.                              */
  /***************************************************************************/
  OSP_GET_LOCK(&sua_file_mutex);

  /***************************************************************************/
  /* Check that the usage record file is open.  If it isn't, this callback   */
  /* should have been deregistered; but there may be a window where we can   */
  /* be called anyway.                                                       */
  /***************************************************************************/
  if (sua_file == NULL)
  {
    TRC_ALERT((TB, "Usage record file is not open"));
    rc = OSP_ERR_FILE;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If the number of usage records written has reached the maximum that we  */
  /* support, close the record file and reopen it again.  Note that this     */
  /* operation discards all existing records.                                */
  /***************************************************************************/
  if (sua_records_written >= sua_max_records)
  {
    TRC_NORMAL((TB, "Maximum number of records reached: %d",
                                                         sua_records_written));

    if (fclose(sua_file) == 0)
    {
      TRC_NORMAL((TB, "Successfully closed record file"));

      /***********************************************************************/
      /* Reopen the record file.                                             */
      /***********************************************************************/
      TRC_ASSERT((sua_file_name != NULL),
                 (TB, "Usage record file name is NULL"));

      sua_file = fopen(sua_file_name, "w");
      if (sua_file != NULL)
      {
        TRC_NORMAL((TB, "File reopened successfully: %p", sua_file));
        sua_records_written = 0;
      }
      else
      {
        TRC_OS_ERROR((TB, "Couldn't reopen usage record file"));
      }
    }
    else
    {
      TRC_OS_ERROR((TB, "Couldn't close usage record file"));
      sua_file = NULL;
    }

    /*************************************************************************/
    /* If the close and reopen operation failed, we should deregister our    */
    /* usage callback, since we won't be able to write any more usage        */
    /* records.                                                              */
    /*************************************************************************/
    if (sua_file == NULL)
    {
      TRC_ALERT((TB, "Reopen failed - deregistering usage callback"));

      /***********************************************************************/
      /* Unregister our usage callback.  There are no documented ways for    */
      /* this call to fail.                                                  */
      /***********************************************************************/
      rc = osp_um_register(NULL);
      TRC_ASSERT((rc == OSP_SUCCESS),
                 (TB, "osp_um_register call failed: rc=%d", rc));

      /***********************************************************************/
      /* Return an OSP_ERR_FILE error to the caller.                         */
      /***********************************************************************/
      rc = OSP_ERR_FILE;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* We are now ready to output usage records to file.  We will generate one */
  /* usage record for each item in the use_detail_list.                      */
  /***************************************************************************/
  for (use_entry = use_detail_list;
       use_entry != NULL;
       use_entry = use_entry->next)
  {
    TRC_NORMAL((TB, "Generate record for usage entry: %p", use_entry));

    /*************************************************************************/
    /* Timestamp.                                                            */
    /*************************************************************************/
    sua_print_field(timestamp, 20);

    /*************************************************************************/
    /* Role.                                                                 */
    /*************************************************************************/
    fprintf(sua_file, "%s ", sua_role_abbreviation(role));

    /*************************************************************************/
    /* Transaction ID.                                                       */
    /*************************************************************************/
    sua_print_field(transaction_id, 14);

    /*************************************************************************/
    /* Call ID.                                                              */
    /*************************************************************************/
    if (!strcmp(call_id->encoding, sua_encoding_cdata))
    {
      fprintf(sua_file, "C:");
    }
    else if (!strcmp(call_id->encoding, sua_encoding_base64))
    {
      fprintf(sua_file, "B:");
    }
    else
    {
      fprintf(sua_file, "-:");
    }
    sua_print_field(call_id->data, 12);

    /*************************************************************************/
    /* Source info.                                                          */
    /*************************************************************************/
    fprintf(sua_file, "%s ", sua_address_type_abbreviation(source_info->type));
    sua_print_field(source_info->address, 16);

    /*************************************************************************/
    /* Source alternate info.                                                */
    /*************************************************************************/
    if (source_alt_list != NULL)
    {
      fprintf(sua_file, "%s ",
              sua_address_type_abbreviation(source_alt_list->type));
      sua_print_field(source_alt_list->address, 16);
    }
    else
    {
      fprintf(sua_file, "%s ", sua_address_type_abbreviation(NULL));
      sua_print_field(NULL, 16);
    }

    /*************************************************************************/
    /* Destination info.                                                     */
    /*************************************************************************/
    fprintf(sua_file, "%s ", sua_address_type_abbreviation(dest_info->type));
    sua_print_field(dest_info->address, 16);

    /*************************************************************************/
    /* Destination alternate info.                                           */
    /*************************************************************************/
    if (dest_alt_list != NULL)
    {
      fprintf(sua_file, "%s ",
              sua_address_type_abbreviation(dest_alt_list->type));
      sua_print_field(dest_alt_list->address, 16);
    }
    else
    {
      fprintf(sua_file, "%s ", sua_address_type_abbreviation(NULL));
      sua_print_field(NULL, 16);
    }

    /*************************************************************************/
    /* Usage detail.                                                         */
    /*************************************************************************/
    sua_print_field(use_entry->amount, 4);
    sua_print_field(use_entry->increment, 4);
    sua_print_field(use_entry->unit, 4);
    sua_print_field(use_entry->start_time, 20);
    sua_print_field(use_entry->end_time, 20);
    if (use_entry->term_cause != NULL)
    {
      sua_print_field(use_entry->term_cause->term_code, 4);
      sua_print_field(use_entry->term_cause->term_desc, 16);
    }
    else
    {
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 16);
    }

    /*************************************************************************/
    /* Enhanced usage statistics.                                            */
    /*************************************************************************/
    if ((use_entry->statistics != NULL) &&
        (use_entry->statistics->loss_sent != NULL))
    {
      sua_print_field(use_entry->statistics->loss_sent->packets, 4);
      sua_print_field(use_entry->statistics->loss_sent->fraction, 4);
    }
    else
    {
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
    }

    if ((use_entry->statistics != NULL) &&
        (use_entry->statistics->loss_received != NULL))
    {
      sua_print_field(use_entry->statistics->loss_received->packets, 4);
      sua_print_field(use_entry->statistics->loss_received->fraction, 4);
    }
    else
    {
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
    }

    if ((use_entry->statistics != NULL) &&
        (use_entry->statistics->one_way_delay != NULL))
    {
      sua_print_field(use_entry->statistics->one_way_delay->minimum, 4);
      sua_print_field(use_entry->statistics->one_way_delay->mean, 4);
      sua_print_field(use_entry->statistics->one_way_delay->variance, 4);
      sua_print_field(use_entry->statistics->one_way_delay->samples, 4);
    }
    else
    {
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
    }

    if ((use_entry->statistics != NULL) &&
        (use_entry->statistics->round_trip_delay != NULL))
    {
      sua_print_field(use_entry->statistics->round_trip_delay->minimum, 4);
      sua_print_field(use_entry->statistics->round_trip_delay->mean, 4);
      sua_print_field(use_entry->statistics->round_trip_delay->variance, 4);
      sua_print_field(use_entry->statistics->round_trip_delay->samples, 4);
    }
    else
    {
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
      sua_print_field(NULL, 4);
    }

    /*************************************************************************/
    /* Client ID.                                                            */
    /*************************************************************************/
    fprintf(sua_file, "%u.%u.%u.%u\n",
            (client_id->ip_address & 0xFF000000) >> 24,
            (client_id->ip_address & 0x00FF0000) >> 16,
            (client_id->ip_address & 0x0000FF00) >> 8,
            (client_id->ip_address & 0x000000FF));

    /*************************************************************************/
    /* Increment the count of usage records written to file.                 */
    /*************************************************************************/
    sua_records_written++;
  }

  /***************************************************************************/
  /* Flush the usage record file handle so that the records that we have     */
  /* just generated are guaranteed to be there.                              */
  /***************************************************************************/
  fflush(sua_file);

  /***************************************************************************/
  /* Check if any errors have occurred in writing these records.             */
  /***************************************************************************/
  if (ferror(sua_file) != 0)
  {
    TRC_OS_ERROR((TB, "Error occurred writing to record file"));

    /*************************************************************************/
    /* Pretend that we have reached the maximum number of records that we    */
    /* can write to the usage record file.  This means that, next time this  */
    /* callback is called, we will close and reopen the record file.  If     */
    /* that fails as well, we will then give up trying to handle further     */
    /* usage information, by deregistering this callback.                    */
    /*************************************************************************/
    sua_records_written = sua_max_records;
    rc = OSP_ERR_FILE;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Finally call osp_um_response to indicate that we have successfully      */
  /* handled the usage information.                                          */
  /***************************************************************************/
  status.code = OSP_STATUS_SUCCESS;
  status.desc = com_status_success;
  rc = osp_um_response(corr, &status, 0, scm_sign_responses);
  TRC_ASSERT((rc == OSP_SUCCESS),
             (TB, "Error from osp_um_response: rc=%d", rc));

EXIT_LABEL:
  OSP_RELEASE_LOCK(&sua_file_mutex);
  TRC_EXIT();
  return(rc);
}

/**PROC+**********************************************************************/
/* Name:      sua_role_abbreviation                                          */
/*                                                                           */
/* Purpose:   Return a 3 letter abbreviation for the supplied role           */
/*                                                                           */
/* Returns:   Pointer to a string containing 3 characters.  The string       */
/*            is part of the global data _sua_role_map_.                     */
/*                                                                           */
/* Params:    IN role         - The role that is to be abbreviated.  See     */
/*                              _sua_role_map_ for valid values.             */
/*                                                                           */
/* Operation: The abbreviation mapping is table-driven, controlled by the    */
/*            global variable _sua_role_map_.  If the passed _role_          */
/*            value is incorrect, sua_role_abbreviation returns the          */
/*            abbreviation for the NULL entry in _sua_role_map_.             */
/*                                                                           */
/**PROC-**********************************************************************/

char *sua_role_abbreviation(char *role)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  char *abbrev;

  TRC_ENTRY("sua_role_abbreviation");

  abbrev = sua_abbreviation(role, sua_role_map);

  TRC_EXIT();
  return(abbrev);
}

/**PROC+**********************************************************************/
/* Name:      sua_address_type_abbreviation                                  */
/*                                                                           */
/* Purpose:   Return a 4 letter abbreviation for the supplied type           */
/*                                                                           */
/* Returns:   Pointer to a string containing 4 characters.  The string       */
/*            is part of the global data _sua_address_type_map_.             */
/*                                                                           */
/* Params:    IN type         - The type that is to be abbreviated.  See     */
/*                              _sua_address_type_map_ for valid values.     */
/*                                                                           */
/* Operation: The abbreviation mapping is table-driven, controlled by the    */
/*            global variable _sua_address_type_map_.  If the passed _type_  */
/*            value is incorrect, sua_address_type_abbreviation returns the  */
/*            abbreviation for the NULL entry in _sua_address_type_map_.     */
/*                                                                           */
/**PROC-**********************************************************************/

char *sua_address_type_abbreviation(char *type)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  char *abbrev;

  TRC_ENTRY("sua_address_type_abbreviation");

  abbrev = sua_abbreviation(type, sua_address_type_map);

  TRC_EXIT();
  return(abbrev);
}

/**PROC+**********************************************************************/
/* Name:      sua_abbreviation                                               */
/*                                                                           */
/* Purpose:   Return an abbreviation for the supplied string, using the      */
/*            specified mapping table                                        */
/*                                                                           */
/* Returns:   Pointer to an abbreviation string, whose memory belongs to     */
/*            the specified mapping table.                                   */
/*                                                                           */
/* Params:    IN string       - The string to be abbreviated.                */
/*            IN table        - The abbreviation mapping table to refer to.  */
/*                                                                           */
/* Operation: If the supplied string is NULL, or does match any of the       */
/*            non-NULL entries in the supplied table, sua_abbreviation       */
/*            returns the abbreviation for the NULL entry in _table_.        */
/*                                                                           */
/**PROC-**********************************************************************/

char *sua_abbreviation(char *string, SUA_ABBREV *table)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  char *abbrev;

  TRC_ENTRY("sua_abbreviation");

  TRC_DETAIL((TB, "Abbreviate string: %p", string));
  TRC_DETAIL((TB, "Using mapping table: %p", table));

  if (string != NULL)
  {
    TRC_DETAIL((TB, "String is: \"%s\"", string));

    /*************************************************************************/
    /* Find this string in the mapping table.                                */
    /*************************************************************************/
    abbrev = NULL;
    for (; table->full != NULL; table++)
    {
      TRC_DETAIL((TB, "Mapping table string is: \"%s\"", table->full));

      if (strcmp(string, table->full) == 0)
      {
        TRC_DETAIL((TB, "Found; abbreviation is: \"%s\"", table->abbrev));
        abbrev = table->abbrev;
        break;
      }
    }

    /*************************************************************************/
    /* If we didn't find the string in the table, return the abbreviation in */
    /* the table's NULL entry.  Conveniently, this is the entry that _table_ */
    /* is now pointing to.                                                   */
    /*************************************************************************/
    if (abbrev == NULL)
    {
      TRC_DETAIL((TB, "Default abbreviation: \"%s\"", table->abbrev));
      abbrev = table->abbrev;
    }
  }
  else
  {
    TRC_DETAIL((TB, "NULL string: find table's NULL entry"));

    /*************************************************************************/
    /* Just run through the table until we find the NULL entry at the end.   */
    /*************************************************************************/
    for (; table->full != NULL; table++)
    {
      TRC_DETAIL((TB, "Skipping table string: \"%s\"", table->full));
    }

    TRC_DETAIL((TB, "NULL string abbreviation: \"%s\"", table->abbrev));
    abbrev = table->abbrev;
  }

  TRC_EXIT();
  return(abbrev);
}

/**PROC+**********************************************************************/
/* Name:      sua_print_field                                                */
/*                                                                           */
/* Purpose:   Write the specified data into a fixed width field in the       */
/*            usage record file, taking account of non-printable             */
/*            characters and the possibility that the data may need to be    */
/*            truncated, followed by a further space to separate the field   */
/*            contents from the following column.                            */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     data         - Null-terminated string of characters     */
/*                                  that is to be written into the record    */
/*                                  file.                                    */
/*            IN     field_width  - The desired field width for the data     */
/*                                  that is to be written.                   */
/*                                                                           */
/* Operation: Non-printable characters are replaced by %XX, where XX is the  */
/*            ASCII character code in hex.  If the data is too long to fit   */
/*            completely within _field_width_ characters, the last three     */
/*            characters are replaced by dots ("...").  If the data does     */
/*            not require _field_width_ characters, the surplus characters   */
/*            are replaced by spaces (on the right).                         */
/*                                                                           */
/**PROC-**********************************************************************/

void sua_print_field(char *data, int field_width)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  char buffer[21];
  int truncated;
  int bufpos;
  int datpos;
  char c;

  TRC_ENTRY("sua_print_field");

  /***************************************************************************/
  /* Trace input parameters.                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Data pointer: %p", data));
  TRC_DETAIL((TB, "Field width: %d", field_width));

  /***************************************************************************/
  /* _data_ is allowed to be NULL.  This means that the output field should  */
  /* be filled with dashes to the required field width.  The reason for      */
  /* handling this inside sua_print_field is so that all the field filling   */
  /* logic is encapsulated within one function.  Detect and handle this      */
  /* special case here.                                                      */
  /***************************************************************************/
  if (data == NULL)
  {
    TRC_NORMAL((TB, "NULL data case; fill buffer with dashes"));
    for (bufpos = 0; bufpos < field_width; bufpos++)
    {
      buffer[bufpos] = '-';
    }
    buffer[field_width] = 0;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now the mainline case, where data is not NULL.                          */
  /***************************************************************************/
  TRC_DETAIL((TB, "Data string length: \"%d\"", strlen(data)));
  truncated = FALSE;
  bufpos = 0;
  for (datpos = 0; (c = data[datpos]) != 0; datpos++)
  {
    TRC_DETAIL((TB, "Buffer position now: %d", bufpos));
    TRC_DETAIL((TB, "Next character: %#x", (unsigned int) (unsigned char) c));

    if ((c >= 0x20) && (c <= 0x7E) && (c != '%'))
    {
      TRC_DETAIL((TB, "Printable character: %c", c));

      /***********************************************************************/
      /* This is a printable character.  It needs one position in the output */
      /* buffer.                                                             */
      /***********************************************************************/
      if (bufpos + 1 <= field_width)
      {
        TRC_DETAIL((TB, "Append printable character"));
        buffer[bufpos] = c;
        bufpos++;
      }
      else
      {
        TRC_DETAIL((TB, "No room for printable character"));
        truncated = TRUE;
        break;
      }
    }
    else
    {
      TRC_DETAIL((TB, "Unprintable character"));

      /***********************************************************************/
      /* This is an unprintable character.  It needs three positions in the  */
      /* output buffer.                                                      */
      /***********************************************************************/
      if (bufpos + 3 <= field_width)
      {
        sprintf(buffer + bufpos, "%%%02X", (unsigned int) (unsigned char) c);
        TRC_DETAIL((TB, "Append unprintable character: %s", buffer + bufpos));
        bufpos += 3;
      }
      else
      {
        /*********************************************************************/
        /* Note that there is no point writing a partial representation of   */
        /* the unprintable character, since the end of the output buffer     */
        /* will be overwritten by three dots.                                */
        /*********************************************************************/
        TRC_DETAIL((TB, "No room for unprintable character"));
        truncated = TRUE;
        break;
      }
    }
  }

  if (truncated)
  {
    TRC_DETAIL((TB, "Buffer is truncated; mark with dots"));

    /*************************************************************************/
    /* This NUL terminates the buffer as well.                               */
    /*************************************************************************/
    sprintf(buffer + (field_width - 3), "...");
  }
  else
  {
    TRC_DETAIL((TB, "Fill unused positions with spaces"));

    /*************************************************************************/
    /* Fill any unused buffer positions with spaces, and NUL terminate.      */
    /*************************************************************************/
    for (; bufpos + 1 <= field_width; bufpos++)
    {
      buffer[bufpos] = ' ';
    }
    buffer[field_width] = 0;
  }

EXIT_LABEL:
  TRC_DETAIL((TB, "Output buffer: \"%s\"", buffer));

  /***************************************************************************/
  /* Finally, write the buffer to the usage record file.                     */
  /***************************************************************************/
  fprintf(sua_file, "%s ", buffer);

  TRC_EXIT();
  return;
}
