
/**MOD+***********************************************************************/
/* Module:    aibdcov.c                                                      */
/*                                                                           */
/* Purpose:   Common AIB - data coverage functions                           */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.4                $ $Modtime::   Jan 11 2001 15:56:24   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <aibincl.h>

#ifdef OSP_DATA_COVERAGE
/**PROC+**********************************************************************/
/* Name:      aib_init_data_coverage                                         */
/*                                                                           */
/* Purpose:   Initialize data coverage routines                              */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - initialization successful                */
/*            OSP_ERROR           - failed to create data coverage mutex     */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_init_data_coverage()
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  AIB_COVERAGE_STRUCTURE *structure;
  AIB_ELEMENT *elements;

  TRC_ENTRY("aib_init_data_coverage");

  TRC_ASSERT((!aib_coverage_initialized), (TB, "Already initialized"));

  /***************************************************************************/
  /* Create the data coverage mutex.                                         */
  /***************************************************************************/
  rc = osp_create_mutex(&aib_coverage_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ALERT((TB, "Failed to create data coverage mutex: rc=%d", rc));
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Created data coverage mutex"));

  OSP_GET_LOCK(&aib_coverage_mutex);

  /***************************************************************************/
  /* Initialize the covered field in all covered AIB_ELEMENTs to 0.          */
  /***************************************************************************/
  for (structure = &(aib_coverage_structures[0]);
       structure->name != NULL;
       structure++)
  {
    TRC_NORMAL((TB, "Initialize coverage flags for \"%s\" structure",
                                                             structure->name));
    for (elements = structure->elements; elements->name != NULL; elements++)
    {
      TRC_DETAIL((TB, "Initialize coverage flag: element=\"%s\"",
                                                              elements->name));
      elements->coverage = 0;
    }
  }

  /***************************************************************************/
  /* Initialize registered structure list.                                   */
  /***************************************************************************/
  SL_INIT_ROOT(aib_registered_structures);

  aib_coverage_initialized = TRUE;
  aib_coverage_error = OSP_SUCCESS;
  TRC_NORMAL((TB, "Data coverage successfully initialized"));

  OSP_RELEASE_LOCK(&aib_coverage_mutex);

EXIT_LABEL:
  TRC_EXIT();
  return(rc);

} /* aib_init_data_coverage */

/**PROC+**********************************************************************/
/* Name:      aib_term_data_coverage                                         */
/*                                                                           */
/* Purpose:   Terminate data coverage routines                               */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_term_data_coverage()
{
  TRC_ENTRY("aib_term_data_coverage");

  if (aib_coverage_initialized)
  {
    /*************************************************************************/
    /* Destroy the data coverage mutex.                                      */
    /*************************************************************************/
    osp_destroy_mutex(&aib_coverage_mutex);

    aib_coverage_initialized = FALSE;
  }
  else
  {
    TRC_ALERT((TB, "Data coverage is not initialized"));
  }

  TRC_EXIT();
  return;

} /* aib_term_data_coverage */

/**PROC+**********************************************************************/
/* Name:      aib_register_data_structure                                    */
/*                                                                           */
/* Purpose:   Register a structure for data coverage                         */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     elements     - Pointer to AIB_ELEMENT array of which    */
/*                                  the structure being registered is an     */
/*                                  instance.                                */
/*            IN     address      - Address of the structure to register.    */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_register_data_structure(AIB_ELEMENT *elements, void *address)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  AIB_COVERAGE_STRUCTURE *structure;
  AIB_REGISTERED_STRUCTURE *reg;

  TRC_ENTRY("aib_register_data_structure");

  TRC_ASSERT((aib_coverage_initialized),
             (TB, "Data coverage is not initialized"));

  /***************************************************************************/
  /* Check parameters are valid.                                             */
  /***************************************************************************/
  TRC_ASSERT((elements != NULL), (TB, "NULL elements parameter"));
  TRC_ASSERT((address != NULL), (TB, "NULL address parameter"));

  /***************************************************************************/
  /* Check that elements is in the list of covered elements.                 */
  /***************************************************************************/
  for (structure = aib_coverage_structures;
       structure->name != NULL;
       structure++)
  {
    if (structure->elements == elements)
    {
      TRC_DETAIL((TB, "Structure is an instance of type \"%s\"",
                                                             structure->name));
      break;
    }
  }
  if (structure->name == NULL)
  {
    TRC_ALERT((TB, "Specified structure type not in covered structures list"));
    aib_coverage_error = OSP_ERR_UNSUPPORTED;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Allocate an AIB_REGISTERED_STRUCTURE to hold the details.               */
  /***************************************************************************/
  reg = rm_get_mem(sizeof(AIB_REGISTERED_STRUCTURE));
  if (reg == NULL)
  {
    TRC_ALERT((TB, "Failed to allocated memory for data coverage structure"));
    aib_coverage_error = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  reg->elements = elements;
  reg->address = address;
  SL_ADD_TO_TAIL(aib_registered_structures, reg);

  TRC_DETAIL((TB, "Structure registered for data coverage: reg=%p", reg));

EXIT_LABEL:
  TRC_EXIT();
  return;

} /* aib_register_data_structure */

/**PROC+**********************************************************************/
/* Name:      aib_scan_registered_structures                                 */
/*                                                                           */
/* Purpose:   Scan registered structures for data coverage                   */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_scan_registered_structures()
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  AIB_REGISTERED_STRUCTURE *reg;
  AIB_ELEMENT *elements;
  OSP_BYTE *address;

  TRC_ENTRY("aib_scan_registered_structures");

  /***************************************************************************/
  /* Loop through list of registered structures, removing as we go.          */
  /***************************************************************************/
  for (reg = SL_HEAD(aib_registered_structures);
       reg != NULL;
       reg = SL_HEAD(aib_registered_structures))
  {
    TRC_DETAIL((TB, "Scan registered structure: reg=%p", reg));

    for (elements = reg->elements; elements->name != NULL; elements++)
    {
      TRC_DETAIL((TB, "Scan \"%s\" field", elements->name));
      address = ((OSP_BYTE *)(reg->address)) + elements->data_offset;
      if (*((void **)address) != NULL)
      {
        elements->coverage++;
        TRC_DETAIL((TB, "Field is covered; coverage count is now %d",
                                                          elements->coverage));
      }
      else
      {
        TRC_DETAIL((TB, "Field is not covered; coverage count still %d",
                                                          elements->coverage));
      }
    }

    SL_REMOVE_HEAD(aib_registered_structures);
  }
  TRC_DETAIL((TB, "Data coverage scan complete"));

  TRC_EXIT();
  return;

} /* aib_scan_registered_structures */

/**PROC+**********************************************************************/
/* Name:      aib_write_data_coverage                                        */
/*                                                                           */
/* Purpose:   Write data coverage information to the specified file stream   */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     fstream      - File stream to write information to.     */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_write_data_coverage(FILE *fstream)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  AIB_COVERAGE_STRUCTURE *structure;
  AIB_ELEMENT *elements;

  TRC_ENTRY("aib_write_data_coverage");

  OSP_GET_LOCK(&aib_coverage_mutex);

  for (structure = aib_coverage_structures;
       structure->name != NULL;
       structure++)
  {
    fprintf(fstream, "Data coverage for \"%s\" structure:\n", structure->name);
    for (elements = structure->elements; elements->name != NULL; elements++)
    {
      fprintf(fstream, "\t%s %d\n", elements->name, elements->coverage);
    }
  }

  if (aib_coverage_error == OSP_SUCCESS)
  {
    fprintf(fstream, "Data coverage statistics are complete.\n");
  }
  else
  {
    fprintf(fstream, "WARNING: Data coverage statistics are incomplete"
            " (error(s) occurred: rc=%d).\n", aib_coverage_error);
  }

  OSP_RELEASE_LOCK(&aib_coverage_mutex);

  TRC_EXIT();
  return;

} /* aib_write_data_coverage */

#endif
