#ifndef AIBEXTRN_INCLUDED
#define AIBEXTRN_INCLUDED

/**INC+***********************************************************************/
/* Header:    aibextrn.h                                                     */
/*                                                                           */
/* Purpose:   AIB Common Function API                                        */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.28               $ $Modtime::   Jan 11 2001 15:56:24   $  */
/*                                                                           */
/**INC-***********************************************************************/

/*****************************************************************************/
/* The maximum possible parsing stack depth for an OSP client component.     */
/* This is determined by the deepest possible nesting of XML elements within */
/* a client component, as described by the OSP specification.  Some of the   */
/* limiting and near-limiting cases are illustrated here:                    */
/*                                                                           */
/*   - UsageIndication               (1)                                     */
/*     - UsageDetail                 (2)                                     */
/*       - Service                   (3)                                     */
/*         - Bandwidth               (4)                                     */
/*       - TerminationCause          (3)                                     */
/*         - TCCode                  (4)                                     */
/*       - Statistics                (3)                                     */
/*         - LossSent                (4)                                     */
/*           - Packets               (5)                                     */
/*                                                                           */
/* This may need increasing if the code is updated to support a later        */
/* revision of the OSP specification that permits greater levels of element  */
/* nesting.                                                                  */
/*****************************************************************************/
#define AIB_MAX_PARSE_DEPTH                       5

/*****************************************************************************/
/* Element occurrence indicators:                                            */
/*                                                                           */
/*   - AIB_OCCURS_ONCE: the element must appear exactly once                 */
/*   - AIB_OCCURS_QUMK: the element may be absent, or else must appear once  */
/*   - AIB_OCCURS_STAR: the element may appear any number of times, or may   */
/*                      be completely absent                                 */
/*   - AIB_OCCURS_PLUS: the element may appear one or more times.            */
/*****************************************************************************/
#define AIB_OCCURS_ONCE                           0x77777771
#define AIB_OCCURS_QUMK                           0x77777772
#define AIB_OCCURS_STAR                           0x77777773
#define AIB_OCCURS_PLUS                           0x77777774

/*****************************************************************************/
/* Externally used structures                                                */
/*****************************************************************************/

/*****************************************************************************/
/* Type definition for element handler routines.                             */
/*****************************************************************************/
typedef OSP_RC (*AIB_HANDLER)(TRANS_INFO *,
                              const char *,
                              const char **,
                              void *,
                              int,
                              void *);

/**STRUCT+********************************************************************/
/* Structure: AIB_ELEMENT                                                    */
/*                                                                           */
/* Description: A structure describing an XML element and how the USE AIB    */
/*              handles it.                                                  */
/*                                                                           */
/*****************************************************************************/

typedef struct aib_element
{
  /***************************************************************************/
  /* The name of the XML element, e.g. "UsageDetail".                        */
  /***************************************************************************/
  const char *name;

  /***************************************************************************/
  /* Handler routine for this element.  The general start element handler    */
  /* calls this function with                                                */
  /*   - the TRANS_INFO * parameter pointing to the current transaction      */
  /*   - the name of the element whose start tag has just been encountered   */
  /*   - the attribute list of the start tag that is being processed         */
  /*   - the void * parameter pointing to the current parse frame's data,    */
  /*     offset by the value of the data_offset field below                  */
  /*   - the int parameter set to the value of the size field below          */
  /*   - the final void * parameter set to the value of the extra_data field */
  /*     below.                                                              */
  /*                                                                         */
  /* In the majority of cases, the handler routine for an element is one of  */
  /*   - aib_start_processing_simple_element, which handles a leaf XML       */
  /*     element by arranging to store its data in a dynamically allocated   */
  /*     string                                                              */
  /*   - aib_start_processing_substructure, which handles a container XML    */
  /*     element by setting up a new parsing frame in which to process the   */
  /*     element's subelements.                                              */
  /***************************************************************************/
  AIB_HANDLER handler;

  /***************************************************************************/
  /* Indicates whether this element may occur exactly once, multiple times,  */
  /* or either once or not at all.  Values AIB_OCCURS_* defined in usedef.h. */
  /***************************************************************************/
  int occurrence;

  /***************************************************************************/
  /* Offset in bytes to the location where the value read for this element   */
  /* should be stored.  The reference position for this offset is the value  */
  /* of the data field in the current parsing frame.  (See AIB_PARSE_FRAME   */
  /* below.)                                                                 */
  /***************************************************************************/
  int data_offset;

  /***************************************************************************/
  /* The third parameter to pass to the handler routine for this element.    */
  /***************************************************************************/
  int size;

  /***************************************************************************/
  /* The final "additional data" parameter to pass to the handler routine    */
  /* for this element.                                                       */
  /***************************************************************************/
  void *extra_data;

  /***************************************************************************/
  /* Data coverage count.  NB not conditioned by OSP_DATA_COVERAGE macro, to */
  /* avoid spurious partial build problems.                                  */
  /***************************************************************************/
  int coverage;

} AIB_ELEMENT;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: AIB_PARSE_FRAME                                                */
/*                                                                           */
/* Description: A structure describing a frame in the overall XML parsing    */
/*              context.                                                     */
/*                                                                           */
/*****************************************************************************/

typedef struct aib_parse_frame
{
  /***************************************************************************/
  /* Pointer to an array of AIB_ELEMENT structures that specifies the XML    */
  /* elements that are expected to occur next.  The array is terminated by   */
  /* an element whose name field is NULL.                                    */
  /***************************************************************************/
  AIB_ELEMENT *elements;

  /***************************************************************************/
  /* A flag that is true if the USE start element handler has been called    */
  /* more recently than the end element handler.                             */
  /***************************************************************************/
  int within_element;

  /***************************************************************************/
  /* The number of previous occurrences in this frame of the element at the  */
  /* head of the elements array (normally zero, but we have to keep track of */
  /* this in order to verify AIB_OCCURS_PLUS occurrence).                    */
  /***************************************************************************/
  int num_occurrences;

  /***************************************************************************/
  /* Reference pointer for storing data associated with elements in this     */
  /* parsing frame.  (See also the data_offset field in AIB_ELEMENT.)        */
  /***************************************************************************/
  void *data;

} AIB_PARSE_FRAME;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: AIB_CHAR_DATA                                                  */
/*                                                                           */
/* Description: Used to manage the memory in which character data is stored  */
/*                                                                           */
/*****************************************************************************/

typedef struct aib_char_data
{
  /***************************************************************************/
  /* Error code set when a problem is encountered in parsing                 */
  /***************************************************************************/
  int error_code;

  /***************************************************************************/
  /* Pointer to start of memory block to be filled with received character   */
  /* data                                                                    */
  /***************************************************************************/
  char **start_ptr;

  /***************************************************************************/
  /* Length of data in memory pointed to by start_ptr                        */
  /***************************************************************************/
  int current_len;

  /***************************************************************************/
  /* Total length of memory pointed to by start_ptr                          */
  /***************************************************************************/
  int total_len;

  /***************************************************************************/
  /* Back pointer to transaction, so that we can get to the string pool.     */
  /***************************************************************************/
  TRANS_INFO *trans;

} AIB_CHAR_DATA;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: AIB_DATA                                                       */
/*                                                                           */
/* Description: A structure containing parsing state information that is     */
/*              common to all AIBs.                                          */
/*                                                                           */
/*****************************************************************************/

typedef struct aib_data
{
  /***************************************************************************/
  /* Used by common AIB code to track storage for character data.            */
  /***************************************************************************/
  AIB_CHAR_DATA char_data;

  /***************************************************************************/
  /* Current element stack depth.                                            */
  /***************************************************************************/
  int depth;

  /***************************************************************************/
  /* Stack of element parsing frames, indexed by depth field above.          */
  /***************************************************************************/
  AIB_PARSE_FRAME frames[AIB_MAX_PARSE_DEPTH];

  /***************************************************************************/
  /* Whenever we ignore an unrecognized non-critical element, we potentially */
  /* have to ignore a whole XML subtree.  This field tracks how deep into    */
  /* such a subtree we are.  In the normal case, where we are not ignoring   */
  /* anything, this is set to zero.                                          */
  /***************************************************************************/
  int ignore_depth;

  /***************************************************************************/
  /* Component response status.  If a component-level error occurs, the      */
  /* status code is set to a value other than OSP_STATUS_SUCCESS, and all    */
  /* further XML elements are ignored until we detect the UsageIndication    */
  /* end tag.  We then send a UsageIndication response containing the        */
  /* appropriate status code.                                                */
  /***************************************************************************/
  OSP_STATUS status;

  /***************************************************************************/
  /* Pointer to correlator that we will use when calling the application's   */
  /* callback for this component.                                            */
  /***************************************************************************/
  OSP_INT_CORRELATOR *corr;

} AIB_DATA;

/**STRUCT-********************************************************************/

#ifdef OSP_DATA_COVERAGE
/**STRUCT+********************************************************************/
/* Structure: AIB_COVERAGE_STRUCTURE                                         */
/*                                                                           */
/* Description: A structure describing a structure that can be covered by    */
/*              the data coverage routines                                   */
/*                                                                           */
/*****************************************************************************/

typedef struct aib_coverage_structure
{
  /***************************************************************************/
  /* Name of this type of structure.                                         */
  /***************************************************************************/
  const char *name;

  /***************************************************************************/
  /* The AIB_ELEMENT array that describes elements of this structure type.   */
  /***************************************************************************/
  AIB_ELEMENT *elements;

} AIB_COVERAGE_STRUCTURE;

/**STRUCT-********************************************************************/
#endif

/*****************************************************************************/
/* Default initial lengths for element data                                  */
/*****************************************************************************/
#define AIB_CODE_LEN                              4
#define AIB_DEF_CALLID_LEN                        32
#define AIB_DEF_ADDRESS_LEN                       16
#define AIB_DEF_ALTERNATE_LEN                     16
#define AIB_DEF_MAX_DESTS_LEN                     4
#define AIB_DEF_TOKEN_LEN                         128
#define AIB_DEF_AUTH_INFO_LEN                     128
#define AIB_DEF_TIME_LEN                          21
#define AIB_DEF_BANDWIDTH_LEN                     16
#define AIB_DEF_ROLE_LEN                          12
#define AIB_DEF_TRANSACTIONID_LEN                 12
#define AIB_DEF_AMOUNT_LEN                        8
#define AIB_DEF_INCREMENT_LEN                     8
#define AIB_DEF_UNIT_LEN                          8
#define AIB_DEF_TERMCODE_LEN                      4
#define AIB_DEF_TERMDESC_LEN                      30
#define AIB_DEF_PACKETS_LEN                       8
#define AIB_DEF_FRACTION_LEN                      8
#define AIB_DEF_MINIMUM_LEN                       8
#define AIB_DEF_MEAN_LEN                          8
#define AIB_DEF_VARIANCE_LEN                      8
#define AIB_DEF_SAMPLES_LEN                       8
#define AIB_DEF_VERSION_LEN                       8
#define AIB_DEF_CAPABILITY_LEN                    40
#define AIB_DEF_CHANNELS_LEN                      4
#define AIB_DEF_ALMOST_OUT_LEN                    8
#define AIB_DEF_SERVICE_TYPE_LEN                  16
#define AIB_DEF_PROT_TYPE_LEN                     16

/*****************************************************************************/
/* Values for element type parameter in XXX_start_processing calls           */
/*****************************************************************************/
#define AIB_AUTH_RQ          1
#define AIB_AUTH_S_RQ        2
#define AIB_AUTH_IND         3
#define AIB_AUTH_S_IND       4
#define AIB_REAUTH_RQ        5
#define AIB_REAUTH_S_RQ      6
#define AIB_PRICING_IND      7
#define AIB_CAPS_IND         8
#define AIB_USAGE_IND        9
#define AIB_SUBS_AUTHEN_RQ  10

/*****************************************************************************/
/* Length of OSP time stamp string                                           */
/*****************************************************************************/
#define AIB_TIME_LEN      21

/*****************************************************************************/
/* aibrdut.c                                                                 */
/*****************************************************************************/
char *aib_find_attr_value(const char **, const char *, const char *);

void aib_free_call_id(OSP_CALL_ID *);

void aib_free_address(OSP_ADDRESS *);

void aib_free_service(OSP_SERVICE_INFO *);

void aib_free_token(OSP_TOKEN *);

void aib_free_usage_detail(OSP_USAGE_DETAIL *);

/*****************************************************************************/
/* aibwrxml.c                                                                */
/*****************************************************************************/
int aib_write_xml_destination(TRANS_INFO *, OSP_DESTINATION *);

int aib_write_xml_address(TRANS_INFO *, const char *, OSP_ADDRESS *);

int aib_write_xml_usage_detail(TRANS_INFO *, OSP_USAGE_DETAIL *);

int aib_write_xml_timestamp(TRANS_INFO *);

int aib_write_xml_status(TRANS_INFO *, OSP_STATUS *);

int aib_write_xml_service_info(TRANS_INFO *, OSP_SERVICE_INFO *);

int aib_write_xml_element(TRANS_INFO *, const char *, const char *);

int aib_write_xml_element_numeric(TRANS_INFO *, const char *, int);

int aib_write_xml_term_cause(TRANS_INFO *, OSP_TERM_CAUSE *);

int aib_write_xml_service(TRANS_INFO *, OSP_SERVICE *);

int aib_write_xml_credit_amount(TRANS_INFO *, OSP_CREDIT_AMOUNT*);

int aib_write_xml_credit_time(TRANS_INFO *, OSP_CREDIT_TIME *);

/*****************************************************************************/
/* aibwrlst.c                                                                */
/*****************************************************************************/
int aib_write_xml_token_list(TRANS_INFO *, char *, OSP_TOKEN *);

int aib_write_xml_destination_list(TRANS_INFO *, OSP_DESTINATION *);

int aib_write_xml_address_list(TRANS_INFO *, const char *, OSP_ADDRESS *);

int aib_write_xml_usage_detail_list(TRANS_INFO *, OSP_USAGE_DETAIL *);

int aib_write_xml_auth_url_list(TRANS_INFO *, OSP_SERVICE_URL *);

int aib_write_xml_service_list(TRANS_INFO *, OSP_SERVICE *);

int aib_write_xml_cert_chain(TRANS_INFO *, OSP_CERTIFICATE *);

/*****************************************************************************/
/* aibstrtp.c                                                                */
/*****************************************************************************/
void aib_start_processing(TRANS_INFO *,
                          XML_Parser *,
                          char *,
                          int,
                          AIB_ELEMENT *,
                          XML_EndElementHandler,
                          OSP_STATUS *);
OSP_RC aib_start_processing_simple_element(TRANS_INFO *,
                                           const char *,
                                           const char **,
                                           void *,
                                           int,
                                           void *);

OSP_RC aib_start_processing_call_id(TRANS_INFO *,
                                    const char *,
                                    const char **,
                                    void *,
                                    int,
                                    void *);

OSP_RC aib_start_processing_address(TRANS_INFO *,
                                    const char *,
                                    const char **,
                                    void *,
                                    int,
                                    void *);

OSP_RC aib_start_processing_token(TRANS_INFO *,
                                  const char *,
                                  const char **,
                                  void *,
                                  int,
                                  void *);

OSP_RC aib_start_processing_subs_auth(TRANS_INFO *,
                                      const char *,
                                      const char **,
                                      void *,
                                      int,
                                      void *);

OSP_RC aib_start_processing_substructure(TRANS_INFO *,
                                         const char *,
                                         const char **,
                                         void *,
                                         int,
                                         void *);

OSP_RC aib_start_processing_list(TRANS_INFO *,
                                 const char *,
                                 const char **,
                                 void *,
                                 int,
                                 void *);

OSP_RC aib_start_processing_usage_detail(TRANS_INFO *,
                                         const char *,
                                         const char **,
                                         void *,
                                         int,
                                         void *);

OSP_RC aib_start_processing_dev_info(TRANS_INFO *,
                                     const char *,
                                     const char **,
                                     void *,
                                     int,
                                     void *);

OSP_RC aib_start_processing_capability(TRANS_INFO *,
                                       const char *,
                                       const char **,
                                       void *,
                                       int,
                                       void *);

OSP_RC aib_start_processing_protocol_type(TRANS_INFO *,
                                          const char *,
                                          const char **,
                                          void *,
                                          int,
                                          void *);

OSP_RC aib_start_processing_resource(TRANS_INFO *,
                                     const char *,
                                     const char **,
                                     void *,
                                     int,
                                     void *);

OSP_RC aib_start_processing_data_rate(TRANS_INFO *,
                                      const char *,
                                      const char **,
                                      void *,
                                      int,
                                      void *);

OSP_RC aib_start_processing_supp_prot(TRANS_INFO *,
                                      const char *,
                                      const char **,
                                      void *,
                                      int,
                                      void *);

/*****************************************************************************/
/* aibhdlrs.c                                                                */
/*****************************************************************************/
void aib_end_element_handler(void *, const XML_Char *);

#ifdef OSP_DATA_COVERAGE
/*****************************************************************************/
/* aibdcov.c                                                                 */
/*****************************************************************************/
OSP_RC aib_init_data_coverage();
void aib_term_data_coverage();
void aib_register_data_structure(AIB_ELEMENT *, void *);
void aib_scan_registered_structures();
void aib_write_data_coverage(FILE *);
#endif

/*****************************************************************************/
/* aibsndut.c                                                                */
/*****************************************************************************/
int aib_space_available(TRANS_INFO *, int);

/*****************************************************************************/
/* Fast inline version of aib_add_data (now removed).  Because this version  */
/* is fast, we can use it to eliminate sprintfs by dividing them into        */
/* sequences of AIB_ADD_DATA.                                                */
/*****************************************************************************/
#define AIB_ADD_DATA(trans, data, len)                                  \
do                                                                      \
{                                                                       \
  int __len = (len);                                                    \
                                                                        \
  memcpy((trans)->send_buffer + (trans)->send_buffer_end, data, __len); \
  (trans)->send_buffer_end += __len;                                    \
  (trans)->send_buffer[(trans)->send_buffer_end] = 0;                   \
}                                                                       \
while (0)

#define AIB_ADD_NULL(trans, data) AIB_ADD_DATA(trans, data, strlen(data))

#endif
