
/**MOD+***********************************************************************/
/* Module:   aibhdlrs.c                                                      */
/*                                                                           */
/* Purpose:  AIB Common XML Start & End element handlers                     */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.12               $ $Modtime::   Jan 11 2001 15:56:24   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <aibincl.h>

/**PROC+**********************************************************************/
/* Name:      aib_start_element_handler                                      */
/*                                                                           */
/* Purpose:   Handle the start tag of an XML element                         */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - A null-terminated string containing the  */
/*                                  XML element name.                        */
/*            IN     attrs        - A null-terminated array of               */
/*                                  null-terminated strings containing any   */
/*                                  attribute names and values that were     */
/*                                  specified with this XML tag.             */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_start_element_handler(void *user_data,
                               const XML_Char *name,
                               const XML_Char **attrs)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  TRANS_INFO *trans = (TRANS_INFO *)user_data;
  AIB_DATA *data;
  AIB_PARSE_FRAME *frame;
  AIB_ELEMENT *element;
  OSP_RC rc;
  char *critical_value = NULL;

  TRC_ENTRY("aib_start_element_handler");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL element name parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute list parameter"));

  /***************************************************************************/
  /* Trace input parameters.                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Input parameters: transaction=%p element name=\"%s\"",
                                                                 trans, name));

  /***************************************************************************/
  /* Extract parse data structure for this transaction.                      */
  /***************************************************************************/
  data = trans->aib_state;

  /***************************************************************************/
  /* Initially set the pointer to where we want character data placed to     */
  /* NULL.  This will stop the char data handler recording the data until    */
  /* we explicitly set up the location.                                      */
  /***************************************************************************/
  if (data != NULL)
  {
    TRC_DETAIL((TB, "Clear character data location"));
    data->char_data.start_ptr = NULL;
  }

  /***************************************************************************/
  /* If the parse data pointer is NULL, or if an error has previously        */
  /* occurred, or if we are currently skipping over an unrecognized element  */
  /* subtree, no further processing required.                                */
  /***************************************************************************/
  if ((data != NULL) &&
      (data->status.code == OSP_STATUS_SUCCESS) &&
      (data->ignore_depth == 0))
  {
    TRC_DETAIL((TB, "Non-NULL parse data: %p", data));

    frame = &(data->frames[data->depth]);
    TRC_DETAIL((TB, "Current parse frame:"
                " depth=%d elements=%p within_element=%d",
                data->depth,
                frame->elements,
                frame->within_element));

    /*************************************************************************/
    /* We shouldn't already be within an element (in this frame).            */
    /*************************************************************************/
    TRC_ASSERT((!frame->within_element),
               (TB, "Already within element:"
                " depth=%d elements=%p within_element=%d",
                data->depth,
                frame->elements,
                frame->within_element));

    frame->within_element = TRUE;

    /*************************************************************************/
    /* Look up the element name in the table of currently expected elements. */
    /*************************************************************************/
    element = aib_lookup_element_name(frame->elements, name);
    if (element != NULL)
    {
      TRC_DETAIL((TB, "Element name found: element=%p", element));

      /***********************************************************************/
      /* Either the element looked up must be the next element that we are   */
      /* expecting, or all of the elements from the one that we are          */
      /* expecting (inclusive) until the one that we have found (exclusive)  */
      /* must be optional.                                                   */
      /***********************************************************************/
      for (; frame->elements != element; frame->elements++)
      {
        TRC_DETAIL((TB, "Skipped element: name=\"%s\" occurrence=%#x",
                          frame->elements->name, frame->elements->occurrence));

        if ((frame->elements->occurrence == AIB_OCCURS_ONCE) ||
            ((frame->elements->occurrence == AIB_OCCURS_PLUS) &&
             (frame->num_occurrences == 0)))
        {
          TRC_ALERT((TB, "Tried to skip a required element:"
                     " name=\"%s\" occurrence=%#x",
                     frame->elements->name,
                     frame->elements->occurrence));
          data->status.code = OSP_STATUS_PARSING_UNSUCCESSFUL;
          data->status.desc = com_status_parsing_unsuccessful;
          goto EXIT_LABEL;
        }

        frame->num_occurrences = 0;
      }

      /***********************************************************************/
      /* Now invoke the element's handler.  Element handler parameters are   */
      /* designed to match aib_start_processing_simple_element, which is the */
      /* most common handler.                                                */
      /***********************************************************************/
      TRC_DETAIL((TB, "Invoke element handler: %p", element->handler));
      rc = (*element->handler)(trans,
                               name,
                               attrs,
                               ((OSP_BYTE *)frame->data)+ element->data_offset,
                               element->size,
                               element->extra_data);
      switch (rc)
      {
        case OSP_SUCCESS:
          TRC_DETAIL((TB, "Handler returned successfully"));
          frame->num_occurrences++;
          break;

        case OSP_ERR_MEMORY:
          TRC_ALERT((TB, "Memory allocation failure"));
          data->status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
          data->status.desc = com_status_transient_problem_in_server;
          goto EXIT_LABEL;
          break;

        default:
          TRC_ALERT((TB, "Unexpected handler return code: rc=%d", rc));
          data->status.code = OSP_STATUS_PARSING_UNSUCCESSFUL;
          data->status.desc = com_status_parsing_unsuccessful;
          goto EXIT_LABEL;
          break;
      }

      /***********************************************************************/
      /* If this is an element that cannot appear multiple times, move the   */
      /* frame's elements pointer on a place so that we don't expect to get  */
      /* the same element again.                                             */
      /***********************************************************************/
      if ((frame->elements->occurrence == AIB_OCCURS_ONCE) ||
          (frame->elements->occurrence == AIB_OCCURS_QUMK))
      {
        TRC_DETAIL((TB, "Advance frame's elements pointer"));
        frame->elements++;
        frame->num_occurrences = 0;
      }
    }
    else
    {
      TRC_NORMAL((TB, "Element name not found: name=\"%s\"", name));

      /***********************************************************************/
      /* Unless the element is marked as non-critical, we have a parsing     */
      /* error.  Note that the element name here could be a name that we     */
      /* don't support at all, or it could be a name from earlier in the     */
      /* current subelement list that is occurring out of order.             */
      /***********************************************************************/
      critical_value = aib_find_attr_value(attrs, com_critical, com_true);
      if (critical_value == NULL)
      {
        TRC_ERROR((TB, "Failed to get critical attribute"));
        data->status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
        data->status.desc = com_status_transient_problem_in_server;
        goto EXIT_LABEL;
      }

      /***********************************************************************/
      /* If the unknown element was critical, then set an error.  Note       */
      /* that we'll still parse as normal to reach the end of the component. */
      /***********************************************************************/
      if (strcasecmp(critical_value, com_true) == 0)
      {
        TRC_ALERT((TB, "Unknown critical element: name=\"%s\"", name));
        data->status.code = OSP_STATUS_PARSING_UNSUCCESSFUL;
        data->status.desc = com_status_parsing_unsuccessful;
        goto EXIT_LABEL;
      }
      else
      {
        /*********************************************************************/
        /* Ignore this element and any potential subtree inside it.          */
        /*********************************************************************/
        TRC_NORMAL((TB, "Start ignoring elements: name=\"%s\"", name));
        data->ignore_depth = 1;
      }
    }
  }
  else
  {
    /*************************************************************************/
    /* Why are we skipping over this element?                                */
    /*************************************************************************/
    if (data != NULL)
    {
      if (data->ignore_depth == 0)
      {
        /*********************************************************************/
        /* Because of an error...                                            */
        /*********************************************************************/
        TRC_DETAIL((TB, "Error previously detected: status=%d",
                                                           data->status.code));
      }
      else
      {
        /*********************************************************************/
        /* Because we're in the subtree of an unrecognized element...        */
        /* However, we do still need to check that the critical attribute is */
        /* present and set false                                             */
        /*********************************************************************/
        TRC_DETAIL((TB, "In unrecognized element subtree: ignore_depth=%d",
                                                          data->ignore_depth));
        critical_value = aib_find_attr_value(attrs, com_critical, com_true);
        if ((critical_value == NULL) ||
                                   (strcasecmp(critical_value, com_true) == 0))
        {
          TRC_ALERT((TB, "Unknown critical sub element: name=\"%s\"", name));
          data->status.code = OSP_STATUS_PARSING_UNSUCCESSFUL;
          data->status.desc = com_status_parsing_unsuccessful;
          goto EXIT_LABEL;
        }
        data->ignore_depth++;
      }
    }
    else
    {
      /***********************************************************************/
      /* Because of an error...                                              */
      /***********************************************************************/
      TRC_DETAIL((TB, "Error previously detected: NULL data"));
    }
  }

EXIT_LABEL:

  /***************************************************************************/
  /* Tidy up                                                                 */
  /***************************************************************************/
  rm_release_mem(critical_value);

  TRC_EXIT();

  return;

} /* aib_start_element_handler */

/**PROC+**********************************************************************/
/* Name:      aib_end_element_handler                                        */
/*                                                                           */
/* Purpose:   Handle the end tag of an XML element                           */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - A null-terminated string containing the  */
/*                                  XML element name.                        */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_end_element_handler(void *user_data,
                             const XML_Char *name)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  TRANS_INFO *trans = (TRANS_INFO *)user_data;
  AIB_DATA *data;
  AIB_CHAR_DATA *char_data;
  AIB_PARSE_FRAME *frame;
  OSP_RC rc;
  OSP_STATUS status;
  char *s;

  TRC_ENTRY("aib_end_element_handler");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL element name parameter"));

  /***************************************************************************/
  /* Trace input parameters.                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Input parameters: transaction=%p element name=\"%s\"",
                                                                 trans, name));

  /***************************************************************************/
  /* Get USE parse data.                                                     */
  /***************************************************************************/
  data = trans->aib_state;
  if (data != NULL)
  {
    /*************************************************************************/
    /* Strip any trailing whitespace from the data that was stored by the    */
    /* character data handler.                                               */
    /*************************************************************************/
    char_data = &(data->char_data);
    if ((char_data->start_ptr != NULL) &&
        (*(char_data->start_ptr) != NULL) &&
        (char_data->current_len > 0))
    {
      TRC_DETAIL((TB, "Check for trailing whitespace"));
      s = (*(char_data->start_ptr)) + (char_data->current_len - 1);
      while ((char_data->current_len > 0) &&
             ((s[0] == ' ') ||
              (s[0] == '\011') ||
              (s[0] == '\012') ||
              (s[0] == '\015')))
      {
        TRC_DETAIL((TB, "Trim whitespace char: code=%d", (int)(s[0])));
        s[0] = 0;
        s--;
        char_data->current_len--;
      }
      TRC_DETAIL((TB, "Trimmed character data: length=%d",
                                                      char_data->current_len));
      TRC_DATA("Character data:", *(char_data->start_ptr), 0);
    }

    /*************************************************************************/
    /* Advance the string pool position.                                     */
    /*************************************************************************/
    if (char_data->start_ptr != NULL)
    {
      trans->pool_position += char_data->current_len + 1;
      TRC_ASSERT((trans->pool_position < trans->pool_size),
                 (TB, "String pool overflowed"));
    }

    /*************************************************************************/
    /* Now clear the character data pointer.  This will stop the char data   */
    /* handler appending any further whitespace characters that occur        */
    /* between the end tag of one element and the start tag of the next.     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Clear character data location"));
    data->char_data.start_ptr = NULL;
  }

  /***************************************************************************/
  /* If USE parse data is NULL, or if an error has previously occurred, or   */
  /* if we are currently skipping over an unrecognized element subtree, no   */
  /* further processing is required except detecting the end of the          */
  /* UsageIndication element.                                                */
  /***************************************************************************/
  if ((data != NULL) &&
      (data->status.code == OSP_STATUS_SUCCESS) &&
      (data->ignore_depth == 0))
  {
    TRC_DETAIL((TB, "Non-NULL parse data: %p", data));

    frame = &(data->frames[data->depth]);
    TRC_DETAIL((TB, "Current parse frame:"
                " depth=%d elements=%p within_element=%d",
                data->depth, frame->elements, frame->within_element));

    /*************************************************************************/
    /* There are two cases to consider here.  (1) If the previous tag        */
    /* handler call was for a start tag, say for <SourceInfo ...>, then this */
    /* frame's within_element flag is currently set, and the end tag for     */
    /* which this handler has just been called matches that start tag and so */
    /* closes the element.  (We don't check for this matching, as it is a    */
    /* required by the well-formedness of the XML, which the XML layer is    */
    /* already checking.)  (2) If the previous tag handler call was for an   */
    /* end tag, this frame's within_element flag is FALSE, and the end tag   */
    /* in this handler call must match a start tag associated with the       */
    /* enclosing frame; hence this end tag closes the whole of the current   */
    /* frame and the current element of the enclosing frame.                 */
    /*************************************************************************/
    if (frame->within_element)
    {
      TRC_DETAIL((TB, "End of element in current frame"));
      frame->within_element = FALSE;
    }
    else
    {
      /***********************************************************************/
      /* Check that all the remaining elements in this frame (if any) are    */
      /* optional.                                                           */
      /***********************************************************************/
      for (; frame->elements->name != NULL; frame->elements++)
      {
        TRC_DETAIL((TB, "Skipped element: name=\"%s\" occurrence=%#x",
                          frame->elements->name, frame->elements->occurrence));

        if ((frame->elements->occurrence == AIB_OCCURS_ONCE) ||
            ((frame->elements->occurrence == AIB_OCCURS_PLUS) &&
             (frame->num_occurrences == 0)))
        {
          TRC_ALERT((TB, "Tried to skip a required element:"
                     " name=\"%s\" occurrence=%#x",
                     frame->elements->name, frame->elements->occurrence));
          data->status.code = OSP_STATUS_PARSING_UNSUCCESSFUL;
          data->status.desc = com_status_parsing_unsuccessful;
          break;
        }

        frame->num_occurrences = 0;
      }

      /***********************************************************************/
      /* Pop the stack.                                                      */
      /***********************************************************************/
      frame->elements = NULL;
      frame->data = NULL;
      data->depth--;
      TRC_DETAIL((TB, "Popped parse frame stack: new depth=%d", data->depth));

      if (data->depth >= 0)
      {
        frame = &(data->frames[data->depth]);
        TRC_DETAIL((TB, "Enclosing frame:"
                    " depth=%d elements=%p within_element=%d",
                    data->depth, frame->elements, frame->within_element));
        frame->within_element = FALSE;
      }
      else
      {
        /*********************************************************************/
        /* Parsing completion is officially detected below, by noting that   */
        /* the end tag is UsageIndication, since that method works both in   */
        /* this case and in the case where trans->aib_state is NULL.         */
        /*********************************************************************/
        TRC_DETAIL((TB, "Exited top level frame: parsing should be complete"));

#ifdef OSP_MEASURE_TRANSACTION
        /*********************************************************************/
        /* Set the transaction finish parse time.                            */
        /*********************************************************************/
        SM_GETTIMES(trans->timings[SM_TIMINGS_FINISH_PARSE]);
#endif
      }
    }
  }
  else
  {
    if (data != NULL)
    {
      if (data->ignore_depth == 0)
      {
        TRC_DETAIL((TB, "Error previously detected: status=%d",
                                                           data->status.code));
      }
      else
      {
        TRC_DETAIL((TB, "Ignoring unrecognized subtree: ignore_depth=%d",
                                                          data->ignore_depth));
        data->ignore_depth--;

        /*********************************************************************/
        /* If we're at the end tag of a tree we were ignoring (ignore depth  */
        /* == 0) then we need to clear the within_element flag)              */
        /*********************************************************************/
        if (data->ignore_depth == 0)
        {
          frame = &(data->frames[data->depth]);
          TRC_ASSERT((frame->within_element),
                     (TB, "Ignore depth reached zero but not within element"));
          frame->within_element = FALSE;
        }

        /*********************************************************************/
        /* If ignore_depth is now back to zero, we will automatically        */
        /* continue processing tags normally next time the start tag handler */
        /* is called.                                                        */
        /*********************************************************************/
      }
    }
    else
    {
      TRC_DETAIL((TB, "Error previously detected: NULL data"));
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return;
}

