
/**MOD+***********************************************************************/
/* Module:   aibrdut.c                                                       */
/*                                                                           */
/* Purpose:  AIB Read-side common functions                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.19               $ $Modtime::   Jan 11 2001 15:56:26   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <aibincl.h>

/**PROC+**********************************************************************/
/* Name:      aib_find_attr_value                                            */
/*                                                                           */
/* Purpose:   Find and copy the value of a specified attribute from an       */
/*            expat-provided XML attribute/value pairs list                  */
/*                                                                           */
/* Returns:   pointer to copy of the value or of supplied default if value   */
/*            not found                                                      */
/*                                                                           */
/* Params:    IN     attr_list    - the list to search                       */
/*            IN     attr         - the attribute to search for              */
/*            IN     def_val      - default value for the attribute          */
/*                                                                           */
/* Operation: Search the list for the attribute; if found, the value is the  */
/*            next item in the list.  Otherwise, return the default value    */
/*                                                                           */
/**PROC-**********************************************************************/

char *aib_find_attr_value(const char **attr_list,
                          const char *attr,
                          const char *def_val)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int found;
  char **pinto_attrs;
  char *value;

  TRC_ENTRY("aib_find_attr_value");

  TRC_DETAIL((TB, "looking for attr '%s'", attr));
  /***************************************************************************/
  /* Search the list                                                         */
  /***************************************************************************/
  pinto_attrs = (char **)attr_list;
  found = FALSE;
  while (*pinto_attrs)
  {
    TRC_DETAIL((TB, "looking at attr '%s'", *pinto_attrs));
    if (strcmp(*pinto_attrs, attr) == 0)
    {
      /***********************************************************************/
      /* Found the attribute - move onto the value                           */
      /***********************************************************************/
      pinto_attrs++;
      TRC_DETAIL((TB, "Found value '%s'", *pinto_attrs));
      found = TRUE;
      break;
    }

    /*************************************************************************/
    /* Move on past the current value and on to the next attribute           */
    /*************************************************************************/
    pinto_attrs += 2;
  }

  /***************************************************************************/
  /* If we didn't find a value, then use the supplied default                */
  /***************************************************************************/
  if (!found)
  {
    TRC_DETAIL((TB, "Return default value '%s'",
                (def_val != NULL) ? def_val : "(null)"));
    *pinto_attrs = (char *)def_val;
  }

  /***************************************************************************/
  /* If we found the value, make a copy                                      */
  /***************************************************************************/
  if (*pinto_attrs)
  {
    TRC_DETAIL((TB, "Copying value"));
    OSP_STRDUP(value, *pinto_attrs);
    if (value == NULL)
    {
      TRC_ERROR((TB, "Copy of value %s failed", *pinto_attrs));
    }
  }
  else
  {
    TRC_DETAIL((TB, "Returning NULL"));
    value = NULL;
  }

  TRC_EXIT();

  return(value);

} /* aib_find_attr_value */

/**PROC+**********************************************************************/
/* Name:      aib_character_data_handler                                     */
/*                                                                           */
/* Purpose:   To handle XML character data from a client component as it is  */
/*            parsed. (Defined in expat's API spec)                          */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to the transaction's control     */
/*                                  structure                                */
/*            IN     s            - the character data (not NULL-terminated) */
/*            IN     len          - length of s                              */
/*                                                                           */
/* Operation: Extract char data structure from trans info                    */
/*            Call on to internal handler                                    */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_character_data_handler(void *user_data, const XML_Char *s, int len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;
  AIB_CHAR_DATA *char_data;

  TRC_ENTRY("aib_character_data_handler");

  /***************************************************************************/
  /* Get the transaction info                                                */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_ASSERT((trans != NULL), (TB, "NULL user data in char data handler"));

  /***************************************************************************/
  /* Get the character data structure from the transaction and pass onto the */
  /* internal handler                                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Calling internal handler"));
  char_data = (AIB_CHAR_DATA *)trans->aib_state;
  aib_internal_character_data_handler((void *)char_data, s, len);

  TRC_EXIT();

  return;

} /* aib_character_data_handler */

/**PROC+**********************************************************************/
/* Name:      aib_internal_character_data_handler                            */
/*                                                                           */
/* Purpose:   To handle XML character data from a client component as it     */
/*            is parsed. (Defined in expat's API spec)                       */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to the char data parsing info    */
/*            IN     s            - the character data (not NULL-terminated) */
/*            IN     len          - length of s                              */
/*                                                                           */
/* Operation: Ensure there's space for the new data                          */
/*            Store the data                                                 */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_internal_character_data_handler(void *user_data,
                                         const XML_Char *s,
                                         int len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  AIB_CHAR_DATA *char_data;
  char *new_mem;
  int new_len;

  TRC_ENTRY("aib_internal_character_data_handler");

  /***************************************************************************/
  /* Get the character data structure                                        */
  /***************************************************************************/
  char_data = (AIB_CHAR_DATA *)user_data;
  if ((char_data == NULL) || (char_data->start_ptr == NULL))
  {
    /*************************************************************************/
    /* If there's nowhere to put the data, its because we're just parsing to */
    /* the end of the element, perhaps because there was no callback         */
    /* registered                                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "nowhere to put char data - ignore it"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If an error has already been sent, then there's no more to do           */
  /***************************************************************************/
  if (char_data->error_code != 0)
  {
    TRC_NORMAL((TB, "Error already sent - done"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Ensure the start pointer points to something                            */
  /***************************************************************************/
  TRC_ASSERT((*(char_data->start_ptr) != NULL),
                                      (TB, "Requested data location is NULL"));

  /***************************************************************************/
  /* If we don't yet have any non-whitespace data, skip over any leading     */
  /* whitespace characters in the current data.                              */
  /***************************************************************************/
  if (char_data->current_len == 0)
  {
    TRC_DETAIL((TB, "Check for leading whitespace"));
    while ((len > 0) &&
           ((s[0] == ' ') ||
            (s[0] == '\011') ||
            (s[0] == '\012') ||
            (s[0] == '\015')))
    {
      TRC_DETAIL((TB, "Skip whitespace char: code=%d", (int)(s[0])));
      s++;
      len--;
    }
  }
  TRC_DETAIL((TB, "Character data now: length=%d", len));

  /***************************************************************************/
  /* Copy the data.  Since the string is NOT null-terminated we have to      */
  /* - use memcpy not strcpy                                                 */
  /* - add our own terminator (we do expect null terminated strings when     */
  /*   we come to use the data later on)                                     */
  /***************************************************************************/
  memcpy(*(char_data->start_ptr) + char_data->current_len, s, len);
  char_data->current_len += len;
  *(*(char_data->start_ptr) + char_data->current_len) = 0;

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* aib_internal_character_data_handler */

/**PROC+**********************************************************************/
/* Name:      aib_init_char_data                                             */
/*                                                                           */
/* Purpose:   Performs initialization of structure used in storing character */
/*            data across successive expat callbacks                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS     - initialization successful                    */
/*            OSP_ERR_MEMORY - failed to allocate storage                    */
/*                                                                           */
/* Params:    IN     char_data    - structure to initialize                  */
/*            IN     len          - initial length to allocate               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_init_char_data(AIB_CHAR_DATA *char_data, int len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("aib_init_char_data");

  /***************************************************************************/
  /* Make sure there's somewhere to put the memory                           */
  /***************************************************************************/
  TRC_ASSERT((char_data != NULL), (TB, "Not given a structure to fill in"));
  TRC_ASSERT((char_data->start_ptr != NULL), (TB, "Nowhere to put memory"));
  TRC_ASSERT((len > 0), (TB, "Invalid initial data length: len=%d", len));

  /***************************************************************************/
  /* Point to next available location in the string pool.                    */
  /***************************************************************************/
  *(char_data->start_ptr) =
                      char_data->trans->pool + char_data->trans->pool_position;
  char_data->total_len =
                 char_data->trans->pool_size - char_data->trans->pool_position;
  char_data->current_len = 0;
  char_data->error_code = 0;
  **(char_data->start_ptr) = 0;
  TRC_DETAIL((TB, "Init string pool location at %p, available space %d",
              *(char_data->start_ptr),
              char_data->total_len));

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_init_char_data */

/**PROC+**********************************************************************/
/* Name:      aib_lookup_element_name                                        */
/*                                                                           */
/* Purpose:   Search for the specified element name in an element table,     */
/*            and return a pointer to the corresponding element structure,   */
/*            or NULL if the table does not contain the named element.       */
/*                                                                           */
/* Returns:   A pointer to the appropriate element structure, or NULL.       */
/*                                                                           */
/* Params:    IN     elements     - A pointer to the elements table in which */
/*                                  to look up the specified element name.   */
/*            IN     name         - A null-terminated string specifying the  */
/*                                  element name to look up.                 */
/*                                                                           */
/**PROC-**********************************************************************/

AIB_ELEMENT *aib_lookup_element_name(AIB_ELEMENT *elements, const char *name)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  AIB_ELEMENT *p;

  TRC_ENTRY("aib_lookup_element_name");

  /***************************************************************************/
  /* Check that the parameters supplied are non-NULL.                        */
  /***************************************************************************/
  TRC_ASSERT((elements != NULL), (TB, "NULL elements parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL element name parameter"));

  /***************************************************************************/
  /* Trace parameters.                                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Input parameters: elements=%p name=\"%s\"",
                                                              elements, name));

  /***************************************************************************/
  /* Scan through the elements table looking for the supplied name.          */
  /***************************************************************************/
  for (p = elements; p->name != NULL; p++)
  {
    TRC_DETAIL((TB, "Element table entry: name=\"%s\"", p->name));
    if (strcmp(p->name, name) == 0)
    {
      TRC_DETAIL((TB, "Found matching entry: p=%p", p));
      break;
    }
  }

  /***************************************************************************/
  /* If we didn't find a matching entry, arrange to return a NULL pointer.   */
  /***************************************************************************/
  if (p->name == NULL)
  {
    TRC_DETAIL((TB, "No matching entry - returning NULL pointer"));
    p = NULL;
  }

EXIT_LABEL:
  TRC_EXIT();
  return(p);

} /* aib_lookup_element_name */

/**PROC+**********************************************************************/
/* Name:      aib_free_call_id                                               */
/*                                                                           */
/* Purpose:   Free an OSP_CALL_ID structure                                  */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     call_id      - Pointer to structure to be freed.        */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_free_call_id(OSP_CALL_ID *call_id)
{
  TRC_ENTRY("aib_free_call_id");

  if (call_id != NULL)
  {
    TRC_DETAIL((TB, "Free call id: %p", call_id));
    rm_release_mem(call_id->encoding);
    rm_release_mem(call_id);
  }

  TRC_EXIT();
  return;

} /* aib_free_call_id */

/**PROC+**********************************************************************/
/* Name:      aib_free_address                                               */
/*                                                                           */
/* Purpose:   Free an OSP_ADDRESS structure                                  */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     address      - Pointer to structure to be freed.        */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_free_address(OSP_ADDRESS *address)
{
  TRC_ENTRY("aib_free_address");

  if (address != NULL)
  {
    TRC_DETAIL((TB, "Free address %p", address));
    rm_release_mem(address->type);
    rm_release_mem(address);
  }

  TRC_EXIT();
  return;

} /* aib_free_address */

/**PROC+**********************************************************************/
/* Name:      aib_free_service                                               */
/*                                                                           */
/* Purpose:   Free an OSP_SERVICE_INFO structure                             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     service      - Pointer to structure to be freed.        */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_free_service(OSP_SERVICE_INFO *service)
{
  OSP_PROTOCOL_TYPE *prot_type, *next;

  TRC_ENTRY("aib_free_service");

  /***************************************************************************/
  /* Check what we've been given                                             */
  /***************************************************************************/
  if (service != NULL)
  {
    TRC_DETAIL((TB, "Free service: %p", service));
    prot_type = service->prot_type;
    while (prot_type != NULL)
    {
      next = prot_type->next;
      rm_release_mem(prot_type);
      prot_type = next;
    }
    rm_release_mem(service);
  }

  TRC_EXIT();
  return;

} /* aib_free_service */

/**PROC+**********************************************************************/
/* Name:      aib_free_token                                                 */
/*                                                                           */
/* Purpose:   Free an OSP_TOKEN structure                                    */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     token      - Pointer to structure to be freed.          */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_free_token(OSP_TOKEN *token)
{
  TRC_ENTRY("aib_free_token");

  if (token != NULL)
  {
    TRC_DETAIL((TB, "Free token: %p", token));
    rm_release_mem(token->encoding);
    rm_release_mem(token);
  }

  TRC_EXIT();
  return;

} /* aib_free_token */

/**PROC+**********************************************************************/
/* Name:      aib_free_usage_detail                                          */
/*                                                                           */
/* Purpose:   Free an OSP_USAGE_DETAIL structure                             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     usage_detail - Pointer to structure to be freed.        */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_free_usage_detail(OSP_USAGE_DETAIL *usage_detail)
{
  TRC_ENTRY("aib_free_usage_detail");

  if (usage_detail != NULL)
  {
    TRC_DETAIL((TB, "Free service: %p", usage_detail->service));
    aib_free_service(usage_detail->service);

    TRC_DETAIL((TB, "Free termination cause: %p", usage_detail->term_cause));
    rm_release_mem(usage_detail->term_cause);

    if (usage_detail->statistics != NULL)
    {
      TRC_DETAIL((TB, "Free loss sent: %p",
                                         usage_detail->statistics->loss_sent));
      rm_release_mem(usage_detail->statistics->loss_sent);

      TRC_DETAIL((TB, "Free loss received: %p",
                                     usage_detail->statistics->loss_received));
      rm_release_mem(usage_detail->statistics->loss_received);

      TRC_DETAIL((TB, "Free one way delay: %p",
                                     usage_detail->statistics->one_way_delay));
      rm_release_mem(usage_detail->statistics->one_way_delay);

      TRC_DETAIL((TB, "Free round trip delay: %p",
                                  usage_detail->statistics->round_trip_delay));
      rm_release_mem(usage_detail->statistics->round_trip_delay);
    }

    TRC_DETAIL((TB, "Free statistics: %p", usage_detail->statistics));
    rm_release_mem(usage_detail->statistics);
  }

  TRC_DETAIL((TB, "Free usage detail: %p", usage_detail));
  rm_release_mem(usage_detail);

  TRC_EXIT();
  return;

} /* aib_free_usage_detail */

