
/**MOD+***********************************************************************/
/* Module:    aibsndut.c                                                     */
/*                                                                           */
/* Purpose:   AIB Send-side common functions                                 */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.14               $ $Modtime::   Jan 11 2001 15:56:26   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <aibincl.h>

/**PROC+**********************************************************************/
/* Name:      aib_space_available                                            */
/*                                                                           */
/* Purpose:   Ensure that space is available in the transaction send buffer  */
/*            for the specified amount of data, reallocating if necessary    */
/*            to make room.                                                  */
/*                                                                           */
/* Returns:   TRUE  - space is available                                     */
/*            FALSE - unable to allocate memory to make space available      */
/*                                                                           */
/* Params:    IN     trans        - transaction for the send buffer          */
/*            IN     len          - amount of space required                 */
/*                                                                           */
/* Operation: Ensure there's space in the send buffer for the new data,      */
/*            reallocing if necessary                                        */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_space_available(TRANS_INFO *trans, int len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  int new_len;
  char *new_buffer;

  TRC_ENTRY("aib_space_available");

  /***************************************************************************/
  /* If there's been an error already, there's no more to do.                */
  /***************************************************************************/
  if (trans->http_status != OSP_STATUS_SUCCESS)
  {
    TRC_NORMAL((TB, "Error already set in transaction - quit"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Check there's space for the data                                        */
  /***************************************************************************/
  if (trans->send_buffer_end + len > trans->send_buffer_size)
  {
    /*************************************************************************/
    /* There's not enough space - make some                                  */
    /*************************************************************************/
    new_len = trans->send_buffer_end + len + AIB_REALLOC_EXTRA_LEN;
    TRC_NORMAL((TB, "Reallocing send buffer %p to new len %d",
                                                 trans->send_buffer, new_len));
    new_buffer = rm_realloc_mem(trans->send_buffer, new_len);
    if (new_buffer != NULL)
    {
      TRC_DETAIL((TB, "Changed send buffer %p to %p",
                                              trans->send_buffer, new_buffer));
      trans->send_buffer = new_buffer;
      trans->send_buffer_size = new_len;
    }
    else
    {
      /***********************************************************************/
      /* Set the error code for the transaction - this will cause the smime  */
      /* layer to send an http failure for the entire transaction            */
      /***********************************************************************/
      TRC_ERROR((TB, "Send buffer realloc failed"));
      trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Space is now available.                                                 */
  /***************************************************************************/
  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_space_available */
