
/**MOD+***********************************************************************/
/* Module:   aibstrtp.c                                                      */
/*                                                                           */
/* Purpose:  AIB Common Start Processing functions                           */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.28               $ $Modtime::   Jan 11 2001 15:56:26   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <aibincl.h>

/**PROC+**********************************************************************/
/* Name:      aib_start_processing                                           */
/*                                                                           */
/* Purpose:   Start processing a top-level client component XML element      */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     parser       - Pointer to XML layer's parser object.    */
/*                                  This is simply passed on to other XML    */
/*                                  layer routines.                          */
/*            IN     comp_id      - The component ID attribute value of the  */
/*                                  XML element that led to this call.       */
/*            IN     size         - Size of parse data structure to          */
/*                                  allocate.                                */
/*            IN     elements     - Pointer to an array of AIB_ELEMENTs that */
/*                                  specifies the expected elements of the   */
/*                                  client component.                        */
/*            IN     end_handler  - Pointer to a function that handles       */
/*                                  end tags for this component.             */
/*            OUT    status       - Status of attempt to start processing    */
/*                                  this client component.  If status->code  */
/*                                  is not OSP_STATUS_SUCCESS when this      */
/*                                  function returns, the caller should      */
/*                                  immediately construct an error response  */
/*                                  using the code and description fields    */
/*                                  of the status.                           */
/*                                                                           */
/* Operation: Allocate a structure to store parsing data; store a pointer    */
/*            to this structure in the transaction; register generic         */
/*            element and character data handlers with the XML layer.        */
/*                                                                           */
/**PROC-**********************************************************************/

void aib_start_processing(TRANS_INFO *trans,
                          XML_Parser *parser,
                          char *comp_id,
                          int size,
                          AIB_ELEMENT *elements,
                          XML_EndElementHandler end_handler,
                          OSP_STATUS *status)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  AIB_DATA *data;

  TRC_ENTRY("aib_start_processing");

  /***************************************************************************/
  /* Assume that function will succeed.                                      */
  /***************************************************************************/
  status->code = OSP_STATUS_SUCCESS;
  status->desc = com_status_success;

  /***************************************************************************/
  /* Register generic AIB element and data handlers with the expat library.  */
  /* Note that end_handler should be a thin AIB-specific wrapper around the  */
  /* common function aib_end_element_handler: the only AIB-specific part in  */
  /* end_handler is that which identifies the end of the client component    */
  /* and calls the appropriate application callback.                         */
  /***************************************************************************/
  XML_SetElementHandler(*parser, aib_start_element_handler, end_handler);
  XML_SetCharacterDataHandler(*parser, aib_character_data_handler);

  /***************************************************************************/
  /* Allocate memory to hold the component data as it is being parsed, and   */
  /* store a pointer to this memory in the transaction.  If a memory         */
  /* allocation fails, we must construct an error response immediately, as   */
  /* we have nowhere to store the component ID; having done that, all we     */
  /* need to do is wait until we detect the end of the top level component.  */
  /* In the normal case, and in the case of parsing errors that occur in the */
  /* middle of the component, we construct the response once we've detected  */
  /* the end of the client component.                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Allocate memory for component parse data"));
  trans->aib_state = rm_get_mem(size);
  if (trans->aib_state != NULL)
  {
    TRC_DETAIL((TB, "Initialize AIB parse data: %p", trans->aib_state));

    /*************************************************************************/
    /* Initialize parse data fields.                                         */
    /*************************************************************************/
    data = trans->aib_state;

    /*************************************************************************/
    /* We don't want to record any character data until we've correctly set  */
    /* up the destinations for it - this will ensure that we don't.          */
    /*************************************************************************/
    data->char_data.start_ptr = NULL;
    data->char_data.current_len = 0;
    data->char_data.total_len = 0;
    data->char_data.error_code = 0;

    /*************************************************************************/
    /* Interpretation:                                                       */
    /*   - parsing begins at the top level looking for the set of elements   */
    /*     provided by the caller                                            */
    /*   - we haven't yet encountered any of those subelements, so we are    */
    /*     expecting next to see the subelement at index 0                   */
    /*   - we aren't in between the start and end tag of any subelement      */
    /*   - the reference memory pointer for storing data associated with     */
    /*     these subelements is the top level USE_DATA structure pointer     */
    /*   - we aren't skipping over any unrecognized element subtrees, so     */
    /*     ignore_depth is zero.                                             */
    /*                                                                       */
    /*************************************************************************/
    data->depth = 0;
    data->frames[0].elements = elements;
    data->frames[0].num_occurrences = 0;
    data->frames[0].within_element = FALSE;
    data->frames[0].data = data;
    data->ignore_depth = 0;

    /*************************************************************************/
    /* No component-level error has occurred yet.                            */
    /*************************************************************************/
    data->status.code = OSP_STATUS_SUCCESS;
    data->status.desc = com_status_success;

    /*************************************************************************/
    /* Allocate a correlator structure for use when calling the relevant     */
    /* application callback.  This is allocated dynamically, rather than     */
    /* being in line in the AIB_DATA structure, because we need it to live a */
    /* little longer than the rest of the AIB_DATA.  It is allocated here,   */
    /* rather than at the point of invoking the callback, in order to reduce */
    /* the code path complexity later on.                                    */
    /*************************************************************************/
    data->corr = rm_get_mem(sizeof(OSP_INT_CORRELATOR));
    if (data->corr != NULL)
    {
      TRC_DETAIL((TB, "Allocated correlator structure: %p", data->corr));
      data->corr->trans = trans;
      data->corr->allocated = TRUE;

      /***********************************************************************/
      /* Copy the component ID: we will need it when composing the component */
      /* reponse.                                                            */
      /***********************************************************************/
      OSP_STRDUP(data->corr->comp_id, comp_id);
      if (data->corr->comp_id == NULL)
      {
        TRC_ERROR((TB, "Failed to allocate memory for component ID"));

        /*********************************************************************/
        /* Free the whole correlator.  This will in turn cause us to free    */
        /* the whole AIB_DATA.                                               */
        /*********************************************************************/
        rm_release_mem(data->corr);
        data->corr = NULL;
      }
    }

    if (data->corr == NULL)
    {
      TRC_ERROR((TB, "Failed to allocate memory for correlator info"));

      /***********************************************************************/
      /* Free the whole structure.  Parsing will continue as though memory   */
      /* allocation for the whole structure had failed.  (See the following  */
      /* block comment.)                                                     */
      /***********************************************************************/
      rm_release_mem(trans->aib_state);
      trans->aib_state = NULL;
      data = NULL;
      status->code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      status->desc = com_status_transient_problem_in_server;
    }

#ifdef OSP_DATA_COVERAGE
    if (data != NULL)
    {
      TRC_DETAIL((TB, "Register structure for data coverage"));
      aib_register_data_structure(elements, data);
    }
#endif
  }
  else
  {
    TRC_ALERT((TB, "Failed to allocate memory for USE parse data"));

    /*************************************************************************/
    /* Arrange for a UsageConfirmation response with an                      */
    /* OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER status code.                   */
    /*************************************************************************/
    status->code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    status->desc = com_status_transient_problem_in_server;
  }

  TRC_EXIT();
  return;

} /* aib_start_processing */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_simple_element                            */
/*                                                                           */
/* Purpose:   Set up char data handling for an element that has neither      */
/*            attributes or child elements                                   */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the element     */
/*                                  data pointer should be stored.           */
/*            IN     initial_len  - Suggested initial length of storage      */
/*                                  to allocate.                             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_simple_element(TRANS_INFO *trans,
                                           const char *name,
                                           const char **attrs,
                                           void *dest,
                                           int initial_len,
                                           void *ignored)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char **dest_addr = (char **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;

  TRC_ENTRY("aib_start_processing_simple_element");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest_addr != NULL), (TB, "NULL location parameter"));

  /***************************************************************************/
  /* Get parse data.  If this was NULL we would not have been called.        */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL data"));

  /***************************************************************************/
  /* Set up where we want the character data to go                           */
  /***************************************************************************/
  char_data->start_ptr = dest_addr;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, initial_len) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for timestamp data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_simple_element */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_call_id                                   */
/*                                                                           */
/* Purpose:   Start processing a call_id element                             */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where a pointer to    */
/*                                  the allocated OSP_CALL_ID structure      */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to array of null-terminated      */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_call_id(TRANS_INFO *trans,
                                    const char *name,
                                    const char **attrs,
                                    void *dest,
                                    int ignored1,
                                    void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_CALL_ID **dest_addr = (OSP_CALL_ID **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;
  OSP_CALL_ID *call_id;

  TRC_ENTRY("aib_start_processing_call_id");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute list parameter"));

  /***************************************************************************/
  /* Assert that we got somewhere to do the setup                            */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL char data ptr supplied"));

  /***************************************************************************/
  /* Assert that the element is what we expect                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Passed element %s", name))
  TRC_ASSERT((strcmp(name, com_call_id) == 0),
                                          (TB, "Unexpected element %s", name));

  /***************************************************************************/
  /* Allocate the memory where we will store resulting call_id information   */
  /***************************************************************************/
  *dest_addr = rm_get_mem(sizeof(**dest_addr));
  if (*dest_addr == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for call_id"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  call_id = *dest_addr;
  call_id->next = NULL;

  /***************************************************************************/
  /* Get the encoding attribute - default is cdata                           */
  /***************************************************************************/
  call_id->encoding = aib_find_attr_value(attrs, com_encoding, com_cdata);
  if (call_id->encoding == NULL)
  {
    /*************************************************************************/
    /* Must be because of a memory error, since we supplied a default        */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to get encoding"));
    rc = OSP_ERR_MEMORY;
  }

  /***************************************************************************/
  /* Set up where we want the call id data to go                             */
  /***************************************************************************/
  char_data->start_ptr = &call_id->data;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, AIB_DEF_CALLID_LEN) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for call_id data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_call_id */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_address                                   */
/*                                                                           */
/* Purpose:   Start processing an address element                            */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to a memory location where the   */
/*                                  newly allocated OSP_ADDRESS structure's  */
/*                                  address (pointer) should be stored.      */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_address(TRANS_INFO *trans,
                                    const char *name,
                                    const char **attrs,
                                    void *dest,
                                    int ignored1,
                                    void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_ADDRESS **dest_addr = (OSP_ADDRESS **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;
  OSP_ADDRESS *address;

  TRC_ENTRY("aib_start_processing_address");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute array parameter"));

  /***************************************************************************/
  /* Assert that we got somewhere to do the setup                            */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL char data ptr supplied"));

  /***************************************************************************/
  /* Assert that the element is one of those we expected                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Passed element %s", name))
  TRC_ASSERT(((strcmp(name, com_dest_info) == 0) ||
              (strcmp(name, com_dest_alt) == 0) ||
              (strcmp(name, com_source_info) == 0) ||
              (strcmp(name, com_source_alt) == 0)),
             (TB, "Unrecognized element %s", name));

  /***************************************************************************/
  /* Allocate the memory where we will store resulting address information   */
  /***************************************************************************/
  *dest_addr = rm_get_mem(sizeof(**dest_addr));
  if (*dest_addr == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for address"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  address = *dest_addr;
  address->next = NULL;

  /***************************************************************************/
  /* Get the type attribute                                                  */
  /***************************************************************************/
  address->type = aib_find_attr_value(attrs, com_type, NULL);
  if (address->type == NULL)
  {
    /*************************************************************************/
    /* Type is a required attribute                                          */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to get type!"));
    rm_release_mem(address);
    *dest_addr = NULL;
    address = NULL;
    rc = OSP_ERR_MISSING_ATTR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up where we want the address string to go                           */
  /***************************************************************************/
  char_data->start_ptr = &address->address;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, AIB_DEF_ADDRESS_LEN) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for address data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_address */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_token                                     */
/*                                                                           */
/* Purpose:   Start processing a token element                               */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_token(TRANS_INFO *trans,
                                  const char *name,
                                  const char **attrs,
                                  void *dest,
                                  int ignored1,
                                  void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_TOKEN **dest_addr = (OSP_TOKEN **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;
  OSP_TOKEN *token;

  TRC_ENTRY("aib_start_processing_token");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute array parameter"));

  /***************************************************************************/
  /* Assert that we got somewhere to do the setup                            */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL char data ptr supplied"));

  /***************************************************************************/
  /* Assert that the element is what we expect                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Passed element %s", name))
  TRC_ASSERT((strcmp(name, com_token) == 0),
                                          (TB, "Unexpected element %s", name));

  /***************************************************************************/
  /* Allocate the memory where we will store resulting token information     */
  /***************************************************************************/
  *dest_addr = rm_get_mem(sizeof(**dest_addr));
  if (*dest_addr == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for token"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  token = *dest_addr;
  token->next = NULL;
  token->token_len = 0;

  /***************************************************************************/
  /* Get the encoding attribute - default is cdata                           */
  /***************************************************************************/
  token->encoding = aib_find_attr_value(attrs, com_encoding, com_cdata);
  if (token->encoding == NULL)
  {
    /*************************************************************************/
    /* Must be because of a memory error, since we supplied a default        */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to get encoding"));
    rc = OSP_ERR_MEMORY;
  }

  /***************************************************************************/
  /* Set up where we want the token data to go                               */
  /***************************************************************************/
  char_data->start_ptr = &token->token_data;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, AIB_DEF_TOKEN_LEN) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for token data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_token */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_subs_auth                                 */
/*                                                                           */
/* Purpose:   Start processing a subscriber authentication info element      */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN     dest_addr    - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_subs_auth(TRANS_INFO *trans,
                                      const char *name,
                                      const char **attrs,
                                      void *dest,
                                      int ignored,
                                      void *extra_data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_AUTH_INFO **dest_addr = (OSP_AUTH_INFO **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;
  OSP_AUTH_INFO *subs_auth;

  TRC_ENTRY("aib_start_processing_subs_auth");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute array parameter"));

  /***************************************************************************/
  /* Assert that we got somewhere to do the setup                            */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL char data ptr supplied"));

  /***************************************************************************/
  /* Assert that the element is what we expect                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Passed element %s", name))
  TRC_ASSERT((strcmp(name, com_subs_authen) == 0),
                                          (TB, "Unexpected element %s", name));

  /***************************************************************************/
  /* Allocate the memory where we will store the resulting authentication    */
  /* information                                                             */
  /***************************************************************************/
  *dest_addr = rm_get_mem(sizeof(**dest_addr));
  if (*dest_addr == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for auth info"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  subs_auth = *dest_addr;
  subs_auth->next = NULL;

  /***************************************************************************/
  /* Get the encoding attribute - default is cdata                           */
  /***************************************************************************/
  subs_auth->encoding = aib_find_attr_value(attrs, com_encoding, com_cdata);
  if (subs_auth->encoding == NULL)
  {
    /*************************************************************************/
    /* Must be because of a memory error, since we supplied a default        */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to get encoding"));
    rc = OSP_ERR_MEMORY;
  }

  /***************************************************************************/
  /* Set up where we want the authentication information to go               */
  /***************************************************************************/
  char_data->start_ptr = &subs_auth->auth_data;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, AIB_DEF_AUTH_INFO_LEN) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for subs auth data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_subs_auth */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_substructure                              */
/*                                                                           */
/* Purpose:   Start processing an XML element that is a container for a      */
/*            group of XML subelements                                       */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - processing started successfully          */
/*            OSP_ERR_MEMORY      - couldn't alloc memory for substructure   */
/*            OSP_ERROR           - an error occurred                        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to a memory location where the   */
/*                                  newly allocated data structure's address */
/*                                  (pointer) should be stored.              */
/*            IN     size         - Amount of memory in bytes to allocate    */
/*                                  for the required substructure.           */
/*            IN     subelements  - Pointer to an array of AIB_ELEMENTs that */
/*                                  specifies the subelements for the        */
/*                                  required substructure.                   */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_substructure(TRANS_INFO *trans,
                                         const char *name,
                                         const char **attrs,
                                         void *dest,
                                         int size,
                                         void *subelements)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  void **location = (void **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_DATA *data;
  AIB_PARSE_FRAME *frame;
  int num_pointers;
  int i;
  char **ptr;

  TRC_ENTRY("aib_start_processing_substructure");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((location != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((size != 0), (TB, "Zero size parameter"));
  TRC_ASSERT((subelements != NULL), (TB, "NULL elements parameter"));

  /***************************************************************************/
  /* Get parse data.  If this was NULL we would not have been called.        */
  /***************************************************************************/
  data = trans->aib_state;
  TRC_ASSERT((data != NULL), (TB, "NULL data"));

  /***************************************************************************/
  /* Initialize a new parsing stack frame in which to read the subelements   */
  /* of the substructure.                                                    */
  /***************************************************************************/
  data->depth++;
  TRC_ASSERT((data->depth < AIB_MAX_PARSE_DEPTH),
             (TB, "Max parsing depth exceeded: depth=%d", data->depth));
  frame = &(data->frames[data->depth]);
  frame->elements = subelements;
  frame->num_occurrences = 0;
  frame->within_element = FALSE;
  frame->data = NULL;
  TRC_DETAIL((TB, "New parse frame:"
              " depth=%d elements=%p within_element=%d",
              data->depth, frame->elements, frame->within_element));

  /***************************************************************************/
  /* Allocate a new structure of the required size, and store its address    */
  /* both in the caller's location pointer and the new frame's data slot.    */
  /***************************************************************************/
  *location = rm_get_mem(size);
  frame->data = *location;
  if ((*location) == NULL)
  {
    TRC_ALERT((TB, "Failed to allocate memory for new structure"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Allocated new structure memory: size=%d ptr=%p",
                                                             size, *location));

  /***************************************************************************/
  /* Initialize all the pointers in the new substructure to NULL.  Since all */
  /* of our substructures contain ONLY pointers, we can do this as follows.  */
  /* Otherwise, we'd have to write a specific initialization function for    */
  /* each substructure.                                                      */
  /***************************************************************************/
  num_pointers = size / (sizeof(char *));
  ptr = (char **)(*location);
  for (i = 0; i < num_pointers; i++)
  {
    TRC_DETAIL((TB, "Initialize pointer to NULL: i=%d", i));
    ptr[i] = NULL;
  }
  TRC_DETAIL((TB, "Substructure initialized: num_pointers=%d", num_pointers));

#ifdef OSP_DATA_COVERAGE
  TRC_DETAIL((TB, "Register structure for data coverage"));
  aib_register_data_structure(subelements, *location);
#endif

EXIT_LABEL:
  TRC_EXIT();
  return(rc);

} /* aib_start_processing_substructure */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_list                                      */
/*                                                                           */
/* Purpose:   Start processing one of the XML elements that can occur        */
/*            multiple times and so has its data stored in a linked list     */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - processing started successfully          */
/*            OSP_ERROR           - an error occurred                        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to a memory location where the   */
/*                                  head of the list is stored.              */
/*            IN     extra_data   - Pointer to function that handles a       */
/*                                  single item of the expected type for     */
/*                                  this XML element.                        */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_list(TRANS_INFO *trans,
                                 const char *name,
                                 const char **attrs,
                                 void *dest,
                                 int ignored,
                                 void *extra_data)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  SLIST_ROOT *list_ptr = (SLIST_ROOT *)dest;
  AIB_HANDLER handler = (AIB_HANDLER)extra_data;
  OSP_RC rc = OSP_SUCCESS;
  void *address;

  TRC_ENTRY("aib_start_processing_list");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute array parameter"));
  TRC_ASSERT((list_ptr != NULL), (TB, "NULL list pointer parameter"));
  TRC_ASSERT((handler != NULL), (TB, "NULL handler parameter"));

  /***************************************************************************/
  /* Call single item handler to allocate memory for this occurrence and     */
  /* initialize character data.                                              */
  /***************************************************************************/
  rc = (*handler)(trans, name, attrs, &address, 0, NULL);
  TRC_DETAIL((TB, "Return from handler: rc=%d", rc));

  /***************************************************************************/
  /* If that call succeeded, add the new structure to the end of the list.   */
  /***************************************************************************/
  if (rc == OSP_SUCCESS)
  {
    TRC_DETAIL((TB, "Add new item to end of list: address=%p", address));
    SL_ADD_TO_TAIL(*list_ptr, address);
  }

EXIT_LABEL:
  TRC_EXIT();
  return(rc);
}

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_usage_detail                              */
/*                                                                           */
/* Purpose:   Start processing a UsageDetail XML element                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - processing started successfully          */
/*            OSP_ERROR           - an error occurred                        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to a memory location where the   */
/*                                  newly allocated OSP_USAGE_DETAIL         */
/*                                  structure's address (pointer) should be  */
/*                                  stored.                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_usage_detail(TRANS_INFO *trans,
                                         const char *name,
                                         const char **attrs,
                                         void *dest,
                                         int ignored1,
                                         void *ignored2)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_USAGE_DETAIL **usage_detail = (OSP_USAGE_DETAIL **)dest;
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("aib_start_processing_usage_detail");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((usage_detail != NULL), (TB, "NULL usage detail parameter"));

  /***************************************************************************/
  /* Call generic routine for allocating a new structure and corresponding   */
  /* parse stack frame.                                                      */
  /***************************************************************************/
  rc = aib_start_processing_substructure(trans,
                                         name,
                                         attrs,
                                         (void **)usage_detail,
                                         sizeof(OSP_USAGE_DETAIL),
                                         aib_usage_detail_elements);

  TRC_EXIT();
  return(rc);

} /* aib_start_processing_usage_detail */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_capability                                */
/*                                                                           */
/* Purpose:   Start processing a capability element                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_capability(TRANS_INFO *trans,
                                       const char *name,
                                       const char **attrs,
                                       void *dest,
                                       int ignored1,
                                       void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_CAPS **dest_addr = (OSP_CAPS **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;
  OSP_CAPS *capability;

  TRC_ENTRY("aib_start_processing_capability");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute array parameter"));

  /***************************************************************************/
  /* Assert that we got somewhere to do the setup                            */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL char data ptr supplied"));

  /***************************************************************************/
  /* Assert that the element is what we expect                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Passed element %s", name))
  TRC_ASSERT((strcmp(name, com_osp_capability) == 0),
                                          (TB, "Unexpected element %s", name));

  /***************************************************************************/
  /* Allocate the memory where we will store the resulting capability        */
  /* information                                                             */
  /***************************************************************************/
  *dest_addr = rm_get_mem(sizeof(**dest_addr));
  if (*dest_addr == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for capability"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  capability = *dest_addr;
  capability->next = NULL;
  capability->cap_name = NULL;

  /***************************************************************************/
  /* Set up where we want the capability name to go                          */
  /***************************************************************************/
  char_data->start_ptr = &capability->cap_name;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, AIB_DEF_CAPABILITY_LEN) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for capability data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_capability */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_dev_info                                  */
/*                                                                           */
/* Purpose:   Start processing a device info element                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_dev_info(TRANS_INFO *trans,
                                     const char *name,
                                     const char **attrs,
                                     void *dest,
                                     int ignored1,
                                     void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_DEVICE_INFO **dest_addr = (OSP_DEVICE_INFO **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;
  OSP_DEVICE_INFO *dev_info;

  TRC_ENTRY("aib_start_processing_dev_info");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute array parameter"));

  /***************************************************************************/
  /* Assert that we got somewhere to do the setup                            */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL char data ptr supplied"));

  /***************************************************************************/
  /* Assert that the element is what we expect                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Passed element %s", name))
  TRC_ASSERT((strcmp(name, com_dev_info) == 0),
                                          (TB, "Unexpected element %s", name));

  /***************************************************************************/
  /* Allocate the memory where we will store the resulting device            */
  /* information                                                             */
  /***************************************************************************/
  *dest_addr = rm_get_mem(sizeof(**dest_addr));
  if (*dest_addr == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for dev_info"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  dev_info = *dest_addr;
  dev_info->type = NULL;
  dev_info->data = NULL;

  /***************************************************************************/
  /* Get the device info type                                                */
  /***************************************************************************/
  dev_info->type = aib_find_attr_value(attrs, com_type, NULL);
  if (dev_info->type == NULL)
  {
    TRC_ERROR((TB, "Failed to get type"));
    rc = OSP_ERR_MEMORY;
  }

  /***************************************************************************/
  /* Set up where we want the data to go                                     */
  /***************************************************************************/
  char_data->start_ptr = &dev_info->data;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, AIB_DEF_ADDRESS_LEN) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for dev_info data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_dev_info */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_protocol_type                             */
/*                                                                           */
/* Purpose:   Start processing a protocol type element                       */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_protocol_type(TRANS_INFO *trans,
                                          const char *name,
                                          const char **attrs,
                                          void *dest,
                                          int ignored1,
                                          void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_PROTOCOL_TYPE **dest_addr = (OSP_PROTOCOL_TYPE **)dest;
  OSP_RC rc = OSP_SUCCESS;
  AIB_CHAR_DATA *char_data;
  OSP_PROTOCOL_TYPE *prot_type;

  TRC_ENTRY("aib_start_processing_protocol_type");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((dest != NULL), (TB, "NULL location parameter"));
  TRC_ASSERT((attrs != NULL), (TB, "NULL attribute array parameter"));

  /***************************************************************************/
  /* Assert that we got somewhere to do the setup                            */
  /***************************************************************************/
  char_data = &(((AIB_DATA *)(trans->aib_state))->char_data);
  TRC_ASSERT((char_data != NULL), (TB, "NULL char data ptr supplied"));

  /***************************************************************************/
  /* Assert that the element is what we expect                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Passed element %s", name))
  TRC_ASSERT((strcmp(name, com_protocol_type) == 0),
                                          (TB, "Unexpected element %s", name));

  /***************************************************************************/
  /* Allocate the memory where we will store the resulting protocol type     */
  /***************************************************************************/
  *dest_addr = rm_get_mem(sizeof(**dest_addr));
  if (*dest_addr == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for prot_type"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  prot_type = *dest_addr;
  prot_type->next = NULL;
  prot_type->prot_name = NULL;

  /***************************************************************************/
  /* Set up where we want the data to go                                     */
  /***************************************************************************/
  char_data->start_ptr = &prot_type->prot_name;
  char_data->trans = trans;
  if (aib_init_char_data(char_data, AIB_DEF_PROT_TYPE_LEN) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get memory for prot_type data"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_protocol_type */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_resource                                  */
/*                                                                           */
/* Purpose:   Start processing a resource element                            */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_resource(TRANS_INFO *trans,
                                     const char *name,
                                     const char **attrs,
                                     void *dest,
                                     int ignored1,
                                     void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  AIB_RESOURCE **resource = (AIB_RESOURCE **)dest;
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("aib_start_processing_resource");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((resource != NULL), (TB, "NULL resource parameter"));

  /***************************************************************************/
  /* Call generic routine for allocating a new structure and corresponding   */
  /* parse stack frame.                                                      */
  /***************************************************************************/
  rc = aib_start_processing_substructure(trans,
                                         name,
                                         attrs,
                                         (void **)resource,
                                         sizeof(AIB_RESOURCE),
                                         aib_resource_elements);

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_resource */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_data_rate                                 */
/*                                                                           */
/* Purpose:   Start processing a data rate element                           */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_data_rate(TRANS_INFO *trans,
                                      const char *name,
                                      const char **attrs,
                                      void *dest,
                                      int ignored1,
                                      void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_DATA_RATE **data_rate = (OSP_DATA_RATE **)dest;
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("aib_start_processing_data_rate");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((data_rate != NULL), (TB, "NULL data rate parameter"));

  /***************************************************************************/
  /* Call generic routine for allocating a new structure and corresponding   */
  /* parse stack frame.                                                      */
  /***************************************************************************/
  rc = aib_start_processing_substructure(trans,
                                         name,
                                         attrs,
                                         (void **)data_rate,
                                         sizeof(OSP_DATA_RATE),
                                         aib_data_rate_elements);

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_data_rate */

/**PROC+**********************************************************************/
/* Name:      aib_start_processing_supp_prot                                 */
/*                                                                           */
/* Purpose:   Start processing a supported protocol element                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - another error occurrred                  */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - Name of the current XML element.         */
/*            IN     attrs        - Attribute list of the current element.   */
/*            IN/OUT dest         - Pointer to address where the results     */
/*                                  should be stored.                        */
/*            IN     extra_data   - Pointer to a null-terminated array of    */
/*                                  XML attribute names and values.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC aib_start_processing_supp_prot(TRANS_INFO *trans,
                                      const char *name,
                                      const char **attrs,
                                      void *dest,
                                      int ignored1,
                                      void *ignored2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_SUPP_PROTOCOL **supp_prot = (OSP_SUPP_PROTOCOL **)dest;
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("aib_start_processing_supp_prot");

  /***************************************************************************/
  /* Check input parameters.                                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL name parameter"));
  TRC_ASSERT((supp_prot != NULL), (TB, "NULL supp prot parameter"));

  /***************************************************************************/
  /* Call generic routine for allocating a new structure and corresponding   */
  /* parse stack frame.                                                      */
  /***************************************************************************/
  rc = aib_start_processing_substructure(trans,
                                         name,
                                         attrs,
                                         (void **)supp_prot,
                                         sizeof(AIB_SUPP_PROTOCOL),
                                         aib_supp_protocol_elements);

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_start_processing_supp_prot */

