
/**MOD+***********************************************************************/
/* Module:    aibwrlst.c                                                     */
/*                                                                           */
/* Purpose:   AIB Common functions to write lists of OSP types as XML        */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.8                $ $Modtime::   Jan 11 2001 15:56:26   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <aibincl.h>

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_token_list                                       */
/*                                                                           */
/* Purpose:   Write a list of tokens as XML                                  */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     name         - element name to use for token            */
/*            IN     token_list   - list of tokens to be written             */
/*                                                                           */
/* Operation: Walk the list of tokens, writing each one as XML               */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_token_list(TRANS_INFO *trans,
                             char *name,
                             OSP_TOKEN *token_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  OSP_TOKEN *curr_token;

  TRC_ENTRY("aib_write_xml_token_list");

  /***************************************************************************/
  /* Walk the list, writing entry each in turn                               */
  /***************************************************************************/
  for (curr_token = (OSP_TOKEN *)token_list;
       (curr_token != NULL) && rc;
       curr_token = (OSP_TOKEN *)curr_token->next)
  {
    TRC_DETAIL((TB, "Adding token %p", curr_token));
    rc = aib_write_encoded_element(trans,
                                   name,
                                   curr_token->encoding,
                                   curr_token->token_len,
                                   curr_token->token_data);
  }

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_token_list */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_destination_list                                 */
/*                                                                           */
/* Purpose:   Write a list of destinations as XML                            */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     dest_list    - list of dests to be written              */
/*                                                                           */
/* Operation: Walk the list of destinations, writing each one as XML         */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_destination_list(TRANS_INFO *trans,
                                   OSP_DESTINATION *dest_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  OSP_DESTINATION *curr_dest;

  TRC_ENTRY("aib_write_xml_destination_list");

  /***************************************************************************/
  /* Walk the list, writing entry each in turn                               */
  /***************************************************************************/
  for (curr_dest = (OSP_DESTINATION *)dest_list;
       (curr_dest != NULL) && rc;
       curr_dest = (OSP_DESTINATION *)curr_dest->next)
  {
    TRC_DETAIL((TB, "Adding dest %p", curr_dest));
    rc = aib_write_xml_destination(trans, curr_dest);
  }

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_destination_list */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_address_list                                     */
/*                                                                           */
/* Purpose:   Write a list of OSP_ADDRESS structures as XML                  */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     element_name - name to use for the elements             */
/*            IN     addr_list    - list of addresses to be written          */
/*                                                                           */
/* Operation: Walk the list of addresses, writing each one as XML            */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_address_list(TRANS_INFO *trans,
                               const char *element_name,
                               OSP_ADDRESS *addr_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  OSP_ADDRESS *curr_addr;

  TRC_ENTRY("aib_write_xml_address_list");

  /***************************************************************************/
  /* Walk the list, writing entry each in turn                               */
  /***************************************************************************/
  for (curr_addr = (OSP_ADDRESS *)addr_list;
       (curr_addr != NULL) && rc;
       curr_addr = (OSP_ADDRESS *)curr_addr->next)
  {
    TRC_DETAIL((TB, "Adding address %p", curr_addr));
    rc = aib_write_xml_address(trans, element_name, curr_addr);
  }

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_address_list */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_usage_detail_list                                */
/*                                                                           */
/* Purpose:   Write a list of usage details as XML                           */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     use_list     - list of usages to be written             */
/*                                                                           */
/* Operation: Walk the list of details, writing each one as XML              */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_usage_detail_list(TRANS_INFO *trans,
                                    OSP_USAGE_DETAIL *use_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  OSP_USAGE_DETAIL *curr_ud;

  TRC_ENTRY("aib_write_xml_usage_detail_list");

  /***************************************************************************/
  /* Walk the list, writing entry each in turn                               */
  /***************************************************************************/
  for (curr_ud = (OSP_USAGE_DETAIL *)use_list;
       (curr_ud != NULL) && rc;
       curr_ud = (OSP_USAGE_DETAIL *)curr_ud->next)
  {
    TRC_DETAIL((TB, "Adding usage detail %p", curr_ud));
    rc = aib_write_xml_usage_detail(trans, curr_ud);
  }

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_usage_detail_list */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_auth_url_list                                    */
/*                                                                           */
/* Purpose:   Write a list of auth urls as XML                               */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     url_list     - list of urls to be written               */
/*                                                                           */
/* Operation: Walk the list of urls, writing each one as XML                 */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_auth_url_list(TRANS_INFO *trans, OSP_SERVICE_URL *url_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  OSP_SERVICE_URL *curr_url;

  TRC_ENTRY("aib_write_xml_auth_url_list");

  /***************************************************************************/
  /* Walk the list, writing entry each in turn                               */
  /***************************************************************************/
  for (curr_url = (OSP_SERVICE_URL *)url_list;
       (curr_url != NULL) && rc;
       curr_url = (OSP_SERVICE_URL *)curr_url->next)
  {
    TRC_DETAIL((TB, "Adding url %p", curr_url));
    rc = aib_write_xml_element(trans, com_authority_url, curr_url->url);
  }

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_auth_url_list */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_service_list                                     */
/*                                                                           */
/* Purpose:   Write a list of OSP services as XML                            */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     service_list - list of services to be written           */
/*                                                                           */
/* Operation: Walk the list, writing each one as XML                         */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_service_list(TRANS_INFO *trans, OSP_SERVICE *service_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  OSP_SERVICE *curr_service;

  TRC_ENTRY("aib_write_xml_service_list");

  /***************************************************************************/
  /* Walk the list, writing entry each in turn                               */
  /***************************************************************************/
  for (curr_service = (OSP_SERVICE *)service_list;
       (curr_service != NULL) && rc;
       curr_service = (OSP_SERVICE *)curr_service->next)
  {
    TRC_DETAIL((TB, "Adding service %p", curr_service));
    rc = aib_write_xml_service(trans, curr_service);
  }

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_service_list */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_cert_chain                                       */
/*                                                                           */
/* Purpose:   Write a chain of certificates as XML                           */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     cert_chain   - chain to be written                      */
/*                                                                           */
/* Operation: Walk the chain, writing each one as XML                        */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_cert_chain(TRANS_INFO *trans, OSP_CERTIFICATE *cert_chain)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  OSP_CERTIFICATE *curr_cert;

  TRC_ENTRY("aib_write_xml_cert_chain");

  /***************************************************************************/
  /* If there's no entries, there's nothing to do                            */
  /***************************************************************************/
  if (cert_chain == NULL)
  {
    TRC_DETAIL((TB, "No certificates to write"));
  }
  else
  {
    /*************************************************************************/
    /* Write the prolog                                                      */
    /*************************************************************************/
    TRC_DETAIL((TB, "Starting cert chain"));

    if (!aib_space_available(trans,
                             1 + 1 + 2 + 3 +
                             strlen(com_cert_chain) + strlen(com_critical) +
                             strlen(com_false)))
    {
      TRC_ERROR((TB, "No room for certificate chain start tag"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt, 1);
    AIB_ADD_NULL(trans, com_cert_chain);
    AIB_ADD_DATA(trans, com_sp, 1);
    AIB_ADD_NULL(trans, com_critical);
    AIB_ADD_DATA(trans, com_eq_qu, 2);
    AIB_ADD_NULL(trans, com_false);
    AIB_ADD_DATA(trans, com_qu_gt_nl, 3);

    /*************************************************************************/
    /* Walk the list, writing entry each in turn                             */
    /*************************************************************************/
    for (curr_cert = (OSP_CERTIFICATE *)cert_chain;
         curr_cert != NULL;
         curr_cert = (OSP_CERTIFICATE *)curr_cert->next)
    {
      TRC_DETAIL((TB, "Adding cert %p", curr_cert));
      if (!aib_write_encoded_element(trans,
                                     com_cert,
                                     curr_cert->encoding,
                                     curr_cert->cert_len,
                                     curr_cert->cert_data))
      {
        TRC_ERROR((TB, "No room for certificates"));
        goto EXIT_LABEL;
      }
    }

    /*************************************************************************/
    /* Write the end of the list.                                            */
    /*************************************************************************/
    if (!aib_space_available(trans, 2 + 2 + strlen(com_cert_chain)))
    {
      TRC_ERROR((TB, "No room for cert_chain close tag"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt_sl, 2);
    AIB_ADD_NULL(trans, com_cert_chain);
    AIB_ADD_DATA(trans, com_gt_nl, 2);
  }

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_cert_chain */
