
/**MOD+***********************************************************************/
/* Module:    aibwrxml.c                                                     */
/*                                                                           */
/* Purpose:   AIB Common functions to write OSP types as XML                 */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.19               $ $Modtime::   Jan 11 2001 15:56:26   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <aibincl.h>

/**PROC+**********************************************************************/
/* Name:      aib_write_encoded_element                                      */
/*                                                                           */
/* Purpose:   Write an XML element with encoded data                         */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     name         - element name                             */
/*            IN     encoding     - data encoding                            */
/*            IN     len          - encoded data length                      */
/*            IN     data         - encoded data                             */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_encoded_element(TRANS_INFO *trans,
                              const char *name,
                              const char *encoding,
                              int len,
                              const char *data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_encoded_element");

  /***************************************************************************/
  /* Build the open and close tags, checking the encoding - if its CDATA we  */
  /* need to add the wrapper                                                 */
  /***************************************************************************/
  TRC_NORMAL((TB, "Adding XML element with encoding %s", encoding));
  if (!strcmp(encoding, com_cdata))
  {
    if (!aib_space_available(trans,
                             1 + 1 + 2 + 3 + 3 + 2 +
                             (2 * strlen(name)) + strlen(com_encoding) +
                             strlen(com_cdata) + len +
                             strlen(com_cdata_start) + strlen(com_cdata_end)))
    {
      TRC_ERROR((TB, "No room to write XML token"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt, 1);
    AIB_ADD_NULL(trans, name);
    AIB_ADD_DATA(trans, com_sp, 1);
    AIB_ADD_NULL(trans, com_encoding);
    AIB_ADD_DATA(trans, com_eq_qu, 2);
    AIB_ADD_NULL(trans, com_cdata);
    AIB_ADD_DATA(trans, com_qu_gt_nl, 3);
    AIB_ADD_NULL(trans, com_cdata_start);

    AIB_ADD_DATA(trans, data, len);

    AIB_ADD_NULL(trans, com_cdata_end);
    AIB_ADD_DATA(trans, com_nl_lt_sl, 3);
    AIB_ADD_NULL(trans, name);
    AIB_ADD_DATA(trans, com_gt_nl, 2);
  }
  else
  {
    if (!aib_space_available(trans,
                             1 + 1 + 2 + 3 + 3 + 2 +
                             (2 * strlen(name)) + strlen(com_encoding) +
                             strlen(encoding) + len))
    {
      TRC_ERROR((TB, "No room to write XML token"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt, 1);
    AIB_ADD_NULL(trans, name);
    AIB_ADD_DATA(trans, com_sp, 1);
    AIB_ADD_NULL(trans, com_encoding);
    AIB_ADD_DATA(trans, com_eq_qu, 2);
    AIB_ADD_NULL(trans, encoding);
    AIB_ADD_DATA(trans, com_qu_gt_nl, 3);

    AIB_ADD_DATA(trans, data, len);

    AIB_ADD_DATA(trans, com_nl_lt_sl, 3);
    AIB_ADD_NULL(trans, name);
    AIB_ADD_DATA(trans, com_gt_nl, 2);
  }

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_encoded_element */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_destination                                      */
/*                                                                           */
/* Purpose:   Write a destination as XML                                     */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     dest         - destination to be written                */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_destination(TRANS_INFO *trans, OSP_DESTINATION *dest)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_destination");

  /***************************************************************************/
  /* Add the open tag                                                        */
  /***************************************************************************/
  if (!aib_space_available(trans, 1 + 2 + strlen(com_destination)))
  {
    TRC_ERROR((TB, "No room for destination start tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_destination);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  /***************************************************************************/
  /* Write any destination information                                       */
  /***************************************************************************/
  if (dest->info != NULL)
  {
    /*************************************************************************/
    /* Build the open and close tags then add these and the address          */
    /*************************************************************************/
    TRC_NORMAL((TB, "Adding dest info"));
    if (!aib_write_xml_address(trans, com_dest_info, dest->info))
    {
      TRC_ERROR((TB, "Failure appending dest info"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Write any destination alternates                                        */
  /***************************************************************************/
  if (!aib_write_xml_address_list(trans, com_dest_alt, dest->alt_list))
  {
    TRC_ERROR((TB, "Failure adding dest alternates"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the signalling address - there MUST be one                          */
  /***************************************************************************/
  TRC_ASSERT((dest->signal_address != NULL), (TB,
                              "Got destination with NULL signalling address"));
  if (!aib_write_xml_element(trans, com_dest_sig_addr, dest->signal_address))
  {
    TRC_ERROR((TB, "Failure appending dest sig address"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add any tokens                                                          */
  /***************************************************************************/
  if (!aib_write_xml_token_list(trans, (char *)com_token, dest->token_list))
  {
    TRC_ERROR((TB, "Failure adding dest tokens"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add any validity limits - these are both optional                       */
  /***************************************************************************/
  if (dest->valid_after != NULL)
  {
    TRC_DETAIL((TB, "Adding valid after"));
    if (!aib_write_xml_element(trans, com_valid_after, dest->valid_after))
    {
      TRC_ERROR((TB, "Failure appending valid after"));
      goto EXIT_LABEL;
    }
  }

  if (dest->valid_until != NULL)
  {
    TRC_DETAIL((TB, "Adding valid until"));
    if (!aib_write_xml_element(trans, com_valid_until, dest->valid_until))
    {
      TRC_ERROR((TB, "Failure appending valid until"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Add the remaining lists - usage details and authority URLs              */
  /***************************************************************************/
  if (!aib_write_xml_usage_detail_list(trans, dest->usage_list))
  {
    TRC_ERROR((TB, "Failure adding usage details"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_auth_url_list(trans, dest->auth_url_list))
  {
    TRC_ERROR((TB, "Failure adding authority URLs"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the call id.  There MUST be one                                     */
  /***************************************************************************/
  TRC_ASSERT((dest->call_id != NULL), (TB,
                                       "Got destination with NULL call id"));

  if (!aib_write_encoded_element(trans,
                                 com_call_id,
                                 dest->call_id->encoding,
                                 strlen(dest->call_id->data),
                                 dest->call_id->data))
  {
    TRC_ERROR((TB, "Failure appending call id"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Finally the close tag                                                   */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_destination)))
  {
    TRC_ERROR((TB, "No room for destination close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_destination);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_destination */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_address                                          */
/*                                                                           */
/* Purpose:   Write an OSP_ADDRESS structure as XML                          */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     element_name - name to use for the elements             */
/*            IN     addr_list    - address to be written                    */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_address(TRANS_INFO *trans,
                          const char *element_name,
                          OSP_ADDRESS *addr)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_address");

  if (!aib_space_available(trans,
                           1 + 1 + 2 + 3 + 3 + 2 +
                           (2 * strlen(element_name)) + strlen(com_type) +
                           strlen(addr->type) + strlen(addr->address)))
  {
    TRC_ERROR((TB, "No room to write XML address"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, element_name);
  AIB_ADD_DATA(trans, com_sp, 1);
  AIB_ADD_NULL(trans, com_type);
  AIB_ADD_DATA(trans, com_eq_qu, 2);
  AIB_ADD_NULL(trans, addr->type);
  AIB_ADD_DATA(trans, com_qu_gt_nl, 3);

  AIB_ADD_NULL(trans, addr->address);

  AIB_ADD_DATA(trans, com_nl_lt_sl, 3);
  AIB_ADD_NULL(trans, element_name);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_address */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_usage_detail                                     */
/*                                                                           */
/* Purpose:   Write a usage detail as XML                                    */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     use          - usage detail to be written               */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_usage_detail(TRANS_INFO *trans, OSP_USAGE_DETAIL *use)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_usage_detail");

  /***************************************************************************/
  /* Add the open tag                                                        */
  /***************************************************************************/
  TRC_NORMAL((TB, "Adding usage detail %p", use));
  if (!aib_space_available(trans, 1 + 2 + strlen(com_usage)))
  {
    TRC_ERROR((TB, "No room for usage start tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_usage);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  /***************************************************************************/
  /* Add the service information                                             */
  /***************************************************************************/
  if (!aib_write_xml_service_info(trans, use->service))
  {
    TRC_ERROR((TB, "Failure appending usage detail service info"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the amount - there MUST be one                                      */
  /***************************************************************************/
  TRC_ASSERT((use->amount != NULL), (TB, "Got usage with NULL amount"));
  if (!aib_write_xml_element(trans, com_amount, use->amount))
  {
    TRC_ERROR((TB, "Failure appending usage amount"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the increment - there MUST be one                                   */
  /***************************************************************************/
  TRC_ASSERT((use->increment != NULL), (TB, "Got usage with NULL increment"));
  if (!aib_write_xml_element(trans, com_increment, use->increment))
  {
    TRC_ERROR((TB, "Failure appending usage increment"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the unit - there MUST be one                                        */
  /***************************************************************************/
  TRC_ASSERT((use->unit != NULL), (TB, "Got usage with NULL unit"));
  if (!aib_write_xml_element(trans, com_unit, use->unit))
  {
    TRC_ERROR((TB, "Failure appending usage unit"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Start and end time are optional                                         */
  /***************************************************************************/
  if (use->start_time != NULL)
  {
    TRC_DETAIL((TB, "Adding start time"));
    if (!aib_write_xml_element(trans, com_start_time, use->start_time))
    {
      TRC_ERROR((TB, "Failure appending start time"));
      goto EXIT_LABEL;
    }
  }

  if (use->end_time != NULL)
  {
    TRC_DETAIL((TB, "Adding end time"));
    if (!aib_write_xml_element(trans, com_end_time, use->end_time))
    {
      TRC_ERROR((TB, "Failure appending end time"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* The termination cause is optional                                       */
  /***************************************************************************/
  if (use->term_cause != NULL)
  {
    TRC_DETAIL((TB, "Adding term cause"));
    if (!aib_write_xml_term_cause(trans, use->term_cause))
    {
      TRC_ERROR((TB, "Failure appending term cause"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* We don't support writing statistics since we only *write* a usage       */
  /* detail as part of a (re)authorization                                   */
  /***************************************************************************/
  TRC_ASSERT((use->statistics == NULL),
                                (TB, "Writing usage statistics no supported"));

  /***************************************************************************/
  /* Finally the close tag                                                   */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_usage)))
  {
    TRC_ERROR((TB, "No room for usage close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_usage);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_usage_detail */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_timestamp                                        */
/*                                                                           */
/* Purpose:   Generates and writes a timestamp as XML                        */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_timestamp(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  time_t curr_time;
  char time_string[AIB_TIME_LEN];
  struct tm gm_time;
  int ret_count;

  TRC_ENTRY("aib_write_xml_timestamp");

  /***************************************************************************/
  /* Get the current time.                                                   */
  /***************************************************************************/
  time(&curr_time);

  /***************************************************************************/
  /* If this is not the same as the cached time, we have to rebuild the      */
  /* timestamp string.                                                       */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->thread->wa_locked.mutex_wa);

  if (curr_time == trans->thread->wa_locked.cached_time)
  {
    TRC_DETAIL((TB, "Copy cached timestamp string"));
    strcpy(time_string, trans->thread->wa_locked.cached_time_string);
  }
#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Since we're timing transactions, avoid including time for the expensive */
  /* time string construction, which under heavy load is only needed once a  */
  /* second, not once per transaction.                                       */
  /***************************************************************************/
  else
  {
    strcpy(time_string, "0000-00-00T00:00:00Z");
  }
#else
  else
  {
    OSP_RELEASE_LOCK(&trans->thread->wa_locked.mutex_wa);

    TRC_DETAIL((TB, "Build timestamp string"));
    gmtime_r(&curr_time, &gm_time);
    ret_count = strftime(time_string, AIB_TIME_LEN, "%Y-%m-%dT%H:%M:%SZ",
                                                                     &gm_time);
    TRC_ASSERT((ret_count != 0), (TB, "Formatted time didn't fit in array"));

    OSP_GET_LOCK(&trans->thread->wa_locked.mutex_wa);
    trans->thread->wa_locked.cached_time = curr_time;
    strcpy(trans->thread->wa_locked.cached_time_string, time_string);
  }
#endif

  OSP_RELEASE_LOCK(&trans->thread->wa_locked.mutex_wa);

  /***************************************************************************/
  /* Get and add the timestamp                                               */
  /***************************************************************************/
  TRC_NORMAL((TB, "Adding timestamp at %s", time_string));
  if (!aib_space_available(trans,
                           1 + 2 + 3 + 2 +
                           (2 * strlen(com_timestamp)) + strlen(time_string)))
  {
    TRC_ERROR((TB, "No room for timestamp element"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_timestamp);
  AIB_ADD_DATA(trans, com_gt_nl, 2);
  AIB_ADD_NULL(trans, time_string);
  AIB_ADD_DATA(trans, com_nl_lt_sl, 3);
  AIB_ADD_NULL(trans, com_timestamp);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_timestamp */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_status                                           */
/*                                                                           */
/* Purpose:   Write a status code as XML                                     */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     status       - status to be written                     */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_status(TRANS_INFO *trans, OSP_STATUS *status)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_status");

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Writing status element start tag"));
  if (!aib_space_available(trans, 1 + 2 + strlen(com_status)))
  {
    TRC_ERROR((TB, "No room for status start tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_status);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  /***************************************************************************/
  /* The code is required                                                    */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding term code %d", status->code));
  if (!aib_write_xml_element_numeric(trans, com_code, status->code))
  {
    TRC_ERROR((TB, "No room for termination code"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* The description is optional                                             */
  /***************************************************************************/
  if (status->desc != NULL)
  {
    TRC_DETAIL((TB, "Adding term desc '%s'", status->desc));
    if (!aib_write_xml_element(trans, com_term_desc, status->desc))
    {
      TRC_ERROR((TB, "No room for termination description"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Add the close tag                                                       */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_status)))
  {
    TRC_ERROR((TB, "No room for status close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_status);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_status */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_service_info                                     */
/*                                                                           */
/* Purpose:   Write a service info structure as XML                          */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     service_info - service info to be written               */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_service_info(TRANS_INFO *trans,
                               OSP_SERVICE_INFO *service_info)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  OSP_PROTOCOL_TYPE *prot_type;

  TRC_ENTRY("aib_write_xml_service_info");

  /***************************************************************************/
  /* The service info may be empty, meaning basic internet telephony.  In    */
  /* this case, the supplied pointer may be NULL, or the contained pointers  */
  /* may all be NULL                                                         */
  /***************************************************************************/
  if ((service_info == NULL) ||
      ((service_info->bandwidth == NULL) &&
       (service_info->service_type == NULL) &&
       (service_info->prot_type == NULL)))
  {
    if (!aib_space_available(trans, 1 + 3 + strlen(com_service)))
    {
      TRC_ERROR((TB, "No room for service element"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt, 1);
    AIB_ADD_NULL(trans, com_service);
    AIB_ADD_DATA(trans, com_sl_gt_nl, 3);
  }
  else
  {
    /*************************************************************************/
    /* Write the service as usual                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Adding non-empty service"));
    if (!aib_space_available(trans, 1 + 2 + strlen(com_service)))
    {
      TRC_ERROR((TB, "No room for service start tag"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt, 1);
    AIB_ADD_NULL(trans, com_service);
    AIB_ADD_DATA(trans, com_gt_nl, 2);

    /*************************************************************************/
    /* Insert the bandwidth                                                  */
    /*************************************************************************/
    if (!aib_write_xml_element(trans, com_bandwidth, service_info->bandwidth))
    {
      TRC_ERROR((TB, "Failure appending service bandwidth"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Insert the service type                                               */
    /*************************************************************************/
    if (!aib_write_xml_element(trans, com_service_type,
                                                  service_info->service_type))
    {
      TRC_ERROR((TB, "Failure appending service type"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Add any protocol types                                                */
    /*************************************************************************/
    for (prot_type = service_info->prot_type;
         prot_type != NULL;
         prot_type = prot_type->next)
    {
      if (!aib_write_xml_element(trans, com_protocol_type,
                                                         prot_type->prot_name))
      {
        TRC_ERROR((TB, "Failure appending service type"));
        goto EXIT_LABEL;
      }
    }

    /*************************************************************************/
    /* Add the close tag                                                     */
    /*************************************************************************/
    if (!aib_space_available(trans, 2 + 2 + strlen(com_service)))
    {
      TRC_ERROR((TB, "No room for service close tag"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt_sl, 2);
    AIB_ADD_NULL(trans, com_service);
    AIB_ADD_DATA(trans, com_gt_nl, 2);
  }

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_service_info */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_element                                          */
/*                                                                           */
/* Purpose:   Write the element and its data as XML                          */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     name         - element name                             */
/*            IN     data         - data to be written                       */
/*                                                                           */
/* Operation: Writes a simple element, ie one with no attributes, as xml     */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_element(TRANS_INFO *trans,
                          const char *name,
                          const char *data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_element");

  /***************************************************************************/
  /* Build the xml, using the <element/> format if data is NULL              */
  /***************************************************************************/
  if (data != NULL)
  {
    TRC_DETAIL((TB, "Adding element '%s', data '%s'", name, data));
    if (!aib_space_available(trans,
                             1 + 2 + 3 + 2 +
                             (2 * strlen(name)) + strlen(data)))
    {
      TRC_ERROR((TB, "No room for XML element"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt, 1);
    AIB_ADD_NULL(trans, name);
    AIB_ADD_DATA(trans, com_gt_nl, 2);
    AIB_ADD_NULL(trans, data);
    AIB_ADD_DATA(trans, com_nl_lt_sl, 3);
    AIB_ADD_NULL(trans, name);
    AIB_ADD_DATA(trans, com_gt_nl, 2);
  }
  else
  {
    TRC_DETAIL((TB, "Adding empty element '%s'", name));
    if (!aib_space_available(trans, 1 + 3 + strlen(name)))
    {
      TRC_ERROR((TB, "No room for XML element"));
      goto EXIT_LABEL;
    }

    AIB_ADD_DATA(trans, com_lt, 1);
    AIB_ADD_NULL(trans, name);
    AIB_ADD_DATA(trans, com_sl_gt_nl, 3);
  }

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_element */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_element_numeric                                  */
/*                                                                           */
/* Purpose:   Write the element and its data as XML                          */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     name         - element name                             */
/*            IN     data         - data to be written                       */
/*                                                                           */
/* Operation: Writes a simple element, ie one with no attributes, as xml.    */
/*            Note that data is an int - use aib_write_xml_element when      */
/*            data is a string                                               */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_element_numeric(TRANS_INFO *trans,
                                  const char *name,
                                  int data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  char xml[256];
  int xml_len;

  TRC_ENTRY("aib_write_xml_element_numeric");

  /***************************************************************************/
  /* Build the xml                                                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding element %s, data %d", name, data));
  xml_len = OSP_SNPRINTF((xml, sizeof(xml), "%d", data));
  TRC_ASSERT((xml_len < sizeof(xml)),
                (TB, "Resulting xml would be too large at %d bytes", xml_len));

  rc = aib_write_xml_element(trans, name, xml);

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_element_numeric */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_term_cause                                       */
/*                                                                           */
/* Purpose:   Write a termination cause structure as XML                     */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     term_cause   - termination cause structure to be        */
/*                                  written                                  */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_term_cause(TRANS_INFO *trans, OSP_TERM_CAUSE *term_cause)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_term_cause");

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Writing termination cause"));
  if (!aib_space_available(trans, 1 + 2 + strlen(com_term_cause)))
  {
    TRC_ERROR((TB, "No room for term_cause start tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_term_cause);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  /***************************************************************************/
  /* Termination code is optional                                            */
  /***************************************************************************/
  if (term_cause->term_code != NULL)
  {
    TRC_DETAIL((TB, "Adding term code %d", term_cause->term_code));
    if (!aib_write_xml_element(trans, com_term_code, term_cause->term_code))
    {
      TRC_ERROR((TB, "Failure appending term code"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Termination description is also optional                                */
  /***************************************************************************/
  if (term_cause->term_desc != NULL)
  {
    TRC_DETAIL((TB, "Adding term desc %s", term_cause->term_desc));
    if (!aib_write_xml_element(trans, com_term_desc, term_cause->term_desc))
    {
      TRC_ERROR((TB, "Failure appending term desc"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Add the close tag                                                       */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_term_cause)))
  {
    TRC_ERROR((TB, "No room for term_cause close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_term_cause);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_term_cause */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_service                                          */
/*                                                                           */
/* Purpose:   Write an OSP service structure as XML                          */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     service      - service to be written                    */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_service(TRANS_INFO *trans, OSP_SERVICE *service)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  OSP_SERVICE_URL *curr_url;

  TRC_ENTRY("aib_write_xml_service");

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding service %p", service));
  if (!aib_space_available(trans, 1 + 2 + strlen(com_osp_service)))
  {
    TRC_ERROR((TB, "No room for service start tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_osp_service);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  /***************************************************************************/
  /* Insert the elements                                                     */
  /***************************************************************************/
  if (!aib_write_xml_element(trans, com_osp_capability, service->capability))
  {
    TRC_ERROR((TB, "Failure appending capability"));
    goto EXIT_LABEL;
  }

  curr_url = service->url_list;
  while (curr_url != NULL)
  {
    TRC_DETAIL((TB, "Adding url %s", curr_url->url));
    if (!aib_write_xml_element(trans, com_osp_service_url, curr_url->url))
    {
      TRC_ERROR((TB, "Failure appending service url"));
      goto EXIT_LABEL;
    }
    curr_url = (OSP_SERVICE_URL *)curr_url->next;
  }

  if (!aib_write_xml_element(trans, com_osp_sig_req,
                                 service->sig_required ? com_true : com_false))
  {
    TRC_ERROR((TB, "Failure appending service bandwidth"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the close tag                                                       */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_osp_service)))
  {
    TRC_ERROR((TB, "No room for service close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_osp_service);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_service */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_credit_amount                                    */
/*                                                                           */
/* Purpose:   Write a creditamount as XML                                    */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     credit       - credit amount to be written              */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_credit_amount(TRANS_INFO *trans, OSP_CREDIT_AMOUNT *credit)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_credit_amount");

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Writing credit amount element start tag"));
  if (!aib_space_available(trans, 1 + 2 + strlen(com_credit_amount)))
  {
    TRC_ERROR((TB, "No room for credit amount start tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_credit_amount);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  /***************************************************************************/
  /* Both currency and amount are required                                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding currency %d", credit->currency));
  if (!aib_write_xml_element(trans, com_currency, credit->currency))
  {
    TRC_ERROR((TB, "No room for currency"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Adding amount %d", credit->amount));
  if (!aib_write_xml_element(trans, com_amount, credit->amount))
  {
    TRC_ERROR((TB, "No room for amount"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the close tag                                                       */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_credit_amount)))
  {
    TRC_ERROR((TB, "No room for credit amount close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_credit_amount);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_credit_amount */

/**PROC+**********************************************************************/
/* Name:      aib_write_xml_credit_time                                      */
/*                                                                           */
/* Purpose:   Write a credit time as XML                                     */
/*                                                                           */
/* Returns:   TRUE if xml written successfully, FALSE otherwise              */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*            IN     credit       - credit time to be written                */
/*                                                                           */
/**PROC-**********************************************************************/

int aib_write_xml_credit_time(TRANS_INFO *trans, OSP_CREDIT_TIME *credit)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("aib_write_xml_credit_time");

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Writing credit time element start tag"));
  if (!aib_space_available(trans, 1 + 2 + strlen(com_credit_time)))
  {
    TRC_ERROR((TB, "No room for credit time start tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_credit_time);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  /***************************************************************************/
  /* All elements are required                                             */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding amount %d", credit->amount));
  if (!aib_write_xml_element(trans, com_amount, credit->amount))
  {
    TRC_ERROR((TB, "No room for amount"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Adding increment %d", credit->increment));
  if (!aib_write_xml_element(trans, com_increment, credit->increment))
  {
    TRC_ERROR((TB, "No room for increment"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Adding unit %d", credit->unit));
  if (!aib_write_xml_element(trans, com_unit, credit->unit))
  {
    TRC_ERROR((TB, "No room for unit"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the close tag                                                       */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_credit_time)))
  {
    TRC_ERROR((TB, "No room for credit_time close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_credit_time);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* aib_write_xml_credit_time */

