
/**MOD+***********************************************************************/
/* Module:    carauthi.c                                                     */
/*                                                                           */
/* Purpose:   Call Authorization and Routing (CAR) Application Interface     */
/*            Layer (AIL) AuthorizationIndication and Confirmation handling  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.9                $ $Modtime::   Jan 11 2001 15:56:26   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <carincl.h>

/**PROC+**********************************************************************/
/* Name:      osp_ar_auth_confirm                                            */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_ar_auth_confirm(OSP_CORRELATOR corr,
                                  OSP_STATUS * status,
                                  char *valid_after,
                                  char *valid_until,
                                  unsigned int sig_required)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_INT_CORRELATOR *int_corr;
  TRANS_INFO *trans;

  TRC_API_ENTRY("osp_ar_auth_confirm");

  /***************************************************************************/
  /* Check we're intialized                                                  */
  /***************************************************************************/
  if (!car_initialized)
  {
    TRC_ERROR((TB, "Auth Conf received when not initialized"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get at the correlator data                                              */
  /***************************************************************************/
  int_corr = (OSP_INT_CORRELATOR *)corr;
  if (int_corr == NULL)
  {
    TRC_ERROR((TB, "Null correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  trans = int_corr->trans;
  if (trans == NULL)
  {
    TRC_ERROR((TB, "Corrupted correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction start build time.                                   */
  /***************************************************************************/
  SM_GETTIMES(trans->timings[SM_TIMINGS_START_BUILD]);
#endif

  /***************************************************************************/
  /* We're about to start using the send buffer - lock everyone else out     */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->mutex);

  /***************************************************************************/
  /* If signing is required, set the flag on the transaction - the whole     */
  /* message is signed, not just this component                              */
  /***************************************************************************/
  if (sig_required)
  {
    TRC_NORMAL((TB, "Signing required on trans %p", trans));
    trans->sign_response = TRUE;
  }

  /***************************************************************************/
  /* Set up the XML for the response, using the supplied parameters and the  */
  /* component id from the correlator                                        */
  /***************************************************************************/
  if (!car_build_auth_confirm(trans, int_corr->comp_type, int_corr->comp_id,
                                      status, valid_after, valid_until))
  {
    TRC_ERROR((TB, "Failed to build xml for Auth Confirm"));
    trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
  }

  /***************************************************************************/
  /* We've finished - update the response count and see if the response      */
  /* needs to be sent.  Note that xml_maybe_send will free the transaction   */
  /* mutex for us.                                                           */
  /***************************************************************************/
  trans->response_count++;
  TRC_NORMAL((TB, "Transaction %p resp count now %d", trans,
                                                       trans->response_count));
  xml_maybe_send(trans);

EXIT_LABEL:
  /***************************************************************************/
  /* Free the correlator component id and the correlator itself              */
  /***************************************************************************/
  rm_release_mem(int_corr->comp_id);
  rm_release_mem(int_corr);

  TRC_API_EXIT();

  return(rc);

} /* osp_ar_auth_confirm */

/**PROC+**********************************************************************/
/* Name:      car_auth_ind_end_elem_handler                                  */
/*                                                                           */
/* Purpose:   End element handler called for the child elements of an        */
/*            authorization indication, & for the end of the auth ind itself */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name         - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

void car_auth_ind_end_elem_handler(void *user_data, const XML_Char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;
  CAR_PARSED_AI *parsed_ai;
  int missing_element;

  TRC_ENTRY("car_auth_ind_end_elem_handler");

  /***************************************************************************/
  /* get the transaction info, trans, by casting the user data               */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_ASSERT((trans != NULL), (TB, "NULL user data in AuthInd end handler"));

  /***************************************************************************/
  /* Do common AIB end tag handling.                                         */
  /***************************************************************************/
  aib_end_element_handler(trans, name);

  /***************************************************************************/
  /* The rest only needs doing if we've reached the end of the auth          */
  /* indication component...                                                 */
  /***************************************************************************/
  if ((!strcmp(name, com_auth_ind)) || (!strcmp(name, com_auth_s_ind)))
  {
    /*************************************************************************/
    /* Reset the xml layer's own handlers                                    */
    /*************************************************************************/
    xml_set_default_handlers(trans);

    /*************************************************************************/
    /* Get the parsed data memory                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got end of %s", name));
    parsed_ai = (CAR_PARSED_AI *)trans->aib_state;
    if (parsed_ai == NULL)
    {
      /***********************************************************************/
      /* there's nothing more we can do                                      */
      /***********************************************************************/
      TRC_ALERT((TB, "End element but no parsed data memory"));
      goto EXIT_LABEL;
    }

    if (parsed_ai->common.status.code != OSP_STATUS_SUCCESS)
    {
      TRC_ALERT((TB, "Unsuccessful parsing status: %d",
                                      parsed_ai->common.status.code));

      car_build_error(trans, parsed_ai->common.corr->comp_type);
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Is the appropriate callback registered?                               */
    /*************************************************************************/
    osp_get_recursive_lock(&car_cb_mutex);
    if (car_auth_ind_cb == NULL)
    {
      osp_release_recursive_lock(&car_cb_mutex);
      TRC_NORMAL((TB, "End auth ind but no callback - done"));
      /***********************************************************************/
      /* add an error response for the transaction                           */
      /***********************************************************************/
      parsed_ai->common.status.code = OSP_STATUS_SERVICE_NOT_AVAILABLE;
      parsed_ai->common.status.desc = car_service_not_available;
      car_build_error(trans, parsed_ai->common.corr->comp_type);
      goto EXIT_LABEL;
    }

#ifdef OSP_DATA_COVERAGE
    aib_scan_registered_structures();
#endif

    /*************************************************************************/
    /* Now we can pass the auth indication on to the AR app                  */
    /*************************************************************************/
    (*car_auth_ind_cb)((OSP_CORRELATOR)(parsed_ai->common.corr),
                       parsed_ai->timestamp,
                       parsed_ai->role,
                       parsed_ai->call_id,
                       parsed_ai->source_info,
                       (OSP_ADDRESS *)SL_HEAD(parsed_ai->source_alt_list),
                       parsed_ai->dest_info,
                       (OSP_ADDRESS *)SL_HEAD(parsed_ai->dest_alt_list),
                       parsed_ai->service,
                       (OSP_TOKEN *)SL_HEAD(parsed_ai->token_list),
                       &trans->client_id);

    osp_release_recursive_lock(&car_cb_mutex);
  }
  else
  {
    /*************************************************************************/
    /* We're processing the end of one of our child elements.  Set the       */
    /* parsed data pointer to NULL so that it's not freed                    */
    /*************************************************************************/
    parsed_ai = NULL;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Free up the memory from any parsed data                                 */
  /***************************************************************************/
  if (parsed_ai != NULL)
  {
    car_free_ai_mem(trans);
  }

  TRC_EXIT();

  return;

} /* car_auth_ind_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      car_build_auth_confirm                                         */
/*                                                                           */
/* Purpose:   Writes an authorization confirmation as XML                    */
/*                                                                           */
/* Returns:   TRUE if write succeeds, FALSE otherwise                        */
/*                                                                           */
/* Params:    IN     trans        - transaction associated with the response */
/*            IN     comp_type    - type, used to choose correct spelling of */
/*                                  the response                             */
/*            IN     comp_id      - the component id for the response        */
/*            IN     status       - OSP status information                   */
/*            IN     valid_after  - time authorization valid from            */
/*            IN     valid_until  - time authorization valid until           */
/*                                                                           */
/**PROC-**********************************************************************/

int car_build_auth_confirm(TRANS_INFO *trans,
                           int comp_type,
                           char *comp_id,
                           OSP_STATUS *status,
                           char *valid_after,
                           char *valid_until)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  const char *comp_name;

  TRC_ENTRY("car_build_auth_confirm");

  /***************************************************************************/
  /* Ensure we got a transaction to write to                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Supplied NULL transaction"));

  /***************************************************************************/
  /* Select the correct spelling of the component name (some older clients   */
  /* may use the International English spelling, AuthorisationConfirmation   */
  /***************************************************************************/
  if (comp_type == AIB_AUTH_IND)
  {
    TRC_DETAIL((TB, "Use US spelling"));
    comp_name = com_auth_cnf;
  }
  else
  {
    TRC_DETAIL((TB, "Use International spelling"));
    comp_name = com_auth_s_cnf;
  }

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  if (!aib_space_available(trans,
                           1 + 1 + 2 + 3 +
                           strlen(comp_name) + strlen(com_comp_id) +
                           strlen(comp_id)))
  {
    TRC_ERROR((TB, "No room for XML component"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, comp_name);
  AIB_ADD_DATA(trans, com_sp, 1);
  AIB_ADD_NULL(trans, com_comp_id);
  AIB_ADD_DATA(trans, com_eq_qu, 2);
  AIB_ADD_NULL(trans, comp_id);
  AIB_ADD_DATA(trans, com_qu_gt_nl, 3);

  /***************************************************************************/
  /* Add each element in turn                                                */
  /***************************************************************************/
  if (!aib_write_xml_timestamp(trans))
  {
    TRC_ERROR((TB, "Failure appending timestamp"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_status(trans, status))
  {
    TRC_ERROR((TB, "Failure appending status"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_element(trans, com_valid_after, valid_after))
  {
    TRC_ERROR((TB, "Failure appending valid after"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_element(trans, com_valid_until, valid_until))
  {
    TRC_ERROR((TB, "Failure appending valid until"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Append the close tag                                                    */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(comp_name)))
  {
    TRC_ERROR((TB, "No room for component close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, comp_name);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* car_build_auth_confirm */

/**PROC+**********************************************************************/
/* Name:      car_free_ai_mem                                                */
/*                                                                           */
/* Purpose:   Frees parsed auth indication and associated memory             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*                                                                           */
/* Operation: Get the auth indication structure from trans.                  */
/*            Free all the memory within the structure, including walking    */
/*            the lists                                                      */
/*                                                                           */
/**PROC-**********************************************************************/

void car_free_ai_mem(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CAR_PARSED_AI *parsed_ai;
  OSP_CALL_ID *call_item;
  OSP_ADDRESS *alt_item;
  OSP_TOKEN *tok_item;

  TRC_ENTRY("car_free_ai_mem");

  /***************************************************************************/
  /* Get the parsed data structure we're to free                             */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Passed NULL transaction info"));
  parsed_ai = trans->aib_state;
  TRC_ASSERT((parsed_ai != NULL), (TB, "Asked to free NULL parsed ai mem"));
  TRC_DETAIL((TB, "Freeing ai data %p", parsed_ai));

  /***************************************************************************/
  /* Free all the easy ones                                                  */
  /***************************************************************************/
  aib_free_call_id(parsed_ai->call_id);

  TRC_DETAIL((TB, "Free source info"));
  aib_free_address(parsed_ai->source_info);

  TRC_DETAIL((TB, "Free dest info"));
  aib_free_address(parsed_ai->dest_info);

  TRC_DETAIL((TB, "Free service"));
  aib_free_service(parsed_ai->service);

  /***************************************************************************/
  /* Now deal with the lists                                                 */
  /***************************************************************************/
  while ((alt_item = (OSP_ADDRESS *)SL_HEAD(parsed_ai->source_alt_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_ai->source_alt_list);
    TRC_DETAIL((TB, "Free source alt %p", alt_item));
    aib_free_address(alt_item);
  }

  while ((alt_item = (OSP_ADDRESS *)SL_HEAD(parsed_ai->dest_alt_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_ai->dest_alt_list);
    TRC_DETAIL((TB, "Free dest alt %p", alt_item));
    aib_free_address(alt_item);
  }

  while ((tok_item = (OSP_TOKEN *)SL_HEAD(parsed_ai->token_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_ai->token_list);
    TRC_DETAIL((TB, "Free token %p", tok_item));
    aib_free_token(tok_item);
  }

  TRC_DETAIL((TB, "Free final structure %p", parsed_ai));
  rm_release_mem(parsed_ai);

  /***************************************************************************/
  /* Having released the memory, NULL the pointer to stop anyone trying to   */
  /* reuse it                                                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freed ai data successfully"));
  trans->aib_state = NULL;

  TRC_EXIT();

  return;

} /* car_free_ai_mem */

