
/**MOD+***********************************************************************/
/* Module:    carauthr.c                                                     */
/*                                                                           */
/* Purpose:   Call Authorization and Routing (CAR) Application Interface     */
/*            Layer (AIL) AuthorizationRequest and Response handling         */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.25               $ $Modtime::   Jan 11 2001 15:56:26   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <carincl.h>

/**PROC+**********************************************************************/
/* Name:      osp_ar_auth_response                                           */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_ar_auth_response(OSP_CORRELATOR corr,
                                   OSP_STATUS *status,
                                   char *trans_id,
                                   OSP_TOKEN *token_list,
                                   OSP_DESTINATION *dest_list,
                                   unsigned int sig_required)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_INT_CORRELATOR *int_corr;
  TRANS_INFO *trans;

  TRC_API_ENTRY("osp_ar_auth_response");

  /***************************************************************************/
  /* Check we're intialized                                                  */
  /***************************************************************************/
  if (!car_initialized)
  {
    TRC_ERROR((TB, "Auth Rsp received when not initialized"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get at the correlator data                                              */
  /***************************************************************************/
  int_corr = (OSP_INT_CORRELATOR *)corr;
  if (int_corr == NULL)
  {
    TRC_ERROR((TB, "Null correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }
  trans = int_corr->trans;
  if (trans == NULL)
  {
    TRC_ERROR((TB, "Corrupted correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We must be given a transaction id                                       */
  /***************************************************************************/
  if (trans_id == NULL)
  {
    TRC_ERROR((TB, "Missing required transaction id."));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Status must be supplied                                                 */
  /***************************************************************************/
  if (status == NULL)
  {
    TRC_ERROR((TB, "Missing required status information."));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction start build time.                                   */
  /***************************************************************************/
  SM_GETTIMES(trans->timings[SM_TIMINGS_START_BUILD]);
#endif

  /***************************************************************************/
  /* We're about to start using the send buffer - lock everyone else out     */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->mutex);

  /***************************************************************************/
  /* If signing is required, set the flag on the transaction - the whole     */
  /* message is signed, not just this component                              */
  /***************************************************************************/
  if (sig_required)
  {
    TRC_NORMAL((TB, "Signing required on trans %p", trans));
    trans->sign_response = TRUE;
  }

  /***************************************************************************/
  /* Set up the XML for the response, using the supplied parameters and the  */
  /* component id from the correlator                                        */
  /***************************************************************************/
  if (!car_build_auth_response(trans, int_corr->comp_type, int_corr->comp_id,
                                      status, trans_id, token_list, dest_list))
  {
    TRC_ERROR((TB, "Failed to build xml for Auth Response"));
    trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
  }

  /***************************************************************************/
  /* We've finished - update the response count and see if the response      */
  /* needs to be sent.  Note that xml_maybe_send will free the transaction   */
  /* mutex for us.                                                           */
  /***************************************************************************/
  trans->response_count++;
  TRC_NORMAL((TB, "Transaction %p resp count now %d", trans,
                                                       trans->response_count));
  xml_maybe_send(trans);

EXIT_LABEL:
  /***************************************************************************/
  /* Free the correlator component id and the correlator itself              */
  /***************************************************************************/
  rm_release_mem(int_corr->comp_id);
  rm_release_mem(int_corr);

  TRC_API_EXIT();

  return(rc);

} /* osp_ar_auth_response */

/**PROC+**********************************************************************/
/* Name:      car_auth_rq_end_elem_handler                                   */
/*                                                                           */
/* Purpose:   End element handler called for the child elements of an        */
/*            authorization request, and for the end of the auth rq itself   */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name         - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

void car_auth_rq_end_elem_handler(void *user_data, const XML_Char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;
  CAR_PARSED_AR *parsed_ar;
  int missing_element;
  int max_dests;
  int num_call_ids;
  SLIST_ENTRY *curr_entry;

  TRC_ENTRY("car_auth_rq_end_elem_handler");

  /***************************************************************************/
  /* get the transaction info, trans, by casting the user data               */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_ASSERT((trans != NULL), (TB, "NULL user data in AuthRq end handler"));

  /***************************************************************************/
  /* Do common AIB end tag handling.                                         */
  /***************************************************************************/
  aib_end_element_handler(trans, name);

  /***************************************************************************/
  /* The rest only needs doing if we've reached the end of the auth request  */
  /* component...                                                            */
  /***************************************************************************/
  if ((!strcmp(name, com_auth_req)) || (!strcmp(name, com_auth_s_req)))
  {
    /*************************************************************************/
    /* Reset the xml layer's own handlers                                    */
    /*************************************************************************/
    xml_set_default_handlers(trans);

    /*************************************************************************/
    /* Get the parsed data memory                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got end of %s", name));
    parsed_ar = (CAR_PARSED_AR *)trans->aib_state;
    if (parsed_ar == NULL)
    {
      /***********************************************************************/
      /* there's nothing more we can do                                      */
      /***********************************************************************/
      TRC_ALERT((TB, "End element but no parsed data memory"));
      goto EXIT_LABEL;
    }

    if (parsed_ar->common.status.code != OSP_STATUS_SUCCESS)
    {
      TRC_ALERT((TB, "Unsuccessful parsing status: %d",
                                      parsed_ar->common.status.code));

      car_build_error(trans, parsed_ar->common.corr->comp_type);
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Check that if max dests is not 0, the client supplied either 1 or max */
    /* dests call ids.                                                       */
    /*************************************************************************/
    max_dests = atoi(parsed_ar->max_dests);
    if (max_dests != 0)
    {
      /***********************************************************************/
      /* Count the call ids supplied                                         */
      /***********************************************************************/
      TRC_DETAIL((TB, "Max dests %d; counting call ids", max_dests));
      num_call_ids = 0;
      for (curr_entry = SL_HEAD(parsed_ar->call_id_list);
           curr_entry != NULL;
           curr_entry = (SLIST_ENTRY *)curr_entry->next)
      {
        num_call_ids++;
      }

      /***********************************************************************/
      /* Check the call id count                                             */
      /***********************************************************************/
      TRC_DETAIL((TB, "Got %d call ids", num_call_ids));
      if ((num_call_ids != 1) && (num_call_ids != max_dests))
      {
        /*********************************************************************/
        /* If they don't tie up, send a failure response                     */
        /*********************************************************************/
        TRC_ERROR((TB, "Max dests %d does not match num call ids %d",
                                                     max_dests, num_call_ids));
        parsed_ar->common.status.code = OSP_STATUS_BAD_REQUEST;
        parsed_ar->common.status.desc = car_num_call_ids_wrong;
        car_build_error(trans, parsed_ar->common.corr->comp_type);
        goto EXIT_LABEL;
      }
    }
    /*************************************************************************/
    /* Is the appropriate callback registered?                               */
    /*************************************************************************/
    osp_get_recursive_lock(&car_cb_mutex);
    if (car_auth_rq_cb == NULL)
    {
      osp_release_recursive_lock(&car_cb_mutex);
      TRC_NORMAL((TB, "End auth rq but no callback - done"));
      /***********************************************************************/
      /* add an error response for the transaction                           */
      /***********************************************************************/
      parsed_ar->common.status.code = OSP_STATUS_SERVICE_NOT_AVAILABLE;
      parsed_ar->common.status.desc = car_service_not_available;
      car_build_error(trans, parsed_ar->common.corr->comp_type);
      goto EXIT_LABEL;
    }

#ifdef OSP_DATA_COVERAGE
    aib_scan_registered_structures();
#endif

    /*************************************************************************/
    /* We captured any supplied protocol types in an SLIST.  We need to      */
    /* point the OSP_SERVICE_INFO at the head of this list before calling    */
    /* across the api                                                        */
    /*************************************************************************/
    parsed_ar->service->osp_service.prot_type =
                                   SL_HEAD(parsed_ar->service->prot_type_list);

    /*************************************************************************/
    /* Now we can pass the auth request on to the AR app                     */
    /*************************************************************************/
    (*car_auth_rq_cb)((OSP_CORRELATOR)(parsed_ar->common.corr),
                      parsed_ar->timestamp,
                      (OSP_CALL_ID *)SL_HEAD(parsed_ar->call_id_list),
                      parsed_ar->source_info,
                      (OSP_ADDRESS *)SL_HEAD(parsed_ar->source_alt_list),
                      parsed_ar->dest_info,
                      (OSP_ADDRESS *)SL_HEAD(parsed_ar->dest_alt_list),
                      &(parsed_ar->service->osp_service),
                      parsed_ar->max_dests,
                      (OSP_TOKEN *)SL_HEAD(parsed_ar->token_list),
                      (OSP_AUTH_INFO *)SL_HEAD(parsed_ar->subs_auth_list),
                      &trans->client_id);

    osp_release_recursive_lock(&car_cb_mutex);
  }
  else
  {
    /*************************************************************************/
    /* We're processing the end of one of our child elements.  Set the       */
    /* parsed data pointer to NULL so that it's not freed                    */
    /*************************************************************************/
    parsed_ar = NULL;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Free up the memory from any parsed data                                 */
  /***************************************************************************/
  if (parsed_ar != NULL)
  {
    car_free_ar_mem(trans);
  }

  TRC_EXIT();

  return;

} /* car_auth_rq_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      car_free_ar_mem                                                */
/*                                                                           */
/* Purpose:   Frees parsed auth request and associated memory                */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*                                                                           */
/* Operation: Get the auth request data structure from trans.                */
/*            Free all the memory within the structure, including walking    */
/*            the lists                                                      */
/*                                                                           */
/**PROC-**********************************************************************/

void car_free_ar_mem(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CAR_PARSED_AR *parsed_ar;
  OSP_CALL_ID *call_item;
  OSP_ADDRESS *alt_item;
  OSP_TOKEN *tok_item;
  OSP_AUTH_INFO *auth_item;

  TRC_ENTRY("car_free_ar_mem");

  /***************************************************************************/
  /* Get the parsed data structure we're to free                             */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Passed NULL transaction info"));
  parsed_ar = trans->aib_state;
  TRC_ASSERT((parsed_ar != NULL), (TB, "Asked to free NULL parsed ar mem"));
  TRC_DETAIL((TB, "Freeing ar data %p", parsed_ar));

  /***************************************************************************/
  /* Free all the easy ones                                                  */
  /***************************************************************************/
  TRC_DETAIL((TB, "Free source info"));
  aib_free_address(parsed_ar->source_info);

  TRC_DETAIL((TB, "Free dest info"));
  aib_free_address(parsed_ar->dest_info);

  TRC_DETAIL((TB, "Free service"));
  if (parsed_ar->service != NULL)
  {
    parsed_ar->service->osp_service.prot_type =
                                   SL_HEAD(parsed_ar->service->prot_type_list);
    aib_free_service(&(parsed_ar->service->osp_service));
  }

  /***************************************************************************/
  /* Now deal with the lists                                                 */
  /***************************************************************************/
  while ((call_item = (OSP_CALL_ID *)SL_HEAD(parsed_ar->call_id_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_ar->call_id_list);
    TRC_DETAIL((TB, "Free call id %p", call_item));
    aib_free_call_id(call_item);
  }

  while ((alt_item = (OSP_ADDRESS *)SL_HEAD(parsed_ar->source_alt_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_ar->source_alt_list);
    TRC_DETAIL((TB, "Free source alt %p", alt_item));
    aib_free_address(alt_item);
  }

  while ((alt_item = (OSP_ADDRESS *)SL_HEAD(parsed_ar->dest_alt_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_ar->dest_alt_list);
    TRC_DETAIL((TB, "Free dest alt %p", alt_item));
    aib_free_address(alt_item);
  }

  while ((tok_item = (OSP_TOKEN *)SL_HEAD(parsed_ar->token_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_ar->token_list);
    TRC_DETAIL((TB, "Free token %p", tok_item));
    aib_free_token(tok_item);
  }

  while ((auth_item = (OSP_AUTH_INFO *)SL_HEAD(parsed_ar->subs_auth_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_ar->subs_auth_list);
    TRC_DETAIL((TB, "Free subs auth %p", auth_item));
    rm_release_mem(auth_item->encoding);
    rm_release_mem(auth_item);
  }

  TRC_DETAIL((TB, "Free final structure %p", parsed_ar));
  rm_release_mem(parsed_ar);

  /***************************************************************************/
  /* Having released the memory, NULL the pointer to stop anyone trying to   */
  /* reuse it                                                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freed ar data successfully"));
  trans->aib_state = NULL;

  TRC_EXIT();

  return;

} /* car_free_ar_mem */

