
/**MOD+***********************************************************************/
/* Module:    carcapsi.c                                                     */
/*                                                                           */
/* Purpose:   Call Authorization and Routing (CAR) Application Interface     */
/*            Layer (AIL) CapabilitiesIndication and Confirmation handling   */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.12               $ $Modtime::   Jan 11 2001 15:56:28   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <carincl.h>

/**PROC+**********************************************************************/
/* Name:      osp_ar_caps_confirm                                            */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_ar_caps_confirm(OSP_CORRELATOR corr,
                                  OSP_STATUS * status,
                                  char *osp_version,
                                  OSP_SERVICE *service_list,
                                  OSP_CERTIFICATE *cert_list,
                                  char *device_id,
                                  unsigned int sig_required)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_INT_CORRELATOR *int_corr;
  TRANS_INFO *trans;

  TRC_API_ENTRY("osp_ar_caps_confirm");

  /***************************************************************************/
  /* Check we're intialized                                                  */
  /***************************************************************************/
  if (!car_initialized)
  {
    TRC_ERROR((TB, "Caps Conf received when not initialized"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get at the correlator data                                              */
  /***************************************************************************/
  int_corr = (OSP_INT_CORRELATOR *)corr;
  if (int_corr == NULL)
  {
    TRC_ERROR((TB, "Null correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  trans = int_corr->trans;
  if (trans == NULL)
  {
    TRC_ERROR((TB, "Corrupted correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Status must be supplied                                                 */
  /***************************************************************************/
  if (status == NULL)
  {
    TRC_ERROR((TB, "Missing required status information."));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction start build time.                                   */
  /***************************************************************************/
  SM_GETTIMES(trans->timings[SM_TIMINGS_START_BUILD]);
#endif

  /***************************************************************************/
  /* We're about to start using the send buffer - lock everyone else out     */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->mutex);

  /***************************************************************************/
  /* If signing is required, set the flag on the transaction - the whole     */
  /* message is signed, not just this component                              */
  /***************************************************************************/
  if (sig_required)
  {
    TRC_NORMAL((TB, "Signing required on trans %p", trans));
    trans->sign_response = TRUE;
  }

  /***************************************************************************/
  /* Set up the XML for the response, using the supplied parameters and the  */
  /* component id from the correlator                                        */
  /***************************************************************************/
  if (!car_build_caps_confirm(trans, int_corr->comp_id, status, osp_version,
                                           service_list, cert_list, device_id))
  {
    TRC_ERROR((TB, "Failed to build xml for Caps Confirm"));
    trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
  }

  /***************************************************************************/
  /* We've finished - update the response count and see if the response      */
  /* needs to be sent.  Note that xml_maybe_send will free the transaction   */
  /* mutex for us.                                                           */
  /***************************************************************************/
  trans->response_count++;
  TRC_NORMAL((TB, "Transaction %p resp count now %d", trans,
                                                       trans->response_count));
  xml_maybe_send(trans);

EXIT_LABEL:
  /***************************************************************************/
  /* Free the correlator component id and the correlator itself              */
  /***************************************************************************/
  rm_release_mem(int_corr->comp_id);
  rm_release_mem(int_corr);

  TRC_API_EXIT();

  return(rc);

} /* osp_ar_caps_confirm */

/**PROC+**********************************************************************/
/* Name:      car_caps_ind_end_elem_handler                                  */
/*                                                                           */
/* Purpose:   End element handler called for the child elements of a         */
/*            capabilities indication and for the end of the caps ind itself */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name         - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

void car_caps_ind_end_elem_handler(void *user_data, const XML_Char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;
  CAR_PARSED_CI *parsed_ci;
  OSP_RESOURCE *osp_res_list;
  int missing_element;

  TRC_ENTRY("car_caps_ind_end_elem_handler");

  /***************************************************************************/
  /* Initialization                                                          */
  /***************************************************************************/
  osp_res_list = NULL;
  parsed_ci = NULL;

  /***************************************************************************/
  /* get the transaction info, trans, by casting the user data               */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_ASSERT((trans != NULL), (TB, "NULL user data in AuthInd end handler"));

  /***************************************************************************/
  /* Do common AIB end tag handling.                                         */
  /***************************************************************************/
  aib_end_element_handler(trans, name);

  /***************************************************************************/
  /* The rest only needs doing if we've reached the end of the caps          */
  /* indication component...                                                 */
  /***************************************************************************/
  if (!strcmp(name, com_caps_ind))
  {
    /*************************************************************************/
    /* Reset the xml layer's own handlers                                    */
    /*************************************************************************/
    xml_set_default_handlers(trans);

    /*************************************************************************/
    /* Get the parsed data memory                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got end of %s", name));
    parsed_ci = (CAR_PARSED_CI *)trans->aib_state;
    if (parsed_ci == NULL)
    {
      /***********************************************************************/
      /* there's nothing more we can do                                      */
      /***********************************************************************/
      TRC_ALERT((TB, "End element but no parsed data memory"));
      goto EXIT_LABEL;
    }

    if (parsed_ci->common.status.code != OSP_STATUS_SUCCESS)
    {
      TRC_ALERT((TB, "Unsuccessful parsing status: %d",
                                      parsed_ci->common.status.code));

      car_build_error(trans, parsed_ci->common.corr->comp_type);
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Is the appropriate callback registered?                               */
    /*************************************************************************/
    osp_get_recursive_lock(&car_cb_mutex);
    if (car_caps_ind_cb == NULL)
    {
      osp_release_recursive_lock(&car_cb_mutex);
      TRC_NORMAL((TB, "End caps ind but no callback - done"));
      /***********************************************************************/
      /* add an error response for the transaction                           */
      /***********************************************************************/
      parsed_ci->common.status.code = OSP_STATUS_SERVICE_NOT_AVAILABLE;
      parsed_ci->common.status.desc = car_service_not_available;
      car_build_error(trans, parsed_ci->common.corr->comp_type);
      goto EXIT_LABEL;
    }

#ifdef OSP_DATA_COVERAGE
    aib_scan_registered_structures();
#endif

    /*************************************************************************/
    /* When we parsed the resources we used an SLIST structure.  We now      */
    /* need to put the information into the OSP form to pass across the api  */
    /* lists off each resource) into the OSP_ form                           */
    /*************************************************************************/
    car_convert_resources(parsed_ci);

    /*************************************************************************/
    /* Now we can pass the caps indication on to the AR app                  */
    /*************************************************************************/
    (*car_caps_ind_cb)((OSP_CORRELATOR)(parsed_ci->common.corr),
                       (OSP_DEVICE_INFO *)SL_HEAD(parsed_ci->dev_info_list),
                       parsed_ci->version,
                       (OSP_CAPS *)SL_HEAD(parsed_ci->caps_list),
                       (OSP_RESOURCE *)SL_HEAD(parsed_ci->resource_list),
                       &trans->client_id);

    osp_release_recursive_lock(&car_cb_mutex);
  }
  else
  {
    /*************************************************************************/
    /* We're processing the end of one of our child elements.  Set the       */
    /* parsed data pointer to NULL so that it's not freed                    */
    /*************************************************************************/
    parsed_ci = NULL;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Free up the memory from any parsed data                                 */
  /***************************************************************************/
  if (parsed_ci != NULL)
  {
    car_free_ci_mem(trans);
  }

  TRC_EXIT();

  return;

} /* car_caps_ind_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      car_build_caps_confirm                                         */
/*                                                                           */
/* Purpose:   Writes an capsorization confirmation as XML                    */
/*                                                                           */
/* Returns:   TRUE if write succeeds, FALSE otherwise                        */
/*                                                                           */
/* Params:    IN     trans        - transaction associated with the response */
/*            IN     comp_id      - the component id for the response        */
/*            IN     status       - OSP status information                   */
/*            IN     osp_version  - negotiated OSP version to use            */
/*            IN     service_list - OSP services supported by the server     */
/*            IN     cert_list    - public key certificates                  */
/*            IN     device_id    - server device identifier                 */
/*                                                                           */
/**PROC-**********************************************************************/

int car_build_caps_confirm(TRANS_INFO *trans,
                           char *comp_id,
                           OSP_STATUS *status,
                           char *osp_version,
                           OSP_SERVICE *service_list,
                           OSP_CERTIFICATE *cert_list,
                           char *device_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("car_build_caps_confirm");

  /***************************************************************************/
  /* Ensure we got a transaction to write to                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Supplied NULL transaction"));

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  if (!aib_space_available(trans,
                           1 + 1 + 2 + 3 +
                           strlen(com_caps_cnf) + strlen(com_comp_id) +
                           strlen(comp_id)))
  {
    TRC_ERROR((TB, "No room for XML component"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_caps_cnf);
  AIB_ADD_DATA(trans, com_sp, 1);
  AIB_ADD_NULL(trans, com_comp_id);
  AIB_ADD_DATA(trans, com_eq_qu, 2);
  AIB_ADD_NULL(trans, comp_id);
  AIB_ADD_DATA(trans, com_qu_gt_nl, 3);

  /***************************************************************************/
  /* Add each element in turn                                                */
  /***************************************************************************/
  if (!aib_write_xml_timestamp(trans))
  {
    TRC_ERROR((TB, "Failure appending timestamp"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_status(trans, status))
  {
    TRC_ERROR((TB, "Failure appending status"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_element(trans, com_osp_version, osp_version))
  {
    TRC_ERROR((TB, "Failure appending osp version"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_service_list(trans, service_list))
  {
    TRC_ERROR((TB, "Failure appending service list"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_cert_chain(trans, cert_list))
  {
    TRC_ERROR((TB, "Failure appending service list"));
    goto EXIT_LABEL;
  }

  if (device_id != NULL)
  {
    if (!aib_write_xml_element(trans, com_dev_id, device_id))
    {
      TRC_ERROR((TB, "Failure appending device id"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Append the close tag                                                    */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_caps_cnf)))
  {
    TRC_ERROR((TB, "No room for component close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_caps_cnf);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* car_build_caps_confirm */

/**PROC+**********************************************************************/
/* Name:      car_free_ci_mem                                                */
/*                                                                           */
/* Purpose:   Frees parsed caps indication and associated memory             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*                                                                           */
/* Operation: Get the caps indication structure from trans.                  */
/*            Free all the memory within the structure, including walking    */
/*            the lists                                                      */
/*                                                                           */
/**PROC-**********************************************************************/

void car_free_ci_mem(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CAR_PARSED_CI *parsed_ci;
  OSP_DEVICE_INFO *dev_item;
  OSP_CAPS *cap_item;
  AIB_RESOURCE *res_item;
  AIB_SUPP_PROTOCOL *prot_item;
  OSP_DATA_RATE *rate_item;

  TRC_ENTRY("car_free_ci_mem");

  /***************************************************************************/
  /* Get the parsed data structure we're to free                             */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Passed NULL transaction info"));
  parsed_ci = trans->aib_state;
  TRC_ASSERT((parsed_ci != NULL), (TB, "Asked to free NULL parsed ci mem"));
  TRC_DETAIL((TB, "Freeing ai data %p", parsed_ci));

  /***************************************************************************/
  /* Now deal with the lists                                                 */
  /***************************************************************************/
  while ((dev_item = (OSP_DEVICE_INFO *)SL_HEAD(parsed_ci->dev_info_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_ci->dev_info_list);
    TRC_DETAIL((TB, "Free dev info %p", dev_item));
    rm_release_mem(dev_item->type);
    rm_release_mem(dev_item);
  }

  while ((cap_item = (OSP_CAPS *)SL_HEAD(parsed_ci->caps_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_ci->caps_list);
    TRC_DETAIL((TB, "Free cap info %p", cap_item));
    rm_release_mem(cap_item);
  }

  while ((res_item = (AIB_RESOURCE *)SL_HEAD(parsed_ci->resource_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_ci->resource_list);
    TRC_DETAIL((TB, "Free res info %p", res_item));

    if (res_item->resource.data_rate != NULL)
    {
      TRC_ASSERT((res_item->resource.data_rate->next == NULL),
                                                  (TB, "Multiple data rates"));
    }
    rm_release_mem(res_item->resource.data_rate);

    while ((prot_item = (AIB_SUPP_PROTOCOL *)
                                SL_HEAD(res_item->supp_protocol_list)) != NULL)
    {
      SL_REMOVE_HEAD(res_item->supp_protocol_list);
      TRC_DETAIL((TB, "Free supp protocol %p", prot_item));
      rm_release_mem(prot_item->supp_prot.prot_type);

      while ((rate_item = (OSP_DATA_RATE *)
                                   SL_HEAD(prot_item->data_rate_list)) != NULL)
      {
        SL_REMOVE_HEAD(prot_item->data_rate_list);
        TRC_DETAIL((TB, "Free data rate %p", rate_item));
        rm_release_mem(rate_item);
      }
      rm_release_mem(prot_item);
    }
    rm_release_mem(res_item);
  }

  TRC_DETAIL((TB, "Free final structure %p", parsed_ci));
  rm_release_mem(parsed_ci);

  /***************************************************************************/
  /* Having released the memory, NULL the pointer to stop anyone trying to   */
  /* reuse it                                                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freed ai data successfully"));
  trans->aib_state = NULL;

  TRC_EXIT();

  return;

} /* car_free_ci_mem */

/**PROC+**********************************************************************/
/* Name:      car_convert_resources                                          */
/*                                                                           */
/* Purpose:   Convert an internal AIB_RESOURCE list to an external           */
/*            OSP_RESOURCE list                                              */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     aib_res_list - list to convert inplace                  */
/*                                                                           */
/* Operation: Each AIB_RESOURCE in the list contains an OSP_RESOURCE.        */
/*            For each OSP_RESOURCE, set its supported protocol list to      */
/*            point to the head of the SLIST version used in parsing         */
/*                                                                           */
/**PROC-**********************************************************************/

void car_convert_resources(CAR_PARSED_CI *parsed_ci)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  AIB_SUPP_PROTOCOL *prot_item;
  AIB_RESOURCE *res_item;

  TRC_ENTRY("car_convert_resources");

  /***************************************************************************/
  /* Walk the resource list                                                  */
  /***************************************************************************/
  TRC_DETAIL((TB, "Walking list"));
  res_item = (AIB_RESOURCE *)SL_HEAD(parsed_ci->resource_list);
  while (res_item != NULL)
  {
    TRC_DETAIL((TB, "Converting res item %p", res_item));

    /*************************************************************************/
    /* Set up the OSP_RESOURCE of the supported protocols list               */
    /*************************************************************************/
    res_item->resource.supp_protocol_list =
                    (OSP_SUPP_PROTOCOL *)SL_HEAD(res_item->supp_protocol_list);

    /*************************************************************************/
    /* Each entry in the protocol list may contain a list of data rates,     */
    /* which also needs converting                                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Converting data rates..."));
    prot_item = (AIB_SUPP_PROTOCOL *)SL_HEAD(res_item->supp_protocol_list);
    while ((prot_item != NULL) && (prot_item->supp_prot.prot_type != NULL))
    {
      TRC_DETAIL((TB, "Convert data rates for '%s'",
                                   prot_item->supp_prot.prot_type->prot_name));
      prot_item->supp_prot.data_rate_list =
                           (OSP_DATA_RATE *)SL_HEAD(prot_item->data_rate_list);
      prot_item = (AIB_SUPP_PROTOCOL *)prot_item->supp_prot.next;
    }

    /*************************************************************************/
    /* Move onto the next resource                                           */
    /*************************************************************************/
    res_item = (AIB_RESOURCE *)res_item->resource.next;
  }

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* car_convert_resources */
