
/**MOD+***********************************************************************/
/* Module:    carmain.c                                                      */
/*                                                                           */
/* Purpose:   Call Authorization and Routing (CAR) Application Interface     */
/*            Layer (AIL) external interfaces                                */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.20               $ $Modtime::   Jan 11 2001 15:56:28   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <carincl.h>

/**PROC+**********************************************************************/
/* Name:      car_init                                                       */
/*                                                                           */
/* Purpose:   To initialize the CAR component.                               */
/*                                                                           */
/* Returns:   OSP_SUCCESS          - CAR was initialized successfully        */
/*            OSP_ERR_ALREADY_INIT - CAR has already been initialized        */
/*            other                - error from osp_create_mutex()           */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC car_init(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("car_init");

  /***************************************************************************/
  /* Check that we haven't already been initialized                          */
  /***************************************************************************/
  if (car_initialized)
  {
    TRC_ERROR((TB, "CAR initialized again"));
    rc = OSP_ERR_ALREADY_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create the mutex used to protect accesses to the callbacks.  This is    */
  /* recursive to allow a callback function to deregister itself.            */
  /***************************************************************************/
  rc = osp_create_recursive_mutex(&car_cb_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create CAR cb mutex"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialization complete.                                                */
  /***************************************************************************/
  car_initialized = TRUE;
  TRC_NORMAL((TB, "CAR initialized successfully"));

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* car_init */

/**PROC+**********************************************************************/
/* Name:      car_term                                                       */
/*                                                                           */
/* Purpose:   Terminates the CAR components                                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - CAR was terminated successfully             */
/*            OSP_ERR_NOT_INIT - CAR has not been initialized                */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC car_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("car_term");

  /***************************************************************************/
  /* check we're initialized                                                 */
  /***************************************************************************/
  if (!car_initialized)
  {
    TRC_ERROR((TB, "CAR Terminate called when not initialized!"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Destroy the callback mutex                                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "destroy cb mutex"));
  osp_destroy_recursive_mutex(&car_cb_mutex);

  /***************************************************************************/
  /* Termaination complete.                                                  */
  /***************************************************************************/
  car_initialized = FALSE;
  TRC_NORMAL((TB, "CAR terminated successfully"));

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* car_term */

/**PROC+**********************************************************************/
/* Name:      osp_ar_register                                                */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_ar_register(POSP_AR_AUTH_RQ_CALLBACK auth_rq_cb,
                              POSP_AR_AUTH_IND_CALLBACK auth_ind_cb,
                              POSP_AR_REAUTH_RQ_CALLBACK reauth_rq_cb,
                              POSP_AR_PRICING_IND_CALLBACK pricing_ind_cb,
                              POSP_AR_CAPS_IND_CALLBACK caps_ind_cb)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;

  TRC_API_ENTRY("osp_ar_register");

  /***************************************************************************/
  /* check we're initialized                                                 */
  /***************************************************************************/
  if (car_initialized)
  {

    /*************************************************************************/
    /* Grab the callback mutex and store the callbacks                       */
    /*************************************************************************/
    osp_get_recursive_lock(&car_cb_mutex);

    car_auth_rq_cb = auth_rq_cb;
    car_auth_ind_cb = auth_ind_cb;
    car_reauth_rq_cb = reauth_rq_cb;
    car_pricing_ind_cb = pricing_ind_cb;
    car_caps_ind_cb = caps_ind_cb;

    osp_release_recursive_lock(&car_cb_mutex);

    rc = OSP_SUCCESS;
  }
  else
  {
    TRC_ERROR((TB, "Register called when CAR not initialized!"));
    rc = OSP_ERR_NOT_INIT;
  }

  TRC_API_EXIT();

  return(rc);

} /* osp_ar_register */

/**PROC+**********************************************************************/
/* Name:      osp_sa_register                                                */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_sa_register(POSP_SA_CALLBACK sa_cb)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("osp_sa_register");

  /***************************************************************************/
  /* check we're initialized                                                 */
  /***************************************************************************/
  if (car_initialized)
  {
    /*************************************************************************/
    /* Grab the callback mutex and store the callbacks                       */
    /*************************************************************************/
    osp_get_recursive_lock(&car_cb_mutex);
    car_subs_auth_rq_cb = sa_cb;
    osp_release_recursive_lock(&car_cb_mutex);

    rc = OSP_SUCCESS;
  }
  else
  {
    TRC_ERROR((TB, "Register called when CAR not initialized!"));
    rc = OSP_ERR_NOT_INIT;
  }

  TRC_EXIT();

  return(rc);

} /* osp_sa_register */

/**PROC+**********************************************************************/
/* Name:      car_start_processing                                           */
/*                                                                           */
/* Purpose:   Tells CAR AIB to allocate resources necessary for parsing the  */
/*            new client component.                                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS - completed successfully                           */
/*            other       - failure; the caller should send an http failure  */
/*                                                                           */
/* Params:    IN     trans        - pointer to the transaction's control     */
/*                                  structure                                */
/*            IN     parser       - XML parser instance                      */
/*            IN     comp_type    - the type of client component.  One of    */
/*                                  - AIB_AUTH_RQ                            */
/*                                  - AIB_AUTH_IND                           */
/*                                  - AIB_REAUTH_RQ                          */
/*                                  - AIB_PRICING_IND                        */
/*                                  - AIB_CAPS_IND                           */
/*            IN     comp_id      - value of the componentId attribute       */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC car_start_processing(TRANS_INFO *trans,
                            XML_Parser *parser,
                            int comp_type,
                            char *comp_id)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  AIB_DATA *parsed_common;
  CAR_PARSED_AR *parsed_ar;
  CAR_PARSED_AI *parsed_ai;
  CAR_PARSED_RR *parsed_rr;
  CAR_PARSED_PI *parsed_pi;
  CAR_PARSED_CI *parsed_ci;
  CAR_PARSED_SAR *parsed_sar;
  OSP_STATUS status;

  TRC_ENTRY("car_start_processing");

  /***************************************************************************/
  /* We must have been initialized                                           */
  /***************************************************************************/
  TRC_ASSERT((car_initialized), (TB,
                          "CAR start processing called when not initialized"));

  /***************************************************************************/
  /* Note that we setup the element handlers even if there's no application  */
  /* callback registered, since we still need to parse the whole component   */
  /* in order to find its end, and hence the start of the next component     */
  /***************************************************************************/
  trans->aib_state = NULL;
  switch (comp_type)
  {
    case AIB_AUTH_RQ:
    case AIB_AUTH_S_RQ:
      /***********************************************************************/
      /* Call common AIB start processing routine to do most of the work.    */
      /* If this succeeds trans->aib_state will be non-NULL; if it fails,    */
      /* the status parameter will have been set accordingly.                */
      /***********************************************************************/
      TRC_DETAIL((TB, "starting auth request"));
      aib_start_processing(trans,
                           parser,
                           comp_id,
                           sizeof(CAR_PARSED_AR),
                           car_authorization_request_elements,
                           car_auth_rq_end_elem_handler,
                           &status);

      if (trans->aib_state != NULL)
      {
        /*********************************************************************/
        /* Initialize the parsed data memory                                 */
        /*********************************************************************/
        parsed_ar = (CAR_PARSED_AR *)trans->aib_state;
        TRC_DETAIL((TB, "Initializing parsed data structure"));
        parsed_ar->timestamp = NULL;
        SL_INIT_ROOT(parsed_ar->call_id_list);
        parsed_ar->source_info = NULL;
        SL_INIT_ROOT(parsed_ar->source_alt_list);
        parsed_ar->dest_info = NULL;
        SL_INIT_ROOT(parsed_ar->dest_alt_list);
        parsed_ar->service = NULL;
        parsed_ar->max_dests = NULL;
        SL_INIT_ROOT(parsed_ar->token_list);
        SL_INIT_ROOT(parsed_ar->subs_auth_list);
      }
      break;

    case AIB_AUTH_IND:
    case AIB_AUTH_S_IND:
      /***********************************************************************/
      /* Call common AIB start processing routine to do most of the work.    */
      /* If this succeeds trans->aib_state will be non-NULL; if it fails,    */
      /* the status parameter will have been set accordingly.                */
      /***********************************************************************/
      TRC_DETAIL((TB, "starting auth indication"));
      aib_start_processing(trans,
                           parser,
                           comp_id,
                           sizeof(CAR_PARSED_AI),
                           car_authorization_indication_elements,
                           car_auth_ind_end_elem_handler,
                           &status);

      if (trans->aib_state != NULL)
      {
        /*********************************************************************/
        /* Make sure we got the memory                                       */
        /*********************************************************************/
        parsed_ai = (CAR_PARSED_AI *)trans->aib_state;

        /*********************************************************************/
        /* Initialize the parsed data memory                                 */
        /*********************************************************************/
        TRC_DETAIL((TB, "Initializing parsed data structure"));
        parsed_ai->timestamp = NULL;
        parsed_ai->role = NULL;
        parsed_ai->call_id = NULL;
        parsed_ai->source_info = NULL;
        SL_INIT_ROOT(parsed_ai->source_alt_list);
        parsed_ai->dest_info = NULL;
        SL_INIT_ROOT(parsed_ai->dest_alt_list);
        parsed_ai->service = NULL;
        SL_INIT_ROOT(parsed_ai->token_list);
      }
      break;

    case AIB_REAUTH_RQ:
    case AIB_REAUTH_S_RQ:
      /***********************************************************************/
      /* Call common AIB start processing routine to do most of the work.    */
      /* If this succeeds trans->aib_state will be non-NULL; if it fails,    */
      /* the status parameter will have been set accordingly.                */
      /***********************************************************************/
      TRC_DETAIL((TB, "starting reauth request"));
      aib_start_processing(trans,
                           parser,
                           comp_id,
                           sizeof(CAR_PARSED_RR),
                           car_reauthorization_request_elements,
                           car_reauth_rq_end_elem_handler,
                           &status);

      if (trans->aib_state != NULL)
      {
        /*********************************************************************/
        /* Make sure we got the memory                                       */
        /*********************************************************************/
        parsed_rr = (CAR_PARSED_RR *)trans->aib_state;

        /*********************************************************************/
        /* Initialize the parsed data memory                                 */
        /*********************************************************************/
        TRC_DETAIL((TB, "Initializing parsed data structure"));
        parsed_rr->timestamp = NULL;
        parsed_rr->role = NULL;
        parsed_rr->call_id = NULL;
        parsed_rr->source_info = NULL;
        SL_INIT_ROOT(parsed_rr->source_alt_list);
        parsed_rr->dest_info = NULL;
        SL_INIT_ROOT(parsed_rr->dest_alt_list);
        parsed_rr->transaction_id = NULL;
        SL_INIT_ROOT(parsed_rr->use_detail_list);
        SL_INIT_ROOT(parsed_rr->token_list);
      }
      break;

    case AIB_PRICING_IND:
      /***********************************************************************/
      /* Call common AIB start processing routine to do most of the work.    */
      /* If this succeeds trans->aib_state will be non-NULL; if it fails,    */
      /* the status parameter will have been set accordingly.                */
      /***********************************************************************/
      TRC_DETAIL((TB, "starting pricing ind"));
      aib_start_processing(trans,
                           parser,
                           comp_id,
                           sizeof(CAR_PARSED_PI),
                           car_pricing_indication_elements,
                           car_pricing_ind_end_elem_handler,
                           &status);

      if (trans->aib_state != NULL)
      {
        /*********************************************************************/
        /* Make sure we got the memory                                       */
        /*********************************************************************/
        parsed_pi = (CAR_PARSED_PI *)trans->aib_state;

        /*********************************************************************/
        /* Initialize the parsed data memory                                 */
        /*********************************************************************/
        TRC_DETAIL((TB, "Initializing parsed data structure"));
        parsed_pi->timestamp = NULL;
        parsed_pi->source_info = NULL;
        parsed_pi->dest_info = NULL;
        parsed_pi->currency = NULL;
        parsed_pi->amount = NULL;
        parsed_pi->increment = NULL;
        parsed_pi->unit = NULL;
        parsed_pi->service = NULL;
        parsed_pi->valid_after = NULL;
        parsed_pi->valid_until = NULL;
      }
      break;

    case AIB_CAPS_IND:
      /***********************************************************************/
      /* Call common AIB start processing routine to do most of the work.    */
      /* If this succeeds trans->aib_state will be non-NULL; if it fails,    */
      /* the status parameter will have been set accordingly.                */
      /***********************************************************************/
      TRC_DETAIL((TB, "starting caps ind"));
      aib_start_processing(trans,
                           parser,
                           comp_id,
                           sizeof(CAR_PARSED_CI),
                           car_caps_indication_elements,
                           car_caps_ind_end_elem_handler,
                           &status);

      if (trans->aib_state != NULL)
      {
        /*********************************************************************/
        /* Make sure we got the memory                                       */
        /*********************************************************************/
        parsed_ci = (CAR_PARSED_CI *)trans->aib_state;

        /*********************************************************************/
        /* Initialize the parsed data memory                                 */
        /*********************************************************************/
        TRC_DETAIL((TB, "Initializing parsed data structure"));
        SL_INIT_ROOT(parsed_ci->dev_info_list);
        parsed_ci->version = NULL;
        SL_INIT_ROOT(parsed_ci->caps_list);
        SL_INIT_ROOT(parsed_ci->resource_list);
      }
      break;

    case AIB_SUBS_AUTHEN_RQ:
      /***********************************************************************/
      /* Call common AIB start processing routine to do most of the work.    */
      /* If this succeeds trans->aib_state will be non-NULL; if it fails,    */
      /* the status parameter will have been set accordingly.                */
      /***********************************************************************/
      TRC_DETAIL((TB, "starting subs authen"));
      aib_start_processing(trans,
                           parser,
                           comp_id,
                           sizeof(CAR_PARSED_SAR),
                           car_subs_authen_request_elements,
                           car_sar_end_elem_handler,
                           &status);

      if (trans->aib_state != NULL)
      {
        /*********************************************************************/
        /* Make sure we got the memory                                       */
        /*********************************************************************/
        parsed_sar = (CAR_PARSED_SAR *)trans->aib_state;

        /*********************************************************************/
        /* Initialize the parsed data memory                                 */
        /*********************************************************************/
        TRC_DETAIL((TB, "Initializing parsed data structure"));
        parsed_sar->timestamp = NULL;
        parsed_sar->source_info = NULL;
        SL_INIT_ROOT(parsed_sar->source_alt_list);
        parsed_sar->dest_info = NULL;
        parsed_sar->service = NULL;
        parsed_sar->subs_authen_info = NULL;
      }
      break;

    default:
      TRC_ASSERT((FALSE), (TB, "unknown component type %d", comp_type));
      break;
  }

  /***************************************************************************/
  /* Do the remaining common initialization of the parsing data              */
  /***************************************************************************/
  if (trans->aib_state != NULL)
  {
    /*************************************************************************/
    /* Store the component type                                              */
    /*************************************************************************/
    TRC_DETAIL((TB, "Storing comp type %d", comp_type));
    parsed_common = (AIB_DATA *)trans->aib_state;
    parsed_common->corr->comp_type = comp_type;
  }
  else
  {
    /*************************************************************************/
    /* If we get here, then we don't have any parsing data in                */
    /* trans->aib_state, because we failed to allocate the memory for it or  */
    /* the component id                                                      */
    /*                                                                       */
    /* Either way, we must send an error response using the values returned  */
    /* in status, but we can't call car_build_error() - it uses              */
    /* trans->aib_state, which we haven't allocated!  Note that we still     */
    /* return a good response code to the xml layer as we have processed the */
    /* component.                                                            */
    /*                                                                       */
    /* As we're accessing the send buffer, we must hold the transaction      */
    /* mutex                                                                 */
    /*************************************************************************/
    OSP_GET_LOCK(&trans->mutex);
    trans->response_count++;
    TRC_NORMAL((TB, "Transaction %08lx resp count now %d", trans,
                                                       trans->response_count));

    /*************************************************************************/
    /* Build the appropriate OSP status response                             */
    /*************************************************************************/
    if ((comp_type == AIB_AUTH_RQ) || (comp_type == AIB_AUTH_S_RQ))
    {
      if (!car_build_auth_response(trans, comp_type, comp_id, &status, "0",
                                                                   NULL, NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Auth Rq"));
      }
    }
    else if ((comp_type == AIB_AUTH_IND) || (comp_type == AIB_AUTH_S_IND))
    {
      if (!car_build_auth_confirm(trans, comp_type, comp_id, &status, NULL,
                                                                         NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Auth Ind"));
      }
    }
    else if ((comp_type == AIB_REAUTH_RQ) || (comp_type == AIB_REAUTH_S_RQ))
    {
      if (!car_build_auth_response(trans, comp_type, comp_id, &status, "0",
                                                                   NULL, NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Reauth Rq"));
      }
    }
    else if (comp_type == AIB_PRICING_IND)
    {
      if (!car_build_price_confirm(trans, comp_id, &status))
      {
        TRC_ERROR((TB, "Failed to build error response for Pri Ind"));
      }
    }
    else if (comp_type == AIB_CAPS_IND)
    {
      if (!car_build_caps_confirm(trans, comp_id, &status, NULL, NULL,
                                                                   NULL, NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Caps Ind"));
      }
    }
    else if (comp_type == AIB_SUBS_AUTHEN_RQ)
    {
      if (!car_build_sa_response(trans, comp_id, &status, NULL, NULL, NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Subs Authen Rq"));
      }
    }
    else
    {
      TRC_ASSERT((FALSE), (TB, "unknown component type %d", comp_type));
    }
    OSP_RELEASE_LOCK(&trans->mutex);
  }

  TRC_EXIT();

  return(rc);

} /* car_start_processing */

/**PROC+**********************************************************************/
/* Name:      car_cancel_processing                                          */
/*                                                                           */
/* Purpose:   Called by the XML layer to inform the AIB that a parsing error */
/*            occurred in the component it was handling.                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - completed successfully                      */
/*            OSP_ERR_NOT_INIT - the AIB was not initialized                 */
/*                                                                           */
/* Params:    IN     trans        - the transaction's control structure      */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC car_cancel_processing(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  AIB_DATA *common_data;

  TRC_ENTRY("car_cancel_processing");

  /***************************************************************************/
  /* set default return code                                                 */
  /***************************************************************************/
  rc = OSP_SUCCESS;

  /***************************************************************************/
  /* Something went wrong in the parsing.  The XML layer will reset the      */
  /* handlers; we just need to free up any resources                         */
  /***************************************************************************/
  if (trans->aib_state)
  {
    /*************************************************************************/
    /* Get the common data from the transaction info so we know what it is   */
    /* we're freeing                                                         */
    /*************************************************************************/
    common_data = (AIB_DATA *)trans->aib_state;

    switch (common_data->corr->comp_type)
    {
      case AIB_AUTH_RQ:
      case AIB_AUTH_S_RQ:
      {
        TRC_DETAIL((TB, "Freeing auth rq data"));
        car_free_ar_mem(trans);
      }
      break;

      case AIB_AUTH_IND:
      case AIB_AUTH_S_IND:
      {
        TRC_DETAIL((TB, "Freeing auth ind data"));
        car_free_ai_mem(trans);
      }
      break;

      case AIB_REAUTH_RQ:
      case AIB_REAUTH_S_RQ:
      {
        TRC_DETAIL((TB, "Freeing reauth rq data"));
        car_free_rr_mem(trans);
      }
      break;

      case AIB_PRICING_IND:
      {
        TRC_DETAIL((TB, "Freeing pricing ind data"));
        car_free_pi_mem(trans);
      }
      break;

      case AIB_CAPS_IND:
      {
        TRC_DETAIL((TB, "Freeing caps ind data"));
        car_free_ci_mem(trans);
      }
      break;

      case AIB_SUBS_AUTHEN_RQ:
      {
        TRC_DETAIL((TB, "Freeing subs authen rq data"));
        car_free_sar_mem(trans);
      }
      break;

      default:
      {
        TRC_ASSERT((FALSE), (TB, "Asked to freeunknown component type %d",
                                            common_data->corr->comp_type));
      }
      break;
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* car_cancel_processing */

