
/**MOD+***********************************************************************/
/* Module:    carrearq.c                                                     */
/*                                                                           */
/* Purpose:   Call Authorization and Routing (CAR) Application Interface     */
/*            Layer (AIL) ReauthorizationRequest and Response handling       */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.9                $ $Modtime::   Jan 11 2001 15:56:28   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <carincl.h>

/**PROC+**********************************************************************/
/* Name:      osp_ar_reauth_response                                         */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_ar_reauth_response(OSP_CORRELATOR corr,
                                     OSP_STATUS * status,
                                     char *transaction_id,
                                     OSP_TOKEN *token_list,
                                     OSP_DESTINATION *dest_list,
                                     unsigned int sig_required)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_INT_CORRELATOR *int_corr;
  TRANS_INFO *trans;

  TRC_API_ENTRY("osp_ar_reauth_response");

  /***************************************************************************/
  /* Check we're intialized                                                  */
  /***************************************************************************/
  if (!car_initialized)
  {
    TRC_ERROR((TB, "Auth Rsp received when not initialized"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get at the correlator data                                              */
  /***************************************************************************/
  int_corr = (OSP_INT_CORRELATOR *)corr;
  if (int_corr == NULL)
  {
    TRC_ERROR((TB, "Null correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  trans = int_corr->trans;
  if (trans == NULL)
  {
    TRC_ERROR((TB, "Corrupted correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We must be given a transaction id                                       */
  /***************************************************************************/
  if (transaction_id == NULL)
  {
    TRC_ERROR((TB, "Missing required transaction id."));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Status must be supplied                                                 */
  /***************************************************************************/
  if (status == NULL)
  {
    TRC_ERROR((TB, "Missing required status information."));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction start build time.                                   */
  /***************************************************************************/
  SM_GETTIMES(trans->timings[SM_TIMINGS_START_BUILD]);
#endif

  /***************************************************************************/
  /* We're about to start using the send buffer - lock everyone else out     */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->mutex);

  /***************************************************************************/
  /* If signing is required, set the flag on the transaction - the whole     */
  /* message is signed, not just this component                              */
  /***************************************************************************/
  if (sig_required)
  {
    TRC_NORMAL((TB, "Signing required on trans %p", trans));
    trans->sign_response = TRUE;
  }

  /***************************************************************************/
  /* Set up the XML for the response, using the supplied parameters and the  */
  /* component id from the correlator                                        */
  /***************************************************************************/
  if (!car_build_auth_response(trans, int_corr->comp_type, int_corr->comp_id,
                                status, transaction_id, token_list, dest_list))
  {
    TRC_ERROR((TB, "Failed to build xml for Reauth Response"));
    trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
  }

  /***************************************************************************/
  /* We've finished - update the response count and see if the response      */
  /* needs to be sent.  Note that xml_maybe_send will free the transaction   */
  /* mutex for us.                                                           */
  /***************************************************************************/
  trans->response_count++;
  TRC_NORMAL((TB, "Transaction %p resp count now %d", trans,
                                                       trans->response_count));
  xml_maybe_send(trans);

EXIT_LABEL:
  /***************************************************************************/
  /* Free the correlator component id and the correlator itself              */
  /***************************************************************************/
  rm_release_mem(int_corr->comp_id);
  rm_release_mem(int_corr);

  TRC_API_EXIT();

  return(rc);

} /* osp_ar_reauth_response */

/**PROC+**********************************************************************/
/* Name:      car_reauth_rq_end_elem_handler                                 */
/*                                                                           */
/* Purpose:   End element handler called for the child elements of a         */
/*            reauthorization request, & for the end of the reauth rq itself */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name         - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

void car_reauth_rq_end_elem_handler(void *user_data, const XML_Char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;
  CAR_PARSED_RR *parsed_rr;
  int missing_element;

  TRC_ENTRY("car_reauth_rq_end_elem_handler");

  /***************************************************************************/
  /* get the transaction info, trans, by casting the user data               */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_ASSERT((trans != NULL), (TB, "NULL user data in AuthRq end handler"));

  /***************************************************************************/
  /* Do common AIB end tag handling.                                         */
  /***************************************************************************/
  aib_end_element_handler(trans, name);

  /***************************************************************************/
  /* The rest only needs doing if we've reached the end of the auth request  */
  /* component...                                                            */
  /***************************************************************************/
  if ((!strcmp(name, com_reauth_req)) || (!strcmp(name, com_reauth_s_req)))
  {
    /*************************************************************************/
    /* Reset the xml layer's own handlers                                    */
    /*************************************************************************/
    xml_set_default_handlers(trans);

    /*************************************************************************/
    /* Get the parsed data memory                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got end of %s", name));
    parsed_rr = (CAR_PARSED_RR *)trans->aib_state;
    if (parsed_rr == NULL)
    {
      /***********************************************************************/
      /* there's nothing more we can do                                      */
      /***********************************************************************/
      TRC_ALERT((TB, "End element but no parsed data memory"));
      goto EXIT_LABEL;
    }

    if (parsed_rr->common.status.code != OSP_STATUS_SUCCESS)
    {
      TRC_ALERT((TB, "Unsuccessful parsing status: %d",
                                        parsed_rr->common.status.code));

      car_build_error(trans, parsed_rr->common.corr->comp_type);
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Is the appropriate callback registered?                               */
    /*************************************************************************/
    osp_get_recursive_lock(&car_cb_mutex);
    if (car_reauth_rq_cb == NULL)
    {
      osp_release_recursive_lock(&car_cb_mutex);
      TRC_NORMAL((TB, "End reauth rq but no callback - done"));
      /***********************************************************************/
      /* add an error response for the transaction                           */
      /***********************************************************************/
      parsed_rr->common.status.code = OSP_STATUS_SERVICE_NOT_AVAILABLE;
      parsed_rr->common.status.desc = car_service_not_available;
      car_build_error(trans, parsed_rr->common.corr->comp_type);
      goto EXIT_LABEL;
    }

#ifdef OSP_DATA_COVERAGE
    aib_scan_registered_structures();
#endif

    /*************************************************************************/
    /* Now we can pass the auth request on to the AR app                     */
    /*************************************************************************/
    (*car_reauth_rq_cb)(
                       (OSP_CORRELATOR)(parsed_rr->common.corr),
                       parsed_rr->timestamp,
                       parsed_rr->role,
                       parsed_rr->call_id,
                       parsed_rr->source_info,
                       (OSP_ADDRESS *)SL_HEAD(parsed_rr->source_alt_list),
                       parsed_rr->dest_info,
                       (OSP_ADDRESS *)SL_HEAD(parsed_rr->dest_alt_list),
                       parsed_rr->transaction_id,
                       (OSP_USAGE_DETAIL *)SL_HEAD(parsed_rr->use_detail_list),
                       (OSP_TOKEN *)SL_HEAD(parsed_rr->token_list),
                       &trans->client_id);

    osp_release_recursive_lock(&car_cb_mutex);
  }
  else
  {
    /*************************************************************************/
    /* We're processing the end of one of our child elements.  Set the       */
    /* parsed data pointer to NULL so that it's not freed                    */
    /*************************************************************************/
    parsed_rr = NULL;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Free up the memory from any parsed data                                 */
  /***************************************************************************/
  if (parsed_rr != NULL)
  {
    car_free_rr_mem(trans);
  }

  TRC_EXIT();

  return;

} /* car_reauth_rq_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      car_free_rr_mem                                                */
/*                                                                           */
/* Purpose:   Frees parsed reauth request and associated memory              */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*                                                                           */
/* Operation: Get the reauth request data structure from trans.              */
/*            Free all the memory within the structure, including walking    */
/*            the lists                                                      */
/*                                                                           */
/**PROC-**********************************************************************/

void car_free_rr_mem(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CAR_PARSED_RR *parsed_rr;
  OSP_ADDRESS *alt_item;
  OSP_TOKEN *tok_item;
  OSP_USAGE_DETAIL *use_item;

  TRC_ENTRY("car_free_rr_mem");

  /***************************************************************************/
  /* Get the parsed data structure we're to free                             */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Passed NULL transaction info"));
  parsed_rr = trans->aib_state;
  TRC_ASSERT((parsed_rr != NULL), (TB, "Asked to free NULL parsed rr mem"));
  TRC_DETAIL((TB, "Freeing rr data %p", parsed_rr));

  /***************************************************************************/
  /* Free all the easy ones                                                  */
  /***************************************************************************/
  aib_free_call_id(parsed_rr->call_id);

  TRC_DETAIL((TB, "Free source info"));
  aib_free_address(parsed_rr->source_info);

  TRC_DETAIL((TB, "Free dest info"));
  aib_free_address(parsed_rr->dest_info);

  /***************************************************************************/
  /* Now deal with the lists                                                 */
  /***************************************************************************/
  while ((alt_item = (OSP_ADDRESS *)SL_HEAD(parsed_rr->source_alt_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_rr->source_alt_list);
    TRC_DETAIL((TB, "Free source alt %p", alt_item));
    aib_free_address(alt_item);
  }

  while ((alt_item = (OSP_ADDRESS *)SL_HEAD(parsed_rr->dest_alt_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_rr->dest_alt_list);
    TRC_DETAIL((TB, "Free dest alt %p", alt_item));
    aib_free_address(alt_item);
  }

  while ((use_item = (OSP_USAGE_DETAIL *)SL_HEAD(parsed_rr->use_detail_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_rr->use_detail_list);
    TRC_DETAIL((TB, "Free usage detail %p", use_item));
    aib_free_usage_detail(use_item);
  }

  while ((tok_item = (OSP_TOKEN *)SL_HEAD(parsed_rr->token_list)) != NULL)
  {
    SL_REMOVE_HEAD(parsed_rr->token_list);
    TRC_DETAIL((TB, "Free token %p", tok_item));
    aib_free_token(tok_item);
  }

  TRC_DETAIL((TB, "Free final structure %p", parsed_rr));
  rm_release_mem(parsed_rr);

  /***************************************************************************/
  /* Having released the memory, NULL the pointer to stop anyone trying to   */
  /* reuse it                                                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freed rr data successfully"));
  trans->aib_state = NULL;

  TRC_EXIT();

  return;

} /* car_free_rr_mem */

