
/**MOD+***********************************************************************/
/* Module:    carsubsa.c                                                     */
/*                                                                           */
/* Purpose:   Call Authorization and Routing (CAR) Application Interface     */
/*            Layer (AIL) SubscriberAuthenticationRequest and Response       */
/*            Handling                                                       */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.9                $ $Modtime::   Jan 11 2001 15:56:30   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <carincl.h>

/**PROC+**********************************************************************/
/* Name:      osp_sa_response                                                */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_sa_response(OSP_CORRELATOR corr,
                              OSP_STATUS * status,
                              OSP_TOKEN *subs_token_list,
                              OSP_CREDIT_AMOUNT *credit_amount,
                              OSP_CREDIT_TIME *credit_time,
                              unsigned int sig_required)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_INT_CORRELATOR *int_corr;
  TRANS_INFO *trans;

  TRC_API_ENTRY("osp_sa_response");

  /***************************************************************************/
  /* Check we're intialized                                                  */
  /***************************************************************************/
  if (!car_initialized)
  {
    TRC_ERROR((TB, "Subs Auth Rq received when not initialized"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get at the correlator data                                              */
  /***************************************************************************/
  int_corr = (OSP_INT_CORRELATOR *)corr;
  if (int_corr == NULL)
  {
    TRC_ERROR((TB, "Null correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  trans = int_corr->trans;
  if (trans == NULL)
  {
    TRC_ERROR((TB, "Corrupted correlator supplied."));
    rc = OSP_ERR_BAD_CORR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Status must be supplied                                                 */
  /***************************************************************************/
  if (status == NULL)
  {
    TRC_ERROR((TB, "Missing required status information."));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction start build time.                                   */
  /***************************************************************************/
  SM_GETTIMES(trans->timings[SM_TIMINGS_START_BUILD]);
#endif

  /***************************************************************************/
  /* We're about to start using the send buffer - lock everyone else out     */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->mutex);

  /***************************************************************************/
  /* If signing is required, set the flag on the transaction - the whole     */
  /* message is signed, not just this component                              */
  /***************************************************************************/
  if (sig_required)
  {
    TRC_NORMAL((TB, "Signing required on trans %p", trans));
    trans->sign_response = TRUE;
  }

  /***************************************************************************/
  /* Set up the XML for the response, using the supplied parameters and the  */
  /* component id from the correlator                                        */
  /***************************************************************************/
  if (!car_build_sa_response(trans, int_corr->comp_id, status,
                                  subs_token_list, credit_amount, credit_time))
  {
    TRC_ERROR((TB, "Failed to build xml for SA Rsp"));
    trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
  }

  /***************************************************************************/
  /* We've finished - update the response count and see if the response      */
  /* needs to be sent.  Note that xml_maybe_send will free the transaction   */
  /* mutex for us.                                                           */
  /***************************************************************************/
  trans->response_count++;
  TRC_NORMAL((TB, "Transaction %p resp count now %d", trans,
                                                       trans->response_count));
  xml_maybe_send(trans);

EXIT_LABEL:
  /***************************************************************************/
  /* Free the correlator component id and the correlator itself              */
  /***************************************************************************/
  rm_release_mem(int_corr->comp_id);
  rm_release_mem(int_corr);

  TRC_API_EXIT();

  return(rc);

} /* osp_sa_response */

/**PROC+**********************************************************************/
/* Name:      car_sar_end_elem_handler                                       */
/*                                                                           */
/* Purpose:   End element handler called for the child elements of a         */
/*            subscriber authentication request, and for the end of the      */
/*            request itself                                                 */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name         - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

void car_sar_end_elem_handler(void *user_data, const XML_Char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *trans;
  CAR_PARSED_SAR *parsed_sar;
  int missing_element;

  TRC_ENTRY("car_sar_end_elem_handler");

  /***************************************************************************/
  /* get the transaction info, trans, by casting the user data               */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_ASSERT((trans != NULL), (TB, "NULL user data in SAR end handler"));

  /***************************************************************************/
  /* Do common AIB end tag handling.                                         */
  /***************************************************************************/
  aib_end_element_handler(trans, name);

  /***************************************************************************/
  /* The rest only needs doing if we've reached the end of the subs auth     */
  /* request component...                                                    */
  /***************************************************************************/
  if (!strcmp(name, com_subs_auth_req))
  {
    /*************************************************************************/
    /* Reset the xml layer's own handlers                                    */
    /*************************************************************************/
    xml_set_default_handlers(trans);

    /*************************************************************************/
    /* Get the parsed data memory                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got end of %s", name));
    parsed_sar = (CAR_PARSED_SAR *)trans->aib_state;
    if (parsed_sar == NULL)
    {
      /***********************************************************************/
      /* there's nothing more we can do                                      */
      /***********************************************************************/
      TRC_ALERT((TB, "End element but no parsed data memory"));
      goto EXIT_LABEL;
    }

    if (parsed_sar->common.status.code != OSP_STATUS_SUCCESS)
    {
      TRC_ALERT((TB, "Unsuccessful parsing status: %d",
                                      parsed_sar->common.status.code));

      car_build_error(trans, parsed_sar->common.corr->comp_type);
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Is the appropriate callback registered?                               */
    /*************************************************************************/
    osp_get_recursive_lock(&car_cb_mutex);
    if (car_subs_auth_rq_cb == NULL)
    {
      osp_release_recursive_lock(&car_cb_mutex);
      TRC_NORMAL((TB, "End sa rq but no callback - done"));
      /***********************************************************************/
      /* add an error response for the transaction                           */
      /***********************************************************************/
      parsed_sar->common.status.code = OSP_STATUS_SERVICE_NOT_AVAILABLE;
      parsed_sar->common.status.desc = car_service_not_available;
      car_build_error(trans, parsed_sar->common.corr->comp_type);
      goto EXIT_LABEL;
    }

#ifdef OSP_DATA_COVERAGE
    aib_scan_registered_structures();
#endif

    /*************************************************************************/
    /* Now we can pass the auth indication on to the AR app                  */
    /*************************************************************************/
    (*car_subs_auth_rq_cb)((OSP_CORRELATOR)(parsed_sar->common.corr),
                           parsed_sar->timestamp,
                           parsed_sar->source_info,
                           (OSP_ADDRESS *)SL_HEAD(parsed_sar->source_alt_list),
                           parsed_sar->dest_info,
                           parsed_sar->service,
                           parsed_sar->subs_authen_info,
                           &trans->client_id);

    osp_release_recursive_lock(&car_cb_mutex);
  }
  else
  {
    /*************************************************************************/
    /* We're processing the end of one of our child elements.  Set the       */
    /* parsed data pointer to NULL so that it's not freed                    */
    /*************************************************************************/
    parsed_sar = NULL;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Free up the memory from any parsed data                                 */
  /***************************************************************************/
  if (parsed_sar != NULL)
  {
    car_free_sar_mem(trans);
  }

  TRC_EXIT();

  return;

} /* car_sar_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      car_build_sa_response                                          */
/*                                                                           */
/* Purpose:   Writes a subs auth response as XML                             */
/*                                                                           */
/* Returns:   TRUE if write succeeds, FALSE otherwise                        */
/*                                                                           */
/* Params:    IN     trans        - transaction associated with the response */
/*            IN     comp_id      - the component id for the response        */
/*            IN     status       - OSP status information                   */
/*            IN     subs_token_list                                         */
/*                                - list of subscriber authentication tokens */
/*                                                                           */
/**PROC-**********************************************************************/

int car_build_sa_response(TRANS_INFO *trans,
                          char *comp_id,
                          OSP_STATUS *status,
                          OSP_TOKEN *subs_token_list,
                          OSP_CREDIT_AMOUNT *credit_amount,
                          OSP_CREDIT_TIME *credit_time)

{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;

  TRC_ENTRY("car_build_sa_response");

  /***************************************************************************/
  /* Ensure we got a transaction to write to                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Supplied NULL transaction"));

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  if (!aib_space_available(trans,
                           1 + 1 + 2 + 3 +
                           strlen(com_subs_auth_rsp) + strlen(com_comp_id) +
                           strlen(comp_id)))
  {
    TRC_ERROR((TB, "No room for XML component"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, com_subs_auth_rsp);
  AIB_ADD_DATA(trans, com_sp, 1);
  AIB_ADD_NULL(trans, com_comp_id);
  AIB_ADD_DATA(trans, com_eq_qu, 2);
  AIB_ADD_NULL(trans, comp_id);
  AIB_ADD_DATA(trans, com_qu_gt_nl, 3);

  /***************************************************************************/
  /* Add each element in turn                                                */
  /***************************************************************************/
  if (!aib_write_xml_timestamp(trans))
  {
    TRC_ERROR((TB, "Failure appending timestamp"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_status(trans, status))
  {
    TRC_ERROR((TB, "Failure appending status"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_token_list(trans, (char *)com_subs_authen,
                                                              subs_token_list))
  {
    TRC_ERROR((TB, "Failure appending subs authen info list"));
    goto EXIT_LABEL;
  }

  if (credit_amount != NULL)
  {
    TRC_DETAIL((TB, "Adding credit amount"));
    if (!aib_write_xml_credit_amount(trans, credit_amount))
    {
      TRC_ERROR((TB, "Failure appending credit amount"));
      goto EXIT_LABEL;
    }
  }

  if (credit_time != NULL)
  {
    TRC_DETAIL((TB, "Adding credit time"));
    if (!aib_write_xml_credit_time(trans, credit_time))
    {
      TRC_ERROR((TB, "Failure appending credit time"));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Append the close tag                                                    */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(com_subs_auth_rsp)))
  {
    TRC_ERROR((TB, "No room for component close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, com_subs_auth_rsp);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* car_build_sa_response */

/**PROC+**********************************************************************/
/* Name:      car_free_sar_mem                                               */
/*                                                                           */
/* Purpose:   Frees parsed auth indication and associated memory             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*                                                                           */
/* Operation: Get the auth indication structure from trans.                  */
/*            Free all the memory within the structure, including walking    */
/*            the lists                                                      */
/*                                                                           */
/**PROC-**********************************************************************/

void car_free_sar_mem(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CAR_PARSED_SAR *parsed_sar;
  OSP_ADDRESS *alt_item;

  TRC_ENTRY("car_free_sar_mem");

  /***************************************************************************/
  /* Get the parsed data structure we're to free                             */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Passed NULL transaction info"));
  parsed_sar = trans->aib_state;
  TRC_ASSERT((parsed_sar != NULL), (TB, "Asked to free NULL parsed ai mem"));
  TRC_DETAIL((TB, "Freeing ai data %p", parsed_sar));

  aib_free_address(parsed_sar->source_info);
  aib_free_address(parsed_sar->dest_info);
  aib_free_service(parsed_sar->service);

  /***************************************************************************/
  /* Now deal with the lists                                                 */
  /***************************************************************************/
  while ((alt_item = (OSP_ADDRESS *)SL_HEAD(parsed_sar->source_alt_list))
                                                                       != NULL)
  {
    SL_REMOVE_HEAD(parsed_sar->source_alt_list);
    TRC_DETAIL((TB, "Free source alt %p", alt_item));
    aib_free_address(alt_item);
  }

  TRC_DETAIL((TB, "Free final structure %p", parsed_sar));
  rm_release_mem(parsed_sar);

  /***************************************************************************/
  /* Having released the memory, NULL the pointer to stop anyone trying to   */
  /* reuse it                                                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freed ai data successfully"));
  trans->aib_state = NULL;

  TRC_EXIT();

  return;

} /* car_free_sar_mem */

