
/**MOD+***********************************************************************/
/* Module:    carutil.c                                                      */
/*                                                                           */
/* Purpose:   Call Authorization and Routing (CAR) Application Interface     */
/*            Layer (AIL) utility functions                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.22               $ $Modtime::   Jan 11 2001 15:56:30   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <carincl.h>

/**PROC+**********************************************************************/
/* Name:      car_build_error                                                */
/*                                                                           */
/* Purpose:   append an error response to a transaction                      */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction                              */
/*            IN     type         - component type                           */
/*                                                                           */
/**PROC-**********************************************************************/

void car_build_error(TRANS_INFO *trans, int type)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CAR_PARSED_AR * parsed_data;

  TRC_ENTRY("car_build_error");

  /***************************************************************************/
  /* Check that we've got a valid transaction                                */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Called with NULL transaction info"));
  parsed_data = trans->aib_state;
  TRC_ASSERT((parsed_data != NULL), (TB, "Called with NULL parsed data"));

  TRC_ALERT((TB, "Building error %d for component %d, transaction %p",
                         parsed_data->common.status.code, type, trans));

  /***************************************************************************/
  /* we're about to start using the send buffer - lock it                    */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->mutex);

  /***************************************************************************/
  /* Even though we're generating an error, we have processed a component,   */
  /* so increment the response count                                         */
  /***************************************************************************/
  trans->response_count++;
  TRC_NORMAL((TB, "Transaction %08lx resp count now %d", trans,
                                                      trans->response_count));

  /***************************************************************************/
  /* Now do the per component processing                                     */
  /***************************************************************************/
  switch (type)
  {
    case AIB_AUTH_RQ:
    case AIB_AUTH_S_RQ:
    case AIB_REAUTH_RQ:
    case AIB_REAUTH_S_RQ:
      /***********************************************************************/
      /* Build the response using the supplied status code                   */
      /***********************************************************************/
      TRC_DETAIL((TB, "Build error response for (Re)Auth Rq"));
      if (!car_build_auth_response(trans,
                                   type,
                                   parsed_data->common.corr->comp_id,
                                   &(parsed_data->common.status),
                                   "0",
                                   NULL,
                                   NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for (re)Auth Rq"));
        trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      }
      break;

    case AIB_AUTH_IND:
    case AIB_AUTH_S_IND:
      /***********************************************************************/
      /* Build the response using the supplied status code                   */
      /***********************************************************************/
      TRC_DETAIL((TB, "Build error response for Auth Rq"));
      if (!car_build_auth_confirm(trans,
                                  type,
                                  parsed_data->common.corr->comp_id,
                                  &(parsed_data->common.status),
                                   NULL,
                                   NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Auth Ind"));
        trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      }
      break;

    case AIB_PRICING_IND:
      /***********************************************************************/
      /* Build the response using the supplied status code                   */
      /***********************************************************************/
      TRC_DETAIL((TB, "Build error response for Pri Ind"));
      if (!car_build_price_confirm(trans,
                                   parsed_data->common.corr->comp_id,
                                   &(parsed_data->common.status)))
      {
        TRC_ERROR((TB, "Failed to build error response for Pri Ind"));
        trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      }
      break;

    case AIB_CAPS_IND:
      /***********************************************************************/
      /* Build the response using the supplied status code                   */
      /***********************************************************************/
      TRC_DETAIL((TB, "Build error response for Caps Ind"));
      if (!car_build_caps_confirm(trans,
                                  parsed_data->common.corr->comp_id,
                                  &(parsed_data->common.status),
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Caps Ind"));
        trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      }
      break;

    case AIB_SUBS_AUTHEN_RQ:
      /***********************************************************************/
      /* Build the response using the supplied status code                   */
      /***********************************************************************/
      TRC_DETAIL((TB, "Build error response for Subs Authen Rq"));
      if (!car_build_sa_response(trans,
                                 parsed_data->common.corr->comp_id,
                                 &(parsed_data->common.status),
                                 NULL,
                                 NULL,
                                 NULL))
      {
        TRC_ERROR((TB, "Failed to build error response for Subs Authen Rq"));
        trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      }
      break;

    default:
      TRC_ERROR((TB, "Asked to build error %d for unknown component %d, " \
           "transaction %08lx", parsed_data->common.status.code, type, trans));
      break;
  }

  /***************************************************************************/
  /* See if it's time to send the buffer.  Note that xml_maybe_send will     */
  /* free the transaction mutex for us.                                      */
  /***************************************************************************/
  xml_maybe_send(trans);

  /***************************************************************************/
  /* Free the correlator component id and the correlator itself              */
  /***************************************************************************/
  rm_release_mem(parsed_data->common.corr->comp_id);
  rm_release_mem(parsed_data->common.corr);

  TRC_EXIT();

  return;

} /* car_build_error */

/**PROC+**********************************************************************/
/* Name:      car_build_auth_response                                        */
/*                                                                           */
/* Purpose:   Writes a (re)authorization response as XML                     */
/*                                                                           */
/* Returns:   TRUE if write succeeds, FALSE otherwise                        */
/*                                                                           */
/* Params:    IN     trans        - transaction associated with the response */
/*            IN     comp_type    - type, used to choose correct spelling of */
/*                                  the response                             */
/*            IN     comp_id      - the component id for the response        */
/*            IN     status       - OSP status information                   */
/*            IN     trans_id     - authorization transaction id             */
/*            IN     token_list   - list of authorization tokens             */
/*            IN     dest_list    - list of destinations for which service   */
/*                                  is authorized                            */
/*                                                                           */
/**PROC-**********************************************************************/

int car_build_auth_response(TRANS_INFO *trans,
                            int comp_type,
                            char *comp_id,
                            OSP_STATUS *status,
                            char *trans_id,
                            OSP_TOKEN *token_list,
                            OSP_DESTINATION *dest_list)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  const char *comp_name;

  TRC_ENTRY("car_build_auth_response");

  /***************************************************************************/
  /* Ensure we got a transaction to write to                                 */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Supplied NULL transaction"));

  /***************************************************************************/
  /* Select the correct spelling of the component name (some older clients   */
  /* may use the International English spelling, (Re)AuthorisationResponse   */
  /***************************************************************************/
  if (comp_type == AIB_AUTH_RQ)
  {
    TRC_DETAIL((TB, "AR: Use US spelling"));
    comp_name = com_auth_rsp;
  }
  else if (comp_type == AIB_AUTH_S_RQ)
  {
    TRC_DETAIL((TB, "AR: Use International spelling"));
    comp_name = com_auth_s_rsp;
  }
  else if (comp_type == AIB_REAUTH_RQ)
  {
    TRC_DETAIL((TB, "RR: Use US spelling"));
    comp_name = com_reauth_rsp;
  }
  else
  {
    TRC_DETAIL((TB, "RR: Use International spelling"));
    comp_name = com_reauth_s_rsp;
  }

  /***************************************************************************/
  /* Write the start tag                                                     */
  /***************************************************************************/
  if (!aib_space_available(trans,
                           1 + 1 + 2 + 3 +
                           strlen(comp_name) + strlen(com_comp_id) +
                           strlen(comp_id)))
  {
    TRC_ERROR((TB, "No room for XML component"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt, 1);
  AIB_ADD_NULL(trans, comp_name);
  AIB_ADD_DATA(trans, com_sp, 1);
  AIB_ADD_NULL(trans, com_comp_id);
  AIB_ADD_DATA(trans, com_eq_qu, 2);
  AIB_ADD_NULL(trans, comp_id);
  AIB_ADD_DATA(trans, com_qu_gt_nl, 3);

  /***************************************************************************/
  /* Add each element in turn                                                */
  /***************************************************************************/
  if (!aib_write_xml_timestamp(trans))
  {
    TRC_ERROR((TB, "Failure appending timestamp"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_status(trans, status))
  {
    TRC_ERROR((TB, "Failure appending status"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_element(trans, com_trans_id, trans_id))
  {
    TRC_ERROR((TB, "Failure appending transaction id"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_token_list(trans, (char *)com_token, token_list))
  {
    TRC_ERROR((TB, "Failure appending token list"));
    goto EXIT_LABEL;
  }

  if (!aib_write_xml_destination_list(trans, dest_list))
  {
    TRC_ERROR((TB, "Failure appending destination list"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Append the close tag                                                    */
  /***************************************************************************/
  if (!aib_space_available(trans, 2 + 2 + strlen(comp_name)))
  {
    TRC_ERROR((TB, "No room for component close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(trans, com_lt_sl, 2);
  AIB_ADD_NULL(trans, comp_name);
  AIB_ADD_DATA(trans, com_gt_nl, 2);

  rc = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* car_build_auth_response */

