
/**MOD+***********************************************************************/
/* Module:    ccmca.c                                                        */
/*                                                                           */
/* Purpose:   CCM certificate authority functions                            */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.18               $ $Modtime::   Jan 11 2001 15:56:30   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_get_cert                                                   */
/*                                                                           */
/* Purpose:   Get a certificate given a subject name                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     subj_name      - the subject name of the required cert  */
/*            OUT    cert_buf       - the resulting certificate (if found)   */
/*                                      buffer must be freed by caller       */
/*            OUT    cert_len       - length of certificate                  */
/*                                                                           */
/* Operation: Call a helper function to do the lookup and convert the        */
/*            result to DER-encoded ASN.1.                                   */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_get_cert(char *subj_name,
                    char **cert_buf,
                    unsigned int *cert_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509 *x509_cert = NULL;
  unsigned char *mem = NULL;
  int len;

  TRC_API_ENTRY("ccm_get_cert");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if (subj_name == NULL)
  {
    TRC_ERROR((TB, "Null pointer passed for subj_name"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((cert_buf == NULL) || (cert_len == NULL))
  {
    TRC_ERROR((TB, "cert_buf and cert_len are both required"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Use our internal function to get an X509 structure.  In the general     */
  /* case, we won't block on this call because the certificate we are        */
  /* getting will have already been loaded into local storage and we         */
  /* therefore won't have to do an LDAP lookup.                              */
  /***************************************************************************/
  rc = ccm_int_get_cert(subj_name, &x509_cert);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get certificate (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now convert the X509 structure to ASN.1.                                */
  /***************************************************************************/
  len = i2d_X509(x509_cert, NULL);
  mem = (unsigned char *)rm_get_mem(len);
  if (mem == NULL)
  {
    TRC_ERROR((TB, "Couldn't get mem to return certificate"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  *cert_buf = (char *)mem;
  *cert_len = i2d_X509(x509_cert, &mem);

  TRC_NORMAL((TB, "Got X.509 cert OK"));

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* Calculate and trace MD5 and SHA-1 fingerprints.                         */
  /***************************************************************************/
  ccm_show_fingerprints(x509_cert);
#endif

  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We failed: pass back a null pointer.                                  */
    /*************************************************************************/
    *cert_buf = NULL;
    *cert_len = 0;
  }

  if (x509_cert != NULL)
  {
    X509_free(x509_cert);
  }

  TRC_API_EXIT();

  return(rc);

} /* ccm_get_cert */

/**PROC+**********************************************************************/
/* Name:      ccm_get_cert_chain                                             */
/*                                                                           */
/* Purpose:   Get a certificate chain up to the root authority, given a      */
/*            subject name                                                   */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     subj_name      - the subject name of the required cert  */
/*            OUT    cert_chain     - certificate chain                      */
/*                                                                           */
/* Operation: Get the certificate for the supplied name                      */
/*            while certificate issuer name not the same as the subject name */
/*              get the issuer's certificate                                 */
/*            convert each certificate to DER form and store in the list as  */
/*            OSP_CERTIFICATEs                                               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_get_cert_chain(char *subj_name,
                          OSP_CERTIFICATE **cert_chain)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  SLIST_ROOT chain;
  OSP_CERTIFICATE *osp_cert;
  X509 *x509_cert;
  X509_NAME *x509_supplied_subj_name;
  X509_NAME *x509_subj_name;
  X509_NAME *x509_issuer_name;
  unsigned char *der_mem, *der_ptr;
  int der_len;

  TRC_API_ENTRY("ccm_get_cert_chain");

  /***************************************************************************/
  /* Initialize variables                                                    */
  /***************************************************************************/
  SL_INIT_ROOT(chain);
  osp_cert = NULL;
  x509_cert = NULL;
  x509_supplied_subj_name = NULL;
  x509_subj_name = NULL;
  x509_issuer_name = NULL;
  der_mem = NULL;
  der_len = 0;

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if (subj_name == NULL)
  {
    TRC_ERROR((TB, "Null pointer passed for subj_name"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (cert_chain == NULL)
  {
    TRC_ERROR((TB, "cert_chain pointer is required"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Getting cert chain for '%s'", subj_name));

  /***************************************************************************/
  /* Convert the subject name into an X509_NAME structure.                   */
  /***************************************************************************/
  rc = ccm_rfc2253_to_x509_name(subj_name, &x509_supplied_subj_name);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to convert subject name to X509_NAME (%u)", rc));
    goto EXIT_LABEL;
  }
  TRC_ASSERT((x509_supplied_subj_name != NULL),
                                       (TB, "x509 name NULL despite good rc"));

  /***************************************************************************/
  /* Now we enter a loop that amounts to                                     */
  /* do                                                                      */
  /*   get cert for subj name                                                */
  /*   store cert in OSP form and add to list                                */
  /*   get cert issue name                                                   */
  /* while issuer name != subj name                                          */
  /***************************************************************************/
  x509_subj_name = x509_supplied_subj_name;
  TRC_DETAIL((TB, "Starting to walk up cert list"));
  while (TRUE)
  {
    /*************************************************************************/
    /* Get the certificate for the subject name                              */
    /*************************************************************************/
    rc = ccm_int_get_cert_x509_name(x509_subj_name, &x509_cert);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Error %u getting certificate", rc));
      goto EXIT_LABEL;
    }
    TRC_ASSERT((x509_cert != NULL), (TB, "x509 cert NULL despite good rc"));
    TRC_DETAIL((TB, "Retrieved certificate OK"));

    /*************************************************************************/
    /* Get the issuer name                                                   */
    /*************************************************************************/
    TRC_DETAIL((TB, "Getting issuer name"));
    x509_issuer_name = X509_get_issuer_name(x509_cert);
    if (x509_issuer_name == NULL)
    {
      TRC_ERROR((TB, "Failed to get issuer name"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Get an osp certificate structure in which to store the certificate    */
    /*************************************************************************/
    TRC_DETAIL((TB, "Get cert storage"));
    osp_cert = rm_get_mem(sizeof(OSP_CERTIFICATE));
    if (osp_cert == NULL)
    {
      TRC_ERROR((TB, "Failed to get mem for OSP certificate"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    osp_cert->next = NULL;
    osp_cert->encoding = NULL;
    osp_cert->cert_len = 0;
    osp_cert->cert_data = NULL;

    /*************************************************************************/
    /* Convert the X509 structure to DER form                                */
    /*************************************************************************/
    TRC_DETAIL((TB, "Copying cert to OSP form"));
    der_len = i2d_X509(x509_cert, NULL);
    der_mem = (unsigned char *)rm_get_mem(der_len);
    if (der_mem == NULL)
    {
      TRC_ERROR((TB, "Couldn't get mem to store certificate"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Pass in a copy of the pointer to the buffer - i2d moves the pointer   */
    /* on past the newly converted certficate, which makes it hard to free   */
    /* later on!                                                             */
    /*************************************************************************/
    der_ptr = der_mem;
    der_len = i2d_X509(x509_cert, &der_ptr);
    if (der_len == 0)
    {
      TRC_ERROR((TB, "Failed to copy cert data to der form"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
    TRC_DATA("DER-form certificate:", der_mem, der_len);

    /*************************************************************************/
    /* Base64 the der buffer                                                 */
    /*************************************************************************/
    rc = osp_base64_encode((char *)der_mem, der_len, &osp_cert->cert_data,
                                        (unsigned int *)&(osp_cert->cert_len));
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failure base64-encoding the certficate"));
      goto EXIT_LABEL;
    }
    TRC_DETAIL((TB, "Base64'd cert len %d", osp_cert->cert_len));

    OSP_STRDUP(osp_cert->encoding, com_base64);
    if (osp_cert->encoding == NULL)
    {
      TRC_ERROR((TB, "Failed to store cert encoding"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Add the certificate to the list                                       */
    /*************************************************************************/
    SL_ADD_TO_TAIL(chain, osp_cert);
    osp_cert = NULL;

    /*************************************************************************/
    /* If we've reached a self-signed certificate (subject name == issuer    */
    /* name), then we've hit the root authority and should stop              */
    /*************************************************************************/
    if (!X509_name_cmp(x509_subj_name, x509_issuer_name))
    {
      TRC_DETAIL((TB, "Reached root authority - stop"));
      break;
    }

    /*************************************************************************/
    /* Tidy up what we no longer need from this pass                         */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing resources from this pass"));
    if (x509_cert != NULL)
    {
      TRC_DETAIL((TB, "Free cert"));
      X509_free(x509_cert);
      x509_cert = NULL;
    }
    rm_release_mem(der_mem);
    der_mem = NULL;

    /*************************************************************************/
    /* Next time through, we want to get the certificate for the issuer      */
    /*************************************************************************/
    x509_subj_name = x509_issuer_name;
    x509_issuer_name = NULL;
  }
  TRC_DETAIL((TB, "Finished walking cert list"));

  /***************************************************************************/
  /* All done                                                                */
  /***************************************************************************/
  *cert_chain = SL_HEAD(chain);
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* We failed: pass back a null pointer, and tidy up anything we did get    */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    *cert_chain = NULL;

    /*************************************************************************/
    /* We might have a partial certificate hanging around, not in the list   */
    /*************************************************************************/
    if (osp_cert != NULL)
    {
      TRC_DETAIL((TB, "Freeing osp cert"));
      rm_release_mem(osp_cert->encoding);
      rm_release_mem(osp_cert->cert_data);
      rm_release_mem(osp_cert);
    }

    /*************************************************************************/
    /* Now free the list                                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing cert chain"));
    ccm_free_cert_chain((OSP_CERTIFICATE *)SL_HEAD(chain));
  }

  /***************************************************************************/
  /* Tidy up everything else.  Note that we do free the x509 version of the  */
  /* supplied subject name, but we MUST NOT free the subject/issuer names -  */
  /* these are not copies but pointers into the certificate...               */
  /***************************************************************************/
  if (x509_supplied_subj_name != NULL)
  {
    TRC_DETAIL((TB, "Freeing x509 version of supplied subj name"));
    X509_NAME_free(x509_supplied_subj_name);
  }

  if (x509_cert != NULL)
  {
    TRC_DETAIL((TB, "Freeing x509 cert"));
    X509_free(x509_cert);
  }

  if (der_mem != NULL)
  {
    TRC_DETAIL((TB, "Freeing der mem"));
    rm_release_mem(der_mem);
  }

  TRC_API_EXIT();

  return(rc);

} /* ccm_get_cert_chain */

/**PROC+**********************************************************************/
/* Name:      ccm_request_new_cert                                           */
/*                                                                           */
/* Purpose:   Kick off the process of issuing a new certificate              */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     req_asn1       - the ASN.1 PKCS#10 certificate request  */
/*            IN     req_asn1_len   - length of request                      */
/*            IN     trans_id       - SCEP transaction ID to check           */
/*                                      (may be NULL)                        */
/*            IN     ca_name        - CA identity to use to sign the cert    */
/*            IN     cert_cb        - function to call with result           */
/*            IN     corr           - correlator to pass on callback         */
/*                                                                           */
/* Operation: If trans_id is supplied, check that it's correct. Then         */
/*            package up the data and kick off a new thread to do the        */
/*            processing asynchronously.                                     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_request_new_cert(char *req_asn1,
                            unsigned int req_asn1_len,
                            char *trans_id,
                            char *ca_name,
                            PCCM_NEW_CERT_CALLBACK cert_cb,
                            void *corr)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509_REQ *req = NULL;
  CCM_NEW_CERT_CONTEXT *context = NULL;
  EVP_PKEY *pub_key = NULL;
  unsigned char digest[MD5_DIGEST_LENGTH];
  unsigned int digest_len = 0;
  pthread_t tid;
  int ii, jj;
  int error;

  TRC_API_ENTRY("ccm_request_new_cert");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if (req_asn1 == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for PKCS#10 ASN.1 request"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (req_asn1_len == 0)
  {
    TRC_ERROR((TB, "Zero length supplied for PKCS#10 ASN.1 request"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (ca_name == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for ca_name"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (cert_cb == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for new certificate callback"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Convert the PKCS#10 from ASN.1 format.                                  */
  /***************************************************************************/
  TRC_DETAIL((TB, "Converting PKCS#10 request to internal format"));
  req = d2i_X509_REQ(NULL, (unsigned char **)&req_asn1, req_asn1_len);
  if (req == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to convert ASN.1 to PKCS#10"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

#ifdef CCM_DEBUG_X509
  /***************************************************************************/
  /* Print out the request in friendly form.                                 */
  /***************************************************************************/
  printf("Request received:\n");
  X509_REQ_print_fp(stdout, req);
#endif

  /***************************************************************************/
  /* Extract the request's public key.                                       */
  /***************************************************************************/
  pub_key = X509_REQ_get_pubkey(req);
  if (pub_key == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Error getting public key from cert request"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If we have been supplied a SCEP transaction ID, check that it really is */
  /* the same as the MD5 digest of the public key.                           */
  /***************************************************************************/
  if (trans_id != NULL)
  {
    TRC_DETAIL((TB, "Checking SCEP trans ID '%s' against public key digest",
                                                                    trans_id));
    if (strlen(trans_id) != (2 * MD5_DIGEST_LENGTH))
    {
      TRC_ERROR((TB, "SCEP transaction ID is wrong length "
                                     "(expecting %d characters; got %d)",
                                     2 * MD5_DIGEST_LENGTH, strlen(trans_id)));
      rc = OSP_ERR_BAD_PARAM;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Calculate the MD5 digest of the public key in the request.            */
    /*************************************************************************/
    error = ASN1_digest((int(*)())i2d_PUBKEY,
                        EVP_md5(),
                        (char *)pub_key,
                        digest,
                        &digest_len);
    if (error == 0)
    {
      TRC_ERROR((TB, "Failed to calc MD5 digest of request's public key"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }

    TRC_ASSERT((digest_len == MD5_DIGEST_LENGTH),
                                 (TB, "Got MD5 digest length %u", digest_len));
    TRC_DATA("MD5 public key digest from request:", digest, digest_len);

    /*************************************************************************/
    /* Now compare the binary bytes we've just calculated with the hex       */
    /* string in trans_id.                                                   */
    /*************************************************************************/
    for (ii = 0; ii < MD5_DIGEST_LENGTH; ii++)
    {
      jj = 2 * ii;

      if ((trans_id[jj] != ccm_hex[(digest[ii] >> 4) & 0x0f]) ||
          (trans_id[jj + 1] != ccm_hex[digest[ii] & 0x0f]))
      {
        TRC_ERROR((TB, "SCEP transaction ID is invalid: rejecting request"));
        rc = OSP_ERR_BAD_REQUEST;
        goto EXIT_LABEL;
      }
    }

    TRC_DETAIL((TB, "SCEP transaction ID agrees with MD5 public key digest"));
  }

  /***************************************************************************/
  /* Create and initialize the context structure that will be passed to the  */
  /* thread we are about to create.                                          */
  /***************************************************************************/
  TRC_NORMAL((TB, "Processing new certificate request for CA '%s'", ca_name));
  TRC_DETAIL((TB, "Creating new certificate context: corr=%p, cb=%p",
                                                               corr, cert_cb));
  context = (CCM_NEW_CERT_CONTEXT *)rm_get_mem(sizeof(CCM_NEW_CERT_CONTEXT));
  if (context == NULL)
  {
    TRC_ERROR((TB, "Couldn't allocate memory for cert request context"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up the structure                                                    */
  /***************************************************************************/
  context->corr = corr;
  context->callback = cert_cb;
  context->req = req;
  context->pub_key = pub_key;
  memcpy(context->digest, digest, MD5_DIGEST_LENGTH);
  context->ca_name = NULL;
  context->ca_cert = NULL;
  context->subj_name = NULL;
  context->new_cert = NULL;

  OSP_STRDUP(context->ca_name, ca_name);
  if (context->ca_name == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory to copy CA name into context"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create a thread to check the request and issue the certificate.         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Creating CCM thread to handle request"));
  rc = rm_request_thread_create(RM_CCM_THREAD);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "RM denied request for new CCM thread (%u)", rc));
    goto EXIT_LABEL;
  }

  error = pthread_create(&tid,
                         NULL,
                         ccm_cert_request_thread,
                         (void *)context);
  if (error != 0)
  {
    TRC_OS_ERROR((TB, "Failed to create worker thread to issue certificate"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  error = pthread_detach(tid);
  if (error != 0)
  {
    /*************************************************************************/
    /* This will just result in a resource leak, so we don't abort.          */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to detach cert issuer thread 0x%x", tid));
  }

  TRC_NORMAL((TB, "Started thread (tid=0x%x) to issue new certificate", tid));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up on failure.                                                    */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    if (context != NULL)
    {
      /***********************************************************************/
      /* We successfully created a request context, so freeing that will     */
      /* take care of everything.                                            */
      /***********************************************************************/
      TRC_ALERT((TB, "Failed to start new cert processing; freeing context"));
      ccm_free_req_context(context);
    }
    else
    {
      /***********************************************************************/
      /* We have to free our objects individually.                           */
      /***********************************************************************/
      TRC_ALERT((TB, "Failed to start new cert processing; freeing objects"));
      if (req != NULL)
      {
        X509_REQ_free(req);
      }

      if (pub_key != NULL)
      {
        TRC_DETAIL((TB, "Freeing pub_key"));
        EVP_PKEY_free(pub_key);
      }
    }
  }

  TRC_API_EXIT();

  return(rc);

} /* ccm_request_new_cert */

/**PROC+**********************************************************************/
/* Name:      ccm_cert_request_thread                                        */
/*                                                                           */
/* Purpose:   Issue a new certificate based on a PKCS#10 request             */
/*            and place the new cert in the LDAP directory.                  */
/*                                                                           */
/* Returns:   none                                                           */
/*                                                                           */
/* Params:    IN     data           - a pointer to the request context       */
/*                                                                           */
/* Operation: This consists of the following steps:                          */
/*            - verify the request                                           */
/*            - check that the requester doesn't already have a valid cert   */
/*            - check that its attributes are acceptable                     */
/*            - create a new signed certificate                              */
/*            - store the new certificate in the directory                   */
/*            - return the new certificate to the caller                     */
/*                                                                           */
/**PROC-**********************************************************************/

void *ccm_cert_request_thread(void *data)
{
  OSP_RC int_rc = OSP_SUCCESS;
  int ok = FALSE;
  CCM_NEW_CERT_CONTEXT *context = (CCM_NEW_CERT_CONTEXT *)data;
  EVP_PKEY *priv_key = NULL;
  X509_NAME *subject = NULL;
  BIGNUM *new_serial = NULL;
  EVP_MD *sign_digest = NULL;
  unsigned char pubkey_digest[MD5_DIGEST_LENGTH];
  unsigned int pubkey_digest_len = 0;
  unsigned char *mem = NULL;
  char *x509_asn1 = NULL;
  unsigned int x509_asn1_len;
  X509V3_CTX ext_ctx;
  int error;
#ifdef OSP_DEBUG
  char *sn_string = NULL;
#endif

  TRC_THREAD_ENTRY("ccm_cert_request_thread");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((context != NULL), (TB, "Caller passed a null context"));

  /***************************************************************************/
  /* Verify the request using the public key that it contains.               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Verifying certificate request"));
  error = X509_REQ_verify(context->req, context->pub_key);
  if (error < 0)
  {
    OSP_OPENSSL_ERROR((TB, "Error verifying certificate request"));
    goto EXIT_LABEL;
  }
  else if (error == 0)
  {
    TRC_ERROR((TB, "Signature did not match the request"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get a copy of the subject name from the request and add any name suffix */
  /* that has been configured.                                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Expanding subject name from certificate request"));
  int_rc = ccm_expand_subject_name(X509_REQ_get_subject_name(context->req),
                                                                     &subject);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to expand subject name"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get the subject name as an LDAP-compatible string.                      */
  /***************************************************************************/
  int_rc = ccm_x509_name_to_rfc2253(subject, &context->subj_name);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to convert request subject name to string"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* The request appears to be valid.                                        */
  /***************************************************************************/
  TRC_NORMAL((TB, "Processing cert request from '%s'", context->subj_name));

  /***************************************************************************/
  /* Do a lookup to see if there's already a certificate for this subject    */
  /* name.                                                                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Checking for an existing certificate"));
  int_rc = ccm_int_get_cert_x509_name(subject, &context->new_cert);
  if (int_rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We found an existing certificate.  If its public key matches that of  */
    /* the request, we just return the existing certificate.  Otherwise we   */
    /* refuse to issue a new certificate.                                    */
    /*************************************************************************/
    TRC_NORMAL((TB, "Found existing certificate; check public key digests"));

    error = ASN1_digest((int(*)())i2d_PUBKEY,
                        EVP_md5(),
                        (char *)X509_get_pubkey(context->new_cert),
                        pubkey_digest,
                        &pubkey_digest_len);
    if (error == 0)
    {
      TRC_ERROR((TB, "Failed to calc MD5 digest of request's public key"));
      goto EXIT_LABEL;
    }

    TRC_ASSERT((pubkey_digest_len == MD5_DIGEST_LENGTH),
                          (TB, "Got MD5 digest length %u", pubkey_digest_len));

    TRC_DATA("MD5 public key digest from existing certificate:",
                                             pubkey_digest, MD5_DIGEST_LENGTH);
    TRC_DATA("MD5 public key digest from request:",
                                           context->digest, MD5_DIGEST_LENGTH);

    error = memcmp(context->digest, pubkey_digest, pubkey_digest_len);
    if (error == 0)
    {
      /***********************************************************************/
      /* The digests match, which means we should return the certificate     */
      /* that we found in the database.                                      */
      /***********************************************************************/
      TRC_NORMAL((TB, "Returning pre-existing certificate for '%s'",
                                                          context->subj_name));

      /***********************************************************************/
      /* Convert it to DER form for return to the caller.                    */
      /***********************************************************************/
      x509_asn1_len = i2d_X509(context->new_cert, NULL);
      mem = (unsigned char *)rm_get_mem(x509_asn1_len);
      if (mem == NULL)
      {
        TRC_ERROR((TB, "Couldn't get %u bytes to convert new cert to ASN.1",
                                                               x509_asn1_len));
        goto EXIT_LABEL;
      }

      x509_asn1 = (char *)mem;
      x509_asn1_len = i2d_X509(context->new_cert, &mem);

      /***********************************************************************/
      /* Set 'ok' to true to indicate success, and the exit branch will do   */
      /* the callback for us.                                                */
      /***********************************************************************/
      ok = TRUE;
      goto EXIT_LABEL;
    }
    else
    {
      /***********************************************************************/
      /* The digests don't match, which means we should return failure       */
      /* because we must not issue another certificate to this entity.       */
      /***********************************************************************/
      TRC_ERROR((TB, "Request for a second certificate received from '%s'",
                                                          context->subj_name));
      goto EXIT_LABEL;
    }
  }

  TRC_DETAIL((TB, "No existing certificate found"));

  /***************************************************************************/
  /* If any constraints are required on the subject's distinguished name,    */
  /* this is the place to enforce them.                                      */
  /***************************************************************************/

  /***************************************************************************/
  /* Get the CA cert that we will sign with.                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting CA cert to sign with"));
  int_rc = ccm_int_get_cert(context->ca_name, &context->ca_cert);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Couldn't get cert (%u) for CA '%s'",
                                                    int_rc, context->ca_name));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get the private key that we will sign with.                             */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting CA private key to sign with"));
  int_rc = ccm_get_priv_key(context->ca_name, &priv_key);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Couldn't get private key (%u) for CA '%s'",
                                                    int_rc, context->ca_name));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Grab the serial number mutex.  We hold this until we have successfully  */
  /* updated the serial number file with a new number.                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting next serial number"));
  OSP_GET_LOCK(&ccm.ca.serial_num_mutex);

  /***************************************************************************/
  /* Get the serial number we should use.                                    */
  /***************************************************************************/
  new_serial = BN_dup(ccm.ca.serial_num);
  if (new_serial == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to copy serial number info"));
    OSP_RELEASE_LOCK(&ccm.ca.serial_num_mutex);
    goto EXIT_LABEL;
  }

#ifdef OSP_DEBUG
  sn_string = BN_bn2hex(new_serial);
  if (sn_string != NULL)
  {
    TRC_NORMAL((TB, "Using serial number 0x%s for subject '%s'",
                                               sn_string, context->subj_name));
    rm_release_mem(sn_string);
  }
  else
  {
    OSP_OPENSSL_ERROR((TB, "Failed to convert serial num for tracing ('%s')",
                                                          context->subj_name));
  }
#endif

  /***************************************************************************/
  /* Increment the global serial number for the next request and immediately */
  /* write it to the config file so we don't lose track of it if something   */
  /* goes wrong.                                                             */
  /***************************************************************************/
  TRC_DETAIL((TB, "Updating serial number file"));
  BN_add_word(ccm.ca.serial_num, 1);
  int_rc = ccm_update_serial_number();
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to update serial number in file (%u)", int_rc));
    OSP_RELEASE_LOCK(&ccm.ca.serial_num_mutex);
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now we can release the serial number mutex.                             */
  /***************************************************************************/
  OSP_RELEASE_LOCK(&ccm.ca.serial_num_mutex);

  /***************************************************************************/
  /* Create the new certificate.                                             */
  /***************************************************************************/
  TRC_NORMAL((TB, "Creating new certificate"));
  context->new_cert = X509_new();
  if (context->new_cert == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create new certificate structure"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set new cert to version 3 (that's what the '2' does here :-)            */
  /***************************************************************************/
  TRC_DETAIL((TB, "Setting standard attributes"));
  error = X509_set_version(context->new_cert, 2);
  if (error != 1)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set new cert's version (%d)", error));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set new cert's serial number.                                           */
  /***************************************************************************/
  context->new_cert->cert_info->serialNumber = BN_to_ASN1_INTEGER(new_serial,
                                                                  NULL);
  if (context->new_cert->cert_info->serialNumber == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set new cert's serial number"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set new cert's issuer name.                                             */
  /***************************************************************************/
  error = X509_set_issuer_name(context->new_cert,
                               X509_get_subject_name(context->ca_cert));
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set new cert's issuer name"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the new cert's notBefore date/time to now.  We don't need the rc.   */
  /***************************************************************************/
  X509_gmtime_adj(X509_get_notBefore(context->new_cert), 0);

  /***************************************************************************/
  /* Set the new cert's duration of validity.  We don't need the rc.         */
  /***************************************************************************/
  TRC_DETAIL((TB, "Validity period: %u days", ccm.ca.valid_duration));
  X509_gmtime_adj(X509_get_notAfter(context->new_cert),
                  (long)(60 * 60 * 24 * ccm.ca.valid_duration));

  /***************************************************************************/
  /* Set the new cert's subject name.                                        */
  /***************************************************************************/
  error = X509_set_subject_name(context->new_cert, subject);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set new cert's subject name"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the new cert's public key.                                          */
  /***************************************************************************/
  error = X509_set_pubkey(context->new_cert, context->pub_key);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set new cert's public key"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Implement the extension policy defined in the config file.  See OpenSSL */
  /* documentation and CA utility for details on this.                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Setting extensions"));
  X509V3_set_ctx(&ext_ctx,
                 context->ca_cert,
                 context->new_cert,
                 context->req,
                 NULL,
                 0);

  X509V3_set_conf_lhash(&ext_ctx, ccm.config);

  error = X509V3_EXT_add_conf(ccm.config,
                              &ext_ctx,
                              (char *)ccm_cfg_group_extensions,
                              context->new_cert);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to apply extension policy from config"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up the right digest algorithm.  This is hard-coded to EVP_dss1      */
  /* (==SHA-1) for DSA keys as SHA-1 is the only digest that is compatible   */
  /* with them.  For other (RSA) keys, the digest was determined at startup  */
  /* from the configuration file.                                            */
  /***************************************************************************/
  if (priv_key->type == EVP_PKEY_DSA)
  {
    TRC_DETAIL((TB, "DSA CA key: signing with SHA-1 digest (EVP_dss1)"));
    sign_digest = EVP_dss1();
  }
  else
  {
    TRC_DETAIL((TB, "RSA CA key: signing with %s digest",
                                    OBJ_nid2sn(ccm.ca.rsa_sign_digest->type)));
    sign_digest = ccm.ca.rsa_sign_digest;
  }

  /***************************************************************************/
  /* Now sign the new certificate.                                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Signing new certificate"));
  error = X509_sign(context->new_cert, priv_key, sign_digest);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to sign new certificate"));
    goto EXIT_LABEL;
  }

#ifdef CCM_DEBUG_X509
  /***************************************************************************/
  /* Get OpenSSL to print the cert details to stdout in friendly format.     */
  /***************************************************************************/
  printf("Newly-issued certificate:\n");
  X509_print_fp(stdout, context->new_cert);
#endif

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* Calculate and trace MD5 and SHA-1 fingerprints.                         */
  /***************************************************************************/
  ccm_show_fingerprints(context->new_cert);
#endif

  /***************************************************************************/
  /* Convert new certificate to ASN.1 DER-encoded form.                      */
  /***************************************************************************/
  x509_asn1_len = i2d_X509(context->new_cert, NULL);
  mem = (unsigned char *)rm_get_mem(x509_asn1_len);
  if (mem == NULL)
  {
    TRC_ERROR((TB, "Couldn't get %u bytes to convert new cert to ASN.1",
                                                               x509_asn1_len));
    goto EXIT_LABEL;
  }

  x509_asn1 = (char *)mem;
  x509_asn1_len = i2d_X509(context->new_cert, &mem);
  TRC_DATA("New certificate in ASN.1 form:", x509_asn1, x509_asn1_len);

  /***************************************************************************/
  /* Store the cert in the LDAP directory.                                   */
  /***************************************************************************/
  TRC_NORMAL((TB, "Storing new certificate in LDAP directory"));
  int_rc = ccm_store_new_cert(context->new_cert,
                                 x509_asn1, x509_asn1_len, context->subj_name);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to store new certificate (%u)", int_rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Created and stored new cert for '%s'", context->subj_name));
  ok = TRUE;

EXIT_LABEL:

  /***************************************************************************/
  /* Call the callback with the appropriate params for success or failure.   */
  /***************************************************************************/
  if (ok)
  {
    TRC_NORMAL((TB, "Calling callback (corr=%p) with X.509 at %p, length %u",
                                     context->corr, x509_asn1, x509_asn1_len));
    context->callback(context->corr,
                      x509_asn1,
                      x509_asn1_len);
  }
  else
  {
    TRC_NORMAL((TB, "Calling callback (corr=%p) with null parameters",
                                                               context->corr));
    context->callback(context->corr,
                      NULL,
                      0);
  }

  /***************************************************************************/
  /* Clean up in all cases.                                                  */
  /* Don't free: priv_key    -- points to global key table                   */
  /*             sign_digest -- points to static OpenSSL data                */
  /***************************************************************************/
  if (subject != NULL)
  {
    X509_NAME_free(subject);
  }

  if (new_serial != NULL)
  {
    BN_free(new_serial);
  }

  if (x509_asn1 != NULL)
  {
    rm_release_mem(x509_asn1);
  }

  /***************************************************************************/
  /* The request context has now come to the end of its life.                */
  /***************************************************************************/
  ccm_free_req_context(context);

  /***************************************************************************/
  /* Clean up OpenSSL per-thread state.                                      */
  /***************************************************************************/
  ERR_remove_state(0);

  /***************************************************************************/
  /* Tell RM we're exiting.  We ignore the return code.                      */
  /***************************************************************************/
  rm_notify_thread_exit(RM_CCM_THREAD);

  TRC_THREAD_EXIT();

  return(NULL);

} /* ccm_cert_request_thread */

/**PROC+**********************************************************************/
/* Name:      ccm_free_cert_chain                                            */
/*                                                                           */
/* Purpose:   Free the supplied supplied chain                               */
/*                                                                           */
/* Returns:   void                                                           */
/*                                                                           */
/* Params:    IN     chain        - chain to free                            */
/*                                                                           */
/**PROC-**********************************************************************/

void ccm_free_cert_chain(OSP_CERTIFICATE *chain)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_CERTIFICATE *cert, *next;

  TRC_ENTRY("ccm_free_cert_chain");

  /***************************************************************************/
  /* Walk the list and free the individual entries                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Walking list..."));
  for (cert = chain; cert != NULL; cert = next)
  {
    TRC_DETAIL((TB, "Freeing cert %p", cert));
    next = cert->next;
    rm_release_mem(cert->encoding);
    rm_release_mem(cert->cert_data);
    rm_release_mem(cert);
  }

  TRC_EXIT();

  return;

} /* ccm_free_cert_chain */
