
/**MOD+***********************************************************************/
/* Module:    ccmcaut.c                                                      */
/*                                                                           */
/* Purpose:   CCM CA-related utility functions                               */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.2                $ $Modtime::   Jan 11 2001 15:56:30   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_expand_subject_name                                        */
/*                                                                           */
/* Purpose:   Add the configured suffix to a request subject name            */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     input_name     - the subject name to expand             */
/*            OUT    output_name    - the expanded subject name              */
/*                                      (caller must free)                   */
/*                                                                           */
/* Operation: If ccm.ca.subject_suffix is non-null, add its name components  */
/*            to the name object supplied.                                   */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_expand_subject_name(X509_NAME *input_name,
                               X509_NAME **output_name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509_NAME_ENTRY *entry = NULL;
  int last_set_num = -1;
  int input_name_entries;
  int set;
  int error;
  int ii;

  TRC_ENTRY("ccm_expand_subject_name");

  /***************************************************************************/
  /* Check paramters.                                                        */
  /***************************************************************************/
  TRC_ASSERT((output_name != NULL), (TB, "Null output name supplied"));

  /***************************************************************************/
  /* The top hierarchical level comes first in an X509_NAME, so it's easiest */
  /* to duplicate our suffix and add the entries from the input_name to it.  */
  /***************************************************************************/
  if (ccm.ca.subject_suffix != NULL)
  {
    *output_name = X509_NAME_dup(ccm.ca.subject_suffix);
  }
  else
  {
    *output_name = X509_NAME_new();
  }

  if (*output_name == NULL)
  {
    TRC_ERROR((TB, "Failed to create new subject name object"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  input_name_entries = X509_NAME_entry_count(input_name);
  for (ii = 0; ii < input_name_entries; ii++)
  {
    entry = X509_NAME_get_entry(input_name, ii);

    set = (last_set_num == entry->set) ? -1 : 0;
    last_set_num = entry->set;

    error = X509_NAME_add_entry(*output_name, entry, -1, set);
    if (error == 0)
    {
      TRC_ERROR((TB, "Failed to add name entry at %p to name at %p",
                                                         entry, *output_name));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }
  }

  TRC_DETAIL((TB, "Expanded subject name from %d to %d entries",
                     input_name_entries, X509_NAME_entry_count(*output_name)));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up on failure.                                                    */
  /***************************************************************************/
  if ((rc != OSP_SUCCESS) && (*output_name != NULL))
  {
    X509_NAME_free(*output_name);
    *output_name = NULL;
  }

  TRC_EXIT();

  return(rc);

} /* ccm_expand_subject_name */

/**PROC+**********************************************************************/
/* Name:      ccm_free_req_context                                           */
/*                                                                           */
/* Purpose:   Free all memory associated with a new cert request context     */
/*                                                                           */
/* Returns:   none                                                           */
/*                                                                           */
/* Params:    IN     context        - the context to free                    */
/*                                                                           */
/**PROC-**********************************************************************/

void ccm_free_req_context(CCM_NEW_CERT_CONTEXT *context)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("ccm_free_req_context");

  if (context != NULL)
  {
    /*************************************************************************/
    /* Free the context members that were dynamically allocated, then free   */
    /* the context itself.                                                   */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing cert request context at %p", context));
    if (context->req != NULL)
    {
      TRC_DETAIL((TB, "Freeing context->req"));
      X509_REQ_free(context->req);
    }

    if (context->pub_key != NULL)
    {
      TRC_DETAIL((TB, "Freeing context->pub_key"));
      EVP_PKEY_free(context->pub_key);
    }

    if (context->ca_name != NULL)
    {
      TRC_DETAIL((TB, "Freeing context->ca_name"));
      rm_release_mem(context->ca_name);
    }

    if (context->ca_cert != NULL)
    {
      TRC_DETAIL((TB, "Freeing context->ca_cert"));
      X509_free(context->ca_cert);
    }

    if (context->subj_name != NULL)
    {
      TRC_DETAIL((TB, "Freeing context->subj_name"));
      rm_release_mem(context->subj_name);
    }

    if (context->new_cert != NULL)
    {
      TRC_DETAIL((TB, "Freeing context->new_cert"));
      X509_free(context->new_cert);
    }

    TRC_DETAIL((TB, "Freeing context memory"));
    rm_release_mem(context);

    TRC_DETAIL((TB, "Freed cert request context at %p", context));
  }

  TRC_EXIT();

  return;

} /* ccm_free_req_context */

/**PROC+**********************************************************************/
/* Name:      ccm_read_serial_number                                         */
/*                                                                           */
/* Purpose:   Read the latest serial number from the configured file         */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Open the file, read the value and store it in the CCM data.    */
/*            Based on load_serial() in OpenSSL's CA utility.                */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_read_serial_number(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  BIO *in_bio = NULL;
  ASN1_INTEGER *asn1_int = NULL;
  char buf[CCM_SERIAL_READ_BUF_SIZE];
  int result;
  long lresult;

  TRC_ENTRY("ccm_read_serial_number");

  in_bio = BIO_new(BIO_s_file());
  if (in_bio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create BIO for loading serial number"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  lresult = BIO_read_filename(in_bio, ccm.ca.serial_file);
  if (lresult <= 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set BIO to read (%ld) from file '%s'",
                                                 lresult, ccm.ca.serial_file));
    rc = OSP_ERR_FILE;
    goto EXIT_LABEL;
  }

  asn1_int = ASN1_INTEGER_new();
  if (asn1_int == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create ASN1_INTEGER for serial number"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  result = a2i_ASN1_INTEGER(in_bio, asn1_int, buf, CCM_SERIAL_READ_BUF_SIZE);
  if (result == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to read serial number from '%s'",
                                                          ccm.ca.serial_file));
    rc = OSP_ERR_FILE;
    goto EXIT_LABEL;
  }

  ccm.ca.serial_num = ASN1_INTEGER_to_BN(asn1_int, NULL);
  if (ccm.ca.serial_num == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Error converting serial number to BIGNUM"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Read and converted serial number ok (from '%s')",
                                                          ccm.ca.serial_file));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (in_bio != NULL)
  {
    BIO_free(in_bio);
  }

  if (asn1_int != NULL)
  {
    ASN1_INTEGER_free(asn1_int);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_read_serial_number */

/**PROC+**********************************************************************/
/* Name:      ccm_update_serial_number                                       */
/*                                                                           */
/* Purpose:   Write a new serial number to the configuration file            */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Open the file and write the value. Based on save_serial() in   */
/*            OpenSSL's CA utility.                                          */
/*            Caller must hold ccm.ca.serial_num_mutex before calling this.  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_update_serial_number(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  BIO *out_bio = NULL;
  ASN1_INTEGER *asn1_int = NULL;
  int result;
  long lresult;

  TRC_ENTRY("ccm_update_serial_number");

  out_bio = BIO_new(BIO_s_file());
  if (out_bio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create BIO for saving serial number"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  lresult = BIO_write_filename(out_bio, ccm.ca.serial_file);
  if (lresult <= 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set BIO to write (%ld) to file '%s'",
                                                 lresult, ccm.ca.serial_file));
    rc = OSP_ERR_FILE;
    goto EXIT_LABEL;
  }

  asn1_int = BN_to_ASN1_INTEGER(ccm.ca.serial_num, NULL);
  if (asn1_int == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to convert BIGNUM to ASN1_INTEGER"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  result = i2a_ASN1_INTEGER(out_bio, asn1_int);
  if (result < 0)
  {
    OSP_OPENSSL_ERROR((TB, "Error saving serial number to file (%d)", result));
    rc = OSP_ERR_FILE;
    goto EXIT_LABEL;
  }

  BIO_puts(out_bio, "\n");

  TRC_NORMAL((TB, "Wrote serial number OK (to file '%s)", ccm.ca.serial_file));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (out_bio != NULL)
  {
    BIO_free(out_bio);
  }

  if (asn1_int != NULL)
  {
    ASN1_INTEGER_free(asn1_int);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_update_serial_number */

/**PROC+**********************************************************************/
/* Name:      ccm_show_fingerprints                                          */
/*                                                                           */
/* Purpose:   Calculate and trace certificate fingerprints                   */
/*                                                                           */
/* Returns:   None.                                                          */
/*                                                                           */
/* Params:    IN     cert           - the certificate on which to operate    */
/*                                                                           */
/* Operation: Calculate both MD5 and SHA-1 fingerprints and trace them.      */
/*                                                                           */
/**PROC-**********************************************************************/

void ccm_show_fingerprints(X509 *cert)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  unsigned char fingerprint[EVP_MAX_MD_SIZE];
  unsigned int fingerprint_len = 0;
  int error;

  TRC_ENTRY("ccm_show_fingerprints");

  error = X509_digest(cert, EVP_md5(), fingerprint, &fingerprint_len);
  if (error == 0)
  {
    TRC_ERROR((TB, "Failed to calculate MD5 certificate fingerprint"));
  }
  else
  {
    TRC_DATA("MD5 fingerprint:", fingerprint, fingerprint_len);
  }

  error = X509_digest(cert, EVP_sha1(), fingerprint, &fingerprint_len);
  if (error == 0)
  {
    TRC_ERROR((TB, "Failed to calculate SHA-1 certificate fingerprint"));
  }
  else
  {
    TRC_DATA("SHA-1 fingerprint:", fingerprint, fingerprint_len);
  }

  TRC_EXIT();

  return;

} /* ccm_show_fingerprints */

