
/**MOD+***********************************************************************/
/* Module:    ccmconf.c                                                      */
/*                                                                           */
/* Purpose:   CCM configuration functions                                    */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.6                $ $Modtime::   Jan 11 2001 15:56:30   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_load_config                                                */
/*                                                                           */
/* Purpose:   Load the configuration file and get the required values into   */
/*            CCM's global data structure.                                   */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Use OpenSSL CONF library for config file handling. The CONF    */
/*            object is not freed until ccm_term() because it contains the   */
/*            certificate extensions that should be added to each new cert.  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_load_config(void)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  char *file_name = NULL;
  FILE *pk_file = NULL;
  char *value = NULL;
  char *value2 = NULL;
  char *token = NULL;
  char *strtok_temp = NULL;
  long error_line;
  int result;
  char id_name[CCM_CFG_ID_VALUE_MAXLEN];
  int ii, jj;
  char *colon;

  TRC_ENTRY("ccm_load_config");

  file_name = osp_common_get_config_filename();
  if (file_name == NULL)
  {
    /*************************************************************************/
    /* No config file configured in the environment.                         */
    /*************************************************************************/
    TRC_ERROR((TB, "Can't get config file name"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Attempting to read config from %s", file_name));

  /***************************************************************************/
  /* Use OpenSSL's config file handling routines to read in the data.        */
  /***************************************************************************/
  ccm.config = CONF_load(NULL, file_name, &error_line);
  if (ccm.config == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Error loading config file %s on line %ld",
                                                       file_name, error_line));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /*                                                                         */
  /* Read our general configuration info.                                    */
  /*                                                                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Reading '[ %s ]' for general config:",
                                                      OSP_CONFIG_STACK_GROUP));

  /***************************************************************************/
  /* Read the name of the random seed file.                                  */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_rand_file);
  if (value == NULL)
  {
    TRC_ERROR((TB, "%s\\%s is required, but was not found",
                                   OSP_CONFIG_STACK_GROUP, ccm_cfg_rand_file));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_rand_file, value));

  OSP_STRDUP(ccm.rand_file, value);
  if (ccm.rand_file == NULL)
  {
    TRC_ERROR((TB, "Failed to get %d bytes for random seed file name",
                                                               strlen(value)));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /*                                                                         */
  /* Read our CA configuration info.                                         */
  /*                                                                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Reading '[ %s ]' for CA config:", OSP_CONFIG_STACK_GROUP));

  /***************************************************************************/
  /* Read the name of the serial number file.                                */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ca_serial_file);
  if (value == NULL)
  {
    TRC_ERROR((TB, "%s\\%s is required, but was not found",
                              OSP_CONFIG_STACK_GROUP, ccm_cfg_ca_serial_file));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ca_serial_file, value));

  OSP_STRDUP(ccm.ca.serial_file, value);
  if (ccm.ca.serial_file == NULL)
  {
    TRC_ERROR((TB, "Failed to get %d bytes for CA serial number file name",
                                                               strlen(value)));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read the latest serial number from the file.                            */
  /***************************************************************************/
  rc = ccm_read_serial_number();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to read serial number (%u)", rc));
    rc = OSP_ERR_FILE;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read the duration of validity that we should set for new certificates.  */
  /* This is in days.                                                        */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ca_duration);
  if (value == NULL)
  {
    TRC_ERROR((TB, "%s\\%s is required, but was not found",
                                 OSP_CONFIG_STACK_GROUP, ccm_cfg_ca_duration));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ca_duration, value));

  result = atoi(value);
  if (result <= 0)
  {
    TRC_ERROR((TB, "Invalid duration value (%d)", result));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }
  ccm.ca.valid_duration = (unsigned int)result;

  /***************************************************************************/
  /* Read the signing digest to use when signing RSA certificates.           */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ca_rsa_sign_digest);
  if (value != NULL)
  {
    result = OBJ_txt2nid(value);
    if (result == NID_sha1)
    {
      TRC_NORMAL((TB, "'%s' is SHA-1", ccm_cfg_ca_rsa_sign_digest));
      ccm.ca.rsa_sign_digest = EVP_sha1();
    }
    else if (result == NID_md5)
    {
      TRC_NORMAL((TB, "'%s' is MD5", ccm_cfg_ca_rsa_sign_digest));
      ccm.ca.rsa_sign_digest = EVP_md5();
    }
    else
    {
      TRC_ERROR((TB, "Invalid config value '%s' for %s\\%s",
                   value, OSP_CONFIG_STACK_GROUP, ccm_cfg_ca_rsa_sign_digest));
      rc = OSP_ERR_CONFIG;
      goto EXIT_LABEL;
    }
  }
  else
  {
    TRC_NORMAL((TB, "'%s' defaulting to MD5", ccm_cfg_ca_rsa_sign_digest));
    ccm.ca.rsa_sign_digest = EVP_md5();
  }

  /***************************************************************************/
  /* Read the suffix to add to certificate request subject names.  Note that */
  /* we always create a name suffix object; it may or may not have entries.  */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ca_subject_suffix);
  if ((value != NULL) && (strlen(value) > 0))
  {
    TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ca_subject_suffix, value));

    rc = ccm_rfc2253_to_x509_name(value, &ccm.ca.subject_suffix);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to store subject name suffix (%u)", rc));
      goto EXIT_LABEL;
    }
  }
  else
  {
    TRC_NORMAL((TB, "'%s' defaulting to <none>", ccm_cfg_ca_subject_suffix));
    ccm.ca.subject_suffix = NULL;
  }

  /***************************************************************************/
  /*                                                                         */
  /* Read the access information for the LDAP certificate repository.        */
  /*                                                                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Reading '[ %s ]' for LDAP directory config:",
                                                      OSP_CONFIG_STACK_GROUP));

  /***************************************************************************/
  /* Read the LDAP directory host name.                                      */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ldap_host);
  if (value == NULL)
  {
    TRC_ERROR((TB, "%s\\%s is required, but was not found",
                                   OSP_CONFIG_STACK_GROUP, ccm_cfg_ldap_host));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ldap_host, value));

  OSP_STRDUP(ccm.ldap.host, value);
  if (ccm.ldap.host == NULL)
  {
    TRC_ERROR((TB, "Failed to get %d bytes for LDAP hostname", strlen(value)));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Extract optional port number from end of hostname string.               */
  /***************************************************************************/
  colon = strchr(ccm.ldap.host, ':');
  if (colon == NULL)
  {
    ccm.ldap.port = LDAP_PORT;
  }
  else
  {
    /*************************************************************************/
    /* Scan the port number from the part of the string after the colon.     */
    /*************************************************************************/
    if (1 != sscanf(colon + 1, "%d", &ccm.ldap.port))
    {
      TRC_ERROR((TB, "Invalid host:port configuration value: %s",
                 ccm.ldap.host));
      rc = OSP_ERR_CONFIG;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Null-terminate the host name part of the string.                      */
    /*************************************************************************/
    *colon = 0;
  }
  TRC_DETAIL((TB, "Using LDAP port number: %d", ccm.ldap.port));

  /***************************************************************************/
  /* Read the user name that we need for LDAP access.                        */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ldap_user);
  if ((value != NULL) && (strlen(value) > 0))
  {
    TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ldap_user, value));
  }
  else
  {
    TRC_NORMAL((TB, "'%s' defaulting to '%s'", ccm_cfg_ldap_user,
                                                      ccm_cfg_ldap_user_dflt));
    value = (char *)ccm_cfg_ldap_user_dflt;
  }

  OSP_STRDUP(ccm.ldap.user, value);
  if (ccm.ldap.user == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for LDAP username"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read the password that we need for LDAP access.                         */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ldap_password);
  if ((value != NULL) && (strlen(value) > 0))
  {
    TRC_NORMAL((TB, "Read '%s' OK", ccm_cfg_ldap_password));
  }
  else
  {
    TRC_NORMAL((TB, "'%s' using default", ccm_cfg_ldap_password));
    value = (char *)ccm_cfg_ldap_password_dflt;
  }

  OSP_STRDUP(ccm.ldap.password, value);
  if (ccm.ldap.password == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for LDAP password"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read the suffix that we add to DNs before doing an LDAP search.         */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ldap_dn_suffix);
  if ((value != NULL) && (strlen(value) > 0))
  {
    TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ldap_dn_suffix, value));

    OSP_STRDUP(ccm.ldap.dn_suffix, value);
    if (ccm.ldap.dn_suffix == NULL)
    {
      TRC_ERROR((TB, "Failed to get memory for LDAP DN suffix"));
      rc = OSP_ERR_MEMORY;
      goto EXIT_LABEL;
    }
  }
  else
  {
    TRC_NORMAL((TB, "'%s' defaulting to <none>", ccm_cfg_ldap_dn_suffix));
    ccm.ldap.dn_suffix = NULL;
  }

  /***************************************************************************/
  /*                                                                         */
  /* Read the SSL group.                                                     */
  /*                                                                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Reading '[ %s ]' for SSL config:", OSP_CONFIG_STACK_GROUP));

  /***************************************************************************/
  /* Read the SSL identity.  This is a required setting.                     */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ssl_identity);
  if (value == NULL)
  {
    TRC_ERROR((TB, "%s\\%s is required, but was not found",
                                OSP_CONFIG_STACK_GROUP, ccm_cfg_ssl_identity));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ssl_identity, value));
  OSP_STRDUP(ccm.ssl.identity, value);
  if (ccm.ssl.identity == NULL)
  {
    TRC_ERROR((TB, "Failed to get %d bytes for SSL identity", strlen(value)));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read the SSL verify mode.                                               */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ssl_verify_client);
  if (value != NULL)
  {
    TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ssl_verify_client, value));
    if (strcasecmp(value, ccm_cfg_vfy_no) == 0)
    {
      TRC_NORMAL((TB, "Client verification turned off"));
      ccm.ssl.verify_mode = SSL_VERIFY_NONE;
    }
    else if (strncasecmp(value, ccm_cfg_vfy_yes, strlen(ccm_cfg_vfy_yes)) == 0)
    {
      /***********************************************************************/
      /* We (hopefully) have either a plain 'yes' or a string like 'yes:     */
      /* once, require'.  We need to work through everything after the 'yes' */
      /* and apply the appropriate options.                                  */
      /***********************************************************************/
      TRC_NORMAL((TB, "Client verification turned on; parsing options"));
      ccm.ssl.verify_mode = SSL_VERIFY_PEER;
      token = strtok_r(value + strlen(ccm_cfg_vfy_yes),
                                         ccm_cfg_vfy_separators, &strtok_temp);
      if (token == NULL)
      {
        TRC_NORMAL((TB, "No options specified"));
      }

      while (token != NULL)
      {
        TRC_DETAIL((TB, "Got token '%s'", token));
        if (strcasecmp(token, ccm_cfg_vfy_once) == 0)
        {
          TRC_NORMAL((TB, "Turning on SSL_VERIFY_CLIENT_ONCE"));
          ccm.ssl.verify_mode |= SSL_VERIFY_CLIENT_ONCE;
        }
        else if (strcasecmp(token, ccm_cfg_vfy_require) == 0)
        {
          TRC_NORMAL((TB, "Turning on SSL_VERIFY_FAIL_IF_NO_PEER_CERT"));
          ccm.ssl.verify_mode |= SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
        }
        else
        {
          TRC_ERROR((TB, "%s\\%s option '%s' is not valid",
                    OSP_CONFIG_STACK_GROUP, ccm_cfg_ssl_verify_client, token));
          rc = OSP_ERR_CONFIG;
          goto EXIT_LABEL;
        }

        token = strtok_r(NULL, ccm_cfg_vfy_separators, &strtok_temp);
      }
    }
    else
    {
      TRC_ERROR((TB, "%s\\%s value of '%s' is not valid",
                    OSP_CONFIG_STACK_GROUP, ccm_cfg_ssl_verify_client, value));
      rc = OSP_ERR_CONFIG;
      goto EXIT_LABEL;
    }
  }
  else
  {
    TRC_NORMAL((TB, "'%s' defaulting to 0x%x",
                   ccm_cfg_ssl_verify_client, CCM_CFG_SSL_VERIFY_CLIENT_DFLT));
    ccm.ssl.verify_mode = CCM_CFG_SSL_VERIFY_CLIENT_DFLT;
  }

  /***************************************************************************/
  /* Read the SSL cipher list.                                               */
  /***************************************************************************/
  value = CONF_get_string(ccm.config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)ccm_cfg_ssl_cipher_list);
  if ((value != NULL) && (strlen(value) > 0))
  {
    TRC_NORMAL((TB, "'%s' is '%s'", ccm_cfg_ssl_cipher_list, value));

  }
  else
  {
    TRC_NORMAL((TB, "'%s' defaulting to '%s'", ccm_cfg_ssl_cipher_list,
                                                ccm_cfg_ssl_cipher_list_dflt));
    value = (char *)ccm_cfg_ssl_cipher_list_dflt;
  }

  OSP_STRDUP(ccm.ssl.cipher_list, value);
  if (ccm.ssl.cipher_list == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for SSL cipher list"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /*                                                                         */
  /* Read the list of identities.                                            */
  /*                                                                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Reading '[ %s ]' for identity list:",
                                                    ccm_cfg_group_identities));

  /***************************************************************************/
  /* We do this twice: the first time to count up the number that are        */
  /* present, the second time to store the identity data.                    */
  /***************************************************************************/
  for (ii = 1; ii <= 2; ii++)
  {
    jj = 0;
    while (TRUE)
    {
      TRC_ASSERT(((strlen(ccm_cfg_identities_id_dn) + CCM_MAX_INT_DIGITS) <=
                                                      CCM_CFG_ID_VALUE_MAXLEN),
                             (TB, "Need to increase CCM_CFG_ID_VALUE_MAXLEN"));
      sprintf(id_name, ccm_cfg_identities_id_dn, jj);
      value = CONF_get_string(ccm.config,
                              (char *)ccm_cfg_group_identities,
                              id_name);

      TRC_ASSERT(((strlen(ccm_cfg_identities_id_pkey) + CCM_MAX_INT_DIGITS) <=
                                                     CCM_CFG_ID_VALUE_MAXLEN),
                             (TB, "Need to increase CCM_CFG_ID_VALUE_MAXLEN"));
      sprintf(id_name, ccm_cfg_identities_id_pkey, jj);
      value2 = CONF_get_string(ccm.config,
                               (char *)ccm_cfg_group_identities,
                               id_name);

      if ((value == NULL) && (value2 == NULL))
      {
        /*********************************************************************/
        /* We haven't found anything for this identity.  If it's the first,  */
        /* that's an error.  If not, we just stop looking.                   */
        /*********************************************************************/
        if (jj == 0)
        {
          TRC_ERROR((TB, "%s and %s are both required",
                ccm_cfg_identities_id_dn, ccm_cfg_identities_id_pkey, jj, jj));
          rc = OSP_ERR_CONFIG;
          goto EXIT_LABEL;
        }

        TRC_DETAIL((TB, "Nothing found for ID #%d; stopping search", jj));
        break;
      }

      if ((value == NULL) || (value2 == NULL))
      {
        /*********************************************************************/
        /* Either the name or the private key wasn't found: that's an error. */
        /*********************************************************************/
        TRC_ERROR((TB, "Incomplete info for ID #%d", jj));
        rc = OSP_ERR_CONFIG;
        goto EXIT_LABEL;
      }

      if (ii == 2)
      {
        /*********************************************************************/
        /* This is the second time through: first store the DN we've got.    */
        /*********************************************************************/
        TRC_NORMAL((TB, "ID #%d: storing name '%s'", jj, value));
        rc = ccm_rfc2253_to_x509_name(value, &(ccm.identities.names[jj]));
        if (rc != OSP_SUCCESS)
        {
          TRC_ERROR((TB, "Failed to store X509_NAME for ID #%d (%u)", jj, rc));
          goto EXIT_LABEL;
        }

        /*********************************************************************/
        /* Attempt to load the private key using the specified file name.    */
        /*********************************************************************/
        TRC_NORMAL((TB, "ID #%d: reading private key from '%s'", jj, value2));
        pk_file = fopen(value2, "r");
        if (pk_file == NULL)
        {
          TRC_ERROR((TB, "Failed to open private key file '%s'", value2));
          rc = OSP_ERR_FILE;
          goto EXIT_LABEL;
        }

        /*********************************************************************/
        /* PEM_read_PrivateKey allows us to specify a customized password    */
        /* callback (3rd param) but we currently allow OpenSSL to use its    */
        /* default (internal) callback, which prompts for any password via   */
        /* stdout/stdin.                                                     */
        /*********************************************************************/
        ccm.identities.privkeys[jj] = PEM_read_PrivateKey(pk_file,
                                                          NULL,
                                                          ccm_passphrase_cb,
                                                          value);
        if (ccm.identities.privkeys[jj] == NULL)
        {
          OSP_OPENSSL_ERROR((TB, "Failed to create EVP_PKEY for ID #%d "
                                               "from file '%s'", jj, pk_file));
          rc = OSP_ERR_NO_RESOURCES;
          goto EXIT_LABEL;
        }

        /*********************************************************************/
        /* Check that we have a key type that we recognize.                  */
        /*********************************************************************/
        if ((ccm.identities.privkeys[jj]->type != EVP_PKEY_RSA) &&
            (ccm.identities.privkeys[jj]->type != EVP_PKEY_DSA))
        {
          TRC_ERROR((TB, "Unrecognized private key type for ID #%d ", jj));
          rc = OSP_ERR_CONFIG;
          goto EXIT_LABEL;
        }

        if (fclose(pk_file) != 0)
        {
          TRC_ERROR((TB, "Failed to close private key file '%s'", value2));
          /*******************************************************************/
          /* Don't exit -- this isn't fatal.                                 */
          /*******************************************************************/
        }
        pk_file = NULL;
      }

      jj++;
    }

    if (ii == 1)
    {
      /***********************************************************************/
      /* This is the first time through: allocate the DN/key indices.  We    */
      /* make sure that they are zero-initialized here so that the freeing   */
      /* code in ccm_term works properly in the event of a malloc failure.   */
      /***********************************************************************/
      ccm.identities.num = jj;

      ccm.identities.names =
                            (X509_NAME **)rm_get_mem(jj * sizeof(X509_NAME *));
      if (ccm.identities.names == NULL)
      {
        TRC_ERROR((TB, "Failed to alloc memory for DN index; %d entries", jj));
        rc = OSP_ERR_MEMORY;
        goto EXIT_LABEL;
      }
      memset(ccm.identities.names, 0, jj * sizeof(X509_NAME *));

      ccm.identities.privkeys =
                             (EVP_PKEY **)rm_get_mem(jj * sizeof(EVP_PKEY *));
      if (ccm.identities.privkeys == NULL)
      {
        TRC_ERROR((TB, "Failed to allocate memory for private key index"));
        rc = OSP_ERR_MEMORY;
        goto EXIT_LABEL;
      }
      memset(ccm.identities.privkeys, 0, jj * sizeof(EVP_PKEY *));

      TRC_NORMAL((TB, "Allocated indices for %d identities", jj));
    }
  }

  TRC_NORMAL((TB, "Read configuration OK (from %s)", file_name));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (pk_file != NULL)
  {
    TRC_NORMAL((TB, "Closing private key file %p", pk_file));
    fclose(pk_file);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_load_config */

/**PROC+**********************************************************************/
/* Name:      ccm_passphrase_cb                                              */
/*                                                                           */
/* Purpose:   Display helpful "enter your passphrase" prompt when an         */
/*            encrypted private key is in use                                */
/*                                                                           */
/* Returns:   number of chars in the passphrase, or -1 if an error occurred  */
/*                                                                           */
/* Params:    OUT    passphrase   - returned passphrase                      */
/*            IN     max_len      - maximum length of passphrase             */
/*            IN     verify       - TRUE if passphrase should be verified    */
/*            IN     userdata     - user data supplied to call back          */
/*                                                                           */
/**PROC-**********************************************************************/

static int ccm_passphrase_cb(char *passphrase,
                             int max_len,
                             int verify,
                             void *userdata)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int pp_len = -1;
  int prompt_len;
  char *dn;
  char *prompt;
  int read_rc;

  TRC_ENTRY("ccm_passphrase_cb");

  /***************************************************************************/
  /* The userdata carries the DN for which we want the passphrase            */
  /***************************************************************************/
  dn = (char *)userdata;
  TRC_ASSERT((dn != NULL), (TB, "Passed NULL DN"));
  TRC_DETAIL((TB, "Getting pass phrase for dn '%s'", dn));

  /***************************************************************************/
  /* Now build the prompt using it                                           */
  /***************************************************************************/
  prompt_len = strlen(ccm_pp_prompt) + strlen(dn);
  TRC_DETAIL((TB, "Getting %d bytes for prompt", prompt_len));
  prompt = rm_get_mem(prompt_len);
  if (prompt == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate memory for prompt"));
    goto EXIT_LABEL;
  }
  OSP_SNPRINTF((prompt, prompt_len, ccm_pp_prompt, dn));
  TRC_DETAIL((TB, "Using prompt '%s'", prompt));

  /***************************************************************************/
  /* Get the passphrase                                                      */
  /***************************************************************************/
  read_rc = EVP_read_pw_string(passphrase, max_len, prompt, verify);
  if (read_rc != 0)
  {
    TRC_ERROR((TB, "Problems getting passphrase"));
    PEMerr(PEM_F_DEF_CALLBACK, PEM_R_PROBLEMS_GETTING_PASSWORD);
    memset(passphrase, 0, (unsigned int)max_len);
  }
  else
  {
    TRC_DETAIL((TB, "Got passphrase OK"));
    pp_len = strlen(passphrase);
  }

EXIT_LABEL:

  TRC_EXIT();

  return(pp_len);

} /* ccm_passphrase_cb */
