
/**MOD+***********************************************************************/
/* Module:    ccmldap.c                                                      */
/*                                                                           */
/* Purpose:   CCM LDAP functions                                             */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.26               $ $Modtime::   Jan 11 2001 15:56:32   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_lookup_ldap_by_subject                                     */
/*                                                                           */
/* Purpose:   Perform certificate or CRL lookup by subject via LDAP          */
/*                                                                           */
/* Returns:   1           - success                                          */
/*            0           - failure                                          */
/*                                                                           */
/* Params:    IN     xl             - the lookup context                     */
/*                                      (this is not currently used)         */
/*            IN     type           - type of object being looked for        */
/*            IN     x509_name      - the subject name to search for         */
/*            OUT    ret            - the resulting cert or CRL object       */
/*                                                                           */
/* Operation: Use OpenLDAP to do the search synchronously; confirm that the  */
/*            name contained in the returned object matches the name under   */
/*            which it was stored in the directory.                          */
/*                                                                           */
/**PROC-**********************************************************************/

int ccm_lookup_ldap_by_subject(X509_LOOKUP *xl,
                               int type,
                               X509_NAME *x509_name,
                               X509_OBJECT *ret)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  int rc = X509_LU_FAIL;
  OSP_RC int_rc = OSP_ERROR;
  char *name = NULL;
  char *ldap_name = NULL;
  char *attrs[3] = {NULL, NULL, NULL};
  LDAPMessage *search_result = NULL;
  LDAPMessage *result_entry = NULL;
  char *search_filter = NULL;
  int num_entries;
  struct berval **attrib_values = NULL;
  unsigned char *data = NULL;
  long data_len;
  int error;
  char *got_name = NULL;

  TRC_ENTRY("ccm_lookup_ldap_by_subject");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT(((type == X509_LU_CRL) || (type == X509_LU_X509)),
                                   (TB, "Unrecognized lookup type: %d", type));

  /***************************************************************************/
  /* Grab the lock so that the entire lookup operation is atomic with        */
  /* respect to other threads.                                               */
  /***************************************************************************/
  OSP_GET_LOCK(&ccm.ldap.access_mutex);

  /***************************************************************************/
  /* Bind to the LDAP server if necessary.                                   */
  /***************************************************************************/
  int_rc = ccm_ldap_bind();
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Search failed because LDAP bind failed"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Construct the string version of the DN we're looking for.               */
  /***************************************************************************/
  int_rc = ccm_x509_name_to_rfc2253(x509_name, &name);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to convert lookup subject name to string"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add any required DN suffix for the directory lookup.                    */
  /***************************************************************************/
  OSP_STRDUP(ldap_name, name);
  if (ldap_name == NULL)
  {
    TRC_ERROR((TB, "Failed to duplicate name string"));
    goto EXIT_LABEL;
  }

  int_rc = ccm_add_ldap_dn_suffix(&ldap_name);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to expand DN for LDAP search (%u)", int_rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Construct the attributes list that we want to retrieve.                 */
  /***************************************************************************/
  if (type == X509_LU_X509)
  {
    /*************************************************************************/
    /* In this case we may be looking up a cert for an end entity or for a   */
    /* CA; we can't tell.  So we ask for attributes userCertificate and      */
    /* cACertificate and see what comes back.                                */
    /*************************************************************************/
    TRC_NORMAL((TB, "Doing cert lookup for '%s'", ldap_name));
    attrs[0] = (char *)ccm_ldap_attr_usercert;
    attrs[1] = (char *)ccm_ldap_attr_cacert;
    attrs[2] = NULL;
    search_filter = (char *)ccm_ldap_search_cert_filter;
  }
  else
  {
    TRC_NORMAL((TB, "Doing CRL lookup for '%s'", ldap_name));
    attrs[0] = (char *)ccm_ldap_attr_crl;
    attrs[1] = NULL;
    search_filter = (char *)ccm_ldap_search_crl_filter;
  }

  /***************************************************************************/
  /* Now we're ready to do the query.                                        */
  /***************************************************************************/
  error = ldap_search_s(ccm.ldap.handle,
                        ldap_name,
                        LDAP_SCOPE_BASE,
                        search_filter,
                        attrs,
                        0,
                        &search_result);
  if (error != LDAP_SUCCESS)
  {
    if ((error != LDAP_NO_SUCH_ATTRIBUTE) &&
        (error != LDAP_NO_SUCH_OBJECT))
    {
      if (error == LDAP_SERVER_DOWN)
      {
        /*********************************************************************/
        /* If the server has gone down we'll have to unbind, then rebind     */
        /* next time an LDAP operation is attempted.                         */
        /*********************************************************************/
        TRC_ERROR((TB, "LDAP search failed: the server appears to be down"));
        ldap_unbind_s(ccm.ldap.handle);
        ccm.ldap.handle = NULL;
      }
      else
      {
        TRC_ERROR((TB, "Error %d (%s) from ldap_search for %s for '%s'",
                   error, ldap_err2string(error),
                   (type == X509_LU_X509) ? "certificate" : "CRL", ldap_name));
      }
    }
    else
    {
      TRC_ALERT((TB, "Couldn't find %s (%d: %s) via LDAP for '%s'",
                                (type == X509_LU_X509) ? "certificate" : "CRL",
                                error, ldap_err2string(error), ldap_name));
    }
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Check that we actually have a search result.                            */
  /***************************************************************************/
  if (search_result == NULL)
  {
    TRC_ERROR((TB, "No search result structure returned"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We have to extract the result from the LDAPMessage structure and place  */
  /* it into the X509_OBJECT for ccm_lookup_ldap_by_subject() to return.     */
  /***************************************************************************/
  num_entries = ldap_count_entries(ccm.ldap.handle, search_result);
  TRC_NORMAL((TB, "Got %d result entries", num_entries));
  if (num_entries == 0)
  {
    TRC_NORMAL((TB, "No results found"));
    goto EXIT_LABEL;
  }
  else if (num_entries != 1)
  {
    TRC_ERROR((TB, "Expecting only one result entry"));
    goto EXIT_LABEL;
  }
  else
  {
    /*************************************************************************/
    /* We searched at base level on a single DN so there should be only one  */
    /* result entry.  Let's get it and look at its attributes.               */
    /*************************************************************************/
    TRC_NORMAL((TB, "Getting result entry"));
    result_entry = ldap_first_entry(ccm.ldap.handle, search_result);

    if (type == X509_LU_X509)
    {
      /***********************************************************************/
      /* First check for a CA certificate.  If there is one, we'll take it.  */
      /***********************************************************************/
      TRC_NORMAL((TB, "Checking for CA certificate first"));
      attrib_values = ldap_get_values_len(ccm.ldap.handle,
                                          result_entry,
                                          (char *)ccm_ldap_attr_cacert);
      if ((attrib_values != NULL) &&
          (ldap_count_values_len(attrib_values) > 0))
      {
        /*********************************************************************/
        /* Got a CA cert.  Convert to an X509_OBJECT.  We assume that there  */
        /* is only one of them.                                              */
        /*********************************************************************/
        TRC_NORMAL((TB, "Found CA cert; convert to internal format"));
        data = (unsigned char *)attrib_values[0]->bv_val;
        data_len = (long)attrib_values[0]->bv_len;

        ret->data.x509 = d2i_X509(NULL, &data, data_len);
        if (ret->data.x509 == NULL)
        {
          /*******************************************************************/
          /* Couldn't convert the data. Behave as if we didn't find a cert.  */
          /*******************************************************************/
          OSP_OPENSSL_ERROR((TB, "Failed to convert ASN.1 CA cert"));
          goto EXIT_LABEL;
        }
      }
      else
      {
        /*********************************************************************/
        /* There was no CA cert -- look for a user cert.                     */
        /*********************************************************************/
        TRC_NORMAL((TB, "No CA cert; checking for user cert"));
        if (attrib_values != NULL)
        {
          TRC_DETAIL((TB, "Freeing attrib_values"));
          ldap_value_free_len(attrib_values);
        }
        attrib_values = ldap_get_values_len(ccm.ldap.handle,
                                            result_entry,
                                            (char *)ccm_ldap_attr_usercert);
        if ((attrib_values != NULL) &&
            (ldap_count_values_len(attrib_values) > 0))
        {
          /*******************************************************************/
          /* Got a user cert.  Convert to an X509_OBJECT.  We assume that    */
          /* there is only one of them.                                      */
          /*******************************************************************/
          TRC_NORMAL((TB, "Got user cert; convert to internal format"));
          data = (unsigned char *)attrib_values[0]->bv_val;
          data_len = (long)attrib_values[0]->bv_len;

          ret->data.x509 = d2i_X509(NULL, &data, data_len);
          if (ret->data.x509 == NULL)
          {
            /*****************************************************************/
            /* Couldn't convert the data.  Behave as if we didn't find a     */
            /* cert.                                                         */
            /*****************************************************************/
            OSP_OPENSSL_ERROR((TB, "Failed to convert ASN.1 user cert"));
            goto EXIT_LABEL;
          }
        }
        else
        {
          /*******************************************************************/
          /* The entry doesn't have a CA or user certificate.                */
          /*******************************************************************/
          TRC_NORMAL((TB, "No CA or user cert found"));
          goto EXIT_LABEL;
        }
      }

      /***********************************************************************/
      /* Extract the subject name from the certificate that we have found.   */
      /***********************************************************************/
      TRC_DETAIL((TB, "Convert certificate subject name to RFC2253 form"));
      ccm_x509_name_to_rfc2253(X509_get_subject_name(ret->data.x509),
                                                                    &got_name);
      if (got_name == NULL)
      {
        TRC_ERROR((TB, "Failed to convert cert subject to text format"));
        goto EXIT_LABEL;
      }
    }
    else
    {
      /***********************************************************************/
      /* Get the CRL.  We assume there's only one.                           */
      /***********************************************************************/
      TRC_NORMAL((TB, "Checking for CRL"));
      attrib_values = ldap_get_values_len(ccm.ldap.handle,
                                          result_entry,
                                          (char *)ccm_ldap_attr_crl);
      if ((attrib_values != NULL) &&
          (ldap_count_values_len(attrib_values) > 0))
      {
        /*********************************************************************/
        /* Got a CRL.  Convert to an X509_OBJECT.                            */
        /*********************************************************************/
        TRC_NORMAL((TB, "Got CRL; convert to internal format"));
        data = (unsigned char *)attrib_values[0]->bv_val;
        data_len = (long)attrib_values[0]->bv_len;

        ret->data.crl = d2i_X509_CRL(NULL, &data, data_len);
        if (ret->data.crl == NULL)
        {
          /*******************************************************************/
          /* Couldn't convert the data. Behave as if we didn't find a cert.  */
          /*******************************************************************/
          OSP_OPENSSL_ERROR((TB, "Failed to convert ASN.1 CRL"));
          goto EXIT_LABEL;
        }
      }
      else
      {
        /*********************************************************************/
        /* The entry doesn't have a CRL.                                     */
        /*********************************************************************/
        TRC_NORMAL((TB, "No CRL found"));
        goto EXIT_LABEL;
      }

      /***********************************************************************/
      /* Extract the issuer name from the CRL that we have found.            */
      /***********************************************************************/
      TRC_DETAIL((TB, "Convert CRL issuer name to RFC2253 form"));
      ccm_x509_name_to_rfc2253(X509_CRL_get_issuer(ret->data.crl), &got_name);
      if (got_name == NULL)
      {
        TRC_ERROR((TB, "Failed to convert CRL subject to text format"));
        goto EXIT_LABEL;
      }
    }
  }

  /***************************************************************************/
  /* Check that the subject name of the cert, or the issuer name of the CRL, */
  /* is the same as the name we were asked to look up.                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Check object's DN matches lookup DN"));
  if (strcmp(name, got_name) != 0)
  {
    TRC_ERROR((TB, "%s for '%s' has wrong name: '%s'",
              (type == X509_LU_X509) ? "Certificate" : "CRL", name, got_name));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Copy the object type (cert or CRL) to the result object.                */
  /***************************************************************************/
  ret->type = type;

  /***************************************************************************/
  /* We now have a cert or CRL and we have to add it to the local store.  If */
  /* this fails we don't really mind.  The cert/CRL will get looked up via   */
  /* LDAP next time it's needed.                                             */
  /***************************************************************************/
  if (type == X509_LU_X509)
  {
    TRC_NORMAL((TB, "Found certificate for '%s'", name));
    error = X509_STORE_add_cert(ccm.cert_store, ret->data.x509);
  }
  else
  {
    TRC_NORMAL((TB, "Found CRL for '%s'", name));
    error = X509_STORE_add_crl(ccm.cert_store, ret->data.crl);
  }

  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to add %s to local store",
                              (type == X509_LU_X509) ? "certificate" : "CRL"));
  }
  else
  {
    /*************************************************************************/
    /* We successfully stored the cert or CRL, so its reference count will   */
    /* have been incremented to 2.  We need to decrement it here because     */
    /* X509_STORE_get_by_subject() expects it to be 1 and increments the     */
    /* count itself before returning the cert/CRL to the original caller.    */
    /*************************************************************************/
    X509_OBJECT_free_contents(ret);
  }

  /***************************************************************************/
  /* Set the return code according to the type of object we're returning.    */
  /***************************************************************************/
  rc = type;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (name != NULL)
  {
    rm_release_mem(name);
  }

  if (ldap_name != NULL)
  {
    rm_release_mem(ldap_name);
  }

  if (search_result != NULL)
  {
    ldap_msgfree(search_result);
  }

  if (attrib_values != NULL)
  {
    ldap_value_free_len(attrib_values);
  }

  if (got_name != NULL)
  {
    rm_release_mem(got_name);
  }

  /***************************************************************************/
  /* We can now release the mutex.                                           */
  /***************************************************************************/
  OSP_RELEASE_LOCK(&ccm.ldap.access_mutex);

  TRC_EXIT();

  return(rc);

} /* ccm_lookup_ldap_by_subject */

/**PROC+**********************************************************************/
/* Name:      ccm_store_new_cert                                             */
/*                                                                           */
/* Purpose:   Store a newly-issued certificate in the LDAP directory         */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     x509           - the cert to store (OpenSSL form)       */
/*            IN     x509_asn1      - the cert to store (DER form)           */
/*            IN     x509_asn1_len  - the length of the cert (DER form)      */
/*            IN     subj_name      - the subject DN to store it under       */
/*                                                                           */
/* Operation: The subject of this certificate may or may not already have an */
/*            entry in the LDAP directory, and any entry may or may not      */
/*            already have a certificate attribute.                          */
/*                                                                           */
/*            If the entry is not present, this function creates it. If the  */
/*            entry is present with a userCertificate attribute, this        */
/*            function replaces the existing certificate. It is up to the    */
/*            caller to check for an existing certificate if required.       */
/*                                                                           */
/*            All LDAP operations are done synchronously using OpenLDAP.     */
/*                                                                           */
/*            In order to minimize the number of LDAP operations, an add     */
/*            operation is attempted first. If this fails because the entry  */
/*            already exists, a modify operation is attempted, to add or     */
/*            replace the certificate attribute.                             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_store_new_cert(X509 *x509,
                          char *x509_asn1,
                          unsigned int x509_asn1_len,
                          char *subj_name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  char *ldap_name = NULL;
  char *oc_values[3] = {NULL, NULL, NULL};
  struct berval cert_value;
  struct berval *cert_values[2] = {NULL, NULL};
  LDAPMod oc_mod;
  LDAPMod cert_mod;
  LDAPMod *mods[3] = {NULL, NULL, NULL};
  int error;

  TRC_ENTRY("ccm_store_new_cert");

  /***************************************************************************/
  /* Grab the lock so that the store operation is atomic with respect to     */
  /* other threads.                                                          */
  /***************************************************************************/
  OSP_GET_LOCK(&ccm.ldap.access_mutex);

  /***************************************************************************/
  /* Bind to the LDAP server if necessary.                                   */
  /***************************************************************************/
  rc = ccm_ldap_bind();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Store failed because LDAP bind failed"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add any required DN suffix for directory access.                        */
  /***************************************************************************/
  OSP_STRDUP(ldap_name, subj_name);
  if (ldap_name == NULL)
  {
    TRC_ERROR((TB, "Failed to duplicate name string"));
    goto EXIT_LABEL;
  }

  rc = ccm_add_ldap_dn_suffix(&ldap_name);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to expand DN for LDAP modify (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Build LDAP modify structure to add the attributes "objectClass" and     */
  /* "userCertificate".                                                      */
  /***************************************************************************/
  oc_values[0] = (char *)ccm_ldap_attr_top;
  oc_values[1] = (char *)ccm_ldap_attr_sauser;
  oc_values[2] = NULL;

  oc_mod.mod_op = LDAP_MOD_ADD;
  oc_mod.mod_type = (char *)ccm_ldap_attr_objclass;
  oc_mod.mod_vals.modv_strvals = oc_values;
  oc_mod.mod_next = NULL;

  cert_value.bv_val = x509_asn1;
  cert_value.bv_len = x509_asn1_len;

  cert_values[0] = &cert_value;
  cert_values[1] = NULL;

  cert_mod.mod_op = LDAP_MOD_ADD | LDAP_MOD_BVALUES;
  cert_mod.mod_type = (char *)ccm_ldap_attr_usercert;
  cert_mod.mod_vals.modv_bvals = cert_values;
  cert_mod.mod_next = NULL;

  mods[0] = &oc_mod;
  mods[1] = &cert_mod;
  mods[2] = NULL;

  /***************************************************************************/
  /* First, attempt to create a new entry.                                   */
  /***************************************************************************/
  TRC_NORMAL((TB, "Attempting create for '%s'", ldap_name));
  error = ldap_add_s(ccm.ldap.handle, ldap_name, mods);
  if (error != LDAP_SUCCESS)
  {
    if (error == LDAP_ALREADY_EXISTS)
    {
      /***********************************************************************/
      /* There's already an entry for this subject name, so try to add the   */
      /* certificate attribute to it, replacing any existing one.            */
      /***********************************************************************/
      TRC_NORMAL((TB, "Subject already has a directory entry: "
                                             "attempting to add certificate"));
      oc_mod.mod_op = LDAP_MOD_REPLACE;
      cert_mod.mod_op = LDAP_MOD_REPLACE | LDAP_MOD_BVALUES;

      error = ldap_modify_s(ccm.ldap.handle, ldap_name, mods);
      if (error != LDAP_SUCCESS)
      {
        if (error == LDAP_SERVER_DOWN)
        {
          /*******************************************************************/
          /* If the server has gone down we'll have to unbind, then rebind   */
          /* next time an LDAP operation is attempted.                       */
          /*******************************************************************/
          TRC_ERROR((TB, "LDAP modify failed: the server appears to be down"));
          ldap_unbind_s(ccm.ldap.handle);
          ccm.ldap.handle = NULL;
        }
        else
        {
          TRC_ERROR((TB, "Error %d (%s) adding cert to LDAP entry for '%s'",
                                    error, ldap_err2string(error), subj_name));
        }
        rc = OSP_ERROR;
        goto EXIT_LABEL;
      }
    }
    else if (error == LDAP_SERVER_DOWN)
    {
      /***********************************************************************/
      /* If the server has gone down we'll have to unbind, then rebind next  */
      /* time an LDAP operation is attempted.                                */
      /***********************************************************************/
      TRC_ERROR((TB, "LDAP add failed: the server appears to be down"));
      ldap_unbind_s(ccm.ldap.handle);
      ccm.ldap.handle = NULL;
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
    else
    {
      TRC_ERROR((TB, "Error %d (%s) storing cert for '%s' in LDAP directory",
                                    error, ldap_err2string(error), subj_name));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* We've stored the certificate.                                           */
  /***************************************************************************/
  TRC_NORMAL((TB, "Stored certificate for '%s'", subj_name));

  /***************************************************************************/
  /* Add the cert to the local store while we have it to hand.               */
  /***************************************************************************/
  TRC_NORMAL((TB, "Placing cert into local store"));
  error = X509_STORE_add_cert(ccm.cert_store, x509);
  if (error == 0)
  {
    /*************************************************************************/
    /* Don't exit as this is not fatal.  The cert will just get looked up    */
    /* via LDAP when the store needs it.                                     */
    /*************************************************************************/
    OSP_OPENSSL_ERROR((TB, "Failed to add certificate to local store"));
  }

  TRC_NORMAL((TB, "Stored new certificate in LDAP directory OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* We can now release the mutex.                                           */
  /***************************************************************************/
  OSP_RELEASE_LOCK(&ccm.ldap.access_mutex);

  /***************************************************************************/
  /* Clean up in all cases.                                                  */
  /***************************************************************************/
  if (ldap_name != NULL)
  {
    rm_release_mem(ldap_name);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_store_new_cert */

/**PROC+**********************************************************************/
/* Name:      ccm_add_ldap_dn_suffix                                         */
/*                                                                           */
/* Purpose:   Add the required suffix to a DN for use with the LDAP store    */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN/OUT name           - IN: the LDAP DN to expand              */
/*                                    OUT: the expanded DN                   */
/*                                                                           */
/* Operation: Reallocate the memory and append the suffix.                   */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_add_ldap_dn_suffix(char **name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  void *new_mem = NULL;
  unsigned int name_len;

  TRC_ENTRY("ccm_add_ldap_dn_suffix");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((name != NULL), (TB, "name is null"));
  TRC_ASSERT((*name != NULL), (TB, "*name is null"));

  /***************************************************************************/
  /* Allocate the required memory, not forgetting the comma separator and    */
  /* the terminating null.                                                   */
  /***************************************************************************/
  if ((ccm.ldap.dn_suffix != NULL) && (strlen(ccm.ldap.dn_suffix) > 0))
  {
    name_len = strlen(*name) + 1 + strlen(ccm.ldap.dn_suffix) + 1;
    new_mem = rm_realloc_mem(*name, name_len);
    if (new_mem == NULL)
    {
      TRC_ERROR((TB, "Failed to realloc memory to store expanded DN"));
      rc = OSP_ERR_MEMORY;
      goto EXIT_LABEL;
    }

    *name = (char *)new_mem;
    strcat(*name, ",");
    strcat(*name, ccm.ldap.dn_suffix);

    TRC_DETAIL((TB, "DN with lookup suffix added is '%s'", *name));
  }
  else
  {
    TRC_DETAIL((TB, "No lookup suffix required"));
  }

  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* ccm_add_ldap_dn_suffix */

/**PROC+**********************************************************************/
/* Name:      ccm_ldap_bind                                                  */
/*                                                                           */
/* Purpose:   Bind to the LDAP server if necessary. Caller must hold mutex   */
/*            ccm.ldap.access_mutex.                                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: See if we're bound already. If not, do the bind operation.     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_ldap_bind(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int error;

  TRC_ENTRY("ccm_ldap_bind");

  if (ccm.ldap.handle == NULL)
  {
    /*************************************************************************/
    /* First, get an LDAP handle.                                            */
    /*************************************************************************/
    TRC_NORMAL((TB, "Binding to LDAP server..."));
    ccm.ldap.handle = ldap_init(ccm.ldap.host, ccm.ldap.port);
    if (ccm.ldap.handle == NULL)
    {
      TRC_ERROR((TB, "Failed LDAP init for host '%s:%d'",
                 ccm.ldap.host,
                 ccm.ldap.port));
      rc = OSP_ERR_BIND_FAILED;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Now do the bind.                                                      */
    /*************************************************************************/
    error = ldap_simple_bind_s(ccm.ldap.handle,
                               ccm.ldap.user,
                               ccm.ldap.password);
    if (error != LDAP_SUCCESS)
    {
      TRC_ERROR((TB, "LDAP bind failed (%u) to host '%s:%d' as user '%s'",
                          error, ccm.ldap.host, ccm.ldap.port, ccm.ldap.user));
      TRC_ERROR((TB, "Check that the LDAP server is running"));
      rc = OSP_ERR_BIND_FAILED;
      goto EXIT_LABEL;
    }

    TRC_NORMAL((TB, "Bound to LDAP server OK"));
  }

  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* ccm_ldap_bind */

