
/**MOD+***********************************************************************/
/* Module:    ccmmain.c                                                      */
/*                                                                           */
/* Purpose:   CCM initialization and termination functions                   */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.20               $ $Modtime::   Jan 11 2001 15:56:32   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#define CCM_DEFINE_VARS
#include <ccmincl.h>
#undef CCM_DEFINE_VARS

/**PROC+**********************************************************************/
/* Name:      ccm_init                                                       */
/*                                                                           */
/* Purpose:   Initialize CCM                                                 */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     ssl_ctx        - the OpenSSL context to initialize      */
/*            IN     ssl_verify_cb  - SM's SSL cert verification callback    */
/*                                                                           */
/* Operation: Initialize LAM; open our own (non-LAM) LDAP access too.        */
/*            Also allocate and initialize the internal data, including:     */
/*            - read the configuration file                                  */
/*            - create the certificate store object                          */
/*            - set up the SSL context to use our store, the correct         */
/*                certificate/private key pair, the configured cipher list   */
/*                and the configured client verification mode                */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_init(SSL_CTX *ssl_ctx,
                PCCM_SSL_VERIFY_CALLBACK ssl_verify_cb)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  OSP_RC int_rc = OSP_SUCCESS;
  X509 *cert = NULL;
  X509 *ssl_cert = NULL;
  EVP_PKEY *ssl_priv_key = NULL;
  X509_LOOKUP *lookup = NULL;
  int error;
  int jj;
  int dh_check;

  TRC_ENTRY("ccm_init");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((ssl_ctx != NULL), (TB, "Missing ssl_ctx"));
  TRC_ASSERT((ssl_verify_cb != NULL), (TB, "Missing ssl_verify_cb"));

  /***************************************************************************/
  /* Initialize the CCM global data structure.                               */
  /***************************************************************************/
  ccm.cert_store = NULL;
  ccm.config = NULL;
  ccm.prng_seeded = FALSE;
  ccm.lam_initialized = FALSE;

  ccm.ca.serial_file = NULL;
  ccm.ca.serial_num = NULL;
  memset(&ccm.ca.serial_num_mutex, 0, sizeof(pthread_mutex_t));
  ccm.ca.valid_duration = 0;
  ccm.ca.rsa_sign_digest = NULL;
  ccm.ca.subject_suffix = NULL;

  ccm.ldap.host = NULL;
  ccm.ldap.user = NULL;
  ccm.ldap.password = NULL;
  ccm.ldap.dn_suffix = NULL;
  ccm.ldap.handle = NULL;
  memset(&ccm.ldap.access_mutex, 0, sizeof(ccm.ldap.access_mutex));

  ccm.ssl.verify_cb = ssl_verify_cb;
  ccm.ssl.identity = NULL;
  ccm.ssl.verify_mode = CCM_CFG_SSL_VERIFY_CLIENT_DFLT;
  ccm.ssl.cipher_list = NULL;

  ccm.identities.num = 0;
  ccm.identities.names = NULL;
  ccm.identities.privkeys = NULL;

  ccm.tempkeys.rsa512 = NULL;
  ccm.tempkeys.rsa1024 = NULL;
  ccm.tempkeys.dh512 = NULL;
  ccm.tempkeys.dh1024 = NULL;

  /***************************************************************************/
  /* Create all the mutexes we need.                                         */
  /***************************************************************************/
  rc = osp_create_mutex(&ccm.ca.serial_num_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create serial number mutex (%u)", rc));
    goto EXIT_LABEL;
  }

  rc = osp_create_mutex(&ccm.ldap.access_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create LDAP access mutex (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read all the configuration info we need.                                */
  /***************************************************************************/
  rc = ccm_load_config();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to read config file (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Seed the random number generator.                                       */
  /***************************************************************************/
  rc = ccm_read_random_state();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to read random state (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize LAM.                                                         */
  /***************************************************************************/
  rc = lam_init();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize LAM (%u)", rc));
    goto EXIT_LABEL;
  }
  ccm.lam_initialized = TRUE;

  /***************************************************************************/
  /* Create the certificate store object.                                    */
  /***************************************************************************/
  ccm.cert_store = X509_STORE_new();
  if (ccm.cert_store == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create certificate store"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add our LDAP lookup method.                                             */
  /***************************************************************************/
  lookup = X509_STORE_add_lookup(ccm.cert_store,
                                 &ccm_lookup_ldap_method);
  if (lookup == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to add LDAP lookup method to cert store"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Attempt to load in the certificate corresponding to each identity that  */
  /* is configured.  We don't need the actual certificate right now, but     */
  /* this lookup allows us to see if it exists and also loads it into local  */
  /* memory so that it'll be to hand when we do need it.                     */
  /***************************************************************************/
  for (jj = 0; jj < ccm.identities.num; jj++)
  {
    TRC_NORMAL((TB, "Looking up certificate for identity #%d", jj));
    rc = ccm_int_get_cert_x509_name(ccm.identities.names[jj], &cert);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Couldn't find cert for ID #%d (%u)", jj, rc));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Check that the certificate and private key are consistent.            */
    /*************************************************************************/
    error = X509_check_private_key(cert, ccm.identities.privkeys[jj]);
    if (error == 0)
    {
      TRC_ERROR((TB, "Cert and private key are inconsistent for ID #%d", jj));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    X509_free(cert);
    cert = NULL;
  }

  /***************************************************************************/
  /* Tell the SSL context to use our store.  We don't ever free the store;   */
  /* it is freed when SM frees the SSL_CTX.  X509_STORE objects don't        */
  /* implement reference counting so we can't use that.                      */
  /***************************************************************************/
  SSL_CTX_set_cert_store(ssl_ctx, ccm.cert_store);

  /***************************************************************************/
  /* Tell the SSL context to use the right certificate and private key.      */
  /***************************************************************************/
  TRC_NORMAL((TB, "Setting SSL certificate"));
  rc = ccm_int_get_cert(ccm.ssl.identity, &ssl_cert);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get SSL certificate (%u)", rc));
    goto EXIT_LABEL;
  }

  error = SSL_CTX_use_certificate(ssl_ctx, ssl_cert);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set SSL certificate"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Setting SSL private key"));
  rc = ccm_get_priv_key(ccm.ssl.identity, &ssl_priv_key);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get SSL private key (%u)", rc));
    goto EXIT_LABEL;
  }

  error = SSL_CTX_use_PrivateKey(ssl_ctx, ssl_priv_key);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set SSL private key"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the SSL verify mode.  This also sets the callback function for our  */
  /* certificate store, which is used for all cert verification.             */
  /***************************************************************************/
  SSL_CTX_set_verify(ssl_ctx,
                     ccm.ssl.verify_mode,
                     ccm_verify_callback);

  /***************************************************************************/
  /* Set the SSL cipher list.                                                */
  /***************************************************************************/
  TRC_NORMAL((TB, "Setting SSL cipher list to '%s'", ccm.ssl.cipher_list));
  error = SSL_CTX_set_cipher_list(ssl_ctx, ccm.ssl.cipher_list);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set cipher list to '%s'",
                                                         ccm.ssl.cipher_list));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Generate temporary RSA keys.                                            */
  /***************************************************************************/
  TRC_NORMAL((TB, "Generating temporary RSA keys"));
  ccm.tempkeys.rsa512 = RSA_generate_key(512, RSA_F4, NULL, NULL);
  if (ccm.tempkeys.rsa512 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to generate temporary 512-bit RSA key"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  ccm.tempkeys.rsa1024 = RSA_generate_key(1024, RSA_F4, NULL, NULL);
  if (ccm.tempkeys.rsa1024 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to generate temporary 1024-bit RSA key"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Load pre-generated temporary DH parameters.                             */
  /***************************************************************************/
  TRC_NORMAL((TB, "Loading and checking temporary DH parameters"));
  ccm.tempkeys.dh512 = ccm_get_dh512();
  if (ccm.tempkeys.dh512 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to load 512-bit DH params"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  ccm.tempkeys.dh1024 = ccm_get_dh1024();
  if (ccm.tempkeys.dh1024 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to load 1024-bit DH params"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Double-check pre-generated temporary DH parameters for consistency.     */
  /***************************************************************************/
  error = DH_check(ccm.tempkeys.dh512, &dh_check);
  if (error != 1)
  {
    TRC_ERROR((TB, "Error while checking 512-bit temporary DH parameters"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }
  if (dh_check != 0)
  {
    TRC_ERROR((TB, "512-bit temp DH parameters failed check (%d)", dh_check));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  error = DH_check(ccm.tempkeys.dh1024, &dh_check);
  if (error != 1)
  {
    TRC_ERROR((TB, "Error while checking 1024-bit temporary DH parameters"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }
  if (dh_check != 0)
  {
    TRC_ERROR((TB, "1024-bit temp DH parameters failed check (%d)", dh_check));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up callbacks for temporary RSA keys and Diffie-Hellman parameters.  */
  /***************************************************************************/
  TRC_NORMAL((TB, "Setting RSA temp key and DH temp parameter callbacks"));
  SSL_CTX_set_tmp_rsa_callback(ssl_ctx, ccm_temp_rsa_callback);
  SSL_CTX_set_tmp_dh_callback(ssl_ctx, ccm_temp_dh_callback);

  TRC_NORMAL((TB, "CCM initialized OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* We can free the SSL cert now since OpenSSL will have upped the          */
  /* reference count on it.                                                  */
  /***************************************************************************/
  if (ssl_cert != NULL)
  {
    TRC_DETAIL((TB, "Freeing certificate"));
    X509_free(ssl_cert);
  }

  /***************************************************************************/
  /* Free the temporary cert we use to retrieve certs for configured IDs.    */
  /***************************************************************************/
  if (cert != NULL)
  {
    TRC_ALERT((TB, "Freeing temp cert (should only be done on error)"));
    X509_free(cert);
  }

  /***************************************************************************/
  /* Don't free ssl_priv_key as it's just a pointer to the permanent copy    */
  /* in our identity list.                                                   */
  /***************************************************************************/

  /***************************************************************************/
  /* Don't free lookup as we don't own the object.                           */
  /***************************************************************************/

  /***************************************************************************/
  /* Clean up if we failed.                                                  */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Terminating because of error during init (%u)", rc));
    int_rc = ccm_term();
    if (int_rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to terminate on error (%u)", int_rc));
    }
  }

  TRC_EXIT();

  return(rc);

} /* ccm_init */

/**PROC+**********************************************************************/
/* Name:      ccm_term                                                       */
/*                                                                           */
/* Purpose:   Terminate CCM                                                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Free all CCM internal data.                                    */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_term(void)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int error;
  int ii;

  TRC_ENTRY("ccm_term");

  /***************************************************************************/
  /* Terminate LAM.                                                          */
  /***************************************************************************/
  if (ccm.lam_initialized)
  {
    TRC_NORMAL((TB, "Terminating LAM"));
    rc = lam_term();
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to terminate LAM (%u)", rc));
    }
  }

  /***************************************************************************/
  /* Save the random number generator state, but only if it was correctly    */
  /* seeded on startup (we don't want to write a low-entropy randfile).      */
  /***************************************************************************/
  TRC_DETAIL((TB, "Writing random state"));
  rc = ccm_write_random_state();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to write random state (%u)", rc));
  }

  /***************************************************************************/
  /* Now free all our global data.  Don't free the certificate and CRL store */
  /* -- this is done when SM frees the SSL_CTX.                              */
  /***************************************************************************/

  /***************************************************************************/
  /* Free the temporary RSA keys and DH params.                              */
  /***************************************************************************/
  if (ccm.tempkeys.rsa512 != NULL)
  {
    TRC_DETAIL((TB, "Freeing temporary 512-bit RSA keypair"));
    RSA_free(ccm.tempkeys.rsa512);
  }
  if (ccm.tempkeys.rsa1024 != NULL)
  {
    TRC_DETAIL((TB, "Freeing temporary 1024-bit RSA keypair"));
    RSA_free(ccm.tempkeys.rsa1024);
  }

  if (ccm.tempkeys.dh512 != NULL)
  {
    TRC_DETAIL((TB, "Freeing temporary 512-bit DH parameters"));
    DH_free(ccm.tempkeys.dh512);
  }
  if (ccm.tempkeys.dh1024 != NULL)
  {
    TRC_DETAIL((TB, "Freeing temporary 1024-bit DH parameters"));
    DH_free(ccm.tempkeys.dh1024);
  }

  /***************************************************************************/
  /* Free the config info table.                                             */
  /***************************************************************************/
  if (ccm.config != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.config"));
    CONF_free(ccm.config);
  }

  /***************************************************************************/
  /* Free the random seed file name.                                         */
  /***************************************************************************/
  if (ccm.rand_file != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.rand_file"));
    rm_release_mem(ccm.rand_file);
  }

  /***************************************************************************/
  /* Free the CA serial number file name.                                    */
  /***************************************************************************/
  if (ccm.ca.serial_file != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ca.serial_file"));
    rm_release_mem(ccm.ca.serial_file);
  }

  /***************************************************************************/
  /* Free the CA serial number structure.                                    */
  /***************************************************************************/
  if (ccm.ca.serial_num != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ca.serial_num"));
    BN_free(ccm.ca.serial_num);
  }

  /***************************************************************************/
  /* Free the subject name suffix structure.                                 */
  /***************************************************************************/
  if (ccm.ca.subject_suffix != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ca.subject_suffix"));
    X509_NAME_free(ccm.ca.subject_suffix);
  }

  /***************************************************************************/
  /* Close LDAP access and free associated memory.                           */
  /***************************************************************************/
  if (ccm.ldap.handle != NULL)
  {
    TRC_NORMAL((TB, "Closing LDAP certificate directory"));
    OSP_GET_LOCK(&ccm.ldap.access_mutex);
    error = ldap_unbind(ccm.ldap.handle);
    OSP_RELEASE_LOCK(&ccm.ldap.access_mutex);
    if (error != LDAP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed ldap_unbind"));
    }
  }

  if (ccm.ldap.host != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ldap.host"));
    rm_release_mem(ccm.ldap.host);
  }

  if (ccm.ldap.user != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ldap.user"));
    rm_release_mem(ccm.ldap.user);
  }

  if (ccm.ldap.password != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ldap.password"));
    rm_release_mem(ccm.ldap.password);
  }

  if (ccm.ldap.dn_suffix != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ldap.dn_suffix"));
    rm_release_mem(ccm.ldap.dn_suffix);
  }

  /***************************************************************************/
  /* Free the memory associated with SSL configuration.                      */
  /***************************************************************************/
  if (ccm.ssl.identity != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ssl.identity"));
    rm_release_mem(ccm.ssl.identity);
  }

  if (ccm.ssl.cipher_list != NULL)
  {
    TRC_DETAIL((TB, "Freeing ccm.ssl.cipher_list"));
    rm_release_mem(ccm.ssl.cipher_list);
  }

  /***************************************************************************/
  /* Free memory for the ID list.                                            */
  /***************************************************************************/
  if (ccm.identities.num > 0)
  {
    TRC_NORMAL((TB, "Freeing %u identities", ccm.identities.num));

    if (ccm.identities.names != NULL)
    {
      /***********************************************************************/
      /* Free the DN index.                                                  */
      /***********************************************************************/
      for (ii = 0;
           (ii < ccm.identities.num) && (ccm.identities.names[ii] != NULL);
           ii++)
      {
        TRC_DETAIL((TB, "Freeing name for ID #%d", ii));
        X509_NAME_free(ccm.identities.names[ii]);
      }

      TRC_DETAIL((TB, "Freeing name index"));
      rm_release_mem(ccm.identities.names);
    }

    if (ccm.identities.privkeys != NULL)
    {
      /***********************************************************************/
      /* Free the private key index.                                         */
      /***********************************************************************/
      for (ii = 0;
           (ii < ccm.identities.num) && (ccm.identities.privkeys[ii] != NULL);
           ii++)
      {
        TRC_DETAIL((TB, "Freeing private key for ID #%d", ii));
        EVP_PKEY_free(ccm.identities.privkeys[ii]);
      }

      TRC_DETAIL((TB, "Freeing private key index"));
      rm_release_mem(ccm.identities.privkeys);
    }
  }

  /***************************************************************************/
  /* Destroy all our mutexes.                                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Destroying mutexes"));
  rc = osp_destroy_mutex(&ccm.ldap.access_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to destroy LDAP access mutex (%u)", rc));
  }

  rc = osp_destroy_mutex(&ccm.ca.serial_num_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to destroy serial number mutex (%u)", rc));
  }

  TRC_NORMAL((TB, "CCM terminated OK"));
  rc = OSP_SUCCESS;

  TRC_EXIT();

  return(rc);

} /* ccm_term */

