
/**MOD+***********************************************************************/
/* Module:    ccmpk7ut.c                                                     */
/*                                                                           */
/* Purpose:   CCM PKCS7-related utility functions                            */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* $Revision::   1.3                $ $Modtime::   Jan 11 2001 16:11:50   $  */
/*                                                                           */
/* The code in this file borrows heavily from that in the OpenSSL file       */
/* crypto\pkcs7\pk7_smime.c, and hence the following license conditions      */
/* apply:                                                                    */
/*                                                                           */
/* ====================================================================
 * Copyright (c) 1999 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_int_pkcs7_sign                                             */
/*                                                                           */
/* Purpose:   Sign data to produce a PKCS#7 SignedData structure             */
/*                                                                           */
/* Returns:   PKCS7 ptr   - success                                          */
/*            NULL        - failure                                          */
/*                                                                           */
/* Params:    IN     signcert       - the certificate to sign with           */
/*            IN     pkey           - the private key to sign with           */
/*            IN     certs          - optional extra certs to add in         */
/*            IN     data           - the data to sign (may be NULL)         */
/*            IN     flags          - (see PKCS7_sign)                       */
/*            IN     digest         - CCM_DIGEST_MD5:     MD5                */
/*                                    CCM_DIGEST_SHA1:    SHA-1              */
/*                                    CCM_DIGEST_DEFAULT: default            */
/*            IN     auth_attribs   - stack of authenticated attribs to add  */
/*                                                                           */
/* Operation: This is essentially the same as the OpenSSL function           */
/*            PKCS7_sign(). The S/MIME-specific bits are removed, and there  */
/*            is support for the addition of arbitrary authenticated attribs */
/*            and selection of the digest algorithm by the caller.           */
/*                                                                           */
/**PROC-**********************************************************************/

PKCS7 *ccm_int_pkcs7_sign(X509 *signcert,
                          EVP_PKEY *pkey,
                          STACK_OF(X509) *certs,
                          BIO *data,
                          int flags,
                          unsigned int digest,
                          STACK_OF(X509_ATTRIBUTE) *auth_attribs)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  PKCS7 *rc = NULL;
  PKCS7 *p7 = NULL;
  PKCS7_SIGNER_INFO *si = NULL;
  BIO *p7bio = NULL;
  EVP_MD *md = NULL;
  int ii;
  int error;

  TRC_ENTRY("ccm_int_pkcs7_sign");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if (!X509_check_private_key(signcert, pkey))
  {
    TRC_ERROR((TB, "Private key doesn't match certificate"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create the structure, set type and create new (empty) content.          */
  /***************************************************************************/
  p7 = PKCS7_new();
  if (p7 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create new PKCS#7 structure"));
    goto EXIT_LABEL;
  }

  error = PKCS7_set_type(p7, NID_pkcs7_signed);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set type of PKCS#7 to SignedData"));
    goto EXIT_LABEL;
  }

  error = PKCS7_content_new(p7, NID_pkcs7_data);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create new content for SignedData"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Decide which digest to use.                                             */
  /***************************************************************************/
  if (digest == CCM_DIGEST_SHA1)
  {
    TRC_NORMAL((TB, "Using SHA-1 digest"));
    md = EVP_sha1();
  }
  else
  {
    TRC_NORMAL((TB, "Using MD5 digest"));
    md = EVP_md5();
  }

  if (md == NULL)
  {
    TRC_ERROR((TB, "Failed to get message digest"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add signer info to the SignedData structure.                            */
  /***************************************************************************/
  si = PKCS7_add_signature(p7, signcert, pkey, md);
  if (si == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to add signerInfo to PKCS#7 SignedData"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add certificates (unless the caller doesn't want them).                 */
  /***************************************************************************/
  if (!(flags & PKCS7_NOCERTS))
  {
    TRC_NORMAL((TB, "Adding signer certificate"));
    PKCS7_add_certificate(p7, signcert);
    if (certs != NULL)
    {
      TRC_NORMAL((TB, "Adding %d additional certs", sk_X509_num(certs)));
      for (ii = 0; ii < sk_X509_num(certs); ii++)
      {
        PKCS7_add_certificate(p7, sk_X509_value(certs, ii));
      }
    }
  }

  /***************************************************************************/
  /* Initialize the digest calculation.                                      */
  /***************************************************************************/
  p7bio = PKCS7_dataInit(p7, NULL);
  if (p7bio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to initialize PKCS#7 digest calculation"));
    goto EXIT_LABEL;
  }

  if (data != NULL)
  {
    /*************************************************************************/
    /* We have some data to sign -- pump it in.                              */
    /*************************************************************************/
    SMIME_crlf_copy(data, p7bio, flags);
  }

  /***************************************************************************/
  /* Add the authenticated attributes.                                       */
  /***************************************************************************/
  if (auth_attribs != NULL)
  {
    TRC_NORMAL((TB, "Adding authenticated attribs"));
    PKCS7_set_signed_attributes(si, auth_attribs);
  }

  /***************************************************************************/
  /* Set the content type to data.                                           */
  /***************************************************************************/
  PKCS7_add_signed_attribute(si, NID_pkcs9_contentType,
        V_ASN1_OBJECT, OBJ_nid2obj(NID_pkcs7_data));

  /***************************************************************************/
  /* Set the signature to be detached if required.                           */
  /***************************************************************************/
  if (flags & PKCS7_DETACHED)
  {
    TRC_NORMAL((TB, "Setting SignedData to be detached"));
    PKCS7_set_detached(p7, 1);
  }

  /***************************************************************************/
  /* Finalize the digest calculation.  This generates the signatures.        */
  /***************************************************************************/
  if (!PKCS7_dataFinal(p7, p7bio))
  {
    OSP_OPENSSL_ERROR((TB, "Failed to calculate signatures"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Done internal sign operation"));
  rc = p7;

EXIT_LABEL:

  if (p7bio != NULL)
  {
    BIO_free_all(p7bio);
  }

  if ((rc == NULL) && (p7 != NULL))
  {
    PKCS7_free(p7);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_int_pkcs7_sign */

/**PROC+**********************************************************************/
/* Name:      ccm_int_pkcs7_verify                                           */
/*                                                                           */
/* Purpose:   Verify a PKCS#7 data structure                                 */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     p7             - the PKCS#7 SignedData structure        */
/*            IN     certs          - optional certs to use                  */
/*            IN     store          - the cert store to use for lookups      */
/*            IN     indata         - data to verify if sig is detached      */
/*            IN     out            - verified data output                   */
/*            IN     flags          - (see PKCS7_verify)                     */
/*            IN     extra_flags    - CCM_FLAG_SELF_SIGNED: allow a single   */
/*                                      self-signed cert                     */
/*                                                                           */
/* Operation: This is essentially the same as the OpenSSL function           */
/*            PKCS7_verify() with modifications so that the verification     */
/*            callback knows it's being called for PKCS#7, not SSL.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_int_pkcs7_verify(PKCS7 *p7,
                            STACK_OF(X509) *certs,
                            X509_STORE *store,
                            BIO *indata,
                            BIO *out,
                            int flags,
                            unsigned int extra_flags)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  STACK_OF(X509) *signers = NULL;
  X509 *signer = NULL;
  BIO *p7bio = NULL;
  STACK_OF(PKCS7_SIGNER_INFO) *sinfos = NULL;
  PKCS7_SIGNER_INFO *si = NULL;
  X509_STORE_CTX cert_ctx;
  char buf[4096];
  int ii;
  int error;
  int store_error = 0;

  TRC_ENTRY("ccm_int_pkcs7_verify");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  rc = OSP_ERR_BAD_PARAM;
  if (p7 == NULL)
  {
    TRC_ERROR((TB, "Null pointer passed for PKCS#7 struct"));
    goto EXIT_LABEL;
  }

  if (!PKCS7_type_is_signed(p7))
  {
    TRC_ERROR((TB, "PKCS#7 type is not SignedData"));
    goto EXIT_LABEL;
  }

  if (PKCS7_get_detached(p7) && (indata == NULL))
  {
    TRC_ERROR((TB, "Detached SignedData and no data supplied"));
    goto EXIT_LABEL;
  }

  if (!PKCS7_get_detached(p7) && indata)
  {
    TRC_ERROR((TB, "Non-detached SignedData but separate data supplied"));
    goto EXIT_LABEL;
  }

  sinfos = PKCS7_get_signer_info(p7);
  if ((sinfos == NULL) || (sk_PKCS7_SIGNER_INFO_num(sinfos) == 0))
  {
    TRC_ERROR((TB, "No signatures found on SignedData"));
    goto EXIT_LABEL;
  }

  signers = PKCS7_get0_signers(p7, certs, flags);
  if (signers == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "No signer certs found"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Assume the worst                                                        */
  /***************************************************************************/
  rc = OSP_ERR_FAILED_VERIFY;

  /***************************************************************************/
  /* Verify all the signer certificates that we have.                        */
  /***************************************************************************/
  TRC_NORMAL((TB, "Verifying %d signer cert(s)", sk_X509_num(signers)))
  for (ii = 0; ii < sk_X509_num(signers); ii++)
  {
    TRC_NORMAL((TB, "Verifying signer cert #%d", ii));
    signer = sk_X509_value(signers, ii);
    if (signer == NULL)
    {
      TRC_ERROR((TB, "Didn't get signer %d", ii));
      goto EXIT_LABEL;
    }

    if (!(flags & PKCS7_NOCHAIN))
    {
      TRC_NORMAL((TB, "Cert chain allowed"));
      X509_STORE_CTX_init(&cert_ctx, store, signer, p7->d.sign->cert);
      X509_STORE_CTX_set_purpose(&cert_ctx, X509_PURPOSE_SMIME_SIGN);
    }
    else
    {
      TRC_NORMAL((TB, "Cert chain not allowed"));
      X509_STORE_CTX_init(&cert_ctx, store, signer, NULL);
    }

    /*************************************************************************/
    /* Set user data so that the verify callback knows that it's verifying   */
    /* PKCS#7 and not a certificate presented via SSL.  OpenSSL internally   */
    /* uses data index 0 to store the SSL pointer when it verifies a cert,   */
    /* so if we set that to NULL the code in ccm_verify_callback() will know */
    /* it's doing a PKCS#7 cert verification and can get the flags from data */
    /* index 1.                                                              */
    /*************************************************************************/
    X509_STORE_CTX_set_ex_data(&cert_ctx, 0, NULL);
    X509_STORE_CTX_set_ex_data(&cert_ctx, 1, &extra_flags);

    error = X509_verify_cert(&cert_ctx);
    if (error <= 0)
    {
      store_error = X509_STORE_CTX_get_error(&cert_ctx);
      OSP_OPENSSL_ERROR((TB, "Verify error (ret=%d); store error=%d ('%s')",
              error, store_error, X509_verify_cert_error_string(store_error)));

      X509_STORE_CTX_cleanup(&cert_ctx);
      goto EXIT_LABEL;
    }

    X509_STORE_CTX_cleanup(&cert_ctx);

    /*************************************************************************/
    /* Checking for revocation status has been done in the verify callback.  */
    /*************************************************************************/
  }

  /***************************************************************************/
  /* Set up an OpenSSL BIO through which the data is filtered to calculate   */
  /* digests.                                                                */
  /***************************************************************************/
  p7bio = PKCS7_dataInit(p7, indata);
  if (p7bio == NULL)
  {
    TRC_ERROR((TB, "Failed to get bio for digest calculation"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We now have to 'read' from p7bio to calculate digests.                  */
  /***************************************************************************/
  TRC_NORMAL((TB, "Calculating digests"));
  for (;;)
  {
    ii = BIO_read(p7bio, buf, sizeof(buf));
    if (ii <= 0)
    {
      TRC_NORMAL((TB, "Finished calculating digests"));
      break;
    }
    if (out)
    {
      BIO_write(out, buf, ii);
    }
  }

  /***************************************************************************/
  /* Now verify all signatures.                                              */
  /***************************************************************************/
  TRC_NORMAL((TB, "Verifying signatures"));
  for (ii = 0; ii < sk_PKCS7_SIGNER_INFO_num(sinfos); ii++)
  {
    TRC_DETAIL((TB, "Verifying signature #%d", ii));
    si = sk_PKCS7_SIGNER_INFO_value(sinfos, ii);
    if (si == NULL)
    {
      TRC_ERROR((TB, "Failed to get signer %d info value", ii));
      goto EXIT_LABEL;
    }
    signer = sk_X509_value(signers, ii);
    if (signer == NULL)
    {
      TRC_ERROR((TB, "Failed to get signer %d value", ii));
      goto EXIT_LABEL;
    }
    error = PKCS7_signatureVerify(p7bio, p7, si, signer);
    if (error <= 0)
    {
      TRC_ERROR((TB, "Signature verification failed, signer %d", ii));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* If we get here, all is well                                             */
  /***************************************************************************/
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Do any tidy up                                                          */
  /***************************************************************************/
  if (signers != NULL)
  {
    TRC_DETAIL((TB, "Freeing signers"));
    sk_X509_free(signers);
  }

  if (p7bio != NULL)
  {
    TRC_DETAIL((TB, "Freeing bio"));
    if (indata)
    {
      BIO_pop(p7bio);
    }
    BIO_free_all(p7bio);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_int_pkcs7_verify */

