
/**MOD+***********************************************************************/
/* Module:    ccmpkcs7.c                                                     */
/*                                                                           */
/* Purpose:   CCM PKCS#7 functions                                           */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.20               $ $Modtime::   Jan 11 2001 15:56:32   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_pkcs7_sign                                                 */
/*                                                                           */
/* Purpose:   Sign some data to produce a PKCS#7 SignedData object           */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     subj_name      - the subject name to sign as            */
/*            IN     sdata          - data to be signed (may be NULL)        */
/*            IN     sdata_len      - number of bytes to be signed           */
/*            IN     digest         - CCM_DIGEST_MD5:     MD5                */
/*                                    CCM_DIGEST_SHA1:    SHA-1              */
/*                                    CCM_DIGEST_DEFAULT: default            */
/*            IN     auth_attribs   - authenticated attribs to include       */
/*            IN     flags          - CCM_FLAG_PKCS7_D: make detached sig    */
/*            OUT    sig_buf        - resulting PKCS#7 SignedData            */
/*            OUT    sig_len        - length of SignedData                   */
/*                                                                           */
/* Operation: Set up and clean up around the helper function that does the   */
/*            real work.                                                     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_pkcs7_sign(char *subj_name,
                      char *sdata,
                      unsigned int sdata_len,
                      unsigned int digest,
                      STACK_OF(X509_ATTRIBUTE) *auth_attribs,
                      unsigned int flags,
                      char **sig_buf,
                      unsigned int *sig_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509 *signer_cert = NULL;
  EVP_PKEY *signer_priv_key = NULL;
  int sign_flags = PKCS7_BINARY;
  BIO *sdata_bio = NULL;
  PKCS7 *p7_sd = NULL;
  unsigned char *mem = NULL;
  int len;

  TRC_API_ENTRY("ccm_pkcs7_sign");

  /***************************************************************************/
  /* Check parameters.  Note that sdata may be NULL if the caller wants an   */
  /* empty SignedData, for example, to sign some attributes but no data.     */
  /***************************************************************************/
  if (subj_name == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for subj_name"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((sdata != NULL) && (sdata_len == 0))
  {
    TRC_ERROR((TB, "Zero length supplied for data to sign"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((digest != CCM_DIGEST_MD5) &&
      (digest != CCM_DIGEST_SHA1) &&
      (digest != CCM_DIGEST_DEFAULT))
  {
    TRC_ERROR((TB, "Unrecognized digest type specified (%u)", digest));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((flags != 0) && (flags != CCM_FLAG_PKCS7_D))
  {
    TRC_ERROR((TB, "Unrecognized flags specified: 0x%x", flags));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((sig_buf == NULL) || (sig_len == NULL))
  {
    TRC_ERROR((TB, "sig_buf and sig_len are both required"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Trace parameters.                                                       */
  /***************************************************************************/
  TRC_NORMAL((TB, "Signing %u bytes of data at %p", sdata_len, sdata));
  TRC_DETAIL((TB, "Caller supplied subj name '%s'", subj_name));
  TRC_DETAIL((TB, "Caller supplied digest %u", digest));
  TRC_DETAIL((TB, "Caller supplied auth_attribs at %p", auth_attribs));
  TRC_DETAIL((TB, "Caller supplied flags 0x%x", flags));
  TRC_DETAIL((TB, "Out: buffer pointer %p, length pointer %p",
                                                            sig_buf, sig_len));

  /***************************************************************************/
  /* We may not have any data to sign if this is a SCEP failure response.    */
  /***************************************************************************/
  if (sdata != NULL)
  {
    /*************************************************************************/
    /* Set up an OpenSSL memory BIO to point to the buffer containing the    */
    /* data to be signed.  It's auto-flagged as read-only.                   */
    /*************************************************************************/
    sdata_bio = BIO_new_mem_buf(sdata, (int)sdata_len);
    if (sdata_bio == NULL)
    {
      OSP_OPENSSL_ERROR((TB, "Failed to create BIO for data to be signed"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Get the certificate and private key that we need.                       */
  /***************************************************************************/
  rc = ccm_int_get_cert(subj_name, &signer_cert);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get local SSL certificate (%u)", rc));
    goto EXIT_LABEL;
  }

  rc = ccm_get_priv_key(subj_name, &signer_priv_key);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get local SSL private key (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Pass on flags as appropriate.                                           */
  /***************************************************************************/
  if (flags & CCM_FLAG_PKCS7_D)
  {
    TRC_NORMAL((TB, "Creating detached SignedData"));
    sign_flags |= PKCS7_DETACHED;
  }

  /***************************************************************************/
  /* Now we can sign the data.                                               */
  /***************************************************************************/
  TRC_NORMAL((TB, "Calling ccm_int_pkcs7_sign()"));
  p7_sd = ccm_int_pkcs7_sign(signer_cert,
                             signer_priv_key,
                             NULL,
                             sdata_bio,
                             sign_flags,
                             digest,
                             auth_attribs);
  if (p7_sd == NULL)
  {
    TRC_ERROR((TB, "Failed to create PKCS#7 SignedData"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now we convert the OpenSSL PKCS7 structure to DER for the caller.       */
  /***************************************************************************/
  len = i2d_PKCS7(p7_sd, NULL);
  mem = (unsigned char *)rm_get_mem(len);
  if (mem == NULL)
  {
    TRC_ERROR((TB, "Couldn't get mem to convert PKCS#7 to ASN.1"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  *sig_buf = (char *)mem;
  *sig_len = i2d_PKCS7(p7_sd, &mem);

  TRC_NORMAL((TB, "Signed data OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (signer_cert != NULL)
  {
    X509_free(signer_cert);
  }

  /***************************************************************************/
  /* Don't free signer_priv_key as it's just a pointer to the permanent      */
  /* copy in our identity list.                                              */
  /***************************************************************************/

  if (sdata_bio != NULL)
  {
    BIO_free(sdata_bio);
  }

  if (p7_sd != NULL)
  {
    PKCS7_free(p7_sd);
  }

  TRC_API_EXIT();

  return(rc);

} /* ccm_pkcs7_sign */

/**PROC+**********************************************************************/
/* Name:      ccm_pkcs7_verify                                               */
/*                                                                           */
/* Purpose:   Verify a PKCS#7 SignedData object                              */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     p7_asn1        - the ASN.1 binary SignedData object     */
/*            IN     p7_asn1_len    - length of SignedData                   */
/*            IN     indata_buf     - input data if SignedData is detached   */
/*            IN     indata_len     - length of input data                   */
/*            IN     flags          - CCM_FLAG_PKCS7_D: detached             */
/*                                    CCM_FLAG_SELF_SIGNED:                  */
/*                                      allow a single self-signed cert      */
/*            OUT    auth_attribs   - stack of authenticated attribs;        */
/*                                      caller may pass null if not required */
/*            OUT    ver_buf        - pointer to verified data (caller must  */
/*                                      free using rm_release_mem())         */
/*            OUT    ver_len        - length of verified data                */
/*            OUT    cert_buf       - the single self-signed cert, if        */
/*                                      there was one                        */
/*            OUT    cert_len       - length of the certificate              */
/*                                                                           */
/* Operation: Set up and clean up around the helper function that does the   */
/*            real work.                                                     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_pkcs7_verify(char *p7_asn1,
                        unsigned int p7_asn1_len,
                        char *indata_buf,
                        unsigned int indata_len,
                        unsigned int flags,
                        STACK_OF(X509_ATTRIBUTE) **auth_attribs,
                        char **ver_buf,
                        unsigned int *ver_len,
                        char **cert_buf,
                        unsigned int *cert_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  PKCS7 *p7 = NULL;
  BIO *inbio = NULL;
  BIO *outbio = NULL;
  unsigned int extra_flags = 0;
  unsigned char *mem = NULL;
  int len;
  STACK_OF(X509) *signer_certs;
  X509 *ss_cert;

  TRC_API_ENTRY("ccm_pkcs7_verify");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if (p7_asn1 == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for p7_asn1"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (p7_asn1_len == 0)
  {
    TRC_ERROR((TB, "Zero length supplied for p7_asn1_len"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((flags != 0) &&
      !(flags & CCM_FLAG_PKCS7_D) &&
      !(flags & CCM_FLAG_SELF_SIGNED))
  {
    TRC_ERROR((TB, "Unrecognized flags specified: 0x%x", flags));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((ver_buf != NULL) && (ver_len == NULL))
  {
    TRC_ERROR((TB, "ver_len is required when ver_buf is specified"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((cert_buf != NULL) && (cert_len == NULL))
  {
    TRC_ERROR((TB, "cert_len is required when cert_buf is specified"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Trace parameters.                                                       */
  /***************************************************************************/
  TRC_NORMAL((TB, "Verifying %u bytes of ASN.1 PKCS#7 at %p",
                                                        p7_asn1_len, p7_asn1));
  TRC_DETAIL((TB, "Caller supplied %u bytes of input data at %p",
                                                      indata_len, indata_buf));
  TRC_DETAIL((TB, "Caller supplied flags 0x%x", flags));
  TRC_DETAIL((TB, "Out: auth attribs %p", auth_attribs));
  TRC_DETAIL((TB, "Out: output data pointer %p, output length pointer %p",
                                                            ver_buf, ver_len));
  TRC_DETAIL((TB, "Out: cert buffer pointer %p, cert length pointer %p",
                                                          cert_buf, cert_len));

  /***************************************************************************/
  /* Translate the ASN.1 to an OpenSSL PKCS7 structure.                      */
  /***************************************************************************/
  p7 = d2i_PKCS7(NULL, (unsigned char **)&p7_asn1, p7_asn1_len);
  if (p7 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to decode ASN.1 PKCS#7 SignedData"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If the signature is detached we have to set up a BIO for the data to    */
  /* be verified.                                                            */
  /***************************************************************************/
  if (flags & CCM_FLAG_PKCS7_D)
  {
    /*************************************************************************/
    /* Set up an OpenSSL memory BIO to point to the buffer containing our    */
    /* data.  It is automatically flagged as read-only.                      */
    /*************************************************************************/
    TRC_NORMAL((TB, "Detached signature"));
    inbio = BIO_new_mem_buf(indata_buf,
                            (int)indata_len);
    if (inbio == NULL)
    {
      OSP_OPENSSL_ERROR((TB, "Failed to create input data BIO"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Now create a new memory BIO to receive the verified data, if the        */
  /* caller wants it.                                                        */
  /***************************************************************************/
  if (ver_buf != NULL)
  {
    TRC_NORMAL((TB, "Creating buffer for verified data"));
    outbio = BIO_new(BIO_s_mem());
    if (outbio == NULL)
    {
      OSP_OPENSSL_ERROR((TB, "Failed to create output BIO"));
      rc = OSP_ERR_NO_RESOURCES;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Pass on flags as appropriate.                                           */
  /***************************************************************************/
  if (flags & CCM_FLAG_SELF_SIGNED)
  {
    TRC_NORMAL((TB, "Allowing self-signed SignedData"));
    extra_flags |= CCM_FLAG_SELF_SIGNED;
  }

  /***************************************************************************/
  /* Now we're ready to verify the data.                                     */
  /***************************************************************************/
  TRC_NORMAL((TB, "Calling ccm_int_pkcs7_verify()"));
  rc = ccm_int_pkcs7_verify(p7,
                            NULL,
                            ccm.cert_store,
                            inbio,
                            outbio,
                            PKCS7_BINARY,
                            extra_flags);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to verify PKCS#7 SignedData"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Pass to the caller the verified data in the output BIO.                 */
  /***************************************************************************/
  if (ver_buf != NULL)
  {
    *ver_len = BIO_get_mem_data(outbio, ver_buf);
    TRC_NORMAL((TB, "Gave caller %u bytes of verified data at %p",
                                                          *ver_len, *ver_buf));
  }

  if (auth_attribs != NULL)
  {
    /*************************************************************************/
    /* The caller also needs the authenticated attributes from the           */
    /* SignedData.  A PKCS#7 SignedData may be signed by multiple entities,  */
    /* each with its own list of authenticated attributes.  Our helper       */
    /* function will set auth_attribs to null if it fails.                   */
    /*************************************************************************/
    TRC_NORMAL((TB, "Getting authenticated attributes"));
    rc = ccm_get_auth_attribs(p7, auth_attribs);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to get authenticated attributes (%u)", rc));
      goto EXIT_LABEL;
    }
  }

  if ((flags & CCM_FLAG_SELF_SIGNED) && (cert_buf != NULL))
  {
    /*************************************************************************/
    /* The caller said they'd accept a single self-signed certificate, and   */
    /* they want a copy of it if that was the case.  This is intended for    */
    /* getting the self-signed certificate from a SCEP PKCSReq.  We will     */
    /* not return anything if there are multiple certificates, self-signed   */
    /* or not.                                                               */
    /*************************************************************************/
    TRC_NORMAL((TB, "Checking for single cert"));
    *cert_buf = NULL;
    signer_certs = p7->d.sign->cert;        /* no macro to get the certs :-( */
    if ((signer_certs != NULL) && (sk_X509_num(signer_certs) == 1))
    {
      /***********************************************************************/
      /* There's one signer certificate.  We're only interested if it's      */
      /* self-signed.                                                        */
      /***********************************************************************/
      TRC_NORMAL((TB, "Single cert: checking for self-signed"));
      ss_cert = sk_X509_value(signer_certs, 0);
      if (X509_NAME_cmp(X509_get_subject_name(ss_cert),
                        X509_get_issuer_name(ss_cert)) == 0)
      {
        /*********************************************************************/
        /* We have something we should return.  We have to convert it to     */
        /* ASN.1 format first though.  Start by allocating the memory.       */
        /*********************************************************************/
        TRC_NORMAL((TB, "Returning single self-signed cert"));
        len = i2d_X509(ss_cert, NULL);
        mem = (unsigned char *)rm_get_mem(len);
        if (mem == NULL)
        {
          TRC_ERROR((TB, "Couldn't get mem to return certificate"));
          rc = OSP_ERR_MEMORY;
          goto EXIT_LABEL;
        }

        *cert_buf = (char *)mem;
        *cert_len = i2d_X509(ss_cert, &mem);
      }
    }
  }

  if (ver_buf != NULL)
  {
    /*************************************************************************/
    /* Set the output BIO's flags to read-only so that its underlying        */
    /* memory is not freed with the BIO -- the caller now owns it.           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Setting output BIO to read-only"));
    BIO_set_flags(outbio, BIO_FLAGS_MEM_RDONLY);
  }

  TRC_NORMAL((TB, "Verified OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Clean up on failure                                                   */
    /*************************************************************************/
    if ((auth_attribs != NULL) && (*auth_attribs != NULL))
    {
      sk_X509_ATTRIBUTE_pop_free(*auth_attribs, X509_ATTRIBUTE_free);
      *auth_attribs = NULL;
    }

    if (ver_buf != NULL)
    {
      /***********************************************************************/
      /* Memory associated with ver_buf will be freed with outbio.           */
      /***********************************************************************/
      *ver_buf = NULL;
    }

    if (ver_len != NULL)
    {
      *ver_len = 0;
    }

    if ((cert_buf != NULL) && (*cert_buf != NULL))
    {
      rm_release_mem(*cert_buf);
      *cert_buf = NULL;
    }

    if (cert_len != NULL)
    {
      *cert_len = 0;
    }
  }

  /***************************************************************************/
  /* Clean up in all cases                                                   */
  /***************************************************************************/
  if (inbio != NULL)
  {
    BIO_free(inbio);
  }

  if (outbio != NULL)
  {
    BIO_free(outbio);
  }

  if (p7 != NULL)
  {
    PKCS7_free(p7);
  }

  TRC_API_EXIT();

  return(rc);

} /* ccm_pkcs7_verify */

/**PROC+**********************************************************************/
/* Name:      ccm_pkcs7_encrypt                                              */
/*                                                                           */
/* Purpose:   Encrypt some data to produce a PKCS#7 EnvelopedData object     */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     subj_name      - the subject name to encrypt for        */
/*            IN     cert           - cert to use (if subj_name is NULL)     */
/*            IN     cert_len       - length of cert (if subj_name is NULL)  */
/*            IN     edata          - data to encrypt                        */
/*            IN     edata_len      - number of bytes to encrypt             */
/*            IN     cipher         - CCM_CIPHER_DES_CBC: DES-CBC            */
/*                                    CCM_CIPHER_RC2_CBC: RC2-CBC            */
/*                                    CCM_CIPHER_DEFAULT: default            */
/*            OUT    enc_buf        - resulting PKCS#7 EnvelopedData         */
/*            OUT    enc_len        - length of EnvelopedData                */
/*                                                                           */
/* Operation: Set up and clean up around the OpenSSL function that does the  */
/*            real work.                                                     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_pkcs7_encrypt(char *subj_name,
                         char *cert,
                         unsigned int cert_len,
                         char *edata,
                         unsigned int edata_len,
                         unsigned int cipher,
                         char **enc_buf,
                         unsigned int *enc_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  STACK_OF(X509) *recip_certs = NULL;
  X509 *recip_cert = NULL;
  PKCS7 *p7_ed = NULL;
  BIO *edata_bio = NULL;
  EVP_CIPHER *evp_cipher;
  unsigned char *mem = NULL;
  int len;

  TRC_API_ENTRY("ccm_pkcs7_encrypt");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if ((subj_name == NULL) &&
      ((cert == NULL) || (cert_len == 0)))
  {
    TRC_ERROR((TB, "cert and cert_len both required when subj_name is null"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (edata == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for data to encrypt"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (edata_len == 0)
  {
    TRC_ERROR((TB, "Zero length supplied for data to encrypt"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((cipher != CCM_CIPHER_DES_CBC) &&
      (cipher != CCM_CIPHER_RC2_CBC))
  {
    TRC_ERROR((TB, "Unrecognized cipher specified (%u)", cipher));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((enc_buf == NULL) || (enc_len == NULL))
  {
    TRC_ERROR((TB, "enc_buf and enc_len are both required"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Trace parameters.                                                       */
  /***************************************************************************/
  TRC_NORMAL((TB, "Encrypting %u bytes of data at %p", edata_len, edata));
  if (subj_name != NULL)
  {
    TRC_DETAIL((TB, "Caller supplied subj name '%s'", subj_name));
  }
  else
  {
    TRC_DETAIL((TB, "Caller supplied %u-byte cert at %p", cert_len, cert));
  }
  TRC_DETAIL((TB, "Caller supplied cipher %u", cipher));
  TRC_DETAIL((TB, "Out: buffer pointer %p, length pointer %p",
                                                            enc_buf, enc_len));

  /***************************************************************************/
  /* Create a stack to store the recipient's certificate.  This is a stack   */
  /* because PKCS#7 EnvelopedData objects can be encrypted for multiple      */
  /* recipients, although we only support one.                               */
  /***************************************************************************/
  recip_certs = sk_X509_new(NULL);
  if (recip_certs == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create stack for recipient cert"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Look up the certificate if a subject name is specified; otherwise use   */
  /* the supplied certificate.                                               */
  /***************************************************************************/
  if (subj_name != NULL)
  {
    TRC_NORMAL((TB, "Getting cert for '%s'", subj_name));
    rc = ccm_int_get_cert(subj_name, &recip_cert);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to find cert (%u) for '%s'", rc, subj_name));
      goto EXIT_LABEL;
    }
  }
  else
  {
    /*************************************************************************/
    /* Convert the cert from ASN.1 format.                                   */
    /*************************************************************************/
    TRC_NORMAL((TB, "Using cert at %p (%u bytes)", cert, cert_len));
    recip_cert = d2i_X509(NULL, (unsigned char **)&cert, cert_len);
    if (recip_cert == NULL)
    {
      OSP_OPENSSL_ERROR((TB, "Failed to convert ASN.1 to X.509"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Add the recipient's certificate to the stack.                           */
  /***************************************************************************/
  if (sk_X509_push(recip_certs, recip_cert) == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to add recipient certificate to stack"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up an OpenSSL memory BIO to point to the buffer containing the      */
  /* data to be signed.  It's auto-flagged as read-only.                     */
  /***************************************************************************/
  edata_bio = BIO_new_mem_buf(edata, (int)edata_len);
  if (edata_bio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create BIO for data to be encrypted"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get the appropriate cipher.                                             */
  /***************************************************************************/
  if (cipher == CCM_CIPHER_DES_CBC)
  {
    TRC_NORMAL((TB, "Using DES-CBC cipher for encryption"));
    evp_cipher = EVP_des_cbc();
  }
  else
  {
    TRC_NORMAL((TB, "Using RC2-CBC cipher for encryption"));
    evp_cipher = EVP_rc2_cbc();
  }

  /***************************************************************************/
  /* Now we're ready to encrypt.                                             */
  /***************************************************************************/
  TRC_NORMAL((TB, "Calling PKCS7_encrypt()"));
  p7_ed = PKCS7_encrypt(recip_certs,
                        edata_bio,
                        evp_cipher,
                        PKCS7_BINARY);
  if (p7_ed == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create PKCS#7 SignedData"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now we convert the OpenSSL PKCS7 structure to DER for the caller.       */
  /***************************************************************************/
  len = i2d_PKCS7(p7_ed, NULL);
  mem = (unsigned char *)rm_get_mem(len);
  if (mem == NULL)
  {
    TRC_ERROR((TB, "Couldn't get mem to convert PKCS#7 to ASN.1"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  *enc_buf = (char *)mem;
  *enc_len = i2d_PKCS7(p7_ed, &mem);

  TRC_NORMAL((TB, "Encrypted data OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (p7_ed != NULL)
  {
    PKCS7_free(p7_ed);
  }

  if (recip_certs != NULL)
  {
    sk_X509_free(recip_certs);
  }

  if (recip_cert != NULL)
  {
    X509_free(recip_cert);
  }

  if (edata_bio != NULL)
  {
    BIO_free(edata_bio);
  }

  TRC_API_EXIT();

  return(rc);

} /* ccm_pkcs7_encrypt */

/**PROC+**********************************************************************/
/* Name:      ccm_pkcs7_decrypt                                              */
/*                                                                           */
/* Purpose:   Decrypt a PKCS#7 EnvelopedData object                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     p7_asn1        - the ASN.1 binary SignedData object     */
/*            IN     p7_asn1_len    - length of SignedData                   */
/*            IN     subj_name      - DN of identity to decrypt as           */
/*            OUT    dec_buf        - pointer to decrypted data (caller must */
/*                                      free with rm_release_mem())          */
/*            OUT    dec_len        - length of decrypted data               */
/*                                                                           */
/* Operation: Set up and clean up around the OpenSSL function that does the  */
/*            real work.                                                     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_pkcs7_decrypt(char *p7_asn1,
                         unsigned int p7_asn1_len,
                         char *subj_name,
                         char **dec_buf,
                         unsigned int *dec_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  PKCS7 *p7 = NULL;
  BIO *outbio = NULL;
  EVP_PKEY *priv_key = NULL;
  X509 *cert = NULL;
  int result;

  TRC_API_ENTRY("ccm_pkcs7_decrypt");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if (p7_asn1 == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for p7_asn1"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (p7_asn1_len == 0)
  {
    TRC_ERROR((TB, "Zero length supplied for p7_asn1_len"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (subj_name == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for subj_name"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if ((dec_buf == NULL) || (dec_len == NULL))
  {
    TRC_ERROR((TB, "dec_buf and dec_len are both required"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Trace parameters.                                                       */
  /***************************************************************************/
  TRC_NORMAL((TB, "Decrypting %u bytes of ASN.1 PKCS#7 at %p",
                                                        p7_asn1_len, p7_asn1));
  TRC_DETAIL((TB, "Caller supplied subj name '%s'", subj_name));
  TRC_DETAIL((TB, "Out: buffer pointer %p, length pointer %p",
                                                            dec_buf, dec_len));

  /***************************************************************************/
  /* Translate the ASN.1 to an OpenSSL PKCS7 structure.                      */
  /***************************************************************************/
  p7 = d2i_PKCS7(NULL, (unsigned char **)&p7_asn1, p7_asn1_len);
  if (p7 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to decode ASN.1 PKCS#7 EnvelopedData"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Find the certificate corresponding to the decrypting identity.  This    */
  /* should not block as this identity should be in the identity list so     */
  /* we'll have already retrieved its certificate at startup.                */
  /***************************************************************************/
  rc = ccm_int_get_cert(subj_name, &cert);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to find certificate (%u) for '%s'", rc, subj_name));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Find the private key corresponding to the decrypting identity.          */
  /***************************************************************************/
  rc = ccm_get_priv_key(subj_name, &priv_key);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to find private key (%u) for '%s'", rc, subj_name));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create a new memory BIO to receive the decrypted data.                  */
  /***************************************************************************/
  outbio = BIO_new(BIO_s_mem());
  if (outbio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create output BIO"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now we're ready to decrypt the data.                                    */
  /***************************************************************************/
  TRC_NORMAL((TB, "Calling PKCS7_decrypt()"));
  result = PKCS7_decrypt(p7,
                         priv_key,
                         cert,
                         outbio,
                         0);
  if (result == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to decrypt PKCS#7 EnvelopedData"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Pass to the caller the verified data in the output BIO.                 */
  /***************************************************************************/
  *dec_len = BIO_get_mem_data(outbio, dec_buf);

  /***************************************************************************/
  /* Set the output BIO's flags to read-only so that its underlying memory   */
  /* is not freed with the BIO -- the caller now owns it.                    */
  /***************************************************************************/
  BIO_set_flags(outbio, BIO_FLAGS_MEM_RDONLY);

  TRC_NORMAL((TB, "Decrypted OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    *dec_buf = NULL;          /* associated memory will be freed with outbio */
    *dec_len = 0;
  }

  if (outbio != NULL)
  {
    TRC_DETAIL((TB, "Freeing bio %p", outbio));
    BIO_free(outbio);
  }

  if (cert != NULL)
  {
    X509_free(cert);
  }

  /***************************************************************************/
  /* Don't free priv_key as it's just a pointer to the permanent copy in     */
  /* our identity list.                                                      */
  /***************************************************************************/

  if (p7 != NULL)
  {
    PKCS7_free(p7);
  }

  TRC_API_EXIT();

  return(rc);

} /* ccm_pkcs7_decrypt */

