
/**MOD+***********************************************************************/
/* Module:    ccmrand.c                                                      */
/*                                                                           */
/* Purpose:   CCM random number functions                                    */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.5                $ $Modtime::   Jan 11 2001 15:56:32   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_random                                                     */
/*                                                                           */
/* Purpose:   Fill the supplied buffer with random bytes                     */
/*                                                                           */
/* Returns:   None                                                           */
/*                                                                           */
/* Params:    IN     buffer         - the buffer to fill                     */
/*            IN     buf_len        - the number of bytes in the buffer      */
/*                                                                           */
/* Operation: Uses OpenSSL's RAND API. This can be set to point to other     */
/*            random number generators by calling OpenSSL's                  */
/*            RAND_set_rand_method() directly.                               */
/*                                                                           */
/**PROC-**********************************************************************/

void ccm_random(void *buffer,
                int buf_len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int result;

  TRC_ENTRY("ccm_random");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  if (buffer == NULL)
  {
    TRC_ERROR((TB, "Null pointer supplied for buffer"));
    goto EXIT_LABEL;
  }

  if (buf_len < 0)
  {
    TRC_ERROR((TB, "Negative value supplied for buf_len (%d)", buf_len));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get the random bytes (a no-op if buf_len is zero).                      */
  /***************************************************************************/
  if (buf_len > 0)
  {
    result = RAND_bytes(buffer, buf_len);
    if (result != 1)
    {
      OSP_OPENSSL_ERROR((TB, "Failed RAND_bytes(%p, %d) (%d)",
                                                     buffer, buf_len, result));
      TRC_ASSERT((FALSE), (TB, "RAND_bytes should always succeed"));
      goto EXIT_LABEL;
    }
  }

  TRC_NORMAL((TB, "Returning %d random bytes at %p", buf_len, buffer));

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* ccm_random */

/**PROC+**********************************************************************/
/* Name:      ccm_read_random_state                                          */
/*                                                                           */
/* Purpose:   Read the random state from a file                              */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Asks OpenSSL what file to use, then tells it to read its state */
/*            from that file.                                                */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_read_random_state(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int result;

  TRC_ENTRY("ccm_read_random_state");

  /***************************************************************************/
  /* Read the data.                                                          */
  /***************************************************************************/
  result = RAND_load_file(ccm.rand_file, -1);
  if (result < CCM_RAND_FILE_LEN_MIN)
  {
    OSP_OPENSSL_ERROR((TB, "Not enough data (%d bytes; min=%d) in random"
             "state file '%s'", result, CCM_RAND_FILE_LEN_MIN, ccm.rand_file));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Seeded random number generator with %d bytes from '%s'",
                                                       result, ccm.rand_file));
  ccm.prng_seeded = TRUE;
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* ccm_read_random_state */

/**PROC+**********************************************************************/
/* Name:      ccm_write_random_state                                         */
/*                                                                           */
/* Purpose:   Write the random state to a file                               */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Asks OpenSSL what file to use, then tells it to write its      */
/*            state to that file.  This is only done if the state was read   */
/*            successfully.                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_write_random_state(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int result;

  TRC_ENTRY("ccm_write_random_state");

  /***************************************************************************/
  /* Only do this if we seeded the random number generator successfully.  We */
  /* will return OSP_ERROR if this is not the case.                          */
  /***************************************************************************/
  if (ccm.prng_seeded)
  {
    /*************************************************************************/
    /* Write the data.                                                       */
    /*************************************************************************/
    result = RAND_write_file(ccm.rand_file);
    if (result <= 0)
    {
      OSP_OPENSSL_ERROR((TB, "Failed to save random state to '%s'",
                                                               ccm.rand_file));
      rc = OSP_ERROR;
    }

    TRC_NORMAL((TB, "Wrote %d bytes random state to '%s'",
                                                       result, ccm.rand_file));
    rc = OSP_SUCCESS;
  }
  else
  {
    TRC_ALERT((TB, "Random number generator wasn't seeded: not saving state"));
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* ccm_write_random_state */

