#ifndef CCMSTRC_INCLUDED
#define CCMSTRC_INCLUDED

/**INC+***********************************************************************/
/* Header:    ccmstrc.h                                                      */
/*                                                                           */
/* Purpose:   CCM structure definitions                                      */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.8                $ $Modtime::   Jan 11 2001 15:56:34   $  */
/*                                                                           */
/**INC-***********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: CCM_DATA                                                       */
/*                                                                           */
/* Description: CCM internal data structure                                  */
/*****************************************************************************/

typedef struct ccm_data
{
  /***************************************************************************/
  /* CCM's certificate and CRL store (an OpenSSL type)                       */
  /***************************************************************************/
  X509_STORE *cert_store;

  /***************************************************************************/
  /* Handle to information read from config file (used for setting           */
  /* extensions on issued certificates).                                     */
  /***************************************************************************/
  LHASH *config;

  /***************************************************************************/
  /* The name of the random number seed file.                                */
  /***************************************************************************/
  char *rand_file;

  /***************************************************************************/
  /* Flag for whether the random number generator was correctly seeded at    */
  /* startup.                                                                */
  /***************************************************************************/
  int prng_seeded;

  /***************************************************************************/
  /* Flag to indicate whether or not LAM was successfully initialized.       */
  /***************************************************************************/
  int lam_initialized;

  /***************************************************************************/
  /* Values relating to our CA functionality.                                */
  /***************************************************************************/
  struct
  {
    /*************************************************************************/
    /* The file that holds the serial number to use for the next             */
    /* newly-issued certificate.                                             */
    /*************************************************************************/
    char *serial_file;

    /*************************************************************************/
    /* Our local copy of the serial number.                                  */
    /*************************************************************************/
    BIGNUM *serial_num;

    /*************************************************************************/
    /* A mutex which must be held to access serial_num.                      */
    /*************************************************************************/
    pthread_mutex_t serial_num_mutex;

    /*************************************************************************/
    /* Number of days for new certificate validity.                          */
    /*************************************************************************/
    unsigned int valid_duration;

    /*************************************************************************/
    /* Signing digest to use when issuing RSA certificates.                  */
    /*************************************************************************/
    EVP_MD *rsa_sign_digest;

    /*************************************************************************/
    /* Additional name components to add to the subject name in a request.   */
    /*************************************************************************/
    X509_NAME *subject_suffix;

  } ca;

  /***************************************************************************/
  /* LDAP-related values, all read from the configuration file.              */
  /***************************************************************************/
  struct
  {
    /*************************************************************************/
    /* Host name of LDAP directory.                                          */
    /*************************************************************************/
    char *host;

    /*************************************************************************/
    /* Port number of LDAP directory.                                        */
    /*************************************************************************/
    int port;

    /*************************************************************************/
    /* User name to access LDAP directory.                                   */
    /*************************************************************************/
    char *user;

    /*************************************************************************/
    /* Password to access LDAP directory.                                    */
    /*************************************************************************/
    char *password;

    /*************************************************************************/
    /* Suffix to add to DNs before doing an LDAP search.                     */
    /*************************************************************************/
    char *dn_suffix;

    /*************************************************************************/
    /* Handle to directory, returned from ldap_init().                       */
    /*************************************************************************/
    LDAP *handle;

    /*************************************************************************/
    /* Mutex to serialise LDAP access.                                       */
    /*************************************************************************/
    pthread_mutex_t access_mutex;

  } ldap;

  /***************************************************************************/
  /* SSL-related values.                                                     */
  /***************************************************************************/
  struct
  {
    /*************************************************************************/
    /* The verification callback to use for SSL cert verification (from SM). */
    /*************************************************************************/
    PCCM_SSL_VERIFY_CALLBACK verify_cb;

    /*************************************************************************/
    /* The distinguished name of the identity to use for SSL (from config    */
    /* file).                                                                */
    /*************************************************************************/
    char *identity;

    /*************************************************************************/
    /* The mode to use for client verification (from config file).           */
    /*************************************************************************/
    int verify_mode;

    /*************************************************************************/
    /* The SSL cipher list to use (from config file).                        */
    /*************************************************************************/
    char *cipher_list;

  } ssl;

  /***************************************************************************/
  /* Identity-related values (indirectly from config file).                  */
  /***************************************************************************/
  struct
  {
    /*************************************************************************/
    /* Number of identities.                                                 */
    /*************************************************************************/
    unsigned int num;

    /*************************************************************************/
    /* Array of pointers to distinguished names of identities.               */
    /*************************************************************************/
    X509_NAME **names;

    /*************************************************************************/
    /* Array of pointers to private keys of identities.                      */
    /*************************************************************************/
    EVP_PKEY **privkeys;

  } identities;

  /***************************************************************************/
  /* Temporary keys that OpenSSL may ask us for.                             */
  /***************************************************************************/
  struct
  {
    /*************************************************************************/
    /* A 512-bit temporary RSA key.                                          */
    /*************************************************************************/
    RSA *rsa512;

    /*************************************************************************/
    /* A 1024-bit temporary RSA key.                                         */
    /*************************************************************************/
    RSA *rsa1024;

    /*************************************************************************/
    /* A 512-bit temporary DH key.                                           */
    /*************************************************************************/
    DH *dh512;

    /*************************************************************************/
    /* A 1024-bit temporary DH key.                                          */
    /*************************************************************************/
    DH *dh1024;

  } tempkeys;

} CCM_DATA;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: CCM_NEW_CERT_CONTEXT                                           */
/*                                                                           */
/* Description:  Holds information that is passed to a CCM thread when it    */
/*               issues a new certificate.                                   */
/*****************************************************************************/

typedef struct ccm_new_cert_context
{
  /***************************************************************************/
  /* The caller's correlator.                                                */
  /***************************************************************************/
  void *corr;

  /***************************************************************************/
  /* The callback function.                                                  */
  /***************************************************************************/
  PCCM_NEW_CERT_CALLBACK callback;

  /***************************************************************************/
  /* The certificate request.                                                */
  /***************************************************************************/
  X509_REQ *req;

  /***************************************************************************/
  /* The public key from the request.                                        */
  /***************************************************************************/
  EVP_PKEY *pub_key;

  /***************************************************************************/
  /* The MD5 digest of the request's public key.                             */
  /***************************************************************************/
  unsigned char digest[MD5_DIGEST_LENGTH];

  /***************************************************************************/
  /* The CA identity to sign as.                                             */
  /***************************************************************************/
  char *ca_name;

  /***************************************************************************/
  /* The CA certificate (looked up from ca_name).                            */
  /***************************************************************************/
  X509 *ca_cert;

  /***************************************************************************/
  /* The subject name from the request, as a plain string.                   */
  /***************************************************************************/
  char *subj_name;

  /***************************************************************************/
  /* The newly-issued certificate.                                           */
  /***************************************************************************/
  X509 *new_cert;

} CCM_NEW_CERT_CONTEXT;

/**STRUCT-********************************************************************/

#endif

