
/**MOD+***********************************************************************/
/* Module:    ccmutil.c                                                      */
/*                                                                           */
/* Purpose:   CCM utility functions                                          */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.22               $ $Modtime::   Jan 11 2001 15:56:34   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_get_auth_attribs                                           */
/*                                                                           */
/* Purpose:   Collect all the authenticated attributes from all the signers  */
/*            in a PKCS#7 SignedData into one stack                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     p7             - the PKCS#7 SignedData structure        */
/*            OUT    all_attribs    - the stack of collected attributes      */
/*                                      or NULL if none. Caller must free.   */
/*                                                                           */
/* Operation: Loop through all the signerinfo blocks and collect all their   */
/*            signed attributes into one stack, which is returned.           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_get_auth_attribs(PKCS7 *p7,
                            STACK_OF(X509_ATTRIBUTE) **attribs_out)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  STACK_OF(PKCS7_SIGNER_INFO) *sinfos;
  int num_sinfos;
  STACK_OF(X509_ATTRIBUTE) *sig_attribs;
  int num_sig_attribs;
  X509_ATTRIBUTE *new_attrib;
  STACK_OF(X509_ATTRIBUTE) *all_attribs;
  int ii, jj;
  int error;

  TRC_ENTRY("ccm_get_auth_attribs");

  /***************************************************************************/
  /* Allocate a new stack for the collected attributes.                      */
  /***************************************************************************/
  all_attribs = sk_X509_ATTRIBUTE_new(NULL);
  if (all_attribs == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create stack for attribute collection"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get the stack of signer infos.                                          */
  /***************************************************************************/
  sinfos = PKCS7_get_signer_info(p7);
  if ((sinfos == NULL) ||
      (sk_PKCS7_SIGNER_INFO_num(sinfos) == 0))
  {
    OSP_OPENSSL_ERROR((TB, "Couldn't get signer info"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Loop through the stack of signer infos.                                 */
  /***************************************************************************/
  num_sinfos = sk_PKCS7_SIGNER_INFO_num(sinfos);
  for (ii = 0; ii < num_sinfos; ii++)
  {
    /*************************************************************************/
    /* Get this signer's stack of auth attribs.                              */
    /*************************************************************************/
    sig_attribs =
          PKCS7_get_signed_attributes(sk_PKCS7_SIGNER_INFO_value(sinfos, ii));
    if ((sig_attribs == NULL) ||
        (sk_X509_ATTRIBUTE_num(sig_attribs) == 0))
    {
      /***********************************************************************/
      /* It's not a PKCS#7 error if the signer has no auth attribs.          */
      /***********************************************************************/
      TRC_NORMAL((TB, "No authenticated attributes for signer %d", ii));
      continue;
    }

    /*************************************************************************/
    /* Loop through the stack of authenticated attributes.                   */
    /*************************************************************************/
    num_sig_attribs = sk_X509_ATTRIBUTE_num(sig_attribs);
    for (jj = 0; jj < num_sig_attribs; jj++)
    {
      /***********************************************************************/
      /* Duplicate the attribute so that it isn't dependent on the PKCS#7    */
      /* structure staying around.                                           */
      /***********************************************************************/
      new_attrib = X509_ATTRIBUTE_dup(sk_X509_ATTRIBUTE_value(sig_attribs,
                                                              jj));
      if (new_attrib == NULL)
      {
        OSP_OPENSSL_ERROR((TB, "Failed to duplicate attrib %d from signer %d",
                                                                      jj, ii));
        rc = OSP_ERR_NO_RESOURCES;
        goto EXIT_LABEL;
      }

      error = sk_X509_ATTRIBUTE_push(all_attribs, new_attrib);
      if (error == 0)
      {
        OSP_OPENSSL_ERROR((TB, "Failed to push attrib %d from signer %d",
                                                                      jj, ii));
        rc = OSP_ERROR;
        goto EXIT_LABEL;
      }
    }
  }

  TRC_NORMAL((TB, "Built list of %d authenticated attribs at %p",
                                            sk_X509_ATTRIBUTE_num(all_attribs),
                                            all_attribs));

  if (sk_X509_ATTRIBUTE_num(all_attribs) == 0)
  {
    /*************************************************************************/
    /* The stack is empty.  We guarantee that any returned stack will have   */
    /* at least one item in it, so we'll free the empty stack & return null  */
    /*************************************************************************/
    sk_X509_ATTRIBUTE_pop_free(all_attribs, X509_ATTRIBUTE_free);
    all_attribs = NULL;
  }

  *attribs_out = all_attribs;
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* The caller owns the stack on success, so we only free it on failure.    */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    if (all_attribs != NULL)
    {
      sk_X509_ATTRIBUTE_pop_free(all_attribs, X509_ATTRIBUTE_free);
    }
    *attribs_out = NULL;
  }

  TRC_EXIT();

  return(rc);

} /* ccm_get_auth_attribs */

/**PROC+**********************************************************************/
/* Name:      ccm_rfc2253_to_x509_name                                       */
/*                                                                           */
/* Purpose:   Convert an LDAP-style name string into an OpenSSL X509_NAME    */
/*            structure                                                      */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     dn_string      - the distinguished name to convert      */
/*            OUT    x509_name      - the resulting X509_NAME structure.     */
/*                                      Caller must free.                    */
/*                                                                           */
/* Operation: Parse an LDAP-style comma-separated DN of the form:            */
/*            CN=John Doe,O=Data Connection Ltd,C=GB                         */
/*            Note that this conversion is very basic and doesn't support    */
/*            quotes, escapes or hex format. It does, however, support       */
/*            multi-valued RDNs having a '+' separating each type/val pair.  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_rfc2253_to_x509_name(char *dn_string,
                                X509_NAME **x509_name)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned int parsed_ok = 0;
  X509_NAME_ENTRY *entry = NULL;
  char *str = NULL;
  char *cc;
  char *type;
  char *value;
  int rdn_flag_next = 0;
  int rdn_flag_this;
  int prev_set;
  int new_set;
  int ii;
  int error;

  TRC_ENTRY("ccm_rfc2253_to_x509_name");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((dn_string != NULL), (TB, "dn_string is required"));
  TRC_ASSERT((x509_name != NULL), (TB, "x509_name is required"));

  /***************************************************************************/
  /* Create the X509_NAME structure.                                         */
  /***************************************************************************/
  *x509_name = X509_NAME_new();
  if (*x509_name == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create X509_NAME structure"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Duplicate the string because we will be inserting nulls into it.        */
  /***************************************************************************/
  OSP_STRDUP(str, dn_string);
  if (str == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for local copy of string"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Search through the string, extracting each type/value pair and adding   */
  /* them to the X509_NAME structure.                                        */
  /***************************************************************************/
  TRC_NORMAL((TB, "Parsing DN string: '%s'", dn_string));
  cc = str;
  while (TRUE)
  {
    /*************************************************************************/
    /* Look for the start of an attributeType.                               */
    /*************************************************************************/
    while (isspace((int)*cc))
    {
      cc++;
    }

    /*************************************************************************/
    /* Check that it's valid; remember where it starts if so.                */
    /*************************************************************************/
    if (!isalpha((int)*cc))
    {
      TRC_ERROR((TB, "Invalid char in attributeType"));
      break;
    }
    type = cc;
    cc++;

    /*************************************************************************/
    /* Look for the end of the attributeType and insert nulls until we find  */
    /* an equals sign.                                                       */
    /*************************************************************************/
    while (isalpha((int)*cc) || isdigit((int)*cc) || (*cc == '-'))
    {
      cc++;
    }

    if (*cc == 0)
    {
      TRC_ERROR((TB, "Nothing after attributeType"));
      break;
    }

    while (isspace((int)*cc))
    {
      *cc = 0;
      cc++;
    }

    if (*cc != '=')
    {
      TRC_ERROR((TB, "Expecting '='"));
      break;
    }
    *cc = 0;
    cc++;

    /*************************************************************************/
    /* Look for the start of an attributeValue.  Note that we will fail to   */
    /* parse an attribute with nothing in it, eg CN=,O=DCL,C=GB.             */
    /*************************************************************************/
    while (isspace((int)*cc))
    {
      cc++;
    }
    value = cc;

    /*************************************************************************/
    /* Look for the end of the attributeValue and insert a null.  If that's  */
    /* the end of the string, set cc to null so that we exit the loop after  */
    /* adding this last value.                                               */
    /*************************************************************************/
    while (strchr(",;+", (int)*cc) == NULL)
    {
      cc++;
    }

    rdn_flag_this = rdn_flag_next;
    if (*cc == 0)
    {
      cc = NULL;
    }
    else
    {
      /***********************************************************************/
      /* Work out whether the next attributeType/Value pair will be part of  */
      /* the same RDN ('+' separator) or a new RDN (',' or ';' separator).   */
      /***********************************************************************/
      if (*cc == '+')
      {
        rdn_flag_next = -1;
      }
      else
      {
        rdn_flag_next = 0;
      }

      *cc = 0;
      cc++;
    }

    /*************************************************************************/
    /* Add the attribute to the X509_NAME structure.                         */
    /*************************************************************************/
    TRC_DETAIL((TB, "Adding %s=%s to X509_NAME (rdn_flag_this=%d)",
                                                  type, value, rdn_flag_this));
    error = X509_NAME_add_entry_by_txt(*x509_name,
                                       type,
                                       MBSTRING_ASC,
                                       (unsigned char *)value,
                                       -1, -1, rdn_flag_this);
    if (error == 0)
    {
      OSP_OPENSSL_ERROR((TB, "Error adding '%s=%s' to X509_NAME",
                                                                 type, value));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    if (cc == NULL)
    {
      TRC_NORMAL((TB, "Finished parsing DN; %d entries",
                                           X509_NAME_entry_count(*x509_name)));
      parsed_ok = 1;
      break;
    }
  }

  /***************************************************************************/
  /* We've broken out of the loop.  If parsed_ok==0 or there are no entries  */
  /* in the name structure, it's an error.                                   */
  /***************************************************************************/
  if ((parsed_ok == 0) || (X509_NAME_entry_count(*x509_name) == 0))
  {
    TRC_ERROR((TB, "Failed to parse DN string '%s'; error at offset %d",
                                                         dn_string, cc - str));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now reverse the order of the entries in the X509_NAME structure, so     */
  /* that the ordering agrees with X.501 (i.e.  opposite to RFC2253).  We    */
  /* can't construct the list of entries backwards in the first place,       */
  /* because the logic in X509_NAME_add_entry() doesn't allow us to set up   */
  /* 'set' correctly for each entry.                                         */
  /***************************************************************************/
  TRC_DETAIL((TB, "Reversing name entries"));
  prev_set = -1;
  new_set = -1;
  for (ii = (X509_NAME_entry_count(*x509_name) - 1); ii >= 0 ; ii--)
  {
    entry = sk_X509_NAME_ENTRY_delete((*x509_name)->entries, ii);
    if (entry->set != prev_set)
    {
      prev_set = entry->set;
      new_set++;
    }
    TRC_DETAIL((TB, "Entry %d set: old %d, new %d", ii, entry->set, new_set));
    entry->set = new_set;
    sk_X509_NAME_ENTRY_push((*x509_name)->entries, entry);
  }

#ifdef CCM_DEBUG_RFC2253
  {
    /*************************************************************************/
    /* Trace to check conversion by translating back to RFC2253 format.      */
    /*************************************************************************/
    char *rfc2253;

    ccm_x509_name_to_rfc2253(*x509_name, &rfc2253);
    printf("Converted name to: '%s'\n", rfc2253);
    rm_release_mem(rfc2253);
  }
#endif

  TRC_DETAIL((TB, "Converted string to X509_NAME OK", dn_string));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    if (*x509_name != NULL)
    {
      X509_NAME_free(*x509_name);
      *x509_name = NULL;
    }
  }

  if (str != NULL)
  {
    rm_release_mem(str);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_rfc2253_to_x509_name */

/**PROC+**********************************************************************/
/* Name:      ccm_x509_name_to_rfc2253                                       */
/*                                                                           */
/* Purpose:   Convert an X509_NAME structure to an LDAP-style (RFC2253) DN,  */
/*            with comma separation and the highest-level component last.    */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     x509_name      - the X509_NAME structure to convert     */
/*            OUT    name_out       - a buffer containing the DN string      */
/*                                      (caller must free)                   */
/*                                                                           */
/* Operation: Similar to the OpenSSL function X509_NAME_oneline(), but       */
/*            creates an RFC2253-style DN. This function doesn't provide for */
/*            the EBCDIC character set or 10646 coding. See RFC2253 for      */
/*            details of this coding method.                                 */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_x509_name_to_rfc2253(X509_NAME *x509_name, char **name_out)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  char *buf = NULL;
  char *new_buf = NULL;
  int buf_size = 200;
  int ii, jj;
  X509_NAME_ENTRY *name_entry = NULL;
  int nid;
  const char *short_name;
  int short_name_len;
  char *ne_data;
  int ne_data_len;
  char *dest_ptr;
  int is_new_set;
  int prev_set = -1;
  int first_entry = TRUE;
  int data_txt_length;
  int total_txt_length = 0;
  int copied_length;
  char temp_buf[1024];

  TRC_ENTRY("ccm_x509_name_to_rfc2253");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((x509_name != NULL), (TB, "x509_name can't be null"));
  TRC_ASSERT((name_out != NULL), (TB, "name_out can't be null"));

  /***************************************************************************/
  /* Initialize.                                                             */
  /***************************************************************************/
  *name_out = NULL;
  buf = (char *)rm_get_mem(buf_size);
  if (buf == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate %d bytes for LDAP DN", buf_size));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }
  *buf = 0;

  TRC_DETAIL((TB, "Converting X509_NAME at %p to LDAP DN string", x509_name));
  TRC_DETAIL((TB, "Name has %d entries", X509_NAME_entry_count(x509_name)));

  /***************************************************************************/
  /* Loop through the name entry list, last first.                           */
  /***************************************************************************/
  for (ii = (X509_NAME_entry_count(x509_name) - 1); ii >= 0 ; ii--)
  {
    /*************************************************************************/
    /* Get a pointer to the name entry.                                      */
    /*************************************************************************/
    name_entry = X509_NAME_get_entry(x509_name, ii);
    if (name_entry == NULL)
    {
      OSP_OPENSSL_ERROR((TB, "Failed to get entry %d from X509_NAME at %p",
                                                               ii, x509_name));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Get a textual name for this entry.                                    */
    /*************************************************************************/
    nid = OBJ_obj2nid(X509_NAME_ENTRY_get_object(name_entry));
    short_name = OBJ_nid2sn(nid);
    if ((nid == NID_undef) || (short_name == NULL))
    {
      OBJ_obj2txt(temp_buf, sizeof(temp_buf),
                                    X509_NAME_ENTRY_get_object(name_entry), 0);
      short_name = temp_buf;
    }
    short_name_len = strlen(short_name);

    /*************************************************************************/
    /* Get information about the entry's data.                               */
    /*************************************************************************/
    ne_data_len = ASN1_STRING_length(X509_NAME_ENTRY_get_data(name_entry));
    ne_data = (char *)ASN1_STRING_data(X509_NAME_ENTRY_get_data(name_entry));
    TRC_DETAIL((TB, "Name entry %d: '%s'; data at %p len %d; set %d",
                       ii, short_name, ne_data, ne_data_len, name_entry->set));

    /*************************************************************************/
    /* See if this entry begins a new RDN set, in which case we will use a   */
    /* comma to delimit it instead of a plus sign.                           */
    /*************************************************************************/
    if (name_entry->set != prev_set)
    {
      is_new_set = TRUE;
      prev_set = name_entry->set;
    }
    else
    {
      is_new_set = FALSE;
    }

    /*************************************************************************/
    /* Find space required to store entry data.                              */
    /*************************************************************************/
    data_txt_length = 0;
    for (jj = 0; jj < ne_data_len; jj++)
    {
      data_txt_length++;

      if (ccm_must_escape(ne_data[jj], jj, ne_data_len))
      {
        data_txt_length += 2;
      }
    }

    /*************************************************************************/
    /* Add required space for this entry to the total string size.           */
    /*************************************************************************/
    copied_length = total_txt_length;
    total_txt_length +=
                  (first_entry ? 0 : 1) + short_name_len + 1 + data_txt_length;

    /*************************************************************************/
    /* Realloc the buffer, allowing for the terminating null.                */
    /*************************************************************************/
    if ((total_txt_length + 1) > buf_size)
    {
      TRC_DETAIL((TB, "Reallocating %d-byte buffer at %p to %d bytes",
                                         buf_size, buf, total_txt_length + 1));
      new_buf = rm_realloc_mem(buf, total_txt_length + 1);
      if (new_buf == NULL)
      {
        TRC_ERROR((TB, "Failed to realloc %d-byte buffer at %p to %d bytes",
                                         buf_size, buf, total_txt_length + 1));
        rc = OSP_ERR_MEMORY;
        goto EXIT_LABEL;
      }
      buf = new_buf;
      buf_size = total_txt_length + 1;
    }

    /*************************************************************************/
    /* Copy in the separator, entry name, '=' and entry data.                */
    /*************************************************************************/
    dest_ptr = &(buf[copied_length]);
    if (!first_entry)
    {
      if (is_new_set)
      {
        *(dest_ptr++) = ',';
      }
      else
      {
        *(dest_ptr++) = '+';
      }
    }

    memcpy(dest_ptr, short_name, (unsigned int)short_name_len);
    dest_ptr += short_name_len;

    *(dest_ptr++) = '=';

    for (jj = 0; jj < ne_data_len; jj++)
    {
      if (ccm_must_escape(ne_data[jj], jj, ne_data_len))
      {
        /*********************************************************************/
        /* Generate the escape sequence.                                     */
        /*********************************************************************/
        *(dest_ptr++) = '\\';
        *(dest_ptr++) = ccm_hex[(ne_data[jj] >> 4) & 0x0f];
        *(dest_ptr++) = ccm_hex[ne_data[jj] & 0x0f];
      }
      else
      {
        *(dest_ptr++) = ne_data[jj];
      }
    }

    /*************************************************************************/
    /* Add a null terminator now, in case this is the last entry.            */
    /*************************************************************************/
    *dest_ptr = 0;

    /*************************************************************************/
    /* After this, we will no longer be processing the first entry.          */
    /*************************************************************************/
    first_entry = FALSE;
  }

  /***************************************************************************/
  /* Return the data buffer to the caller.                                   */
  /***************************************************************************/
  TRC_NORMAL((TB, "Converted X509_NAME to '%s'", buf));
  *name_out = buf;
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if ((rc != OSP_SUCCESS) && (buf != NULL))
  {
    rm_release_mem(buf);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_x509_name_to_rfc2253 */

/**PROC+**********************************************************************/
/* Name:      ccm_must_escape                                                */
/*                                                                           */
/* Purpose:   Determines if the given character needs replacing with an      */
/*            escape sequence for inclusion in an RFC2253-format value.      */
/*                                                                           */
/* Returns:   1           - escape sequence required                         */
/*            0           - escape sequence not required                     */
/*                                                                           */
/* Params:    IN     candidate      - the character to check                 */
/*            IN     pos            - the character's offset from the start  */
/*                                      of the string                        */
/*            IN     length         - the length of the string               */
/*                                                                           */
/* Operation: Check the character against a list of must-escape chars;       */
/*            also check for leading '#' or ' ' and trailing ' '. See        */
/*            RFC2253 for details.                                           */
/*                                                                           */
/**PROC-**********************************************************************/

int ccm_must_escape(char candidate,
                    unsigned int pos,
                    unsigned int length)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int escape = FALSE;

  TRC_ENTRY("ccm_must_escape");

  if ((candidate == ',')  ||
      (candidate == '+')  ||
      (candidate == '"')  ||
      (candidate == '\\') ||
      (candidate == '<')  ||
      (candidate == '>')  ||
      (candidate == ';'))
  {
    escape = TRUE;
  }
  else if ((pos == 0) && ((candidate == ' ') || (candidate == '#')))
  {
    escape = TRUE;
  }
  else if ((pos == length - 1) && (candidate == ' '))
  {
    escape = TRUE;
  }

  TRC_EXIT();

  return(escape);

} /* ccm_must_escape */

/**PROC+**********************************************************************/
/* Name:      ccm_get_priv_key                                               */
/*                                                                           */
/* Purpose:   Look up private key in identity list                           */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     subj_name      - the distinguished name to look for     */
/*            OUT    priv_key       - the private key (if found)             */
/*                                                                           */
/* Operation: Loop through the names in the identity list looking for a      */
/*            match, then return its corresponding private key.              */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_get_priv_key(char *subj_name,
                        EVP_PKEY **priv_key)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509_NAME *x509_name = NULL;
  int ii;

  TRC_ENTRY("ccm_get_priv_key");

  /***************************************************************************/
  /* Convert the subject name into an X509_NAME structure.                   */
  /***************************************************************************/
  rc = ccm_rfc2253_to_x509_name(subj_name, &x509_name);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to convert subject name to X509_NAME (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Spin through the DN index to find one that matches.                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Searching for a match..."));
  for (ii = 0; ii < ccm.identities.num; ii++)
  {
    if (X509_NAME_cmp(x509_name, ccm.identities.names[ii]) == 0)
    {
      break;
    }
  }

  if (ii >= ccm.identities.num)
  {
    TRC_ERROR((TB, "Failed to find '%s' in identity index", subj_name));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Found it -- return the private key structure to the caller.             */
  /***************************************************************************/
  *priv_key = ccm.identities.privkeys[ii];

  TRC_NORMAL((TB, "Found private key for identity '%s'", subj_name));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (x509_name != NULL)
  {
    X509_NAME_free(x509_name);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_get_priv_key */

/**PROC+**********************************************************************/
/* Name:      ccm_int_get_cert                                               */
/*                                                                           */
/* Purpose:   Look up a certificate using the subject's distinguished name.  */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     subj_name      - the distinguished name to look for     */
/*            OUT    cert           - the X.509 certificate (if found)       */
/*                                                                           */
/* Operation: Convert the name string into an X509_NAME and use a helper     */
/*            function to do the actual lookup.  Note that this will block   */
/*            on an LDAP lookup if the cert is not held locally.             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_int_get_cert(char *subj_name,
                        X509 **cert)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509_NAME *x509_name = NULL;

  TRC_ENTRY("ccm_int_get_cert");

  /***************************************************************************/
  /* Convert the subject name into an X509_NAME structure.                   */
  /***************************************************************************/
  rc = ccm_rfc2253_to_x509_name(subj_name, &x509_name);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to convert subject name to X509_NAME (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Call the helper routine.                                                */
  /***************************************************************************/
  rc = ccm_int_get_cert_x509_name(x509_name, cert);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Error getting certificate (%u) for '%s'", rc, subj_name));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Retrieved certificate for '%s' OK", subj_name));

EXIT_LABEL:

  if (x509_name != NULL)
  {
    X509_NAME_free(x509_name);
  }

  TRC_EXIT();

  return(rc);

} /* ccm_int_get_cert */

/**PROC+**********************************************************************/
/* Name:      ccm_int_get_cert_x509_name                                     */
/*                                                                           */
/* Purpose:   Look up a certificate using the subject's distinguished name   */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     x509_name      - the distinguished name to look for     */
/*            OUT    cert           - the X.509 certificate (if found)       */
/*                                                                           */
/* Operation: Call the store to do the lookup. If not present locally, the   */
/*            store will do an LDAP lookup to get it, which will cause this  */
/*            function to block.                                             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_int_get_cert_x509_name(X509_NAME *x509_name,
                                  X509 **cert)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509_STORE_CTX store_ctx;
  X509_OBJECT x509_obj;
  int error;

  TRC_ENTRY("ccm_int_get_cert_x509_name");

  /***************************************************************************/
  /* Initialize a temporary store context.                                   */
  /***************************************************************************/
  X509_STORE_CTX_init(&store_ctx, ccm.cert_store, NULL, NULL);

  /***************************************************************************/
  /* Do the lookup.  This will block on an LDAP directory search if the      */
  /* certificate is not held locally.                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Looking up cert from store (name is at %p)", x509_name));
  error = X509_STORE_get_by_subject(&store_ctx,
                                    X509_LU_X509,
                                    x509_name,
                                    &x509_obj);
  if (error <= 0)
  {
    TRC_ALERT((TB, "Couldn't find certificate (%d)", error));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Return the certificate to the caller if they wanted it.                 */
  /***************************************************************************/
  if (cert != NULL)
  {
    TRC_DETAIL((TB, "Returning cert %p to caller at %p",
                                                    x509_obj.data.x509, cert));
    *cert = x509_obj.data.x509;
  }
  else
  {
    /*************************************************************************/
    /* The caller was just checking to see if the cert was in the directory  */
    /* or to make sure it got pulled into local memory.  We can free it.     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Caller doesn't want the cert at %p", x509_obj.data.x509));
    X509_OBJECT_free_contents(&x509_obj);
  }

  rc = OSP_SUCCESS;

EXIT_LABEL:

  X509_STORE_CTX_cleanup(&store_ctx);

  TRC_EXIT();

  return(rc);

} /* ccm_int_get_cert_x509_name */

/**PROC+**********************************************************************/
/* Name:      ccm_temp_rsa_callback                                          */
/*                                                                           */
/* Purpose:   Pass back a temporary RSA key to OpenSSL. This is normally     */
/*            used when we have a >512-bit certificate but are using an      */
/*            export cipher that requires <=512-bit certs. It can also be    */
/*            called when we have a signing-only certificate.                */
/*                                                                           */
/* Returns:   An OpenSSL RSA structure that contains the temporary key.      */
/*                                                                           */
/* Params:    IN     ssl            - the SSL structure that needs the       */
/*                                      temporary key                        */
/*            IN     is_export      - non-zero if this is for export reasons */
/*            IN     keylength      - the keylength required                 */
/*                                                                           */
/* Operation: Retrieve one of the temporary keys we generated on startup.    */
/*                                                                           */
/**PROC-**********************************************************************/

RSA *ccm_temp_rsa_callback(SSL *ssl, int is_export, int keylength)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  RSA *rsa = NULL;

  TRC_ENTRY("ccm_temp_rsa_callback");

  /***************************************************************************/
  /* Return the appropriate pre-generated key.                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Temp RSA key for ssl %p (%d bits); export=%d",
                                                   ssl, keylength, is_export));
  switch (keylength)
  {
    case 512:
      rsa = ccm.tempkeys.rsa512;
      break;

    case 1024:
      rsa = ccm.tempkeys.rsa1024;
      break;

    default:
      TRC_NORMAL((TB, "OpenSSL wanted %-bit temp RSA key; returning 1024-bit",
                                                                   keylength));
      rsa = ccm.tempkeys.rsa1024;
      break;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rsa);

} /* ccm_temp_rsa_callback */

/**PROC+**********************************************************************/
/* Name:      ccm_temp_dh_callback                                           */
/*                                                                           */
/* Purpose:   Pass back some DH parameters to OpenSSL. This is for when SSL  */
/*            is using a cipher that requires Ephemeral Diffie-Hellman       */
/*            key exchange.                                                  */
/*                                                                           */
/* Returns:   An OpenSSL DH structure that contains the parameters.          */
/*                                                                           */
/* Params:    IN     ssl            - the SSL structure that needs the       */
/*                                      parameters.                          */
/*            IN     is_export      - non-zero if this is for export reasons */
/*                                      (was inverted in OpenSSL-0.9.5a but  */
/*                                       was fixed in later versions)        */
/*            IN     keylength      - the keylength required                 */
/*                                                                           */
/* Operation: Retrieve one of the temporary keys we generated on startup.    */
/*                                                                           */
/**PROC-**********************************************************************/

DH *ccm_temp_dh_callback(SSL *ssl, int is_export, int keylength)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  DH *dh = NULL;

  TRC_ENTRY("ccm_temp_dh_callback");

  /***************************************************************************/
  /* Return the appropriate pre-generated parameters.                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "DH parameters for ssl %p (%d bits); export=%d",
                                                   ssl, keylength, is_export));
  switch (keylength)
  {
    case 512:
      dh = ccm.tempkeys.dh512;
      break;

    case 1024:
      dh = ccm.tempkeys.dh1024;
      break;

    default:
      TRC_NORMAL((TB, "OpenSSL wanted %-bit DH params; returning 1024-bit",
                                                                   keylength));
      dh = ccm.tempkeys.dh1024;
      break;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(dh);

} /* ccm_temp_dh_callback */

/**PROC+**********************************************************************/
/* Name:      ccm_get_dh512                                                  */
/*                                                                           */
/* Purpose:   Return pre-generated temporary 512-bit Diffie-Hellman params.  */
/*                                                                           */
/* Returns:   OpenSSL DH structure containing the parameters, or NULL on     */
/*            failure.                                                       */
/*                                                                           */
/* Params:    None.                                                          */
/*                                                                           */
/* Operation: Based on the auto-generated C code from the following OpenSSL  */
/*            command:                                                       */
/*                                                                           */
/*              openssl dhparam -check 512 -C                                */
/*                                                                           */
/**PROC-**********************************************************************/

DH *ccm_get_dh512(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int ok = FALSE;
  DH *dh;

  TRC_ENTRY("ccm_get_dh512");

  dh = DH_new();
  if (dh == NULL)
  {
    TRC_ERROR((TB, "Failed to create DH structure"));
    goto EXIT_LABEL;
  }

  dh->p = BN_bin2bn(ccm_dh512_p, sizeof(ccm_dh512_p), NULL);
  dh->g = BN_bin2bn(ccm_dh512_g, sizeof(ccm_dh512_g), NULL);

  if ((dh->p == NULL) || (dh->g == NULL))
  {
    TRC_ERROR((TB, "Failed to convert 512-bit temporary DH parameters"));
    goto EXIT_LABEL;
  }

  ok = TRUE;

EXIT_LABEL:

  if (!ok && (dh != NULL))
  {
    DH_free(dh);
    dh = NULL;
  }

  TRC_EXIT();

  return(dh);

} /* ccm_get_dh512 */

/**PROC+**********************************************************************/
/* Name:      ccm_get_dh1024                                                 */
/*                                                                           */
/* Purpose:   Return pre-generated temporary 1024-bit Diffie-Hellman params. */
/*                                                                           */
/* Returns:   OpenSSL DH structure containing the parameters, or NULL on     */
/*            failure.                                                       */
/*                                                                           */
/* Params:    None.                                                          */
/*                                                                           */
/* Operation: Based on the auto-generated C code from the following OpenSSL  */
/*            command:                                                       */
/*                                                                           */
/*              openssl dhparam -check 1024 -C                               */
/*                                                                           */
/**PROC-**********************************************************************/

DH *ccm_get_dh1024(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int ok = FALSE;
  DH *dh;

  TRC_ENTRY("ccm_get_dh1024");

  dh = DH_new();
  if (dh == NULL)
  {
    TRC_ERROR((TB, "Failed to create DH structure"));
    goto EXIT_LABEL;
  }

  dh->p = BN_bin2bn(ccm_dh1024_p, sizeof(ccm_dh1024_p), NULL);
  dh->g = BN_bin2bn(ccm_dh1024_g, sizeof(ccm_dh1024_g), NULL);

  if ((dh->p == NULL) || (dh->g == NULL))
  {
    TRC_ERROR((TB, "Failed to convert 1024-bit temporary DH parameters"));
    goto EXIT_LABEL;
  }

  ok = TRUE;

EXIT_LABEL:

  if (!ok && (dh != NULL))
  {
    DH_free(dh);
    dh = NULL;
  }

  TRC_EXIT();

  return(dh);

} /* ccm_get_dh1024 */

