
/**MOD+***********************************************************************/
/* Module:    ccmvfy.c                                                       */
/*                                                                           */
/* Purpose:   CCM certificate verification functions                         */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.12               $ $Modtime::   Jan 11 2001 15:56:34   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <ccmincl.h>

/**PROC+**********************************************************************/
/* Name:      ccm_verify_callback                                            */
/*                                                                           */
/* Purpose:   Handles all certificate verification callbacks from the CCM    */
/*            certificate store.                                             */
/*                                                                           */
/* Returns:   1           - certificate is OK                                */
/*            0           - certificate is not OK                            */
/*                                                                           */
/* Params:    IN     ok             - 1: cert is OK as far as OpenSSL knows  */
/*                                    0: cert is not OK according to OpenSSL */
/*            IN     store_ctx      - the OpenSSL store context used for     */
/*                                      verifying this chain of certificates */
/*                                                                           */
/* Operation: Work out whether this verification is for SSL or not. If it's  */
/*            an SSL cert, we call the callback we got on ccm_init().        */
/*            When a certificate chain is being verified, this callback is   */
/*            called for each certificate in the chain, starting with the    */
/*            root CA's certificate.                                         */
/*                                                                           */
/**PROC-**********************************************************************/

int ccm_verify_callback(int ok,
                        X509_STORE_CTX *store_ctx)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC int_rc = OSP_SUCCESS;
  void *user_data = NULL;
  X509 *cert = NULL;
  int is_ssl = FALSE;
  SSL *ssl = NULL;
  char *subject_name_txt = NULL;
  char *issuer_name_txt = NULL;
  unsigned int *flags_pointer;
  unsigned int flags;
  unsigned char *mem = NULL;
  char *asn1_cert = NULL;
  unsigned int asn1_cert_len;
  unsigned int chain_len;
  unsigned int chain_pos;
  int store_error;

  TRC_ENTRY("ccm_verify_callback");

  /***************************************************************************/
  /* Trace parameters.                                                       */
  /***************************************************************************/
  store_error = X509_STORE_CTX_get_error(store_ctx);
  TRC_NORMAL((TB, "Verifying cert; ok=%d, context ptr %p, error=%d ('%s')",
      ok, store_ctx, store_error, X509_verify_cert_error_string(store_error)));

  /***************************************************************************/
  /* Get the certificate currently being verified.                           */
  /***************************************************************************/
  cert = X509_STORE_CTX_get_current_cert(store_ctx);

  /***************************************************************************/
  /* Get the chain length and position.                                      */
  /***************************************************************************/
  chain_len = (unsigned int)sk_X509_num(X509_STORE_CTX_get_chain(store_ctx));
  chain_pos = (unsigned int)X509_STORE_CTX_get_error_depth(store_ctx);
  TRC_NORMAL((TB, "Chain pos=%d, len=%d", chain_pos, chain_len));

  /***************************************************************************/
  /* Get the subject and issuer names as plain text for tracing purposes.    */
  /***************************************************************************/
  int_rc = ccm_x509_name_to_rfc2253(X509_get_subject_name(cert),
                                                            &subject_name_txt);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to convert cert subject name to text"));
    ok = 0;
    goto EXIT_LABEL;
  }

  int_rc = ccm_x509_name_to_rfc2253(X509_get_issuer_name(cert),
                                                             &issuer_name_txt);
  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to convert cert issuer name to text"));
    ok = 0;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Subject is '%s'", subject_name_txt));
  TRC_NORMAL((TB, "Issuer is '%s'", issuer_name_txt));

  /***************************************************************************/
  /* Extract the ex data from the store context.  If the verification is for */
  /* SSL, this value will be the SSL structure.  If not, it will be NULL and */
  /* index 1 will contain some flags used to control the verification.       */
  /***************************************************************************/
  ssl = X509_STORE_CTX_get_ex_data(store_ctx, 0);
  TRC_DETAIL((TB, "Got SSL ptr %p", ssl));
  if (ssl != NULL)
  {
    /*************************************************************************/
    /* This is an SSL verification.                                          */
    /*************************************************************************/
    TRC_NORMAL((TB, "SSL cert verification"));
    is_ssl = TRUE;
  }
  else
  {
    /*************************************************************************/
    /* This is a PKCS#7 cert verification.                                   */
    /*************************************************************************/
    TRC_NORMAL((TB, "PKCS#7 cert verification"));
    is_ssl = FALSE;

    /*************************************************************************/
    /* Extract the flags.                                                    */
    /*************************************************************************/
    flags_pointer = (unsigned int *)X509_STORE_CTX_get_ex_data(store_ctx, 1);
    if (flags_pointer == NULL)
    {
      TRC_ERROR((TB, "Got null flags pointer from store context"));
      ok = 0;
      goto EXIT_LABEL;
    }

    flags = *flags_pointer;
    TRC_DETAIL((TB, "Recovered flags: 0x%x", flags));

    /*************************************************************************/
    /* If we have an error due to a depth zero self-signed certificate and   */
    /* the flags allow this, override it.                                    */
    /*************************************************************************/
    if (!ok &&
        (store_error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) &&
        (flags & CCM_FLAG_SELF_SIGNED))
    {
      TRC_NORMAL((TB, "Allowing depth zero self-signed certificate"));
      ok = 1;
    }
  }

  /***************************************************************************/
  /* If we're still ok, do the CRL checks.                                   */
  /***************************************************************************/
  if (ok)
  {
    TRC_DETAIL((TB, "Doing CRL checks"));
    ok = ccm_check_crl(store_ctx, subject_name_txt, issuer_name_txt);
    if (!ok)
    {
      TRC_ERROR((TB, "Failed CRL checks for subject '%s'", subject_name_txt));
    }
  }

  /***************************************************************************/
  /* If we're still OK and this verification is for SSL, call the callback.  */
  /***************************************************************************/
  if (ok && is_ssl)
  {
    TRC_NORMAL((TB, "Cert for SSL is OK; preparing to call callback"));

    /*************************************************************************/
    /* Convert the certificate to ASN.1 DER-encoded form.  If this fails we  */
    /* have to reject the cert as we can't pass it to the callback.          */
    /*************************************************************************/
    asn1_cert_len = i2d_X509(cert, NULL);
    mem = (unsigned char *)rm_get_mem(asn1_cert_len);
    if (mem == NULL)
    {
      OSP_OPENSSL_ERROR((TB, "Couldn't convert cert for '%s' -- rejecting",
                                                            subject_name_txt));
      ok = 0;
      goto EXIT_LABEL;
    }

    asn1_cert = (char *)mem;
    asn1_cert_len = i2d_X509(cert, &mem);

    /*************************************************************************/
    /* Call the callback.                                                    */
    /*************************************************************************/
    TRC_NORMAL((TB, "Calling verification callback; "
                  "asn1_cert=%p, asn1_cert_len=%u, chain_len=%u, chain_pos=%u",
                  asn1_cert, asn1_cert_len, chain_len, chain_pos));
    ok = ccm.ssl.verify_cb(ssl,
                           asn1_cert,
                           asn1_cert_len,
                           chain_len,
                           chain_pos);
    if (!ok)
    {
      TRC_ALERT((TB, "Callback rejected certificate for '%s'",
                                                            subject_name_txt));
    }
  }

  TRC_NORMAL((TB, "Completed ccm_verify_callback(); ok=%d", ok));

#ifdef CCM_DEBUG_X509
  if (!ok)
  {
    printf("This certificate failed verification:\n");
    X509_print_fp(stdout, cert);
  }
#endif

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (asn1_cert != NULL)
  {
    rm_release_mem(asn1_cert);
  }

  if (subject_name_txt != NULL)
  {
    rm_release_mem(subject_name_txt);
  }

  if (issuer_name_txt != NULL)
  {
    rm_release_mem(issuer_name_txt);
  }

  TRC_EXIT();

  return(ok);

} /* ccm_verify_callback */

/**PROC+**********************************************************************/
/* Name:      ccm_check_crl                                                  */
/*                                                                           */
/* Purpose:   Do CRL-related checks for the supplied certificate's subject   */
/*                                                                           */
/* Returns:   1           - CRL checks passed                                */
/*            0           - CRL checks failed                                */
/*                                                                           */
/* Params:    IN     store_ctx      - certificate store context from OpenSSL */
/*            IN     subject_name_txt                                        */
/*                                  - current cert's subject name as text    */
/*            IN     issuer_name_txt                                         */
/*                                  - current cert's issuer name as text     */
/*                                                                           */
/* Operation: First, check for a CRL issued by the current certificate's     */
/*            _subject_, and verify it if there is one. We do this now       */
/*            because we have the certificate to verify the CRL with - this  */
/*            CRL is actually used the next time this function is called.    */
/*                                                                           */
/*            Then retrieve a CRL corresponding to the current certificate's */
/*            _issuer_ and check that the current certificate isn't on it.   */
/*            The CRL will already have been verified in the previous call   */
/*            to this function. It will also be in the local store so this   */
/*            lookup won't impose a significant overhead.                    */
/*                                                                           */
/*            This strategy is borrowed from mod_ssl.                        */
/*                                                                           */
/**PROC-**********************************************************************/

int ccm_check_crl(X509_STORE_CTX *store_ctx,
                  char *subject_name_txt,
                  char *issuer_name_txt)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int ok = 0;
  OSP_RC int_rc;
  X509 *current_cert = NULL;
  X509_NAME *subject_name = NULL;
  X509_NAME *issuer_name = NULL;
  X509_OBJECT x509_obj;
  X509_CRL *crl = NULL;
  X509_REVOKED *revoked = NULL;
  int revoked_num;
  int ii;
  int error;

  TRC_ENTRY("ccm_check_crl");

  /***************************************************************************/
  /* Initialize the X509_OBJECT that we use later.                           */
  /***************************************************************************/
  x509_obj.type = X509_LU_FAIL;
  x509_obj.data.ptr = NULL;

  /***************************************************************************/
  /* Get the information we need.                                            */
  /***************************************************************************/
  current_cert = X509_STORE_CTX_get_current_cert(store_ctx);
  subject_name = X509_get_subject_name(current_cert);
  issuer_name = X509_get_issuer_name(current_cert);

  /***************************************************************************/
  /* First look for a CRL issued by the subject of the current certificate.  */
  /* We only need to do this if the current certificate is not an end entity */
  /* certificate -- end entities will never have CRLs so we don't want to    */
  /* waste time and resource doing a search.                                 */
  /***************************************************************************/
  if (X509_STORE_CTX_get_error_depth(store_ctx) != 0)
  {
    TRC_DETAIL((TB, "Looking up CRL for CA certificate's subject name"));
    int_rc = ccm_get_valid_crl(store_ctx, subject_name,
                                                       subject_name_txt, &crl);
    if (int_rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We couldn't find a valid CRL.  The above function will have set the */
      /* error in the store context, so no need to do that here.             */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to find a valid CRL for '%s' -- "
                                    "failing verification", subject_name_txt));
      goto EXIT_LABEL;
    }

    TRC_DETAIL((TB, "Got CRL for cert subject; checking signature"));
    error = X509_CRL_verify(crl, X509_get_pubkey(current_cert));
    if (error <= 0)
    {
      TRC_ERROR((TB, "Invalid signature on CRL for '%s'", subject_name_txt));
      X509_STORE_CTX_set_error(store_ctx, X509_V_ERR_CRL_SIGNATURE_FAILURE);
      goto EXIT_LABEL;
    }

    TRC_NORMAL((TB, "CRL for '%s' appears to be valid", subject_name_txt));
  }
  else
  {
    TRC_DETAIL((TB, "No need to get CRL for end entity '%s'",
                                                            subject_name_txt));
  }

  /***************************************************************************/
  /* Now look for a CRL issued by the issuer of the current certificate.     */
  /* This will be in the local store because the code above will have got it */
  /* previously.                                                             */
  /***************************************************************************/
  TRC_DETAIL((TB, "Looking up CRL for certificate's issuer name"));
  error = X509_STORE_get_by_subject(store_ctx,
                                    X509_LU_CRL,
                                    issuer_name,
                                    &x509_obj);
  crl = x509_obj.data.crl;
  if ((error > 0) && (crl != NULL))
  {
    /*************************************************************************/
    /* Check that the current certificate isn't in this CRL.                 */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got CRL for cert issuer; searching for current cert"));

    revoked_num = sk_X509_REVOKED_num(X509_CRL_get_REVOKED(crl));
    TRC_DETAIL((TB, "Checking %d entries in CRL", revoked_num));
    for (ii = 0; ii < revoked_num; ii++)
    {
      revoked = sk_X509_REVOKED_value(X509_CRL_get_REVOKED(crl), ii);
      if (ASN1_INTEGER_cmp(revoked->serialNumber,
                                     X509_get_serialNumber(current_cert)) == 0)
      {
        TRC_ERROR((TB, "Certificate revoked for '%s' in CRL from '%s'",
                                           subject_name_txt, issuer_name_txt));
        X509_STORE_CTX_set_error(store_ctx, X509_V_ERR_CERT_REVOKED);
        goto EXIT_LABEL;
      }
    }

    TRC_NORMAL((TB, "Certificate not revoked for '%s'", subject_name_txt));
  }
  else
  {
    TRC_DETAIL((TB, "No CRL found for cert issuer"));
  }

  /***************************************************************************/
  /* All OK.                                                                 */
  /***************************************************************************/
  ok = 1;

EXIT_LABEL:

  TRC_NORMAL((TB, "CRL checks done: ok=%d", ok));

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (crl != NULL)
  {
    X509_CRL_free(crl);
  }

  TRC_EXIT();

  return(ok);

} /* ccm_check_crl */

/**PROC+**********************************************************************/
/* Name:      ccm_get_valid_crl                                              */
/*                                                                           */
/* Purpose:   Get a CRL from the store; force LDAP lookup if it's expired.   */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     store_ctx      - certificate store context from OpenSSL */
/*                                                                           */
/* Operation: Look up a CRL for the name given. If it has expired, delete it */
/*            from the local store and do another lookup. The second lookup  */
/*            will force an LDAP search, which will retrieve the latest CRL  */
/*            from the directory.                                            */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC ccm_get_valid_crl(X509_STORE_CTX *store_ctx,
                         X509_NAME *crl_issuer_name,
                         char *crl_issuer_txt,
                         X509_CRL **crl)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509_OBJECT x509_obj;
  X509_OBJECT *store_obj = NULL;
  int error;

  TRC_ENTRY("ccm_get_valid_crl");

  /***************************************************************************/
  /* Initialize the X509_OBJECT.                                             */
  /***************************************************************************/
  x509_obj.type = X509_LU_FAIL;
  x509_obj.data.ptr = NULL;

  /***************************************************************************/
  /* First, do a normal store lookup to get a CRL.                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting CRL from store"));
  error = X509_STORE_get_by_subject(store_ctx,
                                    X509_LU_CRL,
                                    crl_issuer_name,
                                    &x509_obj);
  *crl = x509_obj.data.crl;
  if ((error <= 0) || (*crl == NULL))
  {
    TRC_ERROR((TB, "Failed to find a CRL for '%s'", crl_issuer_txt));
    X509_STORE_CTX_set_error(store_ctx, X509_V_ERR_UNABLE_TO_GET_CRL);
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Check this CRL's validity.                                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Got CRL; checking validity"));
  error = X509_cmp_current_time(X509_CRL_get_nextUpdate(*crl));
  if (error <= 0)
  {
    /*************************************************************************/
    /* The CRL is either expired or has an invalid nextUpdate field.         */
    /*************************************************************************/
    if (error == 0)
    {
      TRC_ERROR((TB, "CRL for '%s' has an invalid nextUpdate field; "
                               "attempting to get a new CRL", crl_issuer_txt));
    }
    else
    {
      TRC_ALERT((TB, "CRL for '%s' has expired; attempting to get a new one",
                                                              crl_issuer_txt));
    }

    /*************************************************************************/
    /* We need to force an LDAP lookup to get the latest CRL.  We'll do this */
    /* by deleting the CRL from the local store and then doing another get.  */
    /*************************************************************************/
    TRC_DETAIL((TB, "Deleting CRL from local store (%p)", x509_obj.data.crl));
    store_obj = lh_delete(store_ctx->ctx->certs, &x509_obj);
    if (store_obj == NULL)
    {
      TRC_ERROR((TB, "Failed to delete CRL from local store"));
      X509_STORE_CTX_set_error(store_ctx, X509_V_ERR_UNABLE_TO_GET_CRL);
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* We must free the store object twice, because it has a reference count */
    /* of 2: one for the store and another from the get that we just did.    */
    /*************************************************************************/
    X509_OBJECT_free_contents(store_obj);
    X509_OBJECT_free_contents(store_obj);
    TRC_DETAIL((TB, "Deleted CRL object %p", store_obj->data.crl));

    /*************************************************************************/
    /* Since the CRL is no longer in the local store, this get will be       */
    /* forced to do an LDAP lookup and retrieve the latest available CRL.    */
    /*************************************************************************/
    TRC_NORMAL((TB, "Getting new CRL for '%s'", crl_issuer_txt));
    error = X509_STORE_get_by_subject(store_ctx,
                                      X509_LU_CRL,
                                      crl_issuer_name,
                                      &x509_obj);
    *crl = x509_obj.data.crl;
    if ((error <= 0) || (*crl == NULL))
    {
      TRC_ERROR((TB, "Failed to find a new CRL for '%s'", crl_issuer_txt));
      X509_STORE_CTX_set_error(store_ctx, X509_V_ERR_UNABLE_TO_GET_CRL);
      goto EXIT_LABEL;
    }

    error = X509_cmp_current_time(X509_CRL_get_nextUpdate(*crl));
    if (error == 0)
    {
      TRC_ERROR((TB, "New CRL for '%s' has invalid nextUpdate field",
                                                              crl_issuer_txt));
      X509_STORE_CTX_set_error(store_ctx,
                                    X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD);
      goto EXIT_LABEL;
    }
    else if (error < 0)
    {
      TRC_ERROR((TB, "New CRL for '%s' has expired", crl_issuer_txt));
      X509_STORE_CTX_set_error(store_ctx, X509_V_ERR_CRL_HAS_EXPIRED);
      goto EXIT_LABEL;
    }

    TRC_NORMAL((TB, "New CRL is valid"));
  }

  /***************************************************************************/
  /* We've found a valid CRL. It's already been returned to the caller.      */
  /***************************************************************************/
  TRC_DETAIL((TB, "Got a valid CRL"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    X509_OBJECT_free_contents(&x509_obj);
    *crl = NULL;
  }

  TRC_EXIT();

  return(rc);

} /* ccm_get_valid_crl */

