
/**MOD+***********************************************************************/
/* Module:    commain.c                                                      */
/*                                                                           */
/* Purpose:   OSP Common functions                                           */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.14               $ $Modtime::   Jan 11 2001 15:56:36   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <comincl.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

/**PROC+**********************************************************************/
/* Name:      osp_common_init                                                */
/*                                                                           */
/* Purpose:   Initialize the OSP common functions                            */
/*                                                                           */
/* Returns:   OSP_SUCCESS          - common code initialized successfully    */
/*            OSP_ERR_ALREADY_INIT - common code  already initialized        */
/*            OSP_ERROR            - another error occurred                  */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_common_init(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  int error;

  TRC_ENTRY("osp_common_init");

  /***************************************************************************/
  /* Check for re-initialization                                             */
  /***************************************************************************/
  if (com_initialized)
  {
    TRC_ERROR((TB, "COM re-initialized"));
    rc = OSP_ERR_ALREADY_INIT;
    goto EXIT_LABEL;
  }

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* Initialize the count of mutexes created/destroyed                       */
  /***************************************************************************/
  com_mutex_count = 0;
#endif

  TRC_NORMAL((TB, "Common initialized successfully"));
  com_initialized = TRUE;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* osp_common_init */

/**PROC+**********************************************************************/
/* Name:      osp_common_term                                                */
/*                                                                           */
/* Purpose:   Terminate the OSP common functions                             */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - Common code terminated successfully         */
/*            OSP_ERR_NOT_INIT - The common code was not intialized          */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: Destroy the mutex attribute.                                   */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_common_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("osp_common_term");

  /***************************************************************************/
  /* Check we were initialized                                               */
  /***************************************************************************/
  if (com_initialized)
  {
#ifdef OSP_DEBUG
    /*************************************************************************/
    /* Check that all the mutexes created through us have been destroyed     */
    /*************************************************************************/
    if (com_mutex_count == 0)
    {
      TRC_DETAIL((TB, "All mutexes destroyed"));
    }
    else
    {
      TRC_ERROR((TB, "%d mutexes not destroyed", com_mutex_count));
    }
#endif
    /*************************************************************************/
    /* tidy up                                                               */
    /*************************************************************************/
    TRC_NORMAL((TB, "Common terminated successfully"));
    com_initialized = FALSE;
  }
  else
  {
    TRC_ERROR((TB, "COM not intialized"));
    rc = OSP_ERR_NOT_INIT;
  }

  TRC_EXIT();

  return(rc);

} /* osp_common_term */

/**PROC+**********************************************************************/
/* Name:      osp_create_mutex                                               */
/*                                                                           */
/* Purpose:   Create a default mutex                                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS - the mutex was created                            */
/*            OSP_ERROR   - there was a problem creating the mutex           */
/*                                                                           */
/* Params:    OUT    mutex        - pointer to created mutex                 */
/*                                                                           */
/* Operation: Use pthread_mutex_init() with a null attribute pointer.        */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_create_mutex(pthread_mutex_t *mutex)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  int result;

  TRC_ENTRY("osp_create_mutex");

  result = pthread_mutex_init(mutex, NULL);
  if (result == 0)
  {
    TRC_DETAIL((TB, "Created mutex OK"));
#ifdef OSP_DEBUG
    /*************************************************************************/
    /* Increment the mutex count                                             */
    /*************************************************************************/
    com_mutex_count++;
    TRC_DETAIL((TB, "Mutex count now %d", com_mutex_count));
#endif
  }
  else
  {
    TRC_OS_ERROR((TB, "Failed to initialize default mutex (%d)", result));
    rc = OSP_ERROR;
  }

  TRC_EXIT();

  return(rc);

} /* osp_create_mutex */

/**PROC+**********************************************************************/
/* Name:      osp_destroy_mutex                                              */
/*                                                                           */
/* Purpose:   destroy a mutex                                                */
/*                                                                           */
/* Returns:   OSP_SUCCESS - the mutex was destroyed                          */
/*            OSP_ERROR   - there was a problem destroying the mutex         */
/*                                                                           */
/* Params:    IN     mutex        - mutex to destroy                         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_destroy_mutex(pthread_mutex_t *mutex)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  int result;

  TRC_ENTRY("osp_destroy_mutex");

  result = pthread_mutex_destroy(mutex);
  if (result == 0)
  {
    TRC_DETAIL((TB, "Destroyed mutex OK"));
#ifdef OSP_DEBUG
    /*************************************************************************/
    /* Decrement the mutex count                                             */
    /*************************************************************************/
    com_mutex_count--;
    TRC_DETAIL((TB, "Mutex count now %d", com_mutex_count));
#endif
  }
  else
  {
    TRC_OS_ERROR((TB, "Failed to destroy mutex (%d)", result));
    rc = OSP_ERROR;
  }

  TRC_EXIT();

  return(rc);

} /* osp_destroy_mutex */

/**PROC+**********************************************************************/
/* Name:      osp_common_get_config_filename                                 */
/*                                                                           */
/* Purpose:   Gets the name of the config file                               */
/*                                                                           */
/* Returns:   Name of config file or NULL                                    */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: Calls getenv()                                                 */
/*                                                                           */
/**PROC-**********************************************************************/

char *osp_common_get_config_filename(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char *file_name = NULL;

  TRC_ENTRY("osp_common_get_config_filename");

  file_name = getenv(OSP_CONFIG_FILE_ENV);
  if (file_name == NULL)
  {
    /*************************************************************************/
    /* No config file configured in the environment.                         */
    /*************************************************************************/
    TRC_ERROR((TB, "Can't open config file: %s not set", OSP_CONFIG_FILE_ENV));
  }
  else
  {
    TRC_NORMAL((TB, "Config file name is %s", file_name));
  }

  TRC_EXIT();

  return(file_name);

} /* osp_common_get_config_filename */

/**PROC+**********************************************************************/
/* Name:      osp_base64_encode                                              */
/*                                                                           */
/* Purpose:   Encode binary data as base64                                   */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     buffer         - pointer to the raw binary data         */
/*            IN     buf_len        - length of raw binary data              */
/*            OUT    b64_buf        - pointer to the base64 (caller must     */
/*                                      free with rm_release_mem())          */
/*            OUT    b64_len        - number of chars in the base64          */
/*                                                                           */
/* Operation: Use OpenSSL memory and base64 BIOs to do the conversion.       */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_base64_encode(char *buffer,
                         unsigned int buf_len,
                         char **b64_buf,
                         unsigned int *b64_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  BIO *mbio = NULL;
  BIO *b64bio = NULL;
  BIO *encoding_bio = NULL;
  BIO *outbio = NULL;
  char iobuf[1024];
  int len;

  TRC_ENTRY("osp_base64_encode");

  /***************************************************************************/
  /* Set up an OpenSSL memory BIO to point to the buffer containing the      */
  /* binary data.                                                            */
  /***************************************************************************/
  mbio = BIO_new_mem_buf((void *)buffer, (unsigned int)buf_len);
  if (mbio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create mbio"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up an OpenSSL filter BIO to do the base64 encoding.                 */
  /***************************************************************************/
  b64bio = BIO_new(BIO_f_base64());
  if (b64bio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create b64bio"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up an OpenSSL memory BIO to receive the output of the encoding.     */
  /***************************************************************************/
  outbio = BIO_new(BIO_s_mem());
  if (outbio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create outbio"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Then push the base64 BIO on top of the input BIO.  The returned bio     */
  /* handle is actually just the b64 BIO handle.                             */
  /***************************************************************************/
  encoding_bio = BIO_push(b64bio, outbio);

  /***************************************************************************/
  /* Loop to transfer the data from input to output.                         */
  /***************************************************************************/
  while ((len = BIO_read(mbio, iobuf, sizeof(iobuf))) > 0)
  {
    BIO_write(encoding_bio, iobuf, len);
  }
  BIO_flush(encoding_bio);

  /***************************************************************************/
  /* Get the output BIO's underlying memory pointer and length, and mark     */
  /* the BIO as read-only so that it is not freed when we free the BIO.      */
  /***************************************************************************/
  *b64_len = BIO_get_mem_data(outbio, b64_buf);
  BIO_set_flags(outbio, BIO_FLAGS_MEM_RDONLY);

  TRC_NORMAL((TB, "Converted %u bytes of binary to %u bytes of base64 at %p",
                                                 buf_len, *b64_len, *b64_buf));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (encoding_bio != NULL)
  {
    BIO_free_all(encoding_bio);
  }
  else
  {
    /*************************************************************************/
    /* We didn't manage to create the BIO stack, so free these separately.   */
    /*************************************************************************/
    if (outbio != NULL)
    {
      BIO_free(outbio);
    }
    if (b64bio != NULL)
    {
      BIO_free(b64bio);
    }
  }

  if (mbio != NULL)
  {
    BIO_free(mbio);
  }

  TRC_EXIT();

  return(rc);

} /* osp_base64_encode */

/**PROC+**********************************************************************/
/* Name:      osp_base64_decode                                              */
/*                                                                           */
/* Purpose:   Convert base64 to binary                                       */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     b64_buf        - pointer to the base64                  */
/*            IN     b64_len        - number of chars in the base64          */
/*            OUT    buffer         - pointer to the raw binary data (caller */
/*                                      must free with rm_release_mem())     */
/*            OUT    buf_len        - length of raw binary data              */
/*                                                                           */
/* Operation: Use OpenSSL memory and base64 BIOs to do the conversion.       */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_base64_decode(char *b64_buf,
                          unsigned int b64_len,
                          char **buffer,
                          unsigned int *buf_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  BIO *mbio = NULL;
  BIO *b64bio = NULL;
  BIO *inbio = NULL;
  BIO *outbio = NULL;
  char iobuf[1024];
  int len;

  TRC_ENTRY("osp_base64_decode");

  /***************************************************************************/
  /* Set up an OpenSSL memory BIO to point to the buffer containing our      */
  /* base64.  It's auto-flagged as read-only.                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Converting %u bytes of base64 at %p to binary",
                                                            b64_len, b64_buf));
  mbio = BIO_new_mem_buf(b64_buf, (int)b64_len);
  if (mbio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create mbio"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up an OpenSSL filter BIO to do the base64 decoding.  SCEP requests  */
  /* contain linefeeds in the base64, so we don't use BIO_FLAGS_BASE64_NO_NL */
  /***************************************************************************/
  b64bio = BIO_new(BIO_f_base64());
  if (b64bio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create b64bio"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up an OpenSSL memory BIO to receive the output of the decoding.     */
  /***************************************************************************/
  outbio = BIO_new(BIO_s_mem());
  if (outbio == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create outbio"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Push the base64 BIO on top of the input BIO.  The returned bio handle   */
  /* is actually just the base64 BIO handle.                                 */
  /***************************************************************************/
  inbio = BIO_push(b64bio, mbio);

  /***************************************************************************/
  /* Loop to transfer the data from input to output.                         */
  /***************************************************************************/
  while ((len = BIO_read(inbio, iobuf, sizeof(iobuf))) > 0)
  {
    BIO_write(outbio, iobuf, len);
  }

  /***************************************************************************/
  /* Get the output BIO's underlying memory pointer and length, and mark     */
  /* the BIO as read-only so that it is not freed when we free the BIO.      */
  /***************************************************************************/
  *buf_len = BIO_get_mem_data(outbio, buffer);
  BIO_set_flags(outbio, BIO_FLAGS_MEM_RDONLY);

  TRC_NORMAL((TB, "Converted %u bytes of base64 to %u bytes of binary at %p",
                                                  b64_len, *buf_len, *buffer));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (inbio != NULL)
  {
    BIO_free_all(inbio);
  }
  else
  {
    /*************************************************************************/
    /* We didn't manage to create the BIO stack, so free these separately.   */
    /*************************************************************************/
    if (mbio != NULL)
    {
      BIO_free(mbio);
    }
    if (b64bio != NULL)
    {
      BIO_free(b64bio);
    }
  }

  if (outbio != NULL)
  {
    BIO_free(outbio);
  }

  TRC_EXIT();

  return(rc);

} /* osp_base64_decode */

/**PROC+**********************************************************************/
/* Name:      osp_create_recursive_mutex                                     */
/*                                                                           */
/* Purpose:   Create a mutex that may won't block if the owning thread       */
/*            tries to get it again                                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS - the mutex was created                            */
/*            OSP_ERROR   - there was a problem creating the mutex           */
/*                                                                           */
/* Params:    IN     r_mutex      - structure to hold created mutex          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_create_recursive_mutex(OSP_RECURSIVE_MUTEX *r_mutex)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("osp_create_recursive_mutex");

  /***************************************************************************/
  /* Initialize the lock structure                                           */
  /***************************************************************************/
  r_mutex->current_locker = NULL;
  r_mutex->lock_count = 0;

  /***************************************************************************/
  /* Create the underlying mutex                                             */
  /***************************************************************************/
  rc = osp_create_mutex(&r_mutex->mutex);

  TRC_EXIT();

  return(rc);

} /* osp_create_recursive_mutex */

/**PROC+**********************************************************************/
/* Name:      osp_get_recursive_lock                                         */
/*                                                                           */
/* Purpose:   get a lock without blocking if the lock is already owned by    */
/*            the calling thread                                             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     r_mutex      - mutex to lock                            */
/*                                                                           */
/**PROC-**********************************************************************/

void osp_get_recursive_lock(OSP_RECURSIVE_MUTEX *r_mutex)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("osp_get_recursive_lock");

  /***************************************************************************/
  /* if this thread doesn't already own the lock, get it                     */
  /***************************************************************************/
  if (r_mutex->current_locker != pthread_self())
  {
    TRC_DETAIL((TB, "Getting mutex"));
    OSP_GET_LOCK(&(r_mutex->mutex));
    r_mutex->current_locker = pthread_self();
  }

  /***************************************************************************/
  /* increment the lock count                                                */
  /***************************************************************************/
  r_mutex->lock_count++;
  TRC_DETAIL((TB, "Lock count for mutex %p is now %u", r_mutex,
                                                         r_mutex->lock_count));

  TRC_EXIT();

  return;

} /* osp_get_recursive_lock */

/**PROC+**********************************************************************/
/* Name:      osp_release_recursive_lock                                     */
/*                                                                           */
/* Purpose:   Release one level of locking on a lock which may have been     */
/*            got multiple times by the same thread                          */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN     r_mutex      - mutex to release                         */
/*                                                                           */
/**PROC-**********************************************************************/

void osp_release_recursive_lock(OSP_RECURSIVE_MUTEX *r_mutex)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("osp_release_recursive_lock");

  /***************************************************************************/
  /* we really ought to be the owning thread at this juncture                */
  /***************************************************************************/
  TRC_ASSERT((r_mutex->current_locker == pthread_self()), (TB,
                                        "Mutex not owned by current thread!"));

  /***************************************************************************/
  /* decrement the lock count                                                */
  /***************************************************************************/
  r_mutex->lock_count--;
  TRC_DETAIL((TB, "Lock count for mutex %p is now %u", r_mutex,
                                                         r_mutex->lock_count));

  /***************************************************************************/
  /* if the count is now zero, free the lock                                 */
  /***************************************************************************/
  if (r_mutex->lock_count == 0)
  {
    r_mutex->current_locker = NULL;
    OSP_RELEASE_LOCK(&(r_mutex->mutex));
  }

  TRC_EXIT();

  return;

} /* osp_release_recursive_lock */

/**PROC+**********************************************************************/
/* Name:      osp_destroy_recursive_mutex                                    */
/*                                                                           */
/* Purpose:   Destroy the underlying mutex used in recursive locking         */
/*                                                                           */
/* Returns:   OSP_SUCCESS - the mutex was destroyed                          */
/*            OSP_ERROR   - there was a problem destroying the mutex         */
/*                                                                           */
/* Params:    IN     r_mutex      - mutex to destroy                         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC osp_destroy_recursive_mutex(OSP_RECURSIVE_MUTEX *r_mutex)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("osp_destroy_recursive_mutex");

  /***************************************************************************/
  /* Simply destroy the mutex part                                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Destroying mutex %p", r_mutex->mutex));
  rc = osp_destroy_mutex(&(r_mutex->mutex));

  TRC_EXIT();

  return(rc);

} /* osp_destroy_recursive_mutex */
