
/**MOD+***********************************************************************/
/* Module:    comparse.c                                                     */
/*                                                                           */
/* Purpose:   HTTP/MIME header parsing routines                              */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.5                $ $Modtime::   Jan 11 2001 15:56:36   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <comincl.h>

/**PROC+**********************************************************************/
/* Name:      com_parse_headers                                              */
/*                                                                           */
/* Purpose:   Parse a list of HTTP/MIME headers                              */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - success                                */
/*            OSP_ERR_MEMORY        - memory allocation failure              */
/*            OSP_ERR_BAD_REQUEST   - invalid headers                        */
/*            OSP_ERROR             - some other error                       */
/*                                                                           */
/* Params:    IN     header_text    - pointer to the text of the headers     */
/*            IN     length         - maximum length to search for headers   */
/*            OUT    result_out     - parsed list of headers (caller must    */
/*                                      free using com_free_headers())       */
/*            OUT    body_start     - pointer to the first character of the  */
/*                                      body following the headers           */
/*                                                                           */
/* Operation: Parses CRLF-separated headers that end in a double CRLF.       */
/*            Allows line-splitting, when a CRLF is followed by a space or   */
/*            tab. See RFC2616 sections 2.2, 4.1 and 4.2. This routine does  */
/*            not process the header value in any way. Specifically, it does */
/*            not remove quotes.                                             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC com_parse_headers(char *header_text,
                         unsigned int length,
                         COM_HEADERS **result_out,
                         char **body_start)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned int offset = 0;
  unsigned int header_name_start = 0;
  unsigned int header_name_end = 0;
  unsigned int header_value_start = 0;
  unsigned int header_value_end = 0;
  unsigned int max_headers;
  void *new_mem;

  TRC_ENTRY("com_parse_headers");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((result_out != NULL), (TB, "result_out is null"));
  TRC_ASSERT((body_start != NULL), (TB, "body_start is null"));

  /***************************************************************************/
  /* Allocate a COM_HEADERS structure.                                       */
  /***************************************************************************/
  max_headers = COM_LIST_INCREMENT;
  *result_out = (COM_HEADERS *)rm_get_mem(sizeof(COM_HEADERS) +
                                           (max_headers * sizeof(COM_HEADER)));
  if (*result_out == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate memory for headers; max_headers=%u",
                                                                 max_headers));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize the structure.                                               */
  /***************************************************************************/
  (*result_out)->num_headers = 0;
  (*result_out)->headers = (COM_HEADER *)(*result_out + 1);

  /***************************************************************************/
  /* Keep looping for each header.  We break when we're out of headers.      */
  /***************************************************************************/
  while (TRUE)
  {
    /*************************************************************************/
    /* Remember the start of the header name.                                */
    /*************************************************************************/
    header_name_start = offset;

    /*************************************************************************/
    /* See if this is the end of the headers -- we'll be pointing to a       */
    /* CRLF pair if so.                                                      */
    /*************************************************************************/
    if ((offset + 1) >= length)
    {
      /***********************************************************************/
      /* There's not room for a CRLF, and there can't be room for another    */
      /* header, so it's an error.                                           */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to find end of headers"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    if ((header_text[offset] == '\r') &&
        (header_text[offset + 1] == '\n'))
    {
      /***********************************************************************/
      /* We've found the end of the headers, so break out of the loop.       */
      /***********************************************************************/
      TRC_DETAIL((TB, "Found end of headers; got %u total",
                                                  (*result_out)->num_headers));
      break;
    }

    /*************************************************************************/
    /* Find the end of this header (same as the end of the value).  We do    */
    /* this by looking for the first CRLF pair without a space or horizontal */
    /* tab character after it.                                               */
    /*************************************************************************/
    while ((offset + 2) < length)
    {
      if ((header_text[offset] == '\r') &&
          (header_text[offset + 1] == '\n') &&
          (header_text[offset + 2] != ' ') &&
          (header_text[offset + 2] != '\t'))
      {
        /*********************************************************************/
        /* We've found the end of the header.  Remember it.                  */
        /*********************************************************************/
        header_value_end = offset;
        break;
      }
      offset++;
    }

    if ((offset + 2) >= length)
    {
      /***********************************************************************/
      /* We didn't find the end of the header.                               */
      /***********************************************************************/
      TRC_ERROR((TB, "Error parsing header"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Skip back over any spaces after the header value.                     */
    /*************************************************************************/
    while ((header_value_end > header_name_start) &&
           isspace(header_text[header_value_end - 1]))
    {
      header_value_end--;
    }

    if (header_value_end <= header_name_start)
    {
      TRC_ERROR((TB, "Zero-length header"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Find the end of the header name.                                      */
    /*************************************************************************/
    offset = header_name_start;
    while ((offset < header_value_end) && (header_text[offset] != ':'))
    {
      offset++;
    }

    if (offset >= header_value_end)
    {
      TRC_ERROR((TB, "Reached end of buffer while parsing header name"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Remember the end of the header name.                                  */
    /*************************************************************************/
    header_name_end = offset;

    /*************************************************************************/
    /* Skip back over any spaces after the header name.                      */
    /*************************************************************************/
    while ((header_name_end > header_name_start) &&
           isspace(header_text[header_name_end - 1]))
    {
      header_name_end--;
    }

    if (header_name_end <= header_name_start)
    {
      TRC_ERROR((TB, "Zero-length header name"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Skip over any space to the start of the value.                        */
    /*************************************************************************/
    offset++;
    while ((offset < header_value_end) &&
           isspace(header_text[offset]))
    {
      offset++;
    }

    /*************************************************************************/
    /* Remember the start of the value.                                      */
    /*************************************************************************/
    header_value_start = offset;
    if (header_value_start >= header_value_end)
    {
      TRC_ERROR((TB, "Got zero-length header value"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    TRC_DATA("Got header:", header_text + header_name_start,
                                          header_name_end - header_name_start);
    TRC_DATA("Value:", header_text + header_value_start,
                                        header_value_end - header_value_start);

    /*************************************************************************/
    /* We are ready to add this header.  See if the list is big enough.      */
    /*************************************************************************/
    if ((*result_out)->num_headers == max_headers)
    {
      max_headers += COM_LIST_INCREMENT;
      TRC_NORMAL((TB, "Reallocating list for %u headers", max_headers));
      new_mem = rm_realloc_mem(*result_out,
                     sizeof(COM_HEADERS) + (max_headers * sizeof(COM_HEADER)));
      if (new_mem == NULL)
      {
        TRC_ERROR((TB, "Failed to realloc memory for headers; max_headers=%u",
                                                                 max_headers));
        rc = OSP_ERR_MEMORY;
        goto EXIT_LABEL;
      }

      /***********************************************************************/
      /* Use the reallocated memory and update the pointer in the structure. */
      /***********************************************************************/
      *result_out = (COM_HEADERS *)new_mem;
      (*result_out)->headers = (COM_HEADER *)(*result_out + 1);
    }

    /*************************************************************************/
    /* Now add the latest header to the list.                                */
    /*************************************************************************/
    (*result_out)->headers[(*result_out)->num_headers].name.buffer =
                                               header_text + header_name_start;
    (*result_out)->headers[(*result_out)->num_headers].name.len =
                                           header_name_end - header_name_start;
    (*result_out)->headers[(*result_out)->num_headers].value.buffer =
                                              header_text + header_value_start;
    (*result_out)->headers[(*result_out)->num_headers].value.len =
                                         header_value_end - header_value_start;
    (*result_out)->num_headers++;

    /*************************************************************************/
    /* Set 'offset' to indicate the start of the next header.                */
    /*************************************************************************/
    offset = header_value_end + 2;
  }

  /***************************************************************************/
  /* We've successfully parsed the headers.  Indicate the start of the body  */
  /* part and return success.                                                */
  /***************************************************************************/
  *body_start = header_text + offset + 2;
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    TRC_DATA("Failed while parsing this header block:", header_text, length);
    rm_release_mem(*result_out);
    *result_out = NULL;
  }

  TRC_EXIT();

  return(rc);

} /* com_parse_headers */

/**PROC+**********************************************************************/
/* Name:      com_find_header_value                                          */
/*                                                                           */
/* Purpose:   Search for a specific header in a COM_HEADERS structure and    */
/*            return its value, if present. The search is case-insensitive.  */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - success                                */
/*            OSP_ERROR             - header not found                       */
/*                                                                           */
/* Params:    IN     header_list    - the header list structure to search    */
/*            IN     header_name    - the name of the header to search for   */
/*                                      (not null-terminated)                */
/*            IN     header_length  - the length of the header name          */
/*            IN     result         - the structure to fill in with a        */
/*                                      pointer to and length of the result  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC com_find_header_value(const COM_HEADERS *header_list,
                             const char *header_name,
                             const unsigned int header_len,
                             COM_STRING *result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int found = FALSE;
  int ii;

  TRC_ENTRY("com_find_header_value");

  /***************************************************************************/
  /* Go through each header value in turn to find the one we want.           */
  /***************************************************************************/
  for (ii = 0; ii < header_list->num_headers; ii++)
  {
    if (com_strings_match(&header_list->headers[ii].name,
                                                      header_name, header_len))
    {
      found = TRUE;
      break;
    }
  }

  /***************************************************************************/
  /* If we've found what we wanted, fill in the result structure to point to */
  /* the value corresponding to this header.                                 */
  /***************************************************************************/
  if (found)
  {
    result->buffer = header_list->headers[ii].value.buffer;
    result->len = header_list->headers[ii].value.len;
    rc = OSP_SUCCESS;
  }

  TRC_EXIT();

  return(rc);

} /* com_find_header_value */

/**PROC+**********************************************************************/
/* Name:      com_parse_media_type                                           */
/*                                                                           */
/* Purpose:   Parse a media-type value as found in a Content-Type header     */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - success                                */
/*            OSP_ERR_MEMORY        - memory allocation failure              */
/*            OSP_ERR_BAD_REQUEST   - invalid headers                        */
/*            OSP_ERROR             - some other error                       */
/*                                                                           */
/* Params:    IN     string         - the string to parse (not null-         */
/*                                      terminated)                          */
/*            IN     string_len     - the length of the string               */
/*            IN     type_subtype   - structure to fill in with a pointer to */
/*                                      and length of the type/subtype value */
/*            OUT    result_out     - list of parameters                     */
/*                                                                           */
/* Operation: Expects a media-type such as 'text/plain' followed by zero or  */
/*            more parameters. See RFC2045, section 5.1 for the syntax.      */
/*            This routine removes quotes from parameter values.             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC com_parse_media_type(char *string,
                            unsigned int string_len,
                            COM_STRING *type_subtype,
                            COM_PARAMS **params_out)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned int max_params;
  unsigned int offset = 0;
  unsigned int param_attribute_start = 0;
  unsigned int param_attribute_end = 0;
  unsigned int param_value_start = 0;
  unsigned int param_value_end = 0;
  void *new_mem;

  TRC_ENTRY("com_parse_media_type");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((type_subtype != NULL), (TB, "type_subtype is null"));
  TRC_ASSERT((params_out != NULL), (TB, "params_out is null"));

  /***************************************************************************/
  /* We have to have some characters in the string.                          */
  /***************************************************************************/
  if (string_len == 0)
  {
    TRC_ERROR((TB, "Got zero-length string"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Allocate the parameter list.                                            */
  /***************************************************************************/
  max_params = COM_LIST_INCREMENT;
  *params_out = (COM_PARAMS *)rm_get_mem(sizeof(COM_PARAMS) +
                                             (max_params * sizeof(COM_PARAM)));
  if (*params_out == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate memory for params; max_params=%u",
                                                                  max_params));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize the count of headers in the list.                            */
  /***************************************************************************/
  (*params_out)->num_params = 0;
  (*params_out)->params = (COM_PARAM *)(*params_out + 1);

  /***************************************************************************/
  /* Skip any linear white space (LWS) at the start of the string.  The      */
  /* type/subtype value is never quoted, so it starts right after the LWS.   */
  /***************************************************************************/
  COM_SKIP_LWS(string, offset, string_len);
  type_subtype->buffer = string + offset;

  /***************************************************************************/
  /* Find the '/' in the type/subtype string.  It should be the first        */
  /* non-token character.                                                    */
  /***************************************************************************/
  COM_FIND_TOKEN_END(string, offset, string_len);

  if ((offset >= string_len) || (string[offset] != '/'))
  {
    TRC_ERROR((TB, "Invalid type/subtype separator at offset %u", offset));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Find the end of the subtype and save it.                                */
  /***************************************************************************/
  offset++;
  COM_FIND_TOKEN_END(string, offset, string_len);
  type_subtype->len = (string + offset) - type_subtype->buffer;
  TRC_DATA("Got MIME type/subtype:", type_subtype->buffer, type_subtype->len);

  /***************************************************************************/
  /* Skip any LWS after the type/subtype.                                    */
  /***************************************************************************/
  COM_SKIP_LWS(string, offset, string_len);

  /***************************************************************************/
  /* Loop, pulling parameters from the string.                               */
  /***************************************************************************/
  while (TRUE)
  {
    /*************************************************************************/
    /* See if there are any (more) parameters.  There are none if we're at   */
    /* the end of the string, or if the LWS we just skipped was actually     */
    /* just a CRLF pair with no spaces or tabs after it.                     */
    /*************************************************************************/
    if ((offset >= string_len) ||
        ((string[offset - 2] == '\r') && (string[offset - 1] == '\n')))
    {
      TRC_DETAIL((TB, "Finished parsing parameters; got %u",
                                                   (*params_out)->num_params));
      break;
    }

    /*************************************************************************/
    /* We should now have a semicolon separator.                             */
    /*************************************************************************/
    if (string[offset] != ';')
    {
      TRC_ERROR((TB, "Invalid parameter separator at offset %u", offset));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }
    offset++;

    /*************************************************************************/
    /* Skip any LWS after the parameter separator.                           */
    /*************************************************************************/
    COM_SKIP_LWS(string, offset, string_len);

    /*************************************************************************/
    /* Save the start of the attribute.                                      */
    /*************************************************************************/
    param_attribute_start = offset;

    /*************************************************************************/
    /* Find the end of the attribute.                                        */
    /*************************************************************************/
    COM_FIND_TOKEN_END(string, offset, string_len);
    param_attribute_end = offset;
    TRC_DATA("Got attribute:", string + param_attribute_start,
                                  param_attribute_end - param_attribute_start);

    /*************************************************************************/
    /* No LWS is allowed between a parameter, the equals sign and the value. */
    /*************************************************************************/
    if (((offset + 1) >= string_len) ||
        (string[offset] != '=') ||
        (isspace(string[offset + 1])))
    {
      TRC_ERROR((TB, "Invalid parameter syntax at offset %u", offset));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }
    offset++;

    /*************************************************************************/
    /* Remember the start of the value.                                      */
    /*************************************************************************/
    param_value_start = offset;

    /*************************************************************************/
    /* See if this is a token or a quoted-string, find its end and remember  */
    /* it.                                                                   */
    /*************************************************************************/
    if (string[offset] == '"')
    {
      /***********************************************************************/
      /* Note that a quoted string may contain escaped characters.  A caller */
      /* who wants to deal with these will have to take a copy of the string */
      /* and de-escape it.                                                   */
      /***********************************************************************/
      param_value_start++;
      COM_FIND_QUOTED_STRING_END(string, offset, string_len);
      param_value_end = offset;

      /***********************************************************************/
      /* Check that we have the closing quote.                               */
      /***********************************************************************/
      if (string[offset] != '"')
      {
        TRC_ERROR((TB, "Reached end of string looking for closing quote"));
        rc = OSP_ERR_BAD_REQUEST;
        goto EXIT_LABEL;
      }
      offset++;
    }
    else
    {
      COM_FIND_TOKEN_END(string, offset, string_len);
      param_value_end = offset;
    }
    TRC_DATA("Value:", string + param_value_start,
                                          param_value_end - param_value_start);

    /*************************************************************************/
    /* We are ready to add this parameter.  See if the list is big enough.   */
    /*************************************************************************/
    if ((*params_out)->num_params == max_params)
    {
      max_params += COM_LIST_INCREMENT;
      TRC_NORMAL((TB, "Reallocating list for %u params", max_params));
      new_mem = rm_realloc_mem(*params_out,
                        sizeof(COM_PARAMS) + (max_params * sizeof(COM_PARAM)));
      if (new_mem == NULL)
      {
        TRC_ERROR((TB, "Failed to realloc memory for params; max_params=%u",
                                                                  max_params));
        rc = OSP_ERR_MEMORY;
        goto EXIT_LABEL;
      }

      /***********************************************************************/
      /* Use the reallocated memory and update the pointer in the structure. */
      /***********************************************************************/
      *params_out = (COM_PARAMS *)new_mem;
      (*params_out)->params = (COM_PARAM *)(*params_out + 1);
    }

    /*************************************************************************/
    /* Now add the latest parameter to the list.                             */
    /*************************************************************************/
    (*params_out)->params[(*params_out)->num_params].attribute.buffer =
                                                string + param_attribute_start;
    (*params_out)->params[(*params_out)->num_params].attribute.len =
                                   param_attribute_end - param_attribute_start;
    (*params_out)->params[(*params_out)->num_params].value.buffer =
                                                    string + param_value_start;
    (*params_out)->params[(*params_out)->num_params].value.len =
                                           param_value_end - param_value_start;
    (*params_out)->num_params++;

    /*************************************************************************/
    /* Skip any LWS after the parameter we've just parsed.                   */
    /*************************************************************************/
    COM_SKIP_LWS(string, offset, string_len);

    /*************************************************************************/
    /* 'offset' now points to the separator between the parameter we've just */
    /* parsed and the next parameter.                                        */
    /*************************************************************************/
  }

  /***************************************************************************/
  /* We've successfully parsed the parameters.                               */
  /***************************************************************************/
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    TRC_DATA("Failed while parsing this string:", string, string_len);
    rm_release_mem(*params_out);
    *params_out = NULL;
  }

  TRC_EXIT();

  return(rc);

} /* com_parse_media_type */

/**PROC+**********************************************************************/
/* Name:      com_find_param_value                                           */
/*                                                                           */
/* Purpose:   Search for a specific parameter in a COM_PARAMS                */
/*            structure and return its value, if present. The search is      */
/*            case-insensitive.                                              */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - success                                */
/*            OSP_ERROR             - param not found                        */
/*                                                                           */
/* Params:    IN     param_list     - the parameter list structure to search */
/*            IN     param_name     - the parameter name to search for (not  */
/*                                      null-terminated)                     */
/*            IN     param_len      - the length of the parameter name       */
/*            IN     result         - the structure to fill in with a        */
/*                                      pointer to and length of the result  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC com_find_param_value(const COM_PARAMS *param_list,
                            const char *param_name,
                            const unsigned int param_len,
                            COM_STRING *result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int found = FALSE;
  int ii;

  TRC_ENTRY("com_find_param_value");

  /***************************************************************************/
  /* Go through each header value in turn to find the one we want.           */
  /***************************************************************************/
  for (ii = 0; ii < param_list->num_params; ii++)
  {
    if (com_strings_match(&param_list->params[ii].attribute,
                                                        param_name, param_len))
    {
      found = TRUE;
      break;
    }
  }

  /***************************************************************************/
  /* If we've found what we wanted, fill in the result structure to point to */
  /* the value corresponding to this parameter.                              */
  /***************************************************************************/
  if (found)
  {
    result->buffer = param_list->params[ii].value.buffer;
    result->len = param_list->params[ii].value.len;
    rc = OSP_SUCCESS;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* com_find_param_value */

/**PROC+**********************************************************************/
/* Name:      com_strings_match                                              */
/*                                                                           */
/* Purpose:   Compare a COM_STRING with some other string (case-insensitive) */
/*                                                                           */
/* Returns:   1           - the strings are identical                        */
/*            0           - the strings are different                        */
/*                                                                           */
/* Params:    IN     string1        - the first string to compare            */
/*            IN     string2_buffer - a buffer containing the second string  */
/*                                      to compare (not null-terminated)     */
/*            IN     string2_len    - the length of the second string        */
/*                                                                           */
/* Operation: First compares the string lengths, then compares each char.    */
/*                                                                           */
/**PROC-**********************************************************************/

int com_strings_match(const COM_STRING *string1,
                      const char *string2_buffer,
                      const unsigned int string2_len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = 0;

  TRC_ENTRY("com_strings_match");

  /***************************************************************************/
  /* If the lengths are the same, compare the actual strings.  We use a      */
  /* limited-length compare because string1's buffer isn't null-terminated.  */
  /***************************************************************************/
  if ((string1->len == string2_len) &&
      (strncasecmp(string1->buffer, string2_buffer, string2_len) == 0))
  {
    rc = 1;
  }

  TRC_EXIT();

  return(rc);

} /* com_strings_match */

