
/**MOD+***********************************************************************/
/* Module:    lamctrl.c                                                      */
/*                                                                           */
/* Purpose:   LDAP Access Module Control routines                            */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.13               $ $Modtime::   Jan 11 2001 15:56:36   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <lamincl.h>

/**PROC+**********************************************************************/
/* Name:      lam_init                                                       */
/*                                                                           */
/* Purpose:   To initialize the LAM component.                               */
/*                                                                           */
/* Returns:   OSP_SUCCESS - LAM was initialized successfully                 */
/*            OSP_ERROR   - Another error occured                            */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC lam_init(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;

  TRC_ENTRY("lam_init");

  /***************************************************************************/
  /* set default return code                                                 */
  /***************************************************************************/
  rc = OSP_SUCCESS;

  /***************************************************************************/
  /* We don't mind if we're multiply initialized                             */
  /***************************************************************************/
  if (lam_initialized)
  {
    TRC_NORMAL((TB, "LAM already initialized - OK"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create the thread key we will use to identify threads as ours           */
  /***************************************************************************/
  if (pthread_key_create(&lam_thread_key, NULL) != 0)
  {
    TRC_ERROR((TB, "Failed to create lam thread key"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create and lock the lam the mutex                                       */
  /***************************************************************************/
  if (osp_create_mutex(&lam_mutex) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create lam mutex"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  OSP_GET_LOCK(&lam_mutex);

  /***************************************************************************/
  /* Initialize the host connection list                                     */
  /***************************************************************************/
  LIST_INIT(&lam_host_conn_list);

  /***************************************************************************/
  /* All done                                                                */
  /***************************************************************************/
  TRC_NORMAL((TB, "LAM initialized successfully"));
  lam_initialized = TRUE;
  OSP_RELEASE_LOCK(&lam_mutex);

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* lam_init */

/**PROC+**********************************************************************/
/* Name:      lam_term                                                       */
/*                                                                           */
/* Purpose:   To terminate the LAM component.                                */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - LAM was terminated successfully             */
/*            OSP_ERR_NOT_INIT - called before lam_init()                    */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC lam_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  PLAM_HOST_CONN host_conn;

  TRC_ENTRY("lam_term");

  /***************************************************************************/
  /* set default return code                                                 */
  /***************************************************************************/
  rc = OSP_SUCCESS;

  /***************************************************************************/
  /* ensure no one starts any new operations under us                        */
  /***************************************************************************/
  OSP_GET_LOCK(&lam_mutex);

  /***************************************************************************/
  /* check we've been initialized!                                           */
  /***************************************************************************/
  if (!lam_initialized)
  {
    TRC_ERROR((TB, "Called before init"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* apps *should* call lam_close() for all open handles before calling      */
  /* lam_term()...  If they haven't, we'll do it for them, which will also   */
  /* cancel any pending LDAP operations                                      */
  /***************************************************************************/
  while (!LIST_IS_EMPTY(&lam_host_conn_list))
  {
    host_conn = (PLAM_HOST_CONN)LIST_REMOVE_HEAD(&lam_host_conn_list);
    TRC_DETAIL((TB, "Remove conn %p from list %p", host_conn,
                                                         &lam_host_conn_list));
    lam_close_internal(host_conn);
  }

  /***************************************************************************/
  /* Destroy the thread key                                                  */
  /***************************************************************************/
  if (pthread_key_delete(lam_thread_key) != 0)
  {
    TRC_ERROR((TB, "Failed to delete lam thread key"));
    rc = OSP_ERROR;
  }

  /***************************************************************************/
  /* now we really aren't initialized                                        */
  /***************************************************************************/
  TRC_NORMAL((TB, "LAM Termination complete"));
  lam_initialized = FALSE;
  OSP_RELEASE_LOCK(&lam_mutex);
  osp_destroy_mutex(&lam_mutex);

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* lam_term */

/**PROC+**********************************************************************/
/* Name:      lam_open                                                       */
/*                                                                           */
/* Purpose:   To open a directory for access. This does not necessarily      */
/*            imply that a network connection is made to the directory.      */
/*                                                                           */
/* Returns:   OSP_SUCCESS       - the directory was opened successfully      */
/*            OSP_ERR_NOT_INIT  - called before lam_init()                   */
/*            OSP_ERR_BAD_PARAM - host name is NULL                          */
/*            OSP_ERROR         - Another error occured                      */
/*                                                                           */
/* Params:    IN     host         - pointer to a string containing the name  */
/*                                  or ip address of the LDAP host           */
/*            IN     port         - port number to connect to                */
/*            IN     who,         - user dn and password information used to */
/*                   cred           access the directory on host             */
/*            OUT    handle       - pointer to an opaque handle value        */
/*                                  filled in by this function               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC lam_open(char *host,
                int port,
                char *who,
                char *cred,
                LAM_HANDLE *handle)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  PLAM_HOST_CONN host_conn;

  TRC_ENTRY("lam_open");

  /***************************************************************************/
  /* set default return code                                                 */
  /***************************************************************************/
  rc = OSP_SUCCESS;

  /***************************************************************************/
  /* check we've been initialized!                                           */
  /***************************************************************************/
  OSP_GET_LOCK(&lam_mutex);
  if (!lam_initialized)
  {
    TRC_ERROR((TB, "Called before init"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* check there's something to connect to                                   */
  /***************************************************************************/
  if ((host == NULL) || (*host == 0))
  {
    TRC_ERROR((TB, "No host supplied"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* and something to connect to it with                                     */
  /***************************************************************************/
  if (who == NULL)
  {
    TRC_ERROR((TB, "No dn supplied"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  if (cred == NULL)
  {
    TRC_ERROR((TB, "No credentials supplied"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up a thread pool per session (ie per call to lam_open())            */
  /***************************************************************************/
  TRC_NORMAL((TB, "Creating thread pool per call"));
  host_conn = NULL;

  /***************************************************************************/
  /* If we found a connection to this host, then we'll just use that one     */
  /***************************************************************************/
  if (host_conn != NULL)
  {
    TRC_NORMAL((TB, "Found host conn %p for '%s:%d'", host_conn, host, port));
  }
  else
  {
    /*************************************************************************/
    /* If not, we need to create a new connection                            */
    /*************************************************************************/
    rc = lam_create_host_connection(host, port, who, cred, &host_conn);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to create new host connection"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Add the new connection to LAM's list                                  */
    /*************************************************************************/
    TRC_DETAIL((TB, "Insert conn %p in list %p", host_conn,
                                                         &lam_host_conn_list));
    LIST_INSERT_TAIL(&lam_host_conn_list, &host_conn->links);
  }

  /***************************************************************************/
  /* If we got here, we've opened the connection successfully                */
  /***************************************************************************/
  host_conn->open_count++;
  TRC_NORMAL((TB, "Host %p (%s:%d) open count now %d",
              host_conn,
              host_conn->host,
              host_conn->port,
              host_conn->open_count));
  *handle = (LAM_HANDLE)host_conn;
  host_conn = NULL;

EXIT_LABEL:

  /***************************************************************************/
  /* Release the lam_mutex                                                   */
  /***************************************************************************/
  OSP_RELEASE_LOCK(&lam_mutex);

  TRC_EXIT();

  return(rc);

} /* lam_open */

/**PROC+**********************************************************************/
/* Name:      lam_close                                                      */
/*                                                                           */
/* Purpose:   To close an open directory.                                    */
/*                                                                           */
/* Returns:   OSP_SUCCESS        - the directory was closed successfully     */
/*            OSP_ERR_NOT_INIT   - called before lam_init()                  */
/*            OSP_ERR_BAD_HANDLE - bad LAM_HANDLE supplied                   */
/*                                                                           */
/* Params:    IN/OUT handle       - pointer to the handle returned by        */
/*                                  lam_open                                 */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC lam_close(LAM_HANDLE *handle)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  PLAM_HOST_CONN host_conn;
  OSP_RC rc;

  TRC_ENTRY("lam_close");

  /***************************************************************************/
  /* set default return code                                                 */
  /***************************************************************************/
  rc = OSP_SUCCESS;

  /***************************************************************************/
  /* check we've been initialized!                                           */
  /***************************************************************************/
  if (!lam_initialized)
  {
    TRC_ERROR((TB, "Called before init"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* check handle is valid!                                                  */
  /***************************************************************************/
  if (!handle || !(*handle))
  {
    TRC_ERROR((TB, "Called with NULL handle"));
    rc = OSP_ERR_BAD_HANDLE;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* access the underlying host connection - requires that we hold the lam   */
  /* mutex                                                                   */
  /***************************************************************************/
  OSP_GET_LOCK(&lam_mutex);
  host_conn = (PLAM_HOST_CONN)*handle;

  /***************************************************************************/
  /* reduce the open count                                                   */
  /***************************************************************************/
  host_conn->open_count--;
  TRC_NORMAL((TB, "Host %p (%s) open count now %d", host_conn, host_conn->host,
                                                       host_conn->open_count));

  /***************************************************************************/
  /* if that was the last user of this connection...                         */
  /***************************************************************************/
  if (host_conn->open_count == 0)
  {
    /*************************************************************************/
    /* ...shut the connection down                                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Remove conn %p from list %p", host_conn,
                                                         &lam_host_conn_list));
    LIST_REMOVE_ITEM(&host_conn->links);
    lam_close_internal(host_conn);
  }

  /***************************************************************************/
  /* make sure caller doesn't use the handle again                           */
  /***************************************************************************/
  *handle = NULL;

  /***************************************************************************/
  /* release the lam mutex                                                   */
  /***************************************************************************/
  OSP_RELEASE_LOCK(&lam_mutex);

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* lam_close */

/**PROC+**********************************************************************/
/* Name:      lam_close_internal                                             */
/*                                                                           */
/* Purpose:   Terminates threads for a connection and frees the              */
/*            connection resources                                           */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     conn         - connection to shutdown                   */
/*                                                                           */
/**PROC-**********************************************************************/

void lam_close_internal(PLAM_HOST_CONN conn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CANCEL_FD *cfd;
  PLAM_REQUEST_INFO request;
  TRC_ENTRY("lam_close_internal");

  /***************************************************************************/
  /* Check that supplied connection is valid                                 */
  /***************************************************************************/
  TRC_ASSERT((conn != NULL), (TB, "Asked to close NULL connection"));

  /***************************************************************************/
  /* signal that we're going down                                            */
  /***************************************************************************/
  TRC_NORMAL((TB, "Shutting down connection %p", conn));
  OSP_GET_LOCK(&conn->connection_mutex);
  conn->shutting_down = TRUE;

  /***************************************************************************/
  /* writing to/closing the cancel pipes will cause any threads processing   */
  /* a request on this connection to abort.  The thread will free up the     */
  /* request resources - we just tidy up the fd                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Closing active threads"));
  while (!LIST_IS_EMPTY(&conn->cancel_fds))
  {
    cfd = (CANCEL_FD *)LIST_REMOVE_HEAD(&conn->cancel_fds);
    TRC_DETAIL((TB, "Removed cfd %p from list %p", cfd, &conn->cancel_fds));
    close(cfd->fd);
    TRC_DETAIL((TB, "Freed cancel fd %p", cfd));
    rm_release_mem(cfd);
  }

  /***************************************************************************/
  /* Now we empty the request queue.  Any requests that are being processed  */
  /* by a thread won't be on the queue, and the thread will take care of     */
  /* freeing the resources                                                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing requests"));
  while (!LIST_IS_EMPTY(&conn->request_queue))
  {
    request = (PLAM_REQUEST_INFO)LIST_REMOVE_HEAD(&conn->request_queue);
    TRC_DETAIL((TB, "Removed request %p from list %p",
                                               request, &conn->request_queue));
    lam_do_request_callback(request, NULL, NULL);
  }

  /***************************************************************************/
  /* All the connection threads should now be blocked waiting for some data  */
  /* to be placed on the queue.  With shutting_down set, calling             */
  /* cond_broadcast() will cause all the threads on this connection to shut  */
  /* down in turn                                                            */
  /***************************************************************************/
  TRC_NORMAL((TB, "Broadcasting to shutdown threads"));
  pthread_cond_broadcast(&conn->data_on_queue);
  OSP_RELEASE_LOCK(&conn->connection_mutex);

  /***************************************************************************/
  /* Wait until all the threads have gone                                    */
  /***************************************************************************/
  TRC_NORMAL((TB, "Waiting for threads to end"));
  OSP_GET_LOCK(&conn->connection_mutex);
  while (conn->thread_count > 0)
  {
    OSP_RELEASE_LOCK(&conn->connection_mutex);
    sleep(1);
    TRC_DETAIL((TB, "...waiting..."));
    OSP_GET_LOCK(&conn->connection_mutex);
  }
  TRC_NORMAL((TB, "...Done"));

  /***************************************************************************/
  /* finish tidy up by destroying the mutex and condition attribute and      */
  /* finally freeing the connection memory                                   */
  /***************************************************************************/
  OSP_RELEASE_LOCK(&conn->connection_mutex);
  osp_destroy_mutex(&conn->connection_mutex);
  pthread_cond_destroy(&conn->data_on_queue);

  rm_release_mem(conn->host);
  rm_release_mem(conn->who);
  rm_release_mem(conn->cred);
  rm_release_mem(conn);

  TRC_EXIT();

  return;

} /* lam_close_internal */

/**PROC+**********************************************************************/
/* Name:      lam_is_lam_thread                                              */
/*                                                                           */
/* Purpose:   Check if the calling thread is a LAM thread                    */
/*                                                                           */
/* Returns:   TRUE if calling thread was created by LAM                      */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

int lam_is_lam_thread(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  pthread_t calling_thread;
  void *thread_specific;

  TRC_ENTRY("lam_is_lam_thread");

  /***************************************************************************/
  /* Find out who the caller is                                              */
  /***************************************************************************/
  calling_thread = pthread_self();
  TRC_DETAIL((TB, "Testing thread id %u", (unsigned int)calling_thread));

  /***************************************************************************/
  /* Get the lam thread key data.  If this fails, or the key doesn't match   */
  /* up, then we're probably not being called on a lam thread                */
  /***************************************************************************/
  thread_specific = pthread_getspecific(lam_thread_key);
  if (thread_specific != NULL)
  {
    TRC_DETAIL((TB, "Got thread specific data"));
    if (thread_specific == lam_key_name)
    {
      TRC_DETAIL((TB, "Found correct key data - is a lam thread"));
      rc = TRUE;
    }
  }

  TRC_EXIT();

  return(rc);

} /* lam_is_lam_thread */
