
/**MOD+***********************************************************************/
/* Module:    lammain.c                                                      */
/*                                                                           */
/* Purpose:   LDAP Access Module Search routines                             */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.6                $ $Modtime::   Jan 11 2001 15:56:36   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <lamincl.h>

/**PROC+**********************************************************************/
/* Name:      lam_search                                                     */
/*                                                                           */
/* Purpose:   To start an LDAP search operation.                             */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - the search was started successfully      */
/*            OSP_ERR_NOT_INIT    - called before lam_init()                 */
/*            OSP_ERR_BAD_HANDLE  - bad LAM_HANDLE supplied                  */
/*            OSP_ERR_BAD_PARAM   - the callback, base or filter was NULL    */
/*            OSP_ERR_BIND_FAILED - LAM could not bind to the host specified */
/*                                  in lam_open                              */
/*            OSP_ERROR           - Another error occured                    */
/*                                                                           */
/* Params:    IN     handle       - the handle returned from lam_open()      */
/*            IN     base,        - see the OpenLDAP documentation for       */
/*                   scope,         ldap_search() for details of these       */
/*                   filter,         parameters                              */
/*                   attrs,                                                  */
/*                   attrs_only                                              */
/*            IN     user_data    - correlator to supply on results callback */
/*            IN     result_callback                                         */
/*                                - the function to call with search results */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC lam_search(LAM_HANDLE handle,
                  char *base,
                  int scope,
                  char *filter,
                  char **attrs,
                  int attrs_only,
                  void *user_data,
                  LAM_RESULT_CALLBACK result_callback)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  PLAM_HOST_CONN host_conn;
  PLAM_REQUEST_INFO new_rq = NULL;
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("lam_search");

  /***************************************************************************/
  /* check we've been initialized                                            */
  /***************************************************************************/
  if (!lam_initialized)
  {
    TRC_ERROR((TB, "Called before init"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* check handle is valid                                                   */
  /***************************************************************************/
  if (handle == NULL)
  {
    TRC_ERROR((TB, "Called with NULL handle"));
    rc = OSP_ERR_BAD_HANDLE;
    goto EXIT_LABEL;
  }
  host_conn = (PLAM_HOST_CONN)handle;

  /***************************************************************************/
  /* check that at least some of the threads associated with the handle are  */
  /* up and running                                                          */
  /***************************************************************************/
  if (host_conn->thread_count == 0)
  {
    /*************************************************************************/
    /* The threads must have failed to bind to the host - there's no more we */
    /* can do                                                                */
    /*************************************************************************/
    TRC_ALERT((TB, "no threads left for connection"));
    rc = OSP_ERR_BIND_FAILED;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* param checking                                                          */
  /***************************************************************************/
  if ((result_callback == NULL) || (base == NULL) || (filter == NULL))
  {
    TRC_ERROR((TB, "Bad parameters cb:%p, base:%p, filter:%p",
               result_callback, base, filter));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now build the search request                                            */
  /***************************************************************************/
  new_rq = lam_build_search_request(base, scope, filter, attrs, attrs_only,
                                                   user_data, result_callback);
  if (new_rq == NULL)
  {
    TRC_ERROR((TB, "Failed to build new request"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* And submit it to the queue                                              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Submitting new request %p", new_rq));
  lam_submit_new_request(host_conn, new_rq);

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* lam_search */

