#ifndef LAMSTRC_INCLUDED
#define LAMSTRC_INCLUDED

/**INC+***********************************************************************/
/* Header:    lamstrc.h                                                      */
/*                                                                           */
/* Purpose:   LDAP Access Manager (LAM) Structure definitions.               */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.7                $ $Modtime::   Jan 11 2001 15:56:38   $  */
/*                                                                           */
/**INC-***********************************************************************/

/**STRUCT+********************************************************************/
/* Structure:   CANCEL_FD                                                    */
/*                                                                           */
/* Description: Element in list of pre-connection cancel fd's                */
/*                                                                           */
/*****************************************************************************/

typedef struct CANCEL_FD
{
  /***************************************************************************/
  /* Structure containing forward and back links                             */
  /***************************************************************************/
  LIST_ENTRY links;

  /***************************************************************************/
  /* File descriptor for write end of cancel pipe                            */
  /***************************************************************************/
  int         fd;

} CANCEL_FD;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: LAM_HOST_CONN                                                  */
/*                                                                           */
/* Description: LAM Connection Data.  Per-host data, accessed by casting LAM */
/*              Handle                                                       */
/*                                                                           */
/* Allocated:   in lam_open() on first connection to the host                */
/*                                                                           */
/* Freed:       by lam_close_internal() when last connection to this host is */
/*              closed                                                       */
/*                                                                           */
/*****************************************************************************/

typedef struct lam_host_conn
{
  /***************************************************************************/
  /* Structure containing forward and back links, protected by lam_mutex     */
  /***************************************************************************/
  LIST_ENTRY links;

  /***************************************************************************/
  /* Count of opens to this connection, protected by lam_mutex               */
  /***************************************************************************/
  int open_count;

  /***************************************************************************/
  /* Host name for this connection                                           */
  /***************************************************************************/
  char *host;

  /***************************************************************************/
  /* Port number for this connection                                         */
  /***************************************************************************/
  int port;

  /***************************************************************************/
  /* DN used to access host                                                  */
  /***************************************************************************/
  char *who;

  /***************************************************************************/
  /* Credentials (typically a password) used to access host                  */
  /***************************************************************************/
  char *cred;

  /***************************************************************************/
  /* Mutex used to protect the per-connection control variables              */
  /***************************************************************************/
  pthread_mutex_t connection_mutex;

  /***************************************************************************/
  /* Count of threads associated with this connection                        */
  /***************************************************************************/
  int thread_count;

  /***************************************************************************/
  /* Connection shutting down flag; when set, all threads should exit        */
  /***************************************************************************/
  int shutting_down;

  /***************************************************************************/
  /* The queue of LDAP search requests outstanding on this connection        */
  /***************************************************************************/
  LIST_ENTRY request_queue;

  /***************************************************************************/
  /* Condition attribute used to signal new data on the queue                */
  /***************************************************************************/
  pthread_cond_t data_on_queue;

  /***************************************************************************/
  /* List containing the write ends of all the thread cancel pipes           */
  /***************************************************************************/
  LIST_ENTRY cancel_fds;

} LAM_HOST_CONN, * PLAM_HOST_CONN;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: LAM_THREAD_DATA                                                */
/*                                                                           */
/* Description: LAM Thread Info per-thread data                              */
/*                                                                           */
/* Allocated:   in lam_open() on first connection to a host                  */
/*                                                                           */
/* Freed:       by thread before it exits                                    */
/*                                                                           */
/*****************************************************************************/

typedef struct lam_thread_data
{
  /***************************************************************************/
  /* LAM connection with which this thread is associated                     */
  /***************************************************************************/
  LAM_HOST_CONN *conn;

  /***************************************************************************/
  /* LDAP connection to the host                                             */
  /***************************************************************************/
  LDAP *ld;

  /***************************************************************************/
  /* Read end of this thread's cancel pipe                                   */
  /***************************************************************************/
  int read_fd;

} LAM_THREAD_DATA, * PLAM_THREAD_DATA;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: LAM_SEARCH_PARAMS                                              */
/*                                                                           */
/* Description: Structure used to store parameters from search requests      */
/*****************************************************************************/

typedef struct lam_search_params
{
  /***************************************************************************/
  /* Base dn for search                                                      */
  /***************************************************************************/
  char *base;

  /***************************************************************************/
  /* Scope of search                                                         */
  /***************************************************************************/
  int scope;

  /***************************************************************************/
  /* Attribute filter                                                        */
  /***************************************************************************/
  char *filter;

  /***************************************************************************/
  /* Attributes to return                                                    */
  /***************************************************************************/
  char **attrs;

  /***************************************************************************/
  /* Flag set to indicate that only attribute names and not values should be */
  /* returned                                                                */
  /***************************************************************************/
  int attrs_only;

} LAM_SEARCH_PARAMS, * PLAM_SEARCH_PARAMS;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure: LAM_REQUEST_INFO                                               */
/*                                                                           */
/* Description: LAM Request Information.  Used to store search               */
/*              requests on the queue.                                       */
/*                                                                           */
/* Allocated:   in lam_search()                                              */
/*                                                                           */
/* Freed:       in lam_do_request_callback(), called either from             */
/*              lam_thread_proc() or lam_close()                             */
/*                                                                           */
/*****************************************************************************/

typedef struct lam_request_info
{
  /***************************************************************************/
  /* Structure containing forward and back links.  Protected by connection   */
  /* mutex                                                                   */
  /***************************************************************************/
  LIST_ENTRY links;

  /***************************************************************************/
  /* Host connection to be used for this request                             */
  /***************************************************************************/
  PLAM_HOST_CONN lam_host_conn;

  /***************************************************************************/
  /* request parameters                                                      */
  /***************************************************************************/
  LAM_SEARCH_PARAMS search;

  /***************************************************************************/
  /* Userdata supplied with the request; returned as a parameter on the      */
  /* result callback                                                         */
  /***************************************************************************/
  void *user_data;

  /***************************************************************************/
  /* Function called with the results of the requests                        */
  /***************************************************************************/
  LAM_RESULT_CALLBACK callback;

} LAM_REQUEST_INFO, * PLAM_REQUEST_INFO;

/**STRUCT-********************************************************************/

#endif
