
/**MOD+***********************************************************************/
/* Module:    lamthrd.c                                                      */
/*                                                                           */
/* Purpose:   LDAP Access Module Thread routines                             */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.20               $ $Modtime::   Jan 11 2001 15:56:38   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <lamincl.h>

/**PROC+**********************************************************************/
/* Name:      lam_thread_proc                                                */
/*                                                                           */
/* Purpose:   LAM thread procedure                                           */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     thread_info  - pointer to thread info (must be cast)    */
/*                                                                           */
/**PROC-**********************************************************************/

void *lam_thread_proc(void *thread_info)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  PLAM_THREAD_DATA td = (PLAM_THREAD_DATA)thread_info;
  PLAM_REQUEST_INFO request;
  LDAPMessage *ld_msg;
  int ld_corr;
  fd_set readfds;
  int rc;
  int num_fds;
  int bound;
  int first_bind;

  TRC_THREAD_ENTRY("lam_thread_proc");

  /***************************************************************************/
  /* Initialization: set up the thread specific data with the LAM thread     */
  /* identifier.  This is used to prevent calls to OSP term on this thread   */
  /* (which would result in a hang)                                          */
  /***************************************************************************/
  if (pthread_setspecific(lam_thread_key, lam_key_name) != 0)
  {
    TRC_ERROR((TB, "Failed to set thread specific data: quitting"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get the connection mutex, then loop doing thread's real work            */
  /***************************************************************************/
  TRC_DETAIL((TB, "Getting connection_mutex at %p for conn %p",
                                       &td->conn->connection_mutex, td->conn));
  OSP_GET_LOCK(&td->conn->connection_mutex);

  TRC_NORMAL((TB, "Entering work loop"));
  bound = FALSE;
  first_bind = TRUE;
  while (!td->conn->shutting_down)
  {
    /*************************************************************************/
    /* Get a request from the queue                                          */
    /*************************************************************************/
    while (LIST_IS_EMPTY(&td->conn->request_queue))
    {
      TRC_DETAIL((TB, "Waiting for condition variable %p",
                                                    &td->conn->data_on_queue));
      pthread_cond_wait(&td->conn->data_on_queue,
                        &td->conn->connection_mutex);

      TRC_DETAIL((TB, "Condition variable %p signalled",
                                                    &td->conn->data_on_queue));

      /***********************************************************************/
      /* See if we should shut down                                          */
      /***********************************************************************/
      if (td->conn->shutting_down)
      {
        TRC_NORMAL((TB, "Shutting down"));
        goto EXIT_LABEL;
      }
    }

    /*************************************************************************/
    /* Take an item off the queue and release the connection mutex           */
    /*************************************************************************/
    request = (PLAM_REQUEST_INFO)LIST_REMOVE_HEAD(&td->conn->request_queue);
    TRC_DETAIL((TB, "Remove rq %p from list %p", request,
                                                    &td->conn->request_queue));
    OSP_RELEASE_LOCK(&td->conn->connection_mutex);
    TRC_NORMAL((TB, "processing request %p", request));

    /*************************************************************************/
    /* See if we've got a connection to the Directory - this might be the    */
    /* first time through, or the directory may have previously gone down,   */
    /* so we'll need to rebind                                               */
    /*************************************************************************/
    if (!bound)
    {
      if (first_bind)
      {
        TRC_NORMAL((TB, "init ldap"));
        first_bind = FALSE;
      }
      else
      {
        TRC_ALERT((TB, "re-init ldap"));
      }

      bound = lam_bind(td);
      if (!bound)
      {
        TRC_ALERT((TB, "Bind failure"));
        lam_do_request_callback(request, NULL, NULL);
        OSP_GET_LOCK(&td->conn->connection_mutex);
        continue;
      }
    }

    /*************************************************************************/
    /* Now actually do some work!                                            */
    /*************************************************************************/
    TRC_DETAIL((TB, "Searching..."));
    TRC_DETAIL((TB, "Base    : %s", request->search.base));
    TRC_DETAIL((TB, "Filter  : %s", request->search.filter));
    TRC_DETAIL((TB, "Attrs[0]: %s", (*(request->search.attrs) == NULL)
                                  ? "NULL" : *(request->search.attrs)));
    ld_corr = ldap_search(td->ld,
                          request->search.base,
                          request->search.scope,
                          request->search.filter,
                          request->search.attrs,
                          request->search.attrs_only);
    if (ld_corr == -1)
    {
      TRC_ERROR((TB, "LDAP search/mod failed - server maybe unavailable"));
      bound = FALSE;
      ldap_unbind(td->ld);
      lam_do_request_callback(request, NULL, NULL);
      OSP_GET_LOCK(&td->conn->connection_mutex);
      continue;
    }
    else
    {
      TRC_DETAIL((TB, "Got ldap correlator %p", ld_corr));
    }

    /*************************************************************************/
    /* select on both the LDAP and cancel pipe FDs for reading               */
    /*************************************************************************/
    TRC_DETAIL((TB, "Wait for results"));
    FD_ZERO(&readfds);
    FD_SET(td->read_fd, &readfds);
    FD_SET(td->ld->ld_sb.sb_sd, &readfds);

    num_fds = OSP_MAX(td->read_fd, td->ld->ld_sb.sb_sd) + 1;
    rc = select(num_fds, &readfds, NULL, NULL, NULL);

    /*************************************************************************/
    /* See what went on                                                      */
    /*************************************************************************/
    if (rc == -1)
    {
      /***********************************************************************/
      /* Select failed                                                       */
      /***********************************************************************/
      TRC_OS_ERROR((TB, "Select failed"));
      ld_msg = NULL;
    }
    else if (rc == 0)
    {
      /***********************************************************************/
      /* Select timed out - which is odd, given we told it not to!           */
      /***********************************************************************/
      TRC_ERROR((TB, "Select timed out"));
      ld_msg = NULL;
    }
    else if (FD_ISSET(td->read_fd, &readfds))
    {
      /***********************************************************************/
      /* Cancel pipe was written to - free any partial results               */
      /***********************************************************************/
      TRC_ALERT((TB, "Operation canceled"));
      ldap_msgdelete(td->ld, ld_corr);
      ld_msg = NULL;
    }
    else
    {
      /***********************************************************************/
      /* Got some data from the ldap server                                  */
      /***********************************************************************/
      TRC_DETAIL((TB, "Reading results..."));
      rc = ldap_result(td->ld, ld_corr, TRUE, NULL, &ld_msg);
      if (rc == -1)
      {
        /*********************************************************************/
        /* In this case, we try an immediate re-bind in case the server has  */
        /* been lost and then come back since we were last called            */
        /*********************************************************************/
        TRC_ERROR((TB, "ldap_result failed - server unavailable?"));
        ldap_unbind(td->ld);
        bound = lam_bind(td);
        if (bound)
        {
          /*******************************************************************/
          /* Success - resubmit the search at the head of the list           */
          /*******************************************************************/
          TRC_DETAIL((TB, "Rebound successfully - try again"));
          LIST_INSERT_HEAD(&td->conn->request_queue, &request->links);
        }
        else
        {
          /*******************************************************************/
          /* Give up                                                         */
          /*******************************************************************/
          TRC_ALERT((TB, "Bind failure"));
          lam_do_request_callback(request, NULL, NULL);
        }
        OSP_GET_LOCK(&td->conn->connection_mutex);
        continue;
      }
      TRC_DETAIL((TB, "Got result %x", rc));
    }

    /*************************************************************************/
    /* Tell the caller what went on                                          */
    /*************************************************************************/
    lam_do_request_callback(request, td->ld, &ld_msg);

    /*************************************************************************/
    /* and off we go to get another request...                               */
    /*************************************************************************/
    OSP_GET_LOCK(&td->conn->connection_mutex);

  } /* while not shutting down                                               */

EXIT_LABEL:
  /***************************************************************************/
  /* tidy up                                                                 */
  /***************************************************************************/
  TRC_NORMAL((TB, "Shutting down connection %p", td->conn));
  td->conn->thread_count--;
  TRC_DETAIL((TB, "Conn now has %d active threads", td->conn->thread_count));

  /***************************************************************************/
  /* If we're the last thread out, make sure the request queue is empty or   */
  /* any requests will be left in limbo                                      */
  /***************************************************************************/
  if (td->conn->thread_count == 0)
  {
    TRC_DETAIL((TB, "last thread checking request queue"));
    while (!LIST_IS_EMPTY(&td->conn->request_queue))
    {
      request = (PLAM_REQUEST_INFO)LIST_REMOVE_HEAD(&td->conn->request_queue);
      TRC_DETAIL((TB, "Removed request %p from list %p",
                                           request, &td->conn->request_queue));
      lam_do_request_callback(request, NULL, NULL);
    }
  }

  if (bound)
  {
    TRC_DETAIL((TB, "Unbinding"));
    ldap_unbind(td->ld);
  }
  close(td->read_fd);
  OSP_RELEASE_LOCK(&td->conn->connection_mutex);
  TRC_DETAIL((TB, "Free td %p", td));
  rm_release_mem(td);

  /***************************************************************************/
  /* Clean up OpenSSL per-thread state, in case the callback function used   */
  /* OpenSSL.  If it didn't, this call is harmless.                          */
  /***************************************************************************/
  ERR_remove_state(0);

  /***************************************************************************/
  /* all done!                                                               */
  /***************************************************************************/
  rm_notify_thread_exit(RM_LAM_THREAD);

  TRC_THREAD_EXIT();

  pthread_exit(NULL);

  return(NULL);

} /* lam_thread_proc */

/**PROC+**********************************************************************/
/* Name:      lam_do_request_callback                                        */
/*                                                                           */
/* Purpose:   passes the result to the the search callback                   */
/*            and then frees all the memory associated with the request      */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     request      - pointer to the request to processes      */
/*            IN     ld           - ldap handle to use for processing        */
/*            IN     result       - LDAP result                              */
/*                                                                           */
/**PROC-**********************************************************************/

void lam_do_request_callback(PLAM_REQUEST_INFO request,
                             LDAP *ld,
                             LDAPMessage **result)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("lam_do_request_callback");

  /***************************************************************************/
  /* Validate the parameters                                                 */
  /***************************************************************************/
  TRC_ASSERT((request != NULL), (TB, "Called with null request"));
  TRC_ASSERT((request->callback != NULL), (TB, "Called with null callback"));

  /***************************************************************************/
  /* make the callback                                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Making callback..."));
  request->callback(request->user_data, ld, result);

  /***************************************************************************/
  /* Free up any result                                                      */
  /***************************************************************************/
  if ((result != NULL) && (*result != NULL))
  {
    TRC_DETAIL((TB, "Free result %p", *result));
    ldap_msgfree(*result);
  }

  /***************************************************************************/
  /* Free the request's parameters, and then the request itself              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Freeing request %p", request));
  rm_release_mem(request->search.base);
  rm_release_mem(request->search.filter);
  rm_release_mem(request->search.attrs);
  rm_release_mem(request);

  TRC_EXIT();

  return;

} /* lam_do_request_callback */

/**PROC+**********************************************************************/
/* Name:      lam_bind                                                       */
/*                                                                           */
/* Purpose:   init ldap connection and bind to server                        */
/*                                                                           */
/* Returns:   TRUE - bound successfully                                      */
/*            FALSE - a failure occurred                                     */
/*                                                                           */
/* Params:    IN     td           - thread data                              */
/*                                                                           */
/**PROC-**********************************************************************/

int lam_bind(PLAM_THREAD_DATA td)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int bound = FALSE;
  int rc;

  TRC_ENTRY("lam_bind");

  /***************************************************************************/
  /* Initialize the ldap handle                                              */
  /***************************************************************************/
  td->ld = ldap_init(td->conn->host, td->conn->port);
  if (td->ld)
  {
    /*************************************************************************/
    /* Now bind to the server                                                */
    /*************************************************************************/
    TRC_NORMAL((TB, "Got ldap handle %p; binding...", td->ld));
    rc = ldap_bind_s(td->ld, td->conn->who, td->conn->cred, LDAP_AUTH_SIMPLE);
    if (rc == LDAP_SUCCESS)
    {
      TRC_DETAIL((TB, "Bound successfully"));
      bound = TRUE;
    }
    else
    {
      TRC_ERROR((TB, "ldap (re)bind failed, rc = %d", rc));
    }
  }
  else
  {
    TRC_OS_ERROR((TB, "ldap init failed"));
  }

  TRC_EXIT();

  return(bound);

} /* lam_bind */
