
/**MOD+***********************************************************************/
/* Module:    lamutil.c                                                      */
/*                                                                           */
/* Purpose:   LDAP Access Module internal utility routines                   */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.16               $ $Modtime::   Jan 11 2001 15:56:38   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <lamincl.h>

/**PROC+**********************************************************************/
/* Name:      lam_create_host_connection                                     */
/*                                                                           */
/* Purpose:   Creates a new lam connection to the specified host             */
/*                                                                           */
/* Returns:   OSP_SUCCESS       - the connection was created successfully    */
/*            OSP_ERROR         - Another error occured                      */
/*                                                                           */
/* Params:    IN     host         - host name to connect to                  */
/*            IN     port         - port number to connect to                */
/*            IN     who,         - user dn and password information used to */
/*                   cred           access the directory on host             */
/*            OUT    new_conn     - pointer to the new connection            */
/*                                                                           */
/* Operation: Allocate and initialize the host connection structure          */
/*            Create all the lam threads for the connection                  */
/*            Create the pipe used to communicate with the threads           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC lam_create_host_connection(char *host,
                                  int port,
                                  char *who,
                                  char *cred,
                                  LAM_HOST_CONN **new_conn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  LAM_HOST_CONN *host_conn;
  int ii;

  TRC_ENTRY("lam_create_host_connection");

  /***************************************************************************/
  /* Get some space for the connection                                       */
  /***************************************************************************/
  TRC_NORMAL((TB, "Creating new host '%s:%d'", host, port));
  host_conn = (PLAM_HOST_CONN)rm_get_mem(sizeof(LAM_HOST_CONN));
  if (host_conn == NULL)
  {
    TRC_ERROR((TB, "Failed to alloc mem for host conn"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Allocd host conn %p", host_conn));

  /***************************************************************************/
  /* Initialize the connection structure                                     */
  /***************************************************************************/
  LIST_INIT(&host_conn->links);
  host_conn->open_count = 0;
  host_conn->host = NULL;
  host_conn->port = 0;
  host_conn->who = NULL;
  host_conn->cred = NULL;
  memset(&host_conn->connection_mutex, 0, sizeof(host_conn->connection_mutex));
  host_conn->thread_count = 0;
  host_conn->shutting_down = FALSE;
  LIST_INIT(&host_conn->request_queue);
  memset(&host_conn->data_on_queue, 0, sizeof(host_conn->data_on_queue));
  LIST_INIT(&host_conn->cancel_fds);

  /***************************************************************************/
  /* Store the host name and user dn/credentials                             */
  /***************************************************************************/
  OSP_STRDUP(host_conn->host, host);
  if (host_conn->host == NULL)
  {
    TRC_ERROR((TB, "Failed to save host name"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  host_conn->port = port;

  OSP_STRDUP(host_conn->who, who);
  if (host_conn->who == NULL)
  {
    TRC_ERROR((TB, "Failed to save user dn"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  OSP_STRDUP(host_conn->cred, cred);
  if (host_conn->cred == NULL)
  {
    TRC_ERROR((TB, "Failed to save user credentials"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Got host name %p (%s)", host_conn->host, host_conn->host));
  TRC_DETAIL((TB, "    user dn %p (%s)", host_conn->who, host_conn->who));
  TRC_DETAIL((TB, "    user credentials %p (%s)",
                                            host_conn->cred, host_conn->cred));

  /***************************************************************************/
  /* Set up the connection mutex                                             */
  /***************************************************************************/
  if (osp_create_mutex(&host_conn->connection_mutex) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create conn mutex"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Created conn mutex at %p", &host_conn->connection_mutex));
  OSP_GET_LOCK(&host_conn->connection_mutex);

  if (pthread_cond_init(&host_conn->data_on_queue, NULL))
  {
    TRC_ERROR((TB, "Failed to create data on q condition"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now create the threads for the connection.  Try to create all of them,  */
  /* then check that we got at least one                                     */
  /***************************************************************************/
  for (ii = 0; ii < LAM_THREADS_PER_CONNECTION; ii++)
  {
    rc = lam_create_thread(host_conn);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Create lam conn thread %d failed, rc %d", ii, rc));
    }
  }

  /***************************************************************************/
  /* As long as we created some threads, we're OK to carry on                */
  /***************************************************************************/
  if (host_conn->thread_count > 0)
  {
    TRC_NORMAL((TB, "Have %d threads of %d - carry on",
                                                 host_conn->thread_count, ii));
    rc = OSP_SUCCESS;
  }
  else
  {
    TRC_ALERT((TB, "Failed to create any lam conn threads"));
    rc = OSP_ERROR;
  }

  /***************************************************************************/
  /* Free the connection's mutex                                             */
  /***************************************************************************/
  OSP_RELEASE_LOCK(&host_conn->connection_mutex);

EXIT_LABEL:

  if (rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* On success, we pass back the new connection structure                 */
    /*************************************************************************/
    TRC_DETAIL((TB, "Returning host conn %p", host_conn));
    *new_conn = host_conn;
  }
  else
  {
    /*************************************************************************/
    /* We failed.  Make sure we return NULL for the new connection           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Failed"));
    *new_conn = NULL;

    /*************************************************************************/
    /* Free up any resources we allocated                                    */
    /*************************************************************************/
    if (host_conn)
    {
      /***********************************************************************/
      /* free up allocated fields in host conn                               */
      /***********************************************************************/
      TRC_DETAIL((TB, "Freeing host conn resources"));
      rm_release_mem(host_conn->host);
      rm_release_mem(host_conn->who);
      rm_release_mem(host_conn->cred);

      /***********************************************************************/
      /* Destroy the condition attribute and mutex                           */
      /***********************************************************************/
      pthread_cond_destroy(&host_conn->data_on_queue);
      OSP_RELEASE_LOCK(&host_conn->connection_mutex);
      osp_destroy_mutex(&host_conn->connection_mutex);
    }
  }

  TRC_EXIT();

  return(rc);

} /* lam_create_host_connection */

/**PROC+**********************************************************************/
/* Name:      lam_create_thread                                              */
/*                                                                           */
/* Purpose:   Create a lam thread for the supplied host connection           */
/*                                                                           */
/* Returns:   OSP_SUCCESS       - the thread was created successfully        */
/*            OSP_ERROR         - another error occured                      */
/*                                                                           */
/* Params:    IN     host_conn    - host connection for the thread           */
/*                                                                           */
/* Operation: Get create approval from RM                                    */
/*            Allocate and setup thread control structures                   */
/*            Update the host connection data with the new thread info       */
/*                                                                           */
/*            NB: the host connection mutex must be held by the caller of    */
/*                this function                                              */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC lam_create_thread(LAM_HOST_CONN *host_conn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  LAM_THREAD_DATA *td;
  CANCEL_FD *cfd;
  int fds[2];
  pthread_t thread_handle;
  int pipe_created;

  TRC_ENTRY("lam_create_thread");

  /***************************************************************************/
  /* Init locals                                                             */
  /***************************************************************************/
  td = NULL;
  cfd = NULL;
  pipe_created = FALSE;

  /***************************************************************************/
  /* Check with RM that it's OK to create the thread                         */
  /***************************************************************************/
  rc = rm_request_thread_create(RM_LAM_THREAD);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Create thread permission refused"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get buffers for the thread data and cancel FD                           */
  /***************************************************************************/
  td = (PLAM_THREAD_DATA)rm_get_mem(sizeof(LAM_THREAD_DATA));
  if (td == NULL)
  {
    TRC_ERROR((TB, "Failed to alloc thread data"));
    goto EXIT_LABEL;
  }

  cfd = (CANCEL_FD *)rm_get_mem(sizeof(CANCEL_FD));
  if (cfd == NULL)
  {
    TRC_ERROR((TB, "Failed to alloc cancel FD"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Allocd thread data %p", td));
  TRC_DETAIL((TB, "Allocd cancel fd %p", cfd));

  /***************************************************************************/
  /* Create the cancel pipe                                                  */
  /***************************************************************************/
  if (pipe(fds))
  {
    TRC_OS_ERROR((TB, "Failed to create cancel pipe"));
    goto EXIT_LABEL;
  }
  pipe_created = TRUE;

  /***************************************************************************/
  /* Store the fds                                                           */
  /***************************************************************************/
  cfd->fd = fds[0];
  td->read_fd = fds[1];

  /***************************************************************************/
  /* Save the connection in the thread data                                  */
  /***************************************************************************/
  td->conn = host_conn;

  /***************************************************************************/
  /* Now we can create the thread, passing it a pointer to the thread data   */
  /***************************************************************************/
  if (pthread_create(&thread_handle, NULL, lam_thread_proc, td))
  {
    TRC_OS_ERROR((TB, "Failed to create thread"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Success - increment the connection's thread count and add the cancel    */
  /* pipe to the connection's list                                           */
  /***************************************************************************/
  rc = OSP_SUCCESS;
  host_conn->thread_count++;
  TRC_DETAIL((TB, "Insert cfd %p in list %p", cfd, &host_conn->cancel_fds));
  LIST_INSERT_TAIL(&host_conn->cancel_fds, &cfd->links);
  TRC_NORMAL((TB, "created thread %d", host_conn->thread_count));

  /***************************************************************************/
  /* Since we don't care about the return code from the thread, we can       */
  /* detach it                                                               */
  /***************************************************************************/
  pthread_detach(thread_handle);

EXIT_LABEL:

  /***************************************************************************/
  /* If something went wrong, free any resources we did allocate             */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    TRC_DETAIL((TB, "Freeing allocated resource after failure"));
    rm_release_mem(td);
    rm_release_mem(cfd);
    if (pipe_created)
    {
      close(fds[0]);
      close(fds[1]);
    }
  }

  TRC_EXIT();

  return(rc);

} /* lam_create_thread */

/**PROC+**********************************************************************/
/* Name:      lam_submit_new_request                                         */
/*                                                                           */
/* Purpose:   Add a new search request to the connection's                   */
/*            request queue                                                  */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     host_conn    - the connection                           */
/*            IN     request      - the request                              */
/*                                                                           */
/* Operation: add the requeston to the connection's queue and signal the     */
/*            data on queue condition                                        */
/*                                                                           */
/**PROC-**********************************************************************/

void lam_submit_new_request(PLAM_HOST_CONN host_conn,
                            PLAM_REQUEST_INFO request)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("lam_submit_new_request");

  /***************************************************************************/
  /* Ensure we've got a valid host connection                                */
  /***************************************************************************/
  TRC_ASSERT((host_conn != NULL), (TB, "Null connection supplied"));

  /***************************************************************************/
  /* Store the connection in the request                                     */
  /***************************************************************************/
  request->lam_host_conn = host_conn;
  TRC_NORMAL((TB, "Conn %p : adding rq %p", host_conn, request));

  /***************************************************************************/
  /* Lock the list and add the request to it                                 */
  /***************************************************************************/
  OSP_GET_LOCK(&host_conn->connection_mutex);
  TRC_DETAIL((TB, "Insert rq %p in list %p", request,
                                                   &host_conn->request_queue));
  LIST_INSERT_TAIL(&host_conn->request_queue, &request->links);

  /***************************************************************************/
  /* NB: one option at this point would be to check the queue length and     */
  /* grow the number of threads servicing this connection if queue reaches   */
  /* a certain limit or 'high water mark'.  You'd also want a corresponding  */
  /* 'low water mark' which would trigger the shut down of some threads      */
  /***************************************************************************/

  /***************************************************************************/
  /* Tell the LAM threads there's something to do                            */
  /***************************************************************************/
  pthread_cond_signal(&host_conn->data_on_queue);

  OSP_RELEASE_LOCK(&host_conn->connection_mutex);

  TRC_EXIT();

  return;

} /* lam_submit_new_request */

/**PROC+**********************************************************************/
/* Name:      lam_build_search_request                                       */
/*                                                                           */
/* Purpose:   Build a search request from the supplied parameters            */
/*                                                                           */
/* Returns:   pointer to the new request, or NULL                            */
/*                                                                           */
/* Params:    IN     base,        - see the OpenLDAP documentation for       */
/*                   scope,         ldap_search() for details of these       */
/*                   filter,         parameters                              */
/*                   attrs,                                                  */
/*                   attrs_only                                              */
/*            IN     user_data    - correlator to supply on results callback */
/*            IN     result_callback                                         */
/*                                - the function to call with search results */
/*                                                                           */
/* Operation: Allocate memory for the request                                */
/*            Build the search based on the supplied parameters              */
/*                                                                           */
/**PROC-**********************************************************************/

PLAM_REQUEST_INFO lam_build_search_request(char *base,
                                           int scope,
                                           char *filter,
                                           char **attrs,
                                           int attrs_only,
                                           void *user_data,
                                           LAM_RESULT_CALLBACK result_callback)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  PLAM_REQUEST_INFO new_rq = NULL;
  PLAM_SEARCH_PARAMS search;
  int total_len;
  int attr_index;
  char *param_buffer;
  char *pinto_buffer;
  int alloc_failed;

  TRC_ENTRY("lam_build_search_request");

  /***************************************************************************/
  /* Initialize the allocation failure flag, used to tell the tidy up code   */
  /* to free any resources that were allocated                               */
  /***************************************************************************/
  alloc_failed = FALSE;

  /***************************************************************************/
  /* Get some memory for the new request                                     */
  /***************************************************************************/
  new_rq = (PLAM_REQUEST_INFO)rm_get_mem(sizeof(LAM_REQUEST_INFO));
  if (new_rq == NULL)
  {
    TRC_ERROR((TB, "Alloc failed for new request"));
    alloc_failed = TRUE;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Allocd request %p", new_rq));

  /***************************************************************************/
  /* and fill it in with the supplied params, easy ones first                */
  /***************************************************************************/
  TRC_NORMAL((TB, "Building new search request"));
  new_rq->user_data = user_data;
  new_rq->callback = result_callback;

  search = &new_rq->search;
  search->scope = scope;
  search->attrs_only = attrs_only;
  search->scope = scope;

  /***************************************************************************/
  /* Save off the string parameters                                          */
  /***************************************************************************/
  OSP_STRDUP(search->base, base);
  OSP_STRDUP(search->filter, filter);
  if ((search->base == NULL) || (search->filter == NULL))
  {
    TRC_ERROR((TB, "Alloc failed for base/filter"));
    alloc_failed = TRUE;
    goto EXIT_LABEL;
  }
  TRC_DETAIL((TB, "Allocd base name %p", search->base));
  TRC_DETAIL((TB, "Allocd filter %p", search->filter));

  /***************************************************************************/
  /* Get some memory for the attribute list - but how much?  Walk the list   */
  /* of attributes                                                           */
  /***************************************************************************/
  total_len = 0;
  attr_index = 0;
  while (attrs[attr_index] != NULL)
  {
    TRC_DETAIL((TB, "Attribute is %s", attrs[attr_index]));
    total_len += sizeof(char *);
    total_len += strlen(attrs[attr_index]) + 1;
    attr_index++;
  }

  /***************************************************************************/
  /* don't forget the NULL terminator on the attrs list                      */
  /***************************************************************************/
  total_len += sizeof(char*);
  attr_index++;

  /***************************************************************************/
  /* allocate the parameter buffer                                           */
  /***************************************************************************/
  TRC_NORMAL((TB, "Getting %d bytes for %d attrs", total_len, attr_index));
  param_buffer = (char *)rm_get_mem(total_len);
  if (param_buffer == NULL)
  {
    TRC_ERROR((TB, "Alloc failed for search attrs "));
    alloc_failed = TRUE;
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Allocd param buffer %p", param_buffer));
  pinto_buffer = param_buffer;

  /***************************************************************************/
  /* allow for the list of attributes on the front                           */
  /***************************************************************************/
  TRC_NORMAL((TB, "Copying attributes"));
  search->attrs = (char **)pinto_buffer;
  pinto_buffer += attr_index * sizeof(char *);

  /***************************************************************************/
  /* and copy the attributes                                                 */
  /***************************************************************************/
  attr_index = 0;
  while (attrs[attr_index] != NULL)
  {
    search->attrs[attr_index] = pinto_buffer;
    strcpy(pinto_buffer, attrs[attr_index]);
    pinto_buffer += strlen(attrs[attr_index]) + 1;
    attr_index++;
  }

  /***************************************************************************/
  /* don't forget to NULL-terminate the list of attributes                   */
  /***************************************************************************/
  search->attrs[attr_index] = NULL;

EXIT_LABEL:
  /***************************************************************************/
  /* If something went wrong, we need to free up any resources that were     */
  /* allocated.                                                              */
  /***************************************************************************/
  if (alloc_failed && (new_rq != NULL))
  {
    TRC_DETAIL((TB, "Freeing resources after allocation failure"));
    rm_release_mem(new_rq->search.base);
    rm_release_mem(new_rq->search.filter);
    rm_release_mem(new_rq);
    new_rq = NULL;
  }

  TRC_EXIT();

  return(new_rq);

} /* lam_build_search_request */

