
/**MOD+***********************************************************************/
/* Module:    rmmain.c                                                       */
/*                                                                           */
/* Purpose:   Resource Manager main routines                                 */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.11               $ $Modtime::   Jan 11 2001 15:56:38   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <rmincl.h>

/**PROC+**********************************************************************/
/* Name:      rm_init                                                        */
/*                                                                           */
/* Purpose:   To initialize the resource manager                             */
/*                                                                           */
/* Returns:   OSP_SUCCESS          - rm initialized successfully             */
/*            OSP_ERR_ALREADY_INIT - rm already initialized                  */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: (Debug only) initialize the counts                             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC rm_init(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("rm_init");

  /***************************************************************************/
  /* Check for re-initialization                                             */
  /***************************************************************************/
  if (rm_initialized)
  {
    TRC_ERROR((TB, "RM re-initialized"));
    rc = OSP_ERR_ALREADY_INIT;
    goto EXIT_LABEL;
  }

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* Ensure the structure and thread types will fit in the arrays we've set  */
  /* up                                                                      */
  /***************************************************************************/
  TRC_ASSERT(((RM_THREAD_INFO < 4) &&
              (RM_CONN_INFO < 4) &&
              (RM_TRANS_INFO < 4)),
             (TB, "Structure types out of count array range"));

  TRC_ASSERT(((RM_WORKER_THREAD < 4) &&
              (RM_LAM_THREAD < 4) &&
              (RM_CCM_THREAD < 4)),
             (TB, "Thread types out of count array range"));

  /***************************************************************************/
  /* Create lock used to protect counts                                      */
  /***************************************************************************/
  if (osp_create_mutex(&rm_count_mutex) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create mem count mutex"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* initialize our counts                                                   */
  /***************************************************************************/
  OSP_GET_LOCK(&rm_count_mutex);
  count_mem = 0;

  count_struct[RM_THREAD_INFO] = 0;
  count_struct[RM_CONN_INFO] = 0;
  count_struct[RM_TRANS_INFO] = 0;

  count_threads[RM_WORKER_THREAD] = 0;
  count_threads[RM_LAM_THREAD] = 0;
  count_threads[RM_CCM_THREAD] = 0;

  OSP_RELEASE_LOCK(&rm_count_mutex);
#endif

  rm_initialized = TRUE;
  TRC_NORMAL((TB, "RM Initialization complete"));

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* rm_init */

/**PROC+**********************************************************************/
/* Name:      rm_term                                                        */
/*                                                                           */
/* Purpose:   To terminate the RM component.                                 */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - RM terminated successfully                  */
/*            OSP_ERR_NOT_INIT - RM was not intialized                       */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: (Debug only) check the alloc and thread start counts are zero  */
/*                                                                           */
/**PROC-**********************************************************************/

void rm_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("rm_term");

  /***************************************************************************/
  /* Check we were initialized                                               */
  /***************************************************************************/
  if (!rm_initialized)
  {
    TRC_ERROR((TB, "RM not intialized"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* Check everything we're tracking was freed                               */
  /***************************************************************************/
  TRC_NORMAL((TB, "Allocated memory blocks: %d", count_mem));

  TRC_NORMAL((TB, "Allocated structure count:"));
  TRC_NORMAL((TB, "Type %d count %d", RM_THREAD_INFO,
                                                count_struct[RM_THREAD_INFO]));
  TRC_NORMAL((TB, "Type %d count %d", RM_CONN_INFO,
                                                  count_struct[RM_CONN_INFO]));
  TRC_NORMAL((TB, "Type %d count %d", RM_TRANS_INFO,
                                                 count_struct[RM_TRANS_INFO]));

  TRC_NORMAL((TB, "Thread count:"));
  TRC_NORMAL((TB, "Type %d count %d", RM_WORKER_THREAD,
                                             count_threads[RM_WORKER_THREAD]));
  TRC_NORMAL((TB, "Type %d count %d", RM_LAM_THREAD,
                                                count_threads[RM_LAM_THREAD]));
  TRC_NORMAL((TB, "Type %d count %d", RM_CCM_THREAD,
                                                count_threads[RM_CCM_THREAD]));

  osp_destroy_mutex(&rm_count_mutex);
#endif

  rm_initialized = FALSE;
  TRC_NORMAL((TB, "RM Termination complete"));

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* rm_term */

/*****************************************************************************/
/* Suppress tracing within rm_get_mem/rm_release_mem/rm_realloc_mem.         */
/*****************************************************************************/
#define trc_trace_types TRC_LEVEL_ALERT

/**PROC+**********************************************************************/
/* Name:      rm_get_mem                                                     */
/*                                                                           */
/* Purpose:   General purpose memory allocation.                             */
/*                                                                           */
/* Returns:   pointer to memory if successful; NULL otherwise                */
/*                                                                           */
/* Params:    IN     size         - number of bytes to allocate              */
/*                                                                           */
/* Operation: Call onto the system to get the memory                         */
/*                                                                           */
/**PROC-**********************************************************************/

void *rm_get_mem(unsigned int size)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  void *mem;

  TRC_ENTRY("rm_get_mem");

  /***************************************************************************/
  /* Validate the size                                                       */
  /***************************************************************************/
  TRC_ASSERT((size > 0), (TB, "Asked for 0 bytes"));

  /***************************************************************************/
  /* get the memory                                                          */
  /***************************************************************************/
  mem = malloc(size);
  if (mem != NULL)
  {
#ifdef OSP_DEBUG
    /*************************************************************************/
    /* track the number of allocations                                       */
    /*************************************************************************/
    OSP_GET_LOCK(&rm_count_mutex);
    count_mem++;
    TRC_DETAIL((TB, "Now %d memory blocks", count_mem));
    OSP_RELEASE_LOCK(&rm_count_mutex);
#endif

    /*************************************************************************/
    /* We no longer zero the memory.  This allows Purify to detect code that */
    /* is reading it before it has been written.                             */
    /*************************************************************************/
    TRC_DETAIL((TB, "Alloced mem at %p len %d", mem, size));
  }
  else
  {
    TRC_OS_ERROR((TB, "Failed to get mem size %d", size));
  }

  TRC_EXIT();

  return(mem);

} /* rm_get_mem */

/**PROC+**********************************************************************/
/* Name:      rm_release_mem                                                 */
/*                                                                           */
/* Purpose:   Frees memory allocated through rm_get_mem                      */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     mem          - memory to free                           */
/*                                                                           */
/* Operation: Call onto the system to free the memory                        */
/*                                                                           */
/**PROC-**********************************************************************/

void rm_release_mem(void *mem)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("rm_release_mem");

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* track the number of allocations                                         */
  /***************************************************************************/
  if (mem)
  {
    OSP_GET_LOCK(&rm_count_mutex);
    count_mem--;
    TRC_DETAIL((TB, "Now %d memory blocks", count_mem));
    OSP_RELEASE_LOCK(&rm_count_mutex);
  }
#endif

  TRC_DETAIL((TB, "Freeing mem at %p", mem));
  free(mem);

  TRC_EXIT();

  return;

} /* rm_release_mem */

/**PROC+**********************************************************************/
/* Name:      rm_realloc_mem                                                 */
/*                                                                           */
/* Purpose:   Reallocates a buffer previously allocated through rm_get_mem   */
/*                                                                           */
/* Returns:   pointer to memory if successful; NULL otherwise (old pointer   */
/*            remains valid in this case)                                    */
/*                                                                           */
/* Params:    IN     mem          - memory to reallocate                     */
/*            IN     new_size     - new size requested                       */
/*                                                                           */
/* Operation: Call onto the system to realloc the memory                     */
/*                                                                           */
/**PROC-**********************************************************************/

void *rm_realloc_mem(void *mem, unsigned int new_size)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  void *new_mem;

  TRC_ENTRY("rm_realloc_mem");

  /***************************************************************************/
  /* Check what we've been given                                             */
  /***************************************************************************/
  TRC_ASSERT((mem != NULL), (TB, "Asked to realloc null pointer"));
  TRC_ASSERT((new_size > 0), (TB, "Asked to realloc to zero size"));

  /***************************************************************************/
  /* Get the memory                                                          */
  /***************************************************************************/
  new_mem = realloc(mem, new_size);
  if (new_mem != NULL)
  {
    TRC_DETAIL((TB, "Realloced mem at %p to %p, new len %d",
                                                      mem, new_mem, new_size));
  }
  else
  {
    TRC_OS_ERROR((TB, "Realloc of %p to size %d failed", mem, new_size));
  }

  TRC_EXIT();

  return(new_mem);

} /* rm_realloc_mem */

/*****************************************************************************/
/* Restore normal tracing behaviour.                                         */
/*****************************************************************************/
#undef trc_trace_types

/**PROC+**********************************************************************/
/* Name:      rm_get_structure                                               */
/*                                                                           */
/* Purpose:   Allocate memory for stack data structures.  This allows RM to  */
/*            limit the total number of structures of each type that are in  */
/*            existence.                                                     */
/*                                                                           */
/* Returns:   pointer to memory if successful; NULL otherwise                */
/*                                                                           */
/* Params:    IN     size         - number of bytes to allocate              */
/*            IN     type         - type of structure.  Must be one of       */
/*                                  - RM_THREAD_INFO                         */
/*                                  - RM_CONN_INFO                           */
/*                                  - RM_TRANS_INFO                          */
/*                                                                           */
/* Operation: Call onto the system to get the memory                         */
/*                                                                           */
/**PROC-**********************************************************************/

void *rm_get_structure(unsigned int size, int type)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  void *struct_mem;

  TRC_ENTRY("rm_get_structure");

  /***************************************************************************/
  /* Do parameter validation                                                 */
  /***************************************************************************/
  TRC_ASSERT((size > 0), (TB, "Asked for 0 bytes"));
  TRC_ASSERT(((type == RM_THREAD_INFO) ||
              (type == RM_CONN_INFO) ||
              (type == RM_TRANS_INFO)),
             (TB, "Unknown structure type"));

  /***************************************************************************/
  /* get the memory                                                          */
  /***************************************************************************/
  struct_mem = malloc(size);
  if (struct_mem != NULL)
  {
#ifdef OSP_DEBUG
    /*************************************************************************/
    /* track the number of allocations                                       */
    /*************************************************************************/
    OSP_GET_LOCK(&rm_count_mutex);
    count_struct[type]++;
    TRC_DETAIL((TB, "Now %d structs of type %d", count_struct[type], type));
    OSP_RELEASE_LOCK(&rm_count_mutex);
#endif

    /*************************************************************************/
    /* We no longer zero the memory.  This allows Purify to detect code that */
    /* is reading it before it has been written.                             */
    /*************************************************************************/
    TRC_DETAIL((TB, "Alloced mem at %p len %d for struct type %d",
                                                      struct_mem, size, type));
  }
  else
  {
    TRC_OS_ERROR((TB, "Failed to get struct mem size %d", size));
  }

  TRC_EXIT();

  return(struct_mem);

} /* rm_get_structure */

/**PROC+**********************************************************************/
/* Name:      rm_release_structure                                           */
/*                                                                           */
/* Purpose:   Free memory allocated through rm_get_structure                 */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     mem          - memory to free                           */
/*            IN     type         - type of structure.  Must be one of       */
/*                                  - RM_THREAD_INFO                         */
/*                                  - RM_CONN_INFO                           */
/*                                  - RM_TRANS_INFO                          */
/*                                                                           */
/* Operation: Call onto the system to free the memory                        */
/*                                                                           */
/**PROC-**********************************************************************/

void rm_release_structure(void *mem, int type)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("rm_release_structure");

  /***************************************************************************/
  /* Perform validation                                                      */
  /***************************************************************************/
  TRC_ASSERT(((type == RM_THREAD_INFO) ||
              (type == RM_CONN_INFO) ||
              (type == RM_TRANS_INFO)),
             (TB, "Unknown structure type"));

  TRC_DETAIL((TB, "Freeing mem at %p for struct type %d", mem, type));

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* track the number of allocations                                         */
  /***************************************************************************/
  if (mem)
  {
    OSP_GET_LOCK(&rm_count_mutex);
    count_struct[type]--;
    TRC_DETAIL((TB, "Now %d structs of type %d", count_struct[type], type));
    OSP_RELEASE_LOCK(&rm_count_mutex);
  }
#endif
  free(mem);

  TRC_EXIT();

  return;

} /* rm_release_structure */

/**PROC+**********************************************************************/
/* Name:      rm_request_thread_create                                       */
/*                                                                           */
/* Purpose:   Request permission from RM to create a new thread.  This       */
/*            allows RM to limit the total number of threads of each type    */
/*            that are in existence.                                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS - thread may be created                            */
/*            OSP_ERROR   - permission denied                                */
/*                                                                           */
/* Params:    IN     type         - type of thread.  Must be one of          */
/*                                  - RM_WORKER_THREAD                       */
/*                                  - RM_LAM_THREAD                          */
/*                                  - RM_CCM_THREAD                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC rm_request_thread_create(int type)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("rm_request_thread_create");

  /***************************************************************************/
  /* Validate the type                                                       */
  /***************************************************************************/
  TRC_ASSERT(((type == RM_WORKER_THREAD) ||
              (type == RM_LAM_THREAD) ||
              (type == RM_CCM_THREAD)),
             (TB, "Unknown thread type"));

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* track the number of threads                                             */
  /***************************************************************************/
  OSP_GET_LOCK(&rm_count_mutex);
  count_threads[type]++;
  TRC_DETAIL((TB, "Now %d threads of type %d", count_threads[type], type));
  OSP_RELEASE_LOCK(&rm_count_mutex);
#endif

  TRC_EXIT();

  return(rc);

} /* rm_request_thread_create */

/**PROC+**********************************************************************/
/* Name:      rm_notify_thread_exit                                          */
/*                                                                           */
/* Purpose:   Notify RM that a thread is about to exit.  This allows RM      */
/*            to keep its thread counts up-to-date.                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS                                                    */
/*                                                                           */
/* Params:    IN     type         - type of thread.  Must be one of          */
/*                                  - RM_WORKER_THREAD                       */
/*                                  - RM_LAM_THREAD                          */
/*                                  - RM_CCM_THREAD                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC rm_notify_thread_exit(int type)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("rm_notify_thread_exit");

  /***************************************************************************/
  /* Validate the type                                                       */
  /***************************************************************************/
  TRC_ASSERT(((type == RM_WORKER_THREAD) ||
              (type == RM_LAM_THREAD) ||
              (type == RM_CCM_THREAD)),
             (TB, "Unknown thread type"));

#ifdef OSP_DEBUG
  /***************************************************************************/
  /* track the number of threads                                             */
  /***************************************************************************/
  OSP_GET_LOCK(&rm_count_mutex);
  count_threads[type]--;
  TRC_DETAIL((TB, "Now %d threads of type %d", count_threads[type], type));
  OSP_RELEASE_LOCK(&rm_count_mutex);
#endif

  TRC_EXIT();

  return(rc);

} /* rm_notify_thread_exit */

