
/**MOD+***********************************************************************/
/* Module:    scepattr.c                                                     */
/*                                                                           */
/* Purpose:   SCEP utility functions to handle PKCS#7 attributes             */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.6                $ $Modtime::   Jan 11 2001 15:56:38   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scepincl.h>

/**PROC+**********************************************************************/
/* Name:      scep_extract_attribs                                           */
/*                                                                           */
/* Purpose:   Extract the authenticated attributes that we need to store     */
/*            in the transaction info or use for parsing a request           */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     auth_attribs   - the stack of authenticated attribs     */
/*                                      from the PKCS#7 SignedData           */
/*            OUT    trans_id       - transactionId data                     */
/*            OUT    msg_type       - messageType data                       */
/*            OUT    s_nonce        - senderNonce data                       */
/*            OUT    s_nonce_len    - senderNonce length                     */
/*                                                                           */
/* Operation: Use helper functions to get the transaction ID, message type   */
/*            and senderNonce values.                                        */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_extract_attribs(STACK_OF(X509_ATTRIBUTE) *auth_attribs,
                            char **trans_id,
                            char **msg_type,
                            char **s_nonce,
                            unsigned int *s_nonce_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int value_len;

  TRC_ENTRY("scep_extract_attribs");

  /***************************************************************************/
  /* Get transaction ID.                                                     */
  /***************************************************************************/
  rc = scep_get_string_attrib(auth_attribs,
                              scep.nid_transId,
                              V_ASN1_PRINTABLESTRING,
                              trans_id,
                              NULL);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Error while getting transactionId attribute (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get messageType.                                                        */
  /***************************************************************************/
  rc = scep_get_string_attrib(auth_attribs,
                              scep.nid_messageType,
                              V_ASN1_PRINTABLESTRING,
                              msg_type,
                              NULL);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Error while getting messageType attribute (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get senderNonce.                                                        */
  /***************************************************************************/
  rc = scep_get_string_attrib(auth_attribs,
                              scep.nid_senderNonce,
                              V_ASN1_OCTET_STRING,
                              s_nonce,
                              s_nonce_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Error while getting senderNonce attribute (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Extracted all required attributes"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Clean up on failure.                                                  */
    /*************************************************************************/
    if (*trans_id != NULL)
    {
      rm_release_mem(*trans_id);
      *trans_id = NULL;
    }
    if (*msg_type != NULL)
    {
      rm_release_mem(*msg_type);
      *msg_type = NULL;
    }
    if (*s_nonce != NULL)
    {
      rm_release_mem(*s_nonce);
      *s_nonce = NULL;
      *s_nonce_len = 0;
    }
  }

  TRC_EXIT();

  return(rc);

} /* scep_extract_attribs */

/**PROC+**********************************************************************/
/* Name:      scep_get_string_attrib                                         */
/*                                                                           */
/* Purpose:   Extract one attribute from a stack of attributes               */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     attribs        - the stack of attributes                */
/*            IN     nid            - OpenSSL NID value of req'd attribute   */
/*            IN     type           - V_ASN1_OCTET_STRING: octet string      */
/*                                    V_ASN1_PRINTABLESTRING: printable str  */
/*            OUT    buffer         - the value of the attribute,            */
/*                                      or null if it wasn't found           */
/*            OUT    buf_len        - the length of the data in the buffer   */
/*                                                                           */
/* Operation: Use a helper function to find the attribute and then convert   */
/*            it to a byte string.                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_get_string_attrib(STACK_OF(X509_ATTRIBUTE) *attribs,
                              int nid,
                              int type,
                              char **buffer,
                              unsigned int *buf_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  ASN1_TYPE *asn1_type;
  unsigned int len;

  TRC_ENTRY("scep_get_string_attrib");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT(((type == V_ASN1_PRINTABLESTRING) ||
              (type == V_ASN1_OCTET_STRING)),
             (TB, "Unrecognized attribute type: %d", type));

  TRC_ASSERT(((type != V_ASN1_OCTET_STRING) || (buf_len != NULL)),
             (TB, "Must supply length for octet string"));

  /***************************************************************************/
  /* Get the attribute requested.                                            */
  /***************************************************************************/
  rc = scep_get_attrib(attribs,
                       nid,
                       &asn1_type);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Error %u while finding attribute '%s'",
                                                         rc, OBJ_nid2sn(nid)));
    goto EXIT_LABEL;
  }

  if (ASN1_TYPE_get(asn1_type) != type)
  {
    TRC_ERROR((TB, "'%s' was type %d; expected %d",
                             OBJ_nid2sn(nid), ASN1_TYPE_get(asn1_type), type));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  len = ASN1_STRING_length(asn1_type->value.asn1_string);

  if (type == V_ASN1_PRINTABLESTRING)
  {
    /*************************************************************************/
    /* We will add a null terminator to a PrintableString.                   */
    /*************************************************************************/
    *buffer = rm_get_mem(len + 1);
  }
  else
  {
    *buffer = rm_get_mem(len);
  }

  if (*buffer == NULL)
  {
    TRC_ERROR((TB, "Couldn't get memory for '%s'", OBJ_nid2sn(nid)));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Copy the data.                                                          */
  /***************************************************************************/
  memcpy(*buffer, ASN1_STRING_data(asn1_type->value.asn1_string), len);

  /***************************************************************************/
  /* Add that null terminator if it's a PrintableString.                     */
  /***************************************************************************/
  if (type == V_ASN1_PRINTABLESTRING)
  {
    (*buffer)[len] = 0;
    len++;
  }

  /***************************************************************************/
  /* Pass back the buffer length if the caller wants it.                     */
  /***************************************************************************/
  if (buf_len != NULL)
  {
    *buf_len = len;
  }

  TRC_DETAIL((TB, "Got attribute '%s'", OBJ_nid2sn(nid)));
  TRC_DATA("Attribute data:", *buffer, len);
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    rm_release_mem(*buffer);
    *buffer = NULL;
    if (buf_len != NULL)
    {
      *buf_len = 0;
    }
  }

  TRC_EXIT();

  return(rc);

} /* scep_get_string_attrib */

/**PROC+**********************************************************************/
/* Name:      scep_get_attrib                                                */
/*                                                                           */
/* Purpose:   Extract the value of a particular attribute from a stack of    */
/*            attributes                                                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     attribs        - the stack of attributes                */
/*            IN     required_nid   - the required OpenSSL numerical ID      */
/*            OUT    asn1_type      - the resulting ASN1_TYPE, or NULL if    */
/*                                      the attribute wasn't found           */
/*                                                                           */
/* Operation: Search the stack of attributes for a match. Return error if    */
/*            more than one match is found.                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_get_attrib(STACK_OF(X509_ATTRIBUTE) *attribs,
                       int required_nid,
                       ASN1_TYPE **asn1_type)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  ASN1_OBJECT *asn1_obj = NULL;
  X509_ATTRIBUTE *x509_attrib = NULL;
  int num_attribs;
  int ii;

  TRC_ENTRY("scep_get_attrib");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((attribs != NULL), (TB, "Null attributes"));
  TRC_ASSERT((asn1_type != NULL), (TB, "Null return pointer"));

  /***************************************************************************/
  /* Initialize the return parameter in case we don't find anything.         */
  /***************************************************************************/
  *asn1_type = NULL;

  /***************************************************************************/
  /* Convert the supplied NID to an object.                                  */
  /***************************************************************************/
  asn1_obj = OBJ_nid2obj(required_nid);
  if (asn1_obj == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to convert nid to object"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Loop through the stack of attribs, looking for our object type.  We     */
  /* don't jump out on the first match so that we can be sure there aren't   */
  /* any duplicates.                                                         */
  /***************************************************************************/
  num_attribs = sk_X509_ATTRIBUTE_num(attribs);
  for (ii = 0; ii < num_attribs; ii++)
  {
    x509_attrib = sk_X509_ATTRIBUTE_value(attribs, ii);
    if (OBJ_cmp(x509_attrib->object, asn1_obj) == 0)
    {
      /***********************************************************************/
      /* Check that there's a value.  We expect attrib.set to be non-zero,   */
      /* as does PKCS7_get_signed_attribute().                               */
      /***********************************************************************/
      if (x509_attrib->set &&
          (sk_ASN1_TYPE_num(x509_attrib->value.set) != 0))
      {
        if (*asn1_type != NULL)
        {
          /*******************************************************************/
          /* We found a match on a previous iteration and now we've found    */
          /* another one.  This ultimately means a bad request, but we leave */
          /* asn1_type as it is in case the caller wants it.                 */
          /*******************************************************************/
          TRC_ERROR((TB, "Found duplicate attributes (nid=%d)", required_nid));
          rc = OSP_ERR_BAD_REQUEST;
          goto EXIT_LABEL;
        }

        *asn1_type = sk_ASN1_TYPE_value(x509_attrib->value.set, 0);
        TRC_DETAIL((TB, "Got value (at %p) for nid %d",
                                                    *asn1_type, required_nid));
      }
    }
  }

  /***************************************************************************/
  /* Check we found something                                                */
  /***************************************************************************/
  if (*asn1_type == NULL)
  {
    TRC_ALERT((TB, "No attribute found with nid %d", required_nid));
    rc = OSP_ERR_MISSING_ATTR;
  }
  else
  {
    rc = OSP_SUCCESS;
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* scep_get_attrib */

/**PROC+**********************************************************************/
/* Name:      scep_make_common_attribs                                       */
/*                                                                           */
/* Purpose:   Construct a stack of attributes that are common to success     */
/*            and failure responses                                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     state          - the request state information          */
/*            OUT    auth_attribs   - the stack of attributes                */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_make_common_attribs(SCEP_STATE *state,
                                const char *msg_type,
                                const char *pki_status,
                                STACK_OF(X509_ATTRIBUTE) **auth_attribs)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned char sender_n_once[SCEP_N_ONCE_LEN];

  TRC_ENTRY("scep_make_common_attribs");

  /***************************************************************************/
  /* Generate a random senderNonce value.                                    */
  /***************************************************************************/
  ccm_random(sender_n_once, SCEP_N_ONCE_LEN);

  /***************************************************************************/
  /* Create the stack.                                                       */
  /***************************************************************************/
  *auth_attribs = sk_X509_ATTRIBUTE_new(NULL);
  if (*auth_attribs == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create stack for attributes"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the attributes.  Note that contentType (='data') and messageDigest  */
  /* are added automatically during signing.                                 */
  /*                                                                         */
  /* First up is the transactionId, which we saved from the request...       */
  /***************************************************************************/
  rc = scep_add_string_attrib(*auth_attribs,
                              scep.nid_transId,
                              V_ASN1_PRINTABLESTRING,
                              state->transaction_id,
                              0);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add transactionId to attributes (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* ...the messageType, which the caller supplies...                        */
  /***************************************************************************/
  rc = scep_add_string_attrib(*auth_attribs,
                              scep.nid_messageType,
                              V_ASN1_PRINTABLESTRING,
                              msg_type,
                              0);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add messageType to attributes (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* ...the pkiStatus, which the caller supplies...                          */
  /***************************************************************************/
  rc = scep_add_string_attrib(*auth_attribs,
                              scep.nid_pkiStatus,
                              V_ASN1_PRINTABLESTRING,
                              pki_status,
                              0);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add pkiStatus to attributes (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* ...the senderNonce, which we generated above...                         */
  /***************************************************************************/
  rc = scep_add_string_attrib(*auth_attribs,
                              scep.nid_senderNonce,
                              V_ASN1_OCTET_STRING,
                              (char *)sender_n_once,
                              SCEP_N_ONCE_LEN);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add senderNonce to attributes (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* ...and the recipientNonce, which is the same as the senderNonce value   */
  /* that we saved from the request.                                         */
  /***************************************************************************/
  rc = scep_add_string_attrib(*auth_attribs,
                              scep.nid_recipientNonce,
                              V_ASN1_OCTET_STRING,
                              state->n_once,
                              state->n_once_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add recipientNonce to attributes (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Made stack of common attributes"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    if (*auth_attribs != NULL)
    {
      sk_X509_ATTRIBUTE_pop_free(*auth_attribs, X509_ATTRIBUTE_free);
      *auth_attribs = NULL;
    }
  }

  TRC_EXIT();

  return(rc);

} /* scep_make_common_attribs */

/**PROC+**********************************************************************/
/* Name:      scep_add_string_attrib                                         */
/*                                                                           */
/* Purpose:   Add a string attribute to a stack of attributes                */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     attribs        - the stack of attributes                */
/*            IN     nid            - OpenSSL NID value of req'd attribute   */
/*            IN     type           - V_ASN1_OCTET_STRING: octet string      */
/*                                    V_ASN1_PRINTABLESTRING: printable str  */
/*            IN     buffer         - the value for the attribute            */
/*            IN     buf_len        - length of buffer (OctetString only)    */
/*                                                                           */
/* Operation: Convert the data pointed to by buffer into the object type     */
/*            and add it to the stack supplied. The buf_len parameter is     */
/*            used only for OctetString; PrintableString is assumed to be    */
/*            null-terminated (but the null is not copied).                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_add_string_attrib(STACK_OF(X509_ATTRIBUTE) *attribs,
                              int nid,
                              int type,
                              const char *buffer,
                              unsigned int buf_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  ASN1_STRING *asn1_string = NULL;
  int result;

  TRC_ENTRY("scep_add_string_attrib");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT(((type == V_ASN1_PRINTABLESTRING) ||
              (type == V_ASN1_OCTET_STRING)),
             (TB, "Unrecognized string type: %d", type));

  /***************************************************************************/
  /* See what we've got and set up buffer length if it's a PrintableString.  */
  /***************************************************************************/
  if (type == V_ASN1_OCTET_STRING)
  {
    TRC_DETAIL((TB, "Adding OctetString attribute: nid=%d (%s)",
                                                        nid, OBJ_nid2sn(nid)));
    TRC_DATA("OctetString data:", buffer, buf_len);
  }
  else
  {
    TRC_DETAIL((TB, "Adding PrintableString attribute: nid=%d (%s) value='%s'",
                                                nid, OBJ_nid2sn(nid), buffer));
    buf_len = strlen(buffer);
  }

  /***************************************************************************/
  /* Create the data structure we need.                                      */
  /***************************************************************************/
  asn1_string = ASN1_STRING_new();
  if (asn1_string == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create ASN.1 string structure"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  result = ASN1_STRING_set(asn1_string, buffer, buf_len);
  if (result <= 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to put data in ASN.1 string (%d)", result));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now add the ASN.1 structure to the stack.                               */
  /***************************************************************************/
  rc = scep_add_attrib(attribs,
                       nid,
                       type,
                       asn1_string);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add string attribute (%u): nid=%d (%s), type=%d",
                                              rc, nid, OBJ_nid2sn(nid), type));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Added string attribute OK"));

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    ASN1_STRING_free(asn1_string);
  }

  TRC_EXIT();

  return(rc);

} /* scep_add_string_attrib */

/**PROC+**********************************************************************/
/* Name:      scep_add_attrib                                                */
/*                                                                           */
/* Purpose:   Add an attribute to a stack of attributes                      */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     attribs        - the stack of attributes                */
/*            IN     nid            - the required OpenSSL numerical ID      */
/*            IN     type           - type of attribute                      */
/*            IN     value          - OpenSSL ASN1_xxx structure             */
/*                                                                           */
/* Operation: Create the attribute structure and push it onto the stack.     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_add_attrib(STACK_OF(X509_ATTRIBUTE) *attribs,
                       int nid,
                       int type,
                       void *value)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  X509_ATTRIBUTE *x509_attrib = NULL;
  int result;

  TRC_ENTRY("scep_add_attrib");

  /***************************************************************************/
  /* Note that we don't check for existing entries of this type; we just     */
  /* add the attribute to the stack.                                         */
  /***************************************************************************/
  x509_attrib = X509_ATTRIBUTE_create(nid, type, value);
  if (x509_attrib == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create attribute: nid=%d (%s), type=%d",
                                                  nid, OBJ_nid2sn(nid), type));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  result = sk_X509_ATTRIBUTE_push(attribs, x509_attrib);
  if (result <= 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to push attrib (%d): nid=%d (%s), type=%d",
                                          result, nid, OBJ_nid2sn(nid), type));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Added attribute to stack"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if ((rc != OSP_SUCCESS) && (x509_attrib != NULL))
  {
    X509_ATTRIBUTE_free(x509_attrib);
  }

  TRC_EXIT();

  return(rc);

} /* scep_add_attrib */

