
/**MOD+***********************************************************************/
/* Module:    scepmain.c                                                     */
/*                                                                           */
/* Purpose:   SCEP initialization and termination functions                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.9                $ $Modtime::   Jan 11 2001 15:56:40   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#define SCEP_DEFINE_VARS
#include <scepincl.h>
#undef SCEP_DEFINE_VARS

/**PROC+**********************************************************************/
/* Name:      scep_init                                                      */
/*                                                                           */
/* Purpose:   Initialize the SCEP layer                                      */
/*                                                                           */
/* Returns:   OSP_SUCCESS - completed successfully                           */
/*            OSP_ERR_CONFIG                                                 */
/*                        - SCEP initialization failed because of incorrect  */
/*                          configuration data                               */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     config         - configuration data handle              */
/*                                                                           */
/* Operation: Allocate and initialize SCEP internal data, including:         */
/*                                                                           */
/*            - allocate required OIDs that aren't built into OpenSSL        */
/*            - get the CA cert locally for quick retrieval                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_init(LHASH *config)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERR_NO_RESOURCES;
  OSP_RC int_rc = OSP_SUCCESS;
  char *value;

  TRC_ENTRY("scep_init");

  /***************************************************************************/
  /* Check the parameter.                                                    */
  /***************************************************************************/
  TRC_ASSERT((config != NULL),
             (TB, "no configuration data provided"));

  /***************************************************************************/
  /* Initialize the global data structure.                                   */
  /***************************************************************************/
  scep.ca_name = NULL;
  scep.ca_cert = NULL;
  scep.ca_cert_len = 0;

  /***************************************************************************/
  /* OpenSSL should be initialized by now -- we need to add some object IDs. */
  /* See SCEP spec appendix C for details of these.                          */
  /***************************************************************************/
  TRC_NORMAL((TB, "Creating object IDs"));
  scep.nid_messageType = OBJ_create("2.16.840.1.113733.1.9.2",
                                    "messageType",
                                    "messagetype");
  if (scep.nid_messageType == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create object 'messageType'"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  scep.nid_pkiStatus = OBJ_create("2.16.840.1.113733.1.9.3",
                                  "pkiStatus",
                                  "pkistatus");
  if (scep.nid_pkiStatus == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create object 'pkiStatus'"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  scep.nid_failInfo = OBJ_create("2.16.840.1.113733.1.9.4",
                                 "failInfo",
                                 "failinfo");
  if (scep.nid_failInfo == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create object 'failInfo'"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  scep.nid_senderNonce = OBJ_create("2.16.840.1.113733.1.9.5",
                                    "senderNonce",
                                    "sendernonce");
  if (scep.nid_senderNonce == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create object 'senderNonce'"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  scep.nid_recipientNonce = OBJ_create("2.16.840.1.113733.1.9.6",
                                       "recipientNonce",
                                       "recipientnonce");
  if (scep.nid_recipientNonce == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create object 'recipientNonce'"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  scep.nid_transId = OBJ_create("2.16.840.1.113733.1.9.7",
                                "transId",
                                "transid");
  if (scep.nid_transId == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create object 'transId'"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  scep.nid_extensionReq = OBJ_create("2.16.840.1.113733.1.9.8",
                                     "extensionReq",
                                     "extensionreq");
  if (scep.nid_extensionReq == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create object 'extensionReq'"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read the CA name from configuration.                                    */
  /***************************************************************************/
  value = CONF_get_string(config,
                          OSP_CONFIG_STACK_GROUP,
                          (char *)scep_config_ca_name);
  if ((value == NULL) || (value[0] == 0))
  {
    OSP_OPENSSL_ERROR((TB, "No SCEP CA name configured"));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Copy the CA name that was retrieved, including the terminating null.    */
  /***************************************************************************/
  TRC_NORMAL((TB, "Storing CA name %s", value));
  OSP_STRDUP(scep.ca_name, value);
  if (scep.ca_name == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for SCEP CA name"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Ask CCM for the CA certificate.                                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Getting CA certificate from CCM"));
  rc = ccm_get_cert(scep.ca_name,
                    &scep.ca_cert,
                    &scep.ca_cert_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get CA certificate (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "SCEP initialized OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up if we failed.                                                  */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Terminating because of error during init"));
    int_rc = scep_term();
    if (int_rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to terminate on error"));
    }
  }

  TRC_EXIT();

  return(rc);

} /* scep_init */

/**PROC+**********************************************************************/
/* Name:      scep_term                                                      */
/*                                                                           */
/* Purpose:   Terminate the SCEP layer                                       */
/*                                                                           */
/* Returns:   OSP_SUCCESS - completed successfully                           */
/*            This function always returns success.                          */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Free all the SCEP internal data.                               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_term(void)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;

  TRC_ENTRY("scep_term");

  /***************************************************************************/
  /* Free the CA name string.                                                */
  /***************************************************************************/
  if (scep.ca_name != NULL)
  {
    rm_release_mem((void *)(scep.ca_name));
  }

  /***************************************************************************/
  /* Free the CA certificate buffer.                                         */
  /***************************************************************************/
  if (scep.ca_cert != NULL)
  {
    rm_release_mem((void *)(scep.ca_cert));
  }

  /***************************************************************************/
  /* Free up all the OpenSSL objects we created                              */
  /***************************************************************************/
  OBJ_cleanup();

  TRC_NORMAL((TB, "SCEP terminated OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* scep_term */

