
/**MOD+***********************************************************************/
/* Module:    scepreq.c                                                      */
/*                                                                           */
/* Purpose:   SCEP request processing functions                              */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.15               $ $Modtime::   Jan 11 2001 15:56:40   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scepincl.h>

/**PROC+**********************************************************************/
/* Name:      scep_request_received                                          */
/*                                                                           */
/* Purpose:   Process a received SCEP request                                */
/*                                                                           */
/* Returns:   OSP_SUCCESS - completed successfully                           */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     trans          - the transaction's control block        */
/*            IN     operation      - pointer to first char after the URI    */
/*                                                                           */
/* Operation: Parse the operation type (GetCACert or PKIOperation) and pass  */
/*            it to the appropriate handler.                                 */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_request_received(TRANS_INFO *trans,
                             char *operation)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  SCEP_STATE *state = NULL;
  unsigned int op_type;
  unsigned int msg_len;

  TRC_ENTRY("scep_request_received");

  /***************************************************************************/
  /* Trace the transaction pointer.                                          */
  /***************************************************************************/
  TRC_NORMAL((TB, "Processing transaction at %p", trans));

  /***************************************************************************/
  /* First, allocate space for some state info.  This will contain all the   */
  /* information relating to this request, including a pointer to the        */
  /* associated TRANS_INFO structure.                                        */
  /***************************************************************************/
  state = (SCEP_STATE *)rm_get_mem(sizeof(SCEP_STATE));
  if (state == NULL)
  {
    TRC_ERROR((TB, "Failed to alloc memory for SCEP state"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize the state structure.                                         */
  /***************************************************************************/
  state->trans = trans;
  state->transaction_id = NULL;
  state->message_type = NULL;
  state->n_once = NULL;
  state->n_once_len = 0;
  state->ss_cert = NULL;
  state->ss_cert_len = 0;

  /***************************************************************************/
  /* Find out what operation type this is.  The possibilities at this stage  */
  /* are GetCACert or PKIOperation.                                          */
  /***************************************************************************/
  rc = scep_parse_op_type(&operation,
                          trans->recv_buffer + trans->recv_buffer_end,
                          &op_type,
                          &msg_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to parse operation type (%u)", rc));
    goto EXIT_LABEL;
  }

  if (op_type == SCEP_OP_GETCACERT)
  {
    /*************************************************************************/
    /* GetCACert is a special case message, which doesn't require any        */
    /* PKCS#7 unpacking, etc.  This function handles it.                     */
    /*************************************************************************/
    TRC_NORMAL((TB, "Processing GetCACert"));
    rc = scep_process_getcacert(state,
                                operation,
                                msg_len);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to process GetCACert (%u)", rc));
      goto EXIT_LABEL;
    }
  }
  else
  {
    /*************************************************************************/
    /* It's a PKIOperation message.  These are embedded in PKCS#7            */
    /* SignedData and EnvelopedData structures.  This function handles all   */
    /* that, and dispatches the contents to the appropriate handler.         */
    /*************************************************************************/
    TRC_NORMAL((TB, "Processing PKIOperation"));
    rc = scep_process_pkioperation(state,
                                   operation,
                                   msg_len);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to process PKIOperation (%u)", rc));
      goto EXIT_LABEL;
    }
  }

  TRC_NORMAL((TB, "Processed SCEP request (type %u)", op_type));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    if ((rc == OSP_ERR_BAD_REQUEST) || (rc == OSP_ERR_FAILED_VERIFY))
    {
      TRC_ALERT((TB, "Sending 'bad request' response"));
      rc = scep_send_http_failure(trans, SM_RESPONSE_400);
    }
    else
    {
      TRC_ALERT((TB, "Sending 'server error' response"));
      rc = scep_send_http_failure(trans, SM_RESPONSE_500);
    }

    if (state != NULL)
    {
      TRC_DETAIL((TB, "Freeing scep state"));
      scep_free_state(state);
    }
  }

  TRC_EXIT();

  return(rc);

} /* scep_request_received */

/**PROC+**********************************************************************/
/* Name:      scep_parse_op_type                                             */
/*                                                                           */
/* Purpose:   identifies the SCEP operation type                             */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN/OUT ptr            - IN:  pointer to 'operation=...'        */
/*                                    OUT: pointer to char after 'message='  */
/*            IN     buf_end        - pointer to first byte after buffer end */
/*            OUT    op_type        - SCEP_OP_xxx                            */
/*            OUT    msg_len        - length of text after 'message='        */
/*                                                                           */
/* Operation: identify operation type; set pointer to 1st char of message.   */
/*            [On error, op_type is 'SCEP_OP_UNKNOWN' and msg_len is zero;   */
/*            ptr is unchanged]                                              */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_parse_op_type(char **ptr,
                          char *buf_end,
                          unsigned int *op_type,
                          unsigned int *msg_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERR_BAD_REQUEST;
  char *current = *ptr;
  char *msg_end;

  TRC_ENTRY("scep_parse_op_type");

  /***************************************************************************/
  /* We assume that the caller has made sure there's a CRLF terminating the  */
  /* string pointed to by 'operation'.  The SCEP spec doesn't say whether    */
  /* or not the 'operation' and 'message' strings _must_ be lower case.  We  */
  /* are accomodating and case-insensitive.                                  */
  /***************************************************************************/

  /***************************************************************************/
  /* Parse out the operation type.                                           */
  /***************************************************************************/
  if (strncasecmp(current,
                  scep_txt_operation,
                  SCEP_TXT_OPERATION_LEN) != 0)
  {
    /*************************************************************************/
    /* Didn't see the expected 'operation=' parameter.                       */
    /*************************************************************************/
    TRC_ERROR((TB, "Couldn't find 'operation=' in request"));
    goto EXIT_LABEL;
  }
  current += SCEP_TXT_OPERATION_LEN;

  /***************************************************************************/
  /* Work out whether it's a GetCACert or a PKIOperation and set op_type     */
  /* for the caller. It's an error if we can't recognize the request.        */
  /***************************************************************************/
  if (strncasecmp(current,
                  scep_txt_getcacert,
                  SCEP_TXT_GETCACERT_LEN) == 0)
  {
    TRC_NORMAL((TB, "Got GetCACert"));
    *op_type = SCEP_OP_GETCACERT;
    current += SCEP_TXT_GETCACERT_LEN;
  }
  else if (strncasecmp(current,
                       scep_txt_pkioperation,
                       SCEP_TXT_PKIOPERATION_LEN) == 0)
  {
    TRC_NORMAL((TB, "Got PKIOperation"));
    *op_type = SCEP_OP_PKIOPERATION;
    current += SCEP_TXT_PKIOPERATION_LEN;
  }
  else
  {
    TRC_ERROR((TB, "Unrecognized SCEP operation"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Find the message: first we expect '&message='.                          */
  /***************************************************************************/
  if (strncasecmp(current,
                  scep_txt_message,
                  SCEP_TXT_MESSAGE_LEN) != 0)
  {
    TRC_ERROR((TB, "Couldn't find '&message=' in request"));
    goto EXIT_LABEL;
  }
  current += SCEP_TXT_MESSAGE_LEN;

  /***************************************************************************/
  /* 'current' now points to the start of the message.  Let's find the end,  */
  /* which is marked by a space or a CRLF sequence (just a CR will do).      */
  /***************************************************************************/
  msg_end = current;
  while ((msg_end < buf_end) &&
         (*msg_end != 0x0D) &&
         (*msg_end != 0x20))
  {
    msg_end++;
  }

  if (msg_end >= buf_end)
  {
    TRC_ERROR((TB, "Ran off end of buffer while parsing message"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the message length for the caller (msg_end points to the first      */
  /* byte after the message).                                                */
  /***************************************************************************/
  *msg_len = msg_end - current;

  TRC_NORMAL((TB, "Recognized %s OK; message length=%u, data at %p",
                       (*op_type == SCEP_OP_GETCACERT) ? scep_txt_getcacert
                                                       : scep_txt_pkioperation,
                       *msg_len, current));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Set the message pointer for the caller; other params are already set  */
    /*************************************************************************/
    *ptr = current;
  }
  else
  {
    /*************************************************************************/
    /* Failure: set parameters appropriately (ptr is left alone).            */
    /*************************************************************************/
    *op_type = SCEP_OP_UNKNOWN;
    *msg_len = 0;

    TRC_DATA("Contents of invalid message", *ptr, (buf_end - *ptr));
  }

  TRC_EXIT();

  return(rc);

} /* scep_parse_op_type */

/**PROC+**********************************************************************/
/* Name:      scep_process_getcacert                                         */
/*                                                                           */
/* Purpose:   Process a request for the CA's certificate (GetCACert)         */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     state          - the request state information          */
/*            IN     msg            - the message data                       */
/*            IN     msg_len        - length of message data                 */
/*                                                                           */
/* Operation: Just send the certificate as it's already held locally.        */
/*            If we succeed in sending the response then                     */
/*            scep_request_received() will always return success to SM.      */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_process_getcacert(SCEP_STATE *state,
                              char *msg,
                              unsigned int msg_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;

  TRC_ENTRY("scep_process_getcacert");

  /***************************************************************************/
  /* We don't currently do anything with the message data.  If we were       */
  /* emulating multiple CAs, we'd have to use the CA-IDENT info in 'msg' to  */
  /* decide which CA cert to send.                                           */
  /***************************************************************************/
  TRC_DATA("GetCACert CA-IDENT data", msg, msg_len);

  /***************************************************************************/
  /* We got the certificate itself at startup, so all we need to do is send  */
  /* it.  There's no PKCS#7 packaging for this response.  This call will     */
  /* free the state structure.                                               */
  /***************************************************************************/
  rc = scep_send_response(state,
                          SCEP_OP_GETCACERT,
                          scep.ca_cert,
                          scep.ca_cert_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to construct and send CA cert response (%u)", rc));
  }
  else
  {
    TRC_NORMAL((TB, "Sent response to GetCACert"));
  }

  TRC_EXIT();

  return(rc);

} /* scep_process_getcacert */

/**PROC+**********************************************************************/
/* Name:      scep_process_pkioperation                                      */
/*                                                                           */
/* Purpose:   processes a PKIOperation, which may be one of several types    */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     state          - the request state information          */
/*            IN     msg            - the message data                       */
/*            IN     msg_len        - length of message data                 */
/*                                                                           */
/* Operation: Consists of the following steps:                               */
/*            - verify PKCS#7 SignedData wrapper and extract attributes      */
/*            - decrypt PKCS#7 EnvelopedData wrapper to produce the message  */
/*            - switch on the message type                                   */
/*            - call the appropriate handler                                 */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_process_pkioperation(SCEP_STATE *state,
                                 char *msg,
                                 unsigned int msg_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  STACK_OF(X509_ATTRIBUTE) *auth_attribs = NULL;
  char *p7_sd_asn1 = NULL;
  unsigned int p7_sd_asn1_len;
  char *p7_ed_asn1 = NULL;
  unsigned int p7_ed_asn1_len;
  int message_type;
  char *pki_msg_asn1 = NULL;
  unsigned int pki_msg_asn1_len;

  TRC_ENTRY("scep_process_pkioperation");

  /***************************************************************************/
  /* It's a PKIOperation.  We need to remove the HTTP '%' escape sequences.  */
  /* Note that this is done in-place.                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Remove 'percent' escape sequences"));
  rc = scep_remove_escapes(msg,
                           &msg_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Problem removing 'percent' escapes from base64 (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Un-base64 the data.                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Decode base 64"));
  rc = osp_base64_decode(msg,
                         msg_len,
                         &p7_sd_asn1,
                         &p7_sd_asn1_len);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Couldn't decode the PKIOperation structure.                           */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to convert PKIOperation from base64 (%u)", rc));
    goto EXIT_LABEL;
  }

  if (p7_sd_asn1 == NULL)
  {
    TRC_ERROR((TB, "No result after converting from base64"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Assuming we have valid data, p7_sd_asn1 now points to an ASN.1          */
  /* encoding of a PKCS#7 SignedData structure.  We need to verify it.       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Verify PKCS#7 SignedData"));
  rc = ccm_pkcs7_verify(p7_sd_asn1,
                        p7_sd_asn1_len,
                        NULL,
                        0,
                        CCM_FLAG_SELF_SIGNED,
                        &auth_attribs,
                        &p7_ed_asn1,
                        &p7_ed_asn1_len,
                        &state->ss_cert,
                        &state->ss_cert_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to verify PKCS#7 SignedData (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* It's an error if there are no authenticated attributes.                 */
  /***************************************************************************/
  if (auth_attribs == NULL)
  {
    TRC_ERROR((TB, "No authenticated attributes in SignedData"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Extract all of the authenticated attributes we'll need.  One of these   */
  /* is messageType, which we will check after decryption.                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Extract attribs"));
  rc = scep_extract_attribs(auth_attribs,
                            &state->transaction_id,
                            &state->message_type,
                            &state->n_once,
                            &state->n_once_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to extract required authenticated attributes"
                                                              "from request"));
    if (rc == OSP_ERR_MISSING_ATTR)
    {
      /***********************************************************************/
      /* A missing attribute implies a bad request.                          */
      /***********************************************************************/
      TRC_NORMAL((TB, "One of the attributes was missing"));
      rc = OSP_ERR_BAD_REQUEST;
    }
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now decrypt the resulting EnvelopedData.                                */
  /***************************************************************************/
  TRC_DETAIL((TB, "Decrypt PKCS#7 EnvelopedData"));
  rc = ccm_pkcs7_decrypt(p7_ed_asn1,
                         p7_ed_asn1_len,
                         scep.ca_name,
                         &pki_msg_asn1,
                         &pki_msg_asn1_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to decrypt PKCS#7 EnvelopedData"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Check the message type.  We currently only support PKCSReq.             */
  /***************************************************************************/
  message_type = atoi(state->message_type);
  switch (message_type)
  {
    /*************************************************************************/
    /* A PKCSReq message.                                                    */
    /*************************************************************************/
    case SCEP_MSGTYPE_INT_PKCSREQ:
      TRC_NORMAL((TB, "Got PKCSReq"));
      rc = scep_process_pkcsreq(state,
                                pki_msg_asn1,
                                pki_msg_asn1_len);
      break;

    /*************************************************************************/
    /* Message types we recognize but don't support.                         */
    /*************************************************************************/
    case SCEP_MSGTYPE_INT_GETCERTINIT:
    case SCEP_MSGTYPE_INT_GETCERT:
    case SCEP_MSGTYPE_INT_GETCRL:
      TRC_ERROR((TB, "Received unsupported SCEP request %d", message_type));

      /***********************************************************************/
      /* Send a failure response and inherit the resulting return code.      */
      /***********************************************************************/
      rc = scep_send_scep_failure(state, (char *)scep_failinfo_badreq);
      if (rc != OSP_SUCCESS)
      {
        TRC_ERROR((TB, "Failed to send SCEP failure response (error %u)", rc));
      }
      goto EXIT_LABEL;
      break;

    /*************************************************************************/
    /* Unrecognized message types.                                           */
    /*************************************************************************/
    default:
      TRC_ERROR((TB, "Received unrecognized SCEP request %d", message_type));

      /***********************************************************************/
      /* Send a failure response and inherit the resulting return code.      */
      /***********************************************************************/
      rc = scep_send_scep_failure(state, (char *)scep_failinfo_badreq);
      if (rc != OSP_SUCCESS)
      {
        TRC_ERROR((TB, "Failed to send SCEP failure response (error %u)", rc));
      }
      goto EXIT_LABEL;
      break;
  }

  /***************************************************************************/
  /* Now check the results of our processing.                                */
  /***************************************************************************/
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Something went wrong.                                                 */
    /*************************************************************************/
    if (rc == OSP_ERR_BAD_REQUEST)
    {
      /***********************************************************************/
      /* It's a bad request.  We attempt to send a SCEP failure response.    */
      /* If that succeeds, we return success.                                */
      /***********************************************************************/
      TRC_ERROR((TB, "Bad request; sending SCEP failure response"));
      rc = scep_send_scep_failure(state, (char *)scep_failinfo_badreq);
      if (rc != OSP_SUCCESS)
      {
        TRC_ERROR((TB, "Failed to send SCEP failure response (error %u)", rc));
        goto EXIT_LABEL;
      }
    }
    else
    {
      /***********************************************************************/
      /* We have some other error.  Returning failure from this function     */
      /* will result in an HTTP error being sent.                            */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to process SCEP message type %d; error %u",
                                                            message_type, rc));
      goto EXIT_LABEL;
    }
  }

  TRC_NORMAL((TB, "Processed PKIOperation (type %d)", message_type));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (auth_attribs != NULL)
  {
    sk_X509_ATTRIBUTE_pop_free(auth_attribs, X509_ATTRIBUTE_free);
  }

  if (p7_sd_asn1 != NULL)
  {
    rm_release_mem(p7_sd_asn1);
  }

  if (p7_ed_asn1 != NULL)
  {
    rm_release_mem(p7_ed_asn1);
  }

  if (pki_msg_asn1 != NULL)
  {
    rm_release_mem(pki_msg_asn1);
  }

  TRC_EXIT();

  return(rc);

} /* scep_process_pkioperation */

/**PROC+**********************************************************************/
/* Name:      scep_process_pkcsreq                                           */
/*                                                                           */
/* Purpose:   Process a PKCS#10 certificate request (PKCSReq)                */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     state          - the request state information          */
/*            IN     p10_asn1       - the ASN.1 DER PKCS#10 cert request     */
/*            IN     p10_asn1_len   - length of the cert request             */
/*                                                                           */
/* Operation: Pass the request to CCM for validation and signing. The        */
/*            response is asynchronous, via the callback function.           */
/*                                                                           */
/*            Note that the only time we return failure to the caller is if  */
/*            we failed to initiate a request _and_ failed to send a SCEP    */
/*            failure message.                                               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_process_pkcsreq(SCEP_STATE *state,
                            char *p10_asn1,
                            unsigned int p10_asn1_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;

  TRC_ENTRY("scep_process_pkcsreq");

  /***************************************************************************/
  /* There must be a single self-signed certificate in the request, and      */
  /* this should have been returned in state->ss_cert when we called         */
  /* ccm_pkcs7_verify().                                                     */
  /***************************************************************************/
  if (state->ss_cert == NULL)
  {
    TRC_ERROR((TB, "Didn't find expected self-signed cert in PKCSReq"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Dispatch the PKCS#10 request to CCM.                                    */
  /***************************************************************************/
  TRC_NORMAL((TB, "Sending PKCS#10 cert request to CCM"));
  TRC_DATA("PKCS#10 data:", p10_asn1, p10_asn1_len);
  rc = ccm_request_new_cert(p10_asn1,
                            p10_asn1_len,
                            state->transaction_id,
                            scep.ca_name,
                            (PCCM_NEW_CERT_CALLBACK)scep_new_cert_callback,
                            (void *)state);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Something went wrong.                                                 */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to initiate new certificate request (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Initiated new certificate request"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* scep_process_pkcsreq */

/**PROC+**********************************************************************/
/* Name:      scep_new_cert_callback                                         */
/*                                                                           */
/* Purpose:   Pass a newly-generated certificate from CCM to the client      */
/*                                                                           */
/* Returns:   None.                                                          */
/*                                                                           */
/* Params:    IN     corr           - the transaction's control block        */
/*            IN     x509_asn1      - ASN.1 X.509 certificate                */
/*            IN     x509_asn1_len  - length of X.509                        */
/*                                                                           */
/* Operation: This consists of the following steps:                          */
/*            - construct CertRep (certs-only PKCS#7 SignedData)             */
/*            - wrap in PKCS#7 EnvelopedData                                 */
/*            - wrap in PKCS#7 SignedData with appropriate signed attributes */
/*            - send the response                                            */
/*                                                                           */
/**PROC-**********************************************************************/

void scep_new_cert_callback(void *corr,
                            char *x509_asn1,
                            unsigned int x509_asn1_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  SCEP_STATE *state = (SCEP_STATE *)corr;
  STACK_OF(X509_ATTRIBUTE) *auth_attribs = NULL;
  char *certrep_asn1 = NULL;
  unsigned int certrep_asn1_len;
  char *p7_ed_asn1 = NULL;
  unsigned int p7_ed_asn1_len;
  char *p7_sd_asn1 = NULL;
  unsigned int p7_sd_asn1_len;

  TRC_API_ENTRY("scep_new_cert_callback");

  /***************************************************************************/
  /* If a null pointer is passed in, that means CCM failed to issue the      */
  /* certificate.                                                            */
  /***************************************************************************/
  if (x509_asn1 == NULL)
  {
    TRC_ERROR((TB, "CCM failed to issue a new certificate"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* First, we build the CertRep message.                                    */
  /***************************************************************************/
  TRC_NORMAL((TB, "Got new cert from CCM: build and send CertRep message"));
  rc = scep_build_certrep(x509_asn1, x509_asn1_len,
                                             &certrep_asn1, &certrep_asn1_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to build CertRep message (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Wrap the CertRep into a PKCS#7 EnvelopedData using the self-signed      */
  /* certificate that the end entity passed to us in its PKCSReq.            */
  /***************************************************************************/
  TRC_DETAIL((TB, "Encrypt to make PKCS#7 EnvelopedData"));
  rc = ccm_pkcs7_encrypt(NULL,
                         state->ss_cert,
                         state->ss_cert_len,
                         certrep_asn1,
                         certrep_asn1_len,
                         CCM_CIPHER_DES_CBC,
                         &p7_ed_asn1,
                         &p7_ed_asn1_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to wrap CertRep in EnvelopedData (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Make a stack of the attributes that are common to success and failure.  */
  /***************************************************************************/
  rc = scep_make_common_attribs(state,
                                scep_msgtype_certrep,
                                scep_pkistatus_success,
                                &auth_attribs);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to make stack of authenticated attribs (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now wrap the EnvelopedData into a PKCS#7 SignedData.                    */
  /***************************************************************************/
  TRC_DETAIL((TB, "Sign to make PKCS#7 SignedData"));
  rc = ccm_pkcs7_sign(scep.ca_name,
                      p7_ed_asn1,
                      p7_ed_asn1_len,
                      CCM_DIGEST_MD5,
                      auth_attribs,
                      0,
                      &p7_sd_asn1,
                      &p7_sd_asn1_len);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to sign CertRep envelope (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Construct the HTTP and send it.                                         */
  /***************************************************************************/
  TRC_NORMAL((TB, "Sending CertRep response to PKCSReq (%u bytes at %p)",
                                                  p7_sd_asn1_len, p7_sd_asn1));
  rc = scep_send_response(state,
                          SCEP_OP_PKIOPERATION,
                          p7_sd_asn1,
                          p7_sd_asn1_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to send CertRep response (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Sent CertRep response to PKCSReq OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    rc = scep_send_scep_failure(state, (char *)scep_failinfo_badreq);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to send SCEP failure response (error %u)", rc));
    }
  }

  /***************************************************************************/
  /* We don't need any of these buffers now.  scep_send_response() will have */
  /* copied the contents of p7_sd_asn1 into the send buffer.                 */
  /***************************************************************************/
  if (auth_attribs != NULL)
  {
    sk_X509_ATTRIBUTE_pop_free(auth_attribs, X509_ATTRIBUTE_free);
  }

  if (certrep_asn1 != NULL)
  {
    rm_release_mem(certrep_asn1);
  }

  if (p7_ed_asn1 != NULL)
  {
    rm_release_mem(p7_ed_asn1);
  }

  if (p7_sd_asn1 != NULL)
  {
    rm_release_mem(p7_sd_asn1);
  }

  TRC_API_EXIT();

  return;

} /* scep_new_cert_callback */

