
/**MOD+***********************************************************************/
/* Module:    sceputil.c                                                     */
/*                                                                           */
/* Purpose:   SCEP utility functions                                         */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.21               $ $Modtime::   Jan 11 2001 15:56:42   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <scepincl.h>

/**PROC+**********************************************************************/
/* Name:      scep_build_certrep                                             */
/*                                                                           */
/* Purpose:   Build a CertRep structure (certificates-only PKCS#7)           */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     x509_asn1      - the ASN.1 DER-encoded certificate      */
/*            IN     x509_asn1_len  - length of the certificate              */
/*            OUT    cert_rep_asn1  - ASN.1 DER-encoded CertRep structure    */
/*            OUT    cert_rep_asn1_len                                       */
/*                                  - length of the CertRep structure        */
/*                                                                           */
/* Operation: Create a PKCS#7 structure; add the supplied certificate and    */
/*            the SCEP CA certificate; return it in DER form.                */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_build_certrep(char *x509_asn1,
                          unsigned int x509_asn1_len,
                          char **certrep_asn1,
                          unsigned int *certrep_asn1_len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  PKCS7 *p7_sd = NULL;
  X509 *new_x509 = NULL;
  X509 *ca_x509 = NULL;
  unsigned char *mem = NULL;
  unsigned char *data;
  int error;

  TRC_ENTRY("scep_build_certrep");

  /***************************************************************************/
  /* Create the PKCS#7 SignedData structure.                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Creating PKCS#7 SignedData for CertRep"));
  p7_sd = PKCS7_new();
  if (p7_sd == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create PKCS#7 object for CertRep"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  error = PKCS7_set_type(p7_sd, NID_pkcs7_signed);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set type of PKCS#7 to SignedData"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  error = PKCS7_content_new(p7_sd, NID_pkcs7_data);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to create new content for SignedData"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Add the new certificate and the SCEP CA certificate.                    */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding new cert and CA cert to PKCS#7"));
  data = (unsigned char *)x509_asn1;
  new_x509 = d2i_X509(NULL, &data, x509_asn1_len);
  if (new_x509 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to convert new cert to internal format"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  error = PKCS7_add_certificate(p7_sd, new_x509);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to add new cert to CertRep"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  data = (unsigned char *)scep.ca_cert;
  ca_x509 = d2i_X509(NULL, &data, scep.ca_cert_len);
  if (ca_x509 == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to convert CA cert to internal format"));
    rc = OSP_ERR_NO_RESOURCES;
    goto EXIT_LABEL;
  }

  error = PKCS7_add_certificate(p7_sd, ca_x509);
  if (error == 0)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to add CA cert to CertRep"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now we convert the OpenSSL PKCS7 structure to DER for the caller.       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Converting PKCS#7 to ASN.1"));
  *certrep_asn1_len = i2d_PKCS7(p7_sd, NULL);
  mem = (unsigned char *)rm_get_mem(*certrep_asn1_len);
  if (mem == NULL)
  {
    TRC_ERROR((TB, "Couldn't get %u bytes to convert PKCS#7 to ASN.1",
                                                           *certrep_asn1_len));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  *certrep_asn1 = (char *)mem;
  *certrep_asn1_len = i2d_PKCS7(p7_sd, &mem);

  TRC_NORMAL((TB, "Built CertRep OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (p7_sd != NULL)
  {
    PKCS7_free(p7_sd);
  }

  if (new_x509 != NULL)
  {
    X509_free(new_x509);
  }

  if (ca_x509 != NULL)
  {
    X509_free(ca_x509);
  }

  if (rc != OSP_SUCCESS)
  {
    if (*certrep_asn1 != NULL)
    {
      rm_release_mem(*certrep_asn1);
      *certrep_asn1 = NULL;
      *certrep_asn1_len = 0;
    }
  }

  TRC_EXIT();

  return(rc);

} /* scep_build_certrep */

/**PROC+**********************************************************************/
/* Name:      scep_send_scep_failure                                         */
/*                                                                           */
/* Purpose:   Build and send a SCEP failure response                         */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     state          - the request state information          */
/*            IN     fail_info      - the reason for the failure             */
/*                                      (scep_failinfo_xxx)                  */
/*                                                                           */
/* Operation: Construct a PKCS#7 SignedData with no contained data; just     */
/*            some signed attributes including 'failInfo' to indicate the    */
/*            error.                                                         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_send_scep_failure(SCEP_STATE *state,
                              char *fail_info)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  STACK_OF(X509_ATTRIBUTE) *auth_attribs = NULL;
  char *p7_asn1 = NULL;
  unsigned int p7_asn1_len;

  TRC_ENTRY("scep_send_scep_failure");

  /***************************************************************************/
  /* Make a stack of the attribs that are common to success and failure.     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Making common attributes"));
  rc = scep_make_common_attribs(state,
                                scep_msgtype_certrep,
                                scep_pkistatus_failure,
                                &auth_attribs);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to make stack of authenticated attribs (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We now add the failure reason code.                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding failInfo code '%s'", fail_info));
  TRC_NORMAL((TB, "SCEP failure: bad request (%s)", scep_failinfo_badreq));
  rc = scep_add_string_attrib(auth_attribs,
                              scep.nid_failInfo,
                              V_ASN1_PRINTABLESTRING,
                              fail_info,
                              0);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to add failInfo attribute '%s' to "
                                 "SCEP failure response (%u)", fail_info, rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Generate the PKCS#7 SignedData structure.                               */
  /***************************************************************************/
  TRC_DETAIL((TB, "Creating PKCS#7 SignedData for failure response"));
  rc = ccm_pkcs7_sign(scep.ca_name,
                      NULL,
                      0,
                      CCM_DIGEST_MD5,
                      auth_attribs,
                      0,
                      &p7_asn1,
                      &p7_asn1_len);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create SCEP failure response (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Construct and send the response.  This will free the state structure if */
  /* it succeeds.                                                            */
  /***************************************************************************/
  TRC_DETAIL((TB, "Preparing and sending failure response"));
  rc = scep_send_response(state,
                          SCEP_OP_PKIOPERATION,
                          p7_asn1,
                          p7_asn1_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to construct and send failure response (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Sent SCEP failure response"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We couldn't send a SCEP failure response, so we'll send an HTTP       */
    /* failure response instead.  This will also free the state structure.   */
    /*************************************************************************/
    TRC_ALERT((TB, "Failed to send a SCEP failure response; "
                                             "sending an HTTP error instead"));
    rc = scep_send_http_failure(state->trans, SM_RESPONSE_500);

    /*************************************************************************/
    /* We're finished with the state structure, so free it.                  */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing scep state"));
    scep_free_state(state);
  }

  if (p7_asn1 != NULL)
  {
    rm_release_mem(p7_asn1);
  }

  if (auth_attribs != NULL)
  {
    TRC_DETAIL((TB, "Freeing auth attribs"));
    sk_X509_ATTRIBUTE_pop_free(auth_attribs, X509_ATTRIBUTE_free);
  }

  TRC_EXIT();

  return(rc);

} /* scep_send_scep_failure */

/**PROC+**********************************************************************/
/* Name:      scep_send_http_failure                                         */
/*                                                                           */
/* Purpose:   Send an HTTP failure response                                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     state          - the request state information          */
/*            IN     http_code      - the reason for the failure             */
/*                                      (SM_RESPONSE_xxx)                    */
/*                                                                           */
/* Operation: Send the HTTP response and free the state structure.           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_send_http_failure(TRANS_INFO *trans,
                              int http_code)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;

  TRC_ENTRY("scep_send_http_failure");

  TRC_ASSERT((trans != NULL), (TB, "Passed NULL trans info"));

  /***************************************************************************/
  /* Call SM to send the response.                                           */
  /***************************************************************************/
  TRC_DETAIL((TB, "Calling SM to send response"));
  rc = sm_send_response(trans, NULL, http_code);
  if (rc == OSP_SUCCESS)
  {
    TRC_NORMAL((TB, "Sent HTTP failure response"));
  }
  else
  {
    TRC_ERROR((TB, "Failed to send response"));
  }

  TRC_EXIT();

  return(rc);

} /* scep_send_http_failure */

/**PROC+**********************************************************************/
/* Name:      scep_make_response_header                                      */
/*                                                                           */
/* Purpose:   Write a MIME header into the transaction's send buffer         */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     trans          - the transaction's control block        */
/*            IN     type           - SCEP_OP_GETCACERT: a CA cert           */
/*                                    SCEP_OP_PKIOPERATION: a PKI message    */
/*                                                                           */
/* Operation: Construct the MIME header according to the SCEP spec,          */
/*            according to the type of response being sent.                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_make_response_header(TRANS_INFO *trans,
                                 unsigned int type)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;

  TRC_ENTRY("scep_make_response_header");

  /***************************************************************************/
  /* The buffer must be big enough for the standard headers.                 */
  /***************************************************************************/
  if (type == SCEP_OP_GETCACERT)
  {
    TRC_ASSERT(((SCEP_RESP_HEAD_LEN +
                 SCEP_RESP_CACERT_LEN +
                 SCEP_RESP_TAIL_LEN) <=
                (trans->send_buffer_size - trans->send_buffer_end)),
               (TB, "Not enough room in buffer for CA cert header"));
  }
  else
  {
    TRC_ASSERT(((SCEP_RESP_HEAD_LEN +
                 SCEP_RESP_PKIMSG_LEN +
                 SCEP_RESP_TAIL_LEN) <=
                (trans->send_buffer_size - trans->send_buffer_end)),
               (TB, "Not enough room in buffer for PKI message header"));
  }

  /***************************************************************************/
  /* Copy the header in, bit by bit.                                         */
  /***************************************************************************/
  strcpy(trans->send_buffer + trans->send_buffer_end, scep_resp_head);
  trans->send_buffer_end += SCEP_RESP_HEAD_LEN;

  if (type == SCEP_OP_GETCACERT)
  {
    TRC_DETAIL((TB, "Adding header for CA cert response"));
    strcpy(trans->send_buffer + trans->send_buffer_end, scep_resp_cacert);
    trans->send_buffer_end += SCEP_RESP_CACERT_LEN;
  }
  else
  {
    TRC_DETAIL((TB, "Adding header for PKI message response"));
    strcpy(trans->send_buffer + trans->send_buffer_end, scep_resp_pkimsg);
    trans->send_buffer_end += SCEP_RESP_PKIMSG_LEN;
  }

  strcpy(trans->send_buffer + trans->send_buffer_end, scep_resp_tail);
  trans->send_buffer_end += SCEP_RESP_TAIL_LEN;

  TRC_DETAIL((TB, "Made %s header",
                  (type == SCEP_OP_GETCACERT) ? scep_resp_cacert
                                              : scep_resp_pkimsg));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* scep_make_response_header */

/**PROC+**********************************************************************/
/* Name:      scep_send_response                                             */
/*                                                                           */
/* Purpose:   Construct a response and pass it to SM for sending. This also  */
/*            frees the state structure.                                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     state          - the request state information          */
/*            IN     type           - passed to scep_make_response_header()  */
/*            IN     data           - data to append to the header           */
/*            IN     data_len       - length of data                         */
/*                                                                           */
/* Operation: Construct the complete SCEP response and call SM to send it.   */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_send_response(SCEP_STATE *state,
                          unsigned int type,
                          char *data,
                          unsigned int data_len)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  TRANS_INFO *trans = state->trans;
  char *new_mem = NULL;
  char *content_type = NULL;

  TRC_ENTRY("scep_send_response");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT(((type == SCEP_OP_GETCACERT) || (type == SCEP_OP_PKIOPERATION)),
             (TB, "Invalid response type: %u", type));

  /***************************************************************************/
  /* We expect the buffer to be empty.                                       */
  /***************************************************************************/
  TRC_ASSERT((trans->send_buffer_start == trans->send_buffer_end),
             (TB, "Send buffer not empty"));

  if (trans->http_version == SM_HTTP_0_9)
  {
    /*************************************************************************/
    /* We're dealing with HTTP/0.9, so we need to construct a MIME header    */
    /* for the response.                                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Making MIME header for HTTP/0.9 response"));
    rc = scep_make_response_header(trans, type);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to make response header (%u)", rc));
      goto EXIT_LABEL;
    }
  }
  else
  {
    /*************************************************************************/
    /* We're dealing with HTTP/1.0 or higher, in which case we supply the    */
    /* ContentType value to SM instead of putting it in with the data.       */
    /*************************************************************************/
    if (type == SCEP_OP_GETCACERT)
    {
      content_type = (char *)scep_resp_cacert;
    }
    else
    {
      content_type = (char *)scep_resp_pkimsg;
    }
    TRC_DETAIL((TB, "Passing content type '%s' to SM (HTTP/1.x)",
                                                                content_type));
  }

  /***************************************************************************/
  /* Realloc the send buffer if there's not enough space left for the data.  */
  /***************************************************************************/
  if (data_len > (trans->send_buffer_size - trans->send_buffer_end))
  {
    TRC_ALERT((TB, "Need to realloc send buffer"));
    new_mem = rm_realloc_mem(trans->send_buffer,
                             trans->send_buffer_end + data_len);
    if (new_mem == NULL)
    {
      /***********************************************************************/
      /* Couldn't realloc: we must fail.                                     */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to realloc send buffer"));
      rc = OSP_ERR_MEMORY;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Reset the send buffer information to reflect the new buffer           */
    /* (send_buffer_start/end remain valid)                                  */
    /*************************************************************************/
    trans->send_buffer = new_mem;
    trans->send_buffer_size = trans->send_buffer_end + data_len;
  }

  /***************************************************************************/
  /* Copy the message data into the send buffer.                             */
  /***************************************************************************/
  memcpy(trans->send_buffer + trans->send_buffer_end,
         data,
         data_len);
  trans->send_buffer_end += data_len;

  /***************************************************************************/
  /* Trace what we're going to send.                                         */
  /***************************************************************************/
  TRC_DATA("Sending SCEP response:",
           trans->send_buffer + trans->send_buffer_start,
           trans->send_buffer_end - trans->send_buffer_start);

  /***************************************************************************/
  /* Now tell SM to send the response.                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Calling SM to send response"));
  rc = sm_send_response(trans, content_type, SM_RESPONSE_200);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to send response"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Sent response, header type %u", type));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc == OSP_SUCCESS)
  {
    /*************************************************************************/
    /* We're finished with the state structure, so free it.                  */
    /*************************************************************************/
    scep_free_state(state);
  }
  else
  {
    /*************************************************************************/
    /* We failed -- empty the send buffer in case we end up back in this     */
    /* function again.                                                       */
    /*************************************************************************/
    trans->send_buffer_end = trans->send_buffer_start;
  }

  TRC_EXIT();

  return(rc);

} /* scep_send_response */

/**PROC+**********************************************************************/
/* Name:      scep_remove_escapes                                            */
/*                                                                           */
/* Purpose:   Removes HTTP %xx escape sequences in-place                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS           - completed successfully                 */
/*            other                 - OSP result code                        */
/*                                                                           */
/* Params:    IN     msg            - the HTTP from which to remove the      */
/*                                      escapes                              */
/*            IN/OUT msg_len        - IN: the length of data to scan         */
/*                                    OUT: the length of the resulting data  */
/*                                                                           */
/* Operation: Scan for escape sequences, convert them and shift the          */
/*            following bytes back to compensate.                            */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC scep_remove_escapes(char *msg,
                           unsigned int *msg_len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  char *ptr = msg;
  int delta = 0;
  char hex[] = "__";
  long value;

  TRC_ENTRY("scep_remove_escapes");

  /***************************************************************************/
  /* Spin through the data, copying bytes and converting escapes on the way. */
  /***************************************************************************/
  TRC_DETAIL((TB, "Removing escapes from %u bytes at %p", *msg_len, msg));
  while (ptr < (msg + *msg_len))
  {
    if (*ptr == '%')
    {
      if (ptr < (msg + *msg_len - 2))
      {
        /*********************************************************************/
        /* Convert this escape sequence and update the delta and pointer.    */
        /*********************************************************************/
        hex[0] = ptr[1];
        hex[1] = ptr[2];
        value = strtol(hex, NULL, 16);
        *(ptr - delta) = (char)value;

        delta += 2;
        ptr += 2;
      }
      else
      {
        TRC_ERROR((TB, "Got escape sequence too close to end of data"));
        goto EXIT_LABEL;
      }
    }
    else if (delta != 0)
    {
      *(ptr - delta) = *ptr;
    }

    ptr++;
  }

  /***************************************************************************/
  /* Return the new message length.                                          */
  /***************************************************************************/
  *msg_len -= delta;

  TRC_NORMAL((TB, "Converted %d escape sequences OK; returning %u bytes",
                                                         delta / 2, *msg_len));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* scep_remove_escapes */

/**PROC+**********************************************************************/
/* Name:      scep_free_state                                                */
/*                                                                           */
/* Purpose:   Free all memory associate with a SCEP_STATE structure.         */
/*                                                                           */
/* Returns:   none                                                           */
/*                                                                           */
/* Params:    IN     state          - the state structure to free            */
/*                                                                           */
/**PROC-**********************************************************************/

void scep_free_state(SCEP_STATE *state)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("scep_free_state");

  if (state != NULL)
  {
    /*************************************************************************/
    /* Free the context members that were dynamically allocated, then free   */
    /* the context itself.                                                   */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing SCEP state structure at %p", state));

    if (state->transaction_id != NULL)
    {
      TRC_DETAIL((TB, "Freeing state->transaction_id"));
      rm_release_mem(state->transaction_id);
    }

    if (state->message_type != NULL)
    {
      TRC_DETAIL((TB, "Freeing state->message_type"));
      rm_release_mem(state->message_type);
    }

    if (state->n_once != NULL)
    {
      TRC_DETAIL((TB, "Freeing state->n_once"));
      rm_release_mem(state->n_once);
    }

    if (state->ss_cert != NULL)
    {
      TRC_DETAIL((TB, "Freeing state->ss_cert"));
      rm_release_mem(state->ss_cert);
    }

    TRC_DETAIL((TB, "Freeing state memory"));
    rm_release_mem(state);

    TRC_DETAIL((TB, "Freed SCEP state structure at %p", state));
  }

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* scep_free_state */

