
/**MOD+***********************************************************************/
/* Module:    smimain.c                                                      */
/*                                                                           */
/* Purpose:   SMIME layer external routines                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.15               $ $Modtime::   Jan 11 2001 15:56:44   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#define SMIME_DEFINE_VARS
#include <smiincl.h>
#undef SMIME_DEFINE_VARS

/**PROC+**********************************************************************/
/* Name:      smime_init                                                     */
/*                                                                           */
/* Purpose:   Initialize the S/MIME layer                                    */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     config         - pointer to the loaded config file data */
/*                                                                           */
/* Operation: Initalize S/MIME global data and the XML layer.                */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_init(LHASH *config)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  char *value;
  char dummy[2];

  TRC_ENTRY("smime_init");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((config != NULL), (TB, "Null config supplied"));

  /***************************************************************************/
  /* Initialize the S/MIME global data.                                      */
  /***************************************************************************/
  smime.signing_dn = NULL;
  smime.default_digest = CCM_DIGEST_DEFAULT;
  smime.send_header_space = 0;
  memset(&smime.callback_mutex, 0, sizeof(smime.callback_mutex));
  smime.nr_cb = NULL;

  /***************************************************************************/
  /* Create the callback mutex.                                              */
  /***************************************************************************/
  rc = osp_create_recursive_mutex(&smime.callback_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create callback mutex"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Save the signing identity.  This is mandatory.                          */
  /***************************************************************************/
  value = CONF_get_string(config, OSP_CONFIG_STACK_GROUP,
                                                (char *)smime_config_identity);
  if ((value == NULL) || (value[0] == 0))
  {
    OSP_OPENSSL_ERROR((TB, "No S/MIME signing identity configured"));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Copy the identity that was retrieved, including the terminating null.   */
  /***************************************************************************/
  TRC_NORMAL((TB, "Storing S/MIME identity %s", value));
  OSP_STRDUP(smime.signing_dn, value);
  if (smime.signing_dn == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for S/MIME signing identity"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get the default signing digest to use.  This is optional.               */
  /***************************************************************************/
  value = CONF_get_string(config, OSP_CONFIG_STACK_GROUP,
                                          (char *)smime_config_default_digest);
  if (value != NULL)
  {
    if (strcasecmp(value, smime_digest_sha1) == 0)
    {
      TRC_NORMAL((TB, "Default S/MIME digest is SHA-1"));
      smime.default_digest = CCM_DIGEST_SHA1;
    }
    else if (strcasecmp(value, smime_digest_md5) == 0)
    {
      TRC_NORMAL((TB, "Default S/MIME digest is MD5"));
      smime.default_digest = CCM_DIGEST_MD5;
    }
    else
    {
      TRC_ERROR((TB, "Invalid config value '%s' for %s\\%s",
                  value, OSP_CONFIG_STACK_GROUP, smime_config_default_digest));
      rc = OSP_ERR_CONFIG;
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Find the space we'll need to reserve for our headers in each            */
  /* transaction's send buffer, not forgetting the space SM will need to put */
  /* our Content-Type value into the HTTP headers.                           */
  /***************************************************************************/
  smime.send_header_space = snprintf(dummy,
                                     1,
                                     smime_part_header_format,
                                     SMIME_BOUNDARY_MAXULONG_VALUES,
                                     smime_content_plain,
                                     MAX_ULONG);

  smime.send_header_space += snprintf(dummy,
                                    1,
                                    smime_content_type_format,
                                    smime_content_signature,
                                    smime_digest_sha1,
                                    SMIME_BOUNDARY_MAXULONG_VALUES);

  TRC_NORMAL((TB, "Send header space is %d", smime.send_header_space));

  /***************************************************************************/
  /* Initialize the XML layer.                                               */
  /***************************************************************************/
  TRC_NORMAL((TB, "Initializing XML layer"));
  rc = xml_init();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize the XML layer (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "S/MIME layer initialized OK"));
  smime_initialized = TRUE;
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Call the termination routine to clean up.                             */
    /*************************************************************************/
    TRC_NORMAL((TB, "Terminating S/MIME because of init failure"));
    smime_term();
  }

  TRC_EXIT();

  return(rc);

} /* smime_init */

/**PROC+**********************************************************************/
/* Name:      smime_term                                                     */
/*                                                                           */
/* Purpose:   Terminate the S/MIME layer                                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Terminate the XML layer and free global data.                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;

  TRC_ENTRY("smime_term");

  /***************************************************************************/
  /* Disable the non-repudiation callback.                                   */
  /***************************************************************************/
  osp_get_recursive_lock(&smime.callback_mutex);
  smime.nr_cb = NULL;
  osp_release_recursive_lock(&smime.callback_mutex);

  /***************************************************************************/
  /* If any of the steps below fails, we still want to continue termination  */
  /* so there are no gotos.                                                  */
  /***************************************************************************/

  /***************************************************************************/
  /* Free global data.                                                       */
  /***************************************************************************/
  if (smime.signing_dn != NULL)
  {
    rm_release_mem(smime.signing_dn);
  }

  rc = osp_destroy_recursive_mutex(&smime.callback_mutex);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to destroy callback mutex"));
  }

  /***************************************************************************/
  /* Terminate the XML layer.                                                */
  /***************************************************************************/
  rc = xml_term();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to terminate the XML layer (%u)", rc));
  }

  TRC_NORMAL((TB, "S/MIME layer terminated"));
  smime_initialized = FALSE;
  rc = OSP_SUCCESS;

  TRC_EXIT();

  return(rc);

} /* smime_term */

/**PROC+**********************************************************************/
/* Name:      smime_request_received                                         */
/*                                                                           */
/* Purpose:   Handle received requests and pass them up to the XML layer     */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     trans          - the transaction to process             */
/*            IN     content_type   - pointer to Content-Type value in HTTP  */
/*                                      headers; not null-terminated         */
/*                                                                           */
/* Operation: Check to see if the request is signed or not. If not, pass     */
/*            to XML layer; if so, verify signature and pass to XML layer    */
/*            only if signature is OK.                                       */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_request_received(TRANS_INFO *trans,
                              char *content_type)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned int request_length;
  unsigned int offset;
  unsigned int content_type_len;
  char *boundary = NULL;

  TRC_ENTRY("smime_request_received");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "No transaction supplied"));

  if (content_type == NULL)
  {
    TRC_ERROR((TB, "No content type supplied"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Reserve space for signing the response before processing the request.   */
  /***************************************************************************/
  trans->send_buffer_start += smime.send_header_space;
  trans->send_buffer_end += smime.send_header_space;
  TRC_ASSERT((trans->send_buffer_end < trans->send_buffer_size),
                         (TB, "Send buffer not big enough for S/MIME header"));

  trans->signing_digest = smime.default_digest;

  /***************************************************************************/
  /* Remember the length of the original HTTP message in case we have to     */
  /* call the non-repudiation callback.                                      */
  /***************************************************************************/
  request_length = trans->recv_buffer_end;

  /***************************************************************************/
  /* Calculate a limiting length for the Content-Type header we've been      */
  /* passed -- it can't be long enough to reach past                         */
  /* trans->recv_buffer_start, as that indicates the start of the HTTP body. */
  /***************************************************************************/
  content_type_len = (trans->recv_buffer + trans->recv_buffer_start) -
                                                                  content_type;

  /***************************************************************************/
  /* Parse content_type to determine whether the request is signed and       */
  /* extract the boundary string if so.                                      */
  /***************************************************************************/
  rc = smime_parse_content_type(trans,
                                content_type,
                                content_type_len,
                                &boundary);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to parse request"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If the request is signed, we must verify the signature.  We use the     */
  /* presence of a boundary to indicate this.                                */
  /***************************************************************************/
  if (boundary != NULL)
  {
    /*************************************************************************/
    /* Verify the signature on the request.  This will also adjust the recv  */
    /* buffer start and end to the XML body of the request.                  */
    /*************************************************************************/
    TRC_NORMAL((TB, "Verifying signed request"));
    rc = smime_verify_request(trans, boundary);
    if (rc == OSP_ERR_FAILED_VERIFY)
    {
      TRC_ERROR((TB, "Bad signature on S/MIME-signed request from client "
                     "ip: %08x; ver_cookie=%p",
                    trans->client_id.ip_address, trans->client_id.ver_cookie));
      goto EXIT_LABEL;
    }
    else if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Error occurred while verifying S/MIME-signed request."));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Call the non-repudiation callback, if one is registered.              */
    /*************************************************************************/
    osp_get_recursive_lock(&smime.callback_mutex);
    if (smime.nr_cb != NULL)
    {
      TRC_NORMAL((TB, "Calling non-repudiation callback"));
      (*smime.nr_cb)(trans->recv_buffer, request_length, &trans->client_id);
      TRC_NORMAL((TB, "Back from non-repudiation callback"));
    }
    else
    {
      TRC_DETAIL((TB, "No non-repudiation callback registered"));
    }
    osp_release_recursive_lock(&smime.callback_mutex);
  }

  /***************************************************************************/
  /* The request is now ready to go up to the XML layer, which will send a   */
  /* failure response if anything goes wrong.                                */
  /***************************************************************************/
  xml_request_received(trans);

  TRC_NORMAL((TB, "Processed %s request OK",
                                (boundary != NULL) ? "signed" : "plain text"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* No need to free boundary because it points into the receive buffer.     */
  /***************************************************************************/

  TRC_EXIT();

  return(rc);

} /* smime_request_received */

/**PROC+**********************************************************************/
/* Name:      smime_send_response                                            */
/*                                                                           */
/* Purpose:   Sign (if required) and pass on a response to SM for sending    */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     trans          - the transaction to be signed/sent      */
/*                                                                           */
/* Operation: Check if response needs signing; sign if so and then call SM   */
/*            to queue the transaction for sending.                          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_send_response(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  char *content_type = NULL;
  int sm_resp;

  TRC_ENTRY("smime_send_response");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "No transaction supplied"));

  /***************************************************************************/
  /* See if the response needs signing or not.  We never try to sign if      */
  /* we're sending a non-success response.                                   */
  /***************************************************************************/
  if (trans->sign_response && (trans->http_status < 300))
  {
    TRC_NORMAL((TB, "Signing response (trans ptr %p)", trans));
    rc = smime_sign_response(trans, &content_type);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We failed to sign the response: force an (unsigned) http error to   */
      /* be sent instead                                                     */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to sign response (%u)", rc));
      content_type = (char *)smime_content_plain;
      trans->http_status = 500;
    }
  }
  else
  {
    TRC_DETAIL((TB, "No signing required (trans ptr %p)", trans));
    content_type = (char *)smime_content_plain;
  }

  /***************************************************************************/
  /* Translate to SM response codes                                          */
  /*                                                                         */
  /* @@@ SM will eventually read http_status directly                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Mapping http status %d to...", trans->http_status));
  if (trans->http_status < 300)
  {
    TRC_DETAIL((TB, "200 OK"));
    sm_resp = SM_RESPONSE_200;
  }
  else if (trans->http_status < 500)
  {
    TRC_DETAIL((TB, "400 Error"));
    sm_resp = SM_RESPONSE_400;
  }
  else if (trans->http_status < 999)
  {
    TRC_DETAIL((TB, "500 Error"));
    sm_resp = SM_RESPONSE_500;
  }
  else
  {
    TRC_ASSERT((trans->http_status == 999), (TB, "Unknown HTTP status code"));
    TRC_ASSERT((FALSE), (TB, "SM doesn't yet support 999"));
  }

  /***************************************************************************/
  /* Call SM to send the response.                                           */
  /***************************************************************************/
  rc = sm_send_response(trans, content_type, sm_resp);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Error sending response (%u)", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Processed and sent response OK"));

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if ((content_type != NULL) && (content_type != smime_content_plain))
  {
    rm_release_mem(content_type);
  }

  TRC_EXIT();

  return(rc);

} /* smime_send_response */

/**PROC+**********************************************************************/
/* Name:      osp_non_repudiation_register                                   */
/*                                                                           */
/* Purpose:   Register a non-repudiation callback which is called for each   */
/*            successfully verified S/MIME message.                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - success                                     */
/*            OSP_ERR_NOT_INIT - stack was not initialized                   */
/*                                                                           */
/* Params:    IN     nr_cb          - the new callback function              */
/*                                                                           */
/* Operation: Set the new callback, protected by the callback mutex.         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_non_repudiation_register(POSP_NON_REPUDIATION_CALLBACK nr_cb)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_API_ENTRY("osp_non_repudiation_register");

  /***************************************************************************/
  /* Check we're initialized                                                 */
  /***************************************************************************/
  if (!smime_initialized)
  {
    TRC_ERROR((TB, "Called before initialization"));
    rc = OSP_ERR_NOT_INIT;
  }
  else
  {
    /*************************************************************************/
    /* Grab the callback mutex, set the callback and release the mutex.      */
    /*************************************************************************/
    osp_get_recursive_lock(&smime.callback_mutex);

    smime.nr_cb = nr_cb;
    TRC_DETAIL((TB, "Set non-repudiation callback to %p", nr_cb));

    osp_release_recursive_lock(&smime.callback_mutex);
  }

  TRC_API_EXIT();

  return(rc);

} /* osp_non_repudiation_register */

