
/**MOD+***********************************************************************/
/* Module:   smint.c                                                         */
/*                                                                           */
/* Purpose:  SM internal general functions                                   */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.37               $ $Modtime::   Jan 11 2001 15:56:44   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <smincl.h>

/**PROC+**********************************************************************/
/* Name:      sm_get_thread_id                                               */
/*                                                                           */
/* Purpose:   Returns the calling thread's ID.                               */
/*                                                                           */
/* Returns:   The calling thread's ID.                                       */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Uses pthread_self()                                            */
/*                                                                           */
/**PROC-**********************************************************************/

unsigned long sm_get_thread_id(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  unsigned long rc;

  TRC_ENTRY("sm_get_thread_id");

  rc = (unsigned long)pthread_self();

  TRC_DETAIL((TB, "Thread ID is %lu", rc));

  TRC_EXIT();

  return(rc);

} /* sm_get_thread_id */

/*****************************************************************************/
/* Suppress tracing within sm_locking_callback.                              */
/*****************************************************************************/
#define trc_trace_types TRC_LEVEL_ALERT

/**PROC+**********************************************************************/
/* Name:      sm_locking_callback                                            */
/*                                                                           */
/* Purpose:   Locking function required by OpenSSL.                          */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN mode - CRYPTO_LOCK to lock, CRYPTO_UNLOCK to unlock         */
/*            IN type - index that identifies the lock                       */
/*            IN file - NULL terminated calling file name (debug use only)   */
/*            IN line - calling line number (debug use only)                 */
/*                                                                           */
/* Operation: Uses SM locking macros to lock/unlock.                         */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_locking_callback(int mode, int type, const char *file, int line)
{
  TRC_ENTRY("sm_locking_callback");

  if (mode & CRYPTO_LOCK)
  {
    TRC_DETAIL((TB, "Get SSL lock: lock=%d mode=%x", type, mode));
    OSP_GET_LOCK(&(smg_listener_info.ssl_locks[type]));
  }
  else if (mode & CRYPTO_UNLOCK)
  {
    TRC_DETAIL((TB, "Release SSL lock: lock=%d mode=%x", type, mode));
    OSP_RELEASE_LOCK(&(smg_listener_info.ssl_locks[type]));
  }
  else
  {
    TRC_ASSERT((FALSE),
               (TB, "Invalid lock mode: lock=%d mode=%x", type, mode));
  }

  TRC_EXIT();

  return;

} /* sm_locking_callback */

/*****************************************************************************/
/* Restore normal tracing behaviour.                                         */
/*****************************************************************************/
#undef trc_trace_types

/**PROC+**********************************************************************/
/* Name:      sm_ssl_verify_callback                                         */
/*                                                                           */
/* Purpose:   Called by CCM (from SSL) to verify a remote client             */
/*            certificate.                                                   */
/*                                                                           */
/* Returns:   TRUE  - continue with connection.                              */
/*            FALSE - refuse connection.                                     */
/*                                                                           */
/* Params:    IN     ssl        - SSL structure that manages this connection */
/*            IN     cert_len   - length of certificate                      */
/*            IN     cert       - certificate                                */
/*            IN     chain_len  - number of certificates in chain.           */
/*            IN     chain_pos  - position of certificate in chain.          */
/*                                                                           */
/* Operation: If an application has registered their own callback we call    */
/*            that, otherwise we accept all connections.                     */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_ssl_verify_callback(SSL *ssl,
                           char *cert,
                           unsigned int cert_len,
                           unsigned int chain_len,
                           unsigned int chain_pos)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc;

  TRC_ENTRY("sm_ssl_verify_callback");

  rc = sm_call_verify_callback(SSL_get_app_data(ssl),
                               cert,
                               cert_len,
                               chain_len,
                               chain_pos);

  TRC_EXIT();

  return(rc);

} /* sm_ssl_verify_callback */

/**PROC+**********************************************************************/
/* Name:      sm_call_verify_callback                                        */
/*                                                                           */
/* Purpose:   Call the application's client verification callback and        */
/*            return whether or not we should continue with the connection.  */
/*                                                                           */
/* Returns:   TRUE  - continue with connection.                              */
/*            FALSE - refuse connection.                                     */
/*                                                                           */
/* Params:    IN     connection - CONN_INFO structure for this connection    */
/*            IN     cert_len   - length of certificate                      */
/*            IN     cert       - certificate                                */
/*            IN     chain_len  - number of certificates in chain.           */
/*            IN     chain_pos  - position of certificate in chain.          */
/*                                                                           */
/* Operation: If an application has registered their own callback we call    */
/*            that, otherwise we accept all connections.                     */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_call_verify_callback(CONN_INFO *connection,
                            char *cert,
                            unsigned int cert_len,
                            unsigned int chain_len,
                            unsigned int chain_pos)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = TRUE;
  int error;
  OSP_RC int_rc;

  TRC_ENTRY("sm_call_verify_callback");

  osp_get_recursive_lock(&smg_listener_info.callback_mutex);

  if (smg_listener_info.cv_cb != NULL)
  {
    /*************************************************************************/
    /* Application has registered a callback so call it.                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Calling client verification callback"));

    int_rc = (*smg_listener_info.cv_cb)(connection->client_id.ip_address,
                                        &(connection->client_id.ver_cookie),
                                        cert,
                                        cert_len,
                                        chain_len,
                                        chain_pos);

    if (int_rc != OSP_SUCCESS)
    {
      TRC_ALERT((TB, "Connection from %d.%d.%d.%d denied %d",
                    ((char *)(&connection->client_id.ip_address))[0],
                    ((char *)(&connection->client_id.ip_address))[1],
                    ((char *)(&connection->client_id.ip_address))[2],
                    ((char *)(&connection->client_id.ip_address))[3], int_rc));
      rc = FALSE;
    }
  }
  else
  {
    /*************************************************************************/
    /* Application hasn't registered a callback.  Just accept the            */
    /* connection.                                                           */
    /*************************************************************************/
    TRC_DETAIL((TB,
                  "Automatically verifying client as no callback registered"));
  }

  osp_release_recursive_lock(&smg_listener_info.callback_mutex);

  /***************************************************************************/
  /* Note that the client verification callback has been called.             */
  /***************************************************************************/
  connection->called_cv_cb = 1;

  TRC_EXIT();

  return(rc);

} /* sm_call_verify_callback */

/**PROC+**********************************************************************/
/* Name:      sm_create_socket                                               */
/*                                                                           */
/* Purpose:   Creates a new socket for SM.                                   */
/*                                                                           */
/* Returns:   The socket handle or SM_INVALID_SOCKET                         */
/*                                                                           */
/* Params:    IN port - port to bind socket to                               */
/*                                                                           */
/* Operation: Creates a new socket, sets ot to nonblocking and binds the     */
/*            specified address.                                             */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_create_socket(unsigned short port)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int new_socket;
  int file_flags;
  int error;
  struct sockaddr_in sin;
  int value;

  TRC_ENTRY("sm_create_socket");

  /***************************************************************************/
  /* First create the socket - failure code is -1 = SM_INVALID_SOCKET        */
  /***************************************************************************/
  new_socket = socket(AF_INET, SOCK_STREAM, 0);

  TRC_ASSERT((new_socket < FD_SETSIZE),
        (TB, "FD too large for select call: %d < %d", FD_SETSIZE, new_socket));

  if (new_socket == SM_INVALID_SOCKET)
  {
    /*************************************************************************/
    /* Failed to create socket                                               */
    /*************************************************************************/
    TRC_OS_ERROR((TB, "Failed to create socket"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the socket to non-blocking                                          */
  /***************************************************************************/
  file_flags = fcntl(new_socket, F_GETFL, 0);

  TRC_DETAIL((TB, "Original file settings %x", file_flags));

  error = fcntl(new_socket, F_SETFL, file_flags | FNDELAY);

  if (error == -1)
  {
    TRC_OS_ERROR((TB, "Failed to set socket to non-blocking"));
    close(new_socket);
    new_socket = SM_INVALID_SOCKET;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the option to reuse an address that may appear to the OS still to   */
  /* be bound to another socket.  This usually happens when trying to        */
  /* restart the server after an abnormal termination.                       */
  /***************************************************************************/
  value = TRUE;
  error = setsockopt(new_socket,
                     SOL_SOCKET,
                     SO_REUSEADDR,
                     (char *)&value,
                     sizeof(int));
  if (error == -1)
  {
    TRC_OS_ERROR((TB, "Failed to set SO_REUSEADDR socket option"));
    close(new_socket);
    new_socket = SM_INVALID_SOCKET;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now bind to supplied address.                                           */
  /***************************************************************************/
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = htonl(INADDR_ANY);
  sin.sin_port = htons(port);

  TRC_DETAIL((TB, "Binding to %u", port));

  error = bind(new_socket,
               (struct sockaddr *)&sin,
               sizeof(sin));

  if (error != 0)
  {
    TRC_OS_ERROR((TB, "Failed to bind to port %u", port));
    close(new_socket);
    new_socket = SM_INVALID_SOCKET;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Created socket FD %d bound to %u", new_socket, port));

EXIT_LABEL:

  TRC_EXIT();

  return(new_socket);

} /* sm_create_socket */

/**PROC+**********************************************************************/
/* Name:      sm_get_transaction                                             */
/*                                                                           */
/* Purpose:   Get new transaction structure.                                 */
/*                                                                           */
/* Returns:   Pointer to the new TRANS_INFO structure or NULL.               */
/*                                                                           */
/* Params:    IN conn - connection to be associated with the new transaction */
/*                                                                           */
/* Operation: Allocates a new transaction structure along with the send and  */
/*            receive buffers.                                               */
/*                                                                           */
/**PROC-**********************************************************************/

TRANS_INFO *sm_get_transaction(CONN_INFO *conn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  TRANS_INFO *transaction;
  OSP_RC int_rc = OSP_ERR_MEMORY;
  int destroy_mutex = FALSE;

  TRC_ENTRY("sm_get_transaction");

  /***************************************************************************/
  /* Get memory for transaction itself                                       */
  /***************************************************************************/
  transaction =
             (TRANS_INFO *)rm_get_structure(sizeof(TRANS_INFO), RM_TRANS_INFO);

  if (transaction == NULL)
  {
    TRC_ERROR((TB, "Failed to allocate transaction"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize all fields.                                                  */
  /***************************************************************************/
  memset(&transaction->links, 0, sizeof(transaction->links));
  transaction->connection = conn;
  conn->last_assigned_serial++;
  transaction->serial = conn->last_assigned_serial;
  memcpy(&(transaction->client_id),
         &(conn->client_id),
         sizeof(transaction->client_id));
  transaction->xml_state = NULL;
  transaction->message_id = NULL;
  transaction->http_version = conn->http_version;
  transaction->aib_state = NULL;
  transaction->sign_response = FALSE;
  transaction->signing_digest = CCM_DIGEST_DEFAULT;
  transaction->parsing = FALSE;
  transaction->request_count = 0;
  transaction->response_count = 0;
  transaction->on_send_queue = FALSE;
  transaction->http_status = OSP_STATUS_SUCCESS;
  transaction->recv_buffer_size = SM_RECV_BUFFER_SIZE;
  transaction->recv_buffer_start = 0;
  transaction->recv_buffer_end = 0;
  transaction->send_buffer_size = SM_SEND_BUFFER_SIZE;
  transaction->send_buffer_start = SM_RESERVED_HEADER_SIZE;
  transaction->send_buffer_end = SM_RESERVED_HEADER_SIZE;
  transaction->recv_buffer = NULL;
  transaction->send_buffer = NULL;
  transaction->thread = conn->thread;
  transaction->pool = NULL;

  /***************************************************************************/
  /* Set up receive buffer.                                                  */
  /***************************************************************************/
  transaction->recv_buffer = rm_get_mem(SM_RECV_BUFFER_SIZE);

  if (transaction->recv_buffer == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for receive buffer"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up send buffer.                                                     */
  /***************************************************************************/
  transaction->send_buffer = rm_get_mem(SM_SEND_BUFFER_SIZE);

  if (transaction->send_buffer == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for send buffer"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up transaction mutex.                                               */
  /***************************************************************************/
  int_rc = osp_create_mutex(&transaction->mutex);

  if (int_rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create transaction mutex"));
    goto EXIT_LABEL;
  }

#ifdef OSP_MEASURE_TRANSACTION
  memset(&transaction->timings[0], 0, SM_NUM_TIMINGS * sizeof(SM_TIMINGS));
#endif

  TRC_DETAIL((TB, "Created transaction %p", transaction));
  int_rc = OSP_SUCCESS;

EXIT_LABEL:

  if (int_rc != OSP_SUCCESS)
  {
    TRC_DETAIL((TB, "Free what we've allocated so far"));
    sm_free_transaction(transaction, destroy_mutex);
    transaction = NULL;
  }

  TRC_EXIT();

  return(transaction);

} /* sm_get_transaction */

/**PROC+**********************************************************************/
/* Name:      sm_free_transaction                                            */
/*                                                                           */
/* Purpose:   Free the transaction                                           */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT transaction   - to free                                 */
/*            IN     destroy_mutex - TRUE to destroy the mutex               */
/*                                   FALSE if the mutex wasn't initialized   */
/*                                                                           */
/* Operation: Frees the send and receive buffers, destroys the mutex and     */
/*            frees the transaction itself.                                  */
/*            destroy_mutex should always be TRUE except when called during  */
/*            transaction initialization.                                    */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_free_transaction(TRANS_INFO *transaction, int destroy_mutex)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int itimings;

  TRC_ENTRY("sm_free_transaction");

  if (transaction != NULL)
  {
#ifdef OSP_MEASURE_TRANSACTION
    /*************************************************************************/
    /* Avoid tracing timings for this transaction if nothing interesting has */
    /* happened to it yet.  This is the case when a connection ends, since   */
    /* there is always a transaction allocated per connection for the next   */
    /* data to come in.                                                      */
    /*************************************************************************/
    if ((transaction->timings[SM_TIMINGS_FINISH_READ].real.tv_sec != 0) ||
        (transaction->timings[SM_TIMINGS_FINISH_READ].real.tv_usec != 0))
    {
      TRC_ALERT((TB, "Transaction timings for transaction %p:", transaction));
      for (itimings = 0; itimings < SM_NUM_TIMINGS; itimings++)
      {
        TRC_ALERT((TB,
                   "%16.16s @ %ld.%06ld s",
                   smg_timings_description[itimings],
                   transaction->timings[itimings].real.tv_sec,
                   transaction->timings[itimings].real.tv_usec));
      }
    }
#endif

    /*************************************************************************/
    /* transaction exists, free it.                                          */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing transaction %p", transaction));
    TRC_ASSERT((transaction->aib_state == NULL), (TB, "aib_state not NULL"));
    rm_release_mem(transaction->recv_buffer);
    rm_release_mem(transaction->send_buffer);
    rm_release_mem(transaction->pool);

    if (transaction->message_id != NULL)
    {
      rm_release_mem(transaction->message_id);
    }

    if (destroy_mutex)
    {
      TRC_DETAIL((TB, "Destroying mutex"));
      osp_destroy_mutex(&transaction->mutex);
    }

    rm_release_structure(transaction, RM_TRANS_INFO);
    transaction = NULL;
  }

  TRC_EXIT();

  return;

} /* sm_free_transaction */

/**PROC+**********************************************************************/
/* Name:      sm_read_config_file                                            */
/*                                                                           */
/* Purpose:   Reads in the stack's config file.                              */
/*                                                                           */
/* Returns:   OSP_SUCCESS          - completed successfully                  */
/*            OSP_ERR_CONFIG       - failed to read configuration file       */
/*                                                                           */
/* Params:    IN  config      - configuration data hash handler.             */
/*            OUT config_data - pointer to parsed configuration data.        */
/*                                                                           */
/* Operation: Reads the stack configuration from the file and do basic       */
/*            sanity check on it.                                            */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sm_read_config_file(LHASH *config, CONFIG_DATA *config_data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERR_CONFIG;
  char *value;

  TRC_ENTRY("sm_read_config_file");

  TRC_ASSERT((config != NULL), (TB, "NULL config data hash handle"));
  TRC_ASSERT((config_data), (TB, "NULL config_data pointer"));

  /***************************************************************************/
  /* Set default values for configuration which can be overwritten by the    */
  /* config file.                                                            */
  /***************************************************************************/
  config_data->non_secure_listening_port = SM_DEFAULT_NONSECURE_PORT;
  config_data->secure_listening_port = SM_DEFAULT_SECURE_PORT;
  config_data->threads_per_processor = SM_DEFAULT_THREADS_PER_PROCESSOR;

  value = CONF_get_string(config, OSP_CONFIG_STACK_GROUP,
                                                       OSP_CONFIG_SM_SSL_PORT);

  if (value != NULL)
  {
    TRC_DETAIL((TB, "%s is %s", OSP_CONFIG_SM_SSL_PORT, value));
    config_data->secure_listening_port = atoi(value);
  }

  value = CONF_get_string(config, OSP_CONFIG_STACK_GROUP,
                                                    OSP_CONFIG_SM_NONSSL_PORT);

  if (value != NULL)
  {
    TRC_DETAIL((TB, "%s is %s", OSP_CONFIG_SM_NONSSL_PORT, value));
    config_data->non_secure_listening_port = atoi(value);
  }

  value = CONF_get_string(config, OSP_CONFIG_STACK_GROUP,
                                          OSP_CONFIG_SM_THREADS_PER_PROCESSOR);

  if (value != NULL)
  {
    TRC_DETAIL((TB, "%s is %s", OSP_CONFIG_SM_THREADS_PER_PROCESSOR, value));
    config_data->threads_per_processor = atoi(value);
  }

  /***************************************************************************/
  /* Consistency checks on data we've read in.                               */
  /***************************************************************************/
  if ((config_data->non_secure_listening_port == 0) &&
      (config_data->secure_listening_port == 0))
  {
    TRC_ERROR((TB, "Not listening on any ports"));
    goto EXIT_LABEL;
  }

  if ((config_data->threads_per_processor < 1) ||
      (config_data->threads_per_processor > 30))
  {
    TRC_ERROR((TB, "Nonsensical value for %s (%d)",
               OSP_CONFIG_SM_THREADS_PER_PROCESSOR,
               config_data->threads_per_processor));
    goto EXIT_LABEL;
  }

  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sm_read_config_file */

/**PROC+**********************************************************************/
/* Name:      sm_strtok                                                      */
/*                                                                           */
/* Purpose:   SM-specific version of strtok which looks for single instances */
/*            of the delimiter character, ie for a delimiter of space, a     */
/*            "Hello  world" will be broken into three tokens, "Hello", ""   */
/*            and "world", while strtok would return just "Hello" and "World"*/
/*                                                                           */
/* Returns:   pointer to start of                                            */
/*                                                                           */
/* Params:    IN     buffer       - buffer to search (NULL on second and     */
/*                                  subsequent invocations)                  */
/*            IN     delim        - delimiter character                      */
/*            IN     place_holder - used to hold current place in buffer     */
/*                                  across invocations (making the routine   */
/*                                  thread-safe)                             */
/*                                                                           */
/**PROC-**********************************************************************/

char *sm_strtok(char *buffer, char delim, char **place_holder)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  char *curr_char;
  char *start_ptr;

  TRC_ENTRY("sm_strtok");

  /***************************************************************************/
  /* Check what we've been passed                                            */
  /***************************************************************************/
  TRC_ASSERT((delim != NULL), (TB, "NULL delimeter"));
  TRC_ASSERT((place_holder != NULL), (TB, "NULL place holder"));

  /***************************************************************************/
  /* Is this the first time we've been called?                               */
  /***************************************************************************/
  if (buffer != NULL)
  {
    /*************************************************************************/
    /* Yes - start searching in buffer                                       */
    /*************************************************************************/
    TRC_DETAIL((TB, "First time"));
    start_ptr = buffer;
  }
  else
  {
    /*************************************************************************/
    /* No - place holder must be pointing at where we last left off          */
    /*************************************************************************/
    TRC_ASSERT((*place_holder != NULL), (TB, "NULL pointer in place holder"));
    start_ptr = *place_holder;
  }

  /***************************************************************************/
  /* Do the search                                                           */
  /***************************************************************************/
  curr_char = start_ptr;
  TRC_DETAIL((TB, "Searching '%s' for '%c'", curr_char, delim));
  while ((*curr_char != 0) && (*curr_char != delim))
  {
    curr_char++;
  }

  /***************************************************************************/
  /* If we reached the end of the string without finding an instance of      */
  /* delim, we must return NULL                                              */
  /***************************************************************************/
  if (*curr_char == 0)
  {
    TRC_DETAIL((TB, "Delimiter not found"));
    start_ptr = NULL;
  }
  else
  {
    /*************************************************************************/
    /* Set the delimiter we've found to NULL and use the supplied place      */
    /* holder to record where we should start next time round                */
    /*************************************************************************/
    *curr_char = 0;
    *place_holder = curr_char + 1;
    TRC_DETAIL((TB, "Found delim: token is '%s'", start_ptr));
  }

  TRC_EXIT();

  return(start_ptr);

} /* sm_strtok */

/**PROC+**********************************************************************/
/* Name:      sm_queue_transaction                                           */
/*                                                                           */
/* Purpose:   Add a transaction to the send queue for its connection         */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - queued successfully                      */
/*            OSP_ERROR           - failed to write wakeup byte to worker    */
/*                                                                           */
/* Params:    IN     transaction  - transaction to add to send queue         */
/*            IN     decrement_outstanding_count                             */
/*                                - whether to decrement the outstanding     */
/*                                  transaction count for this connection    */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sm_queue_transaction(TRANS_INFO *transaction,
                            int decrement_outstanding_count)

{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  TRANS_INFO *queued;
  int error;

  TRC_ENTRY("sm_queue_transaction");

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction finish build time.                                  */
  /***************************************************************************/
  SM_GETTIMES(transaction->timings[SM_TIMINGS_FINISH_BUILD]);
#endif

  /***************************************************************************/
  /* Put this transaction on the send queue in transaction serial number     */
  /* order.                                                                  */
  /***************************************************************************/
  OSP_GET_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);
  if (decrement_outstanding_count)
  {
    transaction->connection->wa_locked_outstanding_count--;
  }
  for (queued = (TRANS_INFO *)
                LIST_GET_FIRST(&transaction->connection->wa_locked_send_queue);
       queued != NULL;
       queued = (TRANS_INFO *)
                  LIST_GET_NEXT(&transaction->connection->wa_locked_send_queue,
                                &queued->links))
  {
    if (transaction->serial < queued->serial)
    {
      TRC_DETAIL((TB, "Insert transaction with serial %d before queued"
                  " transaction with serial %d",
                  transaction->serial,
                  queued->serial));
      LIST_INSERT_BEFORE(&queued->links, &transaction->links);
      break;
    }
  }
  if (queued == NULL)
  {
    TRC_DETAIL((TB, "Insert transaction with serial %d at end of send queue",
                transaction->serial));
    LIST_INSERT_TAIL(&transaction->connection->wa_locked_send_queue,
                     &transaction->links);
  }
  TRC_NORMAL((TB, "Added trans %p to send queue", transaction));
  TRC_NORMAL((TB, "Outstanding transaction count now %d",
              transaction->connection->wa_locked_outstanding_count));

  FD_SET(transaction->connection->sock_handle,
         &transaction->connection->thread->wa_locked.write_set);

  /***************************************************************************/
  /* Wake up the thread to try to write on this connection.  If the          */
  /* connection is closed, and there are no more outstanding transactions,   */
  /* this will result in the connection being freed.                         */
  /***************************************************************************/
  if (transaction->connection->thread->wa_locked.sent_wakeup == 0)
  {
    TRC_DETAIL((TB, "Kicking thread to send response"));
    transaction->connection->thread->wa_locked.sent_wakeup = 1;
#ifdef OSP_MEASURE_PERFORMANCE
    gettimeofday(&transaction->connection->thread->wa_locked.woken, NULL);
#endif
    error = write(transaction->connection->thread->wa_locked.wakeup_pipe_in,
                                                                       &rc, 1);

    if (error != 1)
    {
      TRC_OS_ERROR((TB, "Failed to write down wakeup pipe"));
      rc = OSP_ERROR;
    }
  }
  OSP_RELEASE_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);

  TRC_EXIT();
  return(rc);

} /* sm_queue_transaction */

#ifdef OSP_MEASURE_PERFORMANCE
/**PROC+**********************************************************************/
/* Name:      sm_print_performance_data                                      */
/*                                                                           */
/* Purpose:   Print information accumulated in the supplied performance      */
/*            data structure                                                 */
/*                                                                           */
/* Returns:   none                                                           */
/*                                                                           */
/* Params:    IN     fstream      - File stream to print on.                 */
/*            IN     perf_data    - Pointer to SM_PERF performance data      */
/*                                  structure.                               */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_print_performance_data(FILE *fstream, SM_PERF *perf_data)
{
  int count;
  int x, y, z;

  TRC_ENTRY("sm_print_performance_data");

  /***************************************************************************/
  /* Avoid division by zero in case where count is still zero.               */
  /***************************************************************************/
  count = perf_data->count ? perf_data->count : 1;

  /***************************************************************************/
  /* Reduce precision of the numbers y and z so that we can calculate        */
  /* ((x * y) / z) without overflowing.                                      */
  /***************************************************************************/
  x = perf_data->accum.tv_sec % count;
  y = 1000000;
  z = count;
  while ((y > 1000) && (z > 1000))
  {
    y /= 10;
    z /= 10;
  }

  /***************************************************************************/
  /* Print performance data to the specified file stream.                    */
  /***************************************************************************/
  fprintf(fstream,
          "%s:\n"
          "\tcount = %d\n"
          "\taccumulated time = %ld.%06ld s\n"
          "\taverage time = %ld.%06ld s\n",
          perf_data->name,
          perf_data->count,
          perf_data->accum.tv_sec,
          perf_data->accum.tv_usec,
          perf_data->accum.tv_sec / count,
          ((x * y) / z) + (perf_data->accum.tv_usec / count));

  TRC_EXIT();
  return;

} /* sm_print_performance_data */
#endif
