
/**MOD+***********************************************************************/
/* Module:    smiutil.c                                                      */
/*                                                                           */
/* Purpose:   S/MIME utility functions                                       */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.11               $ $Modtime::   Jan 11 2001 15:56:46   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <smiincl.h>

/**PROC+**********************************************************************/
/* Name:      smime_sign_response                                            */
/*                                                                           */
/* Purpose:   Generate a complete S/MIME signed message.                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     trans          - the transaction to be processed        */
/*            OUT    content_type   - the Content-Type string to be used in  */
/*                                      the HTTP header (caller must free)   */
/*                                                                           */
/* Operation: Canonicalize the XML; use CCM to generate the signature; add   */
/*            S/MIME headers and signature to buffer; build Content-Type.    */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_sign_response(TRANS_INFO *trans, char **content_type)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned int canonical_length;
  char *sig_buf = NULL;
  unsigned int sig_len = 0;
  unsigned long boundary_rand[SMIME_BOUNDARY_RAND_LONGS];
  char temp_buffer[SMIME_MAX_TP_HEADER];
  int header_length;
  int append_length;
  int content_type_length;
  const char *mic_alg = NULL;
  char *new_mem = NULL;
  int new_bufsize;
  char dummy[2];

  TRC_ENTRY("smime_sign_response");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Null transaction supplied"));
  TRC_ASSERT((content_type != NULL), (TB, "Null content_type supplied"));

  /***************************************************************************/
  /* Canonicalize the XML, using the receive buffer for storage.  We assert  */
  /* that canonicalization of our own XML always succeeds.                   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Canonicalizing response"));
  rc = smime_canonicalize(trans->send_buffer + trans->send_buffer_start,
                          trans->send_buffer_end - trans->send_buffer_start,
                          &trans->recv_buffer,
                          &trans->recv_buffer_size,
                          &canonical_length);

  TRC_ASSERT((rc == OSP_SUCCESS), (TB, "Failed to canonicalize our own XML"));

  /***************************************************************************/
  /* Generate a random MIME boundary.  The chances of this sequence being in */
  /* the encapsulated data are astronomically small.                         */
  /***************************************************************************/
  ccm_random(boundary_rand, sizeof(boundary_rand));

  /***************************************************************************/
  /* Construct the S/MIME text/plain header in a temporary buffer.  We can't */
  /* sprintf directly into the send buffer because sprintf/snprintf always   */
  /* add the terminating null; we don't want that because it will overwrite  */
  /* the data that follows the header.                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Prepending S/MIME text/plain header"));
  header_length = snprintf(temp_buffer,
                           SMIME_MAX_TP_HEADER,
                           smime_part_header_format,
                           SMIME_BOUNDARY_RAND_VALUES(boundary_rand),
                           smime_content_plain,
                           trans->send_buffer_end - trans->send_buffer_start);

  TRC_ASSERT((header_length < SMIME_MAX_TP_HEADER),
                     (TB, "S/MIME text/plain header too big for temp buffer"));
  TRC_ASSERT((header_length <= trans->send_buffer_start),
          (TB, "Not enough room in send buffer for S/MIME text/plain header"));

  /***************************************************************************/
  /* Prepend the header.  Note that header_length (returned from snprintf)   */
  /* doesn't include the terminating null.                                   */
  /***************************************************************************/
  trans->send_buffer_start -= header_length;
  memcpy(trans->send_buffer + trans->send_buffer_start,
                                                   temp_buffer, header_length);

  /***************************************************************************/
  /* Call CCM to generate the signature.                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Generating signature"));
  rc = ccm_pkcs7_sign(smime.signing_dn,
                      trans->recv_buffer,
                      canonical_length,
                      trans->signing_digest,
                      NULL,
                      CCM_FLAG_PKCS7_D,
                      &sig_buf,
                      &sig_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to sign canonicalized XML (%u)", rc));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Work out how much space we'll need for everything we need to append.    */
  /***************************************************************************/
  append_length = 0;
  append_length += snprintf(dummy,
                            1,
                            smime_part_header_format,
                            SMIME_BOUNDARY_RAND_VALUES(boundary_rand),
                            smime_content_signature,
                            sig_len);

  append_length += sig_len;

  append_length += snprintf(dummy,
                            1,
                            smime_end_boundary_format,
                            SMIME_BOUNDARY_RAND_VALUES(boundary_rand));

  /***************************************************************************/
  /* Add one for the null terminator that sprintf adds.                      */
  /***************************************************************************/
  append_length += 1;

  /***************************************************************************/
  /* See if a realloc is needed to fit everything in.                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Need to append %d bytes (%d available)", append_length,
                            trans->send_buffer_size - trans->send_buffer_end));

  if (append_length > (trans->send_buffer_size - trans->send_buffer_end))
  {
    /*************************************************************************/
    /* We need to realloc the send buffer.                                   */
    /*************************************************************************/
    TRC_ALERT((TB, "Reallocating send buffer to fit signature header in "
                   "(got %d bytes; need %d)",
             trans->send_buffer_size - trans->send_buffer_end, append_length));

    new_bufsize = trans->send_buffer_end + append_length;
    new_mem = rm_realloc_mem(trans->send_buffer, new_bufsize);
    if (new_mem == NULL)
    {
      TRC_ERROR((TB, "Failed to realloc send buffer from %d to %d",
                                        trans->send_buffer_size, new_bufsize));
      rc = OSP_ERR_MEMORY;
      goto EXIT_LABEL;
    }

    trans->send_buffer = new_mem;
    trans->send_buffer_size = new_bufsize;
  }

  /***************************************************************************/
  /* Append the signature boundary, signature and end boundary.              */
  /***************************************************************************/
  TRC_DETAIL((TB, "Appending S/MIME signature"));
  trans->send_buffer_end += sprintf(trans->send_buffer+trans->send_buffer_end,
                                    smime_part_header_format,
                                    SMIME_BOUNDARY_RAND_VALUES(boundary_rand),
                                    smime_content_signature,
                                    sig_len);

  memcpy(trans->send_buffer + trans->send_buffer_end, sig_buf, sig_len);
  trans->send_buffer_end += sig_len;

  trans->send_buffer_end += sprintf(trans->send_buffer+trans->send_buffer_end,
                                    smime_end_boundary_format,
                                    SMIME_BOUNDARY_RAND_VALUES(boundary_rand));

  /***************************************************************************/
  /* Build the Content-Type string, using digest names from OpenSSL.  First  */
  /* do a dry run to work out how much space is required, and allocate it.   */
  /***************************************************************************/
  TRC_DETAIL((TB, "Building Content-Type string"));
  TRC_ASSERT(((trans->signing_digest == CCM_DIGEST_SHA1) ||
              (trans->signing_digest == CCM_DIGEST_MD5)),
             (TB, "Unknown digest type"));

  if (trans->signing_digest == CCM_DIGEST_SHA1)
  {
    mic_alg = smime_digest_sha1;
  }
  else
  {
    mic_alg = smime_digest_md5;
  }

  content_type_length = snprintf(dummy,
                                 1,
                                 smime_content_type_format,
                                 smime_content_signature,
                                 mic_alg,
                                 SMIME_BOUNDARY_RAND_VALUES(boundary_rand));

  *content_type = (char *)rm_get_mem(content_type_length + 1);
  if (*content_type == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for Content-Type string"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Now build the real Content-Type string in the buffer we've just got.    */
  /***************************************************************************/
  sprintf(*content_type,
          smime_content_type_format,
          smime_content_signature,
          mic_alg,
          SMIME_BOUNDARY_RAND_VALUES(boundary_rand));

  /***************************************************************************/
  /* That's it.                                                              */
  /***************************************************************************/
  TRC_NORMAL((TB, "Signed response OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (sig_buf != NULL)
  {
    rm_release_mem(sig_buf);
  }

  TRC_EXIT();

  return(rc);

} /* smime_sign_response */

/**PROC+**********************************************************************/
/* Name:      smime_canonicalize                                             */
/*                                                                           */
/* Purpose:   Canonicalizes the given XML, reallocating the destination      */
/*            buffer if necessary                                            */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     xml_in         - the XML to be canonicalized            */
/*            IN     xml_in_len     - length of supplied XML                 */
/*            IN/OUT xml_out        - IN: pointer to a buffer to receive     */
/*                                        the canonicalized XML              */
/*                                    OUT: pointer to a (possibly realloced) */
/*                                         buffer containing the             */
/*                                         canonicalized XML                 */
/*            IN/OUT outbuf_len     - IN: the length available in the        */
/*                                        destination buffer                 */
/*                                    OUT: the length of the destination     */
/*                                         buffer if it was realloced        */
/*            OUT    xml_out_len    - length of canonicalized XML            */
/*                                                                           */
/* Operation: Use xml_canonicalize(); realloc buffer if necessary and try    */
/*            for a 2nd attempt. Give up if this fails.                      */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_canonicalize(char *xml_in,
                          unsigned int xml_in_len,
                          char **xml_out,
                          int *outbuf_len,
                          unsigned int *xml_out_len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned int canonical_length = *outbuf_len;
  char *new_mem = NULL;

  TRC_ENTRY("smime_canonicalize");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((xml_in != NULL), (TB, "xml_in is null"));
  TRC_ASSERT((xml_out != NULL), (TB, "xml_out is null"));
  TRC_ASSERT((outbuf_len != NULL), (TB, "outbuf_len is null"));
  TRC_ASSERT((xml_out_len != NULL), (TB, "xml_out_len is null"));

  /***************************************************************************/
  /* Call through to the XML layer to do the real work.                      */
  /***************************************************************************/
  rc = xml_canonicalize(xml_in,
                        xml_in_len,
                        *xml_out,
                        &canonical_length);
  if (rc != OSP_SUCCESS)
  {
    if (rc == OSP_ERR_BUFSIZE)
    {
      /***********************************************************************/
      /* The canonical form didn't fit entirely into the supplied buffer.    */
      /* Reallocate the buffer and try again.                                */
      /***********************************************************************/
      TRC_ALERT((TB, "Need to realloc supplied buffer from %u to %u bytes",
                                               *outbuf_len, canonical_length));
      new_mem = rm_realloc_mem(*xml_out, canonical_length);
      if (new_mem == NULL)
      {
        TRC_ERROR((TB, "Failed to realloc supplied buffer"));
        rc = OSP_ERR_MEMORY;
        goto EXIT_LABEL;
      }

      *xml_out = new_mem;
      *outbuf_len = canonical_length;

      rc = xml_canonicalize(xml_in,
                            xml_in_len,
                            *xml_out,
                            &canonical_length);
      if (rc != OSP_SUCCESS)
      {
        TRC_ERROR((TB, "Failed to canonicalize on 2nd attempt (%u)", rc));
        goto EXIT_LABEL;
      }
    }
    else
    {
      TRC_ERROR((TB, "Failed to canonicalize on 1st attempt (%u)", rc));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Save the length of the canonical form, subtracting the terminating null */
  /* byte.                                                                   */
  /***************************************************************************/
  *xml_out_len = canonical_length - 1;

  TRC_DETAIL((TB, "Canonicalized XML at %p (len %u) into buf at %p (len %u)",
                                  xml_in, xml_in_len, *xml_out, *xml_out_len));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* smime_canonicalize */

/**PROC+**********************************************************************/
/* Name:      smime_parse_content_type                                       */
/*                                                                           */
/* Purpose:   Parses the Content-Type HTTP header to decide whether or not   */
/*            the message is signed                                          */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     trans          - the transaction containing the message */
/*            IN     content_type   - pointer to Content-Type value          */
/*                                      (from HTTP headers)                  */
/*            IN     content_type_len                                        */
/*                                  - length of Content-Type value           */
/*            OUT    boundary       - the boundary if the message is signed  */
/*                                      (null-terminated; caller must free)  */
/*                                                                           */
/* Operation: Parse the Content-Type field value passed in; check the        */
/*            parameters; save off 'micalg' and return 'boundary' to caller. */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_parse_content_type(TRANS_INFO *trans,
                                char *content_type,
                                unsigned int content_type_len,
                                char **boundary)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  COM_STRING type_subtype;
  COM_PARAMS *params = NULL;
  COM_STRING attrib_value;

  TRC_ENTRY("smime_parse_content_type");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((boundary != NULL), (TB, "Null boundary pointer"));

  /***************************************************************************/
  /* Parse the 'media-type' string from the Content-Type field.              */
  /***************************************************************************/
  rc = com_parse_media_type(content_type,
                            content_type_len,
                            &type_subtype,
                            &params);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to parse Content-Type value"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* See if the request is signed or not.                                    */
  /***************************************************************************/
  if (com_strings_match(&type_subtype,
                        smime_content_plain,
                        strlen(smime_content_plain)))
  {
    /*************************************************************************/
    /* It's not signed.  We just have to null the boundary parameter.        */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got plain message"));
    *boundary = NULL;
  }
  else if (com_strings_match(&type_subtype,
                             smime_content_signed,
                             strlen(smime_content_signed)))
  {
    TRC_NORMAL((TB, "Got signed message"));

    /*************************************************************************/
    /* Get and validate the parameters we expect.  There should be three.    */
    /*************************************************************************/
    if (params->num_params != 3)
    {
      TRC_ERROR((TB, "Got %u params for multipart/signed; expecting 3",
                                                          params->num_params));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Expect protocol=application/pkcs7-signature.                          */
    /*************************************************************************/
    TRC_DETAIL((TB, "Checking 'protocol' attribute"));
    rc = com_find_param_value(params,
                              smime_ct_param_protocol,
                              strlen(smime_ct_param_protocol),
                              &attrib_value);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to find 'protocol' parameter"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    if (!com_strings_match(&attrib_value,
                           smime_content_signature,
                           strlen(smime_content_signature)))
    {
      TRC_ERROR((TB, "Unexpected protocol value"));
      TRC_DATA("Bad protocol value:", attrib_value.buffer, attrib_value.len);
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Expect micalg=sha1 or micalg=md5.                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Parsing value of 'micalg' attribute"));
    rc = com_find_param_value(params,
                              smime_ct_param_micalg,
                              strlen(smime_ct_param_micalg),
                              &attrib_value);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to find 'micalg' parameter"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    if (com_strings_match(&attrib_value,
                          smime_digest_sha1,
                          strlen(smime_digest_sha1)))
    {
      TRC_DETAIL((TB, "Got micalg=sha1"));
      trans->signing_digest = CCM_DIGEST_SHA1;
    }
    else if (com_strings_match(&attrib_value,
                               smime_digest_md5,
                               strlen(smime_digest_md5)))
    {
      TRC_DETAIL((TB, "Got micalg=md5"));
      trans->signing_digest = CCM_DIGEST_MD5;
    }
    else
    {
      TRC_ERROR((TB, "Unexpected micalg value"));
      TRC_DATA("Bad micalg value:", attrib_value.buffer, attrib_value.len);
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Get the boundary string.                                              */
    /*************************************************************************/
    TRC_DETAIL((TB, "Storing value of 'boundary' attribute"));
    rc = com_find_param_value(params,
                              smime_ct_param_boundary,
                              strlen(smime_ct_param_boundary),
                              &attrib_value);
    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to find 'boundary' parameter"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    if (attrib_value.len == 0)
    {
      TRC_ERROR((TB, "Zero-length boundary string"));
      rc = OSP_ERR_BAD_REQUEST;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Construct the string with a null terminator.                          */
    /*************************************************************************/
    *boundary = rm_get_mem(strlen(smime_boundary_delimiter) +
                                                         attrib_value.len + 1);
    if (*boundary == NULL)
    {
      TRC_ERROR((TB, "Failed to get memory to build boundary string"));
      rc = OSP_ERR_MEMORY;
      goto EXIT_LABEL;
    }

    strcpy(*boundary, smime_boundary_delimiter);
    memcpy(*boundary + strlen(smime_boundary_delimiter),
                                        attrib_value.buffer, attrib_value.len);
    *(*boundary + strlen(smime_boundary_delimiter) + attrib_value.len) = 0;

    TRC_DETAIL((TB, "Boundary string: '%s'", *boundary));
  }
  else
  {
    TRC_ERROR((TB, "Unknown MIME type in Content-Type header"));
    TRC_DATA("Bad Content-Type value:",
                                        type_subtype.buffer, type_subtype.len);
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Parsed Content-Type header OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (params != NULL)
  {
    rm_release_mem(params);
  }

  TRC_EXIT();

  return(rc);

} /* smime_parse_content_type */

/**PROC+**********************************************************************/
/* Name:      smime_verify_request                                           */
/*                                                                           */
/* Purpose:   Verify the signature of an S/MIME message                      */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     trans          - the transaction containing the message */
/*            IN     boundary       - the boundary string                    */
/*                                                                           */
/* Operation: Split up the message using the boundary string; check entity   */
/*            headers; canonicalize the XML; check the signature             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_verify_request(TRANS_INFO *trans,
                            char *boundary)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  unsigned int offset;
  unsigned int match_index;
  char *xml_part = NULL;
  char *signature_part = NULL;
  char *end_message = NULL;
  int looking_for;
  int boundary_len;
  COM_HEADERS *xml_headers = NULL;
  COM_HEADERS *signature_headers = NULL;
  char *xml_start = NULL;
  char *signature_start = NULL;
  unsigned int xml_len;
  unsigned int signature_len;
  unsigned int canonical_length;

  TRC_ENTRY("smime_verify_request");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Null transaction supplied"));
  TRC_ASSERT((boundary != NULL), (TB, "Null boundary supplied"));

  /***************************************************************************/
  /* Search for the boundary to split up the message.  Note that we assume   */
  /* that the boundary string is immediately preceded by a CRLF.             */
  /***************************************************************************/
  match_index = 0;
  looking_for = 1;
  boundary_len = strlen(boundary);

  for (offset = trans->recv_buffer_start;
                                     offset < trans->recv_buffer_end; offset++)
  {
    if (trans->recv_buffer[offset] == boundary[match_index])
    {
      match_index++;
    }
    else
    {
      match_index = 0;
    }

    if (match_index >= boundary_len)
    {
      /***********************************************************************/
      /* Found an instance of the boundary string.  Now see which boundary   */
      /* we were looking for.                                                */
      /***********************************************************************/
      switch (looking_for)
      {
        case 1:
          /*******************************************************************/
          /* First boundary (beginning of XML part).  Check that there is a  */
          /* following CRLF.                                                 */
          /*******************************************************************/
          if (((offset + 2) >= trans->recv_buffer_end) ||
              (trans->recv_buffer[offset + 1] != '\r') ||
              (trans->recv_buffer[offset + 2] != '\n'))
          {
            TRC_ERROR((TB, "No CRLF after boundary"));
            rc = OSP_ERR_BAD_REQUEST;
            goto EXIT_LABEL;
          }
          TRC_DETAIL((TB, "Got XML part at offset %u", offset + 3));
          xml_part = trans->recv_buffer + offset + 3;
          looking_for = 2;
          break;

        case 2:
          /*******************************************************************/
          /* Second boundary (beginning of signature part).  Check that      */
          /* there is a following CRLF.                                      */
          /*******************************************************************/
          if (((offset + 2) >= trans->recv_buffer_end) ||
              (trans->recv_buffer[offset + 1] != '\r') ||
              (trans->recv_buffer[offset + 2] != '\n'))
          {
            TRC_ERROR((TB, "No CRLF after boundary"));
            rc = OSP_ERR_BAD_REQUEST;
            goto EXIT_LABEL;
          }
          TRC_DETAIL((TB, "Got signature part at offset %u", offset + 3));
          signature_part = trans->recv_buffer + offset + 3;
          looking_for = 3;
          break;

        case 3:
          /*******************************************************************/
          /* Third boundary (end of message).  Just check for a valid        */
          /* end-separator ('--' suffix).  We ignore anything after that.    */
          /*******************************************************************/
          if (((offset + 2) >= trans->recv_buffer_end) ||
              (trans->recv_buffer[offset + 1] != '-') ||
              (trans->recv_buffer[offset + 2] != '-'))
          {
            TRC_ERROR((TB, "Invalid end separator or wrong number of parts"));
            rc = OSP_ERR_BAD_REQUEST;
            goto EXIT_LABEL;
          }

          TRC_DETAIL((TB, "Got end separator; last char at offset %u",
                                                                  offset + 2));
          end_message = trans->recv_buffer + offset + 2;
          looking_for = -1;
          break;

        default:
          TRC_ASSERT((FALSE), (TB, "Invalid state value during parsing"));
          break;
      }
    }

    if (looking_for == -1)
    {
      break;
    }
  }

  /***************************************************************************/
  /* Check that we got everything we need.                                   */
  /***************************************************************************/
  if ((xml_part == NULL) || (signature_part == NULL) ||
                                  (end_message == NULL) || (looking_for != -1))
  {
    TRC_ERROR((TB, "Failed to parse S/MIME signed message"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Parse headers in each part.                                             */
  /***************************************************************************/
  TRC_DETAIL((TB, "Parsing headers from XML part"));
  rc = com_parse_headers(xml_part,
                         signature_part - xml_part,
                         &xml_headers,
                         &xml_start);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to parse XML part headers"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Parsing headers from signature part"));
  rc = com_parse_headers(signature_part,
                         end_message - signature_part,
                         &signature_headers,
                         &signature_start);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to parse signature part headers"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Sanity check.                                                           */
  /***************************************************************************/
  if (signature_start > end_message)
  {
    TRC_ERROR((TB, "Incomplete message -- couldn't find signature"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Find the Content-Length headers for each part.                          */
  /***************************************************************************/
  rc = smime_get_content_length(xml_headers, &xml_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get valid Content-Length from XML part"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  rc = smime_get_content_length(signature_headers, &signature_len);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to get valid Content-Length from signature part"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Adjust the receive buffer pointers to indicate the XML body.            */
  /***************************************************************************/
  trans->recv_buffer_start = xml_start - trans->recv_buffer;
  trans->recv_buffer_end = trans->recv_buffer_start + xml_len;

#ifndef OSP_DONT_CANONICALIZE_XML
  /***************************************************************************/
  /* Canonicalize the XML ready for verification, using the send buffer for  */
  /* temporary storage.  Note that this may reallocate the send buffer if it */
  /* isn't big enough.                                                       */
  /***************************************************************************/
  rc = smime_canonicalize(xml_start,
                          xml_len,
                          &trans->send_buffer,
                          &trans->send_buffer_size,
                          &canonical_length);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to canonicalize received XML"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }
#else
  if (trans->send_buffer_size < xml_len + 1)
  {
    trans->send_buffer = rm_realloc_mem(trans->send_buffer, xml_len + 1);
    trans->send_buffer_size = xml_len + 1;
  }
  memcpy(trans->send_buffer, xml_start, xml_len);
  trans->send_buffer[xml_len] = 0;
  canonical_length = xml_len;
#endif

  /***************************************************************************/
  /* Verify the signature.                                                   */
  /***************************************************************************/
  TRC_NORMAL((TB, "Calling CCM to verify signature"));
  rc = ccm_pkcs7_verify(signature_start,
                        signature_len,
                        trans->send_buffer,
                        canonical_length,
                        CCM_FLAG_PKCS7_D,
                        NULL,
                        NULL,
                        NULL,
                        NULL,
                        NULL);
  if (rc == OSP_ERR_FAILED_VERIFY)
  {
    TRC_ERROR((TB, "Bad signature on S/MIME-signed request"));
    goto EXIT_LABEL;
  }
  else if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Error occurred while verifying S/MIME-signed request."));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We have successfully split up the message and verified the signature.   */
  /***************************************************************************/
  TRC_NORMAL((TB, "Message successfully verified"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  /***************************************************************************/
  /* Clean up.                                                               */
  /***************************************************************************/
  if (xml_headers != NULL)
  {
    rm_release_mem(xml_headers);
  }

  if (signature_headers != NULL)
  {
    rm_release_mem(signature_headers);
  }

  TRC_EXIT();

  return(rc);

} /* smime_verify_request */

/**PROC+**********************************************************************/
/* Name:      smime_get_content_length                                       */
/*                                                                           */
/* Purpose:   Retrieve a Content-Length value from the supplied headers      */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            other       - OSP result code                                  */
/*                                                                           */
/* Params:    IN     headers        - the list of headers                    */
/*            OUT    content_length - the numerical content length           */
/*                                                                           */
/* Operation: Use COM to find the value, then atoi() to convert it.          */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC smime_get_content_length(COM_HEADERS *headers,
                                unsigned int *content_length)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  COM_STRING content_length_value;
  char temp_buffer[SMIME_ATOI_BUFLEN + 1];
  int length;

  TRC_ENTRY("smime_get_content_length");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((headers != NULL), (TB, "Null header list"));
  TRC_ASSERT((content_length != NULL), (TB, "Null content length pointer"));

  /***************************************************************************/
  /* Find the Content-Length value.                                          */
  /***************************************************************************/
  rc = com_find_header_value(headers,
                             smime_content_length,
                             strlen(smime_content_length),
                             &content_length_value);
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to find Content-Length in headers at %p", headers));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Check that it's sensible.                                               */
  /***************************************************************************/
  if (content_length_value.len > SMIME_ATOI_BUFLEN)
  {
    TRC_ERROR((TB, "Invalid Content-Length"));
    TRC_DATA("Bad Content-Length data:",
                        content_length_value.buffer, content_length_value.len);
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Copy to a null-terminated string and convert it to an integer.          */
  /***************************************************************************/
  memcpy(temp_buffer, content_length_value.buffer, content_length_value.len);
  temp_buffer[content_length_value.len] = 0;

  length = atoi(temp_buffer);

  /***************************************************************************/
  /* Check that the result is sensible.                                      */
  /***************************************************************************/
  if (length <= 0)
  {
    TRC_ERROR((TB, "Invalid Content-Length (%d)", length));
  }

  /***************************************************************************/
  /* Pass back to caller.                                                    */
  /***************************************************************************/
  *content_length = (unsigned int)length;

  TRC_DETAIL((TB, "Got Content-Length value of %d", length));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* smime_get_content_length */

