
/**MOD+***********************************************************************/
/* Module:   smlisten.c                                                      */
/*                                                                           */
/* Purpose:  SM Listener thread functions                                    */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.23               $ $Modtime::   Jan 11 2001 15:56:46   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <smincl.h>

/**PROC+**********************************************************************/
/* Name:      sm_listener_thread                                             */
/*                                                                           */
/* Purpose:   Listener thread procedure.                                     */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Select on the listening sockets and the listener termination   */
/*            pipe.                                                          */
/*            If there is data in the termination pipe, close the listening  */
/*            sockets, close the notification pipe for all worker threads    */
/*            and wait until all of the workers have shut down. Then         */
/*            terminate all of the stack components and close the            */
/*            termination pipe.                                              */
/*                                                                           */
/*            If there is an incoming connection, call                       */
/*            sm_listener_handle_new_connection()  deal with it.             */
/*                                                                           */
/**PROC-**********************************************************************/

void *sm_listener_thread(void *dummy)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  fd_set read_set;
  int error;
  int nfds;
  THREAD_INFO *worker;
  CONN_INFO *new_connection;
  int new_socket;
  struct sockaddr_in remote_address;
  int address_len;
  int file_flags;

  TRC_THREAD_ENTRY("sm_listener_thread");

  /***************************************************************************/
  /* Initialization: set up the thread specific data with the SM thread      */
  /* identifier.  This is used to prevent calls to OSP term on this thread   */
  /* (which would result in a hang)                                          */
  /***************************************************************************/
  if (pthread_setspecific(sm_thread_key, sm_key_name) != 0)
  {
    TRC_ERROR((TB, "Failed to set thread specific data: quitting"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* If we're not listening on a socket, the FD will be zero so this is OK.  */
  /***************************************************************************/
  nfds = OSP_MAX(smg_listener_info.listener_termination_pipe_out,
                 smg_listener_info.secure_listening_socket);
  nfds = OSP_MAX(nfds, smg_listener_info.non_secure_listening_socket);
  nfds++;

  while (TRUE)
  {
    FD_ZERO(&read_set);

    if (smg_listener_info.non_secure_listening_socket != SM_INVALID_SOCKET)
    {
      FD_SET(smg_listener_info.non_secure_listening_socket, &read_set);
    }
    if (smg_listener_info.secure_listening_socket != SM_INVALID_SOCKET)
    {
      FD_SET(smg_listener_info.secure_listening_socket, &read_set);
    }
    FD_SET(smg_listener_info.listener_termination_pipe_out, &read_set);

    error = select(nfds, &read_set, NULL, NULL, NULL);
    if (error == -1)
    {
      TRC_OS_ERROR((TB, "select() error on listener thread"));
      goto EXIT_LABEL;
    }

    if (FD_ISSET(smg_listener_info.listener_termination_pipe_out, &read_set))
    {
      /***********************************************************************/
      /* Time to terminate everything.  First thing to do is stop listening  */
      /* for new connections.                                                */
      /***********************************************************************/
      TRC_NORMAL((TB, "Start terminating the stack"));

      if (smg_listener_info.non_secure_listening_socket != SM_INVALID_SOCKET)
      {
        TRC_DETAIL((TB, "Closing non secure socket"));
        close(smg_listener_info.non_secure_listening_socket);
        smg_listener_info.non_secure_listening_socket = SM_INVALID_SOCKET;
      }

      if (smg_listener_info.secure_listening_socket != SM_INVALID_SOCKET)
      {
        TRC_DETAIL((TB, "Closing secure socket"));
        close(smg_listener_info.secure_listening_socket);
        smg_listener_info.secure_listening_socket = SM_INVALID_SOCKET;
      }

      /***********************************************************************/
      /* Tell all of the worker threads to shut down.                        */
      /***********************************************************************/
      OSP_GET_LOCK(&smg_listener_info.worker_list_mutex);

      for (worker =
              (THREAD_INFO *)LIST_GET_FIRST(&smg_listener_info.worker_threads);
           worker != NULL;
           worker = (THREAD_INFO *)
              LIST_GET_NEXT(&smg_listener_info.worker_threads, &worker->links))
      {
        TRC_DETAIL((TB, "Close worker %p pipe", worker));
        close(worker->notification_pipe_in);
      }

      /***********************************************************************/
      /* Now wait for them to go away.                                       */
      /***********************************************************************/
      while (!LIST_IS_EMPTY(&smg_listener_info.worker_threads))
      {
        OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);
        TRC_DETAIL((TB, "Sleep for %d microseconds", SM_SLEEP_WAIT_INTERVAL));
        usleep(SM_SLEEP_WAIT_INTERVAL);
        OSP_GET_LOCK(&smg_listener_info.worker_list_mutex);
      }
      OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);

      /***********************************************************************/
      /* Now all of the workers have terminated, we can continue with the    */
      /* stack shutdown.                                                     */
      /***********************************************************************/
      sm_internal_term();

      /***********************************************************************/
      /* Now close the termination pipe which tells the calling thread that  */
      /* we are done.                                                        */
      /***********************************************************************/
      error = close(smg_listener_info.listener_termination_pipe_out);
      if (error == -1)
      {
        TRC_OS_ERROR((TB, "close() error on listener thread"));
        goto EXIT_LABEL;
      }

      TRC_NORMAL((TB, "Listener thread exiting normally"));
      goto EXIT_LABEL;
    }
    else
    {
      /***********************************************************************/
      /* Incoming socket connection.                                         */
      /***********************************************************************/
      TRC_DETAIL((TB, "Incoming socket connection"));

      if ((smg_listener_info.non_secure_listening_socket !=
                                                          SM_INVALID_SOCKET) &&
          (FD_ISSET(smg_listener_info.non_secure_listening_socket, &read_set)))
      {
        /*********************************************************************/
        /* Handle non-secure connection                                      */
        /*********************************************************************/
        sm_listener_handle_new_connection(
          smg_listener_info.non_secure_listening_socket, SM_NON_SECURE_SOCKET);
      }

      if ((smg_listener_info.secure_listening_socket != SM_INVALID_SOCKET) &&
          (FD_ISSET(smg_listener_info.secure_listening_socket, &read_set)))
      {
        /*********************************************************************/
        /* Handle secure connection                                          */
        /*********************************************************************/
        sm_listener_handle_new_connection(
                  smg_listener_info.secure_listening_socket, SM_SECURE_SOCKET);
      }
    }
  }

EXIT_LABEL:
  /***************************************************************************/
  /* Clean up OpenSSL per-thread state.                                      */
  /***************************************************************************/
  ERR_remove_state(0);

  TRC_NORMAL((TB, "Listener thread exiting"));

  TRC_THREAD_EXIT();

  return(NULL);

} /* sm_listener_thread */

/**PROC+**********************************************************************/
/* Name:      sm_listener_handle_new_connection                              */
/*                                                                           */
/* Purpose:   Handles an incoming connection                                 */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN   accept_socket - socket on which to issue accept()         */
/*            IN   socket_type   - SM_SECURE_SOCKET or                       */
/*                                 SM_NON_SECURE_SOCKET                      */
/*                                                                           */
/* Operation: accept() the connection and intialize the relevant data        */
/*            structures.  Find a worker that can handle this connection     */
/*            and, if necessary, create a new worker to handle further       */
/*            connections.                                                   */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_listener_handle_new_connection(int accept_socket, int socket_type)

{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CONN_INFO *new_connection;
  int new_socket;
  struct sockaddr_in remote_address;
  int address_len;
  int file_flags;
  THREAD_INFO *worker;
  int error;

  TRC_ENTRY("sm_listener_handle_new_connection");

  address_len = sizeof(remote_address);
  new_socket = accept(accept_socket,
                      (struct sockaddr *)&remote_address,
                      &address_len);

  if (new_socket == SM_INVALID_SOCKET)
  {
    TRC_OS_ERROR((TB, "Failed to accept connection"));
    goto EXIT_LABEL;
  }

  TRC_ASSERT((new_socket < FD_SETSIZE),
        (TB, "FD too large for select call: %d < %d", FD_SETSIZE, new_socket));

  TRC_NORMAL((TB, "Accepted %s connection (handle %d) from %s (%08x)",
                 ((socket_type == SM_SECURE_SOCKET) ? "secure" : "non-secure"),
                 new_socket,
                 inet_ntoa(remote_address.sin_addr),
                 remote_address.sin_addr.S_un.S_addr));

  new_connection = rm_get_structure(sizeof(CONN_INFO), RM_CONN_INFO);
  worker = sm_listener_find_free_worker_thread();

  if ((new_connection == NULL) ||
      (worker == NULL))
  {
    /*************************************************************************/
    /* Can't handle this new connection                                      */
    /*************************************************************************/
    TRC_ERROR((TB, "Can't handle new connection"));
    close(new_socket);

    if (new_connection)
    {
      rm_release_structure(new_connection, RM_CONN_INFO);
    }
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the socket to non-blocking                                          */
  /***************************************************************************/
  file_flags = fcntl(new_socket, F_GETFL, 0);

  TRC_DETAIL((TB, "Original file settings %x", file_flags));

  error = fcntl(new_socket, F_SETFL, file_flags | FNDELAY);

  if (error == -1)
  {
    TRC_OS_ERROR((TB, "Failed to set socket to non-blocking"));
    close(new_socket);
    new_socket = SM_INVALID_SOCKET;
    rm_release_structure(new_connection, RM_CONN_INFO);
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Fill in the connection details.                                         */
  /***************************************************************************/
  memset(&new_connection->links, 0, sizeof(new_connection->links));
  new_connection->ssl = NULL;
  new_connection->current_trans = NULL;
  new_connection->next_renegotiate = COM_RENEGOTIATE_NOT_SET;
  new_connection->conn_state = SM_CONN_STATE_OPEN;
  new_connection->http_version = SM_HTTP_VERSION_UNDETECTED;
  new_connection->persistent = FALSE;
  new_connection->ssl_data_buffered = FALSE;
  new_connection->last_assigned_serial = 0;
  new_connection->last_response_serial = 0;
  new_connection->called_cv_cb = 0;
  new_connection->client_id.ip_address = remote_address.sin_addr.S_un.S_addr;
  new_connection->client_id.ver_cookie = NULL;
  new_connection->thread = worker;
  new_connection->sock_handle = new_socket;
  new_connection->type = socket_type;
  new_connection->wa_locked_outstanding_count = 0;
  LIST_INIT(&new_connection->wa_locked_send_queue);

  sm_listener_hand_connection_to_worker(new_connection, worker);

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* sm_listener_handle_new_connection */

/**PROC+**********************************************************************/
/* Name:      sm_listener_create_worker_thread                               */
/*                                                                           */
/* Purpose:   Called on listener thread to create a new worker thread        */
/*                                                                           */
/* Returns:   OSP_SUCCESS     - completed successfully                       */
/*            OSP_ERROR       - operation failed                             */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Allocate the data structure and create the pipes and mutexes   */
/*            for this thread.                                               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sm_listener_create_worker_thread(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  int error;
  THREAD_INFO *new_worker = NULL;
  int pipe_fds[2];

  TRC_ENTRY("sm_listener_create_worker_thread");

  rc = rm_request_thread_create(RM_WORKER_THREAD);

  if (rc != OSP_SUCCESS)
  {
    TRC_ALERT((TB, "Not allowed to create a new thread"));
    goto EXIT_LABEL;
  }

  new_worker = rm_get_structure(sizeof(THREAD_INFO), RM_THREAD_INFO);

  if (new_worker == NULL)
  {
    TRC_ERROR((TB, "Failed to get memory for THREAD_INFO"));
    rm_notify_thread_exit(RM_WORKER_THREAD);
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize the data structure                                           */
  /***************************************************************************/
  memset(&new_worker->links, 0, sizeof(new_worker->links));
  new_worker->notification_pipe_in = SM_INVALID_SOCKET;
  new_worker->thread_handle = 0;
  new_worker->wakeup_pipe_out = SM_INVALID_SOCKET;
  new_worker->notification_pipe_out = SM_INVALID_SOCKET;
  LIST_INIT(&new_worker->conn_list);
  FD_ZERO(&new_worker->read_set);
  FD_ZERO(&new_worker->error_set);
  new_worker->new_connection_length = 0;
  new_worker->first_renegotiate = COM_RENEGOTIATE_NOT_SET;
  new_worker->nfds = 0;
  new_worker->ssl_data_buffered_count = 0;
  new_worker->parser = NULL;

  /***************************************************************************/
  /* Initialize the lw_locked sub structure                                  */
  /***************************************************************************/
  new_worker->lw_locked.assigned_conns = 0;
  memset(&new_worker->lw_locked.mutex_lw, 0,
                                       sizeof(new_worker->lw_locked.mutex_lw));

  /***************************************************************************/
  /* Initialize the wa_locked sub structure                                  */
  /***************************************************************************/
  memset(&new_worker->wa_locked.mutex_wa, 0,
                                       sizeof(new_worker->wa_locked.mutex_wa));
  FD_ZERO(&new_worker->wa_locked.write_set);
  new_worker->wa_locked.sent_wakeup = 0;
  new_worker->wa_locked.wakeup_pipe_in = SM_INVALID_SOCKET;
  new_worker->wa_locked.cached_time = 0;
  new_worker->wa_locked.cached_time_string[0] = 0;

  /***************************************************************************/
  /* Create the two communication pipes                                      */
  /***************************************************************************/
  error = pipe(pipe_fds);

  if (error != 0)
  {
    TRC_OS_ERROR((TB, "Error creating worker npipe"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  new_worker->notification_pipe_in = pipe_fds[0];
  new_worker->notification_pipe_out = pipe_fds[1];

  error = pipe(pipe_fds);

  if (error != 0)
  {
    TRC_OS_ERROR((TB, "Error creating worker wpipe"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  new_worker->wa_locked.wakeup_pipe_in = pipe_fds[0];
  new_worker->wakeup_pipe_out = pipe_fds[1];

  /***************************************************************************/
  /* Set up the worker's select structures to listen for reads on the wakeup */
  /* and notification pipes.                                                 */
  /***************************************************************************/
  FD_ZERO(&new_worker->read_set);
  FD_SET(new_worker->wakeup_pipe_out, &new_worker->read_set);
  FD_SET(new_worker->notification_pipe_out, &new_worker->read_set);

  FD_ZERO(&new_worker->error_set);
  FD_SET(new_worker->notification_pipe_out, &new_worker->error_set);

  /***************************************************************************/
  /* Initially the worker is only interested in notification_pipe_out and    */
  /* wakeup_pipe_out.                                                        */
  /***************************************************************************/
  SM_RECALCULATE_NFDS(new_worker);

  /***************************************************************************/
  /* Initialize busy-ness information for the new worker thread.  Make it    */
  /* look initially as though the thread is 0% busy.                         */
  /***************************************************************************/
  new_worker->select_start_time.tv_sec = 0;
  new_worker->select_start_time.tv_usec = 0;
  gettimeofday(&(new_worker->select_return_time), NULL);
  new_worker->lw_locked.busyness = 0;

  /***************************************************************************/
  /* Create the two mutexes.                                                 */
  /***************************************************************************/
  if (osp_create_mutex(&new_worker->lw_locked.mutex_lw) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create mutex_lw"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  if (osp_create_mutex(&new_worker->wa_locked.mutex_wa) != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create mutex_wa"));
    osp_destroy_mutex(&new_worker->lw_locked.mutex_lw);
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  OSP_GET_LOCK(&smg_listener_info.worker_list_mutex);

  LIST_INSERT_TAIL(&smg_listener_info.worker_threads, &new_worker->links);

  error = pthread_create(&new_worker->thread_handle, NULL, sm_worker_thread,
                                                                   new_worker);

  if (error != 0)
  {
    /*************************************************************************/
    /* Failed to create the thread.  Clear up the bits we've done so far.    */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to create worker %d", error));
    osp_destroy_mutex(&new_worker->wa_locked.mutex_wa);
    osp_destroy_mutex(&new_worker->lw_locked.mutex_lw);
    LIST_REMOVE_ITEM(&new_worker->links);
    OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);

    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Created thread ID:%d, Ptr:%p", new_worker->thread_handle,
                                                                  new_worker));

  pthread_detach(new_worker->thread_handle);

  OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);

  rc = OSP_SUCCESS;

EXIT_LABEL:

  if ((rc != OSP_SUCCESS) &&
      (new_worker != NULL))
  {
    /*************************************************************************/
    /* Clear up after failure                                                */
    /* First close the pipes                                                 */
    /*************************************************************************/
    if (new_worker->wa_locked.wakeup_pipe_in != SM_INVALID_SOCKET)
    {
      TRC_DETAIL((TB, "Close wakeup_pipe_in"));
      close(new_worker->wa_locked.wakeup_pipe_in);
    }

    if (new_worker->wakeup_pipe_out != SM_INVALID_SOCKET)
    {
      TRC_DETAIL((TB, "Close wakeup_pipe_out"));
      close(new_worker->wakeup_pipe_out);
    }

    if (new_worker->notification_pipe_in != SM_INVALID_SOCKET)
    {
      TRC_DETAIL((TB, "Close notification_pipe_in"));
      close(new_worker->notification_pipe_in);
    }

    if (new_worker->notification_pipe_out != SM_INVALID_SOCKET)
    {
      TRC_DETAIL((TB, "Close notification_pipe_out"));
      close(new_worker->notification_pipe_out);
    }

    /*************************************************************************/
    /* Now actually delete the data structure.                               */
    /*************************************************************************/
    rm_release_structure(new_worker, RM_THREAD_INFO);

    /*************************************************************************/
    /* Tell SM that the thread has gone.                                     */
    /*************************************************************************/
    rm_notify_thread_exit(RM_WORKER_THREAD);
  }

  TRC_EXIT();

  return(rc);

} /* sm_listener_create_worker_thread */

/**PROC+**********************************************************************/
/* Name:      sm_listener_find_free_worker_thread                            */
/*                                                                           */
/* Purpose:   Finds a worker thread that can handle another connection.      */
/*                                                                           */
/* Returns:   THREAD_INFO pointer of worker thread or NULL if there are no   */
/*            free worker threads.                                           */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Find the first worker thread that is not operating at          */
/*            capacity.                                                      */
/*                                                                           */
/**PROC-**********************************************************************/

THREAD_INFO *sm_listener_find_free_worker_thread(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  int min_busyness;
  int min_connections;
  THREAD_INFO *least_busy_worker;
  THREAD_INFO *worker;

  TRC_ENTRY("sm_listener_find_free_worker_thread");

  /***************************************************************************/
  /* Each SM worker thread maintains a busyness measure, which describes as  */
  /* a percentage how much of the time it has recently spent running, as     */
  /* opposed to waiting for select to pop.  The algorithm implemented here   */
  /* is to create a fixed number of worker threads at start of day, based on */
  /* the number of available processors and a configurable multiplier, and   */
  /* to allocate each new connection to the worker thread that is currently  */
  /* least busy.                                                             */
  /*                                                                         */
  /* Rationale:                                                              */
  /*                                                                         */
  /* As the number of connections increases, it is better to place a bound   */
  /* upon the number of threads than upon the number of connections per      */
  /* thread, since (i) each new thread uses up memory for its stack; (ii) we */
  /* have more control over the fairness and latency of scheduling within    */
  /* each worker thread than we do over the OS's scheduling between worker   */
  /* threads.                                                                */
  /*                                                                         */
  /* On the other hand, a certain amount of blocking within worker thread    */
  /* processing is inevitable, and we can only use this blockage time for    */
  /* processing other connections by making use of OS threads.  It's simpler */
  /* and more predictable to create all such OS threads at start of day than */
  /* to attempt a dynamic mechanism which creates threads at times of high   */
  /* load and destroys them again when the load eases.                       */
  /***************************************************************************/

  /***************************************************************************/
  /* Spin through the threads to establish which is the least busy.          */
  /***************************************************************************/
  min_busyness = 101;
  least_busy_worker = NULL;

  OSP_GET_LOCK(&smg_listener_info.worker_list_mutex);

  for (worker = (THREAD_INFO *)
                             LIST_GET_FIRST(&smg_listener_info.worker_threads);
       worker != NULL;
       worker = (THREAD_INFO *)
              LIST_GET_NEXT(&smg_listener_info.worker_threads, &worker->links))
  {
    /*************************************************************************/
    /* Is this worker full?                                                  */
    /*************************************************************************/
    OSP_GET_LOCK(&worker->lw_locked.mutex_lw);

    TRC_DETAIL((TB, "Worker %p has %d connections, busyness %d%%",
                worker,
                worker->lw_locked.assigned_conns,
                worker->lw_locked.busyness));

    if (((worker->lw_locked.busyness < min_busyness) ||
         ((worker->lw_locked.busyness == min_busyness) &&
          (worker->lw_locked.assigned_conns < min_connections))) &&
        (worker->notification_pipe_in != SM_INVALID_SOCKET))
    {
      TRC_DETAIL((TB, "Found less busy thread %p", worker));
      min_busyness = worker->lw_locked.busyness;
      min_connections = worker->lw_locked.assigned_conns;
      least_busy_worker = worker;
    }

    OSP_RELEASE_LOCK(&worker->lw_locked.mutex_lw);
  }

  OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);

  /***************************************************************************/
  /* If we're here, and least_busy_worker is still NULL, then this server as */
  /* a whole isn't going to be very effective.                               */
  /***************************************************************************/
  TRC_ASSERT((least_busy_worker != NULL), (TB, "No worker threads available"));

  TRC_EXIT();

  return(least_busy_worker);

} /* sm_listener_find_free_worker_thread */

/**PROC+**********************************************************************/
/* Name:      sm_listener_hand_connection_to_worker                          */
/*                                                                           */
/* Purpose:   Hands new connection to the worker that will service it.       */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN    conn   - connection to hand off.                         */
/*            IN    worker - thread that will service the connection.        */
/*                                                                           */
/* Operation: Write the CONN_INFO pointer down the worker's notification     */
/*            pipe.                                                          */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_listener_hand_connection_to_worker(CONN_INFO *conn,
                                           THREAD_INFO *worker)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  int error;
  int written;

  TRC_ENTRY("sm_listener_hand_connection_to_worker");

  OSP_GET_LOCK(&smg_listener_info.worker_list_mutex);

  if (worker->notification_pipe_in != SM_INVALID_SOCKET)
  {
    /*************************************************************************/
    /* Notification pipe is still valid so send the connection pointer to    */
    /* the worker.                                                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Sending conn %p to worker %p", conn, worker));
    written = write(worker->notification_pipe_in, &conn, sizeof(conn));

    TRC_ASSERT((written == sizeof(conn)),
                    (TB, "Partial write on notification_pipe_in %d", written));

    OSP_GET_LOCK(&worker->lw_locked.mutex_lw);
    worker->lw_locked.assigned_conns++;
    OSP_RELEASE_LOCK(&worker->lw_locked.mutex_lw);
  }
  else
  {
    TRC_ALERT((TB, "Can't send conn %p to worker %p: invalid pipe handle",
                                                                conn, worker));
  }

  OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* sm_listener_hand_connection_to_worker */

