
/**MOD+***********************************************************************/
/* Module:   smmain.c                                                        */
/*                                                                           */
/* Purpose:  Implementation of SM's externally visible functions             */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.49               $ $Modtime::   Jan 11 2001 15:56:46   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <smincl.h>

/*****************************************************************************/
/* Initialization progress of SM                                             */
/*****************************************************************************/
static int progress = SM_INIT_NONE;

/**PROC+**********************************************************************/
/* Name:      sm_init                                                        */
/*                                                                           */
/* Purpose:   Initialize SM                                                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS - completed successfully                           */
/*            OSP_ERROR   - failed to initialize                             */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Initialize all memory and mutexes.  Then create a worker       */
/*            thread and the listener thread.                                */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sm_init(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int error;
  int pipe_fds[2];
  pthread_t tid;
  OSP_RC rc;
  struct rlimit limits;
  int ithread;

  TRC_ENTRY("sm_init");

  /***************************************************************************/
  /* Check that buffer size is OK before anything else                       */
  /***************************************************************************/
  TRC_ASSERT((SM_SEND_BUFFER_SIZE >= sizeof(sm_const_canned_400_10)),
             (TB, "send buffer size too small. %d < %d",
                         SM_SEND_BUFFER_SIZE, sizeof(sm_const_canned_400_10)));
  TRC_ASSERT((SM_SEND_BUFFER_SIZE >= sizeof(sm_const_canned_500_10)),
             (TB, "send buffer size too small. %d < %d",
                         SM_SEND_BUFFER_SIZE, sizeof(sm_const_canned_500_10)));
  TRC_ASSERT((SM_SEND_BUFFER_SIZE >= sizeof(sm_const_canned_501_10)),
             (TB, "send buffer size too small. %d < %d",
                         SM_SEND_BUFFER_SIZE, sizeof(sm_const_canned_501_10)));
  TRC_ASSERT((SM_SEND_BUFFER_SIZE >= sizeof(sm_const_canned_400_11)),
             (TB, "send buffer size too small. %d < %d",
                         SM_SEND_BUFFER_SIZE, sizeof(sm_const_canned_400_11)));
  TRC_ASSERT((SM_SEND_BUFFER_SIZE >= sizeof(sm_const_canned_500_11)),
             (TB, "send buffer size too small. %d < %d",
                         SM_SEND_BUFFER_SIZE, sizeof(sm_const_canned_500_11)));
  TRC_ASSERT((SM_SEND_BUFFER_SIZE >= sizeof(sm_const_canned_501_11)),
             (TB, "send buffer size too small. %d < %d",
                         SM_SEND_BUFFER_SIZE, sizeof(sm_const_canned_501_11)));

  /***************************************************************************/
  /* Bump the OS file descriptor limit so we are not constrained by the      */
  /* number of file descriptors per process.  Note there is still a limit of */
  /* file descriptors that can be selected on FD_SETSIZE.                    */
  /***************************************************************************/
  limits.rlim_cur = RLIM_INFINITY;
  limits.rlim_max = RLIM_INFINITY;

#ifndef OSP_SUPPORT_PURIFY
  /***************************************************************************/
  /* Disable this code if building with Purify as it causes an assertion     */
  /* failure.                                                                */
  /***************************************************************************/
  error = setrlimit(RLIMIT_NOFILE, &limits);

  if (error != 0)
  {
    /*************************************************************************/
    /* setrlimit failed                                                      */
    /*************************************************************************/
    TRC_OS_ERROR((TB, "Failed to set limit on file descriptors"));
    goto EXIT_LABEL;
  }
#endif

  /***************************************************************************/
  /* First intialize the SM global structure.                                */
  /***************************************************************************/
  smg_listener_info.ssl_method = NULL;
  smg_listener_info.ssl_ctx = NULL;
  memset(smg_listener_info.ssl_locks, 0, sizeof(smg_listener_info.ssl_locks));
  memset(&smg_listener_info.worker_list_mutex, 0,
                                  sizeof(smg_listener_info.worker_list_mutex));
  LIST_INIT(&smg_listener_info.worker_threads);
  smg_listener_info.secure_listening_socket = SM_INVALID_SOCKET;
  smg_listener_info.non_secure_listening_socket = SM_INVALID_SOCKET;
  smg_listener_info.listener_termination_pipe_out = SM_INVALID_SOCKET;
  FD_ZERO(&smg_listener_info.listening_socket_handles);
  smg_listener_info.listener_termination_pipe_in = SM_INVALID_SOCKET;
  memset(&smg_listener_info.callback_mutex, 0,
                                     sizeof(smg_listener_info.callback_mutex));
  smg_listener_info.cv_cb = NULL;
  smg_listener_info.cd_cb = NULL;

  /***************************************************************************/
  /* Find out how many processors are running on this machine.               */
  /***************************************************************************/
  smg_listener_info.processor_count = sysconf(_SC_NPROCESSORS_ONLN);
  TRC_DETAIL((TB, "Online processor count: %d",
              smg_listener_info.processor_count));

  /***************************************************************************/
  /* Now initialize SSL, which we need to do before CCM.                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Initializing SSL"));

  rc = sm_init_ssl();

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to init SSL"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "SSL initialized"));
  progress = SM_INIT_SSL;

  /***************************************************************************/
  /* Initialize CCM.                                                         */
  /***************************************************************************/
  rc = ccm_init(smg_listener_info.ssl_ctx, sm_ssl_verify_callback);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize CCM"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "CCM initialized"));
  progress = SM_INIT_CCM;

  /***************************************************************************/
  /* Initialize all of the mutexes.  The callback mutex is recursive to      */
  /* allow a callback function to deregister itself.                         */
  /***************************************************************************/
  rc = osp_create_recursive_mutex(&smg_listener_info.callback_mutex);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create callback mutex %d", rc));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Created callback mutex"));
  progress = SM_INIT_CB_MUTEX;

  rc = osp_create_mutex(&smg_listener_info.worker_list_mutex);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create worker list mutex %d", rc));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Created worker list mutex"));
  progress = SM_INIT_WL_MUTEX;

  /***************************************************************************/
  /* Create listener termination pipe.                                       */
  /***************************************************************************/
  error = pipe(pipe_fds);

  if (error != 0)
  {
    TRC_OS_ERROR((TB, "Error creating listener termination pipe"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  smg_listener_info.listener_termination_pipe_in = pipe_fds[0];
  smg_listener_info.listener_termination_pipe_out = pipe_fds[1];

  TRC_DETAIL((TB, "Created termination pipe"));
  progress = SM_INIT_PIPE;

  /***************************************************************************/
  /* Create listening sockets.                                               */
  /***************************************************************************/
  if (smg_config_data.non_secure_listening_port != 0)
  {
    /*************************************************************************/
    /* Non-secure port configured - create non-secure listening socket.      */
    /*************************************************************************/
    TRC_DETAIL((TB, "Creating non secure listening socket"));
    smg_listener_info.non_secure_listening_socket =
                   sm_create_socket(smg_config_data.non_secure_listening_port);

    if (smg_listener_info.non_secure_listening_socket == SM_INVALID_SOCKET)
    {
      TRC_ERROR((TB, "Failed to create non-secure listening socket"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
  }

  progress = SM_INIT_NON_SECURE_SOCK;

  if (smg_config_data.secure_listening_port != 0)
  {
    /*************************************************************************/
    /* Secure port configured - create secure listening socket               */
    /*************************************************************************/
    TRC_DETAIL((TB, "Creating secure listening socket"));

    smg_listener_info.secure_listening_socket =
                       sm_create_socket(smg_config_data.secure_listening_port);

    if (smg_listener_info.secure_listening_socket == SM_INVALID_SOCKET)
    {
      TRC_ERROR((TB, "Failed to create secure listening socket"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
  }

  TRC_DETAIL((TB, "Created listening socket(s)"));
  progress = SM_INIT_SOCKETS;

  /***************************************************************************/
  /* Create the thread key we will use to identify threads as ours           */
  /***************************************************************************/
  if (pthread_key_create(&sm_thread_key, NULL) != 0)
  {
    TRC_ERROR((TB, "Failed to create sm thread key"));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create the listener thread.                                             */
  /***************************************************************************/
  error = pthread_create(&tid, NULL, sm_listener_thread, NULL);

  if (error != 0)
  {
    TRC_ERROR((TB, "Failed to create listener thread - error %d", error));
    rc = OSP_ERROR;
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Created listener ID:%d", tid));

  pthread_detach(tid);

  progress = SM_INIT_FULL;

  /***************************************************************************/
  /* Create worker threads.                                                  */
  /***************************************************************************/
  for (ithread = 0;
       ithread < smg_listener_info.processor_count *
                                         smg_config_data.threads_per_processor;
       ithread++)
  {
    TRC_NORMAL((TB, "Create worker thread #%d", ithread));
    rc = sm_listener_create_worker_thread();

    if (rc != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to create worker thread #%d: rc=%d",
                 ithread,
                 rc));
      goto EXIT_LABEL;
    }
  }

  rc = OSP_SUCCESS;
  TRC_NORMAL((TB, "SM is Initialized"));

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* Failed to initialize SM - call sm_term to shut everything down        */
    /*************************************************************************/
    TRC_DETAIL((TB, "Failed to init SM"));
    sm_term();
  }

  TRC_EXIT();

  return(rc);

} /* sm_init */

/**PROC+**********************************************************************/
/* Name:      sm_term                                                        */
/*                                                                           */
/* Purpose:   Terminates SM                                                  */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    None.                                                          */
/*                                                                           */
/* Operation: Writes a byte down the listener's termination pipe.  Then wait */
/*            until the listener closes the pipe which indicates that the    */
/*            stack is terminated.                                           */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int error;
  fd_set error_set;
  fd_set read_set;
  struct timeval timeout;

  TRC_ENTRY("sm_term");

  /***************************************************************************/
  /* If SM has been fully initialized, we need to tell the listener thread   */
  /* to shut everything down and wait until it has finished.  Note that this */
  /* includes the case where one or more of the worker threads could not be  */
  /* created by sm_init(), since the listener thread will automatically      */
  /* terminate however many worker threads do exist.                         */
  /***************************************************************************/
  if (progress == SM_INIT_FULL)
  {
    /*************************************************************************/
    /* Write a byte down the termination pipe.  The actual data is           */
    /* irrelevant, but we set a value so as to avoid an uninitialized memory */
    /* read error from Purify.                                               */
    /*************************************************************************/
    TRC_NORMAL((TB, "SM Fully initialized - tell listener to shut down"));

    error = 0;
    error = write(smg_listener_info.listener_termination_pipe_in, &error, 1);

    if (error != 1)
    {
      TRC_OS_ERROR((TB, "Failed to write down termination pipe"));
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Now we give the listener a few seconds to terminate everything.  This */
    /* is communicated by the listener terminating the termination pipe.     */
    /*************************************************************************/
    FD_ZERO(&error_set);
    FD_SET(smg_listener_info.listener_termination_pipe_in, &error_set);
    FD_ZERO(&read_set);
    FD_SET(smg_listener_info.listener_termination_pipe_in, &read_set);
    timeout.tv_usec = 0;
    timeout.tv_sec = SM_SELECT_TIMEOUT;

    error = select(smg_listener_info.listener_termination_pipe_in + 1,
                   &read_set,
                   NULL,
                   &error_set,
                   &timeout);

    TRC_DETAIL((TB, "Finished waiting for listener to close"));
    if (error == -1)
    {
      TRC_OS_ERROR((TB, "select() error on termination"));
    }
    else if (error == 0)
    {
      TRC_ALERT((TB, "Select returned after timeout"));
    }
    else
    {
      TRC_DETAIL((TB, "Select returned so the pipe must have shut down"));
    }

    /*************************************************************************/
    /* Close the pipe                                                        */
    /*************************************************************************/
    TRC_DETAIL((TB, "Closing term pipe"));
    error = close(smg_listener_info.listener_termination_pipe_in);
    if (error == -1)
    {
      TRC_OS_ERROR((TB, "close() error on listener termination thread"));
    }

    /*************************************************************************/
    /* Destroy the thread key                                                */
    /*************************************************************************/
    if (pthread_key_delete(sm_thread_key) != 0)
    {
      TRC_ERROR((TB, "Failed to delete sm thread key"));
    }
  }
  else
  {
    /*************************************************************************/
    /* Stack wasn't fully initialized.  Just terminate those bits of that    */
    /* were.                                                                 */
    /*************************************************************************/
    TRC_ALERT((TB, "Terminate part initialized stack"));

    /*************************************************************************/
    /* There are now no other OSP threads in existence.  Call                */
    /* sm_internal_term() to free everything else.                           */
    /*************************************************************************/
    sm_internal_term();
  }

  /***************************************************************************/
  /* We're terminated.                                                       */
  /***************************************************************************/
  TRC_NORMAL((TB, "OSP terminated."));

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* sm_term */

/**PROC+**********************************************************************/
/* Name:      sm_internal_term                                               */
/*                                                                           */
/* Purpose:   Terminates the SM internals.  Only called by other SM code.    */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Free everything.                                               */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_internal_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("sm_internal_term");

  /***************************************************************************/
  /* The drop thorugh case statements are deliberate.                        */
  /***************************************************************************/
  switch (progress)
  {
    case SM_INIT_SOCKETS:
      /***********************************************************************/
      /* Close sockets - In the case of SM_INIT_FULL, this will already have */
      /* been done.                                                          */
      /***********************************************************************/
      if (smg_listener_info.secure_listening_socket != SM_INVALID_SOCKET)
      {
        TRC_DETAIL((TB, "Closing secure socket"));
        close(smg_listener_info.secure_listening_socket);
        smg_listener_info.secure_listening_socket = SM_INVALID_SOCKET;
      }

    case SM_INIT_NON_SECURE_SOCK:
      if (smg_listener_info.non_secure_listening_socket != SM_INVALID_SOCKET)
      {
        TRC_DETAIL((TB, "Closing non secure socket"));
        close(smg_listener_info.non_secure_listening_socket);
        smg_listener_info.non_secure_listening_socket = SM_INVALID_SOCKET;
      }

    case SM_INIT_PIPE:
      /***********************************************************************/
      /* The out (listener thread) side of the termination pipe will be      */
      /* closed by our caller once all of sm_internal_term has completed.    */
      /***********************************************************************/

    case SM_INIT_FULL:
      /***********************************************************************/
      /* The in (main thread) side of the termination pipe will be closed by */
      /* the main thread when it detects that the listener thread end has    */
      /* been closed.                                                        */
      /***********************************************************************/

    case SM_INIT_WL_MUTEX:
      /***********************************************************************/
      /* Delete mutex                                                        */
      /***********************************************************************/
      TRC_DETAIL((TB, "Delete worker list mutex"));
      osp_destroy_mutex(&smg_listener_info.worker_list_mutex);

    case SM_INIT_CB_MUTEX:
      /***********************************************************************/
      /* Delete mutex                                                        */
      /***********************************************************************/
      TRC_DETAIL((TB, "Delete callback mutex"));
      osp_destroy_recursive_mutex(&smg_listener_info.callback_mutex);

    case SM_INIT_CCM:
      /***********************************************************************/
      /* Shutdown CCM                                                        */
      /***********************************************************************/
      TRC_DETAIL((TB, "Shutdown CCM"));
      ccm_term();

    case SM_INIT_SSL:
      /***********************************************************************/
      /* Shutdown SSL                                                        */
      /***********************************************************************/
      TRC_DETAIL((TB, "Shutdown SSL"));
      sm_term_ssl();
      break;

    case SM_INIT_NONE:
      /***********************************************************************/
      /* Nothing initialized                                                 */
      /***********************************************************************/
      TRC_DETAIL((TB, "Nothing initialized - nothing to terminate"));
      break;

    default:
      /***********************************************************************/
      /* Error                                                               */
      /***********************************************************************/
      TRC_ASSERT((FALSE), (TB, "illegal progress value %d", progress));
      break;
  }

  TRC_EXIT();

  return;

} /* sm_internal_term */

/**PROC+**********************************************************************/
/* Name:      sm_init_ssl                                                    */
/*                                                                           */
/* Purpose:   Initialize the SSL library                                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS - completed successfully                           */
/*            OSP_ERROR   - failed to initialize SSL                         */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Acquire the resources necessary for the SSL library.           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sm_init_ssl(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int ii;

  TRC_ENTRY("sm_init_ssl");

  /***************************************************************************/
  /* Initialize SSL library                                                  */
  /***************************************************************************/
  if (SSL_library_init() != 1)
  {
    TRC_ERROR((TB, "Failed to initialize SSL"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Load error strings so that we get meaningful text on error              */
  /***************************************************************************/
  SSL_load_error_strings();

  /***************************************************************************/
  /* Create a number of locks for SSL to use                                 */
  /***************************************************************************/
  for (ii = 0; ii < CRYPTO_NUM_LOCKS; ii++)
  {
    if (osp_create_mutex(&smg_listener_info.ssl_locks[ii]) != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failed to create SSL mutex number %d", ii));
      goto EXIT_LABEL;
    }
  }

  /***************************************************************************/
  /* Tell the SSL library to use our locking and thread ID callbacks.        */
  /***************************************************************************/
  CRYPTO_set_locking_callback(sm_locking_callback);
  CRYPTO_set_id_callback(sm_get_thread_id);

  /***************************************************************************/
  /* Get pointers to the SSL functions.                                      */
  /***************************************************************************/
  smg_listener_info.ssl_method = SSLv23_server_method();

  if (smg_listener_info.ssl_method == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to get SSL_METHOD"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Get SSL context                                                         */
  /***************************************************************************/
  smg_listener_info.ssl_ctx = SSL_CTX_new(smg_listener_info.ssl_method);

  if (smg_listener_info.ssl_ctx == NULL)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to get SSL_CTX"));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "Created SSL_CTX at %p", smg_listener_info.ssl_ctx));

  /***************************************************************************/
  /* Disallow use of SSL V2                                                  */
  /***************************************************************************/
  if (SSL_CTX_set_options(smg_listener_info.ssl_ctx, SSL_OP_NO_SSLv2) !=
                                                               SSL_OP_NO_SSLv2)
  {
    OSP_OPENSSL_ERROR((TB, "Failed to set SSL_CTX options"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set the session cache size                                              */
  /***************************************************************************/
  SSL_CTX_sess_set_cache_size(smg_listener_info.ssl_ctx,
                              SM_SESSION_CACHE_SIZE);

  /***************************************************************************/
  /* Set the session ID context.  The exact value isn't important, but this  */
  /* call is required to allow session ID re-use when client verification is */
  /* enabled.                                                                */
  /***************************************************************************/
  SSL_CTX_set_session_id_context(smg_listener_info.ssl_ctx,
                                   (void *)&smg_listener_info, sizeof(void *));

  TRC_NORMAL((TB, "Initialized SSL library OK"));
  rc = OSP_SUCCESS;

EXIT_LABEL:

  if (rc != OSP_SUCCESS)
  {
    TRC_DETAIL((TB, "Failed SSL init: terminate SSL"));
    sm_term_ssl();
  }

  TRC_EXIT();

  return(rc);

} /* sm_init_ssl */

/**PROC+**********************************************************************/
/* Name:      sm_term_ssl                                                    */
/*                                                                           */
/* Purpose:   Terminate the SSL library                                      */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Free all of the resources allocated by SSL. Note that this is  */
/*            not the exact opposite of sm_init_ssl() due to the nature of   */
/*            the OpenSSL API.                                               */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_term_ssl(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int ii;

  TRC_ENTRY("sm_term_ssl");

  TRC_NORMAL((TB, "Freeing SSL_CTX"));
  SSL_CTX_free(smg_listener_info.ssl_ctx);
  smg_listener_info.ssl_ctx = NULL;

  TRC_NORMAL((TB, "Calling EVP_cleanup"));
  EVP_cleanup();

  TRC_NORMAL((TB, "Calling ERR_remove_state"));
  ERR_remove_state(0);

  TRC_NORMAL((TB, "Calling ERR_free_strings"));
  ERR_free_strings();

  TRC_NORMAL((TB, "Calling OBJ_NAME_cleanup"));
  OBJ_NAME_cleanup(-1);

  TRC_NORMAL((TB, "Destroying mutexes"));
  for (ii = 0; ii < CRYPTO_NUM_LOCKS; ii++)
  {
    osp_destroy_mutex(&(smg_listener_info.ssl_locks[ii]));
  }

  TRC_EXIT();

  return;

} /* sm_term_ssl */

/**PROC+**********************************************************************/
/* Name:      sm_send_response                                               */
/*                                                                           */
/* Purpose:   Send response to transaction.                                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERROR           - operation failed                         */
/*                                                                           */
/* Params:    IN/OUT transaction  - transaction being responded to           */
/*            IN     content_type - entire content type header (including    */
/*                                  ContentType =)                           */
/*            IN     response     - HTTP response code - SM_RESPONSE_200 for */
/*                                  success                                  */
/*                                                                           */
/* Operation: If this is an HTTP/1.0 or 1.1 response, fill in the            */
/*            Content-Type and Content-Length headers.                       */
/*                                                                           */
/*            Put the transaction on the send queue and wakeup the worker if */
/*            necessary.                                                     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC sm_send_response(TRANS_INFO *transaction, char *content_type,
                                                                  int response)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  int error;
  int room_needed = 0;
  char *current_ptr;
  struct timeval now;
  char content_length[7];
  int content_length_len;
  const char *conn_header;
  int conn_header_length;

  TRC_ENTRY("sm_send_response");

  TRC_ASSERT((transaction != NULL), (TB, "NULL transaction"));
  TRC_ASSERT((transaction->http_version != SM_HTTP_VERSION_UNDETECTED),
             (TB, "HTTP version undetected"));

#ifdef OSP_MEASURE_PERFORMANCE
  /***************************************************************************/
  /* Accumulate processing time statistics for this transaction.             */
  /***************************************************************************/
  gettimeofday(&now, NULL);
  TRC_DETAIL((TB, "Transaction processed: trans=%p time=%ld:%06ld",
              transaction,
              now.tv_sec,
              now.tv_usec));
  SM_ACCUMULATE(smg_processing, transaction->start_time_processing, now);
  SM_ACCUMULATE(smg_until_send_queue, transaction->start_time_io, now);
#endif

  if (response == SM_RESPONSE_200)
  {
    /*************************************************************************/
    /* OK response                                                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "200 Response"));

    /*************************************************************************/
    /* No extra formatting required for HTTP 0.9                             */
    /*************************************************************************/
    if (transaction->http_version != SM_HTTP_0_9)
    {
      /***********************************************************************/
      /* Now check that there is enough room for the headers - Content-Type, */
      /* Content-Length, Connection (if any) - and the response line, then   */
      /* fill them in.                                                       */
      /***********************************************************************/
      TRC_DETAIL((TB, "HTTP/%s", (transaction->http_version == SM_HTTP_1_0) ?
                                                              "1.0" : "1.1" ));
      OSP_GET_LOCK(&transaction->mutex);

      /***********************************************************************/
      /* Work out what, if any, connection header we should send             */
      /***********************************************************************/
      if ((transaction->http_version == SM_HTTP_1_1) &&
                                          !transaction->connection->persistent)
      {
        TRC_DETAIL((TB, "HTTP/1.1 non-persistent: send connection close"));
        conn_header = sm_const_conn_header_close;
        conn_header_length = sizeof(sm_const_conn_header_close);
        room_needed += conn_header_length;
        room_needed += 2;
      }
      else if ((transaction->http_version == SM_HTTP_1_0) &&
                                           transaction->connection->persistent)
      {
        TRC_DETAIL((TB, "HTTP/1.0 persistent: send connection keep alive"));
        conn_header = sm_const_conn_header_keep_alive;
        conn_header_length = sizeof(sm_const_conn_header_keep_alive);
        room_needed += conn_header_length;
        room_needed += 2;
      }
      else
      {
        TRC_DETAIL((TB, "No conn header"));
        conn_header_length = 0;
        conn_header = NULL;
      }

      /***********************************************************************/
      /* Content type header, space, value and CRLF if required              */
      /***********************************************************************/
      if (content_type != NULL)
      {
        TRC_DETAIL((TB, "Content type supplied: %s", content_type));
        room_needed += sizeof(sm_const_content_type);
        room_needed += 1;
        room_needed += strlen(content_type);
        room_needed += 2;
      }

      /***********************************************************************/
      /* Calculate number of digits required to express content length.      */
      /***********************************************************************/
      TRC_ASSERT((transaction->send_buffer_end -
                                     transaction->send_buffer_start <= 999999),
                 (TB, "send buffer too large %d",
                       transaction->send_buffer_end -
                                    transaction->send_buffer_start <= 999999));
      sprintf(content_length, "%d",
                transaction->send_buffer_end - transaction->send_buffer_start);
      content_length_len = strlen(content_length);
      room_needed += content_length_len;

      /***********************************************************************/
      /* Space for constant header elements:                                 */
      /*                                                                     */
      /* HTTP/1.X 200 OK[CRLF]                         17 bytes              */
      /* Content-Length: <content_length>[CRLFCRLF]    20 bytes              */
      /***********************************************************************/
      room_needed += 37;

      TRC_DETAIL((TB, "Room needed for header is %d", room_needed));

      /***********************************************************************/
      /* Check that the room needed for the header is less than the amount   */
      /* of space that is available.                                         */
      /***********************************************************************/
      TRC_ASSERT((room_needed < transaction->send_buffer_start),
              (TB, "Reserved header size insufficient: needed=%d available=%d",
                                 room_needed, transaction->send_buffer_start));

      /***********************************************************************/
      /* Now fill in buffer which looks like:                                */
      /* HTTP/1.0 200 OK[CRLF]                                               */
      /* Content-Type: blahblahblah[CRLF]                                    */
      /* Content-Length: 999[CRLF]                                           */
      /* Connection: [Close | Keep-Alive][CRLF]                              */
      /* [CRLF]                                                              */
      /* datadatadata                                                        */
      /***********************************************************************/
      current_ptr =
       transaction->send_buffer + transaction->send_buffer_start - room_needed;

      /***********************************************************************/
      /* Copy in HTTP/1.X                                                    */
      /***********************************************************************/
      if (transaction->http_version == SM_HTTP_1_0)
      {
        TRC_DETAIL((TB, "HTTP 1.0 response"));
        memcpy(current_ptr, sm_const_http_10_response,
                                            sizeof(sm_const_http_10_response));
        current_ptr += sizeof(sm_const_http_10_response);
      }
      else
      {
        TRC_DETAIL((TB, "HTTP 1.1 response"));
        memcpy(current_ptr, sm_const_http_11_response,
                                            sizeof(sm_const_http_11_response));
        current_ptr += sizeof(sm_const_http_11_response);
      }

      /***********************************************************************/
      /* Add CRLF at end of response line                                    */
      /***********************************************************************/
      memcpy(current_ptr, sm_const_crlf, sizeof(sm_const_crlf));
      current_ptr += sizeof(sm_const_crlf);

      if (content_type != NULL)
      {
        /*********************************************************************/
        /* Copy in Content-Type header, space, value and CRLF                */
        /*********************************************************************/
        TRC_DETAIL((TB, "Copying in content type"));
        memcpy(current_ptr, sm_const_content_type,
                                                sizeof(sm_const_content_type));
        current_ptr += sizeof(sm_const_content_type);
        *current_ptr = 0x20;
        current_ptr++;
        strcpy(current_ptr, content_type);
        current_ptr += strlen(content_type);
        memcpy(current_ptr, sm_const_crlf, sizeof(sm_const_crlf));
        current_ptr += sizeof(sm_const_crlf);
      }

      /***********************************************************************/
      /* Copy in Content-Length header, space, value and CRLF                */
      /***********************************************************************/
      TRC_DETAIL((TB, "Copy content length"));
      memcpy(current_ptr, sm_const_content_length,
                                             sizeof(sm_const_content_length));
      current_ptr += sizeof(sm_const_content_length);
      *current_ptr = 0x20;
      current_ptr++;

      memcpy(current_ptr, content_length, content_length_len);
      current_ptr += content_length_len;

      memcpy(current_ptr, sm_const_crlf, sizeof(sm_const_crlf));
      current_ptr += sizeof(sm_const_crlf);

      /***********************************************************************/
      /* Add any connection header plus CRLF                                 */
      /***********************************************************************/
      if (conn_header != NULL)
      {
        TRC_DETAIL((TB, "Copy connection header"));
        memcpy(current_ptr, conn_header, conn_header_length);
        current_ptr += conn_header_length;
        memcpy(current_ptr, sm_const_crlf, sizeof(sm_const_crlf));
        current_ptr += sizeof(sm_const_crlf);
      }

      /***********************************************************************/
      /* Add extra CRLF to signify end of headers                            */
      /***********************************************************************/
      memcpy(current_ptr, sm_const_crlf, sizeof(sm_const_crlf));
      current_ptr += sizeof(sm_const_crlf);

      TRC_ASSERT((current_ptr == transaction->send_buffer +
       transaction->send_buffer_start), (TB, "Internal error: %d, %d, %d",
       current_ptr, transaction->send_buffer, transaction->send_buffer_start));

      transaction->send_buffer_start -= room_needed;
      TRC_DATA("Headers:",
               (transaction->send_buffer + transaction->send_buffer_start),
               room_needed);

      OSP_RELEASE_LOCK(&transaction->mutex);
    }

    /*************************************************************************/
    /* Queue the transaction.                                                */
    /*************************************************************************/
    rc = sm_queue_transaction(transaction, TRUE);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* Not much we can do but decrement the outstanding count on the       */
      /* connection and free the transaction                                 */
      /***********************************************************************/
      TRC_ALERT((TB, "Failed to queue transaction: rc=%d", rc));
      OSP_GET_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);
      transaction->connection->wa_locked_outstanding_count--;
      OSP_RELEASE_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);
      TRC_DETAIL((TB, "Freeing transaction after send failure"));
      sm_free_transaction(transaction, TRUE);
      goto EXIT_LABEL;
    }
  }
  else
  {
    /*************************************************************************/
    /* Error response                                                        */
    /*************************************************************************/
    TRC_ALERT((TB, "Error response %d", response));

    if (transaction->http_version == SM_HTTP_0_9)
    {
      /***********************************************************************/
      /* HTTP 0.9 Error response just closes the connection                  */
      /* WE can't do this from the application thread so wake up the worker  */
      /* who will close the connection for us.                               */
      /***********************************************************************/
      TRC_NORMAL((TB, "Waking worker to close HTTP/0.9 connection"));
      OSP_GET_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);
      transaction->connection->wa_locked_outstanding_count--;

      FD_SET(transaction->connection->sock_handle,
             &transaction->connection->thread->wa_locked.write_set);

      /***********************************************************************/
      /* Wakeup the thread to try to write on this connection                */
      /***********************************************************************/
      if (transaction->connection->thread->wa_locked.sent_wakeup == 0)
      {
        TRC_DETAIL((TB, "Write to pipe"));
        transaction->connection->thread->wa_locked.sent_wakeup = 1;
#ifdef OSP_MEASURE_PERFORMANCE
        gettimeofday(&transaction->connection->thread->wa_locked.woken, NULL);
#endif
        error = write(
            transaction->connection->thread->wa_locked.wakeup_pipe_in, &rc, 1);

        if (error != 1)
        {
          TRC_OS_ERROR((TB, "Failed to write down wakeup pipe"));
          rc = OSP_ERROR;
        }
      }
      OSP_RELEASE_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);

      /***********************************************************************/
      /* Since the transaction is not queued (there is nothing to send!), we */
      /* need to free it here                                                */
      /***********************************************************************/
      TRC_DETAIL((TB, "Freeing transaction after sending HTTP/0.9 failure"));
      sm_free_transaction(transaction, TRUE);
    }
    else
    {
      TRC_DETAIL((TB, "Send HTTP 1.X canned response"));
      OSP_GET_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);
      transaction->connection->wa_locked_outstanding_count--;
      OSP_RELEASE_LOCK(&transaction->connection->thread->wa_locked.mutex_wa);
      sm_worker_send_canned_response(transaction, response);
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sm_send_response */

#ifdef OSP_MEASURE_PERFORMANCE
/**PROC+**********************************************************************/
/* Name:      sm_print_performance_measurements                              */
/*                                                                           */
/* Purpose:   Print performance measurements on the specified file stream    */
/*                                                                           */
/* Returns:   none                                                           */
/*                                                                           */
/* Params:    IN     fstream      - File stream to print on.                 */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_print_performance_measurements(FILE *fstream)
{
  TRC_ENTRY("sm_print_performance_measurements");

  sm_print_performance_data(fstream, &smg_processing);
  sm_print_performance_data(fstream, &smg_until_smime_received);
  sm_print_performance_data(fstream, &smg_until_send_queue);
  sm_print_performance_data(fstream, &smg_until_start_send);
  sm_print_performance_data(fstream, &smg_until_sent);
  sm_print_performance_data(fstream, &smg_select_cycle);
  sm_print_performance_data(fstream, &smg_wake_up);

  TRC_EXIT();
  return;

} /* sm_print_performance_measurements */
#endif

/**PROC+**********************************************************************/
/* Name:      sm_is_sm_thread                                                */
/*                                                                           */
/* Purpose:   Check if the calling thread is an sm thread                    */
/*                                                                           */
/* Returns:   TRUE if calling thread was created by sm                       */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_is_sm_thread(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc = FALSE;
  pthread_t calling_thread;
  void *thread_specific;

  TRC_ENTRY("sm_is_sm_thread");

  /***************************************************************************/
  /* Find out who the caller is                                              */
  /***************************************************************************/
  calling_thread = pthread_self();
  TRC_DETAIL((TB, "Testing thread id %u", (unsigned int)calling_thread));

  /***************************************************************************/
  /* Get the sm thread key data.  If this fails, or the key doesn't match   */
  /* up, then we're probably not being called on a sm thread                */
  /***************************************************************************/
  thread_specific = pthread_getspecific(sm_thread_key);
  if (thread_specific != NULL)
  {
    TRC_DETAIL((TB, "Got thread specific data"));
    if (thread_specific == sm_key_name)
    {
      TRC_DETAIL((TB, "Found correct key data - is an sm thread"));
      rc = TRUE;
    }
  }

  TRC_EXIT();

  return(rc);

} /* sm_is_sm_thread */
