
/**MOD+***********************************************************************/
/* Module:   smosp.c                                                         */
/*                                                                           */
/* Purpose:  Implements external osp_ functions                              */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.20               $ $Modtime::   Jan 11 2001 15:56:46   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <smincl.h>

/*****************************************************************************/
/* Keeps track of initialization progress of the stack.                      */
/*****************************************************************************/
static int osp_progress = OSP_INIT_NEVER;

/**PROC+**********************************************************************/
/* Name:      osp_init                                                       */
/*                                                                           */
/* Purpose:   Called to initialize the OSP server stack and its associated   */
/*            libraries.                                                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS          - completed successfully                  */
/*            OSP_ERR_CONFIG       - initialization failed because of an     */
/*                                   error reading configuration data        */
/*            OSP_ERROR            - initialization failed                   */
/*            OSP_ERR_ALREADY_INIT - already initialized                     */
/*                                                                           */
/* Params:    None                                                           */
/*                                                                           */
/* Operation: Initialize each of the OSP components.                         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_init(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  char *config_file_name;
  long config_error_line;
  LHASH *config = NULL;
  int error;

  TRC_API_ENTRY("osp_init");

  /***************************************************************************/
  /* Check for re-initialization                                             */
  /***************************************************************************/
  if (osp_progress != OSP_INIT_NEVER)
  {
    /*************************************************************************/
    /* Note that trace should be initialized, so we can do this              */
    /*************************************************************************/
    TRC_ALERT((TB, "Already initialized"));
    rc = OSP_ERR_ALREADY_INIT;
    goto EXIT_LABEL;
  }
  osp_progress = OSP_INIT_NONE;

  /***************************************************************************/
  /* Initialize tracing.                                                     */
  /***************************************************************************/
  trc_init(TRC_LEVEL_DETAIL);

  /***************************************************************************/
  /* Initialize the common code.                                             */
  /***************************************************************************/
  rc = osp_common_init();

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize COM"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Initialized COM"));
  osp_progress = OSP_INIT_COM;

  /***************************************************************************/
  /* Initialize the Resource Manager component.                              */
  /***************************************************************************/
  rc = rm_init();

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize RM"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Initialized RM"));
  osp_progress = OSP_INIT_RM;

  /***************************************************************************/
  /* Set the memory functions that SSL will use.  This has to be done before */
  /* any other SSL/crypto library functions because OpenSSL will not allow   */
  /* the memory functions to be changed once it has made its first           */
  /* allocation.                                                             */
  /***************************************************************************/
  if (!smg_done_once_only_init)
  {
    TRC_DETAIL((TB, "Setting memory functions"));
    error = CRYPTO_set_mem_functions((void * (*)(size_t))rm_get_mem,
                                    (void * (*)(void *, size_t))rm_realloc_mem,
                                    (void(*)(void *))rm_release_mem);
    if (error == 0)
    {
      TRC_ERROR((TB, "Failed to set SSL memory functions"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    smg_done_once_only_init = 1;
  }
  else
  {
    TRC_DETAIL((TB, "Memory functions already set"));
  }

  /***************************************************************************/
  /* Before we can initialize the rest of the stack, we need to read in the  */
  /* config file.                                                            */
  /***************************************************************************/
  config_file_name = osp_common_get_config_filename();
  if (config_file_name == NULL)
  {
    TRC_ERROR((TB, "No configuration file name"));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Open configuration file: \"%s\"", config_file_name));
  config = CONF_load(NULL, config_file_name, &config_error_line);
  if (config == NULL)
  {
    TRC_ERROR((TB, "Error loading configuration: file=\"%s\" line=%ld",
                                         config_file_name, config_error_line));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Read the parts of the configuration file that are relevant to SM.       */
  /***************************************************************************/
  rc = sm_read_config_file(config, &smg_config_data);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to read config file"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Read config file"));
  osp_progress = OSP_INIT_CONFIG;

  /***************************************************************************/
  /* Initialize the higher level components.                                 */
  /***************************************************************************/
  rc = smime_init(config);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize SMIME"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Initialized SMIME"));
  osp_progress = OSP_INIT_SMIME;

  /***************************************************************************/
  /* Initialize statistics and mutex                                         */
  /***************************************************************************/
  rc = osp_create_mutex(&smg_statistics_mutex);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to create statistics mutex"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initialize statistics                                                   */
  /***************************************************************************/
  smg_statistics.ssl_conns = 0;
  smg_statistics.nonssl_conns = 0;
  smg_statistics.osp_requests = 0;
  smg_statistics.scep_requests = 0;
  TRC_DETAIL((TB, "Initialized statistics mutex"));
  osp_progress = OSP_INIT_STATS;

  /***************************************************************************/
  /* Now initialize SM itself which initializes CCM                          */
  /***************************************************************************/
  rc = sm_init();

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize SM"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Initialized SM"));
  osp_progress = OSP_INIT_SM;

#ifdef OSP_SUPPORT_SCEP
  /***************************************************************************/
  /* Initialize SCEP                                                         */
  /***************************************************************************/
  rc = scep_init(config);

  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Failed to initialize SCEP"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Initialized SCEP"));
  osp_progress = OSP_INIT_SCEP;
#endif

  TRC_NORMAL((TB, "OSP Successfully Initialized"));
  osp_progress = OSP_INIT_FULL;

EXIT_LABEL:

  if (config != NULL)
  {
    /*************************************************************************/
    /* Free the configuration file handle.                                   */
    /*************************************************************************/
    CONF_free(config);
    config = NULL;
  }

  if ((rc != OSP_SUCCESS) && (rc != OSP_ERR_ALREADY_INIT))
  {
    /*************************************************************************/
    /* Initialization failed - osp_term() can clear up from any position.    */
    /*************************************************************************/
    osp_term();
  }

  TRC_API_EXIT();

  return(rc);

} /* osp_init */

/**PROC+**********************************************************************/
/* Name:      osp_listen                                                     */
/*                                                                           */
/* Purpose:   Causes the OSP server stack to begin listening for connections */
/*            from clients                                                   */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - completed successfully                      */
/*            OSP_ERR_NOT_INIT - stack was not initialized                   */
/*            OSP_ERR_CONFIG   - no listen ports were configured             */
/*            OSP_ERROR        - a listen call failed                        */
/*                                                                           */
/* Params:    None.                                                          */
/*                                                                           */
/* Operation: If we have a secure listening socket, listen on this socket.   */
/*            If we have a non-secure listening socket, listen on this       */
/*            socket.                                                        */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_listen(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  int error;

  TRC_API_ENTRY("osp_listen");

  /***************************************************************************/
  /* Make sure that we're ready to go                                        */
  /***************************************************************************/
  if (osp_progress != OSP_INIT_FULL)
  {
    TRC_ERROR((TB, "Called before initialization"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  if ((smg_config_data.secure_listening_port == 0) &&
      (smg_config_data.non_secure_listening_port == 0))
  {
    TRC_ERROR((TB, "No listening ports configured."));
    rc = OSP_ERR_CONFIG;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Only start listening if the ports have been configured.                 */
  /***************************************************************************/
  if (smg_config_data.secure_listening_port != 0)
  {
    TRC_DETAIL((TB, "Listen on secure socket"));

    error = listen(smg_listener_info.secure_listening_socket,
                                                     SM_LISTEN_SOCKET_BACKLOG);

    if (error != 0)
    {
      TRC_OS_ERROR((TB, "Failed to listen on secure listening socket"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
  }

  if (smg_config_data.non_secure_listening_port != 0)
  {
    TRC_DETAIL((TB, "Listen on non secure socket"));
    error = listen(smg_listener_info.non_secure_listening_socket,
                                                     SM_LISTEN_SOCKET_BACKLOG);

    if (error != 0)
    {
      TRC_OS_ERROR((TB, "Failed to listen on non secure listening socket"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
  }

  TRC_NORMAL((TB, "Now listening for incoming connections."));

EXIT_LABEL:

  TRC_API_EXIT();

  return(rc);

} /* osp_listen */

/**PROC+**********************************************************************/
/* Name:      osp_get_stack_statistics                                       */
/*                                                                           */
/* Purpose:   Called by the application to get stack statistics              */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - completed successfully                      */
/*            OSP_ERR_NOT_INIT - stack was not initialized                   */
/*                                                                           */
/* Params:    IN/OUT stats        - pointer to the statistics structure to   */
/*                                  fill in                                  */
/*                                                                           */
/* Operation: Copy the stack's statistics into the structure provided.       */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_get_stack_statistics(OSP_STATS * stats)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_API_ENTRY("osp_get_stack_statistics");

  if (osp_progress != OSP_INIT_FULL)
  {
    TRC_ERROR((TB, "Called before initialization"));
    rc = OSP_ERR_NOT_INIT;
  }
  /***************************************************************************/
  /* Check the stats pointer is not NULL                                     */
  /***************************************************************************/
  else if (stats != NULL)
  {
    OSP_GET_LOCK(&smg_statistics_mutex);

    memcpy(stats, &(smg_statistics), sizeof(*stats));

    OSP_RELEASE_LOCK(&smg_statistics_mutex);

    TRC_DETAIL((TB, "Returned ssl_conns     %u", stats->ssl_conns));
    TRC_DETAIL((TB, "Returned nonssl_conns  %u", stats->nonssl_conns));
    TRC_DETAIL((TB, "Returned osp_requests  %u", stats->osp_requests));
    TRC_DETAIL((TB, "Returned scep_requests %u", stats->scep_requests));
  }
  else
  {
    TRC_ERROR((TB, "NULL pointer supplied for stack statisitics"));
    rc = OSP_ERR_BAD_PARAM;
  }

  TRC_API_EXIT();

  return(rc);

} /* osp_get_stack_statistics */

/**PROC+**********************************************************************/
/* Name:      osp_term                                                       */
/*                                                                           */
/* Purpose:   Called to terminate the OSP server stack and its associated    */
/*            libraries.                                                     */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - completed successfully                   */
/*            OSP_ERR_NOT_INIT    - stack was not initialized.               */
/*            OSP_ERR_BAD_REQUEST - function was called on a callback        */
/*                                                                           */
/* Params:    None.                                                          */
/*                                                                           */
/* Operation: If sm is initialized, we must wait for it to terminate before  */
/*            shutting down the rest of the stack.  Termination is in the    */
/*            reverse order of initialization.                               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_API_ENTRY("osp_term");

  /***************************************************************************/
  /* Check the initialization state.  Only init never is not allowed - we    */
  /* should still be called to handle partial failures                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Initialization progress value %d", osp_progress));
  if (osp_progress == OSP_INIT_NEVER)
  {
    TRC_ERROR((TB, "Never initalized"));
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Need to check calling thread is not one of our SM worker or lam         */
  /* threads, since calling the shutdown process waits for those threads to  */
  /* complete processing...                                                  */
  /***************************************************************************/
  if (lam_is_lam_thread() || sm_is_sm_thread())
  {
    TRC_ERROR((TB, "May not call osp_term() from a callback"));
    rc = OSP_ERR_BAD_REQUEST;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Here, we shut down each component in the stack.  Switch to find how     */
  /* initialized we are.  The drop through case statements are deliberate to */
  /* terminate all components.                                               */
  /***************************************************************************/
  switch (osp_progress)
  {
    case OSP_INIT_FULL:

#ifdef OSP_SUPPORT_SCEP
    case OSP_INIT_SCEP:
      TRC_DETAIL((TB, "Terminate SCEP"));
      scep_term();
#endif

    case OSP_INIT_SM:
      /***********************************************************************/
      /* sm_term() blocks until SM termination is complete                   */
      /***********************************************************************/
      TRC_NORMAL((TB, "Calling sm_term()"));
      sm_term();

    case OSP_INIT_STATS:
      TRC_DETAIL((TB, "Deleting stats mutex"));
      osp_destroy_mutex(&smg_statistics_mutex);

    case OSP_INIT_SMIME:
      TRC_DETAIL((TB, "Terminate SMIME"));
      smime_term();

    case OSP_INIT_CONFIG:
      TRC_DETAIL((TB, "Free config"));
      /***********************************************************************/
      /* Nothing to do here now, but we'll leave it for future expansion.    */
      /***********************************************************************/

    case OSP_INIT_RM:
      TRC_DETAIL((TB, "Terminate RM"));
      rm_term();

    case OSP_INIT_COM:
      TRC_DETAIL((TB, "Terminate COM"));
      osp_common_term();
      break;

    default:
      /***********************************************************************/
      /* We really should never hit this, so we'll asssert rather than       */
      /* returning an error                                                  */
      /***********************************************************************/
      TRC_ASSERT((FALSE), (TB, "Illegal osp_progress value %d", osp_progress));
      break;
  }
  TRC_NORMAL((TB, "OSP Terminated"));
  osp_progress = OSP_INIT_NEVER;

  /***************************************************************************/
  /* Terminate tracing.                                                      */
  /***************************************************************************/
  trc_term();

EXIT_LABEL:

  TRC_API_EXIT();

  return(rc);

} /* osp_term */

/**PROC+**********************************************************************/
/* Name:      osp_client_verify_register                                     */
/*                                                                           */
/* Purpose:   Called by the application to register its client verification  */
/*            callback with the OSP Server.                                  */
/*                                                                           */
/*            Only one CV callback is supported per server.                  */
/*                                                                           */
/*            The application may pass a NULL pointer to deregister its      */
/*            callback.                                                      */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - completed successfully                      */
/*            OSP_ERR_NOT_INIT - stack was not initialized                   */
/*                                                                           */
/* Params:    IN     cv_cb        - pointer to the verification callback     */
/*            IN     cd_cb        - pointer to the disconnection callback    */
/*                                                                           */
/* Operation: Records the callback pointers in smg_listener_info             */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_client_verify_register(POSP_CLIENT_VERIFY_CALLBACK cv_cb,
                                         POSP_CLIENT_DISCONNECT_CALLBACK cd_cb)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_API_ENTRY("osp_client_verify_register");

  /***************************************************************************/
  /* Check we're initialized                                                 */
  /***************************************************************************/
  if (osp_progress != OSP_INIT_FULL)
  {
    TRC_ERROR((TB, "Called before initialization"));
    rc = OSP_ERR_NOT_INIT;
  }
  else
  {
    /*************************************************************************/
    /* Update the callbacks                                                  */
    /*************************************************************************/
    osp_get_recursive_lock(&smg_listener_info.callback_mutex);

    smg_listener_info.cd_cb = cd_cb;
    smg_listener_info.cv_cb = cv_cb;

    TRC_DETAIL((TB, "Set cd_cb to %p", cd_cb));
    TRC_DETAIL((TB, "Set cv_cb to %p", cv_cb));

    osp_release_recursive_lock(&smg_listener_info.callback_mutex);
  }

  TRC_API_EXIT();

  return(rc);

} /* osp_client_verify_register */

