#ifndef SMSTRCT_INCLUDED
#define SMSTRCT_INCLUDED

/**INC+***********************************************************************/
/* Header:    smstrct.h                                                      */
/*                                                                           */
/* Purpose:   SM Structures                                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.17               $ $Modtime::   Jan 11 2001 15:56:46   $  */
/*                                                                           */
/**INC-***********************************************************************/

/*****************************************************************************/
/* Variable naming semantics. When referencing a pipe, in means into the     */
/* pipe and out means out of the pipe.                                       */
/*****************************************************************************/

/**STRUCT+********************************************************************/
/* Structure:   CONFIG_DATA                                                  */
/*                                                                           */
/* Description: Structure that holds all of the configuration data from the  */
/*              config file.                                                 */
/*****************************************************************************/

typedef struct config_data
{
  /***************************************************************************/
  /* Ports on which to listen for secure connections                         */
  /***************************************************************************/
  unsigned short secure_listening_port;

  /***************************************************************************/
  /* Ports on which to listen for non-secure connections                     */
  /***************************************************************************/
  unsigned short non_secure_listening_port;

  /***************************************************************************/
  /* Number of SM worker threads to create per processor.                    */
  /***************************************************************************/
  int threads_per_processor;

} CONFIG_DATA;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure:   LISTENER_INFO                                                */
/*                                                                           */
/* Description: This is the main structure that controls SM operation.       */
/*****************************************************************************/

typedef struct listener_info
{
  /***************************************************************************/
  /* Global SSL Data                                                         */
  /*                                                                         */
  /* Function pointers used to access SSL functions.                         */
  /***************************************************************************/
  SSL_METHOD *ssl_method;

  /***************************************************************************/
  /* Global context for SSL operations.                                      */
  /***************************************************************************/
  SSL_CTX *ssl_ctx;

  /***************************************************************************/
  /* Locks provided for SSL use.                                             */
  /***************************************************************************/
  pthread_mutex_t ssl_locks[CRYPTO_NUM_LOCKS];

  /***************************************************************************/
  /* Lock protecting access to worker_threads list and the thread's          */
  /* notification pipe.                                                      */
  /***************************************************************************/
  pthread_mutex_t worker_list_mutex;

  /***************************************************************************/
  /* List of workers shared with worker threads.  Listener thread adds       */
  /* workers to this list and the workers remove themselves when they exit.  */
  /***************************************************************************/
  LIST_ENTRY worker_threads;

  /***************************************************************************/
  /* File descriptors I own - only accessed by listener thread               */
  /*                                                                         */
  /* Socket listening for secure incoming connections                        */
  /***************************************************************************/
  int secure_listening_socket;

  /***************************************************************************/
  /* Socket listening for non-secure incoming connections                    */
  /***************************************************************************/
  int non_secure_listening_socket;

  /***************************************************************************/
  /* Pipe that listener selects on.  Data read from this pipe tells the      */
  /* listener thread to shut down the stack.                                 */
  /***************************************************************************/
  int listener_termination_pipe_out;

  /***************************************************************************/
  /* The number of processors running on this machine.                       */
  /***************************************************************************/
  int processor_count;

  /***************************************************************************/
  /* Set of file descriptors that the listener selects on.  Saves building   */
  /* the set for each select call.                                           */
  /***************************************************************************/
  fd_set listening_socket_handles;

  /***************************************************************************/
  /* Never accessed by listener thread - application threads write down this */
  /* to initiate termination of the stack.                                   */
  /***************************************************************************/
  int listener_termination_pipe_in;

  /***************************************************************************/
  /* Application Callbacks                                                   */
  /*                                                                         */
  /* Protects cv_cb and cd_cb.                                               */
  /***************************************************************************/
  OSP_RECURSIVE_MUTEX callback_mutex;

  /***************************************************************************/
  /* Client verification callback pointer.                                   */
  /***************************************************************************/
  POSP_CLIENT_VERIFY_CALLBACK cv_cb;

  /***************************************************************************/
  /* Client disconnect callback pointer.                                     */
  /***************************************************************************/
  POSP_CLIENT_DISCONNECT_CALLBACK cd_cb;

} LISTENER_INFO;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure:   CONN_INFO                                                    */
/*                                                                           */
/* Description: Structure that represents a connection.                      */
/*****************************************************************************/

typedef struct conn_info
{
  /***************************************************************************/
  /* Only accessed by the worker thread                                      */
  /*                                                                         */
  /* Entry in worker's conn_list                                             */
  /***************************************************************************/
  LIST_ENTRY links;

  /***************************************************************************/
  /* SSL information for this connection.                                    */
  /***************************************************************************/
  SSL *ssl;

  /***************************************************************************/
  /* Transaction that is currently being received.                           */
  /***************************************************************************/
  TRANS_INFO *current_trans;

  /***************************************************************************/
  /* Time when this connection next needs to perform SSL renegotiation.      */
  /***************************************************************************/
  time_t next_renegotiate;

  /***************************************************************************/
  /* State of connection.                                                    */
  /***************************************************************************/
  int conn_state;

  /***************************************************************************/
  /* HTTP version for this connection.  Although the HTTP version is         */
  /* strictly a property of the connection, described by this field here,    */
  /* the code handles, up to a point, GET/POST transactions within the       */
  /* connection that specify different HTTP versions.  Hence there is an     */
  /* http_version in the TRANS_INFO structure as well.  The connection's     */
  /* http_version is initialized when an HTTP version is detected for the    */
  /* first transaction on the connection.                                    */
  /***************************************************************************/
  int http_version;

  /***************************************************************************/
  /* Persistence flag for this connection.  HTTP/0.9 connections are never   */
  /* persistent.  HTTP/1.0 connections are persistent if the Connection:     */
  /* Keep-Alive header is present HTTP/1.1 connections are persistent unless */
  /* the Connection: Close header is present                                 */
  /***************************************************************************/
  int persistent;

  /***************************************************************************/
  /* For SSL connections, this flag indicates whether data to be read may be */
  /* buffered within SSL.                                                    */
  /***************************************************************************/
  int ssl_data_buffered;

  /***************************************************************************/
  /* Last assigned transaction serial number.                                */
  /***************************************************************************/
  int last_assigned_serial;

  /***************************************************************************/
  /* Serial number of last sent transaction response.                        */
  /***************************************************************************/
  int last_response_serial;

  /***************************************************************************/
  /* Whether the client verification callback has been called yet for this   */
  /* connection.                                                             */
  /***************************************************************************/
  int called_cv_cb;

  /***************************************************************************/
  /* Identifier passed to application to identify the client.                */
  /***************************************************************************/
  OSP_CLIENT_ID client_id;

  /***************************************************************************/
  /* These are not changed after initialization.                             */
  /*                                                                         */
  /* Thread that owns this connection                                        */
  /***************************************************************************/
  THREAD_INFO *thread;

  /***************************************************************************/
  /* Connection's socket handle                                              */
  /***************************************************************************/
  int sock_handle;

  /***************************************************************************/
  /* Type of connection                                                      */
  /* SM_SECURE_SOCKET or SM_NON_SECURE_SOCKET                                */
  /***************************************************************************/
  int type;

  /***************************************************************************/
  /* These are shared between the worker and an application thread.  (must   */
  /* hold worker thread's mutex_wa to access these).                         */
  /*                                                                         */
  /* Count of outstanding transactions (transactions that are owned by       */
  /* higher layers of the stack) for this connection.                        */
  /***************************************************************************/
  unsigned int wa_locked_outstanding_count;

  /***************************************************************************/
  /* Queue of transactions waiting to be sent on this connection.            */
  /***************************************************************************/
  LIST_ENTRY wa_locked_send_queue;

} CONN_INFO;

/**STRUCT-********************************************************************/

/**STRUCT+********************************************************************/
/* Structure:   SM_PERF                                                      */
/*                                                                           */
/* Description: Structure containing performance measurement data.           */
/*****************************************************************************/

typedef struct sm_perf
{
  /***************************************************************************/
  /* A descriptive name for this performance measurement.                    */
  /***************************************************************************/
  const char *name;

  /***************************************************************************/
  /* The number of samples measured.                                         */
  /***************************************************************************/
  int count;

  /***************************************************************************/
  /* Accumulated time for all samples.                                       */
  /***************************************************************************/
  struct timeval accum;

} SM_PERF;

/**STRUCT-********************************************************************/

#endif
