
/**MOD+***********************************************************************/
/* Module:  smwkhttp.c                                                       */
/*                                                                           */
/* Purpose: SM worker thread functions that deal with HTTP                   */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.30               $ $Modtime::   Jan 11 2001 15:56:46   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <smincl.h>

/**PROC+**********************************************************************/
/* Name:      sm_worker_received_data                                        */
/*                                                                           */
/* Purpose:   Processes received data.                                       */
/*                                                                           */
/* Returns:   SM_CALL_AGAIN        - if the caller needs to call this        */
/*                                   function again.                         */
/*            SM_DONT_CALL_AGAIN   - if all of the data that can be          */
/*                                   processed has been processed.           */
/*            SM_CONNECTION_CLOSED - The connection has been closed.  Don't  */
/*                                   use this pointer again.                 */
/*                                                                           */
/* Params:    IN/OUT conn - connection on which data was received            */
/*                                                                           */
/* Operation: If we have a POST, call sm_worker_received_post().             */
/*            If we have a GET, call sm_worker_received_get().               */
/*            If we know it isn't a GET or a POST, return an error response. */
/*            If we don't yet have the entire message, increase the buffer   */
/*            size.                                                          */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_worker_received_data(CONN_INFO *conn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int new_buffer_size = 0;
  char *new_buffer;
  int rc = SM_DONT_CALL_AGAIN;

  TRC_ENTRY("sm_worker_received_data");

  /***************************************************************************/
  /* The minimum buffer size must be enough for us to identify which sort of */
  /* message we are dealing with.  For the SM layer, the data is always at   */
  /* the start of the buffer.                                                */
  /***************************************************************************/
  if ((conn->current_trans->recv_buffer_end > 3) &&
      (strncmp(conn->current_trans->recv_buffer, "POST", 4) == 0))
  {
    TRC_DETAIL((TB, "Got POST"));
    rc = sm_worker_received_post(conn, &new_buffer_size);
  }
  else if ((conn->current_trans->recv_buffer_end > 2) &&
           (strncmp(conn->current_trans->recv_buffer, "GET", 3) == 0))
  {
    TRC_DETAIL((TB, "Got GET"));
    rc = sm_worker_received_get(conn, &new_buffer_size);
  }
  else
  {
    /*************************************************************************/
    /* Not a GET or a POST - only allowable situation is one where we don't  */
    /* have enough message to match GET or POST.                             */
    /*************************************************************************/
    TRC_DETAIL((TB, "Not a GET or a POST"));

    if ((strncmp(conn->current_trans->recv_buffer, "POST",
                  conn->current_trans->recv_buffer_end) != 0) &&
        (strncmp(conn->current_trans->recv_buffer, "GET",
                      conn->current_trans->recv_buffer_end) != 0))
    {
      /***********************************************************************/
      /* We're not a generic HTTP server                                     */
      /***********************************************************************/
      TRC_ALERT((TB, "Received HTTP that we do not support."));
      TRC_DATA("HTTP that we do not support", conn->current_trans->recv_buffer,
                                         conn->current_trans->recv_buffer_end);
      FD_CLR(conn->sock_handle, &conn->thread->read_set);

      conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;

      if (conn->current_trans->http_version == SM_HTTP_0_9)
      {
        /*********************************************************************/
        /* Can't send any response, so just close the connection.            */
        /*********************************************************************/
        TRC_DETAIL((TB, "HTTP 0.9 - close connection"));
        rc = SM_CONNECTION_CLOSED;
        sm_worker_close_connection(conn);
      }
      else
      {
        TRC_DETAIL((TB, "HTTP 1.X - send error response"));
        sm_worker_send_canned_response(conn->current_trans, SM_RESPONSE_501);
        conn->current_trans = NULL;
      }
      goto EXIT_LABEL;
    }
  }

  TRC_DETAIL((TB, "new_buffer_size is %d", new_buffer_size));
  TRC_DETAIL((TB, "rc is %d", rc));

  if (new_buffer_size != 0)
  {
    /*************************************************************************/
    /* We have detected that a larger buffer is required to fit an entire    */
    /* message.                                                              */
    /* If we cannot reallocate the buffer, fail the request.                 */
    /*************************************************************************/
    TRC_ALERT((TB, "Reallocating buffer to %d", new_buffer_size));

    new_buffer =
             rm_realloc_mem(conn->current_trans->recv_buffer, new_buffer_size);

    if (new_buffer == NULL)
    {
      TRC_ERROR((TB, "Failed to allocate bigger buffer (%d)",
                 new_buffer_size));

      if (conn->current_trans->http_version == SM_HTTP_0_9)
      {
        /*********************************************************************/
        /* HTTP 0.9 doesn't have negative responses so we just have to close */
        /* the connection.                                                   */
        /*********************************************************************/
        TRC_DETAIL((TB, "HTTP/0.9 - close connection"));
        rc = SM_CONNECTION_CLOSED;
        sm_worker_close_connection(conn);
      }
      else
      {
        TRC_DETAIL((TB, "HTTP/1.X send error response"));
        sm_worker_send_canned_response(conn->current_trans, SM_RESPONSE_500);
        conn->current_trans = NULL;
      }

      goto EXIT_LABEL;
    }

    conn->current_trans->recv_buffer = new_buffer;
    conn->current_trans->recv_buffer_size = new_buffer_size;
  }

EXIT_LABEL:

  TRC_DETAIL((TB, "sm_worker_received_data returns %d", rc));

  TRC_EXIT();

  return(rc);

} /* sm_worker_received_data */

/**PROC+**********************************************************************/
/* Name:      sm_worker_received_post                                        */
/*                                                                           */
/* Purpose:   Processes received data.                                       */
/*                                                                           */
/* Returns:   SM_CALL_AGAIN        - if the caller needs to call this        */
/*                                   function again.                         */
/*            SM_DONT_CALL_AGAIN   - if all of the data that can be          */
/*                                   processed has been processed.           */
/*            SM_CONNECTION_CLOSED - The connection has been closed.  Don't  */
/*                                   use this pointer again.                 */
/*                                                                           */
/* Params:    IN/OUT conn            - connection on which data was received */
/*               OUT new_buffer_size - If we need to reallocate the receive  */
/*                                     buffer, set this to the required      */
/*                                     size.                                 */
/*                                                                           */
/* Operation: First check if we have the entire message.                     */
/*            Go through the message looking for the content-length and      */
/*            content-type header.  Ignore all other headers.                */
/*            If we have more than one message, set SM_CALL_AGAIN.           */
/*            Call smime_request_received to process the request.            */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_worker_received_post(CONN_INFO *conn, int *new_buffer_size)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int error;
  int match_index;
  char content_length_l[] = "content-length";
  char content_length_u[] = "CONTENT-LENGTH";
  char *content_length_string;
  int content_length;
  char *content_type;
  char *connection;
  TRANS_INFO *this_transaction;
  int rc = SM_DONT_CALL_AGAIN;
  int excess_length;
  char *current_ptr;
  char *crlf_ptr;
  char *lasts;

  TRC_ENTRY("sm_worker_received_post");

  /***************************************************************************/
  /* We have POST - this should be an HTTP request.  First thing we need to  */
  /* know is whether we have the whole message.                              */
  /***************************************************************************/
  current_ptr = conn->current_trans->recv_buffer;
  TRC_DATA("Got POST with data", current_ptr,
                                         conn->current_trans->recv_buffer_end);
  match_index = 0;
  *new_buffer_size = 0;

  /***************************************************************************/
  /* Spin through the message looking for the content-length field.  This    */
  /* will allow us to calculate the total length of the message and so know  */
  /* whether we have received it all.  At the same time, check for the       */
  /* double CRLF that indicates the end of the headers - if we hit that      */
  /* without finding the content length, then we've got a badly formed       */
  /* header                                                                  */
  /***************************************************************************/
  while (current_ptr <
       conn->current_trans->recv_buffer + conn->current_trans->recv_buffer_end)
  {
    /*************************************************************************/
    /* Have we got part of the content length?                               */
    /*************************************************************************/
    if ((*current_ptr == content_length_l[match_index]) ||
        (*current_ptr == content_length_u[match_index]))
    {
      match_index++;

      if (match_index == strlen(content_length_l))
      {
        /*********************************************************************/
        /* Got content-length                                                */
        /*********************************************************************/
        current_ptr++;
        TRC_DETAIL((TB, "Found content length; next char at %p", current_ptr));
        break;
      }
    }
    else
    {
      /***********************************************************************/
      /* No - check for end of headers                                       */
      /***********************************************************************/
      if ((current_ptr < (conn->current_trans->recv_buffer +
                                  conn->current_trans->recv_buffer_end - 4)) &&
          (current_ptr[0] == sm_const_crlf[0]) &&
          (current_ptr[1] == sm_const_crlf[1]) &&
          (current_ptr[2] == sm_const_crlf[0]) &&
          (current_ptr[3] == sm_const_crlf[1]))
      {
        /*********************************************************************/
        /* This marks the end of the headers - we must have a badly formed   */
        /* request                                                           */
        /*********************************************************************/
        TRC_ERROR((TB, "No content length found"));
        TRC_DATA("Buffer is:", conn->current_trans->recv_buffer,
                                         conn->current_trans->recv_buffer_end);
        FD_CLR(conn->sock_handle, &conn->thread->read_set);

        conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
        sm_worker_send_canned_response(conn->current_trans,
                                       SM_RESPONSE_400);
        goto EXIT_LABEL;
        break;
      }

      match_index = 0;
    }

    current_ptr++;
  }

  if (current_ptr !=
       conn->current_trans->recv_buffer + conn->current_trans->recv_buffer_end)
  {
    /*************************************************************************/
    /* Found content-length header.  Check that we get a colon and then a    */
    /* number.                                                               */
    /*************************************************************************/
    while (current_ptr < conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end)
    {
      /***********************************************************************/
      /* Only allowed whitespace or a colon                                  */
      /***********************************************************************/
      if (*current_ptr == ':')
      {
        current_ptr++;
        TRC_DETAIL((TB, "Found colon; next char at %p", current_ptr));
        break;
      }
      else if (!isspace(*current_ptr))
      {
        /*********************************************************************/
        /* Malformed HTTP - send an error response and close the connection  */
        /*********************************************************************/
        TRC_ERROR((TB,
                    "Not space character between content-length and : '%c' %x",
                    *current_ptr, *current_ptr));
        TRC_DATA("Buffer is:", conn->current_trans->recv_buffer,
                                         conn->current_trans->recv_buffer_end);
        FD_CLR(conn->sock_handle, &conn->thread->read_set);

        conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
        sm_worker_send_canned_response(conn->current_trans,
                                       SM_RESPONSE_400);
        goto EXIT_LABEL;
      }
      current_ptr++;
    }

    /*************************************************************************/
    /* Now get the actual length                                             */
    /*************************************************************************/
    while (current_ptr < conn->current_trans->recv_buffer +
                                        conn->current_trans->recv_buffer_end)
    {
      /***********************************************************************/
      /* Only allowed whitespace or a number                                 */
      /***********************************************************************/
      if (isdigit(*current_ptr))
      {
        content_length_string = current_ptr;
        break;
      }
      else if (!isspace(*current_ptr))
      {
        /*********************************************************************/
        /* Malformed HTTP - send an error response and close the connection  */
        /*********************************************************************/
        TRC_ERROR((TB, "Not space character between : and digits '%c' %x",
                                                  *current_ptr, *current_ptr));
        TRC_DATA("Buffer is:", conn->current_trans->recv_buffer,
                                         conn->current_trans->recv_buffer_end);
        FD_CLR(conn->sock_handle, &conn->thread->read_set);

        conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
        sm_worker_send_canned_response(conn->current_trans,
                                       SM_RESPONSE_400);
        goto EXIT_LABEL;
      }
      current_ptr++;
    }

    /*************************************************************************/
    /* If we aren't yet at the end of the buffer, we can find out the        */
    /* content length.                                                       */
    /*************************************************************************/
    if (current_ptr != conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end)
    {
      content_length = atoi(content_length_string);

      TRC_DETAIL((TB, "Content length is %d", content_length));

      /***********************************************************************/
      /* Run on till we hit the message body.  There may be any number of    */
      /* headers before this which we ignore.                                */
      /* End of the headers is two CRLF pairs.                               */
      /***********************************************************************/
      while (current_ptr + 3 < conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end)
      {
        if ((current_ptr[0] == sm_const_crlf[0]) &&
            (current_ptr[1] == sm_const_crlf[1]) &&
            (current_ptr[2] == sm_const_crlf[0]) &&
            (current_ptr[3] == sm_const_crlf[1]))
        {
          /*******************************************************************/
          /* Got the body - advance recv_buffer_start to point to the        */
          /* message body as this is what the upper layers need.             */
          /*******************************************************************/
          TRC_DETAIL((TB, "Message body starts at %p", current_ptr + 4));
          break;
        }
        current_ptr++;
      }

      /***********************************************************************/
      /* We are now in a position to know whether we have an entire message  */
      /* in the buffer.                                                      */
      /***********************************************************************/
      if (current_ptr != conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end)
      {
        excess_length = current_ptr + 4 - conn->current_trans->recv_buffer;
        excess_length = conn->current_trans->recv_buffer_end -
                                                excess_length - content_length;
        TRC_DETAIL((TB, "Excess length is %d", excess_length));

        if (excess_length >= 0)
        {
          /*******************************************************************/
          /* We now have a complete message. We know that we will not need   */
          /* to go through this processing again for this transaction so we  */
          /* can advance recv_buffer_start to point to the message body as   */
          /* this is what the upper layers use.                              */
          /*******************************************************************/
          conn->current_trans->recv_buffer_start =
                            current_ptr + 4 - conn->current_trans->recv_buffer;
          this_transaction = conn->current_trans;
          conn->current_trans = NULL;
          TRC_DATA("Content is", current_ptr + 4, content_length);

          /*******************************************************************/
          /* Check to see if we have more than one message.                  */
          /*******************************************************************/
          if (excess_length > 0)
          {
            /*****************************************************************/
            /* Multiple messages in buffer.  Set rc to SM_CALL_AGAIN so that */
            /* the we process the other messages.                            */
            /*****************************************************************/
            TRC_DETAIL((TB, "Multiple messages in the buffer."));

            /*****************************************************************/
            /* Copy the excess data into another transaction.                */
            /*****************************************************************/
            conn->current_trans = sm_get_transaction(conn);

            if (conn->current_trans == NULL)
            {
              /***************************************************************/
              /* We have enough room for the first transaction in the queue  */
              /* but not for the excess data. WE have to fail the connection */
              /* but first we'll send an HTTP error response.                */
              /***************************************************************/
              TRC_ERROR((TB, "Failed to get new transaction"));
              FD_CLR(conn->sock_handle, &conn->thread->read_set);

              conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
              sm_worker_send_canned_response(this_transaction,
                                             SM_RESPONSE_500);
              this_transaction = NULL;
              goto EXIT_LABEL;
            }

            /*****************************************************************/
            /* Arrange to be called again to process the next transaction.   */
            /*****************************************************************/
            rc = SM_CALL_AGAIN;

#ifdef OSP_MEASURE_PERFORMANCE
            /*****************************************************************/
            /* Set the overflow transaction's IO start time.                 */
            /*****************************************************************/
            gettimeofday(&(conn->current_trans->start_time_io), NULL);
            TRC_DETAIL((TB, "Set transaction IO start time:"
                        " trans=%p time=%ld:%06ld",
                        conn->current_trans,
                        conn->current_trans->start_time_io.tv_sec,
                        conn->current_trans->start_time_io.tv_usec));
#endif

            TRC_ASSERT((excess_length <=
                        conn->current_trans->recv_buffer_size),
                        (TB, "Receive buffer not big enough %d, %d",
                        excess_length, conn->current_trans->recv_buffer_size));

            memcpy(conn->current_trans->recv_buffer,
                               this_transaction->recv_buffer +
                               this_transaction->recv_buffer_end
                                             - excess_length, excess_length);
            this_transaction->recv_buffer_end -= excess_length;
            conn->current_trans->recv_buffer_end = excess_length;

            TRC_DETAIL((TB, "Amount of room left in new transaction is %d",
                       conn->current_trans->recv_buffer_size - excess_length));
          }

          /*******************************************************************/
          /* Got everything we want.  NULL terminate the end of the header   */
          /* section to make it easier to manipulate.                        */
          /*******************************************************************/
          this_transaction->recv_buffer[
                                  this_transaction->recv_buffer_start - 1] = 0;

          /*******************************************************************/
          /* Check that we have a correct version of HTTP in correct format. */
          /* Request line is of form:                                        */
          /* POST uriuriuriuri HTTP/1.X                                      */
          /* So we look for the third token to get the HTTP version.         */
          /*******************************************************************/
          TRC_DETAIL((TB, "Looking for HTTP version..."));
          lasts = NULL;

          current_ptr = sm_strtok(this_transaction->recv_buffer, ' ', &lasts);

          if (current_ptr == NULL)
          {
            /*****************************************************************/
            /* Can't find first token - error. Malformed HTTP - send an      */
            /* error response and close the connection                       */
            /*****************************************************************/
            TRC_ERROR((TB, "Can't find first token in request line."));
            TRC_DATA("Buffer is:", this_transaction->recv_buffer,
                                            this_transaction->recv_buffer_end);
            FD_CLR(conn->sock_handle, &conn->thread->read_set);

            conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
            sm_worker_send_canned_response(this_transaction,
                                           SM_RESPONSE_400);
            this_transaction = NULL;
            goto EXIT_LABEL;
          }

          /*******************************************************************/
          /* HTTP request                                                    */
          /*******************************************************************/
          TRC_DETAIL((TB, "First token is %s", current_ptr));

          current_ptr = sm_strtok(NULL, ' ', &lasts);

          if (current_ptr == NULL)
          {
            /*****************************************************************/
            /* Can't find second token - error. Malformed HTTP - send an     */
            /* error response and close the connection                       */
            /*****************************************************************/
            TRC_ERROR((TB, "Can't find second token in request line."));
            TRC_DATA("Buffer is:", this_transaction->recv_buffer,
                                            this_transaction->recv_buffer_end);
            FD_CLR(conn->sock_handle, &conn->thread->read_set);

            conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
            sm_worker_send_canned_response(this_transaction,
                                           SM_RESPONSE_400);
            this_transaction = NULL;
            goto EXIT_LABEL;
          }

          /*******************************************************************/
          /* URI - we don't care about this                                  */
          /*******************************************************************/
          TRC_DETAIL((TB, "Second token is %s", current_ptr));

          current_ptr = sm_strtok(NULL, ' ', &lasts);

          if (current_ptr == NULL)
          {
            /*****************************************************************/
            /* Can't find third token - error. Malformed HTTP - send an      */
            /* error response and close the connection                       */
            /*****************************************************************/
            TRC_ERROR((TB, "Can't find third token in request line."));
            TRC_DATA("Buffer is:", this_transaction->recv_buffer,
                                            this_transaction->recv_buffer_end);
            FD_CLR(conn->sock_handle, &conn->thread->read_set);

            conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
            sm_worker_send_canned_response(this_transaction,
                                           SM_RESPONSE_400);
            this_transaction = NULL;
            goto EXIT_LABEL;
          }

          TRC_DETAIL((TB, "Third token is %s", current_ptr));

          /*******************************************************************/
          /* Check the HTTP version                                          */
          /*******************************************************************/
          if (strncmp(current_ptr, "HTTP/1.0", 8) == 0)
          {
            TRC_DETAIL((TB, "HTTP/1.0"));
            SM_SET_HTTP_VERSION(this_transaction, SM_HTTP_1_0);
          }
          else if (strncmp(current_ptr, "HTTP/1.1", 8) == 0)
          {
            TRC_DETAIL((TB, "HTTP/1.1"));
            SM_SET_HTTP_VERSION(this_transaction, SM_HTTP_1_1);
          }
          else
          {
            TRC_ERROR((TB, "Unsupported HTTP version"));

            TRC_DATA("Buffer is:", this_transaction->recv_buffer,
                                            this_transaction->recv_buffer_end);
            FD_CLR(conn->sock_handle, &conn->thread->read_set);

            conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
            sm_worker_send_canned_response(this_transaction,
                                           SM_RESPONSE_501);
            this_transaction = NULL;
            goto EXIT_LABEL;
          }

          /*******************************************************************/
          /* sm_strtok will have replaced the space after the http version   */
          /* with a NULL; we need to undo that before we go on               */
          /*******************************************************************/
          current_ptr[strlen(current_ptr)] = ' ';

          /*******************************************************************/
          /* At this point we have successfully parsed the request line and  */
          /* are now looking at the headers.  Given that we've already found */
          /* the content-length, we're only interested in the content-type   */
          /* and any connection header                                       */
          /*******************************************************************/
          TRC_DATA("Looking for headers in", current_ptr, strlen(current_ptr));
          content_type = NULL;
          connection = NULL;
          while (TRUE)
          {
            /*****************************************************************/
            /* Headers are delimited by CRLF and the header name must be at  */
            /* the start of the line.                                        */
            /*****************************************************************/
            current_ptr = strstr(current_ptr, "\r\n");

            if (current_ptr == NULL)
            {
              /***************************************************************/
              /* Reached the end of the header.  Did we get the content      */
              /* type?                                                       */
              /***************************************************************/
              if (content_type == NULL)
              {
                /*************************************************************/
                /* No - malformed HTTP                                       */
                /*************************************************************/
                TRC_ERROR((TB, "Can't find content-type"));
                TRC_DATA("Buffer is:", this_transaction->recv_buffer,
                                            this_transaction->recv_buffer_end);
                sm_worker_send_canned_response(this_transaction,
                                               SM_RESPONSE_400);
                this_transaction = NULL;
                goto EXIT_LABEL;
              }

              /***************************************************************/
              /* Yes - finish looping                                        */
              /***************************************************************/
              break;
            }
            current_ptr += 2;

            /*****************************************************************/
            /* Look for content-type                                         */
            /*****************************************************************/
            if (strncasecmp(current_ptr, sm_const_content_type,
                                           sizeof(sm_const_content_type)) == 0)
            {
              /***************************************************************/
              /* Found it                                                    */
              /***************************************************************/
              TRC_DETAIL((TB, "Got Content-Type at %p", current_ptr));
              content_type = current_ptr + sizeof(sm_const_content_type);
            }
            /*****************************************************************/
            /* Look for connection                                           */
            /*****************************************************************/
            else if (strncasecmp(current_ptr, sm_const_connection,
                                             sizeof(sm_const_connection)) == 0)
            {
              /***************************************************************/
              /* Found it                                                    */
              /***************************************************************/
              TRC_DETAIL((TB, "Got connection at %p", current_ptr));
              connection = current_ptr + sizeof(sm_const_connection);
            }
          }

          /*******************************************************************/
          /* Set up persistence based on http version; then override it with */
          /* any connection type we found                                    */
          /*******************************************************************/
          TRC_DETAIL((TB, "Setting persistence for connection"));
          if (conn->http_version == SM_HTTP_1_1)
          {
            TRC_DETAIL((TB, "HTTP/1.1 persistent by default"));
            conn->persistent = TRUE;
          }
          else
          {
            TRC_DETAIL((TB, "Not HTTP/1.1 - not persistent by default"));
            conn->persistent = FALSE;
          }

          if (connection != NULL)
          {
            /*****************************************************************/
            /* There could be an initial space                               */
            /*****************************************************************/
            TRC_DETAIL((TB, "Checking for conn type in '%s'", connection));
            while (connection < (this_transaction->recv_buffer +
                                            this_transaction->recv_buffer_end))
            {
              if (*connection == ' ')
              {
                connection++;
              }
              else
              {
                break;
              }
            }

            /*****************************************************************/
            /* Now check what we've got.  For http/1.0 we are interested     */
            /* only in 'keep-alive'; for http/1.1 we are interested only in  */
            /* 'close'.  Note that we set the flag each time through because */
            /* a previously persistent connection may change to              */
            /* non-persistent at a later date.                               */
            /*****************************************************************/
            if (conn->http_version == SM_HTTP_1_0)
            {
              TRC_DETAIL((TB, "HTTP/1.0: looking for keep-alive"));
              if (strncasecmp(connection, sm_const_conn_keep_alive,
                                             sizeof(sm_const_conn_keep_alive)))
              {
                TRC_DETAIL((TB, "No keepalive; HTTP/1.0 conn not persistent"));
              }
              else
              {
                TRC_DETAIL((TB, "keep alive - HTTP/1.0 conn persistent"));
                conn->persistent = TRUE;
              }
            }
            else if (conn->http_version == SM_HTTP_1_1)
            {
              TRC_DETAIL((TB, "HTTP/1.1: looking for close"));
              if (strncasecmp(connection, sm_const_conn_close,
                                                  sizeof(sm_const_conn_close)))
              {
                TRC_DETAIL((TB, "Close not found - HTTP/1.1 conn persistent"));
              }
              else
              {
                TRC_DETAIL((TB, "Close found - HTTP/1.1 conn not persistent"));
                conn->persistent = FALSE;
              }
            }

          }

          /*******************************************************************/
          /* If this connection is not persistent, then we should mark it to */
          /* not receive any more.  This means that the connection will be   */
          /* closed once the associated transaction is processed.  Note that */
          /* given that this is a non-persistent connection, there should be */
          /* nothing more to receieve anyway.                                */
          /*******************************************************************/
          if (!conn->persistent)
          {
            TRC_DETAIL((TB, "Non-persistent connection - don't receive"));
            conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
          }

          /*******************************************************************/
          /* We now definitely have a well formatted HTTP request so we can  */
          /* increment the statistics.                                       */
          /*******************************************************************/
          OSP_GET_LOCK(&smg_statistics_mutex);
          smg_statistics.osp_requests++;
          TRC_DETAIL((TB, "Now %u osp_requests", smg_statistics.osp_requests));
          OSP_RELEASE_LOCK(&smg_statistics_mutex);

          /*******************************************************************/
          /* Increment outstanding_count so that we remember to wait for     */
          /* this transaction to complete.                                   */
          /*******************************************************************/
          OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
          conn->wa_locked_outstanding_count++;
          OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);

#ifdef OSP_MEASURE_PERFORMANCE
          /*******************************************************************/
          /* Set the transaction processing start time.                      */
          /*******************************************************************/
          gettimeofday(&(this_transaction->start_time_processing), NULL);
          TRC_DETAIL((TB, "Set transaction processing start time:"
                      " trans=%p time=%ld:%06ld",
                      this_transaction,
                      this_transaction->start_time_processing.tv_sec,
                      this_transaction->start_time_processing.tv_usec));

          /*******************************************************************/
          /* Accumulate time up until this point.                            */
          /*******************************************************************/
          SM_ACCUMULATE(smg_until_smime_received,
                        this_transaction->start_time_io,
                        this_transaction->start_time_processing);
#endif

          /*******************************************************************/
          /* Call SMIME layer to process this request                        */
          /*******************************************************************/
          error = smime_request_received(this_transaction,
                                         content_type);

          switch (error)
          {
            case OSP_SUCCESS:
              TRC_DETAIL((TB, "smime_request_received() was successful"));
              break;

            case OSP_ERR_MEMORY:
              TRC_ERROR((TB,
                        "Failure of smime_request_received() OSP_ERR_MEMORY"));
              OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
              conn->wa_locked_outstanding_count--;
              OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
              sm_worker_send_canned_response(this_transaction,
                                             SM_RESPONSE_500);
              break;

            default:
              TRC_ERROR((TB, "Failure of smime_request_received() %d", error));
              OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
              conn->wa_locked_outstanding_count--;
              OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
              sm_worker_send_canned_response(this_transaction,
                                             SM_RESPONSE_400);
              break;
          }

          this_transaction = NULL;
        }
        else
        {
          /*******************************************************************/
          /* We have the content length and all of the headers but we don't  */
          /* yet have the whole message.  If we don't have room to accept    */
          /* the whole message, reallocate the buffer now.                   */
          /*******************************************************************/
          TRC_DETAIL((TB, "Got content-length (%d) but not whole message",
                                                              content_length));

          if (conn->current_trans->recv_buffer_size - content_length
                        < (current_ptr + 4 - conn->current_trans->recv_buffer))
          {
            /*****************************************************************/
            /* Not enough room in buffer                                     */
            /*****************************************************************/
            TRC_DETAIL((TB, "Buffer not big enough for whole message"));
            *new_buffer_size = content_length +
                          (current_ptr + 4 - conn->current_trans->recv_buffer);
          }
        }
      }
      else
      {
        /*********************************************************************/
        /* We have the content length but we haven't got to the end of the   */
        /* headers so if the buffer is full, we need to reallocate.          */
        /*********************************************************************/
        TRC_DETAIL((TB, "Got content-length (%d) but not all headers",
                                                              content_length));

        if (conn->current_trans->recv_buffer_end ==
                                         conn->current_trans->recv_buffer_size)
        {
          /*******************************************************************/
          /* Buffer filled                                                   */
          /*******************************************************************/
          TRC_DETAIL((TB, "Buffer is full"));
          *new_buffer_size = conn->current_trans->recv_buffer_size * 2;
        }
      }
    }
    else
    {
      /***********************************************************************/
      /* We don't yet know the content length. If we've filled the buffer    */
      /* then we better get a bigger one.                                    */
      /***********************************************************************/
      TRC_DETAIL((TB, "Don't yet have content-length"));

      if (conn->current_trans->recv_buffer_end ==
                                         conn->current_trans->recv_buffer_size)
      {
        /*********************************************************************/
        /* Buffer filled                                                     */
        /*********************************************************************/
        TRC_DETAIL((TB, "Buffer is full"));
        *new_buffer_size = conn->current_trans->recv_buffer_size * 2;
      }
    }
  }
  else
  {
    /*************************************************************************/
    /* We don't have the content length field.  There are three possible     */
    /* reasons for this:                                                     */
    /* - We haven't yet received the whole message                           */
    /* - We haven't received the whole message because the buffer is full    */
    /* - This is not a well formatted HTTP/1.X message (e.g. HTTP 0.9).      */
    /*************************************************************************/
    if (conn->current_trans->recv_buffer_end ==
                                         conn->current_trans->recv_buffer_size)
    {
      TRC_NORMAL((TB, "Buffer Full - need to realloc"));
      *new_buffer_size = conn->current_trans->recv_buffer_size * 2;
    }
    else
    {
      /***********************************************************************/
      /* Buffer isn't full, check for HTTP 0.9.  If we have a CRLF and no    */
      /* HTTP/X.Y then this is 0.9                                           */
      /***********************************************************************/
      crlf_ptr = NULL;
      for (current_ptr = conn->current_trans->recv_buffer;
           current_ptr + 1 < conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end;
           current_ptr++)
      {
        if ((current_ptr[0] == sm_const_crlf[0]) &&
            (current_ptr[1] == sm_const_crlf[1]))
        {
          /*******************************************************************/
          /* Got the CRLF                                                    */
          /*******************************************************************/
          crlf_ptr = current_ptr;
          break;
        }
      }

      if (crlf_ptr != NULL)
      {
        TRC_DETAIL((TB, "Got a CRLF - now look for HTTP"));

        for (current_ptr = conn->current_trans->recv_buffer;
             current_ptr + 7 < crlf_ptr;
             current_ptr++)
        {

          /*******************************************************************/
          /* If we've got the "HTTP/1.x" bit we can get the HTTP version     */
          /*******************************************************************/
          if ((current_ptr[0] == 'H') &&
              (current_ptr[1] == 'T') &&
              (current_ptr[2] == 'T') &&
              (current_ptr[3] == 'P') &&
              (current_ptr[4] == '/') &&
              (current_ptr[5] == '1'))
          {
            if (current_ptr[7] == '1')
            {
              TRC_DETAIL((TB, "Got HTTP/1.1"));
              SM_SET_HTTP_VERSION(conn->current_trans, SM_HTTP_1_1);
            }
            else
            {
              TRC_DETAIL((TB, "Got HTTP/1.0"));
              SM_SET_HTTP_VERSION(conn->current_trans, SM_HTTP_1_0);
            }
            break;
          }
        }

        if (conn->current_trans->http_version == SM_HTTP_0_9)
        {
          /*******************************************************************/
          /* HTTP 0.9 request - just close connection.                       */
          /*******************************************************************/
          TRC_ALERT((TB, "HTTP/0.9 Request"));
          rc = SM_CONNECTION_CLOSED;
          sm_worker_close_connection(conn);
          goto EXIT_LABEL;
        }
      }
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sm_worker_received_post */

/**PROC+**********************************************************************/
/* Name:      sm_worker_received_get                                         */
/*                                                                           */
/* Purpose:   Processes received data.                                       */
/*                                                                           */
/* Returns:   SM_CALL_AGAIN        - if the caller needs to call this        */
/*                                   function again.                         */
/*            SM_DONT_CALL_AGAIN   - if all of the data that can be          */
/*                                   processed has been processed.           */
/*            SM_CONNECTION_CLOSED - The connection has been closed.  Don't  */
/*                                   use this pointer again.                 */
/*                                                                           */
/* Params:    IN/OUT conn            - connection on which data was received */
/*               OUT new_buffer_size - If we need to reallocate the receive  */
/*                                     buffer, set this to the required      */
/*                                     size.                                 */
/*                                                                           */
/* Operation: Check we have all of the message.                              */
/*            For SCEP requests we only ever have one per connection so we   */
/*            can always disable writes and set the connection state to      */
/*            SM_CONN_STATE_DONT_RECEIVE.                                    */
/*            Call scep_request_received() to actually process the data.     */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_worker_received_get(CONN_INFO *conn, int *new_buffer_size)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int error;
  TRANS_INFO *this_transaction;
  int rc = SM_DONT_CALL_AGAIN;
  char *current_ptr;
  char *operation;
  int got_crlf = FALSE;

  TRC_ENTRY("sm_worker_received_get");

  *new_buffer_size = 0;

  /***************************************************************************/
  /* We have GET - this should be a SCEP request.  The request ends with a   */
  /* single CRLF for http/0.9 and with a double crlf for http/1.x            */
  /***************************************************************************/
  current_ptr = conn->current_trans->recv_buffer;

  TRC_DATA("Parsing get:", conn->current_trans->recv_buffer,
                                   conn->current_trans->recv_buffer_end);
  while (current_ptr + 1 < conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end)
  {
    /*************************************************************************/
    /* Check for end of line CRLF                                            */
    /*************************************************************************/
    if ((current_ptr[0] == sm_const_crlf[0]) &&
        (current_ptr[1] == sm_const_crlf[1]))
    {
      /***********************************************************************/
      /* If there's no http header to be found, then we're dealing with      */
      /* HTTP/0.9 and we only get a single crlf.  Put in a temporary NULL to */
      /* limit the search                                                    */
      /***********************************************************************/
      TRC_DETAIL((TB, "Found crlf"));
      current_ptr[1] = 0;
      if (strstr(conn->current_trans->recv_buffer, "HTTP/") == NULL)
      {
        TRC_DETAIL((TB, "Got complete HTTP/0.9 request"));
        got_crlf = TRUE;
        break;
      }
      current_ptr[1] = sm_const_crlf[1];

      /***********************************************************************/
      /* Now check for a second CRLF, which indicates the end of a HTTP/1.x  */
      /* request                                                             */
      /***********************************************************************/
      TRC_DETAIL((TB, "Found http header - check for double CRLF"));
      if ((current_ptr + 3 < conn->current_trans->recv_buffer +
                                       conn->current_trans->recv_buffer_end) &&
          (current_ptr[2] == sm_const_crlf[0]) &&
          (current_ptr[3] == sm_const_crlf[1]))
      {
        TRC_DETAIL((TB, "Got complete HTTP/1.x request"));
        got_crlf = TRUE;
        break;
      }
    }

    current_ptr++;
  }

  if (got_crlf)
  {
    /*************************************************************************/
    /* Got an entire request.  Write a NULL at the end of the transaction to */
    /* allow easier string manipulation.                                     */
    /*************************************************************************/
    TRC_DETAIL((TB, "Got CRLF at %p", current_ptr));
    current_ptr[1] = 0;
    conn->current_trans->recv_buffer_end =
                            current_ptr + 1 - conn->current_trans->recv_buffer;

    /*************************************************************************/
    /* Got a SCEP request.  Check which version of HTTP we have.             */
    /*************************************************************************/
    current_ptr = strstr(conn->current_trans->recv_buffer, "HTTP/");

    /*************************************************************************/
    /* If there is no HTTP/X.Y present this is 0.9 which is fine.  Otherwise */
    /* we need to check we can support the version.                          */
    /*************************************************************************/
    if (current_ptr == NULL)
    {
      TRC_DETAIL((TB, "HTTP/0.9"));
      SM_SET_HTTP_VERSION(conn->current_trans, SM_HTTP_0_9);

      /***********************************************************************/
      /* For HTTP/0.9, there will be no more transactions on this            */
      /* connection.                                                         */
      /***********************************************************************/
      FD_CLR(conn->sock_handle, &conn->thread->read_set);
      conn->persistent = FALSE;
    }
    else
    {
      /***********************************************************************/
      /* Now see what http version we've got                                 */
      /***********************************************************************/
      TRC_DETAIL((TB, "Checking http version, %s", current_ptr));
      if (strncmp(current_ptr, "HTTP/1.0", 8) == 0)
      {
        TRC_DETAIL((TB, "HTTP/1.0"));
        SM_SET_HTTP_VERSION(conn->current_trans, SM_HTTP_1_0);
      }
      else if (strncmp(current_ptr, "HTTP/1.1", 8) == 0)
      {
        TRC_DETAIL((TB, "HTTP/1.1"));
        SM_SET_HTTP_VERSION(conn->current_trans, SM_HTTP_1_1);
      }
      else
      {
        /*********************************************************************/
        /* Don't support this version of HTTP.  Make the response an         */
        /* HTTP/1.1 response as this is the highest version we do support.   */
        /*********************************************************************/
        TRC_ERROR((TB, "Unsupported HTTP version"));
        TRC_DATA("Buffer is:", conn->current_trans->recv_buffer,
                                         conn->current_trans->recv_buffer_end);
        SM_SET_HTTP_VERSION(conn->current_trans, SM_HTTP_1_1);
        sm_worker_send_canned_response(conn->current_trans,
                                       SM_RESPONSE_501);
        conn->current_trans = NULL;
        goto EXIT_LABEL;
      }

      /***********************************************************************/
      /* Check for any connection header                                     */
      /***********************************************************************/
      current_ptr = conn->current_trans->recv_buffer;
      do
      {
        /*********************************************************************/
        /* Headers are delimited by CRLF and the header name must be at the  */
        /* start of the line.                                                */
        /*********************************************************************/
        current_ptr = strstr(current_ptr, "\r\n");
        if (current_ptr == NULL)
        {
          TRC_DETAIL((TB, "Reached end of headers"));
          break;
        }
        current_ptr += 2;

        /*********************************************************************/
        /* Look for connection                                               */
        /*********************************************************************/
        if (strncasecmp(current_ptr, sm_const_connection,
                                         sizeof(sm_const_connection)) == 0)
        {
          /*******************************************************************/
          /* Found it                                                        */
          /*******************************************************************/
          TRC_DETAIL((TB, "Got connection at %p", current_ptr));
          current_ptr += sizeof(sm_const_connection);
          break;
        }
      } while (TRUE);

      /***********************************************************************/
      /* Set up persistence based on http version; then override it with any */
      /* connection type we found                                            */
      /***********************************************************************/
      TRC_DETAIL((TB, "Setting persistence for connection"));
      if (conn->http_version == SM_HTTP_1_1)
      {
        TRC_DETAIL((TB, "HTTP/1.1 persistent by default"));
        conn->persistent = TRUE;
      }
      else
      {
        TRC_DETAIL((TB, "Not HTTP/1.1 - not persistent by default"));
        conn->persistent = FALSE;
      }

      if (current_ptr != NULL)
      {
        /*********************************************************************/
        /* There could be an initial space                                   */
        /*********************************************************************/
        TRC_DETAIL((TB, "Checking for conn type in '%s'", current_ptr));
        while ((*current_ptr != NULL) && (*current_ptr == ' '))
        {
          current_ptr++;
        }

        /*********************************************************************/
        /* Now check what we've got.  For http/1.0 we are interested only in */
        /* 'keep-alive'; for http/1.1 we are interested only in 'close'.     */
        /* Note that we set the flag each time through because a previously  */
        /* persistent connection may change to non-persistent at a later     */
        /* date.                                                             */
        /*********************************************************************/
        if (conn->http_version == SM_HTTP_1_0)
        {
          TRC_DETAIL((TB, "HTTP/1.0: looking for keep-alive"));
          if (strncasecmp(current_ptr, sm_const_conn_keep_alive,
                                             sizeof(sm_const_conn_keep_alive)))
          {
            TRC_DETAIL((TB, "No keepalive; HTTP/1.0 conn not persistent"));
          }
          else
          {
            TRC_DETAIL((TB, "keep alive - HTTP/1.0 conn persistent"));
            conn->persistent = TRUE;
          }
        }
        else if (conn->http_version == SM_HTTP_1_1)
        {
          TRC_DETAIL((TB, "HTTP/1.1: looking for close"));
          if (strncasecmp(current_ptr, sm_const_conn_close,
                                                  sizeof(sm_const_conn_close)))
          {
            TRC_DETAIL((TB, "Close not found - HTTP/1.1 conn persistent"));
          }
          else
          {
            TRC_DETAIL((TB, "Close found - HTTP/1.1 conn not persistent"));
            conn->persistent = FALSE;
          }
        }

      }
    }

    /*************************************************************************/
    /* If this connection is not persistent, then we should mark it to not   */
    /* receive any more.  This means that the connection will be closed once */
    /* the associated transaction is processed.  Note that given that this   */
    /* is a non-persistent connection, there should be nothing more to       */
    /* receieve anyway.                                                      */
    /*************************************************************************/
    if (!conn->persistent)
    {
      TRC_DETAIL((TB, "Non-persistent connection - don't receive"));
      conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;
    }

#ifdef OSP_SUPPORT_SCEP
    operation = strstr(conn->current_trans->recv_buffer, "operation=");
#else
    TRC_ALERT((TB, "HTTP GET operation not supported (SCEP not built in)"));
    operation = NULL;
#endif

    if (operation == NULL)
    {
      TRC_ERROR((TB, "No SCEP operation"));
      TRC_DATA("Buffer is:", conn->current_trans->recv_buffer,
                                         conn->current_trans->recv_buffer_end);

      if (conn->current_trans->http_version == SM_HTTP_0_9)
      {
        /*********************************************************************/
        /* HTTP 0.9 doesn't have negative responses so we just have to close */
        /* the connection.                                                   */
        /*********************************************************************/
        TRC_DETAIL((TB, "HTTP 0.9 Error"));
        rc = SM_CONNECTION_CLOSED;
        sm_worker_close_connection(conn);
      }
      else
      {
        /*********************************************************************/
        /* HTTP/1.X allows us to send error responses                        */
        /*********************************************************************/
        TRC_DETAIL((TB, "HTTP 1.X error"));
#ifdef OSP_SUPPORT_SCEP
        sm_worker_send_canned_response(conn->current_trans, SM_RESPONSE_400);
#else
        sm_worker_send_canned_response(conn->current_trans, SM_RESPONSE_501);
#endif
        conn->current_trans = NULL;
      }
      goto EXIT_LABEL;
    }

#ifdef OSP_SUPPORT_SCEP
    /*************************************************************************/
    /* We now definitely have a well formatted HTTP request so we can        */
    /* increment the statistics.                                             */
    /*************************************************************************/
    OSP_GET_LOCK(&smg_statistics_mutex);
    smg_statistics.scep_requests++;
    TRC_DETAIL((TB, "Now %u scep_requests", smg_statistics.scep_requests));
    OSP_RELEASE_LOCK(&smg_statistics_mutex);

    /*************************************************************************/
    /* Increment outstanding_count so that we remember to wait for this      */
    /* transaction to complete.                                              */
    /*************************************************************************/
    OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
    conn->wa_locked_outstanding_count++;
    OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);

    this_transaction = conn->current_trans;
    conn->current_trans = NULL;

#ifdef OSP_MEASURE_PERFORMANCE
    /*************************************************************************/
    /* Set the transaction processing start time.                            */
    /*************************************************************************/
    gettimeofday(&(this_transaction->start_time_processing), NULL);
    TRC_DETAIL((TB, "Set transaction processing start time:"
                " trans=%p time=%ld:%06ld",
                this_transaction,
                this_transaction->start_time_processing.tv_sec,
                this_transaction->start_time_processing.tv_usec));
#endif

    /*************************************************************************/
    /* Call SCEP to process the transaction.                                 */
    /*************************************************************************/
    error = scep_request_received(this_transaction, operation);

    if (error != OSP_SUCCESS)
    {
      TRC_ERROR((TB, "Failure of scep_request_received() %d", error));
      OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
      conn->wa_locked_outstanding_count--;
      OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);

      if (this_transaction->http_version == SM_HTTP_0_9)
      {
        /*********************************************************************/
        /* HTTP 0.9 doesn't have negative responses so we just have to close */
        /* the connection.                                                   */
        /*********************************************************************/
        TRC_DETAIL((TB, "HTTP0.9 error"));
        rc = SM_CONNECTION_CLOSED;
        sm_worker_close_connection(conn);
      }
      else
      {
        TRC_DETAIL((TB, "HTTP/1.X error"));
        sm_worker_send_canned_response(this_transaction, SM_RESPONSE_400);
      }
      goto EXIT_LABEL;
    }
#endif
  }
  else
  {
    /*************************************************************************/
    /* If the buffer is full, we need a bigger buffer.  Otherwise we just    */
    /* need to wait for some more data to roll in.                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Don't yet have whole SCEP request"));

    if (conn->current_trans->recv_buffer_end ==
                                         conn->current_trans->recv_buffer_size)
    {
      TRC_NORMAL((TB, "Buffer Full - need to realloc"));
      *new_buffer_size = conn->current_trans->recv_buffer_size * 2;
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sm_worker_received_get */

/**PROC+**********************************************************************/
/* Name:      sm_worker_send_canned_response                                 */
/*                                                                           */
/* Purpose:   Queues pre-allocated standard error response.  This cannot be  */
/*            used for HTTP 0.9 as this protocol does not have error         */
/*            responses.                                                     */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT transaction - transaction for which response is         */
/*                                 required                                  */
/*            IN     rsp         - number of canned response                 */
/*                                                                           */
/* Operation: Get the appropriate response for the HTTP version and error    */
/*            code.                                                          */
/*            Copy the precanned response into the send buffer which is      */
/*            guaranteed to be large enough for any response.                */
/*            Queue the transaction on the send queue and wakeup the worker  */
/*            if necessary.                                                  */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_worker_send_canned_response(TRANS_INFO *transaction, int rsp)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char wakeup_byte = ' ';
  int error;
  const char *canned_buffer = NULL;
  OSP_RC rc;

  TRC_ENTRY("sm_worker_send_canned_response");

  TRC_ASSERT((transaction->http_version == SM_HTTP_VERSION_UNDETECTED) ||
             (transaction->http_version == SM_HTTP_1_0) ||
             (transaction->http_version == SM_HTTP_1_1),
             (TB, "No canned responses for HTTP version %d",
              transaction->http_version));

  TRC_NORMAL((TB, "Send %d response version %d on %p", rsp,
                          transaction->http_version, transaction->connection));

  OSP_GET_LOCK(&transaction->mutex);
  transaction->send_buffer_start = 0;

  switch (transaction->http_version)
  {
    case SM_HTTP_VERSION_UNDETECTED:
    case SM_HTTP_1_0:
      /***********************************************************************/
      /* HTTP 1.0                                                            */
      /***********************************************************************/
      TRC_DETAIL((TB, "HTTP 1.0"));

      switch (rsp)
      {
        case SM_RESPONSE_400:
          TRC_DETAIL((TB, "400"));
          canned_buffer = sm_const_canned_400_10;
          transaction->send_buffer_end = sizeof(sm_const_canned_400_10);
          break;

        case SM_RESPONSE_500:
          TRC_DETAIL((TB, "500"));
          canned_buffer = sm_const_canned_500_10;
          transaction->send_buffer_end = sizeof(sm_const_canned_500_10);
          break;

        case SM_RESPONSE_501:
          TRC_DETAIL((TB, "501"));
          canned_buffer = sm_const_canned_501_10;
          transaction->send_buffer_end = sizeof(sm_const_canned_501_10);
          break;

        default:
          TRC_ASSERT((FALSE), (TB, "Unknown response %d", rsp));
          break;
      }
      break;

    case SM_HTTP_1_1:
      /***********************************************************************/
      /* HTTP 1.1                                                            */
      /***********************************************************************/
      TRC_DETAIL((TB, "HTTP 1.1"));

      switch (rsp)
      {
        case SM_RESPONSE_400:
          TRC_DETAIL((TB, "400"));
          canned_buffer = sm_const_canned_400_11;
          transaction->send_buffer_end = sizeof(sm_const_canned_400_11);
          break;

        case SM_RESPONSE_500:
          TRC_DETAIL((TB, "500"));
          canned_buffer = sm_const_canned_500_11;
          transaction->send_buffer_end = sizeof(sm_const_canned_500_11);
          break;

        case SM_RESPONSE_501:
          TRC_DETAIL((TB, "501"));
          canned_buffer = sm_const_canned_501_11;
          transaction->send_buffer_end = sizeof(sm_const_canned_501_11);
          break;

        default:
          TRC_ASSERT((FALSE), (TB, "Unknown response %d", rsp));
          break;
      }
      break;
  }

  /***************************************************************************/
  /* Now copy it into the send buffer                                        */
  /***************************************************************************/
  TRC_ASSERT((canned_buffer != NULL), (TB, "No canned response"));
  TRC_ASSERT((transaction->send_buffer_end <= transaction->send_buffer_size),
             (TB, "canned response doesn't fit in buffer: %d > %d",
                 transaction->send_buffer_end, transaction->send_buffer_size));

  TRC_DETAIL((TB, "Copying %u bytes of canned response into buffer at %p",
                      transaction->send_buffer_end, transaction->send_buffer));
  memcpy(transaction->send_buffer,
         canned_buffer,
         transaction->send_buffer_end);

  OSP_RELEASE_LOCK(&transaction->mutex);

  /***************************************************************************/
  /* Now put the response on the send queue and wakeup the worker.           */
  /***************************************************************************/
  rc = sm_queue_transaction(transaction, FALSE);
  if (rc != OSP_SUCCESS)
  {
    TRC_ALERT((TB, "Failed to queue transaction %p: rc=%d", transaction, rc));
  }

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* sm_worker_send_canned_response */

