
/**MOD+***********************************************************************/
/* Module:   smworker.c                                                      */
/*                                                                           */
/* Purpose:  SM worker thread functions                                      */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.45               $ $Modtime::   Jan 11 2001 15:56:46   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <smincl.h>

/**PROC+**********************************************************************/
/* Name:      sm_worker_thread                                               */
/*                                                                           */
/* Purpose:   Worker thread procedure.                                       */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT this_thread - data structure that controls this thread  */
/*                                                                           */
/* Operation: Select on read_set, write_set and error_set.                   */
/*            Terminate on error.                                            */
/*            Check the wakeup pipe for work                                 */
/*            Check the notify pipe next as this may terminate us so we      */
/*            need not check the other fds.                                  */
/*            Check all of the socket file descriptors for work.             */
/*                                                                           */
/*            When all work is done, see if we need to SSL renegotiate on    */
/*            any connections and recalculate when we next need to           */
/*            renegotiate.                                                   */
/*                                                                           */
/**PROC-**********************************************************************/

void *sm_worker_thread(void *in_param)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  int error;
  fd_set read_set;
  fd_set write_set;
  fd_set error_set;
  struct timeval timeout;
  time_t now;
  CONN_INFO *conn;
  CONN_INFO *next_conn;
  THREAD_INFO *this_thread = (THREAD_INFO *)in_param;
  struct timeval cycle_end_time;
#ifdef OSP_MEASURE_PERFORMANCE
  static struct timeval cycle_start_time = { 0, 0 };
#endif

  TRC_THREAD_ENTRY("sm_worker_thread");

  TRC_ASSERT((this_thread != NULL), (TB, "No thread info pointer"));

  /***************************************************************************/
  /* Initialization: set up the thread specific data with the SM thread      */
  /* identifier.  This is used to prevent calls to OSP term on this thread   */
  /* (which would result in a hang)                                          */
  /***************************************************************************/
  if (pthread_setspecific(sm_thread_key, sm_key_name) != 0)
  {
    TRC_ERROR((TB, "Failed to set thread specific data: quitting"));
    goto EXIT_LABEL;
  }
  timeout.tv_usec = 0;

  while (TRUE)
  {
    OSP_GET_LOCK(&this_thread->wa_locked.mutex_wa);

    memcpy(&read_set,  &this_thread->read_set,  sizeof(read_set));
    memcpy(&write_set, &this_thread->wa_locked.write_set, sizeof(write_set));
    memcpy(&error_set, &this_thread->error_set, sizeof(error_set));

    OSP_RELEASE_LOCK(&this_thread->wa_locked.mutex_wa);

    OSP_GET_LOCK(&this_thread->lw_locked.mutex_lw);

    /*************************************************************************/
    /* Sanity check on connection counts.                                    */
    /*************************************************************************/
    if ((this_thread->ssl_data_buffered_count >= 0) &&
        (this_thread->ssl_data_buffered_count <=
         this_thread->lw_locked.assigned_conns))
    {
      TRC_DETAIL((TB, "Connections with SSL buffered data:"
                  " ssl_data_buffered_count=%d assigned_conns=%d",
                  this_thread->ssl_data_buffered_count,
                  this_thread->lw_locked.assigned_conns));
    }
    else
    {
      /***********************************************************************/
      /* Sanity check failed: reduce ssl_data_buffered_count to equal the    */
      /* number of assigned_conns.  In the case where assigned_conns has     */
      /* dropped to zero, this will clear the error in                       */
      /* ssl_data_buffered_count.                                            */
      /***********************************************************************/
      TRC_ALERT((TB, "Error in SSL data buffered count:"
                 " ssl_data_buffered_count=%d assigned_conns=%d",
                 this_thread->ssl_data_buffered_count,
                 this_thread->lw_locked.assigned_conns));
      this_thread->ssl_data_buffered_count =
                                         this_thread->lw_locked.assigned_conns;
    }

    OSP_RELEASE_LOCK(&this_thread->lw_locked.mutex_lw);

    if (this_thread->ssl_data_buffered_count != 0)
    {
      /***********************************************************************/
      /* Some of our SSL connections may have readable data buffered within  */
      /* SSL.  Set a zero timeout to find out which other connections might  */
      /* be ready for processing without hanging around.                     */
      /***********************************************************************/
      TRC_DETAIL((TB, "Set zero timeout for select"));
      timeout.tv_sec = 0;
    }
    else
    {
      /***********************************************************************/
      /* Set default timeout for dynamic thread busyness measurement.        */
      /***********************************************************************/
      TRC_DETAIL((TB, "Default timeout for select"));
      timeout.tv_sec = SM_SELECT_TIMEOUT;

      /***********************************************************************/
      /* Set smaller timeout for renegotiation if necessary.                 */
      /***********************************************************************/
      now = time(NULL);
      if ((this_thread->first_renegotiate != COM_RENEGOTIATE_NOT_SET) &&
                       (this_thread->first_renegotiate - now < timeout.tv_sec))
      {
        timeout.tv_sec = this_thread->first_renegotiate - now;
      }
      TRC_DETAIL((TB, "Select timeout %ld seconds", timeout.tv_sec));
    }

    /*************************************************************************/
    /* If we are about to block, calculate a measure of how busy this thread */
    /* has been since it last blocked.  The measure we calculate is (elapsed */
    /* time since end of last block) / (elapsed time since start of last     */
    /* block), as a percentage.                                              */
    /*************************************************************************/
    gettimeofday(&cycle_end_time, NULL);
    if (timeout.tv_sec != 0)
    {
      TRC_DETAIL((TB, "Calculate thread busy-ness"));

      /***********************************************************************/
      /* Convert the stored absolute select times for this thread into       */
      /* elapsed times since the beginning and end of the last blocking      */
      /* select call.                                                        */
      /***********************************************************************/
      this_thread->select_start_time.tv_sec =
               (cycle_end_time.tv_sec - this_thread->select_start_time.tv_sec);
      this_thread->select_start_time.tv_usec =
             (cycle_end_time.tv_usec - this_thread->select_start_time.tv_usec);
      this_thread->select_return_time.tv_sec =
              (cycle_end_time.tv_sec - this_thread->select_return_time.tv_sec);
      this_thread->select_return_time.tv_usec =
            (cycle_end_time.tv_usec - this_thread->select_return_time.tv_usec);

      /***********************************************************************/
      /* If the elapsed time in seconds is more than 40 seconds, ignore the  */
      /* microseconds; otherwise combine the tv_sec and tv_usec fields into  */
      /* a single number in units of 10^-5 s.  This is to avoid overflowing  */
      /* longs.                                                              */
      /***********************************************************************/
      if (this_thread->select_start_time.tv_sec < 40)
      {
        this_thread->select_start_time.tv_sec *= 100000;
        this_thread->select_start_time.tv_sec +=
                                   this_thread->select_start_time.tv_usec / 10;
        this_thread->select_return_time.tv_sec *= 100000;
        this_thread->select_return_time.tv_sec +=
                                  this_thread->select_return_time.tv_usec / 10;
        TRC_DETAIL((TB, "Elapsed times (10^-5s): start=%ld return=%ld",
                    this_thread->select_start_time.tv_sec,
                    this_thread->select_return_time.tv_sec));
      }
      else
      {
        TRC_DETAIL((TB, "Elapsed times (s): start=%ld return=%ld",
                    this_thread->select_start_time.tv_sec,
                    this_thread->select_return_time.tv_sec));
      }

      /***********************************************************************/
      /* Calculate busyness percentage, with 80%-20% smoothing.              */
      /***********************************************************************/
      OSP_GET_LOCK(&this_thread->lw_locked.mutex_lw);
      this_thread->lw_locked.busyness *= 4;
      this_thread->lw_locked.busyness /= 5;
      if (this_thread->select_start_time.tv_sec == 0)
      {
        /*********************************************************************/
        /* Avoid division by zero in case where cycle apparently took no     */
        /* time at all.                                                      */
        /*********************************************************************/
        TRC_ALERT((TB, "Cycle appeared to take no time"));
      }
      else
      {
        this_thread->lw_locked.busyness +=
                                (this_thread->select_return_time.tv_sec * 20) /
                                         this_thread->select_start_time.tv_sec;
      }
      TRC_DETAIL((TB, "Calculated thread busy-ness is now %d%%",
                  this_thread->lw_locked.busyness));
      OSP_RELEASE_LOCK(&this_thread->lw_locked.mutex_lw);

      /***********************************************************************/
      /* Store the current time in the select_start_time fields.             */
      /***********************************************************************/
      this_thread->select_start_time.tv_sec = cycle_end_time.tv_sec;
      this_thread->select_start_time.tv_usec = cycle_end_time.tv_usec;
    }

#ifdef OSP_MEASURE_PERFORMANCE
    if (cycle_start_time.tv_usec != 0)
    {
      SM_ACCUMULATE(smg_select_cycle, cycle_start_time, cycle_end_time);
    }
#endif

    error = select(this_thread->nfds,
                   &read_set,
                   &write_set,
                   &error_set,
                   &timeout);

#ifdef OSP_MEASURE_PERFORMANCE
    gettimeofday(&cycle_start_time, NULL);
#endif

    /*************************************************************************/
    /* If that was a blocking select, note the time that the select          */
    /* returned.                                                             */
    /*************************************************************************/
    if (timeout.tv_sec != 0)
    {
      TRC_DETAIL((TB, "Store select return time"));
      gettimeofday(&(this_thread->select_return_time), NULL);
    }

    if (error == -1)
    {
      /***********************************************************************/
      /* Error on select                                                     */
      /***********************************************************************/
      TRC_OS_ERROR((TB, "Error on select"));
      sm_worker_terminate(this_thread);
    }
    else
    {
      /***********************************************************************/
      /* No timeout - actual work to do.  We must now check all of our file  */
      /* descriptors.  If the listener has closed the notify file            */
      /* descriptor, there's no point in checking anything else.             */
      /***********************************************************************/
      TRC_DETAIL((TB, "select returned %d", error));
      sm_worker_check_wakeup_fd(this_thread, &read_set, &write_set);

      TRC_DETAIL((TB, "Checking for notifications"));
      if (sm_worker_check_notify_fd(this_thread, &read_set, &error_set))
      {
        TRC_DETAIL((TB, "Now check all of the connections"));

        for (conn = (CONN_INFO *)LIST_GET_FIRST(&this_thread->conn_list);
             conn != NULL;
             conn = next_conn)
        {
          /*******************************************************************/
          /* sm_worker_check_fd may close the connection and free its        */
          /* CONN_INFO structure, so set up pointer to the next connection   */
          /* in the list before calling it.                                  */
          /*******************************************************************/
          next_conn = (CONN_INFO *)
                          LIST_GET_NEXT(&this_thread->conn_list, &conn->links);

          sm_worker_check_fd(conn, &read_set, &write_set, &error_set);
        }
      }
    }

    /*************************************************************************/
    /* Now check timeouts to ensure that if we are consistently busy on this */
    /* worker, we still look at the timeout values.                          */
    /*************************************************************************/
    now = time(NULL);

    TRC_DETAIL((TB, "now - %ld, first_renegotiate - %ld%s", now,
                                                this_thread->first_renegotiate,
                  (this_thread->first_renegotiate == COM_RENEGOTIATE_NOT_SET) ?
                                                           " (not set)" : ""));

    if ((this_thread->first_renegotiate != COM_RENEGOTIATE_NOT_SET) &&
                                       (this_thread->first_renegotiate <= now))
    {
      this_thread->first_renegotiate = COM_RENEGOTIATE_NOT_SET;

      TRC_DETAIL((TB, "Checking for connections that need renegotiating"));
      for (conn = (CONN_INFO *)LIST_GET_FIRST(&this_thread->conn_list);
           conn != NULL;
           conn = (CONN_INFO *)
                          LIST_GET_NEXT(&this_thread->conn_list, &conn->links))
      {
        /*********************************************************************/
        /* If this is an SSL connection thats due renegotiation, renegotiate */
        /* and reset next_renegotiate.                                       */
        /*********************************************************************/
        if ((conn->type == SM_SECURE_SOCKET) &&
            (conn->next_renegotiate != COM_RENEGOTIATE_NOT_SET) &&
            (conn->next_renegotiate <= now))
        {
          TRC_NORMAL((TB, "Renegotiate on conn %p", conn));
          conn->next_renegotiate = now + SM_RENEGOTIATION_TIME;

          /*******************************************************************/
          /* The return code from SSL_renegotiate would only be something    */
          /* other than 1 if we had disabled renegotiation (which we don't). */
          /* The renegotiation itself happens asynchronously, so any         */
          /* failures will show up during SSL_read/write.                    */
          /*******************************************************************/
          error = SSL_renegotiate(conn->ssl);
          if (error != 1)
          {
            TRC_ALERT((TB, "SSL_renegotiate returned %d", error));
          }

          /*******************************************************************/
          /* Set up the earliest renegotiation time on this connection.      */
          /*******************************************************************/
          if ((this_thread->first_renegotiate == COM_RENEGOTIATE_NOT_SET) ||
                     (conn->next_renegotiate < this_thread->first_renegotiate))
          {
            this_thread->first_renegotiate = conn->next_renegotiate;
            TRC_DETAIL((TB, "first_renegotiate now %ld",
                                              this_thread->first_renegotiate));
          }
        }
      }
    }
  }

EXIT_LABEL:
  TRC_ASSERT((FALSE), (TB, "Worker thread loop %p has exited", this_thread));

  TRC_THREAD_EXIT();

  return(NULL);

} /* sm_worker_thread */

/**PROC+**********************************************************************/
/* Name:      sm_worker_close_connection                                     */
/*                                                                           */
/* Purpose:   Stops sockets operations on this connection.  If there are no  */
/*            outstanding transactions on this connection,  the socket is    */
/*            closed and the connection is deleted.                          */
/*            (An outstanding transaction is one that is not in the control  */
/*            of SM.)                                                        */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT conn - connection to close                              */
/*                                                                           */
/* Operation: Remove this FD from our fd_sets so we do no further work on    */
/*            it.                                                            */
/*            Set the state to SM_CONN_STATE_CLOSED, shutdown the SSL        */
/*            connection and free the current transaction.                   */
/*                                                                           */
/*            If there are no outstanding transactions, call the client      */
/*            disconnect callback, free the transactions owned by this       */
/*            connection, remove the connection from the worker's conn_list  */
/*            and free the connection.                                       */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_worker_close_connection(CONN_INFO *conn)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;
  int error;
  TRANS_INFO *transaction;

  TRC_ENTRY("sm_worker_close_connection");

  TRC_NORMAL((TB, "Closing connection %p", conn));

  OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);

  /***************************************************************************/
  /* Stop any further work on this connection.                               */
  /***************************************************************************/
  FD_CLR(conn->sock_handle, &conn->thread->read_set);
  FD_CLR(conn->sock_handle, &conn->thread->wa_locked.write_set);
  FD_CLR(conn->sock_handle, &conn->thread->error_set);

  /***************************************************************************/
  /* Update the thread's SSL data buffered count if necessary.               */
  /***************************************************************************/
  if (conn->ssl_data_buffered)
  {
    conn->ssl_data_buffered = FALSE;
    conn->thread->ssl_data_buffered_count--;
    TRC_DETAIL((TB, "Now %d SSL connections with data buffered",
                conn->thread->ssl_data_buffered_count));
  }

  if (conn->thread->nfds == conn->sock_handle + 1)
  {
    /*************************************************************************/
    /* Recalculate nfds                                                      */
    /*************************************************************************/
    SM_RECALCULATE_NFDS(conn->thread);
    TRC_DETAIL((TB, "nfds is now %d", conn->thread->nfds));
  }

  /***************************************************************************/
  /* Set the state to closedown.                                             */
  /***************************************************************************/
  if (conn->conn_state != SM_CONN_STATE_CLOSED)
  {
    TRC_DETAIL((TB, "Set state to SM_CONN_STATE_CLOSED"));
    conn->conn_state = SM_CONN_STATE_CLOSED;

    if (conn->type == SM_SECURE_SOCKET)
    {
      /***********************************************************************/
      /* Cleanup SSL - it is safe to do this as we know that no further work */
      /* can occur in this socket.                                           */
      /***********************************************************************/
      TRC_DETAIL((TB, "Connection was secure"));
      if (conn->ssl != NULL)
      {
        TRC_DETAIL((TB, "shutdown SSL"));
        SSL_shutdown(conn->ssl);
        SSL_free(conn->ssl);
        conn->ssl = NULL;
      }
      conn->next_renegotiate = COM_RENEGOTIATE_NOT_SET;
    }

    /*************************************************************************/
    /* Delete the current transaction                                        */
    /*************************************************************************/
    sm_free_transaction(conn->current_trans, TRUE);
    conn->current_trans = NULL;
  }

  TRC_DETAIL((TB, "outstanding count on connection %p is now %u", conn,
                                           conn->wa_locked_outstanding_count));

  /***************************************************************************/
  /* If we don't have any outstanding transactions, it is now safe to delete */
  /* this connection.                                                        */
  /***************************************************************************/
  if (conn->wa_locked_outstanding_count == 0)
  {
    /*************************************************************************/
    /* Tell the application that this connection has gone.                   */
    /*************************************************************************/
    TRC_NORMAL((TB, "Now really close the connection"));

    osp_get_recursive_lock(&smg_listener_info.callback_mutex);
    if (smg_listener_info.cd_cb != NULL)
    {
      TRC_DETAIL((TB, "Calling cd_cb"));
      (*smg_listener_info.cd_cb)(&conn->client_id);
    }
    osp_release_recursive_lock(&smg_listener_info.callback_mutex);

    /*************************************************************************/
    /* Free all queued transactions.                                         */
    /*************************************************************************/
    while (transaction =
                     (TRANS_INFO *)LIST_GET_FIRST(&conn->wa_locked_send_queue))
    {
      LIST_REMOVE_ITEM(&transaction->links);
      sm_free_transaction(transaction, TRUE);
    }

    /*************************************************************************/
    /* Close the socket                                                      */
    /*************************************************************************/
    close(conn->sock_handle);
    conn->sock_handle = SM_INVALID_SOCKET;

    LIST_REMOVE_ITEM(&conn->links);

    OSP_GET_LOCK(&conn->thread->lw_locked.mutex_lw);
    conn->thread->lw_locked.assigned_conns--;
    OSP_RELEASE_LOCK(&conn->thread->lw_locked.mutex_lw);

    OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
    rm_release_structure(conn, RM_CONN_INFO);
    conn = NULL;
  }

EXIT_LABEL:
  if (conn != NULL)
  {
    OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
  }

  TRC_EXIT();

  return;

} /* sm_worker_close_connection */

/**PROC+**********************************************************************/
/* Name:      sm_worker_terminate                                            */
/*                                                                           */
/* Purpose:   Called when listener thread terminates this thread.            */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT this_thread - thread to terminate                       */
/*                                                                           */
/* Operation: Closes each connection and if there are no connections,        */
/*            deletes this thread.                                           */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_worker_terminate(THREAD_INFO *this_thread)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  CONN_INFO *connection;
  OSP_RC rc;
  int error;

  TRC_ENTRY("sm_worker_terminate");

  TRC_NORMAL((TB, "Terminating worker %p", this_thread));

  /***************************************************************************/
  /* Close all connections.                                                  */
  /***************************************************************************/
  for (connection = (CONN_INFO *)LIST_GET_FIRST(&this_thread->conn_list);
       connection != NULL;
       connection = (CONN_INFO *)
                    LIST_GET_NEXT(&this_thread->conn_list, &connection->links))
  {
    sm_worker_close_connection(connection);
  }

  /***************************************************************************/
  /* Need to get worker_list_mutex before mutex_lw to avoid deadlocks.       */
  /***************************************************************************/
  OSP_GET_LOCK(&smg_listener_info.worker_list_mutex);
  OSP_GET_LOCK(&this_thread->lw_locked.mutex_lw);

  TRC_DETAIL((TB, "Assigned conns %d", this_thread->lw_locked.assigned_conns));

  if (this_thread->lw_locked.assigned_conns == 0)
  {
    /*************************************************************************/
    /* All of the connections have gone away.  We can safely delete this     */
    /* worker.                                                               */
    /*************************************************************************/
    TRC_NORMAL((TB, "Now actually delete this worker"));

    close(this_thread->notification_pipe_out);
    close(this_thread->wakeup_pipe_out);

    OSP_GET_LOCK(&this_thread->wa_locked.mutex_wa);
    close(this_thread->wa_locked.wakeup_pipe_in);

    if (this_thread->notification_pipe_in != SM_INVALID_SOCKET)
    {
      TRC_DETAIL((TB, "Close notification pipe"));
      close(this_thread->notification_pipe_in);
    }

    LIST_REMOVE_ITEM(&this_thread->links);

    OSP_RELEASE_LOCK(&this_thread->wa_locked.mutex_wa);
    OSP_RELEASE_LOCK(&this_thread->lw_locked.mutex_lw);
    osp_destroy_mutex(&this_thread->wa_locked.mutex_wa);
    osp_destroy_mutex(&this_thread->lw_locked.mutex_lw);

    rm_release_structure(this_thread, RM_THREAD_INFO);

    OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);

    /*************************************************************************/
    /* Clean up OpenSSL per-thread state.                                    */
    /*************************************************************************/
    ERR_remove_state(0);

    rm_notify_thread_exit(RM_WORKER_THREAD);

    TRC_THREAD_EXIT();

    pthread_exit(NULL);
  }

  OSP_RELEASE_LOCK(&this_thread->lw_locked.mutex_lw);
  OSP_RELEASE_LOCK(&smg_listener_info.worker_list_mutex);

  TRC_EXIT();

  return;

} /* sm_worker_terminate */

/**PROC+**********************************************************************/
/* Name:      sm_worker_check_notify_fd                                      */
/*                                                                           */
/* Purpose:   Checks whether select() has returned work for the notify FD    */
/*                                                                           */
/* Returns:   TRUE  - continue processing                                    */
/*            FALSE - skip processing of sockets as we're shutting down      */
/*                                                                           */
/* Params:    IN    thread    - this thread                                  */
/*            IN    read_set  - set of read FDs returned by select()         */
/*            IN    error_set - set of error FDs returned by select()        */
/*                                                                           */
/* Operation: If an error is set on this FD, terminate the thread and return */
/*            FALSE.                                                         */
/*                                                                           */
/*            If there is data to read on this FD, this is a new connection  */
/*            for this thread to handle. Read the CONN_INFO pointer from the */
/*            pipe and add the new FD to the read_set and error_set.         */
/*                                                                           */
/*            If this is a secure connection, initiate the SSL handshake,    */
/*            otherwise call the client verification callback.               */
/*                                                                           */
/*            If there are any errors, close the connection.                 */
/*                                                                           */
/**PROC-**********************************************************************/

int sm_worker_check_notify_fd(THREAD_INFO *thread,
                              fd_set *read_set,
                              fd_set *error_set)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC int_rc;
  int error;
  int rc = TRUE;
  CONN_INFO *new_connection;
  int now;

  TRC_ENTRY("sm_worker_check_notify_fd");

  if (FD_ISSET(thread->notification_pipe_out, error_set))
  {
    /*************************************************************************/
    /* Pipe to listener has gone.  Shut down this thread.                    */
    /*************************************************************************/
    TRC_ERROR((TB, "Error on notify pipe - shutdown"));
    sm_worker_terminate(thread);
    rc = FALSE;
  }
  else if (FD_ISSET(thread->notification_pipe_out, read_set))
  {
    /*************************************************************************/
    /* Listener thread is sending us a pointer to a new CONN_INFO structure. */
    /*************************************************************************/
    TRC_DETAIL((TB, "Read on notify pipe"));
    error = read(thread->notification_pipe_out,
               thread->new_connection + thread->new_connection_length,
               sizeof(thread->new_connection) - thread->new_connection_length);

    if ((error == 0) || (error == -1))
    {
      /***********************************************************************/
      /* Listener has closed notification pipe.  Time for this worker to     */
      /* terminate.                                                          */
      /***********************************************************************/
      if (error == -1)
      {
        TRC_OS_ERROR((TB, "Error on notification pipe"));
      }
      TRC_NORMAL((TB, "Notification pipe closed - terminating"));
      sm_worker_terminate(thread);
      rc = FALSE;
      goto EXIT_LABEL;
    }

    thread->new_connection_length += error;

    TRC_DETAIL((TB, "Now got %d bytes of pointer",
                                               thread->new_connection_length));

    if (thread->new_connection_length == sizeof(thread->new_connection))
    {
      /***********************************************************************/
      /* Got the connection pointer.                                         */
      /***********************************************************************/
      new_connection = *((CONN_INFO **)thread->new_connection);

      TRC_NORMAL((TB, "Thread %p has connection %p", thread, new_connection));

      memset(thread->new_connection, 0, sizeof(thread->new_connection));
      thread->new_connection_length = 0;

      LIST_INSERT_TAIL(&thread->conn_list, &new_connection->links);

      FD_SET(new_connection->sock_handle, &thread->read_set);

      if (thread->nfds <= new_connection->sock_handle)
      {
        thread->nfds = new_connection->sock_handle + 1;
        TRC_DETAIL((TB, "Increase nfds to %d", thread->nfds));
      }
      FD_SET(new_connection->sock_handle, &thread->error_set);

      /***********************************************************************/
      /* If this is a secure connection, start the SSL handshake             */
      /***********************************************************************/
      if (new_connection->type == SM_SECURE_SOCKET)
      {
        TRC_DETAIL((TB, "Secure connection"));

        new_connection->ssl = SSL_new(smg_listener_info.ssl_ctx);
        TRC_DETAIL((TB, "New SSL: connection=%p ssl=%p",
                                         new_connection, new_connection->ssl));

        if (new_connection->ssl == NULL)
        {
          OSP_OPENSSL_ERROR((TB, "SSL_new failed"));
          sm_worker_close_connection(new_connection);
          goto EXIT_LABEL;
        }

        error = SSL_set_fd(new_connection->ssl, new_connection->sock_handle);
        TRC_DETAIL((TB, "Set SSL FD: connection=%p ssl=%p fd=%d",
            new_connection, new_connection->ssl, new_connection->sock_handle));

        if (error == 0)
        {
          OSP_OPENSSL_ERROR((TB, "SSL_set_fd failed"));
          sm_worker_close_connection(new_connection);
          goto EXIT_LABEL;
        }

        /*********************************************************************/
        /* Set the connection info pointer as SSL's user data.  This allows  */
        /* us later to get the correct connection on the client verify       */
        /* callback.                                                         */
        /*********************************************************************/
        SSL_set_app_data(new_connection->ssl, new_connection);
        TRC_DETAIL((TB, "Set SSL app data: connection=%p ssl=%p",
                                         new_connection, new_connection->ssl));

        error = SSL_accept(new_connection->ssl);
        TRC_DETAIL((TB, "Accept SSL connection: connection=%p ssl=%p rc=%d",
                                  new_connection, new_connection->ssl, error));

        if (error <= 0)
        {
          error = SSL_get_error(new_connection->ssl, error);
          TRC_DETAIL((TB, "More to do: connection=%p ssl=%p error=%d",
                                  new_connection, new_connection->ssl, error));

          switch (error)
          {
            case SSL_ERROR_NONE:
            case SSL_ERROR_ZERO_RETURN:
            case SSL_ERROR_SYSCALL:
            case SSL_ERROR_SSL:
              /***************************************************************/
              /* These return codes all indicate an error in this state.     */
              /***************************************************************/
              OSP_OPENSSL_ERROR((TB, "SSL_accept failure"));
              sm_worker_close_connection(new_connection);
              goto EXIT_LABEL;
              break;

            case SSL_ERROR_WANT_READ:
              /***************************************************************/
              /* Negotiation is going OK and we now need to read on this     */
              /* socket.  This doesn't require any action as we are already  */
              /* selectin for reads.                                         */
              /***************************************************************/
              TRC_DETAIL((TB, "SSL_ERROR_WANT_READ"));
              break;

            case SSL_ERROR_WANT_WRITE:
              /***************************************************************/
              /* Negotiation is going OK and we now need to write on this    */
              /* socket.                                                     */
              /***************************************************************/
              TRC_DETAIL((TB, "SSL_ERROR_WANT_WRITE"));
              OSP_GET_LOCK(&thread->wa_locked.mutex_wa);
              FD_SET(new_connection->sock_handle,
                     &thread->wa_locked.write_set);
              OSP_RELEASE_LOCK(&thread->wa_locked.mutex_wa);
              break;

            default:
              OSP_OPENSSL_ERROR((TB, "Illegal return code %d", error));
              sm_worker_close_connection(new_connection);
              goto EXIT_LABEL;
              break;
          }
        }
        else
        {
          TRC_ASSERT((error == 1), (TB, "Unknown SSL rc %d", error));
          TRC_DETAIL((TB, "SSL Handshake completed"));
        }

        now = time(NULL);

        new_connection->next_renegotiate = now + SM_RENEGOTIATION_TIME;

        if ((thread->first_renegotiate == COM_RENEGOTIATE_NOT_SET) ||
                (new_connection->next_renegotiate < thread->first_renegotiate))
        {
          TRC_DETAIL((TB, "New renegotiation time %d",
                                            new_connection->next_renegotiate));
          thread->first_renegotiate = new_connection->next_renegotiate;
        }

        /*********************************************************************/
        /* Update statistics                                                 */
        /*********************************************************************/
        OSP_GET_LOCK(&smg_statistics_mutex);
        smg_statistics.ssl_conns++;
        TRC_DETAIL((TB, "Now %u ssl_conns", smg_statistics.ssl_conns));
        OSP_RELEASE_LOCK(&smg_statistics_mutex);
      }
      else
      {
        TRC_DETAIL((TB, "Non-secure connection"));

        /*********************************************************************/
        /* Update statistics                                                 */
        /*********************************************************************/
        OSP_GET_LOCK(&smg_statistics_mutex);
        smg_statistics.nonssl_conns++;
        TRC_DETAIL((TB, "Now %u nonssl_conns", smg_statistics.nonssl_conns));
        OSP_RELEASE_LOCK(&smg_statistics_mutex);
      }

      /***********************************************************************/
      /* For non-secure connections, and SSL connections where the callback  */
      /* is not called during connection establishment, we need to           */
      /* explicitly call the applications client verification callback.      */
      /***********************************************************************/
      if (((new_connection->type == SM_SECURE_SOCKET) &&
           (!new_connection->called_cv_cb) &&
           SSL_is_init_finished(new_connection->ssl)) ||
          (new_connection->type == SM_NON_SECURE_SOCKET))
      {
        TRC_DETAIL((TB, "Call client verification callback"));

        if (!sm_call_verify_callback(new_connection, 0, NULL, 0, 0))
        {
          /*******************************************************************/
          /* Client does not accept connection.                              */
          /*******************************************************************/
          TRC_ALERT((TB, "Connection not accepted"));
          sm_worker_close_connection(new_connection);
        }
      }
    }
  }

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* sm_worker_check_notify_fd */

/**PROC+**********************************************************************/
/* Name:      sm_worker_check_wakeup_fd                                      */
/*                                                                           */
/* Purpose:   Checks whether select() has returned work for the wakeup FD    */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN    thread    - this thread                                  */
/*            IN    read_set  - set of read FDs returned by select()         */
/*            IN    write_set - set of write FDs returned by select()        */
/*                                                                           */
/* Operation: If there is data to read, read it, reset the sent_wakeup       */
/*            flag, and modify read_set and write_set so as to encourage     */
/*            the rapid dispatch of pending responses.                       */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_worker_check_wakeup_fd(THREAD_INFO *thread,
                               fd_set *read_set,
                               fd_set *write_set)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char dummy;
  int error;
  int fd;
#ifdef OSP_MEASURE_PERFORMANCE
  struct timeval now;
#endif

  TRC_ENTRY("sm_worker_check_wakeup_fd");

  /***************************************************************************/
  /* This pipe is just used to knock the worker out of the select call so    */
  /* that the select FDs can be updated.                                     */
  /***************************************************************************/
  if (FD_ISSET(thread->wakeup_pipe_out, read_set))
  {
    TRC_DETAIL((TB, "Woken up - read from wakeup pipe"));
    error = read(thread->wakeup_pipe_out, &dummy, 1);
    TRC_ASSERT((error == 1),
               (TB, "Failed to read wakeup - error = %d", error));

    OSP_GET_LOCK(&thread->wa_locked.mutex_wa);

    thread->wa_locked.sent_wakeup = 0;

    /*************************************************************************/
    /* In order to attempt to send out pending responses as soon as possible */
    /* (without waiting for the next select cycle), (i) set the write_set    */
    /* bits for all connections for which responses are pending ...          */
    /*************************************************************************/
    for (fd = 0; fd < thread->nfds; fd++)
    {
      if (FD_ISSET(fd, &thread->wa_locked.write_set))
      {
        FD_SET(fd, write_set);
      }
    }

    /*************************************************************************/
    /* ... and (ii) zero the current read_set, so that the whole of this     */
    /* select cycle is devoted to sending.                                   */
    /*************************************************************************/
    FD_ZERO(read_set);

#ifdef OSP_MEASURE_PERFORMANCE
    /*************************************************************************/
    /* Measure the time since the byte was written down this wakeup pipe.    */
    /*************************************************************************/
    gettimeofday(&now, NULL);
    SM_ACCUMULATE(smg_wake_up, thread->wa_locked.woken, now);
#endif

    OSP_RELEASE_LOCK(&thread->wa_locked.mutex_wa);

    TRC_DETAIL((TB, "Gone to sleep again"));
  }

  TRC_EXIT();

  return;

} /* sm_worker_check_wakeup_fd */

/**PROC+**********************************************************************/
/* Name:      sm_worker_check_fd                                             */
/*                                                                           */
/* Purpose:   Checks whether select() has returned work for a socket         */
/*                                                                           */
/* Returns:   Nothing                                                        */
/*                                                                           */
/* Params:    IN    conn      - this connection                              */
/*            IN    read_set  - set of read FDs returned by select()         */
/*            IN    write_set - set of write FDs returned by select()        */
/*            IN    error_set - set of error FDs returned by select()        */
/*                                                                           */
/* Operation: If there is an error on this FD close the connection.          */
/*                                                                           */
/*            If there is a read on this FD, get a new transaction if        */
/*            necessary.                                                     */
/*            For a non-secure connection, recv() into the current           */
/*            transaction.                                                   */
/*            For a secure connection, SSL_read() may read data or require   */
/*            a write operation for which the write_set needs to be updated. */
/*            If we received data, call sm_worker_received_data until it can */
/*            do no more processing.                                         */
/*                                                                           */
/*            If we can write on this FD, attempt to send all of the         */
/*            transactions in the send queue. If this succeeds, clear this   */
/*            FD from the write_set.                                         */
/*            Note for a secure connection SSL_write may require a read.     */
/*                                                                           */
/**PROC-**********************************************************************/

void sm_worker_check_fd(CONN_INFO *conn,
                        fd_set *read_set,
                        fd_set *write_set,
                        fd_set *error_set)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int error;
  int num_bytes;
  TRANS_INFO *transaction;
  int receive_result;
  struct timeval now;

  TRC_ENTRY("sm_worker_check_fd");

  TRC_DETAIL((TB, "Checking FD %d (conn %p)", conn->sock_handle, conn));

  if (FD_ISSET(conn->sock_handle, error_set))
  {
    /*************************************************************************/
    /* Error on this FD.                                                     */
    /*************************************************************************/
    TRC_ERROR((TB, "Error on file descriptor %d", conn->sock_handle));
    sm_worker_close_connection(conn);
    goto EXIT_LABEL;
  }

  if (FD_ISSET(conn->sock_handle, write_set))
  {
    /*************************************************************************/
    /* We can write on this connection.                                      */
    /*************************************************************************/
    TRC_DETAIL((TB, "Write set"));
    OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);

    /*************************************************************************/
    /* This connection may be marked closed if the underlying socket went    */
    /* away while there were still outstanding transactions.                 */
    /* sm_queue_transaction() will have added our socket handle to           */
    /* write_set, bringing us here.                                          */
    /*                                                                       */
    /* If all the outstanding transactions have completed, all we need to do */
    /* is make a final call to sm_close_connection() and this will free the  */
    /* CONN_INFO structure and all the unsent queued transactions.           */
    /*************************************************************************/
    if (conn->conn_state == SM_CONN_STATE_CLOSED)
    {
      if (conn->wa_locked_outstanding_count == 0)
      {
        TRC_NORMAL((TB, "All transactions returned for closed conn %p", conn));
        OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
        sm_worker_close_connection(conn);
        goto EXIT_LABEL;
      }
      else
      {
        TRC_DETAIL((TB, "Not attempting write for closed conn %p", conn));
        OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
        goto EXIT_LABEL;
      }
    }

    while (transaction =
                     (TRANS_INFO *)LIST_GET_FIRST(&conn->wa_locked_send_queue))
    {
      /***********************************************************************/
      /* Check that the transaction has the serial number that we expect for */
      /* the next response to send.  If it doesn't, we must wait for the     */
      /* expected next transaction to be inserted into the send queue.       */
      /***********************************************************************/
      if (transaction->serial != conn->last_response_serial + 1)
      {
        TRC_NORMAL((TB, "Out of order transaction serial numbers: %d, %d",
                    conn->last_response_serial, transaction->serial));
        FD_CLR(conn->sock_handle, &conn->thread->wa_locked.write_set);
        break;
      }

      OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);

      TRC_DETAIL((TB, "Attempting send for transaction %p", transaction));

#ifdef OSP_MEASURE_PERFORMANCE
      /***********************************************************************/
      /* Accumulate time up until this point.                                */
      /***********************************************************************/
      gettimeofday(&now, NULL);
      SM_ACCUMULATE(smg_until_start_send,
                    transaction->start_time_io,
                    now);
#endif

#ifdef OSP_MEASURE_TRANSACTION
      /***********************************************************************/
      /* Set the transaction start send time.                                */
      /***********************************************************************/
      SM_GETTIMES(transaction->timings[SM_TIMINGS_START_SEND]);
#endif

      if (conn->type == SM_NON_SECURE_SOCKET)
      {
        /*********************************************************************/
        /* Normal send                                                       */
        /*********************************************************************/
        TRC_DETAIL((TB, "Try to send() %d bytes",
               transaction->send_buffer_end - transaction->send_buffer_start));
        num_bytes =
           send(conn->sock_handle,
                transaction->send_buffer + transaction->send_buffer_start,
                transaction->send_buffer_end - transaction->send_buffer_start,
                0);

        if (num_bytes == -1)
        {
          if (errno == EWOULDBLOCK)
          {
            /*****************************************************************/
            /* Socket is full for the moment: we should select for writing   */
            /* again and carry on when that pops.                            */
            /*****************************************************************/
            TRC_ALERT((TB, "Socket temporarily full for writing"));
            OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
            FD_SET(conn->sock_handle, &conn->thread->wa_locked.write_set);
            OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
            goto EXIT_LABEL;
          }

          /*******************************************************************/
          /* Failed to write                                                 */
          /*******************************************************************/
          TRC_OS_ERROR((TB,
           "Failed to write %d bytes on socket %d, connection %p",
                 transaction->send_buffer_end - transaction->send_buffer_start,
                 conn->sock_handle, conn));
          sm_worker_close_connection(conn);
          goto EXIT_LABEL;
        }
      }
      else
      {
        /*********************************************************************/
        /* SSL send                                                          */
        /*********************************************************************/
        TRC_DETAIL((TB, "Try to SSL_write() %d bytes",
               transaction->send_buffer_end - transaction->send_buffer_start));
        num_bytes = SSL_write(conn->ssl,
                transaction->send_buffer + transaction->send_buffer_start,
                transaction->send_buffer_end - transaction->send_buffer_start);

        error = SSL_get_error(conn->ssl, num_bytes);

        switch (error)
        {
          case SSL_ERROR_NONE:
            /*****************************************************************/
            /* Successful write                                              */
            /*****************************************************************/
            TRC_DETAIL((TB, "SSL_write %d bytes", num_bytes));
            break;

          case SSL_ERROR_SSL:
          case SSL_ERROR_SYSCALL:
            /*****************************************************************/
            /* SSL error                                                     */
            /*****************************************************************/
            OSP_OPENSSL_ERROR((TB, "Error on SSL_write"));
            sm_worker_close_connection(conn);
            goto EXIT_LABEL;
            break;

          case SSL_ERROR_ZERO_RETURN:
            /*****************************************************************/
            /* Connection closed                                             */
            /*****************************************************************/
            TRC_DETAIL((TB, "Connection closed"));
            sm_worker_close_connection(conn);
            goto EXIT_LABEL;
            break;

          case SSL_ERROR_WANT_READ:
            /*****************************************************************/
            /* No data was sent as we need to read from that call - now need */
            /* to read.  Check we are actually listening on this socket (We  */
            /* might be flow controlled).                                    */
            /*****************************************************************/
            TRC_DETAIL((TB, "SSL_ERROR_WANT_READ"));
            FD_SET(conn->sock_handle, &conn->thread->read_set);
            goto EXIT_LABEL;
            break;

          case SSL_ERROR_WANT_WRITE:
            /*****************************************************************/
            /* No data from that call - now need to write.                   */
            /*****************************************************************/
            TRC_DETAIL((TB, "SSL_ERROR_WANT_WRITE"));
            OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
            FD_SET(conn->sock_handle, &conn->thread->wa_locked.write_set);
            OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
            goto EXIT_LABEL;
            break;

          default:
            OSP_OPENSSL_ERROR((TB, "Unknown error code %d", error));
            sm_worker_close_connection(conn);
            goto EXIT_LABEL;
            break;
        }
      }

      TRC_DETAIL((TB, "Sent %d bytes", num_bytes));
      OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);

      if (num_bytes !=
                 transaction->send_buffer_end - transaction->send_buffer_start)
      {
        /*********************************************************************/
        /* Haven't sent all of the response - stop reading on this           */
        /* connection to give some back pressure.                            */
        /*********************************************************************/
        TRC_DETAIL((TB,
              "Sent part of response - clear read bit to give back pressure"));
        transaction->send_buffer_start += num_bytes;

        FD_CLR(conn->sock_handle, &conn->thread->read_set);
        break;
      }

      /***********************************************************************/
      /* Increment the serial number of the last sent response.              */
      /***********************************************************************/
      conn->last_response_serial++;
      TRC_DETAIL((TB, "Sent response for conn %p transaction serial %d",
                  conn,
                  conn->last_response_serial));

#ifdef OSP_MEASURE_PERFORMANCE
      /***********************************************************************/
      /* Accumulate processing + I/O time statistics for this transaction.   */
      /***********************************************************************/
      gettimeofday(&now, NULL);
      TRC_DETAIL((TB, "Transaction finished: trans=%p time=%ld:%06ld",
                  transaction,
                  now.tv_sec,
                  now.tv_usec));
      SM_ACCUMULATE(smg_until_sent, transaction->start_time_io, now);
#endif

#ifdef OSP_MEASURE_TRANSACTION
      /***********************************************************************/
      /* Set the transaction finish send time.                               */
      /***********************************************************************/
      SM_GETTIMES(transaction->timings[SM_TIMINGS_FINISH_SEND]);
#endif

      /***********************************************************************/
      /* Sent the whole transaction - remove from list.                      */
      /***********************************************************************/
      LIST_REMOVE_ITEM(&transaction->links);
      if (transaction == conn->current_trans)
      {
        conn->current_trans = NULL;
      }
      sm_free_transaction(transaction, TRUE);
    }

    if (LIST_IS_EMPTY(&conn->wa_locked_send_queue))
    {
      /***********************************************************************/
      /* Sent all of the transactions that were queued.                      */
      /***********************************************************************/
      if (conn->conn_state == SM_CONN_STATE_DONT_RECEIVE)
      {
        TRC_DETAIL((TB,
              "List is empty and dont receive is set - now close connection"));
        OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
        sm_worker_close_connection(conn);
        goto EXIT_LABEL;
      }
      else
      {
        /*********************************************************************/
        /* We can now read on this socket                                    */
        /*********************************************************************/
        TRC_DETAIL((TB, "Cleared transaction list"));
        FD_SET(conn->sock_handle, &conn->thread->read_set);
        FD_CLR(conn->sock_handle, &conn->thread->wa_locked.write_set);
      }
    }

    OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
  }

  if (conn->ssl_data_buffered || FD_ISSET(conn->sock_handle, read_set))
  {
    /*************************************************************************/
    /* Check we have a transaction buffer to read into.                      */
    /*************************************************************************/
    TRC_DETAIL((TB, "Read set"));

    if (conn->current_trans == NULL)
    {
      TRC_DETAIL((TB, "Need to get a new transaction buffer"));

      conn->current_trans = sm_get_transaction(conn);

      if (conn->current_trans == NULL)
      {
        /*********************************************************************/
        /* Failed to get memory to accept a new transaction.  Ideally RM     */
        /* should be prioritizing so that this is unlikely to happen.  We    */
        /* can't ignore the read event because that could lead to us looping */
        /* on the select call.  Best we can do in the situation is to mark   */
        /* this connection to terminate on completion of this transaction.   */
        /*********************************************************************/
        TRC_ERROR((TB, "Failed to get mem for current transaction (conn %p)",
                                                                        conn));
        OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);

        /*********************************************************************/
        /* Make sure no more reads happen on this connection                 */
        /*********************************************************************/
        FD_CLR(conn->sock_handle, &conn->thread->read_set);
        conn->conn_state = SM_CONN_STATE_DONT_RECEIVE;

        /*********************************************************************/
        /* If there are no outstanding transactions then we can drop the     */
        /* connection immediately; otherwise we must wait for them to        */
        /* complete                                                          */
        /*********************************************************************/
        if (conn->wa_locked_outstanding_count == 0)
        {
          /*******************************************************************/
          /* Note that we must release the mutex before we terminate the     */
          /* connection!                                                     */
          /*******************************************************************/
          TRC_DETAIL((TB, "Terminate connection now"));
          OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
          sm_worker_close_connection(conn);
        }
        else
        {
          TRC_DETAIL((TB, "Terminate connection when %d transactions are done",
                                           conn->wa_locked_outstanding_count));
          OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
        }
        goto EXIT_LABEL;
      }
    }

#ifdef OSP_MEASURE_PERFORMANCE
    /*************************************************************************/
    /* If this is the first data that we're receiving for this transaction,  */
    /* set the transaction IO start time.                                    */
    /*************************************************************************/
    if (conn->current_trans->recv_buffer_end == 0)
    {
      gettimeofday(&(conn->current_trans->start_time_io), NULL);
      TRC_DETAIL((TB, "Set transaction IO start time:"
                  " trans=%p time=%ld:%06ld",
                  conn->current_trans,
                  conn->current_trans->start_time_io.tv_sec,
                  conn->current_trans->start_time_io.tv_usec));
    }
#endif

#ifdef OSP_MEASURE_TRANSACTION
    /*************************************************************************/
    /* If this is the first data that we're receiving for this transaction,  */
    /* set the transaction start read time.                                  */
    /*************************************************************************/
    if (conn->current_trans->recv_buffer_end == 0)
    {
      SM_GETTIMES(conn->current_trans->timings[SM_TIMINGS_START_READ]);
    }
#endif

    if (conn->type == SM_NON_SECURE_SOCKET)
    {
      /***********************************************************************/
      /* Non-SSL Socket Connection                                           */
      /***********************************************************************/
      TRC_DETAIL((TB, "Read (non-secure) into buffer: end=%d size=%d",
                  conn->current_trans->recv_buffer_end,
                  conn->current_trans->recv_buffer_size));

      num_bytes = recv(conn->sock_handle,
                       conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end,
                       conn->current_trans->recv_buffer_size -
                                          conn->current_trans->recv_buffer_end,
                       0);

      if ((num_bytes == -1) && (errno == EWOULDBLOCK))
      {
        TRC_NORMAL((TB, "No more socket data to read now"));
        goto EXIT_LABEL;
      }
      else if (num_bytes == -1)
      {
        TRC_OS_ERROR((TB, "Socket error"));
        sm_worker_close_connection(conn);
        goto EXIT_LABEL;
      }
      else if (num_bytes == 0)
      {
        TRC_NORMAL((TB, "Socket closed"));
        sm_worker_close_connection(conn);
        goto EXIT_LABEL;
      }

      TRC_DETAIL((TB, "Read successful: num_bytes=%d", num_bytes));
    }
    else
    {
      /***********************************************************************/
      /* SSL                                                                 */
      /***********************************************************************/
      TRC_DETAIL((TB, "Read (SSL) into buffer: end=%d size=%d",
                  conn->current_trans->recv_buffer_end,
                  conn->current_trans->recv_buffer_size));

      num_bytes = SSL_read(conn->ssl,
                           conn->current_trans->recv_buffer +
                                          conn->current_trans->recv_buffer_end,
                           conn->current_trans->recv_buffer_size -
                                         conn->current_trans->recv_buffer_end);
      TRC_DETAIL((TB, "SSL_read %d bytes, checking for errors", num_bytes));
      error = SSL_get_error(conn->ssl, num_bytes);

      switch (error)
      {
        case SSL_ERROR_NONE:
          /*******************************************************************/
          /* Successful read                                                 */
          /*******************************************************************/
          TRC_DETAIL((TB, "No error"));

          /*******************************************************************/
          /* More data may now be buffered within SSL.                       */
          /*******************************************************************/
          if (!conn->ssl_data_buffered)
          {
            conn->ssl_data_buffered = TRUE;
            conn->thread->ssl_data_buffered_count++;
            TRC_DETAIL((TB, "Now %d SSL connections with data buffered",
                        conn->thread->ssl_data_buffered_count));
          }
          break;

        case SSL_ERROR_SSL:
        case SSL_ERROR_SYSCALL:
          /*******************************************************************/
          /* SSL error                                                       */
          /*******************************************************************/
          OSP_OPENSSL_ERROR((TB, "SSL error on SSL_read"));
          sm_worker_close_connection(conn);
          goto EXIT_LABEL;
          break;

        case SSL_ERROR_ZERO_RETURN:
          /*******************************************************************/
          /* Connection closed                                               */
          /*******************************************************************/
          TRC_DETAIL((TB, "Connection closed"));
          sm_worker_close_connection(conn);
          goto EXIT_LABEL;
          break;

        case SSL_ERROR_WANT_READ:
          /*******************************************************************/
          /* No data from that call - now need to read.  No action required  */
          /* as we are already listening on this socket.                     */
          /*******************************************************************/
          TRC_DETAIL((TB, "SSL_ERROR_WANT_READ"));
          TRC_NORMAL((TB, "No more socket data to read now"));

          /*******************************************************************/
          /* We now know that there is definitely no read data buffered for  */
          /* this connection.                                                */
          /*******************************************************************/
          if (conn->ssl_data_buffered)
          {
            conn->ssl_data_buffered = FALSE;
            conn->thread->ssl_data_buffered_count--;
            TRC_DETAIL((TB, "Now %d SSL connections with data buffered",
                        conn->thread->ssl_data_buffered_count));
          }
          goto EXIT_LABEL;
          break;

        case SSL_ERROR_WANT_WRITE:
          /*******************************************************************/
          /* No data from that call - now need to write.                     */
          /*******************************************************************/
          TRC_DETAIL((TB, "SSL_ERROR_WANT_WRITE"));
          OSP_GET_LOCK(&conn->thread->wa_locked.mutex_wa);
          FD_SET(conn->sock_handle, &conn->thread->wa_locked.write_set);
          OSP_RELEASE_LOCK(&conn->thread->wa_locked.mutex_wa);
          goto EXIT_LABEL;
          break;

        default:
          OSP_OPENSSL_ERROR((TB, "Unknown error code %d", error));
          sm_worker_close_connection(conn);
          goto EXIT_LABEL;
          break;
      }
    }

#ifdef OSP_MEASURE_TRANSACTION
    /*************************************************************************/
    /* Set the transaction finish read time.                                 */
    /*************************************************************************/
    SM_GETTIMES(conn->current_trans->timings[SM_TIMINGS_FINISH_READ]);
#endif

    conn->current_trans->recv_buffer_end += num_bytes;
    TRC_DETAIL((TB, "Read %d bytes into buffer at %p (offset 0x%x); "
                                                      "now got %d bytes total",
                num_bytes,
                conn->current_trans->recv_buffer,
                conn->current_trans->recv_buffer_start,
                conn->current_trans->recv_buffer_end -
                                      conn->current_trans->recv_buffer_start));

    /*************************************************************************/
    /* Call the client verification callback if we haven't already.          */
    /*************************************************************************/
    if ((!conn->called_cv_cb) && SSL_is_init_finished(conn->ssl))
    {
        TRC_DETAIL((TB, "Call client verification callback"));

        if (!sm_call_verify_callback(conn, 0, NULL, 0, 0))
        {
          /*******************************************************************/
          /* Client does not accept connection.                              */
          /*******************************************************************/
          TRC_ALERT((TB, "Connection not accepted"));
          sm_worker_close_connection(conn);
          goto EXIT_LABEL;
        }
    }

    /*************************************************************************/
    /* Call sm_worker_received_data() to process the receive buffer.  Note   */
    /* that this may process more than one HTTP request, but it is limited   */
    /* (and therefore "fair" to other connections) by the amount of data     */
    /* that can be fitted in the transaction read buffer.                    */
    /*************************************************************************/
    do
    {
      receive_result = sm_worker_received_data(conn);
      TRC_DETAIL((TB, "Receive result: %d", receive_result));

      /***********************************************************************/
      /* Encourage a smaller thread scheduling granularity than the OS       */
      /* normally gives us, to avoid large latencies for transactions being  */
      /* processed on other threads.                                         */
      /***********************************************************************/
      yield();
    }
    while (receive_result == SM_CALL_AGAIN);
  }

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* sm_worker_check_fd */

