
/**MOD+***********************************************************************/
/* Module:    usehdlrs.c                                                     */
/*                                                                           */
/* Purpose:   USE (Usage Metering) AIB - internal functions                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.11               $ $Modtime::   Jan 11 2001 15:56:48   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <useincl.h>

/**PROC+**********************************************************************/
/* Name:      use_end_element_handler                                        */
/*                                                                           */
/* Purpose:   Handle the end tag of a UsageIndication XML subelement         */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     name         - A null-terminated string containing the  */
/*                                  XML element name.                        */
/*                                                                           */
/**PROC-**********************************************************************/

void use_end_element_handler(void *user_data,
                             const XML_Char *name)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  TRANS_INFO *trans = (TRANS_INFO *)user_data;
  USE_DATA *data;
  OSP_RC rc;
  OSP_STATUS status;

  TRC_ENTRY("use_end_element_handler");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_ASSERT((name != NULL), (TB, "NULL element name parameter"));

  /***************************************************************************/
  /* Trace input parameters.                                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Input parameters: transaction=%p element name=\"%s\"",
                                                                 trans, name));

  /***************************************************************************/
  /* Do common AIB end tag handling.                                         */
  /***************************************************************************/
  aib_end_element_handler(trans, name);

  /***************************************************************************/
  /* Detect the UsageIndication end tag: this signals parsing completion.    */
  /***************************************************************************/
  if (strcmp(name, com_usage_ind) == 0)
  {
    TRC_NORMAL((TB, "Detected UsageIndication end tag"));

    /*************************************************************************/
    /* Reset the XML layer's own handlers.                                   */
    /*************************************************************************/
    xml_set_default_handlers(trans);

    /*************************************************************************/
    /* If the USE parse data is NULL, then we already called                 */
    /* osp_um_response, and so constructed an XML response, in               */
    /* use_start_processing.  All of the possible code paths for the (data   */
    /* != NULL) branch below must eventually call osp_um_response.           */
    /*************************************************************************/
    data = trans->aib_state;
    if (data != NULL)
    {
      TRC_DETAIL((TB, "Non-NULL data: %p", data));

      if (data->common.status.code == OSP_STATUS_SUCCESS)
      {
        TRC_DETAIL((TB, "Check for usage metering callback"));

        /*********************************************************************/
        /* Complete UsageIndication component parsed successfully: check     */
        /* whether the application callback is registered and, if so, call   */
        /* it.                                                               */
        /*********************************************************************/
        osp_get_recursive_lock(&use_cb_mutex);

        if (use_cb != NULL)
        {
#ifdef OSP_DATA_COVERAGE
          aib_scan_registered_structures();
#endif

          TRC_NORMAL((TB, "Call registered usage metering callback: %p",
                                                                      use_cb));
          rc = (*use_cb)(data->common.corr,
                         data->timestamp,
                         data->role,
                         data->transaction_id,
                         data->call_id,
                         data->source_info,
                         SL_HEAD(data->source_alt_list),
                         data->dest_info,
                         SL_HEAD(data->dest_alt_list),
                         SL_HEAD(data->use_detail_list),
                         &(trans->client_id));

          if (rc != OSP_SUCCESS)
          {
            TRC_ALERT((TB, "Callback invocation failed: rc=%d", rc));

            /*****************************************************************/
            /* The callback has not called osp_um_response, so we will do so */
            /* ourselves with the following status.                          */
            /*****************************************************************/
            data->common.status.code = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
            data->common.status.desc = com_status_transient_problem_in_server;
            rc = osp_um_response(data->common.corr,
                                 &(data->common.status),
                                 0,
                                 0);
            TRC_ASSERT((rc == OSP_SUCCESS),
                       (TB, "Error from osp_um_response: rc=%d", rc));
          }
        }
        else
        {
          TRC_ALERT((TB, "No usage metering callback registered"));
          data->common.status.code = OSP_STATUS_SERVICE_NOT_AVAILABLE;
          data->common.status.desc = com_status_service_not_available;
          rc = osp_um_response(data->common.corr,
                               &(data->common.status),
                               0,
                               0);
          TRC_ASSERT((rc == OSP_SUCCESS),
                     (TB, "Error from osp_um_response: rc=%d", rc));
        }

        osp_release_recursive_lock(&use_cb_mutex);
      }
      else
      {
        TRC_ALERT((TB, "Unsuccessful parsing status: %d",
                                                    data->common.status.code));

        /*********************************************************************/
        /* Call osp_um_response appropriately.                               */
        /*********************************************************************/
        rc = osp_um_response(data->common.corr, &(data->common.status), 0, 0);
        TRC_ASSERT((rc == OSP_SUCCESS),
                   (TB, "Error from osp_um_response: rc=%d", rc));
      }
    }
    else
    {
      TRC_DETAIL((TB, "NULL data - response already generated"));

      /***********************************************************************/
      /* The response was generated immediately in use_start_processing.     */
      /* Nothing further to do here.                                         */
      /***********************************************************************/
    }

    /*************************************************************************/
    /* Done.  We will never be called again for the same UsageIndication     */
    /* component, so we should free our parsing data now.                    */
    /*************************************************************************/
    use_free_use_data(trans->aib_state);
    trans->aib_state = NULL;
  }

EXIT_LABEL:
  TRC_EXIT();
  return;

} /* use_end_element_handler */

/**PROC+**********************************************************************/
/* Name:      use_free_use_data                                              */
/*                                                                           */
/* Purpose:   Free all memory associated with a USE_DATA structure           */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     data         - Pointer to USE_DATA structure to free.   */
/*                                                                           */
/**PROC-**********************************************************************/

void use_free_use_data(USE_DATA *data)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_ADDRESS *address;
  OSP_USAGE_DETAIL *usage_detail;

  TRC_ENTRY("use_free_use_data");

  /***************************************************************************/
  /* Check the supplied pointer                                              */
  /***************************************************************************/
  if (data == NULL)
  {
    TRC_DETAIL((TB, "Supplied NULL data"));
    goto EXIT_LABEL;
  }

  TRC_DETAIL((TB, "Free data: %p", data));

  TRC_DETAIL((TB, "Free call ID: %p", data->call_id));
  aib_free_call_id(data->call_id);

  TRC_DETAIL((TB, "Free source info: %p", data->source_info));
  aib_free_address(data->source_info);

  address = SL_HEAD(data->source_alt_list);
  while (address != NULL)
  {
    TRC_DETAIL((TB, "Free alternate source info: %p", address));
    SL_REMOVE_HEAD(data->source_alt_list);
    aib_free_address(address);
    address = SL_HEAD(data->source_alt_list);
  }

  TRC_DETAIL((TB, "Free destination info: %p", data->dest_info));
  aib_free_address(data->dest_info);

  address = SL_HEAD(data->dest_alt_list);
  while (address != NULL)
  {
    TRC_DETAIL((TB, "Free alternate destination info: %p", address));
    SL_REMOVE_HEAD(data->dest_alt_list);
    aib_free_address(address);
    address = SL_HEAD(data->dest_alt_list);
  }

  usage_detail = SL_HEAD(data->use_detail_list);
  while (usage_detail != NULL)
  {
    TRC_DETAIL((TB, "Free usage detail: %p", usage_detail));
    SL_REMOVE_HEAD(data->use_detail_list);
    aib_free_usage_detail(usage_detail);
    usage_detail = SL_HEAD(data->use_detail_list);
  }

  /***************************************************************************/
  /* USE_DATA fields that we don't free, and why not...                      */
  /*                                                                         */
  /*   - char_data.start_ptr      - This is simply a pointer to another      */
  /*                                memory location, not a pointer to        */
  /*                                independently allocated memory.          */
  /*   - name, attrs              - These are convenience pointers to memory */
  /*                                owned by the XML layer.                  */
  /*   - status.desc              - This char pointer is only ever set to    */
  /*                                constant strings.                        */
  /*   - corr                     - This is specifically designed to have a  */
  /*                                lifetime independent of the USE_DATA     */
  /*                                structure (otherwise it would be an      */
  /*                                inline field rather than a pointer       */
  /*                                field), because the correlator needs to  */
  /*                                remain valid until the application calls */
  /*                                osp_um_response.  In principle, this can */
  /*                                be an arbitrarily long time after the    */
  /*                                USE_DATA structure is destroyed.  The    */
  /*                                separately allocated correlator memory   */
  /*                                is eventually freed by osp_um_response.  */
  /*                                                                         */
  /***************************************************************************/

  TRC_DETAIL((TB, "Free data: %p", data));
  rm_release_mem(data);

EXIT_LABEL:

  TRC_EXIT();
  return;

} /* use_free_use_data */

