
/**MOD+***********************************************************************/
/* Module:    usemain.c                                                      */
/*                                                                           */
/* Purpose:   USE (Usage Metering) AIB - external API functions              */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.17               $ $Modtime::   Jan 11 2001 15:56:48   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <useincl.h>

/**PROC+**********************************************************************/
/* Name:      use_init                                                       */
/*                                                                           */
/* Purpose:   Initialize USE (Usage Metering) component of OpenOSP stack     */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - initialized successfully                 */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC use_init()
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("use_init");

  /***************************************************************************/
  /* Check for previous component initialization, to avoid initializing      */
  /* twice.                                                                  */
  /***************************************************************************/
  TRC_ASSERT((!use_initialized),
             (TB, "USE component already initialized: %d", use_initialized));

  TRC_NORMAL((TB, "Initialize USE component of OpenOSP stack"));

  /***************************************************************************/
  /* Initialize global variables.                                            */
  /***************************************************************************/
  use_initialized = TRUE;
  use_cb = NULL;

  /***************************************************************************/
  /* Create the callback mutex.  This is recursive to allow a callback       */
  /* function to deregister itself.                                          */
  /***************************************************************************/
  rc = osp_create_recursive_mutex(&use_cb_mutex);
  TRC_DETAIL((TB, "Mutex creation: rc=%d", rc));

  TRC_EXIT();
  return(rc);

} /* use_init */

/**PROC+**********************************************************************/
/* Name:      use_term                                                       */
/*                                                                           */
/* Purpose:   Terminate USE (Usage Metering) component of OpenOSP stack      */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - terminated successfully                  */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC use_term()
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("use_term");

  /***************************************************************************/
  /* Check that we have previously been initialized.                         */
  /***************************************************************************/
  if (!use_init)
  {
    TRC_ERROR((TB, "Called when not init"));
    rc = OSP_ERR_NOT_INIT;

  }
  else
  {
    TRC_NORMAL((TB, "Terminate USE component of OpenOSP stack"));

    /*************************************************************************/
    /* Terminate remaining global variables.                                 */
    /*************************************************************************/
    use_initialized = FALSE;
    use_cb = NULL;

    /*************************************************************************/
    /* Destroy the callback mutex.                                           */
    /*************************************************************************/
    rc = osp_destroy_recursive_mutex(&use_cb_mutex);
    TRC_DETAIL((TB, "Mutex destroyed: rc=%d", rc));
  }

  TRC_EXIT();
  return(rc);

} /* use_term */

/**PROC+**********************************************************************/
/* Name:      osp_um_register                                                */
/*                                                                           */
/* Purpose:   Register application callback for usage metering indications   */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - callback registered successfully         */
/*            OSP_ERR_NOT_INIT    - the USE AIB is not yet initialized       */
/*                                                                           */
/* Params:    IN     um_cb        - Pointer to the application's callback    */
/*                                  routine.  This parameter may be NULL,    */
/*                                  which indicates that the currently       */
/*                                  registered callback (if any) should be   */
/*                                  deregistered.                            */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_um_register(POSP_UM_CALLBACK um_cb)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  TRC_API_ENTRY("osp_um_register");

  /***************************************************************************/
  /* Check that the USE AIB has been initialized.                            */
  /***************************************************************************/
  if (use_initialized)
  {
    /*************************************************************************/
    /* Acquire USE callback mutex.                                           */
    /*************************************************************************/
    osp_get_recursive_lock(&use_cb_mutex);

    /*************************************************************************/
    /* Register new callback function pointer.                               */
    /*************************************************************************/
    TRC_DETAIL((TB, "Old USE callback: %p", use_cb));
    use_cb = um_cb;
    TRC_DETAIL((TB, "New USE callback: %p", use_cb));

    /*************************************************************************/
    /* Release USE callback mutex.                                           */
    /*************************************************************************/
    osp_release_recursive_lock(&use_cb_mutex);
  }
  else
  {
    TRC_ERROR((TB, "USE AIB has not been initialized"));
    rc = OSP_ERR_NOT_INIT;
  }

  TRC_API_EXIT();
  return(rc);

} /* osp_um_register */

/**PROC+**********************************************************************/
/* Name:      osp_um_response                                                */
/*                                                                           */
/* Purpose:   See openosp.h                                                  */
/*                                                                           */
/* Returns:   See openosp.h                                                  */
/*                                                                           */
/* Params:    See openosp.h                                                  */
/*                                                                           */
/* Operation: The correlator value is a pointer to an OSP_INT_CORRELATOR     */
/*            struct containing pointers to the current transaction and the  */
/*            component ID attribute of the UsageIndication element that is  */
/*            being responded to.                                            */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC OSPAPI osp_um_response(OSP_CORRELATOR corr,
                              OSP_STATUS *status,
                              unsigned int audit_signal,
                              unsigned int sig_required)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  OSP_INT_CORRELATOR *int_corr;

  TRC_API_ENTRY("osp_um_response");

  /***************************************************************************/
  /* Check parameters; the correlator actually contains a transaction        */
  /* pointer and a component ID.                                             */
  /***************************************************************************/
  if (corr == NULL)
  {
    TRC_ERROR((TB, "NULL correlator parameter"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }
  if (status == NULL)
  {
    TRC_ERROR((TB, "NULL status parameter"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }
  int_corr = (OSP_INT_CORRELATOR *)corr;
  if (int_corr->trans == NULL)
  {
    TRC_ERROR((TB, "NULL correlator transaction pointer"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }
  if (int_corr->comp_id == NULL)
  {
    TRC_ERROR((TB,"NULL correlator component ID"));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }
  if (audit_signal &&
      (audit_signal != OSP_UM_AUDIT_START) &&
      (audit_signal != OSP_UM_AUDIT_STOP))
  {
    TRC_ERROR((TB,"Bad audit signal value: %u", audit_signal));
    rc = OSP_ERR_BAD_PARAM;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Trace input parameters.                                                 */
  /***************************************************************************/
  TRC_NORMAL((TB,
              "Response: trans=%p comp_id=\"%s\" allocated=%d"
              " code=%d desc=\"%s\" audit_signal=%u sig_required=%u",
              int_corr->trans, int_corr->comp_id, int_corr->allocated,
              status->code, status->desc, audit_signal, sig_required));

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction start build time.                                   */
  /***************************************************************************/
  SM_GETTIMES(int_corr->trans->timings[SM_TIMINGS_START_BUILD]);
#endif

  /***************************************************************************/
  /* We're about to start using the transaction send buffer, so grab the     */
  /* mutex.                                                                  */
  /***************************************************************************/
  OSP_GET_LOCK(&(int_corr->trans->mutex));

  /***************************************************************************/
  /* If signing is required, set the flag on the transaction - the whole     */
  /* message is signed, not just this component.                             */
  /***************************************************************************/
  if (sig_required)
  {
    TRC_NORMAL((TB, "Signing required on transaction: %p", int_corr->trans));
    int_corr->trans->sign_response = TRUE;
  }

  /***************************************************************************/
  /* Write the UsageConfirmation start tag.                                  */
  /***************************************************************************/
  if (!aib_space_available(int_corr->trans,
                           1 + 1 + 2 + 3 +
                           strlen(com_usage_cnf) + strlen(com_comp_id) +
                           strlen(int_corr->comp_id)))
  {
    TRC_ERROR((TB, "No room for XML component"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(int_corr->trans, com_lt, 1);
  AIB_ADD_NULL(int_corr->trans, com_usage_cnf);
  AIB_ADD_DATA(int_corr->trans, com_sp, 1);
  AIB_ADD_NULL(int_corr->trans, com_comp_id);
  AIB_ADD_DATA(int_corr->trans, com_eq_qu, 2);
  AIB_ADD_NULL(int_corr->trans, int_corr->comp_id);
  AIB_ADD_DATA(int_corr->trans, com_qu_gt_nl, 3);

  /***************************************************************************/
  /* Write timestamp.                                                        */
  /***************************************************************************/
  if (!aib_write_xml_timestamp(int_corr->trans))
  {
    TRC_ERROR((TB, "No room for timestamp"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Write status.                                                           */
  /***************************************************************************/
  if (!aib_write_xml_status(int_corr->trans, status))
  {
    TRC_ERROR((TB, "No room for status"));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Write AuditSignal element if required.                                  */
  /***************************************************************************/
  if (audit_signal == OSP_UM_AUDIT_START)
  {
    TRC_DETAIL((TB, "Write audit signal start element"));
    if (!aib_space_available(int_corr->trans,
                             strlen(use_audit_signal_start)))
    {
      TRC_ERROR((TB, "No room for audit signal start"));
      goto EXIT_LABEL;
    }
    AIB_ADD_NULL(int_corr->trans, use_audit_signal_start);
  }
  else if (audit_signal == OSP_UM_AUDIT_STOP)
  {
    TRC_DETAIL((TB, "Write audit signal stop element"));
    if (!aib_space_available(int_corr->trans,
                             strlen(use_audit_signal_stop)))
    {
      TRC_ERROR((TB, "No room for audit signal stop"));
      goto EXIT_LABEL;
    }
    AIB_ADD_NULL(int_corr->trans, use_audit_signal_stop);
  }

  /***************************************************************************/
  /* Write UsageConfirmation end tag.                                        */
  /***************************************************************************/
  if (!aib_space_available(int_corr->trans, 2 + 2 + strlen(com_usage_cnf)))
  {
    TRC_ERROR((TB, "No room for component close tag"));
    goto EXIT_LABEL;
  }

  AIB_ADD_DATA(int_corr->trans, com_lt_sl, 2);
  AIB_ADD_NULL(int_corr->trans, com_usage_cnf);
  AIB_ADD_DATA(int_corr->trans, com_gt_nl, 2);

  TRC_DETAIL((TB, "Finished writing response"));

EXIT_LABEL:
  /***************************************************************************/
  /* We've finished - update the response count and see if the response      */
  /* needs to be sent.  Note that xml_maybe_send will free the transaction   */
  /* mutex for us.                                                           */
  /***************************************************************************/
  if (rc == OSP_SUCCESS)
  {
    int_corr->trans->response_count++;
    TRC_NORMAL((TB, "Incremented response count: trans=%p response_count=%d",
                            int_corr->trans, int_corr->trans->response_count));
    xml_maybe_send(int_corr->trans);

    /*************************************************************************/
    /* Free the correlator component id and the correlator itself.           */
    /*************************************************************************/
    if (int_corr->allocated)
    {
      rm_release_mem(int_corr->comp_id);
      rm_release_mem(int_corr);
    }
  }

  TRC_API_EXIT();
  return(rc);

} /* osp_um_response */

/**PROC+**********************************************************************/
/* Name:      use_start_processing                                           */
/*                                                                           */
/* Purpose:   Start processing a UsageIndication XML element                 */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - processing started successfully          */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  This routine */
/*                                  uses the transaction's _aib_state_ field */
/*                                  to store parsing state information.      */
/*            IN     parser       - Pointer to XML layer's parser object.    */
/*                                  This is simply passed on to other XML    */
/*                                  layer routines.                          */
/*            IN     comp_type    - The OSP component type of the XML        */
/*                                  element for which this function is being */
/*                                  called by the XML layer.                 */
/*            IN     comp_id      - The component ID attribute value of the  */
/*                                  XML element that led to this call.       */
/*                                                                           */
/* Operation: Allocate a USE_DATA structure to store parsing data; store a   */
/*            pointer to this structure in the transaction; register         */
/*            USE-specific element and character data handlers with the XML  */
/*            layer.                                                         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC use_start_processing(TRANS_INFO *trans,
                            XML_Parser *parser,
                            int         comp_type,
                            char       *comp_id)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  USE_DATA *data;
  OSP_STATUS status;
  OSP_INT_CORRELATOR corr;

  TRC_ENTRY("use_start_processing");

  /***************************************************************************/
  /* Verify the component type for which we are being called - it should     */
  /* only be a UsageIndication.                                              */
  /***************************************************************************/
  TRC_ASSERT((comp_type == AIB_USAGE_IND),
             (TB, "Incorrect component type: %d", comp_type));

  /***************************************************************************/
  /* Call common AIB start processing routine to do most of the work.  If    */
  /* this succeeds trans->aib_state will be non-NULL; if it fails, the       */
  /* status parameter will have been set accordingly.                        */
  /***************************************************************************/
  aib_start_processing(trans,
                       parser,
                       comp_id,
                       sizeof(USE_DATA),
                       use_usage_indication_elements,
                       use_end_element_handler,
                       &status);

  if (trans->aib_state != NULL)
  {
    TRC_DETAIL((TB, "Initialize USE parse data: %p", trans->aib_state));

    /*************************************************************************/
    /* Initialize parse data fields.                                         */
    /*************************************************************************/
    data = trans->aib_state;

    data->timestamp = NULL;
    data->role = NULL;
    data->transaction_id = NULL;
    data->call_id = NULL;
    data->source_info = NULL;
    SL_INIT_ROOT(data->source_alt_list);
    data->dest_info = NULL;
    SL_INIT_ROOT(data->dest_alt_list);
    SL_INIT_ROOT(data->use_detail_list);
  }
  else
  {
    TRC_NORMAL((TB, "Prepare UsageConfirmation response immediately"));

    corr.trans = trans;
    corr.comp_id = comp_id;
    corr.allocated = FALSE;
    rc = osp_um_response(&corr, &status, 0, 0);
    TRC_ASSERT((rc == OSP_SUCCESS),
               (TB, "Error from osp_um_response: rc=%d", rc));
  }

  TRC_EXIT();
  return(rc);

} /* use_start_processing */

/**PROC+**********************************************************************/
/* Name:      use_cancel_processing                                          */
/*                                                                           */
/* Purpose:   Cancel the processing of a UsageIndication XML component       */
/*                                                                           */
/* Returns:   OSP_SUCCESS         - processing cancelled successfully        */
/*                                                                           */
/* Params:    IN/OUT trans        - Pointer to the transaction that is       */
/*                                  currently being processed.  Data         */
/*                                  allocated and used by the USE component  */
/*                                  is stored in the _aib_state_ field of    */
/*                                  the transaction.                         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC use_cancel_processing(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local variables.                                                        */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  USE_DATA *use_data;
  OSP_INT_CORRELATOR *int_corr;

  TRC_ENTRY("use_cancel_processing");

  /***************************************************************************/
  /* Check parameters.                                                       */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction parameter"));
  TRC_DETAIL((TB, "Cancel transaction: trans=%p aib_state=%p",
                                                     trans, trans->aib_state));

  /***************************************************************************/
  /* Free USE AIB data.                                                      */
  /***************************************************************************/
  use_data = (USE_DATA *)trans->aib_state;
  if (use_data != NULL)
  {
    /*************************************************************************/
    /* Free the correlator - since the processing has been cancelled, it     */
    /* won't be needed again                                                 */
    /*************************************************************************/
    TRC_DETAIL((TB, "Freeing correlator"));
    int_corr = use_data->common.corr;
    TRC_ASSERT((int_corr != NULL), (TB, "Correlator must be present"));
    TRC_ASSERT((int_corr->allocated), (TB, "Correlator must be allocated"));
    rm_release_mem(int_corr->comp_id);
    rm_release_mem(int_corr);

    /*************************************************************************/
    /* Free the rest of the data                                             */
    /*************************************************************************/
    TRC_DETAIL((TB, "Free USE AIB data"));
    use_free_use_data(use_data);

  }
  trans->aib_state = NULL;

  /***************************************************************************/
  /* Whenever the XML layer calls XXX_cancel_processing, it also handles the */
  /* sending of an appropriate HTTP failure response.  Therefore there is    */
  /* nothing else that the USE AIB (or any AIB) needs to do in this case.    */
  /***************************************************************************/

  TRC_EXIT();
  return(rc);

} /* use_cancel_processing */
