
/**MOD+***********************************************************************/
/* Module:    xmlcanon.c                                                     */
/*                                                                           */
/* Purpose:   XML Canonicalization routines                                  */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.12               $ $Modtime::   Jan 11 2001 15:56:48   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <xmlincl.h>
#include <stdarg.h>

/**PROC+**********************************************************************/
/* Name:      xml_canonicalize                                               */
/*                                                                           */
/* Purpose:   Convert the supplied XML document to canonical form            */
/*                                                                           */
/* Returns:   OSP_SUCCESS     - success                                      */
/*            OSP_ERR_MEMORY  - failure, due to a resource shortage          */
/*            OSP_ERROR       - failure, due to an error in the supplied XML */
/*            OSP_ERR_BUFSIZE - failure, due to the output buffer            */
/*                              being too small                              */
/*                                                                           */
/* Params:    IN     inbuf        - pointer to a buffer containing the XML   */
/*                                  to be canonicalized.  Must be NULL       */
/*                                  terminated                               */
/*            IN     inbuf_len    - number of bytes in inbuf                 */
/*            IN     outbuf       - pointer to a buffer in which to place    */
/*                                  the canonicalized XML                    */
/*            IN/OUT outbuf_len   - IN: size of buffer pointed to by outbuf  */
/*                                  OUT: on success, number of bytes copied  */
/*                                       into outbuf; on OSP_ERR_BUFSIZE,    */
/*                                       this indicates the required size    */
/*                                  Note that, for internal implementation   */
/*                                  reasons, this size must always include   */
/*                                  one byte for a null terminator character */
/*                                  at the end of the buffer.  Callers that  */
/*                                  do not want to process the terminating   */
/*                                  character (e.g. for message digest       */
/*                                  calculation) should therefore subtract   */
/*                                  one from the returned outbuf_len.        */
/*                                  (See also xml_maybe_add_to_outbuf.)      */
/*                                                                           */
/* Operation: Create and initialize a parser instance                        */
/*            Parse the DTD                                                  */
/*            Parse the data, writing the output to outbuf                   */
/*            free resources                                                 */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC xml_canonicalize(char *inbuf,
                        unsigned int inbuf_len,
                        char *outbuf,
                        unsigned int *outbuf_len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  XML_Parser parser;
  enum XML_Error xml_error;
  XML_CANON_INFO canon_info;
  int ii;

  TRC_ENTRY("xml_canonicalize");

  /***************************************************************************/
  /* Validate the parameters we've been given                                */
  /***************************************************************************/
  TRC_ASSERT((inbuf != NULL), (TB, "Inbuf is NULL"));
  TRC_ASSERT((outbuf != NULL), (TB, "Outbuf is NULL"));
  TRC_ASSERT((outbuf_len != NULL), (TB, "Outbuf len is NULL"));
  TRC_ASSERT((*outbuf_len > 0), (TB, "Outbuf is zero length"));

  /***************************************************************************/
  /* Initialize variables                                                    */
  /***************************************************************************/
  xml_error = XML_ERROR_NONE;

  /***************************************************************************/
  /* Create the parser instance                                              */
  /***************************************************************************/
  parser = XML_ParserCreate(NULL);
  if (parser == NULL)
  {
    TRC_ERROR((TB, "Failed to create parser instance"));
    rc = OSP_ERR_MEMORY;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Set up the structure we use to store instance data while parsing        */
  /***************************************************************************/
  canon_info.outbuf = outbuf;
  canon_info.outbuf_len = *outbuf_len;
  canon_info.current_pos = outbuf;
  canon_info.current_len = 0;
  canon_info.overflowed = FALSE;

  /***************************************************************************/
  /* Set up the handlers                                                     */
  /***************************************************************************/
  XML_SetUserData(parser, &canon_info);
  XML_SetElementHandler(parser, xml_canon_start_elem_handler,
                                                   xml_canon_end_elem_handler);
  XML_SetCharacterDataHandler(parser, xml_canon_char_data_handler);
  XML_SetProcessingInstructionHandler(parser, xml_canon_pi_handler);

  /***************************************************************************/
  /* Zero the outbuf so we can just copy a single character into it without  */
  /* having to worry about NULL terminators                                  */
  /***************************************************************************/
  memset(outbuf, 0, *outbuf_len);

  /***************************************************************************/
  /* Now zero the returned buffer length in case of failure                  */
  /***************************************************************************/
  *outbuf_len = 0;

  /***************************************************************************/
  /* We need to effectively insert the DTD into the XML, after the version   */
  /* declaration.  First, find out where the declaration ends by searching   */
  /* for '?>'                                                                */
  /***************************************************************************/
  for (ii = 0; ii < (inbuf_len - 1); ii++)
  {
    if ((inbuf[ii] == '?') && (inbuf[ii + 1] == '>'))
    {
      TRC_DETAIL((TB, "Last character of XMLDecl is at offset %d", ii + 1));
      break;
    }
  }

  if (ii == (inbuf_len - 1))
  {
    TRC_ERROR((TB, "Failed to find end of XMLDecl in inbuf"));
    rc = OSP_ERROR;
  }

  /***************************************************************************/
  /* Increment ii to point to the first character after the XMLDecl          */
  /***************************************************************************/
  ii += 2;

  /***************************************************************************/
  /* Parse the XMLDecl                                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Parsing XMLDecl"));
  if (XML_Parse(parser, inbuf, ii, FALSE))
  {
    /*************************************************************************/
    /* Parse the DTD so the parser knows what the canonical form is supposed */
    /* to look like                                                          */
    /*************************************************************************/
    TRC_DETAIL((TB, "Parsing dtd"));
    if (XML_Parse(parser, xml_dtd, strlen(xml_dtd), FALSE))
    {
      /***********************************************************************/
      /* And now we can parse the remainder of the supplied data             */
      /***********************************************************************/
      TRC_DETAIL((TB, "Parsing data"));
      if (XML_Parse(parser, inbuf + ii, inbuf_len - ii, TRUE))
      {
        /*********************************************************************/
        /* We parsed the data OK, but did it all fit in to outbuf?           */
        /*********************************************************************/
        TRC_DETAIL((TB, "Parsing completed"));
        if (canon_info.overflowed)
        {
          /*******************************************************************/
          /* No: set the return code accordingly                             */
          /*******************************************************************/
          TRC_ALERT((TB, "Data len %d bigger than outbuf len %d",
                               canon_info.current_len, canon_info.outbuf_len));
          rc = OSP_ERR_BUFSIZE;
        }

        /*********************************************************************/
        /* Return the length whether or not we overflowed, including a byte  */
        /* for the terminating null character.                               */
        /*********************************************************************/
        *outbuf_len = canon_info.current_len + 1;
      }
      else
      {
        /*********************************************************************/
        /* Get the error code                                                */
        /*********************************************************************/
        xml_error = XML_GetErrorCode(parser);
        TRC_ERROR((TB, "Got error %d (%s) while parsing XML data", xml_error,
                                          (char *)XML_ErrorString(xml_error)));

      }
    }
    else
    {
      /***********************************************************************/
      /* Parsing of the DTD failed: given that the DTD is compiled in, we    */
      /* assert here unless this is a memory error                           */
      /***********************************************************************/
      xml_error = XML_GetErrorCode(parser);
      TRC_ERROR((TB, "Got error %d (%s) parsing DTD", xml_error,
                                          (char *)XML_ErrorString(xml_error)));
      TRC_ASSERT(( xml_error == XML_ERROR_NO_MEMORY),
                           (TB, "Non-memory error %d parsing DTD", xml_error));
    }
  }
  else
  {
    /*************************************************************************/
    /* Get the error code                                                    */
    /*************************************************************************/
    xml_error = XML_GetErrorCode(parser);
    TRC_ERROR((TB, "Got error %d (%s) whil parsing XMLDecl", xml_error,
                                          (char *)XML_ErrorString(xml_error)));

  }

  /***************************************************************************/
  /* Return the appropriate error code                                       */
  /***************************************************************************/
  if (xml_error != XML_ERROR_NONE)
  {
    if (xml_error == XML_ERROR_NO_MEMORY)
    {
      TRC_DETAIL((TB, "returning memory error"));
      rc = OSP_ERR_MEMORY;
    }
    else
    {
      TRC_DETAIL((TB, "non-memory error"));
      rc = OSP_ERROR;
    }
  }
  else
  {
    TRC_DETAIL((TB, "Parsing completed OK"));
    TRC_DATA("Canonicalized XML:", outbuf, *outbuf_len);
  }

EXIT_LABEL:

  /***************************************************************************/
  /* Free the parser if we allocated it                                      */
  /***************************************************************************/
  if (parser != NULL)
  {
    TRC_DETAIL((TB, "Freeing parser"));
    XML_ParserFree(parser);
  }

  TRC_EXIT();

  return(rc);

} /* xml_canonicalize */

/**PROC+**********************************************************************/
/* Name:      xml_canon_char_data_handler                                    */
/*                                                                           */
/* Purpose:   Expat character data handler                                   */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to XML_CANON_INFO sructure       */
/*            IN     data         - character data, not NULL terminated      */
/*            IN     len          - length of data                           */
/*                                                                           */
/* Operation: While there's space in outbuf,                                 */
/*              Convert any special characters to character references       */
/*              Write the character to the outbuf                            */
/*                                                                           */
/**PROC-**********************************************************************/

static void xml_canon_char_data_handler(void *user_data,
                                        const XML_Char *data,
                                        int len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  XML_CANON_INFO *canon_info;
  char new_string[10];
  int new_len;
  char *new_data;

  TRC_ENTRY("xml_canon_char_data_handler");

  /***************************************************************************/
  /* Check the supplied user data                                            */
  /***************************************************************************/
  canon_info = (XML_CANON_INFO *)user_data;
  TRC_ASSERT((canon_info != NULL), (TB, "Passed NULL info block"));

  /***************************************************************************/
  /* If there's room, copy the data to outbuf, converting any special        */
  /* characters as we go                                                     */
  /***************************************************************************/
  TRC_DATA("Looking at char data:", (char *)data, len);
  while (len > 0)
  {
    /*************************************************************************/
    /* Check for special characters in the data                              */
    /*************************************************************************/
    switch (*data)
    {
      case '&':
        new_data = "&amp;";
        break;

      case '<':
        new_data = "&lt;";
        break;

      case '>':
        new_data = "&gt;";
        break;

      case '"':
        new_data = "&quot;";
      break;

      case 9:
      case 13:
        sprintf(new_string, "&#x%x;", *data);
        new_data = new_string;
        break;

      default:
        new_string[0] = *data;
        new_string[1] = 0;
        new_data = new_string;
        break;
    }

    /*************************************************************************/
    /* Copy the character and move onto the next                             */
    /*************************************************************************/
    xml_maybe_add_to_outbuf(canon_info, new_data);
    data++;
    len--;
  }

  TRC_EXIT();

  return;

} /* xml_canon_char_data_handler */

/**PROC+**********************************************************************/
/* Name:      xml_canon_compare_attrs                                        */
/*                                                                           */
/* Purpose:   qsort callback used to compare two attribute strings           */
/*                                                                           */
/* Returns:   strcmp result                                                  */
/*                                                                           */
/* Params:    IN     attr1, attr2 - attributes to compare                    */
/*                                                                           */
/**PROC-**********************************************************************/

static int xml_canon_compare_attrs(const void *attr1, const void *attr2)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int rc;

  TRC_ENTRY("xml_canon_compare_attrs");

  /***************************************************************************/
  /* Simply compare the two attributes                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Comparing %s with %s", attr1, attr2));
  rc = strcmp(*(const XML_Char **)attr1, *(const XML_Char **)attr2);

  TRC_EXIT();

  return(rc);

} /* xml_canon_compare_attrs */

/**PROC+**********************************************************************/
/* Name:      xml_canon_start_elem_handler                                   */
/*                                                                           */
/* Purpose:   Handles start elements and attributes during canonicalization  */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to XML_CANON_INFO sructure       */
/*            IN     name,                                                   */
/*                   attrs        - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

static void xml_canon_start_elem_handler(void *user_data,
                                         const XML_Char *name,
                                         const XML_Char **atts)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  XML_CANON_INFO *canon_info;
  int attr_count;
  const XML_Char **curr_attr;

  TRC_ENTRY("xml_canon_start_elem_handler");

  /***************************************************************************/
  /* Check the supplied user data                                            */
  /***************************************************************************/
  canon_info = (XML_CANON_INFO *)user_data;
  TRC_ASSERT((canon_info != NULL), (TB, "Passed NULL info block"));

  /***************************************************************************/
  /* Begin the start tag                                                     */
  /***************************************************************************/
  TRC_DETAIL((TB, "Begin tag %s", name));
  xml_maybe_add_to_outbuf(canon_info, "<%s", name);

  /***************************************************************************/
  /* Determine the number of attributes                                      */
  /***************************************************************************/
  curr_attr = atts;
  while (*curr_attr)
  {
    ++curr_attr;
  }
  attr_count = (curr_attr - atts) >> 1;

  /***************************************************************************/
  /* If there are multiple attributes, sort them alphabetically              */
  /***************************************************************************/
  if (attr_count > 1)
  {
    qsort((void *)atts, attr_count, sizeof(XML_Char *) * 2,
                                                    xml_canon_compare_attrs);
  }

  /***************************************************************************/
  /* Now write the sorted attribute list                                     */
  /***************************************************************************/
  while (*atts)
  {
    /*************************************************************************/
    /* First we put ' attr_name="'                                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "Add attr %s", *atts));
    xml_maybe_add_to_outbuf(canon_info, " %s=\"", *atts);

    /*************************************************************************/
    /* Move on to the value, getting the char data handler to write it to    */
    /* the outbuf since that code already knows how to handle any special    */
    /* characters.                                                           */
    /*************************************************************************/
    atts++;
    xml_canon_char_data_handler(user_data, *atts, strlen(*atts));

    /*************************************************************************/
    /* Now add the closing " and move on to the next value                   */
    /*************************************************************************/
    xml_maybe_add_to_outbuf(canon_info, "\"");
    atts++;
  }

  /***************************************************************************/
  /* Lastly close the tag                                                    */
  /***************************************************************************/
  TRC_DETAIL((TB, "Closing begin tag"));
  xml_maybe_add_to_outbuf(canon_info, ">");

  TRC_EXIT();

  return;

} /* xml_canon_start_elem_handler */

/**PROC+**********************************************************************/
/* Name:      xml_canon_end_elem_handler                                     */
/*                                                                           */
/* Purpose:   Handles end elements during canonicalization                   */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to XML_CANON_INFO sructure       */
/*            IN     name,        - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

static void xml_canon_end_elem_handler(void *user_data, const XML_Char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  XML_CANON_INFO *canon_info;

  TRC_ENTRY("xml_canon_end_elem_handler");

  /***************************************************************************/
  /* Check the supplied user data                                            */
  /***************************************************************************/
  canon_info = (XML_CANON_INFO *)user_data;
  TRC_ASSERT((canon_info != NULL), (TB, "Passed NULL info block"));

  /***************************************************************************/
  /* Add the close tag                                                       */
  /***************************************************************************/
  TRC_DETAIL((TB, "Closing element %s", name));
  xml_maybe_add_to_outbuf(canon_info, "</%s>", name);

  TRC_EXIT();

  return;

} /* xml_canon_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      xml_canon_pi_handler                                           */
/*                                                                           */
/* Purpose:   Handles processing instructions during canonicalization        */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to XML_CANON_INFO sructure       */
/*            IN     target, data - see expat documentation                  */
/*                                                                           */
/**PROC-**********************************************************************/

static void xml_canon_pi_handler(void *user_data,
                                 const XML_Char *target,
                                 const XML_Char *data)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  XML_CANON_INFO *canon_info;

  TRC_ENTRY("xml_canon_pi_handler");

  /***************************************************************************/
  /* Check the supplied user data                                            */
  /***************************************************************************/
  canon_info = (XML_CANON_INFO *)user_data;
  TRC_ASSERT((canon_info != NULL), (TB, "Passed NULL info block"));

  /***************************************************************************/
  /* Write out the PI                                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Adding PI %s", target));
  xml_maybe_add_to_outbuf(canon_info, "<?%s %s?>", target, data);

  TRC_EXIT();

  return;

} /* xml_canon_pi_handler */

/**PROC+**********************************************************************/
/* Name:      xml_maybe_add_to_outbuf                                        */
/*                                                                           */
/* Purpose:   Updates length of outbuf required for string, and adds string  */
/*            if there's space                                               */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     canon_info   - pointer to outbuf control structure      */
/*            IN     string       - string to add                            */
/*            IN     string_len   - length of string                         */
/*                                                                           */
/* Operation: Update the required outbuf len                                 */
/*            if there's space,                                              */
/*              copy the string to the outbuf                                */
/*            else                                                           */
/*              update the overflowed flag if required                       */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_maybe_add_to_outbuf(XML_CANON_INFO *canon_info, char *string, ...)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  va_list ap;
  int avail_len;
  int req_len;
  char buf[1];

  TRC_ENTRY("xml_maybe_add_to_outbuf");

  /***************************************************************************/
  /* Initialize the vriable argument list so we can pass it on for           */
  /* sprintf-ing                                                             */
  /***************************************************************************/
  va_start(ap, string);

  /***************************************************************************/
  /* Determine the available space in the outbuf                             */
  /***************************************************************************/
  avail_len = canon_info->outbuf_len - canon_info->current_len;
  if (avail_len > 0)
  {
    /*************************************************************************/
    /* Some buffer space is available, so try to write the string to the     */
    /* outbuf.  OSP_VSNPRINTF returns the space required (excluding the      */
    /* terminating NUL) whether or not there actually is enough room.        */
    /*************************************************************************/
    req_len = OSP_VSNPRINTF((canon_info->current_pos, avail_len, string, ap));
  }
  else
  {
    /*************************************************************************/
    /* No buffer space is available, but we still want to find out how much  */
    /* space is required.  OSP_VSNPRINTF returns -1 if we specify a buffer   */
    /* size of 0, so instead use a one-character buffer from the stack.      */
    /*************************************************************************/
    req_len = OSP_VSNPRINTF((buf, 1, string, ap));
  }
  TRC_DETAIL((TB, "%d bytes available, %d required", avail_len, req_len));

  /***************************************************************************/
  /* We use < rather than <= here because vsnprintf always includes a        */
  /* terminating null byte as the last character in the available space,     */
  /* even if there was not room to include all of the preceding characters.  */
  /* Therefore it is possible to have N bytes available, to try to write N   */
  /* characters into the buffer, and vsnprintf returns N, but has in fact    */
  /* written only the first N-1 characters you asked for, followed by a      */
  /* terminating null.  (Yes, this is in direct contradiction to the man     */
  /* page for (v)snprintf.)                                                  */
  /***************************************************************************/
  if (req_len < avail_len)
  {
    /*************************************************************************/
    /* Success - update the current position in the buffer                   */
    /*************************************************************************/
    TRC_DETAIL((TB, "wrote string"));
    canon_info->current_pos += req_len;
  }
  else
  {
    /*************************************************************************/
    /* Failed - set the overflow flag if required.                           */
    /*************************************************************************/
    TRC_DETAIL((TB, "No space - skipping string"));
    if (!canon_info->overflowed)
    {
      TRC_ALERT((TB, "Overflowed outbuf"));
      canon_info->overflowed = TRUE;
    }
  }

  /***************************************************************************/
  /* Always update the buffer length so we can return the required size if   */
  /* outbuf turns out not to be big enough                                   */
  /***************************************************************************/
  canon_info->current_len += req_len;

  /***************************************************************************/
  /* Tidy up the arg list                                                    */
  /***************************************************************************/
  va_end(ap);

  TRC_EXIT();

  return;

} /* xml_maybe_add_to_outbuf */
