
/**MOD+***********************************************************************/
/* Module:    xmlhdlrs.c                                                     */
/*                                                                           */
/* Purpose:   XML Expat Parsing Handlers                                     */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.16               $ $Modtime::   Jan 11 2001 15:56:50   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <xmlincl.h>

/**PROC+**********************************************************************/
/* Name:      xml_start_elem_handler                                         */
/*                                                                           */
/* Purpose:   Start element handler called for the message and each          */
/*            component                                                      */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name,                                                   */
/*                   attrs        - see expat documentation                  */
/*                                                                           */
/* Operation: For the start of a message, get the message id attribute       */
/*            otherwise call the appropriate AIB start processing routine,   */
/*            passing the component id attribue                              */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_start_elem_handler(void *user_data,
                            const char *name,
                            const char **attrs)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char *comp_id = NULL;
  TRANS_INFO *trans;
  XML_STATE *xml_state;
  char *random;
  OSP_RC rc;
  int status_ok;

  TRC_ENTRY("xml_start_elem_handler");

  /***************************************************************************/
  /* Check the parameters                                                    */
  /***************************************************************************/
  trans = (TRANS_INFO *)user_data;
  TRC_DETAIL((TB, "Start element for trans %p", trans));
  TRC_ASSERT((trans != NULL), (TB, "NULL transaction info for start element"));
  xml_state = (XML_STATE *)trans->xml_state;
  TRC_ASSERT((xml_state != NULL), (TB, "NULL xml state for start element"));

  /***************************************************************************/
  /* Set up default values                                                   */
  /***************************************************************************/
  rc = OSP_SUCCESS;
  xml_state->cancel_fn = NULL;

  /***************************************************************************/
  /* If there's already been an http-level error, there's no point in doing  */
  /* further parsing                                                         */
  /***************************************************************************/
  TRC_DETAIL((TB, "Check http status"));
  OSP_GET_LOCK(&trans->mutex);
  status_ok = (trans->http_status == OSP_STATUS_SUCCESS);
  OSP_RELEASE_LOCK(&trans->mutex);
  if (!status_ok)
  {
    TRC_NORMAL((TB, "Skipping element %s following http error", name));
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Check the element name to decide what to do.  For the message, we need  */
  /* to extract the message id attribute; for any components we need to call */
  /* on to the relevant AIB handlers.                                        */
  /***************************************************************************/
  TRC_DETAIL((TB, "Got start of element %s", name));
  if (!strcmp(name, com_message))
  {
    /*************************************************************************/
    /* Message wrapper: get and store the message id.  If there isn't one,   */
    /* we must fail the whole message                                        */
    /*************************************************************************/
    trans->message_id = aib_find_attr_value(attrs, com_message_id, NULL);
    if (trans->message_id == NULL)
    {
      TRC_ERROR((TB, "Failed to get message id from message wrapper"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
    TRC_DETAIL((TB, "Got message id %s", trans->message_id));

    /*************************************************************************/
    /* Check there is a random attribute.  We don't need to do anything with */
    /* it, but it is required, so we'll fail any message without one         */
    /*************************************************************************/
    random = aib_find_attr_value(attrs, com_random, NULL);
    if (random == NULL)
    {
      TRC_ERROR((TB, "No random attribute with message - failing"));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }
    rm_release_mem(random);
  }
  else
  {
    /*************************************************************************/
    /* This should be a component: get the component id                      */
    /*************************************************************************/
    comp_id = aib_find_attr_value(attrs, com_comp_id, NULL);
    if (comp_id != NULL)
    {
      TRC_DETAIL((TB, "Got component id %s", comp_id));
    }
    else
    {
      /***********************************************************************/
      /* If we fail to get this required attribute, we must fail the whole   */
      /* message, not just the component: even though we may know how to     */
      /* generate a failure for the component, we have no way of identifying */
      /* this particular component which is potentially one of many in the   */
      /* message                                                             */
      /***********************************************************************/
      TRC_ERROR((TB, "Failed to get id for component %s", name));
      rc = OSP_ERROR;
      goto EXIT_LABEL;
    }

    /*************************************************************************/
    /* Now call on to the appropriate AIB                                    */
    /*************************************************************************/
    if (!strcmp(name, com_pricing_ind))
    {
      /***********************************************************************/
      /* Pricing Indication                                                  */
      /***********************************************************************/
      trans->request_count++;
      xml_state->cancel_fn = car_cancel_processing;

      TRC_DETAIL((TB,
                  "Handing PricingIndication to CAR AIB: request_count=%d",
                  trans->request_count));
      rc = car_start_processing(trans, xml_state->parser, AIB_PRICING_IND,
                                                                      comp_id);
    }
    else if ((!strcmp(name, com_auth_req)) || (!strcmp(name, com_auth_s_req)))
    {
      /***********************************************************************/
      /* Authorization Request                                               */
      /***********************************************************************/
      trans->request_count++;
      xml_state->cancel_fn = car_cancel_processing;

      TRC_DETAIL((TB,
                  "Handing AuthorizationRequest to CAR AIB: request_count=%d",
                  trans->request_count));
      rc = car_start_processing(trans,
                                xml_state->parser,
                                (!strcmp(name, com_auth_req)) ? AIB_AUTH_RQ
                                                              : AIB_AUTH_S_RQ,
                                comp_id);
    }
    else if ((!strcmp(name, com_auth_ind)) || (!strcmp(name, com_auth_s_ind)))
    {
      /***********************************************************************/
      /* Authorization Indication                                            */
      /***********************************************************************/
      trans->request_count++;
      xml_state->cancel_fn = car_cancel_processing;

      TRC_DETAIL((TB,
                "Handing AuthorizationIndication to CAR AIB: request_count=%d",
                trans->request_count));
      rc = car_start_processing(trans,
                                xml_state->parser,
                                (!strcmp(name, com_auth_ind)) ? AIB_AUTH_IND
                                                              : AIB_AUTH_S_IND,
                                comp_id);
    }
    else if (!strcmp(name, com_usage_ind))
    {
      /***********************************************************************/
      /* Usage Indication                                                    */
      /***********************************************************************/
      trans->request_count++;
      xml_state->cancel_fn = use_cancel_processing;

      TRC_DETAIL((TB, "Handing UsageIndication to USE AIB: request_count=%d",
                  trans->request_count));
      rc = use_start_processing(trans,
                                xml_state->parser,
                                AIB_USAGE_IND,
                                comp_id);
    }
    else if ((!strcmp(name, com_reauth_req)) ||
                                             (!strcmp(name, com_reauth_s_req)))
    {
      /***********************************************************************/
      /* Reauthorization Request                                             */
      /***********************************************************************/
      trans->request_count++;
      xml_state->cancel_fn = car_cancel_processing;

      TRC_DETAIL((TB,
                 "Handing ReauthorizationRequest to CAR AIB: request_count=%d",
                 trans->request_count));
      rc = car_start_processing(trans,
                                xml_state->parser,
                                (!strcmp(name, com_reauth_req)) ?
                                                               AIB_REAUTH_RQ
                                                             : AIB_REAUTH_S_RQ,
                                comp_id);
    }
    else if (!strcmp(name, com_subs_auth_req))
    {
      /***********************************************************************/
      /* Subscriber Authentication Request                                   */
      /***********************************************************************/
      trans->request_count++;
      xml_state->cancel_fn = car_cancel_processing;

      TRC_DETAIL((TB,
                  "Handing Subs Authen Request to CAR AIB: request_count=%d",
                  trans->request_count));
      rc = car_start_processing(trans, xml_state->parser, AIB_SUBS_AUTHEN_RQ,
                                                                      comp_id);
    }
    else if (!strcmp(name, com_caps_ind))
    {
      /***********************************************************************/
      /* Capabilities Indication                                             */
      /***********************************************************************/
      trans->request_count++;
      xml_state->cancel_fn = car_cancel_processing;

      TRC_DETAIL((TB,
                  "Handing Caps Indication to CAR AIB: request_count=%d",
                  trans->request_count));
      rc = car_start_processing(trans, xml_state->parser, AIB_CAPS_IND,
                                                                      comp_id);
    }
    else
    {
      /***********************************************************************/
      /* Unrecognized component.  We have to fail the entire message since   */
      /* we've got no way of sending a response for just the failing         */
      /* component.  (E.g.  consider a new component                         */
      /* RoamingAuthorizationRequest: we don't even know what the response   */
      /* component name should be, let alone the internal format required to */
      /* send a failure.                                                     */
      /***********************************************************************/
      TRC_ALERT((TB, "Got unhandled component <%s> - failing message", name));
      rc = OSP_ERROR;
    }

  }

EXIT_LABEL:
  /***************************************************************************/
  /* The AIBs must have made a copy of comp_id if they need it for longer    */
  /* than the duration of XXX_start_processing.                              */
  /***************************************************************************/
  rm_release_mem(comp_id);

  /***************************************************************************/
  /* The AIBs are responsible for building any component level error         */
  /* responses, but if they fail in some other way, we must send an http     */
  /* response                                                                */
  /***************************************************************************/
  if (rc == OSP_SUCCESS)
  {
    TRC_DETAIL((TB, "Started processing <%s>", name));
  }
  else
  {
    TRC_ERROR((TB, "AIB start processing error %d for <%s>", rc, name));
    trans->http_status = OSP_STATUS_PARSING_UNSUCCESSFUL;
  }

  TRC_EXIT();

  return;

} /* xml_start_elem_handler */

/**PROC+**********************************************************************/
/* Name:      xml_end_elem_handler                                           */
/*                                                                           */
/* Purpose:   End element handler.  Should only be called for the end of the */
/*            message or any unknown components.                             */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to transaction control structure */
/*            IN     name         - see expat documentation                  */
/*                                                                           */
/* Operation: Do nothing!  The AIBs should handle the end of any components; */
/*            if its an unknown component or the end of the message, then    */
/*            there's still nothing to do - our next action occurs when      */
/*            parsing completes.                                             */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_end_elem_handler(void *user_data, const char *name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("xml_end_elem_handler");

  /***************************************************************************/
  /* Just trace what we got                                                  */
  /***************************************************************************/
  TRC_DETAIL((TB, "Got end of %s in trans %p", name, user_data));

  TRC_EXIT();

  return;

} /* xml_end_elem_handler */

/**PROC+**********************************************************************/
/* Name:      xml_character_data_handler                                     */
/*                                                                           */
/* Purpose:   Called with XML character data.  Should only be called for any */
/*            unrecognized components                                        */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     user_data    - pointer to the transaction's control     */
/*                                  structure                                */
/*            IN     data         - the character data (not NULL-terminated) */
/*            IN     len          - length of data                           */
/*                                                                           */
/* Operation: Nothing to do.                                                 */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_character_data_handler(void *user_data, const char *data, int len)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("xml_character_data_handler");

  /***************************************************************************/
  /* Just trace what we got                                                  */
  /***************************************************************************/
  TRC_DETAIL((TB, "Unhandled char data for trans %p", user_data));
  TRC_DATA("Data:", data, len);

  TRC_EXIT();

  return;

} /* xml_character_data_handler */

/**PROC+**********************************************************************/
/* Name:      xml_set_default_handlers                                       */
/*                                                                           */
/* Purpose:   Called by AIBs to indicated that expat's parsing callbacks     */
/*            should be redirected to point to the XML layer.                */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction info, which contains the     */
/*                                  parser instance                          */
/*                                                                           */
/* Operation: Check the supplied parameters                                  */
/*            Restore the xml layer expat handlers                           */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_set_default_handlers(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  XML_STATE *xml_state;
  OSP_RC rc = OSP_SUCCESS;

  TRC_ENTRY("xml_set_default_handlers");

  /***************************************************************************/
  /* Check we've got valid parameters                                        */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Passed NULL trans info"));
  xml_state = (XML_STATE *)trans->xml_state;
  TRC_ASSERT((xml_state != NULL), (TB, "Passed NULL xml state info"));
  TRC_ASSERT((xml_state->parser != NULL), (TB, "Passed NULL parser"));

  /***************************************************************************/
  /* Reset the cancel function.  Doing this here (i) ensures that            */
  /* XXX_cancel_processing is not called twice (ii) gets the                 */
  /* request/response counts right, in the case where an XML parse error     */
  /* occurs just after we have finished successfully parsing an OSP          */
  /* component.                                                              */
  /***************************************************************************/
  xml_state->cancel_fn = NULL;

  /***************************************************************************/
  /* Set the handlers                                                        */
  /***************************************************************************/
  XML_SetElementHandler(*(xml_state->parser), xml_start_elem_handler,
                                                         xml_end_elem_handler);
  XML_SetCharacterDataHandler(*(xml_state->parser),
                                                   xml_character_data_handler);
  XML_SetUserData(*(xml_state->parser), trans);
  TRC_DETAIL((TB, "Reset handlers for trans %p", trans));

  TRC_EXIT();

  return;

} /* xml_set_default_handlers */

/**PROC+**********************************************************************/
/* Name:      xml_set_null_handlers                                          */
/*                                                                           */
/* Purpose:   Sets the default expat element and char data handlers          */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     parser       - parser to set handlers for               */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_set_null_handlers(XML_Parser parser)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  TRC_ENTRY("xml_set_null_handlers");

  /***************************************************************************/
  /* NULL means "use expat default handlers"                                 */
  /***************************************************************************/
  TRC_DETAIL((TB, "Setting handlers to NULL"));
  XML_SetElementHandler(parser, NULL, NULL);
  XML_SetCharacterDataHandler(parser, NULL);

  TRC_EXIT();

  return;

} /* xml_set_null_handlers */
