
/**MOD+***********************************************************************/
/* Module:    xmlmain.c                                                      */
/*                                                                           */
/* Purpose:   XML Layer control routines                                     */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.25               $ $Modtime::   Jan 11 2001 15:56:50   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <xmlincl.h>

/**PROC+**********************************************************************/
/* Name:      xml_init                                                       */
/*                                                                           */
/* Purpose:   Initializes the xml layer                                      */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            OSP_ERROR   - failure                                          */
/*            other       - return code from AIB component                   */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: initialize the AIB components                                  */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC xml_init(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;

  TRC_ENTRY("xml_init");

  /***************************************************************************/
  /* Initialize the AIB components                                           */
  /***************************************************************************/
#ifdef OSP_DATA_COVERAGE
  rc = aib_init_data_coverage();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Data coverage init failed with return code %d", rc));
    goto EXIT_LABEL;
  }
#endif

  rc = car_init();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Car init failed with return code %d", rc));
    goto EXIT_LABEL;
  }

  rc = use_init();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "USE init failed with return code %d", rc));
    goto EXIT_LABEL;
  }

  TRC_NORMAL((TB, "XML initialized successfully"));

EXIT_LABEL:

  TRC_EXIT();

  return(rc);

} /* xml_init */

/**PROC+**********************************************************************/
/* Name:      xml_term                                                       */
/*                                                                           */
/* Purpose:   terminates the xml layer                                       */
/*                                                                           */
/* Returns:   OSP_SUCCESS - success                                          */
/*            OSP_ERROR   - failure                                          */
/*            other       - return code from AIB component                   */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: Terminate the AIB components                                   */
/*                                                                           */
/*            Note that any parsing operation in progress when this function */
/*            is called can safely run to completion                         */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC xml_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;

  TRC_ENTRY("xml_term");

  /***************************************************************************/
  /* Terminate the AIB components.  Note that we'll continue terminating     */
  /* susbsequent components after a failure to avoid potential resource      */
  /* leaks                                                                   */
  /***************************************************************************/
  rc = use_term();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "USE term failed with return code %d", rc));
  }

  rc = car_term();
  if (rc != OSP_SUCCESS)
  {
    TRC_ERROR((TB, "Car term failed with return code %d", rc));
  }

#ifdef OSP_DATA_COVERAGE
  aib_term_data_coverage();
#endif

  TRC_NORMAL((TB, "XML Terminated successfully"));

  TRC_EXIT();

  return(rc);

} /* xml_term */

/**PROC+**********************************************************************/
/* Name:      xml_request_received                                           */
/*                                                                           */
/* Purpose:   To pass an OSP request up to the XML layer to be parsed and    */
/*            dispatched to the appropriate AIBs.                            */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction information containing the   */
/*                                  xml to be parsed                         */
/*                                                                           */
/* Operation: create a parser instance                                       */
/*            reserve space for message header in transactions send buffer   */
/*            set up the expat handlers                                      */
/*            parse the xml                                                  */
/*            call xml_maybe_send() to see if response should be sent        */
/*                                                                           */
/*            A note on error handling: in case of                           */
/*                                                                           */
/*            - an XML parsing error, the XML layer sends an http failure.   */
/*            (Parsing errors include XML that is not well formed, missing   */
/*            required attributes (message and component ids and random      */
/*            data), and unrecognized components.)                           */
/*                                                                           */
/*            - an error within a component, the component start handler is  */
/*            responsible for sending the error.  (For an unrecognized       */
/*            critical child element or for a memory error that prevents the */
/*            parsing completing, this should be a component-level failure.) */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_request_received(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  XML_Parser parser;
  XML_STATE xml_state;
  enum XML_Error xml_error;
  char *xml_error_string;
  OSP_RC cancel_rc;
  int parse_rc;
  const char *xml_pi = XML_PI;
  char *into_recv_buf;

  TRC_ENTRY("xml_request_received");

  /***************************************************************************/
  /* Check what we've been passed                                            */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "XML passed NULL transaction info"));
  TRC_NORMAL((TB, "Parsing transaction %p", trans));

#ifdef OSP_MEASURE_TRANSACTION
  /***************************************************************************/
  /* Set the transaction start parsing time.                                 */
  /***************************************************************************/
  SM_GETTIMES(trans->timings[SM_TIMINGS_START_PARSE]);
#endif

  /***************************************************************************/
  /* Initialize the transaction                                              */
  /***************************************************************************/
  trans->aib_state = NULL;
  trans->sign_response = FALSE;
  trans->parsing = FALSE;
  trans->request_count = 0;
  trans->response_count = 0;
  trans->http_status = OSP_STATUS_SUCCESS;
  trans->message_id = NULL;

  /***************************************************************************/
  /* Allocate a string pool for parsed character data.  Note that it should  */
  /* be impossible for the sum of the sizes of parsed strings to exceed the  */
  /* total size of the XML to be parsed.                                     */
  /***************************************************************************/
  trans->pool_size = trans->recv_buffer_end - trans->recv_buffer_start + 1;
  trans->pool = (char *)rm_get_mem(trans->pool_size);
  if (trans->pool != NULL)
  {
    TRC_DETAIL((TB, "String pool allocated: trans=%p pool=%p size=%d",
                trans,
                trans->pool,
                trans->pool_size));
    trans->pool_position = 0;
  }
  else
  {
    TRC_ERROR((TB, "Couldn't allocate string pool memory: trans=%p size=%d",
                trans,
                trans->pool_size));
    trans->pool_size = 0;
    trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Create the expat parser instance, if we don't already have one for this */
  /* thread.                                                                 */
  /***************************************************************************/
  parser = trans->thread->parser;
  if (parser != NULL)
  {
    TRC_DETAIL((TB, "Reuse this thread's parser: %p", parser));
  }
  else
  {
    parser = XML_ParserCreate(NULL);
    TRC_DETAIL((TB, "Created new parser: %p", parser));

    if (parser != NULL)
    {
      /***********************************************************************/
      /* Parse a top-level wrap element start tag.  Note that XML only       */
      /* allows one top-level element.                                       */
      /***********************************************************************/
      parse_rc = XML_Parse(parser,
                           XML_WRAP_ELEMENT,
                           XML_WRAP_ELEMENT_LEN,
                           FALSE);
      if (parse_rc != 0)
      {
        /*********************************************************************/
        /* Save parser in thread storage for use next time around.           */
        /*********************************************************************/
        trans->thread->parser = parser;
      }
      else
      {
        xml_error = XML_GetErrorCode(parser);
        xml_error_string = (char *)XML_ErrorString(xml_error);
        TRC_ERROR((TB, "Failed to parse wrap element: error %d (%s)",
                   xml_error,
                   (xml_error_string != NULL) ? xml_error_string : "NULL"));
        XML_ParserFree(parser);
        parser = NULL;
        trans->http_status = OSP_STATUS_PARSING_UNSUCCESSFUL;
      }
    }
  }

  if (parser != NULL)
  {
    /*************************************************************************/
    /* Reserve some space in the transaction's send buffer for the message   */
    /* header that will be pre-pended.  Note that access to the transaction  */
    /* limited to just this thread until we have made some callups into the  */
    /* applications, so we can access these variables without holding the    */
    /* mutex.                                                                */
    /*************************************************************************/
    TRC_ASSERT(((trans->send_buffer_end + XML_WRAPPER_START_LEN) <
                                                      trans->send_buffer_size),
        (TB, "Not enough space for XML wrapper in pre-allocated send buffer"));
    trans->send_buffer_start += XML_WRAPPER_START_LEN;
    trans->send_buffer_end += XML_WRAPPER_START_LEN;

    /*************************************************************************/
    /* Initialize the parsing state state information                        */
    /*************************************************************************/
    xml_state.parser = &parser;
    xml_state.cancel_fn = NULL;

    /*************************************************************************/
    /* Note that we're using a stack variable rather than allocated memory   */
    /* for the state.  This is legitimate because parsing will complete      */
    /* within this function, i.e.  the stack variable will not go out of     */
    /* scope                                                                 */
    /*************************************************************************/
    trans->xml_state = &xml_state;

    /*************************************************************************/
    /* Skip over the XML introductory processing instruction, if it's there. */
    /*************************************************************************/
    for (into_recv_buf = trans->recv_buffer + trans->recv_buffer_start;
         (trans->recv_buffer_start < trans->recv_buffer_end) &&
         (*into_recv_buf == *xml_pi);
         trans->recv_buffer_start++, into_recv_buf++, xml_pi++)
    {
    }

    /*************************************************************************/
    /* And start parsing                                                     */
    /*************************************************************************/
    TRC_DATA("Parsing buffer",
             trans->recv_buffer + trans->recv_buffer_start,
             trans->recv_buffer_end - trans->recv_buffer_start);
    trans->parsing = TRUE;

    /*************************************************************************/
    /* We wrap the message in an extra set of tags so we can detect both     */
    /* missing and partial end tags.  Initially we use the expat default     */
    /* handlers which will just validate the XML                             */
    /*************************************************************************/
    xml_set_null_handlers(parser);
    parse_rc = XML_Parse(parser,
                         XML_MESSAGE_WRAP_STAG,
                         XML_MESSAGE_WRAP_STAG_LEN,
                         FALSE);
    if (parse_rc != 0)
    {
      /***********************************************************************/
      /* Set up our expat handlers                                           */
      /***********************************************************************/
      xml_set_default_handlers(trans);
      parse_rc = XML_Parse(parser,
                           trans->recv_buffer + trans->recv_buffer_start,
                           trans->recv_buffer_end - trans->recv_buffer_start,
                           FALSE);
      if (parse_rc != 0)
      {
        /*********************************************************************/
        /* Reset the handlers                                                */
        /*********************************************************************/
        xml_set_null_handlers(parser);
        parse_rc = XML_Parse(parser,
                             XML_MESSAGE_WRAP_ETAG,
                             XML_MESSAGE_WRAP_ETAG_LEN,
                             FALSE);
      }
    }

    TRC_DETAIL((TB, "...finished parsing: rc=%d", parse_rc));

    /*************************************************************************/
    /* Now see what happened during parsing                                  */
    /*************************************************************************/
    if (parse_rc != 0)
    {
      /***********************************************************************/
      /* We parsed the buffer.  See if we need to send a response            */
      /***********************************************************************/
      TRC_DETAIL((TB, "Parsed data successfully"));
    }
    else
    {
      /***********************************************************************/
      /* Record the error                                                    */
      /***********************************************************************/
      xml_error = XML_GetErrorCode(parser);
      xml_error_string = (char *)XML_ErrorString(xml_error);
      TRC_ERROR((TB, "Got parsing error %d (%s)", xml_error,
                      (xml_error_string != NULL) ? xml_error_string : "NULL"));

      /***********************************************************************/
      /* Call any current cancel processing function                         */
      /***********************************************************************/
      if ((*(xml_state.cancel_fn)) != NULL)
      {
        /*********************************************************************/
        /* The transaction request count is incremented whenever cancel_fn   */
        /* is set non-NULL, immediately before calling one of the AIBs'      */
        /* XXX_start_processing functions.  For balance, therefore, we need  */
        /* to increment the transaction response count here.                 */
        /*********************************************************************/
        OSP_GET_LOCK(&(trans->mutex));
        trans->response_count++;
        TRC_DETAIL((TB, "Incremented transaction response count:"
                    " request_count=%d response_count=%d",
                    trans->request_count,
                    trans->response_count));
        OSP_RELEASE_LOCK(&(trans->mutex));

        /*********************************************************************/
        /* This should never fail                                            */
        /*********************************************************************/
        TRC_ALERT((TB, "Calling AIB cancel fn"));
        cancel_rc = (*(xml_state.cancel_fn))(trans);
        TRC_ASSERT((cancel_rc == OSP_SUCCESS),
                               (TB, "Cancel fn failed with rc %d", cancel_rc));
      }

      /***********************************************************************/
      /* All we can do now is send an http failure code, based on the expat  */
      /* error                                                               */
      /***********************************************************************/
      if (xml_error == XML_ERROR_NO_MEMORY)
      {
        TRC_DETAIL((TB, "Mapping expat no memory to http transient problem"));
        trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
      }
      else
      {
        TRC_DETAIL((TB, "Return parsing unsucessful"));
        trans->http_status = OSP_STATUS_PARSING_UNSUCCESSFUL;
      }

      /***********************************************************************/
      /* Since parsing failed, the parser may be left in an unusual state.   */
      /* Therefore we don't want to use this parser instance again.          */
      /***********************************************************************/
      XML_ParserFree(parser);
      trans->thread->parser = NULL;
    }

    /*************************************************************************/
    /* Clear the parsing flag                                                */
    /*************************************************************************/
    OSP_GET_LOCK(&trans->mutex);
    trans->parsing = FALSE;
    OSP_RELEASE_LOCK(&trans->mutex);

    /*************************************************************************/
    /* We're done.                                                           */
    /*************************************************************************/
    TRC_NORMAL((TB, "Parsing complete"));
  }
  else
  {
    /*************************************************************************/
    /* If we can't parse the message, all we can do is send an http failure  */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to create parser instance"));
    trans->http_status = OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
  }

EXIT_LABEL:
  /***************************************************************************/
  /* See if the transaction response is ready to be sent.  Note that         */
  /* xml_maybe_send() will release the lock for us                           */
  /***************************************************************************/
  OSP_GET_LOCK(&trans->mutex);
  xml_maybe_send(trans);

  TRC_EXIT();

  return;

} /* xml_request_received */

/**PROC+**********************************************************************/
/* Name:      xml_maybe_send                                                 */
/*                                                                           */
/* Purpose:   Pass transaction to S/MIME if complete.                        */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     trans        - transaction to be examined               */
/*                                                                           */
/* Operation: If transaction is fully processed and all responses have       */
/*            been received, wrap the data in an OSP Message element and     */
/*            send it                                                        */
/*                                                                           */
/*            THE CALLER MUST HOLD THE TRANSACTION MUTEX                     */
/*                                                                           */
/**PROC-**********************************************************************/

void xml_maybe_send(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc;

  TRC_ENTRY("xml_maybe_send");

  /***************************************************************************/
  /* Has the transaction already been placed on the send queue?              */
  /***************************************************************************/
  if (trans->on_send_queue)
  {
    TRC_DETAIL((TB, "Already on send queue"));
    OSP_RELEASE_LOCK(&trans->mutex);
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Are all the components in the message fully parsed?                     */
  /***************************************************************************/
  if (trans->parsing)
  {
    TRC_NORMAL((TB, "Transaction %p still being parsed", trans));
    OSP_RELEASE_LOCK(&trans->mutex);
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* And have all the responses been built?                                  */
  /***************************************************************************/
  TRC_NORMAL((TB, "Transaction %p fully parsed", trans));
  TRC_DETAIL((TB, "Request count %d, response count %d",
                               trans->request_count, trans->response_count));

  if (trans->request_count != trans->response_count)
  {
    TRC_NORMAL((TB, "but still has outstanding responses"));
    OSP_RELEASE_LOCK(&trans->mutex);
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* We could get here with no failures, but having never seen anything we   */
  /* actually recognized as a message.  In this case there will be no        */
  /* requests                                                                */
  /***************************************************************************/
  TRC_NORMAL((TB, "and ready to send"));
  if (trans->request_count == 0)
  {
    TRC_ALERT((TB, "Trans %p contained no requests", trans));
    trans->http_status = OSP_STATUS_BAD_REQUEST;
  }

  /***************************************************************************/
  /* If there's not been any http-level failure, we add the Message start    */
  /* and end tags                                                            */
  /***************************************************************************/
  if (trans->http_status == OSP_STATUS_SUCCESS)
  {
    TRC_NORMAL((TB, "Attaching <Message> wrapper"));
    rc = xml_add_message_wrapper(trans);
    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* We'll have to fail the whole message now by setting the http error  */
      /* code                                                                */
      /***********************************************************************/
      TRC_ERROR((TB, "Failure adding Message wrapper to trans %p", trans));
      trans->http_status =  OSP_STATUS_TRANSIENT_PROBLEM_IN_SERVER;
    }
  }

  /***************************************************************************/
  /* Set the on send queue flag, then release the mutex and send the         */
  /* response                                                                */
  /***************************************************************************/
  trans->on_send_queue = TRUE;
  OSP_RELEASE_LOCK(&trans->mutex);
  rc = smime_send_response(trans);
  if (rc != OSP_SUCCESS)
  {
    /*************************************************************************/
    /* There's no more we can do.  It's the lower layer's job to free up the */
    /* transaction info, so we just record the failure                       */
    /*************************************************************************/
    TRC_ERROR((TB, "Failed to send response for transaction %p", trans));
  }

EXIT_LABEL:

  TRC_EXIT();

  return;

} /* xml_maybe_send */

/**PROC+**********************************************************************/
/* Name:      xml_add_message_wrapper                                        */
/*                                                                           */
/* Purpose:   Prepends the Message start tag and appends the close tag       */
/*                                                                           */
/* Returns:   OSP_SUCCESS - Message wrapper added successfully               */
/*            OSP_ERROR   - A problem occurred                               */
/*                                                                           */
/* Params:    IN     trans        - transaction info                         */
/*                                                                           */
/* Operation: THE CALLER MUST HOLD THE TRANSACTION MUTEX                     */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC xml_add_message_wrapper(TRANS_INFO *trans)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;
  char message[XML_WRAPPER_START_LEN];
  int message_len;
  char *dest;
  unsigned long random_data;
  XML_STATE *xml_state;

  TRC_ENTRY("xml_add_message_wrapper");

  /***************************************************************************/
  /* Check what we've been given                                             */
  /***************************************************************************/
  TRC_ASSERT((trans != NULL), (TB, "Supplied with NULL trans info"));
  TRC_ASSERT((trans->message_id != NULL),
                                  (TB, "Trans %p had NULL message id", trans));
  xml_state = (XML_STATE *)trans->xml_state;
  TRC_ASSERT((xml_state != NULL),
                              (TB, "Trans %p had NULL xml state info", trans));

  /***************************************************************************/
  /* Set up the random data                                                  */
  /***************************************************************************/
  ccm_random((void *)&random_data, sizeof(random_data));

  /***************************************************************************/
  /* Build the start of the XML message, with attributes                     */
  /* - message ID (stored in trans info)                                     */
  /* - random string                                                         */
  /***************************************************************************/
  message_len = OSP_SNPRINTF((
                        message,
                        XML_WRAPPER_START_LEN,
                     "<?xml version=\"1.0\"?>\n<%s %s=\"%s\" %s=\"%010lu\">\n",
                        com_message,
                        com_message_id,
                        trans->message_id,
                        com_random,
                        random_data));
  TRC_ASSERT((message_len < XML_WRAPPER_START_LEN),
        (TB, "Resulting message would be too large at %d bytes", message_len));

  /***************************************************************************/
  /* Work out where to put the message: it has to go before the start of the */
  /* existing data, which is not the same as at the start of the buffer      */
  /***************************************************************************/
  TRC_ASSERT((trans->send_buffer_start > strlen(message)),
                                 (TB, "Not enough space for message header!"));
  trans->send_buffer_start -= strlen(message);
  dest = trans->send_buffer + trans->send_buffer_start;

  /***************************************************************************/
  /* Memcpy the string into place (we can't use strcpy as that would         */
  /* NULL-terminate the buffer and lose all the components)                  */
  /***************************************************************************/
  TRC_NORMAL((TB, "Prepending message header '%s'", message));
  memcpy(dest, message, strlen(message));

  /***************************************************************************/
  /* Append the message close tag                                            */
  /***************************************************************************/
  if (aib_space_available(trans, 2 + 2 + strlen(com_message)))
  {
    AIB_ADD_DATA(trans, com_lt_sl, 2);
    AIB_ADD_NULL(trans, com_message);
    AIB_ADD_DATA(trans, com_gt_nl, 2);
  }
  else
  {
    TRC_ERROR((TB, "No space for close message tag"));
    rc = OSP_ERROR;
  }

  TRC_EXIT();

  return(rc);

} /* xml_add_message_wrapper */


