#ifndef TRCEXTRN_INCLUDED
#define TRCEXTRN_INCLUDED

/**INC+***********************************************************************/
/* Header:    trcextrn.h                                                     */
/*                                                                           */
/* Purpose:   OSP Tracing - external API and definitions                     */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.12               $ $Modtime::   Jan 11 2001 15:56:14   $  */
/*                                                                           */
/**INC-***********************************************************************/

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <assert.h>

/*****************************************************************************/
/* Trace types                                                               */
/*****************************************************************************/
#define TRC_TYPE_NONE    0x00
#define TRC_TYPE_ERROR   0x01
#define TRC_TYPE_ALERT   0x02
#define TRC_TYPE_NORM    0x04
#define TRC_TYPE_DETAIL  0x08
#define TRC_TYPE_FLOW    0x10

/*****************************************************************************/
/* Pre-defined trace levels                                                  */
/*****************************************************************************/
#define TRC_LEVEL_DETAIL      (TRC_TYPE_ERROR | TRC_TYPE_ALERT  |   \
                               TRC_TYPE_NORM  | TRC_TYPE_DETAIL |   \
                               TRC_TYPE_FLOW)

#define TRC_LEVEL_NORMAL      (TRC_TYPE_ERROR | TRC_TYPE_ALERT  |   \
                               TRC_TYPE_NORM  | TRC_TYPE_FLOW)

#define TRC_LEVEL_ALERT       (TRC_TYPE_ERROR | TRC_TYPE_ALERT)

#define TRC_LEVEL_ERROR        TRC_TYPE_ERROR

#define TRC_LEVEL_NONE         TRC_TYPE_NONE

#define TRC_LEVEL_DEFAULT      TRC_LEVEL_ALERT

/*****************************************************************************/
/* global variable used to hold trace type flags                             */
/*****************************************************************************/
extern int trc_trace_types;

/*****************************************************************************/
/* global initialization flag                                                */
/*****************************************************************************/
extern int trc_initialized;

/*****************************************************************************/
/* Trace macros - set to nothing if tracing is not enabled                   */
/*                                                                           */
/* Usage is of the form                                                      */
/*                                                                           */
/*   TRC_ENTRY("fn_name"_;                                                   */
/*   TRC_NORMAL((TB, "Useful value %d", variable));                          */
/*   TRC_DATA("Some useful data", data_buffer, length_of_data);              */
/*   TRC_ASSERT((a == b), (TB, "A (%d) not equal to B (%d)", a, b));         */
/*   TRC_EXIT();                                                             */
/*                                                                           */
/* Note that the TRC_API_XXX and TRC_THREAD_XXX forms also set up and        */
/* restore the signal mask to prevent interuptions inside our external API   */
/* functions                                                                 */
/*****************************************************************************/

#define TRC_ENTRY(FN_NAME)        char *trc_fn = FN_NAME;                     \
                                  TRC_FLOW((TB, "Entry {"));

#define TRC_EXIT()                TRC_FLOW((TB, "} Exit"));

#define TRC_API_ENTRY(FN_NAME)    char *trc_fn = FN_NAME;                     \
                                  sigset_t set, old;                          \
                                  sigfillset(&set);                           \
                                  sigemptyset(&old);                          \
                                  pthread_sigmask(SIG_SETMASK, &set, &old);   \
                                  if (!trc_initialized)                       \
                                  {                                           \
                                    trc_init(TRC_LEVEL_DEFAULT);              \
                                  }                                           \
                                  TRC_FLOW((TB, "API Entry {{"));

#define TRC_API_EXIT()            TRC_FLOW((TB, "}} API Exit"));              \
                                  pthread_sigmask(SIG_SETMASK, &old, NULL);

#define TRC_THREAD_ENTRY(FN_NAME) char *trc_fn = FN_NAME;                     \
                                  sigset_t set;                               \
                                  sigfillset(&set);                           \
                                  pthread_sigmask(SIG_SETMASK, &set, NULL);   \
                                  if (!trc_initialized)                       \
                                  {                                           \
                                    trc_init(TRC_LEVEL_DEFAULT);              \
                                  }                                           \
                                  TRC_FLOW((TB, "Thread Entry {{"));          \

#define TRC_THREAD_EXIT()         TRC_FLOW((TB, "}} Thread Exit"));

/*****************************************************************************/
/* The release build doesn't include the most detailed levels of trace       */
/*****************************************************************************/
#ifdef OSP_DEBUG
#define TRC_FLOW(PARAMS)        TRACE(TRC_TYPE_FLOW, PARAMS)
#define TRC_DETAIL(PARAMS)      TRACE(TRC_TYPE_DETAIL, PARAMS)
#define TRC_NORMAL(PARAMS)      TRACE(TRC_TYPE_NORM, PARAMS)
#define TRC_DATA(TEXT, PTR, LENGTH)                                           \
  if (trc_trace_types & TRC_TYPE_DETAIL)                                      \
  {                                                                           \
    trc_trace_data(TEXT, (const char *)PTR, LENGTH,                           \
                   __LINE__, trc_fn, __FILE__);                               \
  }
#else
#define TRC_FLOW(PARAMS)
#define TRC_DETAIL(PARAMS)
#define TRC_NORMAL(PARAMS)
#define TRC_DATA(TEXT, PTR, LENGTH)
#endif

#define TRC_ALERT(PARAMS)       TRACE(TRC_TYPE_ALERT, PARAMS)
#define TRC_ERROR(PARAMS)       TRACE(TRC_TYPE_ERROR, PARAMS)
#define TRC_ASSERT(X, PARAMS)        \
  if (!(X))                          \
  {                                  \
    TRACE(TRC_TYPE_ERROR, PARAMS);   \
    assert(X);                       \
  }

/*****************************************************************************/
/* Trace OS error information on failure of an OS call                       */
/*****************************************************************************/
#define TRC_OS_ERROR(X)                                           \
{                                                                 \
  int trc_err = errno;                                            \
  TRC_ERROR(X);                                                   \
  TRC_ERROR((TB, "errno %d = %s", trc_err, strerror(trc_err)));   \
}

/*****************************************************************************/
/* Now the clever bits...                                                    */
/*                                                                           */
/* Combined with the TB macro below, 'sprintf PARAMS' will expand to         */
/* something like                                                            */
/*                                                                           */
/*    sprintf(trc_get_trace_buffer(), "Useful variable %d", variable)        */
/*                                                                           */
/* trc_get_trace_buffer() grabs a mutex to lock the trace buffer before      */
/* returning a pointer to that buffer; trc_trace() formats the output line   */
/* and writes it to the trace file before freeing the mutex                  */
/*****************************************************************************/
#define TB  trc_get_trace_buffer()

#define TRACE(TYPE, PARAMS)                        \
  if (trc_trace_types & (TYPE))                    \
  {                                                \
    sprintf PARAMS;                                \
    trc_trace((TYPE), __LINE__, trc_fn, __FILE__); \
  }

/*****************************************************************************/
/* trcmain.c                                                                 */
/*****************************************************************************/
OSP_RC trc_init(int);

OSP_RC trc_term(void);

char *trc_get_trace_buffer(void);

void trc_trace(int, int, char *, char *);

void trc_trace_data(char *, const char *, int, int line, char *, char *);

#endif

