
/**MOD+***********************************************************************/
/* Module:    trcmain.c                                                      */
/*                                                                           */
/* Purpose:   OSP Tracing - function code                                    */
/*                                                                           */
/* Copyright (c) 2000, 2001 Data Connection Limited                          */
/* Copyright (c) 2000, 2001 Cisco Systems, Inc. All Rights Reserved.         */
/*                                                                           */
/* All software incorporated in the "OpenOSP Server" (except for those       */
/* portions incorporating third party software specifically identified as    */
/* being licensed under separate license) are licensed under the terms of    */
/* the Cisco Public Source License ver. 1.0 (2000) found in                  */
/* $OSPROOT/license.txt.  Additional copies of the license are available     */
/* from opensource@cisco.com.                                                */
/*                                                                           */
/* Also, before using this software, please see $OSPROOT/x509_ipr.txt for    */
/* information regarding intellectual property rights.                       */
/*                                                                           */
/* $Revision::   1.12               $ $Modtime::   Jan 11 2001 15:56:14   $  */
/*                                                                           */
/**MOD-***********************************************************************/

#include <trcincl.h>

/**PROC+**********************************************************************/
/* Name:      trc_init                                                       */
/*                                                                           */
/* Purpose:   Initialize the trace component                                 */
/*                                                                           */
/* Returns:   OSP_SUCCESS          - trace was initialized successfully      */
/*            OSP_ERR_ALREADY_INIT - trace was already initialized           */
/*            OSP_ERROR            - there was a problem; trace will not be  */
/*                                   available                               */
/*                                                                           */
/* Params:    IN     types        - types of trace to output                 */
/*                                                                           */
/* Operation: Set up the trace buffer mutex and the trace file               */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC trc_init(int types)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_ERROR;
  int mutex_created = FALSE;

  /***************************************************************************/
  /* Check for reinitialization                                              */
  /***************************************************************************/
  if (trc_initialized)
  {
    /*************************************************************************/
    /* Just update the trace level                                           */
    /*************************************************************************/
    trc_trace_types = types;
    rc = OSP_ERR_ALREADY_INIT;
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Initially set the trace type to 0.  This will minimize subsequent       */
  /* problems if anything goes wrong later on                                */
  /***************************************************************************/
  trc_trace_types = 0;

  /***************************************************************************/
  /* Try to create the trace mutex.  If this fails, we will not trace, but   */
  /* we'll leave it to the caller to decide it it wants to continue without  */
  /* tracing                                                                 */
  /*                                                                         */
  /* Note that we can't use osp_create_mutex here since that uses tracing... */
  /***************************************************************************/
  if (pthread_mutex_init(&trc_buffer_mutex, NULL))
  {
    goto EXIT_LABEL;
  }
  mutex_created = TRUE;

  /***************************************************************************/
  /* Get the buffer mutex - this makes sure that no-one tries to write to    */
  /* the file before us!                                                     */
  /***************************************************************************/
  trc_get_trace_mutex_lock();

  /***************************************************************************/
  /* Open the trace file - with a similar note about failures                */
  /***************************************************************************/
  tfp = fopen("osptrc", "a");
  if (!tfp)
  {
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Finally, if we get this far, we actually set up the trace types         */
  /***************************************************************************/
  trc_trace_types = types;

  /***************************************************************************/
  /* Record the start time                                                   */
  /***************************************************************************/
  fprintf(tfp, "\nTrace begun at "TRC_TIMESTAMP_FMT"\n", trc_get_timestamp());
  fflush(tfp);
  trc_initialized = TRUE;
  rc = OSP_SUCCESS;

EXIT_LABEL:
  /***************************************************************************/
  /* Tidy up.                                                                */
  /***************************************************************************/
  if (mutex_created)
  {
    pthread_mutex_unlock(&trc_buffer_mutex);

    if (rc != OSP_SUCCESS)
    {
      /***********************************************************************/
      /* Note that we don't use the osp common function because that         */
      /* includes trace...                                                   */
      /***********************************************************************/
      pthread_mutex_destroy(&trc_buffer_mutex);
    }
  }

  return(rc);

} /* trc_init */

/**PROC+**********************************************************************/
/* Name:      trc_term                                                       */
/*                                                                           */
/* Purpose:   Terminate the trace component                                  */
/*                                                                           */
/* Returns:   OSP_SUCCESS      - Trace terminated successfully               */
/*            OSP_ERR_NOT_INIT - Trace was not initialized                   */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: Close the trace file and destroy the trace buffer mutex        */
/*                                                                           */
/**PROC-**********************************************************************/

OSP_RC trc_term(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  OSP_RC rc = OSP_SUCCESS;

  /***************************************************************************/
  /* Check we were initialized                                               */
  /***************************************************************************/
  if (!trc_initialized)
  {
    rc = OSP_ERR_NOT_INIT;
    goto EXIT_LABEL;
  }
  trc_initialized = FALSE;

  /***************************************************************************/
  /* Set the trace type to 0.  This will minimize subsequent problems if     */
  /* someone traces after term is called                                     */
  /***************************************************************************/
  trc_trace_types = 0;

  /***************************************************************************/
  /* Get the buffer mutex - this makes sure that no-one is writing to the    */
  /* file when we close it                                                   */
  /***************************************************************************/
  trc_get_trace_mutex_lock();
  fprintf(tfp, "Trace ended at "TRC_TIMESTAMP_FMT"\n", trc_get_timestamp());;

  /***************************************************************************/
  /* Close the trace file                                                    */
  /***************************************************************************/
  fclose(tfp);
  tfp = NULL;

  /***************************************************************************/
  /* Now destroy the mutex.  Note that we don't use the osp common function  */
  /* because that includes trace...                                          */
  /***************************************************************************/
  pthread_mutex_unlock(&trc_buffer_mutex);
  pthread_mutex_destroy(&trc_buffer_mutex);

EXIT_LABEL:

  return(rc);

} /* trc_term */

/**PROC+**********************************************************************/
/* Name:      trc_get_trace_buffer                                           */
/*                                                                           */
/* Purpose:   gets trace buffer mutex before returning a pointer to the      */
/*            trace buffer.                                                  */
/*                                                                           */
/* Returns:   pointer to trace buffer                                        */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

char *trc_get_trace_buffer(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  /***************************************************************************/
  /* get the trace buffer lock and return a pointer to the trace buffer      */
  /***************************************************************************/
  trc_get_trace_mutex_lock();

  return(trace_buffer);

} /* trc_get_trace_buffer */

/**PROC+**********************************************************************/
/* Name:      trc_trace                                                      */
/*                                                                           */
/* Purpose:   formats supplied trace string and writes it to file before     */
/*            freeing the trace buffer mutex                                 */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     level        - trace level, used in formatting          */
/*            IN     line         - line number of trace statement           */
/*            IN     fn_name      - function name                            */
/*            IN     file_name    - file name                                */
/*                                                                           */
/**PROC-**********************************************************************/

void trc_trace(int level, int line, char *fn_name, char *file_name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  char sep;

  /***************************************************************************/
  /* Check we're initialized and that we hae a file to write to              */
  /***************************************************************************/
  if (!trc_initialized || (tfp == NULL))
  {
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* Format is hh:mm:ss#filename#fn_name#line#threadid#text                  */
  /*                                                                         */
  /* # represents a level based separator - * for error, ! for alert and     */
  /* space otherwise                                                         */
  /*                                                                         */
  /* Column widths are controlled through the TRC_XXX_FMT macros, defined in */
  /* trcdef.h                                                                */
  /***************************************************************************/
  if (level & TRC_TYPE_ERROR)
  {
    sep = '*';
  }
  else if (level & TRC_TYPE_ALERT)
  {
    sep = '!';
  }
  else
  {
    sep = ' ';
  }

  fprintf(tfp, TRC_TIMESTAMP_FMT "%c" \
               TRC_THREAD_ID_FMT "%c" \
               TRC_FILE_NAME_FMT "%c" \
               TRC_FUNC_NAME_FMT "%c" \
               TRC_LINE_NUMB_FMT "%c" \
               "%s\n",
               trc_get_timestamp(), sep,
               (unsigned int)pthread_self(), sep,
               file_name, sep,
               fn_name, sep,
               line, sep,
               trace_buffer);

  /***************************************************************************/
  /* Flush the file and release the trace buffer lock                        */
  /***************************************************************************/
  fflush(tfp);
  pthread_mutex_unlock(&trc_buffer_mutex);

EXIT_LABEL:

  return;

} /* trc_trace */

/**PROC+**********************************************************************/
/* Name:      trc_trace_data                                                 */
/*                                                                           */
/* Purpose:   writes a data buffer to the trace file as hex                  */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     text         - descriptive text supplied with the data  */
/*            IN     data         - the data to trace                        */
/*            IN     data_len     - length of data (remember this is binary  */
/*            IN                    data and won't be NULL-terminated)       */
/*            IN     line         - line number of trace statement           */
/*            IN     fn_name      - function name                            */
/*            IN     file_name    - file name                                */
/*                                                                           */
/**PROC-**********************************************************************/

void trc_trace_data(char *text,
                    const char *data,
                    int data_len,
                    int line,
                    char *fn_name,
                    char *file_name)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int ii;
  char *pinto_data;

  /***************************************************************************/
  /* Check we've got a file!                                                 */
  /***************************************************************************/
  if (tfp == NULL)
  {
    goto EXIT_LABEL;
  }

  /***************************************************************************/
  /* get the trace buffer lock                                               */
  /***************************************************************************/
  trc_get_trace_mutex_lock();

  /***************************************************************************/
  /* Format is hh:mm:ss filename fn_name line threadid text                  */
  /*                                                                         */
  /* Column widths are controlled through the TRC_XXX_FMT macros, defined in */
  /* trcdef.h                                                                */
  /***************************************************************************/
  fprintf(tfp, TRC_TIMESTAMP_FMT " " \
               TRC_THREAD_ID_FMT " " \
               TRC_FILE_NAME_FMT " " \
               TRC_FUNC_NAME_FMT " " \
               TRC_LINE_NUMB_FMT " " \
               "%s\n",
               trc_get_timestamp(),
               (unsigned int)pthread_self(),
               file_name,
               fn_name,
               line,
               text);

  /***************************************************************************/
  /* data_len zero means that the data is null-terminated                    */
  /***************************************************************************/
  if (data_len == 0)
  {
    data_len = strlen(data);
  }

  /***************************************************************************/
  /* Limit the amount of data we trace                                       */
  /***************************************************************************/
  if (data_len > TRC_DATA_LIMIT)
  {
    fprintf(tfp, "(Truncated data len %d at %d)\n", data_len, TRC_DATA_LIMIT);
    data_len = TRC_DATA_LIMIT;
  }

  /***************************************************************************/
  /* Now write the data, a whole line at at a time                           */
  /***************************************************************************/
  pinto_data = (char *)data;
  for ( ii = 0; (ii + 15) < data_len; ii += 16)
  {
    trc_write_line(pinto_data, 16, ii);
    pinto_data += 16;
  }

  /***************************************************************************/
  /* any partial lines?                                                      */
  /***************************************************************************/
  if ((data_len % 16) > 0)
  {
    trc_write_line(pinto_data, (data_len % 16), ii);
  }

  /***************************************************************************/
  /* Flush the trace file and release the trace buffer lock                  */
  /***************************************************************************/
  fflush(tfp);
  pthread_mutex_unlock(&trc_buffer_mutex);

EXIT_LABEL:

  return;

} /* trc_trace_data */

/**PROC+**********************************************************************/
/* Name:      trc_get_trace_mutex_lock                                       */
/*                                                                           */
/* Purpose:   Get the trace mutex                                            */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/**PROC-**********************************************************************/

void trc_get_trace_mutex_lock(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/

  /***************************************************************************/
  /* get the trace buffer lock                                               */
  /***************************************************************************/
  if (pthread_mutex_lock(&trc_buffer_mutex))
  {
    /*************************************************************************/
    /* if we failed, there's not a lot we can do beyond set trace types to   */
    /* 0 to stop traceing                                                    */
    /*************************************************************************/
    trc_trace_types = 0;
  }

EXIT_LABEL:

  return;

} /* trc_get_trace_mutex_lock */

/**PROC+**********************************************************************/
/* Name:      trc_get_timestamp                                              */
/*                                                                           */
/* Purpose:   generate a timestamp                                           */
/*                                                                           */
/* Returns:   pointer to a global string                                     */
/*                                                                           */
/* Params:    none                                                           */
/*                                                                           */
/* Operation: Use system time functions to get the current time.             */
/*                                                                           */
/**PROC-**********************************************************************/

char *trc_get_timestamp(void)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  struct timeb curr_time;
  struct tm local_time;
  char millisecs[5];

  /***************************************************************************/
  /* Get the time since 1 Jan 1970                                           */
  /***************************************************************************/
  ftime(&curr_time);

  /***************************************************************************/
  /* Convert it to something more readable                                   */
  /***************************************************************************/
  localtime_r(&curr_time.time, &local_time);
  strftime(time_string, TRC_FULL_TIME_LEN - 4, "%H:%M:%S", &local_time);

  /***************************************************************************/
  /* append the milliseconds that strftime doesn't give us                   */
  /***************************************************************************/
  sprintf(millisecs, ".%3.3d", curr_time.millitm);
  strcat(time_string, millisecs);

  /***************************************************************************/
  /* return the global time string - since the caller holds the trace lock,  */
  /* there's no problem with re-use of this string                           */
  /***************************************************************************/
  return(time_string);

} /* trc_get_timestamp */

/**PROC+**********************************************************************/
/* Name:      trc_write_line                                                 */
/*                                                                           */
/* Purpose:   write a line of data                                           */
/*                                                                           */
/* Returns:   nothing                                                        */
/*                                                                           */
/* Params:    IN     data         - the line of data to trace                */
/*            IN     length       - length of data                           */
/*            IN     offset       - offset of line in source data            */
/*                                                                           */
/* Operation: Write a line of data as "offset hex ascii".  Output layout is: */
/*            0010  61626364 65666768 696A6B6C 6D6E6F70 abcd efgh ijkl mnop  */
/*                                                                           */
/**PROC-**********************************************************************/

void trc_write_line(char *data, int length, int offset)
{
  /***************************************************************************/
  /* Local Variables                                                         */
  /***************************************************************************/
  int ii;
  char *hex_part;
  char *ascii_part;

  /***************************************************************************/
  /* Start with the offset of the line                                       */
  /***************************************************************************/
  sprintf(trace_buffer, "  %04x  ", offset);

  /***************************************************************************/
  /* Fill the hex part of the line with spaces                               */
  /***************************************************************************/
  hex_part = trace_buffer + 8;
  memset((void *)hex_part, ' ', 36);

  /***************************************************************************/
  /* Write the hex part using a lookup table indexed on hi and lo halves of  */
  /* each byte, moving on a space every 4th byte                             */
  /***************************************************************************/
  for (ii = 0; ii < length; ii++)
  {
    *hex_part = hex_lookup[ ((data[ii] >> 4) & 0x0f)];
    *(hex_part + 1) = hex_lookup[  (data[ii] & 0x0f)];
    hex_part += 2;
    if (0 == ((ii + 1) % 4))
    {
      hex_part++;
    }
  }

  /***************************************************************************/
  /* Now write the data as ascii                                             */
  /***************************************************************************/
  ascii_part = trace_buffer + 8 + 36;
  memset((void *)ascii_part, ' ', 20);
  memcpy((void *)ascii_part, data, length);

  /***************************************************************************/
  /* Convert any non-printing characters to '.'                              */
  /***************************************************************************/
  for (ii = 0; ii < length; ii++)
  {
    if ((ascii_part[ii] < 0x20) || (ascii_part[ii] > 0x7E))
    {
      ascii_part[ii] = '.';
    }
  }

  /***************************************************************************/
  /* Finally append a new line and write the data to file.  Note that we     */
  /* printf via %s, so that any % characters in the data will not be         */
  /* interpreted as formatting characters.                                   */
  /***************************************************************************/
  ascii_part[20] = 0;
  strcat(trace_buffer, "\n");
  fprintf(tfp, "%s", trace_buffer);

  return;

} /* trc_write_line */

